#!/usr/bin/env python
# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --

import os
from setuptools import setup, find_packages
import setuptools

import sys, platform, configparser, subprocess
import numpy as np
from glob import glob
from setuptools import Extension
from setuptools.command.build_ext import build_ext
from distutils.command.install_headers import install_headers
import distutils.log


class EnvInterpolation(configparser.BasicInterpolation):
    """Interpolation which expands environment variables in values."""

    def before_get(self, parser, section, option, value, defaults):
        return os.path.expandvars(value)


# Global specifications
# ---------------------

__version__ = "1.0.0"
__name__ = "pybest"


# Utility functions
# -----------------


def get_sources(dirname, module):
    """Get all cpp files and the binding file of a package"""
    # avoid accidental inclusion of in-place build files and inc files
    if module == "basis":
        result = [
            dirname + "/py_basis.cpp",
            dirname + "/basis.cpp",
            dirname + "/libint_utils.cpp",
        ]
    elif module == "ints":
        result = [
            dirname + "/py_ints.cpp",
            dirname + "/external.cpp",
            dirname + "/overlap.cpp",
            dirname + "/kin.cpp",
            dirname + "/nuclear.cpp",
            dirname + "/pvp.cpp",
            dirname + "/emultipole.cpp",
            dirname + "/eri.cpp",
            dirname + "/cholesky_eri.cpp",
            dirname + "/libint_utils.cpp",
            dirname + "/basis.cpp",
        ]
    elif module == "slice":
        result = [dirname + "/py_slice.cpp", dirname + "/slicing.cpp"]
    elif module == "contract":
        result = [dirname + "/py_contract.cpp", dirname + "/contraction.cpp"]
    elif module == "transform":
        result = [dirname + "/py_transform.cpp", dirname + "/fourindex.cpp"]
    elif module == "sapt":
        result = [dirname + "/py_sapt_utils.cpp", dirname + "/sapt_utils.cpp"]

    return result


def get_sources_fortran(dirname):
    """Get all f files of a package"""
    # avoid accidental inclusion of in-place build files and inc files
    result = [fn for fn in glob("%s/*.f" % dirname)]
    return result


def get_headers():
    """Get all header-like files that need to be installed"""
    result = []
    for dn in ["src/pybest/"] + glob("src/pybest/*/"):
        result.extend(glob("%s/*.h" % dn))
    return result


headers = get_headers()


# Library configuration functions
# -------------------------------

lib_config_keys = [
    "include_dirs",
    "library_dirs",
    "libraries",
    "extra_objects",
    "extra_compile_args",
    "extra_link_args",
]


def print_lib_config(heading, lib_config):
    """Print (partial) lib_config"""
    print("   %s" % heading)
    if len(lib_config) == 0:
        print("      -")
    else:
        for key, value in sorted(lib_config.items()):
            if len(value) > 0:
                print("      %s: %s" % (key, value))


def get_lib_config_setup(prefix, fn_setup_cfg):
    """Get library configuration from a setup.cfg"""
    lib_config = {}
    if os.path.isfile(fn_setup_cfg):
        # config = configparser.ConfigParser()
        config = configparser.ConfigParser(interpolation=EnvInterpolation())
        config.read(fn_setup_cfg)
        if config.has_section(prefix):
            for key in lib_config_keys:
                if config.has_option(prefix, key):
                    value = config.get(prefix, key).strip()
                    if value is not None and len(value) > 0:
                        lib_config[key] = value.split(":")
        print_lib_config("From %s" % fn_setup_cfg, lib_config)
    else:
        print("   File %s not found. Skipping." % fn_setup_cfg)
    return lib_config


def get_lib_config_env(prefix):
    """Read library config from the environment variables"""
    lib_config = {}
    for key in lib_config_keys:
        varname = ("%s_%s" % (prefix, key)).upper()
        value = os.getenv(varname)
        if value is not None:
            lib_config[key] = value.split(":")
    print_lib_config("From environment variables", lib_config)
    return lib_config


class PkgConfigError(Exception):
    pass


def run_pkg_config(libname, option):
    """Safely try to call pkg-config"""
    try:
        return subprocess.check_output(
            ["pkg-config", libname, "--" + option], stderr=subprocess.STDOUT
        )
    except subprocess.CalledProcessError:
        raise PkgConfigError("pkg-config did not exit properly")
    except OSError:
        raise PkgConfigError("pkg-config not installed")


def get_lib_config_pkg(libname):
    """Get library config from the pkg-config program"""
    lib_config = {
        "include_dirs": [
            word[2:]
            for word in run_pkg_config(libname, "cflags-only-I").split()
        ],
        "library_dirs": [
            word[2:] for word in run_pkg_config(libname, "libs-only-L").split()
        ],
        "libraries": [
            word[2:] for word in run_pkg_config(libname, "libs-only-l").split()
        ],
        "extra_compile_args": run_pkg_config(
            libname, "cflags-only-other"
        ).split(),
        "extra_link_args": run_pkg_config(libname, "libs-only-other").split(),
    }
    print_lib_config("From pkg-config", lib_config)
    return lib_config


def all_empty(lib_config):
    """Test if all lib_config fields are empty"""
    return all(len(value) == 0 for value in lib_config.values())


def all_exist(lib_config):
    """Test if all paths in the lib_config exist"""
    # not all keys are paths
    PATH_CHECK_SKIP = ["extra_link_args", "extra_compile_args"]

    for key, value in lib_config.items():
        if key not in PATH_CHECK_SKIP:
            for path in value:
                if not os.path.exists(path):
                    return False
    return True


def detect_machine():
    """Return a description of the machine name, used for setup_cfgs/..."""
    if sys.platform == "linux":
        try:
            dist = platform.linux_distribution()
        except:
            import distro

            dist = distro.linux_distribution()
        return (
            "Linux-%s-%s-%s" % (dist[0], dist[1], platform.machine())
        ).replace(" ", "_")
    elif sys.platform == "darwin":
        mac_ver = platform.mac_ver()
        return "Darwin-%s-%s" % (mac_ver[0], mac_ver[2])
    else:
        return "unknown"


def lib_config_magic(prefix, libname, static_config={}):
    """Detect the configuration of a given library

       **Arguments:**

       prefix
            The prefix for this library. This is a name that PyBEST uses to
            refer to the library.

       libname
            The library name as it is known to the compiler and to pkg-config.
            For example, if the shared object is libfoo.so, then the library
            name is foo.

       **Optional arguments**

       static_config
            If given, this static library configuration is attempted. Ignored
            when empty, or when it contains non-existing files.
    """
    print(f"{prefix.upper()} Configuration")

    # Start with empty dictionary
    lib_config = dict((key, []) for key in lib_config_keys)

    # First try: Update with info from setup.cfg
    lib_config.update(get_lib_config_setup(prefix, "setup.cfg"))

    # Second try: Override with environment variables
    lib_config.update(get_lib_config_env(prefix))

    # If no environment variables were set, fianl attempt to guess the right settings.
    if all_empty(lib_config):
        if all_exist(static_config) and not all_empty(static_config):
            # If the static build is present, use it.
            lib_config.update(static_config)
        else:
            try:
                # Try to get dynamic link info from pkg-config
                lib_config.update(get_lib_config_pkg(libname))
            except PkgConfigError:
                # Try to find a machine-specific config file
                machine = detect_machine()
                fn_setup_cfg = f"setup_cfgs/setup.{machine}.cfg"
                lib_config.update(get_lib_config_setup(prefix, fn_setup_cfg))
                if all_empty(lib_config):
                    raise PkgConfigError(
                        f"No specifications found for {prefix}. Aborting."
                    )
    #                   # Uber-dumb fall back
    #                   lib_config['libraries'] = [libname]

    print_lib_config("Final", lib_config)
    return lib_config


# Print the Machine name on screen
# --------------------------------

print("MACHINE=%s" % detect_machine())
machine_info = detect_machine()


# Configuration of pybind
# -----------------------
def get_pybind_include(user=False):
    """Helper class to determine the pybind11 include path

    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    import pybind11

    return [pybind11.get_include(user)]


# Configuration of LibInt2
# ------------------------

libint2_dir = "depends/libint-upstream"
libint2_static_config = {
    "extra_objects": ["{}/installdir/lib/libint2.a".format(libint2_dir)],
    "include_dirs": ["{}/installdir/include".format(libint2_dir)],
}
libint2_config = lib_config_magic("libint2", "int2", libint2_static_config)


# Configuration of BLAS
# ---------------------

# First try to get the BLAS Configuration from environment variables.
blas_config = lib_config_magic("blas", "atlas")

# Detect which BLAS implementation is used and set a corresponding preprocessor
# flag.
blas_names = blas_config["libraries"] + blas_config["extra_objects"]
if any(("mkl" in l) for l in blas_names):
    blas_precompiler = ("BLAS_MKL", "1")
elif any(("atlas" in l) for l in blas_names):
    blas_precompiler = ("BLAS_ATLAS", "1")
elif any(("openblas" in l) for l in blas_names):
    blas_precompiler = ("BLAS_OPENBLAS", "1")
else:
    print(
        "   Unknown BLAS implementation. Assuming Netlib-compatible headers."
    )
    blas_precompiler = ("BLAS_OTHER", "1")
print("BLAS precompiler directive: -D%s" % blas_precompiler[0])


# Configuration of BOOST
# ----------------------

boost_config = lib_config_magic("boost", "boost")

# Configuration of EIGEN3
# ----------------------
eigen_config = lib_config_magic("eigen", "eigen")

# Configuration of HighFive
# -------------------------

highfive_config = lib_config_magic("highfive", "highfive")


# Configuration of HDF5 required for HighFive
# -------------------------------------------

hdf5_config = lib_config_magic("hdf5", "hdf5")


# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile

    with tempfile.NamedTemporaryFile("w", suffix=".cpp") as f:
        f.write("int main (int argc, char **argv) { return 0; }")
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11] compiler flag.
    """
    flags = ["-std=c++11"]

    for flag in flags:
        if has_flag(compiler, flag):
            return flag

    raise RuntimeError(
        "Unsupported compiler -- at least C++11 support is needed!"
    )


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""

    # Defaults, used in all packages:
    c_opts = {
        "unix": blas_config["extra_compile_args"],
        "intel": blas_config["extra_compile_args"],
    }
    l_opts = {
        "unix": blas_config["extra_link_args"],
        "intel": blas_config["extra_link_args"],
    }
    i_dirs = {
        "unix": [np.get_include(), "."]
        + get_pybind_include()
        + get_pybind_include(user=True)
        + blas_config["include_dirs"],
        "intel": [np.get_include(), "."]
        + get_pybind_include()
        + get_pybind_include(user=True)
        + blas_config["include_dirs"],
    }
    l_dirs = {
        "unix": blas_config["library_dirs"],
        "intel": blas_config["library_dirs"],
    }
    lib = {
        "unix": blas_config["libraries"],
        "intel": blas_config["libraries"],
    }
    e_obj = {
        "unix": blas_config["extra_objects"],
        "intel": blas_config["extra_objects"],
    }
    macros = {
        "unix": [blas_precompiler],
        "intel": [blas_precompiler],
    }

    build_ext.user_options += [
        # The format is (long option, short option, description)
        ("enable-cholesky=", None, "enables CD-ERI via libchol")
    ]

    def check_dependecies(self):
        # Configuration of LibChol is conditional
        if self.enable_cholesky:
            libchol_dir = "depends/libchol-upstream"
            libchol_static_config = {
                "extra_objects": [f"{libchol_dir}/installdir/lib/libchol.a"],
                "include_dirs": [f"{libchol_dir}/installdir/include"],
            }
            self.libchol_config = lib_config_magic(
                "libchol", "chol", libchol_static_config
            )

    def initialize_options(self):
        """Set default values for options."""
        super().initialize_options()
        # Each user option must be listed here with their default value.
        self.enable_cholesky = False
        self.enable_pvp = False

    def finalize_options(self):
        """Post-process options."""
        super().finalize_options()
        self.announce(
            "CD-ERI ENABLED: {}".format(self.enable_cholesky),
            level=distutils.log.INFO,
        )

    if sys.platform == "darwin":
        darwin_opts = ["-stdlib=libc++", "-mmacosx-version-min=10.8"]
        c_opts["unix"] += darwin_opts
        l_opts["unix"] += darwin_opts

    def build_extensions(self):
        # grab optional dependecies
        self.check_dependecies()

        ct = self.compiler.compiler_type
        com_opts = self.c_opts.get(ct, [])
        link_opts = self.l_opts.get(ct, [])
        incl_dirs = self.i_dirs.get(ct, [])
        lib_dirs = self.l_dirs.get(ct, [])
        lib_opts = self.lib.get(ct, [])
        extra_opts = self.e_obj.get(ct, [])
        macros_opts = self.macros.get(ct, [])
        if ct in ["unix", "intel"]:
            com_opts.append(
                '-DVERSION_INFO="%s"' % self.distribution.get_version()
            )
            com_opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, "-fvisibility=hidden"):
                com_opts.append("-fvisibility=hidden")
        for ext in self.extensions:
            ext.extra_compile_args += com_opts
            ext.extra_link_args += link_opts
            ext.include_dirs += incl_dirs
            ext.library_dirs += lib_dirs
            ext.libraries += lib_opts
            ext.extra_objects += extra_opts
            ext.define_macros += macros_opts
            if "gbasis" in ext.name:
                # link against libint2
                ext.extra_compile_args += libint2_config["extra_compile_args"]
                ext.extra_link_args += libint2_config["extra_link_args"]
                ext.include_dirs += (
                    libint2_config["include_dirs"]
                    + boost_config["include_dirs"]
                    + eigen_config["include_dirs"]
                    + highfive_config["include_dirs"]
                    + hdf5_config["include_dirs"]
                )
                ext.library_dirs += (
                    libint2_config["library_dirs"]
                    + hdf5_config["library_dirs"]
                )
                ext.libraries += (
                    libint2_config["libraries"] + hdf5_config["libraries"]
                )
                ext.extra_objects += libint2_config["extra_objects"]
                # compile pVp ints
                if self.enable_pvp:
                    ext.define_macros += [("PYBEST_ENABLE_PVP", "1")]
                else:
                    # 1) delete source files from sources
                    ext.sources = [
                        item for item in ext.sources if "pvp.cpp" not in item
                    ]

                # link against libchol
                if self.enable_cholesky:
                    # do link against libchol:
                    # 1) set macro avaible during pybind11 compilation (enabling cholesky interface)
                    ext.define_macros += [("PYBEST_ENABLE_CHOLESKY", "1")]
                    ext.extra_compile_args += self.libchol_config[
                        "extra_compile_args"
                    ]
                    ext.extra_link_args += self.libchol_config[
                        "extra_link_args"
                    ]
                    ext.include_dirs += self.libchol_config["include_dirs"]
                    ext.library_dirs += self.libchol_config["library_dirs"]
                    ext.libraries += self.libchol_config["libraries"]
                    ext.extra_objects += self.libchol_config["extra_objects"]
                else:
                    # do not link against libchol:
                    # 1) delete source files from sources
                    ext.sources = [
                        item
                        for item in ext.sources
                        if "cholesky_eri.cpp" not in item
                    ]

        build_ext.build_extensions(self)


class InstallHeaders(install_headers):
    """Use custom header installer because the default one flattens subdirectories"""

    def run(self):
        if not self.distribution.headers:
            return

        for header in self.distribution.headers:
            subdir = os.path.dirname(os.path.relpath(header, "src/pybest"))
            install_dir = os.path.join(self.install_dir, subdir)
            self.mkpath(install_dir)

            (out, _) = self.copy_file(header, install_dir)
            self.outfiles.append(out)


# Call distutils setup
# --------------------

setup(
    name=__name__,
    version=__version__,
    description="PyBEST: Pythonic Black-box Electronic Structure Tool.",
    author="Katharina Boguslawski <k.boguslawski@fizyka.umk.pl>, \
            Dariusz Kedziera <teodar@chem.umk.pl>, \
            Pawel Tecmer <ptecmer@fizyka.umk.pl>, \
            Piotr Zuchowski <pzuch@fizyka.umk.pl>, \
            Filip Brzek <filip.brzek@gmail.com>, \
            Aleksandra Leszczyk <aleksandra.leszczyk@gmail.com>, \
            Artur Nowak <nowak.art90@gmail.com>",
    # url='',
    scripts=glob("scripts/*.py"),
    package_dir={"": "src"},
    packages=find_packages("src"),
    include_package_data=True,
    package_data={
        "": [
            "data/basis/*.g94",
            "data/elements.csv",
            "data/*.bib",
            "data/examples/*/*.py",
            "data/test/*.*",
        ],
    },
    headers=headers,
    cmdclass={"install_headers": InstallHeaders, "build_ext": BuildExt,},
    ext_modules=[
        Extension(
            "pybest.linalg.pycontract",
            sources=get_sources("src/pybest/linalg", "contract"),
            language="c++",
        ),
        Extension(
            "pybest.linalg.pyslice",
            sources=get_sources("src/pybest/linalg", "slice"),
            language="c++",
        ),
        Extension(
            "pybest.linalg.pytransform",
            sources=get_sources("src/pybest/linalg", "transform"),
            language="c++",
        ),
        Extension(
            "pybest.gbasis.pybasis",
            sources=get_sources("src/pybest/gbasis", "basis"),
            language="c++",
        ),
        Extension(
            "pybest.gbasis.pyints",
            sources=get_sources("src/pybest/gbasis", "ints"),
            language="c++",
        ),
        Extension(
            "pybest.sapt.pysapt",
            sources=get_sources("src/pybest/sapt", "sapt"),
            language="c++",
        ),
    ],
    setup_requires=["nose>=1.0", "pip==18.1"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: C++",
        "Topic :: Science/Engineering :: Molecular Science",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Chemistry",
    ],
)
