# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Coupled Cluster implementations

   Variables used in this module:
    :nocc:      number of (active) occupied orbitals in the principle configuration
    :nvirt:     number of (active) virtual orbitals in the principle configuration
    :ncore:     number of frozen core orbitals in the principle configuration
    :nbasis:    total number of basis functions
    :energy:    the energy correction, list that can contain different
                contributions
    :amplitudes:the optimized amplitudes, list that can contain different
                contributions ([doubles] or [singles, doubles])

    Indexing convention:
     :i,j,k,..: occupied orbitals of principle configuration
     :a,b,c,..: virtual orbitals of principle configuration
     :p,q,r,..: general indices (occupied, virtual)

    P^bc_jk performs a pair permutation, i.e., P^bc_jk o_(bcjk) = o_(cbkj)

    Abbreviations used (if not mentioned in doc-strings):
     :L_pqrs: 2<pq|rs>-<pq|sr>
     :g_pqrs: <pq|rs>
"""

import numpy as np
import warnings
import math
from scipy import optimize as opt
from scipy import linalg as linalg

from pybest.cache import Cache
from pybest.io.iodata import IOData
from pybest.featuredlists import OneBodyHamiltonian, TwoBodyHamiltonian
from pybest.exceptions import ArgumentError, RestartError, UnknownOption
from pybest.log import log, timer
from pybest.utils import (
    check_type,
    check_options,
    unmask_orb,
    unmask,
)
from pybest.io.checkpoint import CheckPoint
from pybest.orbital_utils import transform_integrals, split_core_active
from pybest.linalg.cholesky import CholeskyFourIndex, CholeskyLinalgFactory
from pybest.linalg.dense import DenseLinalgFactory
from pybest.pt.mp2 import RMP2
from pybest.pt.perturbation_utils import get_epsilon
from pybest.auxmat import get_fock_matrix
from pybest.linalg.base import OneIndex, TwoIndex, FourIndex, Orbital


__all__ = [
    "RHFCCS",
    "RpCCDCCS",
]


class RCC(object):
    """Restricted Coupled Cluster class

       Purpose:
       Optimize cluster amplitudes and determine energy correction to some
       reference wavefunction.

       Currently supported wavefunction models:
        * RHF
        * RpCCD

       Currently supported cluster operators:
        * Doubles with occupied-virtual separation (RLCCD) and choice of frozen
          core and frozen virtual orbitals
        * Singles and Doubles with occupied-virtual separation (RLCCSD);
          only for RpCCD
    """

    def __init__(self, lf, occ_model, ncore=None):
        """
           **Arguments:**

           lf
                A LinalgFactory instance.

           occ_model
                The occupation model.

           **Optional arguments:**

           ncore
                Number of core orbitals. If provided, the core orbitals up
                to index ncore are excluded from the cluster operator.
        """
        check_type("frozencore", ncore, int, type(None))
        self._lf = lf
        self._denself = DenseLinalgFactory(lf.default_nbasis)
        if ncore is None:
            self._nocc = occ_model.noccs[0]
            self._ncore = 0
        else:
            self._nocc = occ_model.noccs[0] - ncore
            self._ncore = ncore
        if self.nocc <= 0 or self.nocc > occ_model.noccs[0]:
            raise ConsistencyError(
                f"Impossible number of frozen core orbitals! You have chosen {icore} "
                f"frozen core orbitals out of {occ_model.noccs[0]} occupied orbitals"
            )
        self._nbasis = lf.default_nbasis - self.ncore
        self._nvirt = self.nbasis - self.nocc
        if (
            self.nvirt <= 0
            or self.nvirt > lf.default_nbasis - occ_model.noccs[0]
        ):
            raise ConsistencyError(
                f"Impossible number of virtual orbitals: {self.nvirt}!"
            )
        self._occ_model = occ_model
        self._cache = Cache()
        self._energy = {}
        self._amplitudes = []
        self._freeze = None
        self._tco = "td"
        self._e_ref = 0.0
        self._e_core = 0.0
        self._solver = "krylov"
        self._checkpoint = CheckPoint({})
        self._checkpoint_fn = f"checkpoint_{self.get_model()}.h5"

    @property
    def lf(self):
        """The linalg factory"""
        return self._lf

    @lf.setter
    def lf(self, new):
        self._lf = new

    @property
    def cache(self):
        """A Cache instance"""
        return self._cache

    @cache.setter
    def cache(self, new):
        self._cache = new

    @property
    def occ_model(self):
        """The linalg factory"""
        return self._occ_model

    @occ_model.setter
    def occ_model(self, new):
        self._occ_model = new

    @property
    def denself(self):
        """The dense linalg factory"""
        return self._denself

    @denself.setter
    def denself(self, new):
        self._denself = new

    @property
    def nbasis(self):
        """The number of basis functions"""
        return self._nbasis

    @nbasis.setter
    def nbasis(self, new):
        self._nbasis = new

    @property
    def nocc(self):
        """The number of occupied orbitals"""
        return self._nocc

    @nocc.setter
    def nocc(self, new):
        self._nocc = new

    @property
    def nvirt(self):
        """The number of virtual orbitals"""
        return self._nvirt

    @nvirt.setter
    def nvirt(self, new):
        self._nvirt = new

    @property
    def ncore(self):
        """The number of core orbitals"""
        return self._ncore

    @ncore.setter
    def ncore(self, new):
        self._ncore = new

    @property
    def energy(self):
        """The coupled cluster energy contributions"""
        return self._energy

    @energy.setter
    def energy(self, energy_dict):
        self._energy.update(energy_dict)
        for key in energy_dict:
            self._checkpoint.update(key, energy_dict[key])

    @energy.deleter
    def energy(self):
        self._energy = {}

    @property
    def singles(self):
        """Single excitations"""
        return self._get_singles()

    @property
    def amplitudes(self):
        """The coupled cluster amplitudes"""
        return self._amplitudes

    @property
    def freeze(self):
        """The freezed coupled cluster amplitudes"""
        return self._freeze

    @freeze.setter
    def freeze(self, args):
        self._freeze = args

    @property
    def multireference(self):
        """multireference wavefunction"""
        return True

    @property
    def solver(self):
        """The solver"""
        return self._solver

    @solver.setter
    def solver(self, new):
        self._solver = new

    @property
    def e_ref(self):
        """Reference energy"""
        return self._e_ref

    @e_ref.setter
    def e_ref(self, new):
        self._e_ref = new

    @property
    def e_core(self):
        """The core energy"""
        return self._e_core

    @e_core.setter
    def e_core(self, new):
        self._e_core = new

    @property
    def tco(self):
        """The tensor contraction operation used"""
        return self._tco

    @tco.setter
    def tco(self, args):
        if args is not None:
            self._tco = args

    @property
    def checkpoint(self):
        """The iodata container that contains all data dump to disk"""
        return self._checkpoint

    @checkpoint.setter
    def checkpoint(self, new):
        self._checkpoint = new

    @property
    def checkpoint_fn(self):
        """The filename that will be dumped to disk"""
        return self._checkpoint_fn

    @checkpoint_fn.setter
    def checkpoint_fn(self, new):
        self._checkpoint_fn = f"checkpoint_{self.get_model()}{new}.h5"

    def get_guess(self):
        raise NotImplementedError

    def clear(self):
        """Clear all wavefunction information"""
        self._cache.clear()

    def clear_aux_matrix(self):
        """Clear the auxiliary matrices"""
        self._cache.clear(tags="m", dealloc=True)

    def calculate_aux_matrix(self, mo1, mo2, args, **kwargs):
        """Compute auxiliary matrices

           **Arguments:**

           mo1, mo2
                One- and two-electron integrals (some Hamiltonian matrix
                elements) in the MO basis.

           args
                List of arguements. Only geminal coefficients are used.
        """
        self.clear_aux_matrix()
        return self.update_aux_matrix(mo1, mo2, args, **kwargs)

    def get_aux_matrix(self, select):
        """Get a matrix.

           **Arguments:**

           select
                'fock'.
        """
        if not select in self._cache:
            raise NotImplementedError
        return self.cache.load(select)

    def check_result(self, **kwargs):
        raise NotImplementedError

    def fix_guess(self, guess):
        raise NotImplementedError

    def check_input(self, **kwargs):
        """Check input parameters."""
        raise NotImplementedError

    def get_range(self, string, start=0):
        """ Returns dictionary with keys beginX, endX, begin(X+1), etc.
            *  string - any sequence of 'o' (occupied) and 'v' (virtual)
        """
        range_ = {}
        ind = start
        for char in string:
            if char == "o":
                range_["begin" + str(ind)] = 0
                range_["end" + str(ind)] = self.nocc
            elif char == "v":
                range_["begin" + str(ind)] = self.nocc
                range_["end" + str(ind)] = self.nbasis
            elif char == "V":
                range_["begin" + str(ind)] = 0
                range_["end" + str(ind)] = self.nvirt
            elif char == "n":
                range_["begin" + str(ind)] = 0
                range_["end" + str(ind)] = self.nbasis
            else:
                raise UnknownOption(f"Do not know how to handle choice {char}")
            ind += 1
        return range_

    @timer.with_section("CC")
    def __call__(self, *args, **kwargs):
        """Optimize cluster amplitudes and determine energy correction to some
           reference wavefunction.

           Cluster amplitudes are determined by finding the root of a set of
           equations of the form A.x+b = 0

           Currently supported wavefunction models (Psi_0):
            * RHF
            * RpCCD

           Currently supported cluster operators:
            * Singles with occupied-virtual separation (RCCS);

           **Arguments:**

           one, two
                One- and two-body integrals (some Hamiltonian matrix elements)
                expressed in the AO basis.

           args
                It contains the AO/MO coefficients (if Psi_0 = RHF) and the
                geminal coefficients (if Psi_0 = pCCD) as second arguement.

           **Keywords:**
                Contains reference energy and solver specific input parameters:
                 * e_ref:        reference energy (default value nan)
                 * threshold:   tolerance for amplitudes (default 1e-6)
                 * maxiter:     maximum number of iterations (default 200)
                 * guess:       initial guess for amplitudes, if not provided,
                                a random guess will be generated

        """
        #
        # Set default keyword arguements to class attributes:
        #
        self.e_ref = unmask("e_tot", *args, **kwargs) or unmask(
            "e_ref", *args, **kwargs
        )
        #
        # Core energy is NOT passed as kwargs as LCC uses e_ref.
        # e_core can only be passed from an IOData container of a previous calculation.
        # e_core is only used when defining an active space.
        #
        self.e_core = unmask("e_core", *args, **kwargs)
        self.tco = kwargs.get("tco", "td")
        self.freeze = kwargs.get("freeze", None)
        self.solver = kwargs.get("solver", "krylov")
        self.checkpoint_fn = kwargs.get("checkpoint_fn", "")
        self.jacobian_approximation = kwargs.get("jacobian", 1)
        #
        # Set remaining default keyword arguements:
        #
        tthresh = kwargs.get("tthreshold", 1e-6)
        ethresh = kwargs.get("ethreshold", 1e-7)
        maxiter = kwargs.get("maxiter", 200)
        initguess = kwargs.get("guess", None)
        skipmo = kwargs.get("skipmo", False)
        deletemo = kwargs.get("deletemo", True)
        guesstype = kwargs.get("guesstype", "mp2")
        diismax = kwargs.get("diismax", 2)
        diisstart = kwargs.get("diisstart", 0)
        indextrans = kwargs.get("indextrans", "tensordot")
        diisreset = kwargs.get("diisreset", True)
        restart = kwargs.get("restart", False)

        #
        # Clear everything
        #
        del self.amplitudes
        del self.energy
        if log.do_medium:
            log.hline("=")
            log(" ")
            log("Entering CC module")
            log(" ")

        #
        # Unmask arguments
        #
        fargs = []
        t_p = None
        # olp
        olp = unmask("olp", *args, **kwargs)
        self.checkpoint.update("olp", olp)
        # orb
        orbs = unmask_orb(*args, **kwargs)
        if orbs:
            orbs = orbs[0]
            self.checkpoint.update("orb_a", orbs.copy())
        # t_p
        if self.multireference:
            t_p = unmask("t_p", *args, **kwargs)
            if t_p is not None:
                self.checkpoint.update("t_p", t_p)
        # 1-e ints and 2-e ints
        one = self.lf.create_two_index(label="one")
        for arg in args:
            if isinstance(arg, TwoIndex):
                if arg.label in OneBodyHamiltonian:
                    one.iadd(arg)
            elif isinstance(arg, FourIndex):
                two = arg
                # Check if lf and two are of the same type
                if isinstance(two, CholeskyFourIndex):
                    if not isinstance(self.lf, CholeskyLinalgFactory):
                        raise ArgumentError(
                            "LinalgFactory and eri are of different type."
                        )
                elif isinstance(self.lf, CholeskyLinalgFactory):
                    if not isinstance(two, CholeskyFourIndex):
                        raise ArgumentError(
                            "LinalgFactory and eri are of different type."
                        )
        #
        # Print some initial information, method specific:
        #
        self.print_info(**kwargs)

        #
        # Check input parameters:
        #
        self.check_input(**kwargs)

        #
        # Get initial guess:
        #
        if restart:
            guess, olp, orbs, t_p = self.do_restart(restart, olp, orbs)
        elif initguess is not None:
            guess = initguess
        else:
            guess = self.get_guess(guesstype, one, two, orbs)

        #
        # Append to function arguments to work with old implementation
        #
        fargs.append(orbs)
        #
        # Check if Tp amplitudes are there if required
        #
        if self.multireference and t_p is None:
            raise ArgumentError(
                "Cannot find pCCD amplitudes in LCC function call."
            )
        else:
            fargs.append(t_p)
        if olp is None:
            raise ArgumentError(
                "Cannot find overlap integrals in LCC function call."
            )
        #
        # Transform integrals:
        #
        if skipmo:
            mo1 = one
            mo2 = two
        else:
            if self.ncore > 0:
                cas = split_core_active(
                    one,
                    two,
                    orbs,
                    e_core=self.e_core,
                    ncore=self.ncore,
                    nactive=self.nbasis,
                    indextrans=indextrans,
                )
                mo1 = cas.one
                mo2 = cas.two
                self.e_core = cas.e_core
            else:
                ti = transform_integrals(one, two, orbs, indextrans=indextrans)
                mo1 = ti.one[0]
                mo2 = ti.two[0]

        #
        # Construct auxilary matrices:
        #
        matrix = self.calculate_aux_matrix(mo1, mo2, t_p, **kwargs)

        #
        # Append arguements, used as arguments in root finding:
        #
        try:
            for mat in matrix:
                fargs.append(mat)
        except:
            pass

        # Update amplitudes and energies for first iteration:
        del self.amplitudes
        self.amplitudes = guess
        self.energy = self.calculate_energy()
        #
        # Solve matrix equataion using scipy.optimize.root routine:
        #
        if self.solver == "krylov":
            if log.do_medium:
                log(
                    f"{'residual 2-norm':>18} {'residual max norm':>20} "
                    f"{'Ecorr':>12} {'Etot':>18} {'D(E)':>14}"
                )
            amplitudes = opt.root(
                self.vfunction,
                guess,
                args=(fargs),
                method="krylov",
                callback=self.callback,
                options={"fatol": tthresh, "maxiter": maxiter},
            )
            if amplitudes.success:
                del self.amplitudes
                self.amplitudes = amplitudes.x
                if log.do_medium:
                    log(
                        f"Optimization of cluster amplitudes converged in "
                        f"{amplitudes.nit} iterations."
                    )
                    log(" ")
                    log("Calculating energy correction:")
            else:
                #
                # Update amplitudes anyway for restart possibility
                #
                del self.amplitudes
                self.amplitudes = amplitudes.x
                log.warn(
                    f"Error in solving amplitude equations: {amplitudes.message}"
                )
        elif self.solver == "pbqn":
            del self.amplitudes
            self.amplitudes = guess
            from pybest.solvers.pbquasinewton import PBQausiNewton

            qn = PBQausiNewton(
                self.lf,
                maxiter=maxiter,
                ethreshold=ethresh,
                tthreshold=tthresh,
            )
            e, t = qn(
                self,
                guess,
                fargs,
                diismax=diismax,
                diisstart=diisstart,
                suffix=self.checkpoint_fn,
                diisreset=diisreset,
            )
            del self.amplitudes
            self.amplitudes = t

        #
        # Get correction to reference energy:
        #
        self.energy = self.calculate_energy()

        #
        # Print some output information for user:
        #
        self.print_energy()

        #
        # Make sanity checks. If somethings is wrong, abort calculation:
        #
        self.check_result(**kwargs)

        #
        # Do Checkpoint
        #
        self.checkpoint.to_file(self.checkpoint_fn)

        #
        # Clear auxiliary matrices from Cache to save memory
        #
        self.clear_aux_matrix()

        if log.do_medium:
            log(" ")
            log.hline("=")
        return self.checkpoint()

    def callback(self, x, f):
        norm2 = linalg.norm(f)
        normmax = linalg.norm(f, ord=np.inf)
        #
        # Get current amplitudes
        #
        del self.amplitudes
        self.amplitudes = x
        #
        # Get previous energies
        #
        e_old = self.energy["e_tot"]
        #
        # Get current energy
        #
        self.energy = self.calculate_energy()
        if log.do_medium:
            e_corr = self.energy["e_corr"]
            e_tot = self.energy["e_tot"]
            log(
                f"{norm2:18.12f} {normmax:18.12f} {e_corr:18.12f} "
                f"{e_tot:18.12f} {(e_old - e_tot):12.3e}"
            )
        #
        # Dump current solution
        #
        self.checkpoint.to_file(self.checkpoint_fn)

    def init_aux_matrix(
        self, select, dim0=None, dim1=None, dim2=None, dim3=None
    ):
        """Initialize auxiliary matrices

           **Arguments:**

           select
                One of ``fock``, ``ocjbc``, ``vcjkb``, ``ocjkb``, ``vcjbc``,
                ``dcjb``, ``vcij``, ``ocab``, ``ocjb``, ``vcjb``, ``vcjkl``,
                ``ocabc``, vcjbk``, ``ocajb``

           **Optional arguments:**

           dim, dim2
                The dimension of the auxiliary matrices for specific axes.
        """
        check_options(
            "select",
            select,
            "fock",
            "vcij",
            "ocab",
            "ocjb",
            "vcjb",
            "govvo",
            "govov",
            "govvv",
            "gvooo",
        )
        dim0 = dim0 or self.nbasis
        dim1 = dim1 or self.nbasis
        dim2 = dim2 or self.nbasis
        dim3 = dim3 or self.nbasis

        if select in ["fock", "vcij", "ocab", "ocjb", "vcjb"]:
            alloc_size = (self.lf.create_two_index, dim0, dim1)
        elif select in ["govvo", "govov", "gvooo", "govvv"]:
            alloc_size = (
                self.denself.create_four_index,
                dim0,
                dim1,
                dim2,
                dim3,
            )

        matrix, new = self._cache.load(select, alloc=alloc_size, tags="m",)
        if not new:
            raise RuntimeError(
                f"The matrix {select} already exists. Call clear prior "
                f"to updating the wfn."
            )
        return matrix

    @timer.with_section("UpdatingAuxMat")
    def update_aux_matrix(self, mo1, mo2, cia, **kwargs):
        """Derive all auxiliary matrices.
           fock_pq:     one_pq + sum_m(2<pm|qm> - <pm|mq>),
           lint_pqrs:   <pq|rs> - <pq|sr>,
           vc_ij:       sum_d <ij|dd> c_j^d,
           oc_ab:       sum_m <ab|mm> c_m^a,
           oc_jb:       sum_m <jb|mm> c_m^b,
           vc_jb:       sum_d <jb|dd> c_j^d,

           **Arguments:**

           mo1, mo2
                one- and two-electron integrals to be sorted.

           cia
                The geminal coefficients. A TwoIndex instance
        """
        if log.do_medium:
            log("Computing auxiliary matrices.")
        deletemo = kwargs.get("deletemo", True)
        #
        # Get ranges for contract
        #
        oov = self.get_range("oov")
        vvo = self.get_range("vvo")
        ovo = self.get_range("ovo")
        ovv = self.get_range("ovv")
        ovvo = self.get_range("ovvo")
        ovov = self.get_range("ovov")
        #
        # Inactive Fock matrix
        #
        auxmat0 = self.init_aux_matrix("fock")
        get_fock_matrix(auxmat0, mo1, mo2, self.nocc)
        if deletemo:
            del mo1
        #
        # g_ovvo
        #
        auxmat1 = self.init_aux_matrix(
            "govvo", self.nocc, self.nvirt, self.nvirt, self.nocc
        )
        mo2.slice("abcd->abcd", auxmat1, **ovvo)
        #
        # g_ovov
        #
        auxmat2 = self.init_aux_matrix(
            "govov", self.nocc, self.nvirt, self.nocc, self.nvirt
        )
        mo2.slice("abcd->abcd", auxmat2, **ovov)
        #
        # pCCD reference function:
        #
        if self.multireference:
            # use 3-index intermediate (will be used several times)
            # This also works with Cholesky
            # tmpabc
            tmpabc = self.denself.create_three_index(self.nbasis)
            mo2.slice("abcc->abc", tmpabc, 1.0, False)
            #
            # vc_ij = sum_d <ij|dd> c_j^d
            #
            auxmat3 = self.init_aux_matrix("vcij", self.nocc, self.nocc)
            tmpabc.contract("abc,bc->ab", cia, auxmat3, **oov)
            #
            # oc_ab = sum_m <ab|mm> c_m^a
            #
            auxmat4 = self.init_aux_matrix("ocab", self.nvirt, self.nvirt)
            tmpabc.contract("abc,ca->ab", cia, auxmat4, **vvo)
            #
            # oc_jb = sum_m <jb|mm> c_m^b
            #
            auxmat5 = self.init_aux_matrix("ocjb", self.nocc, self.nvirt)
            tmpabc.contract("abc,cb->ab", cia, auxmat5, **ovo)
            #
            # vc_jb = sum_d <jb|dd> c_j^d
            #
            auxmat6 = self.init_aux_matrix("vcjb", self.nocc, self.nvirt)
            tmpabc.contract("abc,ac->ab", cia, auxmat6, **ovv)
        #
        # for singles only
        #
        if self.singles:
            #
            # g_vooo
            #
            auxmat7 = self.init_aux_matrix(
                "gvooo", self.nvirt, self.nocc, self.nocc, self.nocc
            )
            mo2.slice(
                "abcd->abcd",
                auxmat7,
                1.0,
                True,
                begin0=self.nocc,
                end0=self.nbasis,
                end1=self.nocc,
                end2=self.nocc,
                end3=self.nocc,
            )
            #
            # g_ovvv
            #
            if not isinstance(self.lf, CholeskyLinalgFactory):
                auxmat8 = self.init_aux_matrix(
                    "govvv", self.nocc, self.nvirt, self.nvirt, self.nvirt
                )
                mo2.slice(
                    "abcd->abcd",
                    auxmat8,
                    1.0,
                    True,
                    begin0=0,
                    end0=self.nocc,
                    begin1=self.nocc,
                    begin2=self.nocc,
                    begin3=self.nocc,
                )
                if deletemo:
                    del mo2
            else:
                return [mo2]

    @timer.with_section("VecFctLCC")
    def vfunction(self, amplitudes, args):
        """Vector function to find coupled cluster doubles amplitudes.

           **Arguments:**

           amplitudes
                Cluster amplitudes. Need to be determined.

           args
                All function arguments needed to calculated the vector
                function:
                    * [0]:  wfn expansion coefficients
                    * [1]:  geminal coefficients
        """
        #
        # Get ranges
        #
        oov = self.get_range("oov")
        vvo = self.get_range("vvo")
        ooo = self.get_range("ooo")
        ovv = self.get_range("ovv")
        ovvo = self.get_range("ovvo")
        ovov = self.get_range("ovov")
        oo2 = self.get_range("oo", start=2)
        vv2 = self.get_range("vv", start=2)
        vo2 = self.get_range("vo", start=2)
        ov2 = self.get_range("ov", start=2)
        if isinstance(self.lf, CholeskyLinalgFactory):
            ovvv = self.get_range("ovvv")
        else:
            ovvv = self.get_range("oVVV")
        #
        # Get aux matrices
        #
        fock = self.get_aux_matrix("fock")
        if self.multireference:
            vcij = self.get_aux_matrix("vcij")
            ocab = self.get_aux_matrix("ocab")
            ocjb = self.get_aux_matrix("ocjb")
            vcjb = self.get_aux_matrix("vcjb")
        govvo = self.get_aux_matrix("govvo")
        govov = self.get_aux_matrix("govov")
        if self.singles:
            gvooo = self.get_aux_matrix("gvooo")
        try:
            govvv = self.get_aux_matrix("govvv")
        except:
            govvv = args[-1]
        #
        # output
        #
        end = 0
        # singles
        if self.singles:
            singles = self.lf.create_two_index(self.nocc, self.nvirt)
            tjb = self.lf.create_two_index(self.nocc, self.nvirt)
            end = self.nocc * self.nvirt
            if type(amplitudes) == np.ndarray:
                tjb.assign(amplitudes[:end])
            else:
                tjb.assign(amplitudes._array[:end])
        #
        # Get active orbital indices:
        # iocc: final active occupied orbital index
        # iend: final active virtual orbital index
        #
        iocc = self.nocc
        iend = self.nbasis
        #
        # temporary storage
        #
        tmp2ind = self.lf.create_two_index(self.nvirt, self.nvirt)
        tmp2ind1 = self.lf.create_two_index(self.nocc, self.nocc)
        tmp2ind2 = self.lf.create_two_index(self.nocc, self.nvirt)

        if self.singles:
            # sum_ai A^{jb}_{ia}*t_i^a
            #
            # t_kc <ic||ak>
            #
            govvo.contract("abcd,db->ac", tjb, singles, factor=2.0, clear=True)
            govov.contract("abcd,cb->ad", tjb, singles, factor=-1.0)
            #
            # Fca,ic ; Fki,ka
            #
            tjb.contract("ab,bc->ac", fock, singles, **vv2)
            tjb.contract("ab,ac->cb", fock, singles, factor=-1.0, **oo2)
            #
            # Fia
            #
            singles.iadd(fock, 1.0, **ov2)
            #
            # quadratic terms
            #
            # tic Fck tka = fik tka
            tjb.contract("ab,bc->ac", fock, tmp2ind1, clear=True, **vo2)
            tmp2ind1.contract("ab,bc->ac", tjb, singles, factor=-1.0)
            #
            # <ci||lk> tlc tka = gik tka
            #
            gvooo.contract(
                "abcd,ca->bd", tjb, tmp2ind1, factor=2.0, clear=True
            )
            gvooo.contract("abcd,da->bc", tjb, tmp2ind1, factor=-1.0)
            tmp2ind1.contract("ab,bc->ac", tjb, singles, factor=-1.0)
            #
            # <ka||cd> tkc tid = tid gad
            #
            govvv.contract(
                "abcd,ac->bd", tjb, tmp2ind, factor=2.0, clear=True, **ovvv
            )
            govvv.contract("abcd,ad->bc", tjb, tmp2ind, factor=-1.0, **ovvv)
            tjb.contract("ab,cb->ac", tmp2ind, singles)
            #
            # cubic terms
            #
            # <kl||cd> tld tic tka = tic Lkc tka = tLik tka
            govvo.contract(
                "abcd,db->ac", tjb, tmp2ind2, factor=2.0, clear=True
            )
            govvo.contract("abcd,dc->ab", tjb, tmp2ind2, factor=-1.0)
            tjb.contract("ab,cb->ac", tmp2ind2, tmp2ind1, clear=True)
            tmp2ind1.contract("ab,bc->ac", tjb, singles, factor=-1.0)
            #
            # pCCD reference function
            #
            if self.multireference:
                #
                # <ki||ca> c_ia t_kc
                #
                govvo.contract(
                    "abcd,db->ac", tjb, tmp2ind2, factor=2.0, clear=True
                )
                govvo.contract("abcd,dc->ab", tjb, tmp2ind2, factor=-1.0)
                tmp2ind2.imul(args[1])
                singles.iadd(tmp2ind2)
                #
                # <ac|ll> t_ic c_la = t_ic ocac
                #
                tjb.contract("ab,cb->ac", ocab, singles, factor=-1.0)
                #
                # <li|cc> t_la c_ic = t_la vcli
                #
                tjb.contract("ab,ac->cb", vcij, singles, factor=-1.0)
                #
                # Fia cia
                #
                singles.iadd_mult(args[1], fock, 1.0, **ov2)
                #
                # <ia|kk> cka = ocia
                #
                singles.iadd(ocjb, -1.0)
                #
                # <ia|cc> cic = vcia
                #
                singles.iadd(vcjb, 1.0, 0)
        #
        # Freeze amplitudes
        #
        if self.freeze is not None:
            for row in self.freeze:
                singles.set_element(row[0], row[1], 0.0, symmetry=1)

        # FIXME: don't access matrix class directly
        if self.solver == "krylov":
            return singles._array.ravel(order="C")
        else:
            out = self.lf.create_one_index(self.nunknowns)
            out.assign(singles._array.ravel(order="C"))
            return out

    @timer.with_section("JacobianCC")
    def jacobian(self, amplitudes, args):
        """Jacobian approximation to find coupled cluster doubles amplitudes.

           **Arguments:**

           amplitudes
                Cluster amplitudes.

           args
                All function arguments needed to calculated the vector
        """
        if log.do_medium:
            log("Computing Jacobian approximation for Quasi-Newton solver.")
        fock = self.get_aux_matrix("fock")
        fi = fock.copy_diagonal(end=self.nocc)
        fa = fock.copy_diagonal(begin=self.nocc)
        out = self.lf.create_one_index(self.nunknowns)
        if self.singles:
            # The function returns fi-fa and fi-fa+fj-fb
            fia, fiajb = get_epsilon(
                self.denself,
                [fi, fa],
                singles=self.singles,
                shift=[1e-12, 1e-12],
            )
            fia.iscale(-1.0)
            eps1 = fia._array.reshape(self.nocc * self.nvirt)
            out.assign(eps1, end0=self.nocc * self.nvirt)
        return out


class RCCS(RCC):
    @property
    def nunknowns(self):
        """Single excitations"""
        return self.nvirt * self.nocc

    def _get_singles(self):
        """Single excitations"""
        return True

    @property
    def amplitudes(self):
        """The coupled cluster amplitudes"""
        return self._amplitudes

    @amplitudes.setter
    def amplitudes(self, new):
        """Update singles and doubles cluster amplitudes in
           self._amplitudes[S,D]

           **Arguments:**

           new
                Singles and doubles amplitudes. A np.ndarray.
        """
        if isinstance(new, np.ndarray) or isinstance(new, OneIndex):
            if self._amplitudes:
                raise NonEmptyData(
                    "Warning: List of LCC amplitudes not empty!"
                )

            singles = self.lf.create_two_index(
                self.nocc, self.nvirt, label="t_1"
            )

            singles.assign(new, end2=(self.nocc * self.nvirt))
            self._amplitudes.append(singles)
            # for checkpointing
            self.checkpoint.update("t_1", singles)
        elif isinstance(new, list):
            self._amplitudes = new

    @amplitudes.deleter
    def amplitudes(self):
        del self._amplitudes[:]
        self._amplitudes = []

    def print_energy(self):
        if log.do_medium:
            log(f"E_ref:        {self.e_ref: 16.8f} a.u.")
            log.hline("-")
            log(f"E_corr(S):    {self.energy['e_corr']: 16.8f} a.u.")
            log.hline("-")
            log(f"E_tot:        {self.energy['e_tot']: 16.8f} a.u.")

    def get_guess(self, select="mp2", *args):
        """Generate inital guess for cluster amplitudes"""
        if select in ["mp2"]:
            if log.do_medium:
                log("Performing an MP2 guess.")
                log.hline("~")
            level = log.level
            log.level = 1
            mp2 = RMP2(self.lf, self.occ_model, ncore=self.ncore)
            mp2_ = mp2(
                args[0],
                args[1],
                args[2],
                IOData(e_core=self.e_core),
                singles=True,
                natorb=False,
                relaxation=False,
                e_ref=0.0,
            )
            mp2.clear()
            log.level = level
            if log.do_medium:
                log("Resuming CC calculation.")
                log.hline("~")
            return mp2_.t_1._array.ravel(order="C")
        elif select in ["random"]:
            singles = np.random.rand(self.nocc * self.nvirt) * 0.1
            return singles.ravel()
        else:
            raise ArgumentError(f"Unknown guess {select}")

    def do_restart(self, restart, olp, orbs):
        """Collect data for restart"""

        data = IOData.from_file(restart)
        t_1 = unmask("t_1", data)
        if t_1 is None:
            raise RestartError("Cannot find CC amplitudes in restart file.")
        else:
            guess = t_1._array.ravel()
        t_p = unmask("t_p", data)
        if t_p is not None:
            self.checkpoint.update("t_p", t_p)
        if hasattr(data, "olp") and hasattr(data, "orb_a"):
            if isinstance(olp, TwoIndex) and isinstance(orbs, Orbital):
                project_orbitals(data.olp, olp, data.orb_a, orbs)
            else:
                olp = data.olp
                orbs = data.orb_a
        self.checkpoint.update("olp", olp)
        self.checkpoint.update("orb_a", orbs.copy())
        if log.do_medium:
            log.warn(
                f"Restarting from orbitals read from file, while Hamiltonian is not"
                f" provided from restart file. Make sure that you know what you are"
                f" doing."
            )
        if hasattr(data, "e_ref") and hasattr(data, "e_core"):
            self.e_ref = data.e_ref
            self.e_core = data.e_core
        else:
            raise RestartError("Cannot find e_ref or e_core in restart file.")
        return guess, olp, orbs, t_p

    def calculate_energy(self, t=None):
        """Calculate energy contribution"""
        ov2 = self.get_range("ov", start=2)
        vv2 = self.get_range("vv", start=2)
        if t is not None:
            del self.amplitudes
            self.amplitudes = t
        #
        # Get active orbital indices
        #
        iocc = self.nocc
        iend = self.nbasis

        fock = self.get_aux_matrix("fock")
        govvo = self.get_aux_matrix("govvo")
        t1 = self.amplitudes[0]

        #
        # E_singles = sum_md F_md t_md
        #
        esingles = 2.0 * t1.contract("ab,ab", fock, **ov2)

        tmp = self.lf.create_two_index(self.nocc, self.nvirt)
        govvo.contract("abcd,ac->db", t1, tmp, factor=2.0, clear=True)
        govvo.contract("abcd,ab->dc", t1, tmp, factor=-1.0)
        esingles += t1.contract("ab,ab", tmp)

        return {
            "e_tot": esingles + self.e_ref,
            "e_corr": esingles,
            "e_corr_s": esingles,
            "e_ref": self.e_ref,
            "e_core": self.e_core,
        }

    def check_input(self, **kwargs):
        """Check input parameters."""
        for name, value in kwargs.items():
            check_options(
                name,
                name,
                "e_ref",
                "tthreshold",
                "maxiter",
                "guess",
                "skipmo",
                "deletemo",
                "freeze",
                "tco",
                "solver",
                "diisstart",
                "diismax",
                "indextrans",
                "diisreset",
                "guesstype",
                "ethreshold",
                "restart",
                "checkpoint_fn",
            )
        solver = kwargs.get("solver", "krylov")
        check_options("solver", solver, "krylov", "pbqn")
        guess = kwargs.get("guess", None)
        if guess is not None:
            check_type("guess", guess, np.ndarray)
            dim = self.nocc * self.nvirt
            if guess.size != dim:
                raise ValueError(
                    "Warning: Number of unknowns in initial guess is wrong!"
                )
        guesstype = kwargs.get("guesstype", "mp2")
        check_options("guesstype", guesstype, "mp2", "random")
        tco = kwargs.get("tco", "td")
        if tco not in ["td", "einsum"]:
            raise ArgumentError("Warning: tco can either be td or einsum!")
        indextrans = kwargs.get("indextrans", "tensordot")
        if indextrans not in ["einsum", "tensordot", "blas"]:
            raise ArgumentError("Choise of 4-index transformation unknown!")
        diisreset = kwargs.get("diisreset", True)
        check_options("diisreset", diisreset, True, False, 0, 1)


class RHFCCS(RCCS):
    def print_info(self, **kwargs):
        """Print optimization options.
        """
        thresh = kwargs.get("threshold", 1e-7)
        maxiter = kwargs.get("maxiter", 200)
        guesstype = kwargs.get("guesstype", "mp2")
        log.hline()
        log("Coupled Cluster Singles correction on top of RHF")
        log.hline()
        log("OPTIMIZATION PARAMETERS:")
        log("Cluster Operator:            T1")
        log(f"Number of occupied orbitals: {self.nocc}")
        log(f"Number of frozen cores:      {self.ncore}")
        log(f"Number of virtuals:          {self.nvirt}")
        log(f"Initial guess:               {guesstype}")
        log("Solvers:")
        log("  cluster amplitudes:        krylov")
        log("Optimization thresholds:")
        log(f"  cluster amplitudes:        {thresh:1.2e}")
        log(f"  maxiter:                   {maxiter}")
        log.hline()

    def fix_guess(self, guess):
        """Set diagonal of double amplitudes to Zero. This avoids any errors
           induced by user-specified guess

           **Arguements:**

           guess:
                The initial guess. A numpy ndarray
        """
        singles = np.reshape(
            guess[: (self.nocc * self.nvirt)], (self.nocc, self.nvirt)
        )
        guess = singles.ravel()
        return guess

    # FIXME: use matrix class
    def check_result(self, **kwargs):
        """Check if amplitudes are reasonable."""
        pass

    def get_model(self):
        return "RHFCCS"

    @property
    def multireference(self):
        """multireference wavefunction"""
        return False


class RpCCDCCS(RCCS):
    def print_info(self, **kwargs):
        """Print optimization options.
        """
        thresh = kwargs.get("threshold", 1e-7)
        maxiter = kwargs.get("maxiter", 200)
        log.hline()
        log("Coupled Cluster Singles correction on top of RpCCD")
        log.hline()
        log("OPTIMIZATION PARAMETERS:")
        log("Cluster Operator:            T1")
        log(f"Number of occupied orbitals: {self.nocc}")
        log(f"Number of frozen cores:      {self.ncore}")
        log(f"Number of virtuals:          {self.nvirt}")
        log("Initial guess:               MP2")
        log("Solvers:")
        log("  cluster amplitudes:        krylov")
        log("Optimization thresholds:")
        log(f"  cluster amplitudes:        {thresh:1.2e}")
        log(f"  maxiter:                   {maxiter}")
        log.hline()

    def fix_guess(self, guess):
        """Set diagonal of double amplitudes to Zero. This avoids any errors
           induced by user-specified guess

           **Arguements:**

           guess:
                The initial guess. A numpy ndarray
        """
        singles = np.reshape(
            guess[: (self.nocc * self.nvirt)], (self.nocc, self.nvirt)
        )
        guess = singles.ravel()
        return guess

    @property
    def multireference(self):
        """multireference wavefunction"""
        return True

    def check_result(self, **kwargs):
        """Check if amplitudes are reasonable."""
        pass

    def get_model(self):
        return "RpCCDCCS"
