# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Restricted Coupled Cluster Singles Doubles implementations

   Variables used in this module:
    :nocc:      number of occupied orbitals in the principle configuration
    :nvirt:     number of virtual orbitals in the principle configuration
    :ncore:     number of frozen core orbitals in the principle configuration
    :nbasis:    total number of basis functions
    :energy:    the CCSD energy, dictionary that contains different
                contributions
    :t_1, t_2:  the optimized amplitudes

    Indexing convention:
    :o:        matrix block corresponding to occupied orbitals of principle
               configuration
    :v:        matrix block corresponding to virtual orbitals of principle
               configuration


    EXAMPLE APPLICATION

    1) Restricted closed-shell CCSD with default parameters

    solver = RCCSD(linalg_factory, occupation_model)
    result = solver(AO_one, AO_two, orbital_orbital, eref=hf_energy)
"""

import numpy as np
import pathlib
import time
from scipy import optimize as opt

from pybest.auxmat import get_fock_matrix
from pybest.cache import Cache
from pybest.log import log, timer
from pybest.io.iodata import IOData
from pybest.io.checkpoint import CheckPoint
from pybest.linalg.base import TwoIndex, FourIndex
from pybest.linalg.dense import DenseLinalgFactory, DenseOneIndex
from pybest.linalg.dense import DenseTwoIndex, DenseFourIndex
from pybest.linalg.cholesky import CholeskyFourIndex, CholeskyLinalgFactory
from pybest.scf import AufbauOccModel
from pybest.pt.mp2 import RMP2
from pybest.utils import unmask, unmask_orb
from pybest.featuredlists import OneBodyHamiltonian, TwoBodyHamiltonian


__all__ = [
    "RCCD",
    "RCCSD",
]


class RCC(object):
    """ Restricted Coupled Cluster Class

        Purpose:
        Optimize cluster amplitudes and determine energy corrections to some
        reference wave function.

        Currently supported reference wave function models:
            * RHF

        Currently supported cluster operators:
            * doubles
            * singles and doubles

    """

    def __init__(self, lf, occ_model, **kwargs):
        """ Arguments
                
                lf : DenseLinalgFactory or CholeskyLinalgFactory
                
                occ_model : AufbauOccModel

            Keyword arguments

                ncore : int
                    number of frozen core orbitals
                
                active : int
                    number of active orbitals
                
                actocc : int
                    number of active occupied orbitals
                
                actvir : int
                    number of active virtual orbitals
                
                external : int
                    number of frozen virtual orbitals
        """
        self._lf = lf
        self._occ_model = occ_model
        self._nbasis = lf.default_nbasis
        self._cache = Cache()

        self._energy = {}
        self._amplitudes = None

        self._singles = self._get_singles()
        self._doubles = True
        self._method_abbrev = self._get_method_abbrev()
        self._method_name = self._get_method_name()

        self._nocc = occ_model.noccs[0]
        self._nvirt = lf.default_nbasis - occ_model.noccs[0]
        self._ncore = kwargs.get("ncore", 0)
        self._nact = kwargs.get("active", self.nbasis - self.ncore)
        self._nacto = kwargs.get("actocc", self.nocc - self.ncore)
        self._nactv = kwargs.get("actvir", self.nact - self.nacto)
        self._next = kwargs.get("external", self.nvirt - self.nactv)
        self._nov = self.nacto * self.nactv

        self._iocc = (self.ncore, self.nocc)
        self._ivir = (self.nocc, self.nocc + self.nactv)
        self._iter = 0
        self._iodata = {"converged": False}

        if not (self.ncore + self.nacto) == self.nocc:
            log("Input error. At least one of following is incorrect:")
            log(f"- the number of frozen core orbitals ({self.ncore})")
            log(f"- the number of active occupied orbitals ({self.nacto})")
            log(f"- the total number of occupied orbitals ({self.nocc})")
            raise ValueError("Input error: number of occupied orbitals.")

        if not (self.nactv + self.next) == self.nvirt:
            log("Input error. At least one of following is incorrect:")
            log(f"- the number of frozen virtual orbitals ({self.next})")
            log(f"- the number of active virtual orbitals ({self.nactv})")
            log(f"- the total number of virtual orbitals ({self.nvirt})")
            raise ValueError("Input error: number of virtual orbitals.")

        if not (self.nacto + self.nactv) == self.nact:
            log("Input error. At least one of following is incorrect:")
            log(f"- the number of total active orbitals {self.nact})")
            log(f"- the number of active occupied orbitals ({self.nacto})")
            log(f"- the number of active virtual orbitals ({self.nactv})")
            raise ValueError("Input error: number of active orbitals.")

    def _get_singles(self):
        return False

    @property
    def singles(self):
        """Consider single excitations."""
        return self._singles

    @property
    def lf(self):
        """The linalg factory"""
        return self._lf

    @property
    def occ_model(self):
        """The occupation model"""
        return self._occ_model

    @property
    def nbasis(self):
        """The number of basis functions."""
        return self._nbasis

    @property
    def nocc(self):
        """The number of occupied orbitals."""
        return self._nocc

    @property
    def nvirt(self):
        """The number of unoccupied (virtual) orbitals."""
        return self._nvirt

    @property
    def ncore(self):
        """The number of core orbitals."""
        return self._ncore

    @property
    def nact(self):
        """The number of active orbitals."""
        return self._nact

    @property
    def nacto(self):
        """The number of active occupied orbitals."""
        return self._nacto

    @property
    def nactv(self):
        """The number of active unoccupied (virtual) orbitals."""
        return self._nactv

    @property
    def next(self):
        """The number of external orbitals."""
        return self._next

    @property
    def nov(self):
        """The number of occ * virt orbitals."""
        return self._nov

    @property
    def energy(self):
        """ Dictionary containing energies """
        return self._energy

    @energy.setter
    def energy(self, energy_dict):
        self._energy.update(energy_dict)
        self._iodata.update(energy_dict)

    @energy.deleter
    def energy(self):
        self._energy = {}

    @property
    def iodata(self):
        """The dictionary that contains all data dump to disk"""
        return self._iodata

    @iodata.setter
    def iodata(self, new):
        # new should be passed as {key: value}
        self._iodata.update(new)

    def dump_iodata(self):
        """Save data (amplitudes, energies) to disk."""
        # Use CheckPoint class to dump
        checkpoint = CheckPoint(self.iodata)
        checkpoint.to_file(self.dump_to)

    @property
    def amplitudes(self):
        """ Returns flatten t_1 and t_2 amplitudes. """
        return self.ravel_amplitudes(self.iodata.t_1, self.iodata.t_2)

    @amplitudes.setter
    def amplitudes(self, x):
        """ Allows to set amplitudes from flat vector. """
        t_1, t_2 = self.unravel_amplitudes(x)
        self.iodata.update({"t_2": t_2})
        if self.singles:
            self.iodata.update({"t_1": t_1})

    @property
    def t_1(self):
        """ Single excitation cluster amplitudes. """
        if self.singles:
            return self.iodata.t_1
        else:
            raise AttributeError("RCCD does not have t_1 attribute.")

    @t_1.setter
    def t_1(self, t_1):
        if self.singles:
            self.iodata.t_1 = t_1
        else:
            raise AttributeError("RCCD does not have t_1 attribute.")

    @property
    def t_2(self):
        """ Single excitation cluster amplitudes. """
        return self.iodata.t_2
        
    @t_2.setter
    def t_2(self):
        self.iodata.update({"t_2": t_2})

    def __call__(self, *args, **kwargs):
        """Optimize cluster amplitudes and determine energy correction to some
        Hamiltonian and reference wavefunction.

           Cluster amplitudes are determined by finding the root of a set of
           equations of the form A.x = 0

           Currently supported reference wavefunction models (Psi_0):

               * RHF

           Arguments

               one : DenseTwoIndex
                   One-electron integrals in the AO basis.

               two : DenseFourIndex or CholeskyFourIndex
                   Electron repulsion integrals in the AO basis.

               orb : DenseOrbital
                   MO coefficients.

            OR

               IOData instance containing all above information.


           Keyword arguments

           Contains reference energy and solver specific input parameters:

               checkpoint : int
                   frequency of checkpointing (int). If > 0, writes energies
                   and amplitudes to a checkpoint file (default 1).

               checkpoint_path : str
                   path to file where current solution is dumped
                   (default ./checkpoint_method.h5).

               e_ref : float
                   Reference energy (default value 0.0).

               initguess : str or numpy.ndarray
                   Initial guess for amplitudes, one of:
                   * "random" - random numbers,
                   * "mp2" - performs MP2 calculations before,
                   * "const" - uses constant number, used for debugging,
                   * path to file - path to .h5 file containing amplitudes
                       (saved as checkpoint_method.h5 by default),
                   * numpy.ndarray - 1d vector with amplitudes.

               maxiter : int
                   Maximum number of iterations (default: 100).

               solver : str
                   one of scipy.optimize.root solvers (default: "krylov").

               threshold : float
                   Tolerance for amplitudes (default: 1e-8).

        """

        # Set default keyword arguments:

        one, eri, orb = self.read_input(*args, **kwargs)
        self._iter = 0

        # Print info
        log(" ")
        log(" ")
        if log.do_medium:
            log.hline("=")
            log(" ")
            log("RCC module")
            log(" ")
            self.print_info(**kwargs)

        # Get guess
        guess = self.get_guess(orbital=orb, ao1=one, ao2=eri)

        # Get Fock matrix and Hamiltonian blocks
        self.get_mo_integrals(one, eri, orb)
        del one, eri

        self._time = time.time()

        # Solve matrix equation using scipy.optimize.root routine
        options = {
            "callback": self.callback,
            "method": self.solver,
            "tol": self.thresh,
            "options": {"maxiter": self.maxiter},
        }

        if self.singles == True:
            solution = opt.root(self.vfunction_sd, guess, **options)
        else:
            solution = opt.root(self.vfunction_d, guess, **options)

        if solution.success:
            if log.do_medium:
                log(" ")
                log(f"Amplitudes converged in {solution.nit-2} iterations.")
            self.iodata.update({"converged": True})
        else:
            log(" ")
            log.warn("Program terminated. Amplitudes did not converge.")
        log(" ")

        t_1, t_2 = self.unravel_amplitudes(solution.x)
        energy = self.calculate_energy(self.e_ref, t_1, t_2)
        self.del_integrals()
        self.energy = energy
        self.print_energy()
        self.amplitudes = solution.x
        if log.do_medium:
            self.print_amplitudes()
        self.dump_iodata()
        return IOData(**self.iodata)

    def read_input(self, *args, **kwargs):

        # Obligatory arguments
        orb = unmask_orb(*args)[0]
        one = self.lf.create_two_index(label="one")

        for label in OneBodyHamiltonian:
            one_body_term = unmask(label, *args, **kwargs)
            if isinstance(one_body_term, TwoIndex):
                one.iadd(one_body_term)

        for label in TwoBodyHamiltonian:
            eri = unmask(label, *args, **kwargs)
            if eri != None:
                break

        assert orb != None, "Please, provide AO/MO coefficients."
        assert eri != None, "Please, provide electron repulsion integrals."

        # Saving options
        self.checkpoint = kwargs.get("checkpoint", True)
        dump_to = "./checkpoint_" + self._get_method_abbrev() + ".h5"
        self.dump_to = kwargs.get("checkpoint_fn", dump_to)

        # Other options
        self.initguess = kwargs.get("initguess", "random")
        self.maxiter = kwargs.get("maxiter", 100) + 1
        self.solver = kwargs.get("solver", "krylov")
        self.thresh = kwargs.get("threshold", 1e-8)

        # Reference determinant energy
        self.e_ref = unmask("e_ref", *args, **kwargs)
        if self.e_ref == None:
            self.e_ref = kwargs.get("e_ref", 0.0)
        return one, eri, orb

    def callback(self, x, f):
        """ Inform about current status of optimization process.
            
            Arguments:
               
               x : numpy.ndarray
                   current solution
               
               f : numpy.ndarray
                   vector of residuals
        """
        t_1, t_2 = self.unravel_amplitudes(x)
        old_energy = self.energy
        new_energy = self.calculate_energy(self.e_ref, t_1, t_2)
        time_ = time.time() - self._time

        if self._iter == 0:
            if log.do_medium:
                log(" ")
                log(f"Energy of reference determinant: {new_energy['e_ref']}")
                log(f"Initial total energy: {new_energy['e_tot']}")
                log(f"Initial correlation energy: {new_energy['e_corr']}")
                log(" ")
                log(
                    f"{'I':>6} {'E_tot':>16} {'D(E)':>11} {'E_corr':>15} "
                    f"{'Residual':>11} {'Time':>12}"
                )
                log(" ")
        else:
            res = np.sum(np.absolute(f))
            diff = old_energy["e_tot"] - new_energy["e_tot"]
            if log.do_medium:
                log(
                    f"{self._iter:6} {new_energy['e_tot']:16.10f} {diff:11.2e} "
                    f"{new_energy['e_corr']:15.10f} {res:>11.2e} {time_:12.2}"
                )

            if self._iter % self.checkpoint == 0:
                self.dump_iodata()

        self.amplitudes = x
        self.energy = new_energy
        self._iter += 1
        self._time = time.time()

    def print_info(self, **kwargs):
        """Print information about model."""
        log.hline()
        log("PARAMETERS")
        log(f"{'Wave function model':30} {self._method_abbrev}")
        log(f"{'Full name of model':30} {self._method_name}")
        if self.singles:
            log(f"{'Cluster operator':30} {'T1 + T2 (singles and doubles)'}")
        else:
            log(f"{'Cluster operator':30} {'T2 (doubles)'}")
        log(f"{'Total number of electron pairs':30} {self.nocc}")
        log(f"{'Total number of orbitals':30} {self.nocc + self.nvirt}")
        log(f"{'Number of active orbitals':30} {self.nact}")
        log(f"{'Number of core orbitals':30} {self.ncore}")
        log(f"{'Number of active occupied':30} {self.nacto}")
        log(f"{'Number of active virtual':30} {self.nvirt}")
        log(f"{'Number of inactive virtual':30} {self.next}")
        log(f"{'Initial guess':30} {self.initguess}")
        log(f"{'Solver':30} {self.solver}")
        log(f"{'Optimization thresholds':30} {self.thresh}")
        log(f"{'Maximum number of iterations':30} {self.maxiter}")
        log.hline(".")

    ################################################################################
    # Energy - calculate, print
    ################################################################################

    def calculate_energy(self, e_ref, t_1, t_2):
        """Returns a dictionary of energies:
            e_tot: total energy,
            e_corr: correlation energy,
            e_ref: energy of reference determinant,
            e_corr_s: part of correlation energy,
            e_corr_d: part of correlation energy.
        """
        energy = {
            "e_ref": e_ref,
            "e_tot": 0.0,
            "e_corr": 0.0,
            "e_corr_s": 0.0,
            "e_corr_d": 0.0,
        }

        ov = self.get_range("ov")
        e_oovv = self._exch_oovv

        energy["e_corr_d"] = -e_oovv.contract("abcd,adbc", t_2)

        if self.singles:
            e1 = 2 * self.fok.contract("ab,ab", t_1, **ov)
            tmp_ov = DenseTwoIndex(self.nacto, self.nactv)
            e_oovv.contract("abcd,bc->ad", t_1, out=tmp_ov)
            e11 = -tmp_ov.contract("ab,ab", t_1)
            energy["e_corr_s"] = e1 + e11

        energy["e_corr"] = energy["e_corr_d"] + energy["e_corr_s"]
        energy["e_tot"] = e_ref + energy["e_corr"]
        return energy

    def print_energy(self):
        if log.do_medium:
            log.hline("-")
            log(f"{self._method_abbrev} energy")
            log(f"{'Total':21} {self.energy['e_tot']:16.8f} a.u.")
            log(f"{'Reference determinant'} {self.energy['e_ref']:16.8f} a.u.")
            log(f"{'Correlation':21} {self.energy['e_corr']:16.8f} a.u.")
            log(f"{'Singles':21} {self.energy['e_corr_s']:16.8f} a.u.")
            log(f"{'Doubles':21} {self.energy['e_corr_d']:16.8f} a.u.")
            log.hline("-")
            log(" ")

    def print_amplitudes(self, threshold=1e-4):

        nocc = self.nacto + self.ncore
        double = self.iodata["t_2"]
        max_single = []

        if self.singles:
            single = self.iodata["t_1"]
            max_single = single.get_max_values(absolute=True)

        max_double = double.get_max_values(absolute=True)

        if max_double:
            log("Leading double excitation amplitudes")
            log(" ")
            log(f"{'amplitude':>13}{'i':>4}{'j':>4}  ->{'a':>4}{'b':>4}")
            log(" ")

            for amplitude_tuple in max_double:
                index, value = amplitude_tuple
                i, a, j, b = index
                if value > threshold:
                    log(f"{value:13.6f}{i:4}{j:4}  ->{a + nocc:4}{b + nocc:4}")
            log.hline("-")

        if self.singles and max_single:
            log("Leading single excitation amplitudes")
            log(" ")
            log(f"{'amplitude':>13}{'i':>4}  ->{'a':>4}")
            log(" ")
            for amplitude_tuple in max_single:
                index, value = amplitude_tuple
                i, a = index
                if value > threshold:
                    log(f"{value:13.6f}{i:>4}  ->{a + nocc:4}")
            log.hline("-")

    ##############################################################################
    # Initial guess
    ##############################################################################

    def get_guess(self, orbital=None, ao1=None, ao2=None, constant=0.0625):
        """Generate the initial guess for CC amplitudes

        Keyword arguments:
            
            orbital : DenseOrbital            
                Expansion coefficients (required for mp2 guess)
 
            ao1 : DenseTwoIndex
                One-body Hamiltonian in AO basis (required for mp2 guess)

            ao2 : DenseFourIndex or CholeskyFourIndex
                Two-body Hamiltonian in AO basis (required for mp2 guess)
	        * required for guess from file:

            constrant : float
	        (optional for constant guess)

        Returns a 1-dimensional numpy.ndarray
             
             * RCCD
                - t_2 amplitudes [:]
             
             * RCCSD
                - t_1 amplitudes [:self.nacto * self.nactv]
                - t_2 amplitudes [self.nacto * self.nactv:]
        """
        msg = f"Initial guess {self.initguess} has been not recognized."

        if type(self.initguess) == np.ndarray:
            return np.ndarray

        elif pathlib.Path(self.initguess).is_file():
            t_2 = IOData.from_file(self.initguess).t_2
            t_1 = None
            if self.singles:
                t_1 = IOData.from_file(self.initguess).t_1
            log(f"The amplitudes have been read from file {self.initguess}.")
            log.hline(".")

        elif not isinstance(self.initguess, str):
            raise UnknownOption(msg)

        elif self.initguess.lower() in ["random", None]:
            t_1, t_2 = self.get_random_guess()

        elif self.initguess.lower() in ["mp2"]:
            mo1 = ao1.new()
            mo1.assign_two_index_transform(ao1, orbital)
            mo2 = ao2.new()
            mo2.assign_four_index_transform(ao2, orbital)
            t_1, t_2 = self.get_mp2_guess(mo1, mo2, orbital)

        elif self.initguess.lower() in ["constant", "const"]:
            t_1, t_2 = self.get_constant_guess(constant)

        else:
            raise UnknownOption(msg)

        return self.ravel_amplitudes(t_1, t_2)

    def get_mp2_guess(self, mo1, mo2, orbital):
        """Generate the MP2 initial guess for CC amplitudes"""

        if log.do_medium:
            log("Performing an MP2 calculations for an initial guess.")
        # Only pring warnings in MP2 guess, e_ref/e_core do not matter here
        loglevel = log.level
        log.level = 1
        mp2 = RMP2(self.lf, self.occ_model, ncore=self.ncore)
        mp2_ = mp2(
            mo1,
            mo2,
            orbital,
            IOData(e_core=0.0),
            singles=self.singles,
            natorb=False,
            relaxation=False,
            e_ref=0.0,
        )
        mp2.clear()
        log.level = loglevel

        if log.do_medium:
            log("Resuming CC calculation.")
            log.hline("~")

        if not self.singles:
            return None, mp2_.t_2

        return mp2_.t_1, mp2_.t_2

    def get_random_guess(self):
        """Generate random guess for t_1 ov matrix and t_2 ovov matrix."""

        nacto = self.nacto
        nactv = self.nactv
        coeff = 1.0 / (self.nov)
        t_1 = DenseTwoIndex(nacto, nactv)
        t_1.randomize()
        t_1.iscale(coeff)
        t_2 = DenseFourIndex(nacto, nactv, nacto, nactv)
        t_2.randomize()
        t_2.iscale(coeff)
        t_2.iadd_transpose((2, 3, 0, 1))
        return t_1, t_2

    def get_constant_guess(self, constant):
        """Generate constant guess for t_1 ov matrix and t_2 ovov matrix."""

        nacto = self.nacto
        nactv = self.nactv
        t_1 = DenseTwoIndex(nacto, nactv)
        t_1.assign(constant)
        t_2 = DenseFourIndex(nacto, nactv, nacto, nactv)
        t_2.assign(constant)
        return t_1, t_2

    ################################################################################
    # Vector functions
    ################################################################################

    @timer.with_section("VecFct RCCSD")
    def vfunction_sd(self, amplitudes):
        """ Vector function to find coupled cluster amplitudes.
            Need to be solved.

            Arguments
 
                amplitudes : numpy.ndarray
                    vector containing singles and doubles cluster amplitudes.

            Abbreviations

            * o - number of active occupied orbitals
            * v - number of active virtual orbitals
            * t_1, t_2 - current solution for CC amplitudes
            * out_s, out_d - vector function containers (singles, doubles)
            * aux - auxilary matrix t_ia_jb + t_ia * t_jb
            * mat_ov - two-index temporary matrix (ov is occupied x virtual)
            * e_oovv - four index integrals with exchange part
        """
        o = self.nacto
        v = self.nactv

        # Create matrices of amplitudes - ov and ovov, and assign current guess

        t_1, t_2 = self.unravel_amplitudes(amplitudes)
        aux = t_2.copy()
        t_1.contract("ab,cd->abcd", t_1, out=aux)

        # Fock matrix, ERI, exchange terms

        fok = self.fok
        eri = self.eri
        e_ooov = self._exch_ooov
        e_oovv = self._exch_oovv

        oo = self.get_range("oo")
        ov = self.get_range("ov")
        vv = self.get_range("vv")
        ov_ = self.get_range("ov", offset=2)
        vvvv = self.get_range("vvvv")
        vovv = self.get_range("vovv")
        oovv = self.get_range("oovv")
        ovov = self.get_range("ovov")
        ooov = self.get_range("ooov")
        oooo = self.get_range("oooo")

        ## SINGLES equations ##
        # 6.0
        out_s = fok.copy(**ov)
        to_s = {"out": out_s, "clear": False}
        # 6.1
        e_ooov.contract("abcd,becd->ae", aux, **to_s)
        # 6.2
        eri.contract("abcd,edbc->ea", aux, out_s, factor=-1, **vovv)
        eri.contract("abcd,ecbd->ea", aux, out_s, factor=2, **vovv)
        # 7.0
        eri.contract("abcd,cd->ab", t_1, out_s, factor=-1, **ovov)
        # 7.1
        eri.contract("abcd,bd->ac", t_1, out_s, factor=2, **oovv)
        # 5 intermediate matrix
        mat_ov = fok.copy(**ov)
        e_oovv.contract("abcd,bc->ad", t_1, out=mat_ov, factor=-1)
        # 7.2
        t_term = aux.copy()
        t_term.iscale(factor=-0.5)
        t_term.itranspose((0, 3, 2, 1))
        t_term.iadd(t_2)
        t_term.contract("abcd,cd->ab", mat_ov, factor=2, **to_s)
        t_term = None
        mat_ov.clear()
        # 3 intermediate matrix
        mat_oo = fok.copy(**oo)
        mat_oo.iscale(-1)
        to_o = {"out": mat_oo, "clear": False}
        e_oovv.contract("abcd,ecad->eb", t_2, **to_o)
        # 7.3
        mat_oo.contract("ab,bc->ac", t_1, **to_s)
        # 4 intermediate matrix
        mat_vv = fok.copy(**vv)
        to_v = {"out": mat_vv, "clear": False}
        e_oovv.contract("abcd,bead->ec", t_2, **to_v)
        # 7.4
        t_1.contract("ab,cb->ac", mat_vv, **to_s)

        ## DOUBLES equations ##
        # Create doubles vector function ovov
        out_d = DenseFourIndex(o, v, o, v)
        to_d = {"out": out_d, "clear": False}
        # 8.1 intermediate
        t_1.contract("ab,cb->ac", fok, factor=-1, **ov_, **to_o)
        # 8.2
        e_ooov.contract("abcd,cd->ab", t_1, **to_o)
        # 8.3
        e_oovv.contract("abcd,ad->bc", t_1, out=mat_ov, clear=True)
        t_1.contract("ab,cb->ac", mat_ov, **to_o)
        # 12.0
        t_2.contract("abcd,ec->edab", mat_oo, **to_d)
        mat_oo = None
        # 9.1 intermediate
        t_1.contract("ab,ac->bc", fok, factor=-1, **ov_, **to_v)
        # 9.2
        eri.contract("abcd,bc->ad", t_1, mat_vv, factor=-1, **vovv)
        eri.contract("abcd,bd->ac", t_1, mat_vv, factor=2, **vovv)
        # 9.3
        e_oovv.contract("abcd,ad->bc", t_1, out=mat_ov, clear=True)
        t_1.contract("ab,ac->bc", mat_ov, factor=1, **to_v)
        # 12.1
        t_2.contract("abcd,ed->ceab", mat_vv, **to_d)
        mat_vv = None
        # 10.0 intermediate
        w_oovo = DenseFourIndex(o, o, v, o)
        to_w = {"out": w_oovo, "clear": False}
        eri.contract("abcd->badc", w_oovo, select="einsum", factor=-1, **ooov)
        # 10.1
        e_ooov.contract("abcd,efcd->eafb", t_2, **to_w)
        # 10.2
        eri.contract("abcd,befd->faec", t_2, w_oovo, **ooov)
        # 10.3
        eri.contract("abcd,cefd->afeb", t_2, w_oovo, **ooov)
        # 10.4
        eri.contract("abcd,ecfd->efab", t_2, w_oovo, factor=-1, **vovv)
        # 10.5
        t_ooov = DenseFourIndex(o, o, o, v)
        eri.contract("abcd,ec->eabd", t_1, out=t_ooov, clear=True, **oovv)
        t_ooov.contract("abcd,cefd->faeb", t_2, **to_w)
        # 10.6
        eri.contract("abcd,ed->eabc", t_1, out=t_ooov, clear=True, **oovv)
        t_ooov.contract("abcd,cefd->afeb", t_2, **to_w)
        # 12.2
        w_oovo.contract("abcd,de->beac", t_1, **to_d)
        w_oovo = None
        # 11.0
        u_ovov = DenseFourIndex(o, v, o, v)
        eri.contract("abcd->abcd", u_ovov, factor=-1, **ovov)
        to_u = {"out": u_ovov, "clear": False}
        # 11.1
        eri.contract("abcd->acbd", out=u_ovov, factor=2, **oovv)
        # 11.2
        eri.contract("abcd,ed->eabc", t_1, u_ovov, factor=-1, **vovv)
        eri.contract("abcd,ec->eabd", t_1, u_ovov, factor=2, **vovv)
        # 11.3
        t_term = t_2.copy()
        t_term.itranspose((2, 1, 0, 3))
        t_term.iscale(-1)
        t_term.iadd(aux)
        e_oovv.contract("abcd,befc->fead", t_term, **to_u)
        t_term = None
        # 12.3
        u_ovov.contract("abcd,efcd->abef", t_2, **to_d)
        u_ovov = None
        # 13.0
        eri.contract("abcd,ec->bdea", t_1, out_d, **vovv)
        # 13.1
        eri.contract("abcd,befd->acfe", aux, out_d, factor=-1, **oovv)
        # 13.2
        eri.contract("abcd,cefd->aefb", aux, out_d, factor=-1, **ovov)
        # 13.3
        intmat = DenseFourIndex(o, o, o, o)
        eri.contract("abcd,ed->aebc", t_1, out=intmat, clear=True, **ooov)
        intmat.contract("abcd,cedf->afbe", aux, **to_d)
        intmat = None
        # 13.4
        intmat = DenseFourIndex(o, v, o, v)
        eri.contract("abcd,ed->eabc", t_1, out=intmat, clear=True, **vovv)
        intmat.contract("abcd,cefd->fbae", aux, factor=-1, **to_d)
        # 13.5
        eri.contract("abcd,ec->eabd", t_1, out=intmat, clear=True, **vovv)
        intmat.contract("abcd,cefd->abfe", t_2, factor=-1, **to_d)
        # Permutation
        out_d.iadd_transpose((2, 3, 0, 1))
        # 14.0
        eri.contract("abcd,befd->feac", t_2, out=intmat, clear=True, **oovv)
        intmat.contract("abcd,cefd->abfe", t_2, **to_d)
        # 14.1
        eri.contract("abcd,edaf->efbc", t_2, out=intmat, clear=True, **oovv)
        intmat.contract("abcd,cefd->aefb", t_2, **to_d)
        intmat = None
        # 14.2 and 14.5
        intmat = DenseFourIndex(o, o, o, o)
        intmat.clear()
        eri.contract("abcd,ecfd->efab", aux, out=intmat, clear=True, **oovv)
        eri.contract("abcd->abcd", out=intmat, **oooo)
        intmat.contract("abcd,cedf->aebf", aux, **to_d)
        # 14.3
        eri.contract("abcd->acbd", out=out_d, **oovv)
        # 14.4 bottleneck contraction
        eri.contract("abcd,ecfd->eafb", aux, out=out_d, **vvvv)

        ## Matrices of equations -> Vector (ndarray) of equations
        return self.ravel_amplitudes(out_s, out_d)

    @timer.with_section("VecFct RCCD")
    def vfunction_d(self, amplitudes):
        """ Vector function to find coupled cluster amplitudes.
            Need to be solved.

            Arguments
        
                amplitudes : numpy.ndarray
                    vector containing singles and doubles cluster amplitudes.

            Abbreviations

            * o - number of active occupied orbitals
            * v - number of active virtual orbitals
            * t_2 - current solution for CC amplitudes
            * out_d - vector function
            * mat_oo - two-index temporary matrix (ov is occupied x virtual)
            * e_oovv - four index integrals with exchange part
        """
        o = self.nacto
        v = self.nactv

        # Create matrices of amplitudes - ov and oovv, and assign current guess

        t_1, t_2 = self.unravel_amplitudes(amplitudes)

        # Fock matrix, ERI, exchange terms

        fok = self.fok
        eri = self.eri
        e_oovv = self._exch_oovv

        oo = self.get_range("oo")
        vv = self.get_range("vv")
        vvvv = self.get_range("vvvv")
        oovv = self.get_range("oovv")
        ovov = self.get_range("ovov")
        oooo = self.get_range("oooo")

        ## DOUBLEs ##

        # Create vector function
        out_d = DenseFourIndex(o, v, o, v)
        out_d.clear()
        to_d = {"out": out_d, "clear": False}
        # 3 intermediate matrix
        mat_oo = fok.copy(**oo)
        mat_oo.iscale(-1)
        to_o = {"out": mat_oo, "clear": False}
        e_oovv.contract("abcd,ecad->eb", t_2, **to_o)
        # 12.0
        t_2.contract("abcd,ec->edab", mat_oo, **to_d)
        mat_oo = None
        # 4 intermediate matrix
        mat_vv = fok.copy(**vv)
        to_v = {"out": mat_vv, "clear": False}
        e_oovv.contract("abcd,bead->ec", t_2, **to_v)
        # 12.1
        t_2.contract("abcd,ed->ceab", mat_vv, **to_d)
        mat_vv = None
        # 11.0
        u_ovov = DenseFourIndex(o, v, o, v)
        eri.contract("abcd->abcd", u_ovov, factor=-1, **ovov)
        to_u = {"out": u_ovov, "clear": False}
        # 11.1
        eri.contract("abcd->acbd", out=u_ovov, factor=2, **oovv)
        # 11.3
        t_term = t_2.copy()
        t_term.itranspose((2, 1, 0, 3))
        t_term.iscale(-1)
        t_term.iadd(t_2)
        e_oovv.contract("abcd,befc->fead", t_term, **to_u)
        t_term = None
        # 12.3
        u_ovov.contract("abcd,efcd->abef", t_2, **to_d)
        u_ovov = None
        # 13.1
        eri.contract("abcd,befd->acfe", t_2, out_d, factor=-1, **oovv)
        # 13.2
        eri.contract("abcd,cefd->aefb", t_2, out_d, factor=-1, **ovov)
        # Permutation
        out_d.iadd_transpose((2, 3, 0, 1))
        # 14.0
        intmat = DenseFourIndex(o, v, o, v)
        eri.contract("abcd,befd->feac", t_2, out=intmat, clear=True, **oovv)
        intmat.contract("abcd,cefd->abfe", t_2, **to_d)
        # 14.1
        eri.contract("abcd,edaf->efbc", t_2, out=intmat, clear=True, **oovv)
        intmat.contract("abcd,cefd->aefb", t_2, **to_d)
        intmat = None
        # 14.2 and 14.5
        intmat = DenseFourIndex(o, o, o, o)
        eri.contract("abcd,ecfd->efab", t_2, out=intmat, **oovv)
        eri.contract("abcd->abcd", out=intmat, **oooo)
        intmat.contract("abcd,cedf->aebf", t_2, **to_d)
        # 14.3
        eri.contract("abcd->acbd", out=out_d, **oovv)
        # 14.4 bottleneck contraction
        eri.contract("abcd,ecfd->eafb", t_2, out=out_d, **vvvv)

        ## Matrices of equations -> Vector (ndarray) of equations
        return self.ravel_amplitudes(out_d)

    ################################################################################
    # Integrals
    ################################################################################

    def get_mo_integrals(self, one, two, transform):
        """ Transforms integrals in AO basis to Fock matrix and ERI in MO
            basis and save them as attributes (self.fok and self.eri).
        """
        mo1 = one.new()
        mo1.assign_two_index_transform(one, transform)
        eri = two.new()
        eri.assign_four_index_transform(two, transform)
        self.fok = get_fock_matrix(self.lf, mo1, eri, self.nocc)
        self.eri = eri
        del mo1
        self.get_exchange()

    def get_exchange(self):
        """ ERI with exchange part are now attributes:
            * self._exch_oovv = <ijab> - 2 <ijba>
            * self._exch_ooov = <ijka> - 2 <ikja>
        """
        eri = self.eri
        fok = self.fok

        o = self.nacto
        v = self.nactv
        ooov = self.get_range("ooov")
        oovv = self.get_range("oovv")

        e_ooov = DenseFourIndex(o, o, o, v)
        eri.contract("abcd->abcd", out=e_ooov, **ooov)
        e_ooov.iadd_transpose((0, 2, 1, 3), factor=-2)
        self._exch_ooov = e_ooov

        e_oovv = DenseFourIndex(o, o, v, v)
        eri.contract("abcd->abcd", out=e_oovv, **oovv)
        e_oovv.iadd_transpose((0, 1, 3, 2), factor=-2)
        self._exch_oovv = e_oovv

    def get_range(self, string, offset=0):
        """ Returns dictionary with keys beginX, endX, begin(X+1), etc.
        
            Arguments

                string : string
                    any sequence of "o" (occupied) and "v" (virtual)

            Keyword arguments
 
                offset : int
                    keys in returned dictionary start from offset value.
        """
        range_ = {}
        # ind = start
        for ind, char in enumerate(string):
            if char == "o":
                range_[f"begin{(ind + offset)}"] = self._iocc[0]
                range_[f"end{(ind + offset)}"] = self._iocc[1]
            elif char == "v":
                range_[f"begin{(ind + offset)}"] = self._ivir[0]
                range_[f"end{(ind + offset)}"] = self._ivir[1]
            else:
                pass
            # ind +=1
        return range_

    def del_integrals(self):
        self.fok = None
        self.eri = None

    def ravel_amplitudes(self, *args):
        """ Returns one-dimensional numpy.ndarray containing flatten data from
        input operands.

        Arguments
            t_1 : DenseTwoIndex
            t_2 : DenseFourIndex
        """
        for arg in args:
            if isinstance(arg, DenseTwoIndex):
                t_1 = arg
            elif isinstance(arg, DenseFourIndex):
                t_2 = arg

        if self.singles:
            t_2_triu = t_2.get_triu()
            variable_number = self.nov + len(t_2_triu)
            flat_amplitudes = DenseOneIndex(variable_number)
            flat_amplitudes.assign(t_1._array.ravel(), end0=self.nov)
            flat_amplitudes.assign(t_2_triu, begin0=self.nov)
            return flat_amplitudes._array

        else:
            return t_2.get_triu()

    def unravel_amplitudes(self, flat_ndarray):
        """ Returns DenseTwoIndex and DenseFourIndex instances filled out with
        data from input flat_ndarray.

        Arguments
            flat_ndarray : numpy.ndarray 
        """

        occ = self.nacto
        vir = self.nactv
        t_2 = DenseFourIndex(occ, vir, occ, vir)

        if self.singles:
            t_1 = DenseTwoIndex(occ, vir)
            t_1.assign(flat_ndarray[: self.nov])
            t_2.assign_triu(flat_ndarray[self.nov :])
            tp_ = t_2.slice("abab->ab")
            t_2.iadd_transpose((2, 3, 0, 1))
            ind1, ind2 = np.indices((occ, vir))
            indp = [ind1, ind2, ind1, ind2]
            t_2.assign(tp_, indp)
            return t_1, t_2

        else:
            t_2.assign_triu(flat_ndarray[:])
            tp_ = t_2.slice("abab->ab")
            t_2.iadd_transpose((2, 3, 0, 1))
            ind1, ind2 = np.indices((occ, vir))
            indp = [ind1, ind2, ind1, ind2]
            t_2.assign(tp_, indp)
            return None, t_2


################################################################################
##### Classes for different methods
################################################################################


class RCCD(RCC):
    """ Restricted Coupled Cluster Doubles """

    def _get_method_abbrev(self):
        return "RCCD"

    def _get_method_name(self):
        return "Restricted Coupled Cluster Doubles"


class RCCSD(RCC):
    """ Restricted Coupled Cluster Singles Doubles """

    def _get_method_abbrev(self):
        return "RCCSD"

    def _get_method_name(self):
        return "Restricted Coupled Cluster Singles Doubles"

    def _get_singles(self):
        return True
