# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Linearized Coupled Cluster implementations

   Variables used in this module:
    :nocc:      number of (active) occupied orbitals in the principle configuration
    :nvirt:     number of (active) virtual orbitals in the principle configuration
    :ncore:     number of frozen core orbitals in the principle configuration
    :nbasis:    total number of basis functions
    :energy:    the energy correction, list that can contain different
                contributions
    :amplitudes:the optimized amplitudes, list that can contain different
                contributions ([doubles] or [singles, doubles])

    Indexing convention:
     :i,j,k,..: occupied orbitals of principle configuration
     :a,b,c,..: virtual orbitals of principle configuration
     :p,q,r,..: general indices (occupied, virtual)

    P^bc_jk performs a pair permutation, i.e., P^bc_jk o_(bcjk) = o_(cbkj)

    Abbreviations used (if not mentioned in doc-strings):
     :L_pqrs: 2<pq|rs>-<pq|sr>
     :g_pqrs: <pq|rs>
"""

import numpy as np
import warnings
import math
import gc
from scipy import optimize as opt
from scipy import linalg as linalg

from pybest.cache import Cache
from pybest.filemanager import filemanager
from pybest.io.iodata import IOData
from pybest.featuredlists import OneBodyHamiltonian, TwoBodyHamiltonian
from pybest.exceptions import (
    ArgumentError,
    RestartError,
    NonEmptyData,
    ConsistencyError,
    DirectoryError,
)
from pybest.log import log, timer
from pybest.utils import (
    check_type,
    check_options,
    unmask_orb,
    unmask,
)
from pybest.io.checkpoint import CheckPoint
from pybest.linalg.cholesky import CholeskyFourIndex, CholeskyLinalgFactory
from pybest.linalg.dense import DenseLinalgFactory
from pybest.pt.mp2 import RMP2
from pybest.pt.perturbation_utils import get_epsilon
from pybest.orbital_utils import (
    transform_integrals,
    split_core_active,
    project_orbitals,
)
from pybest.auxmat import get_fock_matrix
from pybest.linalg.base import OneIndex, TwoIndex, FourIndex, Orbital
from pybest.helperclass import PropertyHelper


# import memory_profiler


__all__ = [
    "RHFLCCD",
    "RHFLCCSD",
    "RpCCDLCCD",
    "RpCCDLCCSD",
]


class RLCC(object):
    """Restricted Linearized Coupled Cluster base class

       Purpose:
       Optimize cluster amplitudes and determine energy correction to some
       reference wavefunction.

       Currently supported wavefunction models:
        * RHF
        * RpCCD

       Currently supported cluster operators:
        * Doubles with occupied-virtual separation (RLCCD) and choice of frozen
          core orbitals
        * Singles and Doubles with occupied-virtual separation (RLCCSD);
          only for RpCCD
    """

    def __init__(self, lf, occ_model, ncore=None):
        """
           **Arguments:**

           lf
                A LinalgFactory instance.

           occ_model
                The occupation model.

           **Optional arguments:**

           ncore
                Number of frozen core orbitals. If provided, the core orbitals
                are excluded from the cluster operator.

        """
        check_type("frozencore", ncore, int, type(None))
        self._lf = lf
        self._denself = DenseLinalgFactory(lf.default_nbasis)
        if ncore is None:
            self._nocc = occ_model.noccs[0]
            self._ncore = 0
        else:
            self._nocc = occ_model.noccs[0] - ncore
            self._ncore = ncore
        if self.nocc <= 0 or self.nocc > occ_model.noccs[0]:
            raise ConsistencyError(
                f"Impossible number of frozen core orbitals! You have chosen {icore} "
                f"frozen core orbitals out of {occ_model.noccs[0]} occupied orbitals"
            )
        self._nbasis = lf.default_nbasis - self.ncore
        self._nvirt = self.nbasis - self.nocc
        if (
            self.nvirt <= 0
            or self.nvirt > lf.default_nbasis - occ_model.noccs[0]
        ):
            raise ConsistencyError(
                f"Impossible number of virtual orbitals: {self.nvirt}!"
            )
        self._cache = Cache()
        self._energy = {}
        self._amplitudes = []
        self._l_amplitudes = []
        self._l = False
        self._freeze = None
        self._tco = "td"
        self._occ_model = occ_model
        self._e_ref = 0.0
        self._e_core = 0.0
        self._solver = "krylov"
        self._jacobian_approximation = 1
        self._checkpoint = CheckPoint({})
        self._checkpoint_fn = f"checkpoint_{self.get_model()}.h5"

        log.cite("the linearized CC correction", "boguslawski2015b")

    @property
    def lf(self):
        """The linalg factory"""
        return self._lf

    @lf.setter
    def lf(self, new):
        self._lf = new

    @property
    def cache(self):
        """A Cache instance"""
        return self._cache

    @cache.setter
    def cache(self, new):
        self._cache = new

    @property
    def occ_model(self):
        """The linalg factory"""
        return self._occ_model

    @occ_model.setter
    def occ_model(self, new):
        self._occ_model = new

    @property
    def denself(self):
        """The dense linalg factory"""
        return self._denself

    @denself.setter
    def denself(self, new):
        self._denself = new

    @property
    def nbasis(self):
        """The number of basis functions"""
        return self._nbasis

    @nbasis.setter
    def nbasis(self, new):
        self._nbasis = new

    @property
    def nocc(self):
        """The number of occupied orbitals"""
        return self._nocc

    @nocc.setter
    def nocc(self, new):
        self._nocc = new

    @property
    def nvirt(self):
        """The number of virtual orbitals"""
        return self._nvirt

    @nvirt.setter
    def nvirt(self, new):
        self._nvirt = new

    @property
    def ncore(self):
        """The number of core orbitals"""
        return self._ncore

    @ncore.setter
    def ncore(self, new):
        self._ncore = new

    @property
    def energy(self):
        """The linearized coupled cluster energy contributions"""
        return self._energy

    @energy.setter
    def energy(self, energy_dict):
        self._energy.update(energy_dict)
        for key in energy_dict:
            self._checkpoint.update(key, energy_dict[key])

    @energy.deleter
    def energy(self):
        self._energy = {}

    @property
    def amplitudes(self):
        """The linearized coupled cluster amplitudes"""
        return self._amplitudes

    @amplitudes.setter
    def amplitudes(self, new):
        """Update singles and doubles cluster amplitudes in
           self._amplitudes[S,D] as well as in the IOData container used for
           checkpointing
        """
        if isinstance(new, np.ndarray) or isinstance(new, OneIndex):
            if self._amplitudes:
                raise NonEmptyData(
                    "Warning: List of LCC amplitudes not empty!"
                )

            end = 0
            if self.singles:
                singles = self.lf.create_two_index(
                    self.nocc, self.nvirt, label="t_1"
                )
                end = self.nocc * self.nvirt
            doubles = self.denself.create_four_index(
                self.nocc, self.nvirt, self.nocc, self.nvirt, label="t_2"
            )

            if self.singles:
                singles.assign(new, end2=(self.nocc * self.nvirt))
                self._amplitudes.append(singles)
                # for checkpointing
                self.checkpoint.update("t_1", singles)
            self.assign_doubles(doubles, new, end)
            self._amplitudes.append(doubles)
            # for checkpointing
            self.checkpoint.update("t_2", doubles)
        elif isinstance(new, list):
            self._amplitudes = new

    @amplitudes.deleter
    def amplitudes(self):
        del self._amplitudes[:]
        self._amplitudes = []

    @property
    def l_amplitudes(self):
        """The lambda amplitudes"""
        return self._l_amplitudes

    @l_amplitudes.setter
    def l_amplitudes(self, new):
        """Update singles and doubles lambda amplitudes in
           self.l_amplitudes[S,D] as well as in the IOData container used for
           checkpointing
        """
        if isinstance(new, np.ndarray) or isinstance(new, OneIndex):
            if self._l_amplitudes:
                raise NonEmptyData(
                    "Warning: List of LCC Lambda amplitudes not empty!"
                )

            end = 0
            if self.singles:
                singles = self.lf.create_two_index(
                    self.nocc, self.nvirt, label="l_1"
                )
                end = self.nocc * self.nvirt
            doubles = self.denself.create_four_index(
                self.nocc, self.nvirt, self.nocc, self.nvirt, label="l_2"
            )

            if self.singles:
                singles.assign(new, end2=(self.nocc * self.nvirt))
                self._l_amplitudes.append(singles)
                # for checkpointing
                self.checkpoint.update("l_1", singles)
            self.assign_doubles(doubles, new, end)
            self._l_amplitudes.append(doubles)
            # for checkpointing
            self.checkpoint.update("l_2", doubles)
        elif isinstance(new, list):
            self._l_amplitudes = new

    @l_amplitudes.deleter
    def l_amplitudes(self):
        del self._l_amplitudes[:]
        self._l_amplitudes = []

    @property
    def checkpoint(self):
        """The iodata container that contains all data dump to disk"""
        return self._checkpoint

    @checkpoint.setter
    def checkpoint(self, new):
        self._checkpoint = new

    @property
    def checkpoint_fn(self):
        """The filename that will be dumped to disk"""
        return self._checkpoint_fn

    @checkpoint_fn.setter
    def checkpoint_fn(self, new):
        if self.singles:
            self._checkpoint_fn = f"checkpoint_{self.get_model()}{new}.h5"
        else:
            self._checkpoint_fn = f"checkpoint_{self.get_model()}{new}.h5"

    @property
    def l(self):
        """The lambda amplitudes"""
        return self._l

    @l.setter
    def l(self, args):
        if args is not None:
            self._l = args

    @property
    def freeze(self):
        """The freezed linearized coupled cluster amplitudes"""
        return self._freeze

    @freeze.setter
    def freeze(self, args):
        self._freeze = args

    @property
    def tco(self):
        """The tensor contraction operation used"""
        return self._tco

    @tco.setter
    def tco(self, args):
        if args is not None:
            self._tco = args

    @property
    def e_ref(self):
        """Reference energy"""
        return self._e_ref

    @e_ref.setter
    def e_ref(self, new):
        self._e_ref = new

    @property
    def e_core(self):
        """Core energy"""
        return self._e_core

    @e_core.setter
    def e_core(self, new):
        self._e_core = new

    @property
    def solver(self):
        """Reference energy"""
        return self._solver

    @solver.setter
    def solver(self, new):
        self._solver = new

    @property
    def jacobian_approximation(self):
        return self._jacobian_approximation

    @jacobian_approximation.setter
    def jacobian_approximation(self, new):
        self._jacobian_approximation = new

    def get_ndm(self, select):
        """Get a n-particle density matrix."""
        if not select in self.cache:
            raise ArgumentError(f"Density matrix {select} not found.")
        return self.cache.load(select)

    one_dm_pccd_pp = PropertyHelper(
        get_ndm,
        "one_dm_pccd_pp",
        "Diagonal 1-RDM for alpha/beta spin for pCCD",
    )
    two_dm_pccd_ppqq = PropertyHelper(
        get_ndm, "two_dm_pccd_ppqq", "2-RDM for alpha/beta spin for pCCD"
    )
    two_dm_pccd_pqpq = PropertyHelper(
        get_ndm, "two_dm_pccd_pqpq", "2-RDM for alpha/beta spin for pCCD"
    )
    one_dm_pp = PropertyHelper(
        get_ndm, "one_dm_pp", "Diagonal 1-RDM for alpha/beta spin"
    )
    one_dm_pq = PropertyHelper(
        get_ndm, "one_dm_pq", "1-RDM for alpha/beta spin"
    )
    two_dm_pPPp = PropertyHelper(
        get_ndm, "two_dm_pPPp", "2-RDM for alpha/beta spin"
    )
    two_dm_pqqp = PropertyHelper(
        get_ndm, "two_dm_pqqp", "2-RDM for alpha/beta spin"
    )
    two_dm_pQQp = PropertyHelper(
        get_ndm, "two_dm_pQQp", "2-RDM for alpha/beta spin"
    )
    two_dm_qQQp = PropertyHelper(
        get_ndm, "two_dm_qQQp", "2-RDM for alpha/beta spin"
    )
    two_dm_pQPp = PropertyHelper(
        get_ndm, "two_dm_pQPp", "2-RDM for alpha/beta spin"
    )
    two_dm_qQPq = PropertyHelper(
        get_ndm, "two_dm_qQPq", "2-RDM for alpha/beta spin"
    )
    two_dm_qQPp = PropertyHelper(
        get_ndm, "two_dm_qQPp", "2-RDM for alpha/beta spin"
    )
    two_dm_pQPq = PropertyHelper(
        get_ndm, "two_dm_pQPq", "2-RDM for alpha/beta spin"
    )
    two_dm_qPPp = PropertyHelper(
        get_ndm, "two_dm_qPPp", "2-RDM for alpha/beta spin"
    )
    three_dm_qPQQPp = PropertyHelper(
        get_ndm, "three_dm_qPQQPp", "3-RDM for alpha/beta spin"
    )
    three_dm_qpPPpq = PropertyHelper(
        get_ndm, "three_dm_qpPPpq", "3-RDM for alpha/beta spin"
    )
    four_dm_pPqQQqPp = PropertyHelper(
        get_ndm, "four_dm_pPqQQqPp", "4-RDM for alpha/beta spin"
    )

    def init_ndm(self, select):
        """Initialize a n-RDM

           **Arguments**

           select
                Type of n-RDM initialized (str).
        """
        alloc_size = {
            "one_dm_pp": (self.lf.create_one_index, self.nbasis),
            "one_dm_pq": (self.lf.create_two_index, self.nbasis),
            "two_dm_pPPp": (self.lf.create_one_index, self.nbasis),
            "two_dm_pqqp": (self.lf.create_two_index, self.nbasis),
            "two_dm_pQQp": (self.lf.create_two_index, self.nbasis),
            "two_dm_qQQp": (self.lf.create_two_index, self.nbasis),
            "two_dm_pQPp": (self.lf.create_two_index, self.nbasis),
            "two_dm_qQPq": (self.lf.create_two_index, self.nbasis),
            "two_dm_qQPp": (self.lf.create_two_index, self.nbasis),
            "two_dm_pQPq": (self.lf.create_two_index, self.nbasis),
            "two_dm_qPPp": (self.lf.create_two_index, self.nbasis),
            "three_dm_qPQQPp": (self.lf.create_two_index, self.nbasis),
            "three_dm_qpPPpq": (self.lf.create_two_index, self.nbasis),
            "four_dm_pPqQQqPp": (self.lf.create_two_index, self.nbasis),
        }
        alloc_method = {
            "one_dm": self.compute_1dm,
            "two_dm": self.compute_2dm,
            "three_dm": self.compute_3dm,
            "four_dm": self.compute_4dm,
        }
        for key in alloc_method:
            if key in select:
                method = alloc_method[key]
        dm, new = self.cache.load(select, alloc=alloc_size[select], tags="d",)
        if not new:
            raise RuntimeError(
                f"The matrix {select} already exists. Call clear prior updating DMs."
            )

        return dm, method

    def update_ndm(self, select, option, *args):
        cached_ndm, method = self.init_ndm(select)
        method(
            cached_ndm, self.l_amplitudes, self.amplitudes, 1.0, option, *args,
        )

    def init_ndm_pccd(self, select):
        """Initialize a n-RDM of pCCD

           **Arguments**

           select
                Type of n-RDM initialized (str).
        """
        alloc_size = {
            "one_dm_pccd_pp": (self.lf.create_one_index, self.nbasis),
            "two_dm_pccd_ppqq": (self.lf.create_two_index, self.nbasis),
            "two_dm_pccd_pqpq": (self.lf.create_two_index, self.nbasis),
        }
        alloc_method = {
            "one_dm": self.assign_pccd_one_dm,
            "two_dm": self.assign_pccd_two_dm,
        }
        for key in alloc_method:
            if key in select:
                method = alloc_method[key]
        dm, new = self.cache.load(select, alloc=alloc_size[select], tags="d",)

        return dm, method

    def update_ndm_pccd(self, select, option, *args):
        cached_ndm, method = self.init_ndm_pccd(select)
        method(
            cached_ndm, option, *args,
        )

    def assign_pccd_one_dm(self, cached_dm, select, *args):
        cached_dm.assign(unmask("dm_1", *args))
        cached_dm.iadd(-1.0, end0=self.nocc)

    def assign_pccd_two_dm(self, cached_dm, select, *args):
        pccd_2dm = unmask("dm_2", *args)
        cached_dm.assign(pccd_2dm[select])

        if select in ["pqpq"]:
            cached_dm.iadd(-1.0, end0=self.nocc, end1=self.nocc)

    def get_guess(self, select="mp2", *args):
        """Generate inital guess for cluster amplitudes"""
        if select in ["mp2"]:
            level = log.level
            if log.do_medium:
                log("Performing an MP2 guess.")
                log.hline("~")
            # still print warnings
            log.level = 1
            # set e_ref to 0 to get rid of warning
            mp2 = RMP2(self.lf, self.occ_model, ncore=self.ncore)
            mp2_ = mp2(
                args[0],
                args[1],
                args[2],
                IOData(e_core=self.e_core),
                singles=self.singles,
                natorb=False,
                relaxation=False,
                e_ref=0.0,
            )
            mp2.clear()
            log.level = level
            if log.do_medium:
                log("Resuming CC calculation.")
                log.hline("~")
            if self.singles:
                return np.hstack(
                    (
                        mp2_.t_1._array.ravel(order="C"),
                        mp2_.t_2._array.ravel(order="C"),
                    )
                )
            else:
                return mp2_.t_2._array.ravel(order="C")
        elif select in ["random"]:
            doubles = (
                np.random.rand(self.nocc * self.nvirt, self.nocc * self.nvirt)
                * 0.1
            )
            if self.singles:
                singles = np.random.rand(self.nocc * self.nvirt) * 0.1
                return np.hstack((singles.ravel(), doubles.ravel()))
            else:
                return doubles.ravel()
        else:
            raise UnkownOption(f"Unknown guess {select}")

    def do_restart(self, restart, olp, orbs, onlyt=False):
        """Collect data for restart"""
        data = IOData.from_file(restart)
        t_2 = unmask("t_2", data)
        if t_2 is None:
            raise RestartError("Cannot find CC T2 amplitudes in restart file.")
        if self.singles:
            t_1 = unmask("t_1", data)
            if t_1 is None:
                raise RestartError(
                    "Cannot find CC T1 amplitudes in restart file."
                )
            guess = np.hstack((t_1._array.ravel(), t_2._array.ravel()))
        else:
            guess = t_2._array.ravel()
        t_p = unmask("t_p", data)
        if t_p is not None:
            self.checkpoint.update("t_p", t_p)
        if onlyt:
            if log.do_medium:
                log("Restarting from amplitudes read from file.")
            if not orbs:
                raise ArgumentError(
                    "Cannot find orbitals in LCC function call."
                )
        else:
            if hasattr(data, "olp") and hasattr(data, "orb_a"):
                if isinstance(olp, TwoIndex) and isinstance(orbs, Orbital):
                    project_orbitals(data.olp, olp, data.orb_a, orbs)
                else:
                    olp = data.olp
                    orbs = data.orb_a
            self.checkpoint.update("olp", olp)
            self.checkpoint.update("orb_a", orbs.copy())
            if log.do_medium:
                log.warn(
                    f"Restarting from orbitals read from file, while Hamiltonian is not"
                    f" provided from restart file. Make sure that you know what you are"
                    f" doing."
                )
        if hasattr(data, "e_ref") and hasattr(data, "e_core"):
            self.e_ref = data.e_ref
            self.e_core = data.e_core
        elif hasattr(data, "e_ref") or hasattr(data, "e_core"):
            raise ArgumentError(
                "Either missing e_ref or e_core in restart file. Both have to be "
                "provided."
            )
        return guess, olp, orbs, t_p

    def clear(self):
        """Clear all wavefunction information"""
        self._cache.clear()

    def clear_aux_matrix(self):
        """Clear the auxiliary matrices"""
        self._cache.clear(tags="m", dealloc=True)

    def clear_dm(self):
        """Clear RDM information"""
        self._cache.clear(tags="d", dealloc=True)

    def calculate_aux_matrix(self, mo1, mo2, args, **kwargs):
        """Compute auxiliary matrices

           **Arguments:**

           mo1, mo2
                One- and two-electron integrals (some Hamiltonian matrix
                elements) in the MO basis.

           args
                List of arguements. Only geminal coefficients are used.
        """
        self.clear_aux_matrix()
        return self.update_aux_matrix(mo1, mo2, args, **kwargs)

    def get_aux_matrix(self, select):
        """Get a matrix.

           **Arguments:**

           select
                'fock'.
        """
        if not select in self.cache:
            raise NotImplementedError
        return self.cache.load(select)

    def check_result(self, **kwargs):
        """Check if amplitudes are reasonable."""
        thresh = 1e-6
        if self.singles:
            t2 = self.amplitudes[1]._array
        else:
            t2 = self.amplitudes[0]._array

        if self.multireference:
            # check if diagonal amplitudes are zero:
            if np.einsum("abab", t2) > thresh:
                raise ConsistencyError(
                    "Diagonal cluster amplitudes too big. Aborting optimization!"
                )

    def fix_guess(self, guess):
        """Set diagonal of double amplitudes to Zero. This avoids any errors
           induced by user-specified guess

           **Arguements:**

           guess:
                The initial guess. A numpy ndarray
        """
        if self.singles:
            singles = np.reshape(
                guess[: (self.nocc * self.nvirt)], (self.nocc, self.nvirt)
            )
            doubles = np.reshape(
                guess[(self.nocc * self.nvirt) :],
                (self.nocc, self.nvirt, self.nocc, self.nvirt),
            )
        else:
            doubles = np.reshape(
                guess, (self.nocc, self.nvirt, self.nocc, self.nvirt)
            )

        if self.multireference:
            #
            # set pairs to zero
            #
            ind1, ind2 = np.indices((self.nocc, self.nvirt))
            indices = [ind1, ind2, ind1, ind2]
            doubles[tuple(indices)] = 0.0
        #
        # symmetrize
        #
        doubles = (doubles + doubles.transpose(2, 3, 0, 1)) / 2.0

        indtriu = np.triu_indices(self.nocc * self.nvirt)
        doubles = doubles.reshape(
            self.nocc * self.nvirt, self.nocc * self.nvirt
        )[indtriu]
        if self.singles:
            return np.hstack((singles.ravel(), doubles.ravel()))
        else:
            return doubles.ravel()

    def check_input(self, **kwargs):
        """Check input parameters."""
        for name, value in kwargs.items():
            check_options(
                name,
                name,
                "e_ref",
                "tthreshold",
                "maxiter",
                "guess",
                "skipmo",
                "deletemo",
                "tco",
                "guesstype",
                "memoryopt",
                "solver",
                "diismax",
                "ethreshold",
                "freeze",
                "checkpoint_fn",
                "diisstart",
                "jacobian",
                "indextrans",
                "diisreset",
                "restart",
                "restart_t",
                "l",
            )
        guess = kwargs.get("guess", None)
        solver = kwargs.get("solver", "krylov")
        check_options("solver", solver, "krylov", "pbqn")
        if guess is not None:
            check_type("guess", guess, np.ndarray)
            if self.singles:
                dim = (
                    self.nocc * self.nocc * self.nvirt * self.nvirt
                    + self.nocc * self.nvirt
                )
            else:
                dim = self.nocc * self.nocc * self.nvirt * self.nvirt
            if guess.size != dim:
                raise ConsistencyError(
                    "Number of unknowns in initial guess is wrong!"
                )
        guesstype = kwargs.get("guesstype", "mp2")
        check_options("guesstype", guesstype, "mp2", "random")
        diismax = kwargs.get("diismax", 2)
        if diismax < 0:
            raise UnkownOption("At least one diis vector has to be used!")
        diisstart = kwargs.get("diisstart", 0)
        if diisstart < 0:
            raise UnkownOption(
                "First DIIS iteration has to be larger than or equal to 0!"
            )
        jacobian = kwargs.get("jacobian", 1)
        check_type("jacobian", jacobian, int)
        if jacobian not in [1, 2]:
            raise UnkownOption("Jacobian option can only have values 1 or 2!")
        indextrans = kwargs.get("indextrans", "tensordot")
        if indextrans not in ["einsum", "tensordot", "blas"]:
            raise UnkownOption("Choise of 4-index transformation unknown!")
        diisreset = kwargs.get("diisreset", True)
        check_options("diisreset", diisreset, True, False, 0, 1)
        restart = kwargs.get("restart", False)
        restart_t = kwargs.get("restart_t", False)
        if restart and restart_t:
            raise ArgumentError(
                f"PyBEST only supports one restart option. "
                f"Provide either restart are restart_t."
            )
        l = kwargs.get("l", False)
        check_type("lambda", l, int, bool)

    def dump_eri(self, eri, checkpoint_fn="checkpoint_eri.h5"):
        """Dump orbitals for restart

           **Arguments:**

           orb
                An expansion instance. AO/MO coefficients stored to disk

           olp
                A TwoIndex instance. The AO overlap matrix. Required for restart
                at different molecular geometry.

           checkpoint_fn
                The filename of the checkpoint file.
        """
        #
        # Use the pybest.io interface to write the checkpoint file
        #
        filename = filemanager.temp_path(checkpoint_fn)
        #
        # dump
        #
        mol = IOData(eri=eri)
        mol.to_file(filename)

    def print_info(self, **kwargs):
        """Print optimization options.
        """
        tthresh = kwargs.get("tthreshold", 1e-6)
        ethresh = kwargs.get("ethreshold", 1e-6)
        maxiter = kwargs.get("maxiter", 200)
        memopt = kwargs.get("memoryopt", False)
        guesstype = kwargs.get("guesstype", "mp2")
        solver = kwargs.get("solver", "krylov")
        jacobian = kwargs.get("jacobian", 1)
        indextrans = kwargs.get("indextrans", "tensordot")
        restart = kwargs.get("restart", None)
        restart_t = kwargs.get("restart_t", None)
        if restart:
            guesstype = "read from " + restart
        if restart_t:
            guesstype = "read from " + restart_t

        log.hline()
        if self.multireference:
            log("Linearized Coupled Cluster correction on top of RpCCD")
        else:
            log("Linearized Coupled Cluster correction on top of RHF")
        log.hline()
        log("OPTIMIZATION PARAMETERS:")
        if self.singles:
            log(f"{'Cluster Operator:':<30} T1+T2")
        else:
            log(f"{'Cluster Operator:':<30} T2")
        log(f"{'Number of occupied orbitals:':<30} {self.nocc}")
        log(f"{'Number of frozen cores:':<30} {self.ncore}")
        log(f"{'Number of virtuals:':<30} {self.nvirt}")
        log(f"{'Initial guess:':<30} {guesstype}")
        log("Solvers:")
        log(f"{'  cluster amplitudes:':<30} {solver}")
        if solver in ["pbqn"]:
            log(f"{'  jacobian approximation:':<30} {jacobian}")
        log("Optimization thresholds:")
        log(f"{'  cluster amplitudes:':<30} {tthresh:1.2e}")
        if solver == "pbqn":
            log(f"{'  energy:':<30} {ethresh:1.2e}")
        log(f"{'  maxiter:':<30} {maxiter:1.2e}")
        log(f"{'4-index transformation:':<30} {indextrans}")
        log(f"{'Memory optimization:':<30} {memopt}")
        log(f"{'Tensor contraction operations:':<30} {self.tco}")
        log.hline()

    def print_energy(self):
        if log.do_medium:
            log(f"{'E_ref:':<20} {self.e_ref:16.8f} a.u.")
            log.hline("-")
            log(
                f"{'E_(Seniority 2):':<20} {self.energy['e_corr_s2']:16.8f} a.u."
            )
            log(
                f"{'E_(Seniority 4):':<20} {self.energy['e_corr_s4']:16.8f} a.u."
            )
            log.hline("-")
            log(f"{'E_corr(D):':<20} {self.energy['e_corr_d']:16.8f} a.u.")
            if self.singles:
                log(f"{'E_corr(S):':<20} {self.energy['e_corr_s']:16.8f} a.u.")
            log.hline("-")
            log(f"{'E_tot:':<20} {self.energy['e_tot']:16.8f} a.u.")

    def calculate_energy(self, t=None):
        """Calculate energy contribution and update checkpoint info"""
        if t is not None:
            del self.amplitudes
            self.amplitudes = t
        #
        # Get active orbital indices
        #
        iocc = self.nocc
        iend = self.nbasis

        govvo = self.get_aux_matrix("govvo")
        if self.singles:
            fock = self.get_aux_matrix("fock")
            t1 = self.amplitudes[0]
        t2 = self.amplitudes[-1]

        #
        # E_singles = sum_md F_md t_md
        #
        ov2 = self.get_range("ov", start=2)
        e_corr = 0.0
        if self.singles:
            esingles = 2.0 * t1.contract("ab,ab", fock, **ov2)
            e_corr += esingles

        #
        # E_doubles = sum_mkde L_mkde t_mdke
        #
        edoubles = t2.contract("abcd,adbc", govvo) * 2.0
        edoubles -= t2.contract("abcd,abdc", govvo)
        e_corr += edoubles
        #
        # Seniority-2 sector
        #
        e_seniority_2 = t2.contract("abac,acba", govvo)
        e_seniority_2 += t2.contract("abcb,abbc", govvo)
        #
        # Seniority-4 sector
        #
        e_seniority_4 = edoubles - e_seniority_2

        if self.singles:
            return {
                "e_tot": e_corr + self.e_ref,
                "e_corr": e_corr,
                "e_corr_d": edoubles,
                "e_corr_s2": e_seniority_2,
                "e_corr_s4": e_seniority_4,
                "e_corr_s": esingles,
                "e_ref": self.e_ref,
                "e_core": self.e_core,
            }
        else:
            return {
                "e_tot": e_corr + self.e_ref,
                "e_corr": e_corr,
                "e_corr_d": edoubles,
                "e_corr_s2": e_seniority_2,
                "e_corr_s4": e_seniority_4,
                "e_ref": self.e_ref,
                "e_core": self.e_core,
            }

    def get_range(self, string, start=0):
        """ Returns dictionary with keys beginX, endX, begin(X+1), etc.
            *  string - any sequence of 'o' (occupied) and 'v' (virtual)
        """
        range_ = {}
        ind = start
        for char in string:
            if char == "o":
                range_["begin" + str(ind)] = 0
                range_["end" + str(ind)] = self.nocc
            elif char == "v":
                range_["begin" + str(ind)] = self.nocc
                range_["end" + str(ind)] = self.nbasis
            elif char == "V":
                range_["begin" + str(ind)] = 0
                range_["end" + str(ind)] = self.nvirt
            elif char == "n":
                range_["begin" + str(ind)] = 0
                range_["end" + str(ind)] = self.nbasis
            else:
                raise ArgumentError(
                    f"Do not know how to handle choice {char}."
                )
            ind += 1
        return range_

    def assign_doubles(self, d, t, begin):
        """Assign upper triangular block of doubles amplitudes to full (dense)
           amplitude tensor.
        """
        d.assign_triu(t, begin4=begin)
        if self.multireference:
            d.iadd_transpose((2, 3, 0, 1))
        else:
            tp_ = d.slice("abab->ab")
            d.iadd_transpose((2, 3, 0, 1))
            ind1, ind2 = np.indices((self.nocc, self.nvirt))
            indp = [ind1, ind2, ind1, ind2]
            # assign diagonal
            d.assign(tp_, indp)

    @timer.with_section("LCC")
    def __call__(self, *args, **kwargs):
        """Optimize cluster amplitudes and determine energy correction to some
           reference wavefunction.

           Cluster amplitudes are determined by finding the root of a set of
           equations of the form A.x+b = 0

           Currently supported wavefunction models (Psi_0):
            * RHF
            * RpCCD

           Currently supported cluster operators:
            * Doubles with occupied-virtual separation (RLCCD)
            * Singles and Doubles with occupied-virtual separation (RLCCSD)

           **Arguments:**

           args
                One- and two-body integrals (some Hamiltonian matrix elements)
                expressed in the AO basis, the AO/MO coefficients and the
                geminal coefficients (if Psi_0 = pCCD). Order of arguments does
                not matter.

           **Keywords:**
                Contains reference energy and solver specific input parameters:
                 * e_ref:        reference energy (default value nan)
                 * tthreshold:  tolerance for amplitudes (default 1e-6)
                 * ethreshold:  tolerance for energy (default 1e-7); only
                                used in PBQN solver
                 * maxiter:     maximum number of iterations (default 200)
                 * guess:       initial guess for amplitudes, if not provided,
                                an MP2 guess will be generated

        """
        #
        # Set default keyword arguements to class attributes:
        #
        self.e_ref = unmask("e_tot", *args, **kwargs) or unmask(
            "e_ref", *args, **kwargs
        )
        #
        # Core energy is NOT passed as kwargs as LCC uses e_ref.
        # e_core can only be passed from an IOData container of a previous calculation.
        # e_core is only used when defining an active space.
        # Otherwise it is useless and only passed for reasons of consistency
        #
        self.e_core = unmask("e_core", *args, **kwargs)
        self.tco = kwargs.get("tco", "td")
        self.freeze = kwargs.get("freeze", None)
        self.solver = kwargs.get("solver", "krylov")
        self.checkpoint_fn = kwargs.get("checkpoint_fn", "")
        self.jacobian_approximation = kwargs.get("jacobian", 1)
        self.l = kwargs.get("l", False)
        #
        # Set remaining default keyword arguements:
        #
        tthresh = kwargs.get("tthreshold", 1e-6)
        ethresh = kwargs.get("ethreshold", 1e-7)
        maxiter = kwargs.get("maxiter", 200)
        initguess = kwargs.get("guess", None)
        skipmo = kwargs.get("skipmo", False)
        memopt = kwargs.get("memoryopt", True)
        guesstype = kwargs.get("guesstype", "mp2")
        diismax = kwargs.get("diismax", 2)
        diisstart = kwargs.get("diisstart", 0)
        indextrans = kwargs.get("indextrans", "tensordot")
        diisreset = kwargs.get("diisreset", True)
        restart = kwargs.get("restart", False)
        restart_t = kwargs.get("restart_t", False)

        #
        # Clear everything
        #
        del self.amplitudes
        del self.energy
        if log.do_medium:
            log.hline("=")
            log(" ")
            log("Entering LCC module")
            log(" ")
        #
        # Unmask arguments
        #
        fargs = []
        t_p = None
        # olp
        olp = unmask("olp", *args, **kwargs)
        self.checkpoint.update("olp", olp)
        # orb
        orbs = unmask_orb(*args, **kwargs)
        if orbs:
            orbs = orbs[0]
            self.checkpoint.update("orb_a", orbs.copy())
        # t_p
        if self.multireference:
            t_p = unmask("t_p", *args, **kwargs)
            if t_p is not None:
                self.checkpoint.update("t_p", t_p)
        # 1-e ints and 2-e ints
        one = self.lf.create_two_index(label="one")
        for arg in args:
            if isinstance(arg, TwoIndex):
                if arg.label in OneBodyHamiltonian:
                    one.iadd(arg)
            elif isinstance(arg, FourIndex):
                two = arg
                # Check if lf and two are of the same type
                if isinstance(two, CholeskyFourIndex):
                    if not isinstance(self.lf, CholeskyLinalgFactory):
                        raise ArgumentError(
                            "LinalgFactory and eri are of different type."
                        )
                elif isinstance(self.lf, CholeskyLinalgFactory):
                    if not isinstance(two, CholeskyFourIndex):
                        raise ArgumentError(
                            "LinalgFactory and eri are of different type."
                        )
        #
        # Print some initial information, method specific:
        #
        if log.do_medium:
            self.print_info(**kwargs)

        #
        # Check input parameters:
        #
        self.check_input(**kwargs)

        #
        # Get initial guess or do restart:
        #
        if restart:
            guess, olp, orbs, t_p = self.do_restart(restart, olp, orbs)
        elif restart_t:
            guess, olp, orbs, t_p2 = self.do_restart(
                restart_t, olp, orbs, True
            )
            # Update pairs if required
            if self.multireference and t_p2 is not None:
                t_p = tp2
                # checkpoint stores the pointer of t_p, so update should be redundant
                self.checkpoint.update("t_p", t_p)
        elif initguess is not None:
            guess = initguess
        else:
            guess = self.get_guess(guesstype, one, two, orbs)
        #
        # Append to function arguments to work with old implementation
        #
        fargs.append(orbs)
        #
        # Check if Tp amplitudes are there if required
        #
        if self.multireference and t_p is None:
            raise ArgumentError(
                "Cannot find pCCD amplitudes in LCC function call."
            )
        else:
            fargs.append(t_p)
        if olp is None:
            raise ArgumentError(
                "Cannot find overlap integrals in LCC function call."
            )
        #
        # For pCCD: Set diagonal amplitudes to zero if we start with a 'wrong'
        # guess
        # Else: symmetrize doubles
        guess = self.fix_guess(guess)

        #
        # Transform integrals:
        #
        if skipmo:
            mo1 = one
            mo2 = two
        else:
            if self.ncore > 0:
                cas = split_core_active(
                    one,
                    two,
                    orbs,
                    e_core=self.e_core,
                    ncore=self.ncore,
                    nactive=self.nbasis,
                    indextrans=indextrans,
                )
                mo1 = cas.one
                mo2 = cas.two
                self.e_core = cas.e_core
            else:
                ti = transform_integrals(one, two, orbs, indextrans=indextrans)
                mo1 = ti.one[0]
                mo2 = ti.two[0]
            if memopt:
                self.dump_eri(two)
                two.__del__()
        #
        # Update for checkpointing:
        #
        self.checkpoint.update("e_core", self.e_core)

        #
        # Construct auxilary matrices:
        #
        matrix = self.calculate_aux_matrix(mo1, mo2, t_p, **kwargs)

        #
        # Append arguements, used as arguments in root finding:
        #
        try:
            for mat in matrix:
                fargs.append(mat)
        except:
            pass

        #
        # Update amplitudes and energies for first iteration:
        #
        del self.amplitudes
        self.amplitudes = guess
        self.energy = self.calculate_energy()
        #
        # Solve matrix equataion using scipy.optimize.root routine:
        #
        if self.solver == "krylov":
            if log.do_medium:
                log(
                    f"{'residual 2-norm':>18} {'residual max norm':>20} "
                    f"{'Ecorr':>12} {'Etot':>18} {'D(E)':>14}"
                )
            amplitudes = opt.root(
                self.vfunction,
                guess,
                args=(fargs),
                method="krylov",
                callback=self.callback,
                options={"fatol": tthresh, "maxiter": maxiter},
            )
            if amplitudes.success:
                del self.amplitudes
                self.amplitudes = amplitudes.x
                if log.do_medium:
                    log(
                        f"Optimization of cluster amplitudes converged in "
                        f"{amplitudes.nit} iterations."
                    )
                    log(" ")
                    log("Calculating energy correction:")
            else:
                #
                # Update amplitudes anyway for restart possibility
                #
                del self.amplitudes
                self.amplitudes = amplitudes.x
                log.warn(
                    f"program terminated. Error in solving amplitude equations: "
                    f"{amplitudes.message}"
                )
            #
            # overwrite for Lambda equations
            #
            if self.l:
                guess = amplitudes.x
            else:
                del guess
        elif self.solver == "pbqn":
            from pybest.solvers.pbquasinewton import PBQausiNewton

            qn = PBQausiNewton(
                self.lf,
                maxiter=maxiter,
                ethreshold=ethresh,
                tthreshold=tthresh,
            )
            e, t = qn(
                self,
                guess,
                fargs,
                diismax=diismax,
                diisstart=diisstart,
                suffix=self.checkpoint_fn,
                diisreset=diisreset,
            )
            del self.amplitudes
            self.amplitudes = t
            #
            # overwrite for Lambda equations
            #
            if self.l:
                guess = t._array

        #
        # Get correction to reference energy:
        #
        self.energy = self.calculate_energy()

        #
        # Print some output information for user:
        #
        self.print_energy()

        #
        # Make sanity checks. If somethings is wrong, abort calculation:
        #
        self.check_result(**kwargs)

        #
        # If required, calculate lambda equations
        #
        if self.l == True:
            log.hline("=")
            log(" ")
            log("Entering Lambda Equation")
            log(" ")
            if log.do_medium:
                log(f"{'residual 2-norm':>18} {'residual max norm':>20}")
            l_amplitudes = opt.root(
                self.vector_lagrange,
                guess,
                args=(fargs),
                method="krylov",
                callback=self.callback_l,
                options={"fatol": tthresh, "maxiter": maxiter},
            )

            if l_amplitudes.success:
                del self.l_amplitudes
                self.l_amplitudes = l_amplitudes.x

                # print l_amplitudesx
                if log.do_medium:
                    log(
                        f"Optimization of cluster amplitudes converged in "
                        f"{l_amplitudes.nit} iterations."
                    )

                ###  DENSITY MATRICES   ###
                self.clear_dm()
                if self.multireference:
                    self.update_ndm_pccd("one_dm_pccd_pp", "pp", *args)
                    self.update_ndm_pccd("two_dm_pccd_ppqq", "ppqq", *args)
                    self.update_ndm_pccd("two_dm_pccd_pqpq", "pqpq", *args)

                self.update_ndm("one_dm_pp", "pp", *fargs)
                self.update_ndm("one_dm_pq", "pq", *fargs)
                self.update_ndm("two_dm_pPPp", "pPPp", *fargs)
                self.update_ndm("two_dm_pqqp", "pqqp", *fargs)
                self.update_ndm("two_dm_pQQp", "pQQp", *fargs)
                self.update_ndm("two_dm_qQQp", "qQQp", *fargs)
                self.update_ndm("two_dm_pQPp", "pQPp", *fargs)
                self.update_ndm("two_dm_qQPq", "qQPq", *fargs)
                self.update_ndm("two_dm_qQPp", "qQPp", *fargs)
                self.update_ndm("two_dm_pQPq", "pQPq", *fargs)
                self.update_ndm("two_dm_qPPp", "qPPp", *fargs)
                self.update_ndm("three_dm_qPQQPp", "qPQQPp", *fargs)
                self.update_ndm("three_dm_qpPPpq", "qpPPpq", *fargs)
                self.update_ndm("four_dm_pPqQQqPp", "pPqQQqPp", *fargs)

                ###IO DATA Updating RDM's ([pCCD - HF], [LCCSD + HF])
                if self.multireference:
                    self.checkpoint.update(
                        "dm_1",
                        {
                            "pccd_pp": self.one_dm_pccd_pp,
                            "pp": self.one_dm_pp,
                            "pq": self.one_dm_pq,
                        },
                    )
                    self.checkpoint.update(
                        "dm_2",
                        {
                            "pccd_ppqq": self.two_dm_pccd_ppqq,
                            "pccd_pqpq": self.two_dm_pccd_pqpq,
                            "pPPp": self.two_dm_pPPp,
                            "pqqp": self.two_dm_pqqp,
                            "pQQp": self.two_dm_pQQp,
                            "qQQp": self.two_dm_qQQp,
                            "pQPp": self.two_dm_pQPp,
                            "qQPq": self.two_dm_qQPq,
                            "qQPp": self.two_dm_qQPp,
                            "pQPq": self.two_dm_pQPq,
                            "qPPp": self.two_dm_qPPp,
                        },
                    )
                else:
                    self.checkpoint.update(
                        "dm_1", {"pp": self.one_dm_pp, "pq": self.one_dm_pq}
                    )
                    self.checkpoint.update(
                        "dm_2",
                        {
                            "pPPp": self.two_dm_pPPp,
                            "pqqp": self.two_dm_pqqp,
                            "pQQp": self.two_dm_pQQp,
                            "qQQp": self.two_dm_qQQp,
                            "pQPp": self.two_dm_pQPp,
                            "qQPq": self.two_dm_qQPq,
                            "qQPp": self.two_dm_qQPp,
                            "pQPq": self.two_dm_pQPq,
                            "qPPp": self.two_dm_qPPp,
                        },
                    )
                self.checkpoint.update(
                    "dm_3",
                    {
                        "qPQQPp": self.three_dm_qPQQPp,
                        "qpPPpq": self.three_dm_qpPPpq,
                    },
                )
                self.checkpoint.update(
                    "dm_4", {"pPqQQqPp": self.four_dm_pPqQQqPp}
                )

            else:
                #
                # Update amplitudes anyway for restart possibility
                #
                del self.l_amplitudes
                self.l_amplitudes = l_amplitudes.x
                log.warn(
                    f"Error in solving amplitude equations: {l_amplitudes.message}"
                )

        #
        # Clear auxiliary matrices from Cache to save memory
        #
        self.clear_aux_matrix()

        #
        # Do Checkpoint
        #
        self.checkpoint.to_file(self.checkpoint_fn)

        #
        # Read integrals back in
        #
        if memopt:
            eri = IOData.from_file(filemanager.temp_path("checkpoint_eri.h5"))
            two.replace_array(eri.eri)

        if log.do_medium:
            log(" ")
            log.hline("=")
        return self.checkpoint()

    def callback(self, x, f):
        norm2 = linalg.norm(f)
        normmax = linalg.norm(f, ord=np.inf)
        #
        # Get current amplitudes
        #
        del self.amplitudes
        self.amplitudes = x
        #
        # Get previous energies
        #
        e_old = self.energy["e_tot"]
        #
        # Get current energy
        #
        self.energy = self.calculate_energy()
        if log.do_medium:
            e_corr = self.energy["e_corr"]
            e_tot = self.energy["e_tot"]
            log(
                f"{norm2:18.12f} {normmax:18.12f} {e_corr:18.12f} "
                f"{e_tot:18.12f} {(e_old - e_tot):12.3e}"
            )
        #
        # Dump current solution
        #
        self.checkpoint.to_file(self.checkpoint_fn)

    def callback_l(self, x, f):
        norm2 = linalg.norm(f)
        normmax = linalg.norm(f, ord=np.inf)
        #
        #
        # Lagrange_multipliers
        del self.l_amplitudes
        self.l_amplitudes = x
        if log.do_medium:
            log(f"{norm2:18.12f} {normmax:18.12f}")
        #
        # Dump current solution
        #
        self.checkpoint.to_file(self.checkpoint_fn)

    ###############################################################################################
    ############################   Density Matrix LCCSD   #########################################
    ###############################################################################################
    def compute_1dm(
        self, one_dm_out, l_amplitudes, amplitudes, factor, select, *args
    ):

        if self.singles:
            lag_mult1 = l_amplitudes[0]
            cluster_ampli1 = amplitudes[0]
        lag_mult2 = l_amplitudes[-1]
        cluster_ampli2 = amplitudes[-1]
        end = self.nocc + self.nvirt
        # assign Tp to T2
        if self.multireference:
            tp = unmask("t_p", *args)
            ind1, ind2 = np.indices((self.nocc, self.nvirt))
            indices = [ind1, ind2, ind1, ind2]
            cluster_ampli2.assign(tp, indices)

        check_options("select", select, "pp", "pq")
        if select == "pp":
            #
            # Calculate occupied block
            #
            tmp_occ = self.lf.create_one_index(self.nocc)
            # 1
            if self.singles:
                cluster_ampli1.contract(
                    "ab,ab->a", lag_mult1, tmp_occ, factor=-0.5
                )
            # 2a
            cluster_ampli2.contract(
                "abcd,abcd->c", lag_mult2, tmp_occ, factor=-0.5
            )
            #
            #
            # Calculate virtual block
            #
            tmp_virt = self.lf.create_one_index(self.nvirt)
            # 1
            if self.singles:
                cluster_ampli1.contract(
                    "ab,ab->b", lag_mult1, tmp_virt, factor=0.5
                )
            # 2a
            cluster_ampli2.contract(
                "abcd,abcd->b", lag_mult2, tmp_virt, factor=0.5
            )
            #
            # Combine both blocks and scale
            #
            one_dm_out.assign(tmp_occ, 0, self.nocc)
            one_dm_out.assign(tmp_virt, self.nocc, self.nbasis)
            # dpi contribution
            one_dm_out.iadd(1.0, end0=self.nocc)

            one_dm_out.iscale(factor)
        elif select == "pq":

            oo = self.lf.create_two_index(self.nocc, self.nocc)
            vv = self.lf.create_two_index(self.nvirt, self.nvirt)
            vo = self.lf.create_two_index(self.nvirt, self.nocc)
            ov = self.lf.create_two_index(self.nocc, self.nvirt)
            #
            # Calculate occupied block
            #
            # rho(ij)
            # 1 -tic.ljc
            if self.singles:
                cluster_ampli1.contract(
                    "ab,cb->ac", lag_mult1, oo, factor=-0.5, clear=True
                )
            # 2a -tmeif . lmejf
            cluster_ampli2.contract(
                "abcd,abed->ce", lag_mult2, oo, factor=-0.5, clear=False
            )
            #
            # Calculate virtual block
            # rho(ab)
            # 1 lka tkb
            if self.singles:
                lag_mult1.contract(
                    "ab,ac->bc", cluster_ampli1, vv, factor=0.5, clear=True
                )
            # 2a lmanf . tmbnf
            lag_mult2.contract("abcd,aecd->be", cluster_ampli2, vv, factor=0.5)
            #
            # vo blocks
            #
            # 3a lia
            if self.singles:
                vo.clear()
                vo.iadd_t(lag_mult1, 0.5)
                lag_mult2.contract(
                    "abcd,cd->ba", cluster_ampli1, vo, factor=0.5
                )
            #
            # ov blocks
            #
            if self.singles:
                # 4a tia
                ov.assign(cluster_ampli1)
                # 4b tianf . lnf
                cluster_ampli2.contract(
                    "abcd,cd->ab", lag_mult1, ov, factor=1.0
                )
                # 4c -tmaif . lmf
                cluster_ampli2.contract(
                    "abcd,ad->cb", lag_mult1, ov, factor=-0.5
                )
                if self.multireference:
                    # 4d -lifmf . tma . cif -> iaf . cif
                    tmp = lag_mult2.contract(
                        "abcb,ce->aeb",
                        cluster_ampli1,
                        out=None,
                        factor=1.0,
                        clear=True,
                    )
                    tmp.contract("abc,ac->ab", tp, ov, factor=-0.5)
                    # 4e -lnena . tie . cna -> ian . cna
                    tmp = lag_mult2.contract(
                        "abac,eb->eca",
                        cluster_ampli1,
                        out=None,
                        factor=1.0,
                        clear=True,
                    )
                    tmp.contract("abc,cb->ab", tp, ov, factor=-0.5)
                    # 4f ljame . tme . cja -> ja . cja
                    tmp = lag_mult2.contract(
                        "abcd,cd->ab",
                        cluster_ampli1,
                        out=None,
                        factor=1.0,
                        clear=True,
                    )
                    ov.iadd_mult(tp, tmp, 0.5)

            # one_dm_out = self.lf.create_two_index(self.nbasis, self.nbasis)

            one_dm_out.iadd(oo, 1.0, 0, self.nocc, 0, self.nocc)
            one_dm_out.iadd(
                vv, 1.0, self.nocc, self.nbasis, self.nocc, self.nbasis
            )
            one_dm_out.iadd(vo, 1.0, self.nocc, self.nbasis, 0, self.nocc)
            one_dm_out.iadd(ov, 1.0, 0, self.nocc, self.nocc, self.nbasis)
            # dpq contribution
            one_dm_out.iadd_diagonal(1.0, end0=self.nocc)
            # scale
            one_dm_out.iscale(factor)
        # remove Tp to T2
        if self.multireference:
            ind1, ind2 = np.indices((self.nocc, self.nvirt))
            indices = [ind1, ind2, ind1, ind2]
            cluster_ampli2.assign(0.0, indices)

    def compute_2dm(
        self, two_dm_out, l_amplitudes, amplitudes, factor, select, *args
    ):

        if self.singles:
            lag_mult1 = l_amplitudes[0]
            cluster_ampli1 = amplitudes[0]
        lag_mult2 = l_amplitudes[-1]
        cluster_ampli2 = amplitudes[-1]
        tmp_vv = self.lf.create_two_index(self.nvirt, self.nvirt)
        tmp_oo = self.lf.create_two_index(self.nocc, self.nocc)
        tmp_ov = self.lf.create_two_index(self.nocc, self.nvirt)
        tmp_vo = self.lf.create_two_index(self.nvirt, self.nocc)

        # assign Tp to T2
        if self.multireference:
            tp = unmask("t_p", *args)
            ind1, ind2 = np.indices((self.nocc, self.nvirt))
            indices = [ind1, ind2, ind1, ind2]
            cluster_ampli2.assign(tp, indices)

        check_options(
            "select",
            select,
            "pPPp",
            "pqqp",
            "pQQp",
            "qQQp",
            "pQPp",
            "qQPq",
            "qQPp",
            "pQPq",
            "qPPp",
        )

        if select == "pPPp":

            tmp_occ = self.lf.create_one_index(self.nocc)
            tmp_virt = self.lf.create_one_index(self.nvirt)
            #
            # Calculate occupied block
            #
            cluster_ampli2.contract(
                "abac,abac->a", lag_mult2, tmp_occ, factor=0.5, clear=True
            )
            #
            # Calculate virtual block
            #
            cluster_ampli2.contract(
                "abcb,abcb->b", lag_mult2, tmp_virt, factor=0.5, clear=True
            )
            #
            # Lower dimensional contributions
            #
            # ii + II
            if self.singles:
                cluster_ampli1.contract(
                    "ab,ab->a", lag_mult1, tmp_occ, factor=-1.0
                )
            cluster_ampli2.contract(
                "abcd,abcd->c", lag_mult2, tmp_occ, factor=-1.0
            )
            #
            # HF
            #
            two_dm_out.iadd(1.0, end0=self.nocc)
            #
            # Combine both blocks and scale
            #
            two_dm_out.iadd(tmp_occ, end0=self.nocc)
            two_dm_out.iadd(tmp_virt, begin0=self.nocc)

        if select == "pqqp":

            f = 1.0 / 6.0
            ###1
            # a
            cluster_ampli2.contract(
                "abcd,abcd->ca", lag_mult2, tmp_oo, factor=f
            )
            # b
            cluster_ampli2.contract(
                "abcd,adcb->ca", lag_mult2, tmp_oo, factor=-f
            )

            ###2
            # a
            if self.singles:
                cluster_ampli1.contract(
                    "ab,ab->ab", lag_mult1, tmp_ov, factor=-0.5,
                )
            # b
            cluster_ampli2.contract(
                "abcd,abcd->ab", lag_mult2, tmp_ov, factor=-0.5
            )
            # c
            cluster_ampli2.contract(
                "abcd,abcd->cb", lag_mult2, tmp_ov, factor=-f
            )
            # d
            cluster_ampli2.contract(
                "abcd,adcb->cb", lag_mult2, tmp_ov, factor=f
            )

            ###3
            # a
            if self.singles:
                cluster_ampli1.contract(
                    "ab,ab->ba", lag_mult1, tmp_vo, factor=-0.5
                )
            # b
            cluster_ampli2.contract(
                "abcd,abcd->ba", lag_mult2, tmp_vo, factor=-0.5
            )
            # c
            cluster_ampli2.contract(
                "abcd,abcd->da", lag_mult2, tmp_vo, factor=-f
            )
            # d
            cluster_ampli2.contract(
                "abcd,adcb->da", lag_mult2, tmp_vo, factor=f
            )

            ###4
            # a
            cluster_ampli2.contract(
                "abcd,abcd->db", lag_mult2, tmp_vv, factor=f
            )
            # b
            cluster_ampli2.contract(
                "abcd,adcb->db", lag_mult2, tmp_vv, factor=-f
            )

            ############################################################################
            #########--------------------------HF-------------------------##############
            ############################################################################
            two_dm_out.iadd(1.0, 1.0, end0=self.nocc, end1=self.nocc)

            ############################################################################
            ##############-------------------lower RDM's-----------------------#########
            ############################################################################
            ### ii + jj + aa + bb

            # 1 i/j
            # a
            tmp2 = cluster_ampli2.contract(
                "abcd,abcd->c", lag_mult2, out=None, factor=-0.5
            )
            # b
            if self.singles:
                cluster_ampli1.contract(
                    "ab,ab->a", lag_mult1, out=tmp2, factor=-0.5
                )

            # 2 a/b
            # a
            tmp4 = cluster_ampli2.contract(
                "abcd,abcd->b", lag_mult2, out=None, factor=0.5
            )
            # b
            if self.singles:
                cluster_ampli1.contract(
                    "ab,ab->b", lag_mult1, out=tmp4, factor=0.5
                )
            #
            # Combine all blocks and scale
            two_dm_out.iadd(tmp_oo, 1.0, 0, self.nocc, 0, self.nocc)
            two_dm_out.iadd(tmp_vo, 1.0, self.nocc, self.nbasis, 0, self.nocc)
            two_dm_out.iadd(
                tmp_vv, 1.0, self.nocc, self.nbasis, self.nocc, self.nbasis
            )
            two_dm_out.iadd(tmp_ov, 1.0, 0, self.nocc, self.nocc, self.nbasis)
            # add ii/jj to [o,o] block
            two_dm_out.iadd(tmp2, 1.0, 0, self.nocc, 0, self.nocc)
            two_dm_out.iadd_t(tmp2, 1.0, 0, self.nocc, 0, self.nocc)
            # add aa/bb to [o,v]/[v,o] block
            two_dm_out.iadd_t(tmp4, 1.0, 0, self.nocc, self.nocc, self.nbasis)
            two_dm_out.iadd(tmp4, 1.0, self.nocc, self.nbasis, 0, self.nocc)
            two_dm_out.iscale(factor)

        if select == "pQQp":
            f = 1.0 / 6.0

            ###1
            # a
            cluster_ampli2.contract(
                "abcd,abcd->ca", lag_mult2, tmp_oo, factor=2 * f
            )

            # b
            cluster_ampli2.contract(
                "abcd,adcb->ca", lag_mult2, tmp_oo, factor=f
            )

            ###2
            # a
            cluster_ampli2.contract(
                "abcd,abcd->cb", lag_mult2, tmp_ov, factor=-2 * f
            )

            # b
            cluster_ampli2.contract(
                "abcd,adcb->cb", lag_mult2, tmp_ov, factor=-f
            )

            ###3
            # a
            cluster_ampli2.contract(
                "abcd,abcd->da", lag_mult2, tmp_vo, factor=-2 * f
            )

            # b
            cluster_ampli2.contract(
                "abcd,adcb->da", lag_mult2, tmp_vo, factor=-f
            )

            ###4
            # a
            cluster_ampli2.contract(
                "abcd,abcd->db", lag_mult2, tmp_vv, factor=2 * f
            )

            # b
            cluster_ampli2.contract(
                "abcd,cbad->db", lag_mult2, tmp_vv, factor=f
            )

            ############################################################################
            ##############--------------------------HF-------------------------#########
            ############################################################################
            two_dm_out.iadd(1.0, 1.0, end0=self.nocc, end1=self.nocc)
            ############################################################################
            ##############-------------------lower RDM's-----------------------#########
            ############################################################################

            ### ii + jj + aa + bb +HF

            # 1 i/j
            # a
            tmp2 = cluster_ampli2.contract(
                "abcd,abcd->c", lag_mult2, out=None, factor=-0.5
            )
            # b
            if self.singles:
                cluster_ampli1.contract(
                    "ab,ab->a", lag_mult1, out=tmp2, factor=-0.5
                )

            # 2 a/b
            # a
            tmp4 = cluster_ampli2.contract(
                "abcd,abcd->b", lag_mult2, out=None, factor=0.5
            )
            # b
            if self.singles:
                cluster_ampli1.contract(
                    "ab,ab->b", lag_mult1, out=tmp4, factor=0.5
                )

            # Combine all blocks and scale
            two_dm_out.iadd(tmp_oo, 1.0, 0, self.nocc, 0, self.nocc)
            two_dm_out.iadd(tmp_ov, 1.0, 0, self.nocc, self.nocc, self.nbasis)
            two_dm_out.iadd(
                tmp_vv, 1.0, self.nocc, self.nbasis, self.nocc, self.nbasis
            )
            two_dm_out.iadd(tmp_vo, 1.0, self.nocc, self.nbasis, 0, self.nocc)
            # add ii/jj to [o,o] block
            two_dm_out.iadd(tmp2, 1.0, 0, self.nocc, 0, self.nocc)
            two_dm_out.iadd_t(tmp2, 1.0, 0, self.nocc, 0, self.nocc)
            # add aa/bb to [o,v]/[v,o] block
            two_dm_out.iadd_t(tmp4, 1.0, 0, self.nocc, self.nocc, self.nbasis)
            two_dm_out.iadd(tmp4, 1.0, self.nocc, self.nbasis, 0, self.nocc)

            two_dm_out.iscale(factor)

        if select in ["qQQp", "qQPq"]:

            f = 1.0 / 6.0

            ###1
            cluster_ampli2.contract(
                "abcd,adcd->db", lag_mult2, tmp_vv, factor=0.5
            )

            ###2
            cluster_ampli2.contract(
                "abac,abdc->ad", lag_mult2, tmp_oo, factor=0.5
            )

            if self.singles:
                ###3

                # a
                cluster_ampli2.contract(
                    "abac,ac->ab", lag_mult1, tmp_ov, factor=-0.5
                )

                # b
                if self.multireference:
                    tmp = lag_mult2.contract(
                        "abac,ac->ab", cluster_ampli1, out=None, factor=1.0
                    )
                    tmp_ov.iadd_mult(tmp, tp, 0.5)

                    # c
                    tmp = lag_mult2.contract(
                        "abcb,cd->abd", cluster_ampli1, out=None, factor=1.0
                    )
                    tmp.contract(
                        "abc,ab->ac", tp, tmp_ov, factor=0.5, clear=False
                    )

                    # d
                    tmp = lag_mult2.contract(
                        "abcd,ab->cd", cluster_ampli1, out=None, factor=1.0
                    )
                    tmp_ov.iadd_mult(tmp, tp, -0.5, False)

                ###4
                lag_mult2.contract(
                    "abcb,cb->ba", cluster_ampli1, tmp_vo, factor=0.5
                )

            ### NO HF ###
            ############################################################################
            ##############------------------lower RDM's------------------------#########
            ############################################################################
            ### ji +ja

            # 1 ji
            # a
            if self.singles:
                cluster_ampli1.contract(
                    "ab,cb->ac", lag_mult1, tmp_oo, factor=-0.5
                )
            # b -tmeif . lmejf
            cluster_ampli2.contract(
                "abcd,abed->ce", lag_mult2, tmp_oo, factor=-0.5
            )

            # 2 ja
            # a
            if self.singles:
                tmp_ov.iadd(cluster_ampli1)
                # b tianf . lnf
                cluster_ampli2.contract(
                    "abcd,cd->ab", lag_mult1, tmp_ov, factor=1.0
                )
                # c -tmaif . lmf
                cluster_ampli2.contract(
                    "abcd,ad->cb", lag_mult1, tmp_ov, factor=-0.5
                )
                if self.multireference:
                    # d -lifmf . tma . cif -> iaf . cif
                    tmp = lag_mult2.contract(
                        "abcb,ce->aeb",
                        cluster_ampli1,
                        out=None,
                        factor=1.0,
                        clear=True,
                    )
                    tmp.contract("abc,ac->ab", tp, tmp_ov, factor=-0.5)
                    # e -lnena . tie . cna -> ian . cna
                    tmp = lag_mult2.contract(
                        "abac,eb->eca",
                        cluster_ampli1,
                        out=None,
                        factor=1.0,
                        clear=True,
                    )
                    tmp.contract("abc,cb->ab", tp, tmp_ov, factor=-0.5)
                    # f ljame . tme . cja ->  ja . cja
                    tmp = lag_mult2.contract(
                        "abcd,cd->ab",
                        cluster_ampli1,
                        out=None,
                        factor=1.0,
                        clear=True,
                    )
                    tmp_ov.iadd_mult(tmp, tp, 0.5)

            ###
            # Combine all blocks and scale
            two_dm_out.iadd(tmp_oo, 1.0, 0, self.nocc, 0, self.nocc)
            two_dm_out.iadd(tmp_vo, 1.0, self.nocc, self.nbasis, 0, self.nocc)
            two_dm_out.iadd(
                tmp_vv, 1.0, self.nocc, self.nbasis, self.nocc, self.nbasis
            )
            two_dm_out.iadd(tmp_ov, 1.0, 0, self.nocc, self.nocc, self.nbasis)

            two_dm_out.iscale(factor)

        if select in ["pQPp", "qPPp"]:

            ov = self.lf.create_two_index(self.nocc, self.nvirt)
            vo = self.lf.create_two_index(self.nvirt, self.nocc)

            ###1
            cluster_ampli2.contract(
                "abcd,abad->ca", lag_mult2, tmp_oo, factor=0.5
            )

            ###2
            cluster_ampli2.contract(
                "abcb,abcd->db", lag_mult2, tmp_vv, factor=0.5
            )

            ###3
            if self.singles:
                lag_mult2.contract(
                    "abac,ab->ca", cluster_ampli1, tmp_vo, factor=-0.5
                )

            ###4
            # a
            if self.singles:
                cluster_ampli2.contract(
                    "abcb,cb->ab", lag_mult1, tmp_ov, factor=0.5
                )
                if self.multireference:
                    # b lnane,je->naj, naj,na->aj
                    tmp = lag_mult2.contract(
                        "abac,dc->abd", cluster_ampli1, out=None, factor=1.0
                    )
                    tmp.contract("abc,ab->cb", tp, tmp_ov, factor=-0.5)
                    # c -ljama . tma . cif -> iaf . cif
                    tmp = lag_mult2.contract(
                        "abcb,cb->ab", cluster_ampli1, out=None, factor=1.0
                    )
                    tmp_ov.iadd_mult(tmp, tp, -0.5, False)
                    # d lmeja,me->ja
                    lag_mult2.contract(
                        "abcd,ab->cd",
                        cluster_ampli1,
                        ov,
                        factor=1.0,
                        clear=True,
                    )
                    tmp_ov.iadd_mult(ov, tp, 0.5, False)

            ### NO HF ###
            ############################################################################
            ##############-----------------Lower RDM's-------------------------#########
            ############################################################################
            # ji +bi
            # 1 ji
            # a
            if self.singles:
                cluster_ampli1.contract(
                    "ab,cb->ac", lag_mult1, tmp_oo, factor=-0.5
                )
            # b
            cluster_ampli2.contract(
                "abcd,abed->ce", lag_mult2, tmp_oo, factor=-0.5
            )

            if self.singles:
                # 2 bi
                # a
                tmp_vo.iadd_t(lag_mult1, 0.5)
                # b
                lag_mult2.contract(
                    "abcd,cd->ba", cluster_ampli1, tmp_vo, factor=0.5
                )

            ###
            # Combine all blocks and scale
            two_dm_out.iadd(tmp_oo, 1.0, 0, self.nocc, 0, self.nocc)
            two_dm_out.iadd(tmp_vo, 1.0, self.nocc, self.nbasis, 0, self.nocc)
            two_dm_out.iadd(
                tmp_vv, 1.0, self.nocc, self.nbasis, self.nocc, self.nbasis
            )
            two_dm_out.iadd(tmp_ov, 1.0, 0, self.nocc, self.nocc, self.nbasis)
            two_dm_out.iscale(factor)

        if select == "pQPq":
            f = 1.0 / 6.0

            ###1
            # a
            cluster_ampli2.contract(
                "abcd,adcb->db", lag_mult2, tmp_vv, factor=2.0 * f
            )
            # b
            cluster_ampli2.contract(
                "abcd,abcd->db", lag_mult2, tmp_vv, factor=f
            )

            ###2
            # a
            cluster_ampli2.contract(
                "abcd,cbad->ca", lag_mult2, tmp_oo, factor=2.0 * f
            )
            # b
            cluster_ampli2.contract(
                "abcd,abcd->ca", lag_mult2, tmp_oo, factor=f
            )

            ###3
            if self.singles:
                # a
                tmp_ov.iadd_mult(cluster_ampli1, lag_mult1, 0.5)
            # b
            cluster_ampli2.contract(
                "abcd,abcd->ab", lag_mult2, tmp_ov, factor=0.5
            )
            # c
            cluster_ampli2.contract(
                "abcd,cbad->cb", lag_mult2, tmp_ov, factor=-2.0 * f
            )
            # d
            cluster_ampli2.contract(
                "abcd,abcd->cb", lag_mult2, tmp_ov, factor=-f
            )

            ### NO HF ###

            # Combine all blocks and scale
            two_dm_out.iadd(tmp_oo, 1.0, 0, self.nocc, 0, self.nocc)
            two_dm_out.iadd(
                tmp_vv, 1.0, self.nocc, self.nbasis, self.nocc, self.nbasis
            )
            two_dm_out.iadd(tmp_ov, 1.0, 0, self.nocc, self.nocc, self.nbasis)
            two_dm_out.iadd_t(
                tmp_ov, 1.0, self.nocc, self.nbasis, 0, self.nocc
            )
            two_dm_out.iscale(factor)

        if select == "qQPp":
            ###1
            cluster_ampli2.contract(
                "abcb,adcd->db", lag_mult2, tmp_vv, factor=0.5
            )

            ###2
            cluster_ampli2.contract(
                "abac,dbdc->ad", lag_mult2, tmp_oo, factor=0.5
            )

            ###3
            if self.singles:
                # a
                if self.multireference:
                    fac = tp.copy()
                    fac.imul(lag_mult1)
                    fac.imul(cluster_ampli1)
                    tmp_ov.iadd(fac, 1.0)
                    # b
                    tmp = cluster_ampli1.contract(
                        "ab,ab->a", lag_mult1, out=None, factor=1.0
                    )
                    tmp_1 = tp.contract("ab,a->ab", tmp, factor=-1.0)
                    tmp_ov.iadd(tmp_1)
                    # c
                    tmp = cluster_ampli1.contract(
                        "ab,ab->b", lag_mult1, out=None, factor=1.0
                    )
                    tmp_1 = tp.contract("ab,b->ab", tmp, out=None, factor=-1.0)
                    tmp_ov.iadd(tmp_1)
            if self.multireference:
                # d
                tmp = cluster_ampli2.contract(
                    "abcd,abcd->ab", lag_mult2, out=None, factor=1.0,
                )
                tmp_ov.iadd_mult(tp, tmp, 1.0)
                # e
                tmp = cluster_ampli2.contract(
                    "abcd,abcd->ad", lag_mult2, out=None, factor=1.0,
                )
                tmp_ov.iadd_mult(tp, tmp, 1.0)
                # f
                tmp = cluster_ampli2.contract(
                    "abcd,abcd->a", lag_mult2, out=None, factor=1.0
                )
                tmp_1 = tp.contract("ab,a->ab", tmp, out=None, factor=-1.0)
                tmp_ov.iadd(tmp_1)
                # g
                tmp = lag_mult2.contract(
                    "abcb,ab->ac", tp, out=None, factor=1.0
                )
                cluster_ampli2.contract(
                    "abcb,ac->ab", tmp, tmp_ov, factor=-1.0
                )
                # h
                tmp = cluster_ampli2.contract(
                    "abcd,abcd->b", lag_mult2, out=None, factor=1.0
                )
                tmp_1 = tp.contract("ab,b->ab", tmp, out=None, factor=-1.0)
                tmp_ov.iadd(tmp_1)
                # i
                tmp = lag_mult2.contract(
                    "abac,ab->bc", tp, out=None, factor=1.0
                )
                cluster_ampli2.contract(
                    "abac,bc->ab", tmp, tmp_ov, factor=-1.0
                )

                # j
                tmp = cluster_ampli2.contract(
                    "abcb,adcd->bd", lag_mult2, out=None, factor=1.0,
                )
                tmp.contract("ab,cb->ca", tp, tmp_ov, factor=0.5)
                # k
                tmp = cluster_ampli2.contract(
                    "abac,dbdc->ad", lag_mult2, out=None, factor=1.0
                )
                tmp.contract("ab,bc->ac", tp, tmp_ov, factor=0.5)
            ### NO HF ###

            # Combine all blocks and scale
            two_dm_out.iadd(tmp_oo, 1.0, 0, self.nocc, 0, self.nocc)
            two_dm_out.iadd(
                tmp_vv, 1.0, self.nocc, self.nbasis, self.nocc, self.nbasis
            )
            two_dm_out.iadd(tmp_ov, 1.0, 0, self.nocc, self.nocc, self.nbasis)
            two_dm_out.iscale(factor)

        # clear Tp from T2
        if self.multireference:
            ind1, ind2 = np.indices((self.nocc, self.nvirt))
            indices = [ind1, ind2, ind1, ind2]
            cluster_ampli2.assign(0.0, indices)

    def compute_3dm(
        self, three_dm_out, l_amplitudes, amplitudes, factor, select, *args
    ):
        if self.singles:
            lag_mult1 = l_amplitudes[0]
            cluster_ampli1 = amplitudes[0]
        lag_mult2 = l_amplitudes[-1]
        cluster_ampli2 = amplitudes[-1]
        tmp_vv = self.lf.create_two_index(self.nvirt, self.nvirt)
        tmp_oo = self.lf.create_two_index(self.nocc, self.nocc)
        tmp_ov = self.lf.create_two_index(self.nocc, self.nvirt)
        tmp_vo = self.lf.create_two_index(self.nvirt, self.nocc)

        check_options("select", select, "qPQQPp", "qpPPpq")
        # assign Tp to T2
        if self.multireference:
            tp = unmask("t_p", *args)
            ind1, ind2 = np.indices((self.nocc, self.nvirt))
            indices = [ind1, ind2, ind1, ind2]
            cluster_ampli2.assign(tp, indices)

        if select in ["qPQQPp"]:

            ov = self.lf.create_two_index(self.nocc, self.nvirt)
            if self.singles:
                if self.multireference:
                    ###1
                    # a
                    tmp_ov.iadd_mult(lag_mult1, tp, -0.5)

                    # b
                    lag_mult2.contract(
                        "abcb,ab->cb", cluster_ampli1, ov, factor=1.0
                    )
                    tmp_ov.iadd_mult(ov, tp, 0.5)

                    # c
                    lag_mult2.contract(
                        "abac,ac->ab",
                        cluster_ampli1,
                        ov,
                        factor=1.0,
                        clear=True,
                    )
                    tmp_ov.iadd_mult(ov, tp, 0.5)

                    # d
                    lag_mult2.contract(
                        "abcd,cd->ab",
                        cluster_ampli1,
                        ov,
                        factor=1.0,
                        clear=True,
                    )
                    tmp_ov.iadd_mult(ov, tp, -0.5)
            ############################################################################
            ##############------------------lower RDM's------------------------#########
            ############################################################################
            ###   jAAa dqj + bBBi dpi + jIIi dqj + jJJi dpi + ji dqj dpi

            if self.singles:
                ###1 bBBi
                lag_mult2.contract(
                    "abcb,cb->ba", cluster_ampli1, tmp_vo, factor=0.5
                )

                ###2 jAAa
                # a
                cluster_ampli2.contract(
                    "abcb,cb->ab", lag_mult1, tmp_ov, factor=0.5
                )
                # b lnane,je->naj, naj,na->aj
                if self.multireference:
                    tmp = lag_mult2.contract(
                        "abac,dc->abd", cluster_ampli1, out=None, factor=1.0
                    )
                    tmp.contract("abc,ab->cb", tp, tmp_ov, factor=-0.5)
                    # c -ljama . tma . cif -> iaf . cif
                    tmp = lag_mult2.contract(
                        "abcb,cb->ab", cluster_ampli1, out=None, factor=1.0
                    )
                    tmp_ov.iadd_mult(tmp, tp, -0.5)
                    # d lmeja,me->ja
                    lag_mult2.contract(
                        "abcd,ab->cd",
                        cluster_ampli1,
                        ov,
                        factor=1.0,
                        clear=True,
                    )
                    tmp_ov.iadd_mult(ov, tp, 0.5)

            ###3 jIIi
            # a
            cluster_ampli2.contract(
                "abcd,cbcd->ac", lag_mult2, tmp_oo, factor=0.5
            )

            ###4 jJJi
            # a
            cluster_ampli2.contract(
                "abac,abdc->ad", lag_mult2, tmp_oo, factor=0.5
            )

            ###5 ji
            if self.singles:
                # a
                cluster_ampli1.contract(
                    "ab,cb->ac", lag_mult1, tmp_oo, factor=-0.5
                )
            # b
            cluster_ampli2.contract(
                "abcd,abed->ce", lag_mult2, tmp_oo, factor=-0.5
            )

            ###
            # Combine all blocks and scale
            three_dm_out.iadd(tmp_oo, 1.0, 0, self.nocc, 0, self.nocc)
            three_dm_out.iadd(
                tmp_ov, 1.0, 0, self.nocc, self.nocc, self.nbasis
            )
            three_dm_out.iadd(
                tmp_vo, 1.0, self.nocc, self.nbasis, 0, self.nocc
            )
            three_dm_out.iscale(factor)

        if select in ["qpPPpq"]:

            f = 1.0 / 6.0
            # 1
            cluster_ampli2.contract(
                "abac,abac->ba", lag_mult2, tmp_vo, factor=0.5
            )
            # 2
            cluster_ampli2.contract(
                "abcb,abcb->ab", lag_mult2, tmp_ov, factor=-0.5
            )
            ############################################################################
            ##############--------------------------HF-------------------------#########
            ############################################################################
            # HF: dqj dpi
            three_dm_out.iadd(1.0, 1.0, end0=self.nocc, end1=self.nocc)

            ############################################################################
            ##############------------------lower RDM's------------------------#########
            ############################################################################
            ### aAAa dqj + ibbi dpi + iBBi dPI + bb dpidPI + IjjI dpi + ijji dPI+
            ### iIIi dqj + jj dpidPI + ii dPIdqj + II dpidqj

            ###1 ibbi
            # a
            if self.singles:
                cluster_ampli1.contract(
                    "ab,ab->ba", lag_mult1, tmp_vo, factor=-0.5
                )
            # b
            cluster_ampli2.contract(
                "abcd,abcd->ba", lag_mult2, tmp_vo, factor=-0.5
            )
            # c
            cluster_ampli2.contract(
                "abcd,abcd->da", lag_mult2, tmp_vo, factor=-f
            )
            # d
            cluster_ampli2.contract(
                "abcd,adcb->da", lag_mult2, tmp_vo, factor=f
            )

            ###2 iBBi
            # a
            cluster_ampli2.contract(
                "abcd,abcd->da", lag_mult2, tmp_vo, factor=-2 * f
            )
            # b
            cluster_ampli2.contract(
                "abcd,adcb->da", lag_mult2, tmp_vo, factor=-f
            )

            ###3 bb
            tmp_bb = cluster_ampli2.contract(
                "abcd,abcd->b", lag_mult2, out=None, factor=0.5
            )
            if self.singles:
                cluster_ampli1.contract(
                    "ab,ab->b", lag_mult1, out=tmp_bb, factor=0.5
                )

            ###4 aAAa
            tmp_a = cluster_ampli2.contract(
                "abcb,abcb->b", lag_mult2, out=None, factor=0.5, clear=True
            )

            ###5 IjjI
            # a
            cluster_ampli2.contract(
                "abcd,abcd->ca", lag_mult2, tmp_oo, factor=2 * f, clear=True
            )
            # b
            cluster_ampli2.contract(
                "abcd,adcb->ca", lag_mult2, tmp_oo, factor=f
            )

            ###6 ijji
            # a
            cluster_ampli2.contract(
                "abcd,abcd->ca", lag_mult2, tmp_oo, factor=f
            )
            # b
            cluster_ampli2.contract(
                "abcd,adcb->ca", lag_mult2, tmp_oo, factor=-f,
            )

            ###7 iIIi
            tmp_i = cluster_ampli2.contract(
                "abac,abac->a", lag_mult2, out=None, factor=0.5, clear=True
            )

            ###8 jj
            # a
            tmp_jj = cluster_ampli2.contract(
                "abcd,abcd->c", lag_mult2, out=None, factor=-0.5
            )
            # b
            if self.singles:
                cluster_ampli1.contract(
                    "ab,ab->a", lag_mult1, out=tmp_jj, factor=-0.5
                )
            ###9 ii + II
            if self.singles:
                cluster_ampli1.contract(
                    "ab,ab->a", lag_mult1, tmp_i, factor=-1.0
                )
            # b
            cluster_ampli2.contract(
                "abcd,abcd->c", lag_mult2, tmp_i, factor=-1.0
            )

            ###
            # Combine all blocks and scale

            three_dm_out.iadd(tmp_oo, 1.0, 0, self.nocc, 0, self.nocc)
            three_dm_out.iadd(
                tmp_ov, 1.0, 0, self.nocc, self.nocc, self.nbasis
            )
            three_dm_out.iadd(
                tmp_vo, 1.0, self.nocc, self.nbasis, 0, self.nocc
            )

            three_dm_out.iadd(tmp_jj, 1.0, 0, self.nocc, 0, self.nocc)
            three_dm_out.iadd_t(tmp_i, 1.0, 0, self.nocc, 0, self.nocc)

            # add to [o,v] block
            three_dm_out.iadd_t(
                tmp_a, 1.0, 0, self.nocc, self.nocc, self.nbasis
            )
            # add to [v,o] block
            three_dm_out.iadd(
                tmp_bb, 1.0, self.nocc, self.nbasis, 0, self.nocc
            )

            three_dm_out.iscale(factor)
        # clear Tp from T2
        if self.multireference:
            ind1, ind2 = np.indices((self.nocc, self.nvirt))
            indices = [ind1, ind2, ind1, ind2]
            cluster_ampli2.assign(0.0, indices)

    def compute_4dm(
        self, four_dm_out, l_amplitudes, amplitudes, factor, select, *args
    ):
        if self.singles:
            lag_mult1 = l_amplitudes[0]
            cluster_ampli1 = amplitudes[0]
        lag_mult2 = l_amplitudes[-1]
        cluster_ampli2 = amplitudes[-1]

        tmp_v = self.lf.create_one_index(self.nvirt)
        tmp_vv = self.lf.create_two_index(self.nvirt, self.nvirt)
        tmp_ov = self.lf.create_two_index(self.nocc, self.nvirt)
        tmp_vo = self.lf.create_two_index(self.nvirt, self.nocc)
        tmp_oo = self.lf.create_two_index(self.nocc, self.nocc)
        tmp = self.lf.create_two_index(self.nbasis)

        check_options("select", select, "pPqQQqPp")

        # assign Tp to T2
        if self.multireference:
            tp = unmask("t_p", *args)
            ind1, ind2 = np.indices((self.nocc, self.nvirt))
            indices = [ind1, ind2, ind1, ind2]
            cluster_ampli2.assign(tp, indices)

        if select == "pPqQQqPp":
            f = 1.0 / 6.0
            ############################################################################
            ##############--------------------------HF-------------------------#########
            ############################################################################
            four_dm_out.iadd(1.0, 1.0, end0=self.nocc, end1=self.nocc)
            ############################################################################
            ##############------------------lower RDM's------------------------#########
            ############################################################################
            ### aAJJAa dqj + aAjjAa dQJ + aAAa dqj dQJ + bBBb dpi dPI + ibBBbi dIP +
            ### IbBBbI dip + iIjjIi +iIJJIi + ijJJji + IjJJjI +
            ### iIIi djq + ijji dIP dQJ+ jJJj dip + IJJI dip dqj  + IjjI dip dQJ +
            ### iJJi dIP djq + ii dip dqj + II dip dqj + jj dip dqj + JJ dip dqj + HF dip dqj

            ###1
            # a aAJJAa +  aAjjAa
            cluster_ampli2.contract(
                "abcb,abcb->ab", lag_mult2, tmp_ov, factor=-1.0
            )  #
            # b aAAa
            tmp_a = cluster_ampli2.contract(
                "abcb,abcb->b", lag_mult2, out=None, factor=0.5
            )  #

            ###2 bBBb
            tmp_b = cluster_ampli2.contract(
                "abcb,abcb->b", lag_mult2, out=None, factor=0.5
            )

            ###3 ibBBbi
            cluster_ampli2.contract(
                "abcb,abcb->ba", lag_mult2, tmp_vo, factor=-0.5
            )

            ###4 IbBBbI
            cluster_ampli2.contract(
                "abcb,abcb->ba", lag_mult2, tmp_vo, factor=-0.5
            )

            ###5 iIjjIi = 0
            ###6 iIJJIi = 0
            ###7 ijJJji = 0
            ###8 IjJJjI = 0

            ###9 iIIi
            tmp_i = cluster_ampli2.contract(
                "abac,abac->a", lag_mult2, out=None, factor=0.5
            )

            ###10 ijji + IJJI
            # a
            cluster_ampli2.contract(
                "abcd,abcd->ca", lag_mult2, tmp_oo, factor=2 * f
            )
            # b
            cluster_ampli2.contract(
                "abcd,adcb->ca", lag_mult2, tmp_oo, factor=-2 * f
            )

            ###11 jJJj
            tmp_j = cluster_ampli2.contract(
                "abac,abac->a", lag_mult2, out=None, factor=0.5
            )

            ###12 IjjI + iJJi
            # a
            cluster_ampli2.contract(
                "abcd,abcd->ca", lag_mult2, tmp_oo, factor=4 * f
            )
            # b
            cluster_ampli2.contract(
                "abcd,adcb->ca", lag_mult2, tmp_oo, factor=2 * f
            )

            ###13 ii + II
            if self.singles:
                # a
                cluster_ampli1.contract(
                    "ab,ab->a", lag_mult1, tmp_i, factor=-1.0
                )
                # b
            cluster_ampli2.contract(
                "abcd,abcd->c", lag_mult2, tmp_i, factor=-1.0
            )

            ###14 jj + JJ
            if self.singles:
                # a
                cluster_ampli1.contract(
                    "ab,ab->a", lag_mult1, tmp_j, factor=-1.0
                )
            # b
            cluster_ampli2.contract(
                "abcd,abcd->c", lag_mult2, tmp_j, factor=-1.0
            )

            #
            # Combine all blocks and scale

            four_dm_out.iadd(tmp_j, 1.0, 0, self.nocc, 0, self.nocc)
            four_dm_out.iadd_t(tmp_i, 1.0, 0, self.nocc, 0, self.nocc)

            # add to [v,o] block
            four_dm_out.iadd(tmp_b, 1.0, self.nocc, self.nbasis, 0, self.nocc)
            # add to [o,v] block
            four_dm_out.iadd_t(
                tmp_a, 1.0, 0, self.nocc, self.nocc, self.nbasis
            )

            four_dm_out.iadd(tmp_ov, 1.0, 0, self.nocc, self.nocc, self.nbasis)
            four_dm_out.iadd(tmp_vo, 1.0, self.nocc, self.nbasis, 0, self.nocc)
            four_dm_out.iadd(tmp_oo, 1.0, 0, self.nocc, 0, self.nocc)

            four_dm_out.iscale(factor)
        # clear Tp from T2
        if self.multireference:
            ind1, ind2 = np.indices((self.nocc, self.nvirt))
            indices = [ind1, ind2, ind1, ind2]
            cluster_ampli2.assign(0.0, indices)

    def init_aux_matrix(
        self, select, dim0=None, dim1=None, dim2=None, dim3=None
    ):
        """Initialize auxiliary matrices

           **Arguments:**

           select
                One of ``fock``, ``ocjbc``, ``vcjkb``, ``ocjkb``, ``vcjbc``,
                ``vcij``, ``ocab``, ``ocjb``, ``vcjb``, ``vcjkl``,
                ``ocabc``, vcjbk``, ``ocajb``

           **Optional arguments:**

           dim0, dim1, dim2, dim3
                The dimension of the auxiliary matrices for specific axes.
        """
        check_options(
            "select",
            select,
            "fock",
            "ocjbc",
            "vcjkb",
            "ocjkb",
            "vcjbc",
            "vcij",
            "ocab",
            "ocjb",
            "vcjb",
            "vcjkl",
            "ocabc",
            "vcjbk",
            "ocajb",
            "govvo",
            "govov",
            "goooo",
            "gvvvv",
            "govvv",
            "gvooo",
            "gpqpq",
            "gpqrr",
        )
        dim0 = dim0 or self.nbasis
        dim1 = dim1 or self.nbasis
        dim2 = dim2 or self.nbasis
        dim3 = dim3 or self.nbasis

        if select in ["fock", "gpqpq", "vcij", "ocab", "ocjb", "vcjb"]:
            alloc_size = (self.lf.create_two_index, dim0, dim1)
        elif select in [
            "gpqrr",
            "ocjbc",
            "vcjbc",
            "vcjkl",
            "ocabc",
            "vcjbk",
            "ocajb",
            "vcjkb",
            "ocjkb",
        ]:
            alloc_size = (self.lf.create_three_index, dim0, dim1, dim2)
        elif select in ["govvo", "govov", "goooo", "gvvvv", "gvooo", "govvv"]:
            alloc_size = (
                self.denself.create_four_index,
                dim0,
                dim1,
                dim2,
                dim3,
            )

        matrix, new = self._cache.load(select, alloc=alloc_size, tags="m",)
        if not new:
            raise RuntimeError(
                f"The matrix {select} already exists. Call clear prior "
                f"to updating the wfn."
            )
        return matrix

    @timer.with_section("UpdatingAuxMat")
    def update_aux_matrix(self, mo1, mo2, cia, **kwargs):
        """Derive all auxiliary matrices.
           fock_pq:     one_pq + sum_m(2<pm|qm> - <pm|mq>),
           lint_pqrs:   <pq|rs> - <pq|sr>,
           oc_jbc:      sum_m(<mm|bc> c_jm^bc),
           vc_jkb:      sum_d(<dd|jk> c_jk^bd),
           vc_jbc:      sum_d(<dd|bc> c_j^d),
           oc_jkb:      sum_m(<mm|jk> c_m^b),
           dc_jb:       sum_md(<mm|dd> c_jm^bd),
           vc_ij:       sum_d <ij|dd> c_j^d,
           oc_ab:       sum_m <ab|mm> c_m^a,
           oc_jb:       sum_m <jb|mm> c_m^b,
           vc_jb:       sum_d <jb|dd> c_j^d,
           oc_abc:      sum_m <bc|mm> c_m^a,
           vc_jkl:      sum_d <jk|dd> c_l^d

           **Arguments:**

           mo1, mo2
                one- and two-electron integrals to be sorted.

           cia
                The geminal coefficients. A TwoIndex instance
        """
        #
        # Get ranges
        #
        oov = self.get_range("oov")
        vvo = self.get_range("vvo")
        ovo = self.get_range("ovo")
        ovv = self.get_range("ovv")
        vvv = self.get_range("vvv")
        ooo = self.get_range("ooo")
        voo = self.get_range("voo")
        ovvo = self.get_range("ovvo")
        ovov = self.get_range("ovov")
        oooo = self.get_range("oooo")
        vooo = self.get_range("vooo")
        ovvv = self.get_range("ovvv")
        vvvv = self.get_range("vvvv")
        if log.do_medium:
            log("Computing auxiliary matrices.")
        deletemo = kwargs.get("deletemo", True)
        tco = self.tco

        #
        # Inactive Fock matrix
        #
        auxmat0 = self.init_aux_matrix("fock")
        get_fock_matrix(auxmat0, mo1, mo2, self.nocc)
        if deletemo:
            mo1.__del__()
        #
        # <pq|pq>
        #
        gpqpq = self.init_aux_matrix("gpqpq")
        mo2.slice("abab->ab", gpqpq)
        #
        # pCCD reference function:
        #
        if self.multireference:
            # use 3-index intermediate (will be used several times)
            # This also works with Cholesky
            # tmpabc
            tmpabc = self.denself.create_three_index(self.nbasis)
            mo2.slice("abcc->abc", tmpabc)
            #
            # vc_ij = sum_d <ij|dd> c_j^d
            #
            auxmat1 = self.init_aux_matrix("vcij", self.nocc, self.nocc)
            tmpabc.contract("abc,bc->ab", cia, auxmat1, **oov)
            #
            # oc_ab = sum_m <ab|mm> c_m^a
            #
            auxmat2 = self.init_aux_matrix("ocab", self.nvirt, self.nvirt)
            tmpabc.contract("abc,ca->ab", cia, auxmat2, **vvo)
            #
            # oc_jb = sum_m <jb|mm> c_m^b
            #
            if self.singles:
                auxmat3 = self.init_aux_matrix("ocjb", self.nocc, self.nvirt)
                tmpabc.contract("abc,cb->ab", cia, auxmat3, **ovo)
            #
            # vc_jb = sum_d <jb|dd> c_j^d
            #
            if self.singles:
                auxmat4 = self.init_aux_matrix("vcjb", self.nocc, self.nvirt)
                tmpabc.contract("abc,ac->ab", cia, auxmat4, **ovv)
            #
            # oc_jbc = sum_m <mm|bc> c_jm^bc
            #
            tmp = self.denself.create_two_index(self.nvirt, self.nvirt)
            auxmat5 = self.init_aux_matrix(
                "ocjbc", self.nocc, self.nvirt, self.nvirt
            )
            tmpabc.contract("abc,cb->ab", cia, tmp, clear=True, **vvo)
            cia.contract("ab,bc->abc", tmp, auxmat5, select="einsum")
            # P_jm
            tmpabc.contract("abc,ca->ab", cia, tmp, clear=True, **vvo)
            cia.contract("ac,bc->abc", tmp, auxmat5, select="einsum")
            #
            # vc_jkb = sum_d <dd|jk> c_jk^bd
            #
            auxmat6 = self.init_aux_matrix(
                "vcjkb", self.nocc, self.nocc, self.nvirt
            )
            # tmp storage
            tmp = self.denself.create_two_index(self.nocc, self.nocc)
            tmpabc.contract("abc,bc->ab", cia, tmp, clear=True, **oov)
            tmp.contract("ab,ac->abc", cia, auxmat6, select="einsum")
            # P_jm
            tmpabc.contract("abc,ac->ab", cia, tmp, clear=True, **oov)
            tmp.contract("ab,bc->abc", cia, auxmat6, select="einsum")
            #
            # vc_jbc = sum_d <bc|dd> c_j^d
            #
            auxmat7 = self.init_aux_matrix(
                "vcjbc", self.nocc, self.nvirt, self.nvirt
            )
            tmpabc.contract("abc,dc->dab", cia, auxmat7, **vvv)
            #
            # oc_jkb = sum_m <mm|jk> c_m^b
            #
            auxmat8 = self.init_aux_matrix(
                "ocjkb", self.nocc, self.nocc, self.nvirt
            )
            tmpabc.contract("abc,cd->abd", cia, auxmat8, **ooo)
            #
            # vc_jkl = sum_d <jk|dd> c_l^d
            #
            auxmat9 = self.init_aux_matrix(
                "vcjkl", self.nocc, self.nocc, self.nocc
            )
            tmpabc.contract("abc,dc->abd", cia, auxmat9, **oov)
            #
            # oc_abc = sum_m <bc|mm> c_m^a
            #
            auxmat10 = self.init_aux_matrix(
                "ocabc", self.nvirt, self.nvirt, self.nvirt
            )
            tmpabc.contract("abc,cd->dab", cia, auxmat10, **vvo)
            if self.singles:
                #
                # vc_jbk = sum_d c_k^d <jb|dd>
                #
                auxmat11 = self.init_aux_matrix(
                    "vcjbk", self.nocc, self.nvirt, self.nocc
                )
                tmpabc.contract("abc,dc->abd", cia, auxmat11, **ovv)
                #
                # oc_ajb = sum_m c_m^b <aj|mm>
                #
                auxmat12 = self.init_aux_matrix(
                    "ocajb", self.nvirt, self.nocc, self.nvirt
                )
                tmpabc.contract("abc,cd->abd", cia, auxmat12, **voo)
            del tmpabc
            if self.l:
                auxmat13 = self.init_aux_matrix("gpqrr")
                mo2.slice("abcc->abc", auxmat13)
            del tmp
        #
        # g_ovvo
        #
        auxmat15 = self.init_aux_matrix(
            "govvo", self.nocc, self.nvirt, self.nvirt, self.nocc
        )
        mo2.slice("abcd->abcd", auxmat15, **ovvo)
        #
        # g_ovov
        #
        auxmat16 = self.init_aux_matrix(
            "govov", self.nocc, self.nvirt, self.nocc, self.nvirt
        )
        mo2.slice("abcd->abcd", auxmat16, **ovov)
        #
        # g_oooo
        #
        auxmat17 = self.init_aux_matrix(
            "goooo", self.nocc, self.nocc, self.nocc, self.nocc
        )
        mo2.slice("abcd->abcd", auxmat17, **oooo)
        #
        # for singles only
        #
        if self.singles:
            #
            # g_vooo
            #
            auxmat18 = self.init_aux_matrix(
                "gvooo", self.nvirt, self.nocc, self.nocc, self.nocc
            )
            mo2.slice("abcd->abcd", auxmat18, **vooo)
        #
        # g_vvvv, g_ovvv
        #
        if isinstance(self.lf, CholeskyLinalgFactory):
            # Use Cholesky functions (slow!):
            return [mo2]
        else:
            # Use dense functions (lots of memory!):
            auxmat19 = self.init_aux_matrix(
                "gvvvv", self.nvirt, self.nvirt, self.nvirt, self.nvirt
            )
            mo2.slice("abcd->abcd", auxmat19, **vvvv)
            # only needed for singles
            if self.singles:
                auxmat20 = self.init_aux_matrix(
                    "govvv", self.nocc, self.nvirt, self.nvirt, self.nvirt
                )
                mo2.slice("abcd->abcd", auxmat20, **ovvv)
            if deletemo:
                mo2.__del__()
        gc.collect()

    @timer.with_section("VecFctLCC")
    def vfunction(self, amplitudes, args):
        """Vector function to find coupled cluster doubles amplitudes.

           **Arguments:**

           amplitudes
                Cluster amplitudes. Need to be determined.

           args
                All function arguments needed to calculated the vector
                function:
                    * [0]:  wfn expansion coefficients
                    * [1]:  geminal coefficients
        """
        #
        # Get ranges
        #
        oo = self.get_range("oo")
        ov = self.get_range("ov")
        vv = self.get_range("vv")
        ov2 = self.get_range("ov", start=2)
        ov4 = self.get_range("ov", start=4)
        oo2 = self.get_range("oo", start=2)
        oo4 = self.get_range("oo", start=4)
        vv2 = self.get_range("vv", start=2)
        vv4 = self.get_range("vv", start=4)
        oov = self.get_range("oov")
        vvo = self.get_range("vvo")
        ovo = self.get_range("ovo")
        ovv = self.get_range("ovv")
        vvv = self.get_range("vvv")
        ooo = self.get_range("ooo")
        voo = self.get_range("voo")
        ovvo = self.get_range("ovvo")
        ovov = self.get_range("ovov")
        oooo = self.get_range("oooo")
        vooo = self.get_range("vooo")
        if isinstance(self.lf, CholeskyLinalgFactory):
            vvvv = self.get_range("vvvv")
            ovvv = self.get_range("ovvv")
        else:
            vvvv = self.get_range("VVVV")
            ovvv = self.get_range("oVVV")

        tco = self.tco
        #
        # Get auxiliary matrices
        #
        fock = self.get_aux_matrix("fock")
        if self.multireference:
            vcij = self.get_aux_matrix("vcij")
            ocab = self.get_aux_matrix("ocab")
            ocjbc = self.get_aux_matrix("ocjbc")
            vcjkb = self.get_aux_matrix("vcjkb")
            vcjbc = self.get_aux_matrix("vcjbc")
            ocjkb = self.get_aux_matrix("ocjkb")
            vcjkl = self.get_aux_matrix("vcjkl")
            ocabc = self.get_aux_matrix("ocabc")
        govvo = self.get_aux_matrix("govvo")
        govov = self.get_aux_matrix("govov")
        goooo = self.get_aux_matrix("goooo")
        if self.singles:
            gvooo = self.get_aux_matrix("gvooo")
            if self.multireference:
                ocjb = self.get_aux_matrix("ocjb")
                vcjb = self.get_aux_matrix("vcjb")
                vcjbk = self.get_aux_matrix("vcjbk")
                ocajb = self.get_aux_matrix("ocajb")
        try:
            gvvvv = self.get_aux_matrix("gvvvv")
        except:
            gvvvv = args[-1]
        try:
            govvv = self.get_aux_matrix("govvv")
        except:
            govvv = args[-1]
        #
        # output
        #
        end = 0
        #
        # singles
        #
        if self.singles:
            singles = self.lf.create_two_index(self.nocc, self.nvirt)
            tjb = self.lf.create_two_index(self.nocc, self.nvirt)
            end = self.nocc * self.nvirt
            if type(amplitudes) == np.ndarray:
                tjb.assign(amplitudes[:end])
            else:
                tjb.assign(amplitudes._array[:end])
        #
        # doubels
        #
        doubles = self.denself.create_four_index(
            self.nocc, self.nvirt, self.nocc, self.nvirt
        )
        tjbkc = self.denself.create_four_index(
            self.nocc, self.nvirt, self.nocc, self.nvirt
        )
        #
        # Assign triangular blocks
        #
        self.assign_doubles(tjbkc, amplitudes, end)
        del amplitudes
        #
        # temporary storage
        #
        if self.multireference:
            tmp4ind = self.denself.create_four_index(
                self.nocc, self.nvirt, self.nocc, self.nvirt
            )
            tmp3ind0 = self.lf.create_three_index(
                self.nocc, self.nvirt, self.nvirt
            )
            tmp3ind1 = self.lf.create_three_index(
                self.nocc, self.nocc, self.nvirt
            )
            tmp2ind = self.lf.create_two_index(self.nvirt, self.nvirt)
            tmp2ind1 = self.lf.create_two_index(self.nocc, self.nocc)
            tmp2ind2 = self.lf.create_two_index(self.nocc, self.nvirt)

        if self.singles:
            #
            # t_kc L<ic|ak> (s4)
            #
            govvo.contract(
                "abcd,db->ac", tjb, singles, factor=2.0, clear=True, select=tco
            )
            govov.contract(
                "abcd,cb->ad", tjb, singles, factor=-1.0, select=tco
            )
            #
            # pCCD reference function
            #
            if self.multireference:
                #
                # L<ik|ac> c_ia t_kc (s12) (icak, iack)
                #
                govvo.contract(
                    "abcd,db->ac",
                    tjb,
                    tmp2ind2,
                    factor=2.0,
                    clear=True,
                    select=tco,
                )
                govvo.contract(
                    "abcd,dc->ab", tjb, tmp2ind2, factor=-1.0, select=tco
                )
                tmp2ind2.imul(args[1], 1.0)
                singles.iadd(tmp2ind2)
            #
            # Fac tic; Fki tka (s2,3)
            #
            tjb.contract("ab,bc->ac", fock, singles, **vv2, select=tco)
            tjb.contract(
                "ab,ac->cb", fock, singles, factor=-1.0, **oo2, select=tco
            )
            if self.multireference:
                # tic (a,c) (s14)
                tjb.contract(
                    "ab,cb->ac", ocab, singles, factor=-1.0, select=tco
                )
                # tka (k,i) (s13)
                tjb.contract(
                    "ab,ac->cb", vcij, singles, factor=-1.0, select=tco
                )
            # F_ia (s1)
            singles.iadd(fock, 1.0, **ov2)
            if self.multireference:
                # Fia cia (s9-3)
                singles.iadd_mult(args[1], fock, 1.0, **ov2)
                # (i,a) (s11-2)
                singles.iadd(ocjb, -1.0)
                # (i,a) (s10-2)
                singles.iadd(vcjb, 1.0)
            # Coupling between amplitudes: Singles-Doubles:
            # Fkc Theta_iakc (s9-1,2)
            tjbkc.contract(
                "abcd,ab->cd", fock, singles, factor=2.0, **ov4, select=tco
            )
            tjbkc.contract(
                "abcd,cb->ad", fock, singles, factor=-1.0, **ov4, select=tco
            )
            # tkcla L<ci|kl> (s11-1)
            tjbkc.contract(
                "abcd,beac->ed", gvooo, singles, factor=-2.0, select=tco
            )
            tjbkc.contract("abcd,beca->ed", gvooo, singles, select=tco)
            # tkcid L<ka|cd> (s10-1)
            govvv.contract(
                "abcd,aced->eb", tjbkc, singles, factor=2.0, **ovvv, select=tco
            )
            govvv.contract(
                "abcd,adec->eb",
                tjbkc,
                singles,
                factor=-1.0,
                **ovvv,
                select=tco,
            )
            # Coupling between amplitudes: Singles-Doubles:
            #
            # P <jc|ba> t_ic (d1)
            #
            govvv.contract("abcd,ed->ebac", tjb, doubles, **ovvv, select=tco)
            #
            # <ja||bc> t_ic c_jb and -<jbca> t_ic c_ja (d12)
            #
            if self.multireference:
                # (iajb)
                govvv.contract(
                    "abcd,ec->ebad",
                    tjb,
                    tmp4ind,
                    clear=True,
                    **ovvv,
                    select=tco,
                )
                # c_ja
                tmp4ind.contract(
                    "abcd,cb->abcd",
                    args[1],
                    doubles,
                    factor=-1.0,
                    select="einsum",
                )
                # c_jb
                govvv.contract(
                    "abcd,ed->ebac",
                    tjb,
                    tmp4ind,
                    factor=-1.0,
                    **ovvv,
                    select=tco,
                )
                tmp4ind.contract(
                    "abcd,cd->abcd",
                    args[1],
                    doubles,
                    factor=-1.0,
                    select="einsum",
                )
            #
            # g_bkji t_ka (d2)
            #
            gvooo.contract(
                "abcd,be->deca", tjb, doubles, factor=-1.0, select=tco
            )
            #
            # pCCD reference function
            #
            if self.multireference:
                #
                # <bj|ki> tka cib <bi||jk> tka cjb (d13)
                #
                # bikj t_ka (iajb)
                gvooo.contract(
                    "abcd,ce->beda", tjb, tmp4ind, clear=True, select=tco
                )
                # c_ib
                tmp4ind.contract(
                    "abcd,ad->abcd", args[1], doubles, select="einsum"
                )
                # bijk tka cjb
                gvooo.contract(
                    "abcd,de->beca", tjb, tmp4ind, factor=-1.0, select=tco
                )
                tmp4ind.contract(
                    "abcd,cd->abcd", args[1], doubles, select="einsum"
                )
                #
                # delta_ij L_kacb t_kc c_ib (d11)
                #
                # (ab)
                govvv.contract(
                    "abcd,ac->bd",
                    tjb,
                    tmp2ind,
                    factor=2.0,
                    clear=True,
                    **ovvv,
                    select=tco,
                )
                govvv.contract(
                    "abcd,ad->bc",
                    tjb,
                    tmp2ind,
                    factor=-1.0,
                    **ovvv,
                    select=tco,
                )
                # (ab) cib
                args[1].contract(
                    "ac,bc->abc", tmp2ind, tmp3ind0, select="einsum"
                )
                #
                # delta_ij f_kb t_ka c_ib (d9)
                # (ab) ; ib (ac)
                tjb.contract(
                    "ab,ac->bc", fock, tmp2ind, clear=True, **ov2, select=tco
                )
                args[1].contract(
                    "ac,bc->abc",
                    tmp2ind,
                    tmp3ind0,
                    factor=-1.0,
                    select="einsum",
                )
                #
                # delta_ij (kbi) t_ka (d15)
                #
                vcjbk.contract(
                    "abc,ad->cdb", tjb, tmp3ind0, factor=-1.0, select=tco
                )
                #
                # delta_ac f_jc t_ic c_jb (d8)
                # (ij) cjb
                tjb.contract(
                    "ab,cb->ac", fock, tmp2ind1, clear=True, **ov2, select=tco
                )
                tmp2ind1.contract(
                    "ab,bc->abc",
                    args[1],
                    tmp3ind1,
                    factor=-1.0,
                    select="einsum",
                )
                #
                # delta_ab (cja) t_ic (d14)
                #
                ocajb.contract("abc,da->dbc", tjb, tmp3ind1, select=tco)
                #
                # delta_ab L_cikj t_kc c_ja (d10)
                # (ij) cja
                gvooo.contract(
                    "abcd,ca->bd",
                    tjb,
                    tmp2ind1,
                    factor=2.0,
                    clear=True,
                    select=tco,
                )
                gvooo.contract(
                    "abcd,da->bc", tjb, tmp2ind1, factor=-1.0, select=tco
                )
                tmp2ind1.contract(
                    "ab,bc->abc",
                    args[1],
                    tmp3ind1,
                    factor=-1.0,
                    select="einsum",
                )

        #
        # <ab|ij> (d-d1)
        #
        govvo.slice("abcd->acdb", doubles, 0.5, False)
        #
        # pCCD reference function:
        #
        if self.multireference:
            #
            # (d-d7)
            # c_jb <ja||bi>
            #
            govvo.contract("abcd,db->acdb", args[1], doubles, select="einsum")
            govov.contract(
                "abcd,cb->adcb", args[1], doubles, factor=-1.0, select="einsum"
            )
            #
            # -c_ib <ia|jb>
            #
            govov.contract(
                "abcd,ab->adcb", args[1], doubles, factor=-1.0, select="einsum"
            )
            # (d-d9)
            # c_iajb <ib|aj>
            # (iajb)
            govvo.contract(
                "abcd,ac->acdb", args[1], tmp4ind, clear=True, select="einsum"
            )
            tmp4ind.contract(
                "abcd,cd->abcd", args[1], doubles, factor=0.5, select="einsum"
            )
            govvo.contract(
                "abcd,ab->acdb", args[1], tmp4ind, clear=True, select="einsum"
            )
            tmp4ind.contract(
                "abcd,cb->abcd", args[1], doubles, factor=0.5, select="einsum"
            )
            #
            # (d-d2-2)
            # delta_ij [ c_jb F_ab ]
            #
            args[1].contract(
                "ac,bc->abc", fock, tmp3ind0, **vv2, select="einsum"
            )
            # (d-d16)
            # delta_ij [ oc_jbc ]
            # (iba) = (iab)
            tmp3ind0.iadd(ocjbc, -0.5)
            # (d-d5-2)
            # delta_ij [ vc_jbc ]
            # (iab)
            tmp3ind0.iadd(vcjbc, 0.5)
            # (d-d3-2)
            # delta_ac [ c_ia F_ij ]
            #
            args[1].contract(
                "ab,ac->acb",
                fock,
                tmp3ind1,
                factor=-1.0,
                **oo2,
                select="einsum",
            )
            # (d-d13)
            # delta_ab [ vc_jkb ]
            # (ija)
            tmp3ind1.iadd(vcjkb, -0.5)
            # (d-d4-2)
            # delta_bc [ oc_jkb ]
            # (ija)
            tmp3ind1.iadd(ocjkb, 0.5)

        # (d-d3-1)
        # sum_k t_iakb*F_kj
        #
        tjbkc.contract(
            "abcd,ce->abed", fock, doubles, factor=-1.0, **oo4, select=tco
        )
        # (d-d2-1)
        # sum_c t_iajc*F_bc
        #
        tjbkc.contract("abcd,ed->abce", fock, doubles, **vv4, select=tco)
        # (d-d6-1)
        # sum_kc t_iakc*L_kbcj
        #
        tjbkc.contract(
            "abcd,cedf->abfe", govvo, doubles, factor=2.0, select=tco
        )
        tjbkc.contract(
            "abcd,cefd->abfe", govov, doubles, factor=-1.0, select=tco
        )
        # (d-d6-3)
        # sum_kc (t_kajc*g_ickb)
        #
        tjbkc.contract(
            "abcd,edaf->ebcf", govov, doubles, factor=-1.0, select=tco
        )
        # (d-d6-2)
        # sum_kc (t_kaic*g_kbcj)
        #
        tjbkc.contract(
            "abcd,aedf->cbfe", govvo, doubles, factor=-1.0, select=tco
        )
        # (d-d4-1)
        # sum_kl t_kalb*g_klij
        #
        tjbkc.contract(
            "abcd,acef->ebfd", goooo, doubles, factor=0.5, select=tco
        )
        # (d-d5-1)
        # sum_cd t_icjd*g_abcd
        #
        gvvvv.contract(
            "abcd,ecfd->eafb", tjbkc, doubles, factor=0.5, **vvvv, select=tco
        )
        #
        # pCCD reference function
        #
        if self.multireference:
            # (d-d12)
            # sum_k t_iakb*vc_ik
            # (kj)
            tjbkc.contract(
                "abcd,ce->abed", vcij, doubles, factor=-1.0, select=tco
            )
            # (d-d15)
            # sum_c t_iajc oc_ca
            # (bc)
            tjbkc.contract(
                "abcd,ed->abce", ocab, doubles, factor=-1.0, select=tco
            )
            # (d-d8-1)
            # sum_kc t_iakc L_kjcb c_jb
            # (iajb, cjb)
            tjbkc.contract(
                "abcd,cedf->abfe", govvo, tmp4ind, clear=True, select=tco
            )
            tmp4ind.contract(
                "abcd,cd->abcd", args[1], doubles, factor=2.0, select="einsum"
            )
            tjbkc.contract(
                "abcd,cdef->abfe", govvo, tmp4ind, clear=True, select=tco
            )
            tmp4ind.contract(
                "abcd,cd->abcd", args[1], doubles, factor=-1.0, select="einsum"
            )
            # (d-d8-3)
            # sum_kc t_kajc g_ikcb c_ib
            # (iajb,ib)
            tjbkc.contract(
                "abcd,efda->ebcf", govvo, tmp4ind, clear=True, select=tco
            )
            tmp4ind.contract(
                "abcd,ad->abcd", args[1], doubles, select="einsum"
            )
            # (d-d8-2)
            # sum_kc t_kaic g_jkbc c_jb
            # (iajb,jb)
            tjbkc.contract(
                "abcd,edfa->cbef", govvo, tmp4ind, clear=True, select=tco
            )
            tmp4ind.contract(
                "abcd,cd->abcd", args[1], doubles, factor=-1.0, select="einsum"
            )
            # part of (d-d8-2)
            # sum_kc t_kaic g_jkcb c_jb
            # (iajb,jb)
            tjbkc.contract(
                "abcd,efda->cbef", govvo, tmp4ind, clear=True, select=tco
            )
            tmp4ind.contract(
                "abcd,cd->abcd", args[1], doubles, select="einsum"
            )
            # (d-d18)
            # delta_ij [ sum_kl t_kalb vc_jkl ]
            # (kli, kalb)
            tjbkc.contract(
                "abcd,ace->ebd", vcjkl, tmp3ind0, factor=0.5, select=tco
            )
            # (d-d14)
            # delta_ij [ sum_klc t_lakc L_lkbc c_ib ]
            # (ab,ib)
            tjbkc.contract(
                "abcd,adec->be", govvo, tmp2ind, clear=True, select=tco
            )
            args[1].contract(
                "ac,bc->abc", tmp2ind, tmp3ind0, factor=-2.0, select="einsum"
            )
            tjbkc.contract(
                "abcd,aedc->be", govvo, tmp2ind, clear=True, select=tco
            )
            args[1].contract("ac,bc->abc", tmp2ind, tmp3ind0, select="einsum")
            # (d-d17)
            # delta_ab [ sum_cd t_icjd vc_abc ]
            # (acd, ticjd)
            tjbkc.contract(
                "abcd,ebd->ace", ocabc, tmp3ind1, factor=0.5, select=tco
            )
            # (d-d11)
            # delta_ab [ sum_kcd t_idkc L_jkdc c_ja ]
            # (ij,ja)
            tjbkc.contract(
                "abcd,edbc->ae", govvo, tmp2ind1, clear=True, select=tco
            )
            tmp2ind1.contract(
                "ab,bc->abc", args[1], tmp3ind1, factor=-2.0, select="einsum"
            )
            tjbkc.contract(
                "abcd,ebdc->ae", govvo, tmp2ind1, clear=True, select=tco
            )
            tmp2ind1.contract("ab,bc->abc", args[1], tmp3ind1, select="einsum")

            doubles.iadd_expand_three_to_four("1-3-1-2", tmp3ind0, 1.0)
            doubles.iadd_expand_three_to_four("0-2-0-1", tmp3ind1, 1.0)
        #
        # Add permutation
        #
        doubles.iadd_transpose((2, 3, 0, 1))
        #
        # Set all diagonal amplitudes zero
        #
        if self.multireference:
            ind1, ind2 = np.indices((self.nocc, self.nvirt))
            indices = [ind1, ind2, ind1, ind2]
            doubles.assign(0.0, indices)
        if self.freeze is not None:
            for row in self.freeze:
                doubles.set_element(
                    row[0], row[1], row[2], row[3], 0.0, symmetry=1
                )
                doubles.set_element(
                    row[2], row[3], row[0], row[1], 0.0, symmetry=1
                )

        gc.collect()
        if self.singles:
            t = np.hstack(
                (
                    singles._array.ravel(order="C"),
                    doubles.get_triu().ravel(order="C"),
                )
            )
            if self.solver == "krylov":
                return t
            else:
                out = self.lf.create_one_index(t.shape[0])
                out.assign(t)
                return out
        t = doubles.get_triu().ravel(order="C")
        if self.solver == "krylov":
            return t
        else:
            out = self.lf.create_one_index(t.shape[0])
            out.assign(t)
            return out

    @timer.with_section("JacobianLCC")
    def jacobian(self, amplitudes, args):
        """Jacobian approximation to find coupled cluster doubles amplitudes.

           **Arguments:**

           amplitudes
                Cluster amplitudes.

           args
                All function arguments needed to calculated the vector
        """
        if log.do_medium:
            log("Computing Jacobian approximation for Quasi-Newton solver.")
        fock = self.get_aux_matrix("fock")
        govvo = self.get_aux_matrix("govvo")
        govov = self.get_aux_matrix("govov")
        gpqpq = self.get_aux_matrix("gpqpq")
        giaai = govvo.slice("abba->ab")
        giaia = govov.slice("abab->ab")
        fi = fock.copy_diagonal(end=self.nocc)
        fa = fock.copy_diagonal(begin=self.nocc)
        if self.multireference:
            cia = args[1]
        out = self.lf.create_one_index(self.nunknowns)
        if self.singles:
            # The function returns fi-fa and fi-fa+fj-fb
            fia, fiajb = get_epsilon(
                self.denself,
                [fi, fa],
                singles=self.singles,
                shift=[1e-12, 1e-12],
            )
            fia.iscale(-1.0)
            fiajb.iscale(-1.0)
            # Add other terms:
            if self.jacobian_approximation == 2:
                fia.iadd(giaai, factor=2.0)
                fia.iadd(giaia, factor=-1.0)
                if self.multireference:
                    tmp = self.lf.create_one_index(self.nocc)
                    giaai.contract("ab,ab->a", cia, tmp)
                    fia.iadd_expand_one_to_two("a->ab", tmp, -1.0)
                    tmp = self.lf.create_one_index(self.nvirt)
                    giaai.contract("ab,ab->b", cia, tmp)
                    fia.iadd_expand_one_to_two("b->ab", tmp, -1.0)
        else:
            fiajb = get_epsilon(
                self.denself,
                [fi, fa],
                singles=self.singles,
                shift=[1e-12, 1e-12],
            )
            fiajb.iscale(-1.0)
        # temporary copy
        eps2 = self.denself.create_four_index(
            self.nocc, self.nvirt, self.nocc, self.nvirt
        )
        eps2.assign(
            fiajb._array.reshape(self.nocc, self.nvirt, self.nocc, self.nvirt)
        )
        del fiajb
        if self.jacobian_approximation == 2:
            eps2_ = eps2.new()
            # Libaj
            govvo.slice("abcd->acdb", eps2_, 2.0, False)
            govov.slice("abcd->adcb", eps2_, -1.0, False)
            eps2_.iadd(govov, factor=-1.0)
            # <ab|ab>
            eps2_.iadd_expand_two_to_four(
                "bd->abcd", gpqpq, 0.5, begin0=self.nocc, begin1=self.nocc
            )
            # <ij|ij>
            eps2_.iadd_expand_two_to_four(
                "ac->abcd", gpqpq, 0.5, end0=self.nocc, end1=self.nocc
            )
            if self.multireference:
                tmp = self.lf.create_one_index(self.nocc)
                giaai.contract("ab,ab->a", cia, tmp)
                eps2_.iadd_expand_one_to_four("2", tmp, -1.0)
                eps2_.iadd_expand_one_to_four("a->abac", tmp, 0.5)
                tmp = self.lf.create_one_index(self.nvirt)
                giaai.contract("ab,ab->b", cia, tmp)
                eps2_.iadd_expand_one_to_four("3", tmp, -1.0)
                eps2_.iadd_expand_one_to_four("b->abcb", tmp, 0.5)
                govov.contract("abcd,cd->abcd", cia, eps2_, factor=2.0)
                govvo.contract("abcd,dc->abdc", cia, eps2_, factor=-1.0)
                govvo.contract("abcd,dc->acdb", cia, eps2_)
            # Permutation
            eps2_.iadd_transpose((2, 3, 0, 1))
            eps2.iadd(eps2_)
            del eps2_
        if self.multireference:
            ind1, ind2 = np.indices((self.nocc, self.nvirt))
            indices = [ind1, ind2, ind1, ind2]
            eps2.assign(1.0, indices)
        if self.singles:
            eps1 = fia._array.reshape(self.nocc * self.nvirt)
            out.assign(eps1, end0=self.nocc * self.nvirt)
            out.assign(eps2.get_triu(), begin0=self.nocc * self.nvirt)
        else:
            out.assign(eps2.get_triu())

        return out

    @timer.with_section("VecFctLambdaLCC")
    def vector_lagrange(self, amplitudes, args):
        #
        # Get ranges
        #
        ov2 = self.get_range("ov", start=2)
        vv2 = self.get_range("vv", start=2)
        oo2 = self.get_range("oo", start=2)
        vv4 = self.get_range("vv", start=4)
        oo4 = self.get_range("oo", start=4)
        vvo = self.get_range("vvo")
        oov = self.get_range("oov")
        ovv = self.get_range("ovv")
        voo = self.get_range("voo")
        if isinstance(self.lf, CholeskyLinalgFactory):
            vvvv = self.get_range("vvvv")
            ovvv = self.get_range("ovvv")
        else:
            vvvv = self.get_range("VVVV")
            ovvv = self.get_range("oVVV")
        #
        tco = self.tco
        fock = self.get_aux_matrix("fock")
        govvo = self.get_aux_matrix("govvo")
        govov = self.get_aux_matrix("govov")
        goooo = self.get_aux_matrix("goooo")
        if self.singles:
            gvooo = self.get_aux_matrix("gvooo")
        if self.multireference:
            gpqrr = self.get_aux_matrix("gpqrr")
        try:
            gvvvv = self.get_aux_matrix("gvvvv")
        except:
            gvvvv = args[-1]
        try:
            govvv = self.get_aux_matrix("govvv")
        except:
            govvv = args[-1]

        end = 0

        # singles
        if self.singles:
            end = self.nocc * self.nvirt
            Lambda_ia = self.lf.create_two_index(self.nocc, self.nvirt)
            lamb_jb = self.lf.create_two_index(self.nocc, self.nvirt)
            if type(amplitudes) == np.ndarray:
                lamb_jb.assign(amplitudes[:end])
            else:
                lamb_jb.assign(amplitudes._array[:end])
        # doubles
        Lambda_iajb = self.denself.create_four_index(
            self.nocc, self.nvirt, self.nocc, self.nvirt
        )
        lamb_jbkc = self.denself.create_four_index(
            self.nocc, self.nvirt, self.nocc, self.nvirt
        )
        if self.multireference:
            tp = unmask("t_p", *args)
        self.assign_doubles(lamb_jbkc, amplitudes, end)
        del amplitudes

        if self.singles:
            # intermediates:
            tmp_vv = self.lf.create_two_index(self.nvirt, self.nvirt)
            tmp_oo = self.lf.create_two_index(self.nocc, self.nocc)
        #
        # Lambda_ia
        #
        if self.singles:
            Lambda_ia.clear()
            # 1
            Lambda_ia.iadd(fock, 2.0, **ov2)
            # 2
            lamb_jb.contract(
                "ab,bc->ac", fock, Lambda_ia, factor=1.0, **vv2, select=tco
            )
            # 3
            lamb_jb.contract(
                "ab,ac->cb", fock, Lambda_ia, factor=-1.0, **oo2, select=tco
            )
            # 4
            # ic||ak lkc
            govvo.contract(
                "abcd,db->ac", lamb_jb, Lambda_ia, factor=2.0, select=tco
            )
            govov.contract(
                "abcd,cb->ad", lamb_jb, Lambda_ia, factor=-1.0, select=tco
            )
            # 5
            govvv.contract(
                "abcd,edac->eb",
                lamb_jbkc,
                Lambda_ia,
                factor=1.0,
                select=tco,
                **ovvv,
            )
            # 6
            gvooo.contract(
                "abcd,deca->be", lamb_jbkc, Lambda_ia, factor=-1.0, select=tco
            )
            if self.multireference:
                # 7
                gpqrr.contract(
                    "abc,cb->ab", tp, tmp_vv, factor=-1.0, clear=True, **vvo
                )
                lamb_jb.contract(
                    "ab,cb->ac", tmp_vv, Lambda_ia, factor=1.0, select=tco
                )
                # 8
                gpqrr.contract(
                    "abc,bc->ab", tp, tmp_oo, factor=-1.0, clear=True, **oov
                )
                tmp_oo.contract(
                    "ab,bc->ac", lamb_jb, Lambda_ia, factor=1.0, select=tco
                )
                # 9
                tpl = tp.copy()
                tpl.imul(lamb_jb)
                govvo.contract(
                    "abcd,db->ac", tpl, Lambda_ia, factor=2.0, select=tco
                )
                govvo.contract(
                    "abcd,dc->ab", tpl, Lambda_ia, factor=-1.0, select=tco
                )
                # 10
                lamb_jbkc.contract(
                    "abac,ac->bc", tp, tmp_vv, factor=-1.0, clear=True
                )
                tmp_vv.contract(
                    "ab,cb->ca", fock, Lambda_ia, factor=1.0, **ov2, select=tco
                )
                # 11
                lamb_jbkc.contract(
                    "abcb,cb->ac", tp, tmp_oo, factor=-1.0, clear=True
                )
                tmp_oo.contract(
                    "ab,bc->ac", fock, Lambda_ia, factor=1.0, **ov2, select=tco
                )
                # 12
                lamb_jbkc.contract(
                    "abac,ab->bc", tp, tmp_vv, factor=1.0, clear=True
                )
                govvv.contract(
                    "abcd,db->ac",
                    tmp_vv,
                    Lambda_ia,
                    factor=2.0,
                    select=tco,
                    **ovvv,
                )
                govvv.contract(
                    "abcd,cd->ab",
                    tmp_vv,
                    Lambda_ia,
                    factor=-1.0,
                    select=tco,
                    **ovvv,
                )
                # 13
                tmp = gpqrr.contract(
                    "abc,dc->adb", tp, out=None, factor=1.0, select=tco, **ovv
                )
                lamb_jbkc.contract("abac,dac->db", tmp, Lambda_ia, factor=-1.0)
                del tmp
                # 14
                tmp = lamb_jbkc.contract(
                    "abcb,db->acd", tp, out=None, factor=1.0
                )
                gpqrr.contract(
                    "abc,dbc->da",
                    tmp,
                    Lambda_ia,
                    factor=1.0,
                    select=tco,
                    **voo,
                )
                del tmp
                # 15
                lamb_jbkc.contract(
                    "abcb,cb->ac", tp, tmp_oo, factor=1.0, clear=True
                )
                # 16
                gvooo.contract(
                    "abcd,bd->ca", tmp_oo, Lambda_ia, factor=-2.0, select=tco
                )
                gvooo.contract(
                    "abcd,dc->ba", tmp_oo, Lambda_ia, factor=1.0, select=tco
                )
                del tmp_oo
                # 17
                tmp = lamb_jbkc.contract(
                    "abcd,cd->abcd", tp, out=None, factor=1.0
                )
                # 18
                govvv.contract(
                    "abcd,ebac->ed",
                    tmp,
                    Lambda_ia,
                    factor=1.0,
                    select=tco,
                    **ovvv,
                )
                # 19
                govvv.contract(
                    "abcd,ebad->ec",
                    tmp,
                    Lambda_ia,
                    factor=-1.0,
                    select=tco,
                    **ovvv,
                )
                del tmp
                # 20
                tmp = lamb_jbkc.contract(
                    "abcd,cb->adcb", tp, out=None, factor=1.0
                )
                govvv.contract(
                    "abcd,ebad->ec",
                    tmp,
                    Lambda_ia,
                    factor=-1.0,
                    select=tco,
                    **ovvv,
                )
                del tmp
                # 21
                tmp = lamb_jbkc.contract(
                    "abcd,ad->abcd", tp, out=None, factor=1.0
                )
                # 22
                gvooo.contract(
                    "abcd,deba->ce", tmp, Lambda_ia, factor=1.0, select=tco
                )
                del tmp
                # 23
                tmp = lamb_jbkc.contract(
                    "abcd,cd->bacd", tp, out=None, factor=1.0
                )
                gvooo.contract(
                    "abcd,ebca->de", tmp, Lambda_ia, factor=-1.0, select=tco
                )
                gvooo.contract(
                    "abcd,ebda->ce", tmp, Lambda_ia, factor=1.0, select=tco
                )
                del tmp

        #
        # Lambda_iajb
        #

        # 1
        govvo.slice("abcd->acdb", Lambda_iajb, 2.0, True)
        govvo.slice("abcd->abdc", Lambda_iajb, -1.0, False)
        if self.singles:
            # 2
            Lambda_iajb.iadd_expand_two_to_four(
                "ab,cd->abcd",
                fock,
                2.0,
                lamb_jb,
                end0=self.nocc,
                begin1=self.nocc,
            )
            Lambda_iajb.iadd_expand_two_to_four(
                "ab,cd->adcb",
                fock,
                -1.0,
                lamb_jb,
                end0=self.nocc,
                begin1=self.nocc,
            )
            # 3
            govvv.contract(
                "abcd,eb->edac",
                lamb_jb,
                Lambda_iajb,
                factor=2.0,
                select=tco,
                **ovvv,
            )
            govvv.contract(
                "abcd,eb->ecad",
                lamb_jb,
                Lambda_iajb,
                factor=-1.0,
                select=tco,
                **ovvv,
            )
            # 4
            gvooo.contract(
                "abcd,be->cade", lamb_jb, Lambda_iajb, factor=-2.0, select=tco
            )
            gvooo.contract(
                "abcd,be->dace", lamb_jb, Lambda_iajb, factor=1.0, select=tco
            )
        # 5
        lamb_jbkc.contract(
            "abcd,ed->abce", fock, Lambda_iajb, factor=1.0, **vv4, select=tco
        )
        # 6
        lamb_jbkc.contract(
            "abcd,ec->abed", fock, Lambda_iajb, factor=-1.0, **oo4, select=tco
        )
        # 7
        gvvvv.contract(
            "abcd,ecfd->eafb",
            lamb_jbkc,
            Lambda_iajb,
            factor=0.5,
            select=tco,
            **vvvv,
        )
        # 8
        goooo.contract(
            "abcd,cedf->aebf", lamb_jbkc, Lambda_iajb, factor=0.5, select=tco
        )
        # 9
        govvo.contract(
            "abcd,efdb->acef", lamb_jbkc, Lambda_iajb, factor=2.0, select=tco
        )
        govov.contract(
            "abcd,efcb->adef", lamb_jbkc, Lambda_iajb, factor=-1.0, select=tco
        )
        # 10
        govvo.contract(
            "abcd,efdb->afec", lamb_jbkc, Lambda_iajb, factor=-1.0, select=tco
        )
        # 11
        govov.contract(
            "abcd,ebcf->edaf", lamb_jbkc, Lambda_iajb, factor=-1.0, select=tco
        )
        if self.multireference:
            # 12
            tmp = lamb_jbkc.contract("abac,ac->bc", tp, out=None, factor=1.0)
            govvo.contract(
                "abcd,ec->aedb", tmp, Lambda_iajb, factor=-2.0, select=tco
            )
            govvo.contract(
                "abcd,eb->aedc", tmp, Lambda_iajb, factor=1.0, select=tco
            )
            del tmp
            # 13
            tmp = lamb_jbkc.contract("abcb,cb->ac", tp, out=None, factor=1.0)
            govvo.contract(
                "abcd,ed->ebac", tmp, Lambda_iajb, factor=-2.0, select=tco
            )
            govvo.contract(
                "abcd,ed->ecab", tmp, Lambda_iajb, factor=1.0, select=tco
            )
            del tmp
            # 14
            tmp = gpqrr.contract(
                "abc,dc->abd", tp, out=None, factor=1.0, select=tco, **oov
            )
            lamb_jbkc.contract("abac,efa->ebfc", tmp, Lambda_iajb, factor=0.5)
            del tmp
            # 15
            tmp = lamb_jbkc.contract("abcb,db->acd", tp, out=None, factor=1.0)
            gpqrr.contract(
                "abc,efc->eafb",
                tmp,
                Lambda_iajb,
                factor=0.5,
                select=tco,
                **vvo,
            )
            del tmp
            # 16
            tmp = gpqrr.contract(
                "abc,cb->ab", tp, out=None, factor=-1.0, **vvo
            )
            lamb_jbkc.contract(
                "abcd,eb->aecd", tmp, Lambda_iajb, factor=1.0, select=tco
            )
            del tmp
            # 17
            tmp = gpqrr.contract(
                "abc,bc->ab", tp, out=None, factor=-1.0, **oov
            )
            lamb_jbkc.contract(
                "abcd,ea->ebcd", tmp, Lambda_iajb, factor=1.0, select=tco
            )
            del tmp
            # 18
            tmp = govvo.contract("abcd,dc->abdc", tp, out=None, factor=1.0)
            lamb_jbkc.contract(
                "abcd,efcb->afed", tmp, Lambda_iajb, factor=1.0, select=tco
            )
            del tmp
            # 19
            tmp = govvo.contract("abcd,db->acdb", tp, out=None, factor=-2.0)
            govvo.contract("abcd,dc->abdc", tp, tmp, factor=1.0)
            lamb_jbkc.contract(
                "abcd,efcd->ebaf", tmp, Lambda_iajb, factor=1.0, select=tco,
            )
            del tmp
            # 20
            tmp = govvo.contract("abcd,db->acdb", tp, out=None, factor=1.0)
            lamb_jbkc.contract(
                "abcd,efcd->ebaf", tmp, Lambda_iajb, factor=1.0, select=tco,
            )
            del tmp
            # 21
            tmp = govvo.contract("abcd,db->acdb", tp, out=None, factor=2.0)
            govvo.contract("abcd,dc->abdc", tp, tmp, factor=-1.0)
            lamb_jbkc.contract(
                "abcd,efcd->abef", tmp, Lambda_iajb, factor=1.0, select=tco
            )
            del tmp
        # Sum
        Lambda_iajb.iadd_transpose((2, 3, 0, 1))

        #
        # Set all diagonal amplitudes zero
        #
        if self.multireference:
            ind1, ind2 = np.indices((self.nocc, self.nvirt))
            indices = [ind1, ind2, ind1, ind2]
            Lambda_iajb.assign(0.0, indices)

        if self.singles:
            t = np.hstack(
                (
                    Lambda_ia._array.ravel(order="C"),
                    Lambda_iajb.get_triu().ravel(order="C"),
                )
            )
            return t
        t = np.hstack(Lambda_iajb.get_triu().ravel(order="C"))
        return t

    # Not to be released:
    @timer.with_section("VecFctLambdaLCC")
    def vector_lagrange_old(self, amplitudes, t, args):
        #
        # This is old code. It contains additional terms that are needed for
        # CCSD. It might be useful in the future
        #

        tco = self.tco
        fock = self.get_aux_matrix("fock")
        govvo = self.get_aux_matrix("govvo")
        govov = self.get_aux_matrix("govov")
        goooo = self.get_aux_matrix("goooo")
        if self.singles:
            gvooo = self.get_aux_matrix("gvooo")
            if self.multireference:
                gpqrr = self.get_aux_matrix("gpqrr")
        try:
            gvvvv = self.get_aux_matrix("gvvvv")
        except:
            pass
        try:
            govvv = self.get_aux_matrix("govvv")
        except:
            pass
        #
        # Get active orbital indices:
        # iocc: final active occupied orbital index
        # iend: final active virtual orbital index
        #
        iocc = self.nocc
        iend = self.nbasis
        end = 0

        # singles
        if self.singles:
            t1 = t[0]
            end = self.nocc * self.nvirt
            Lambda_ia = self.lf.create_two_index(self.nocc, self.nvirt)
            lamb_jb = self.lf.create_two_index(self.nocc, self.nvirt)
            if type(amplitudes) == np.ndarray:
                lamb_jb.assign(amplitudes[:end])
            else:
                lamb_jb.assign(amplitudes._array[:end])
        # doubles
        Lambda_iajb = self.denself.create_four_index(
            self.nocc, self.nvirt, self.nocc, self.nvirt
        )
        lamb_jbkc = self.denself.create_four_index(
            self.nocc, self.nvirt, self.nocc, self.nvirt
        )
        lamb_jbkc.assign_triu(amplitudes, begin0=end)
        lamb_jbkc.iadd_transpose((2, 3, 0, 1))
        # T2
        if isinstance(t[-1], FourIndex):
            t2 = t[-1].copy()
        else:
            t2 = lamb_jbkc.new()
            t2.assign_triu(t[-1])
            if self.multireference:
                t2.iadd_transpose((2, 3, 0, 1))
            else:
                tp_ = t2.slice("abab->ab")
                t2.iadd_transpose((2, 3, 0, 1))
                ind1, ind2 = np.indices((self.nocc, self.nvirt))
                indp = [ind1, ind2, ind1, ind2]
                # assign diagonal
                t2.assign(tp_, indp)
        # assign pair amplitudes to t2 explicitly as these terms are not added
        # separately
        if self.multireference:
            ind1, ind2 = np.indices((self.nocc, self.nvirt))
            indices = [ind1, ind2, ind1, ind2]
            t2.assign(args[1], indices)
            t2p = t2.new()
            t2p.assign(args[1], indices)
        else:
            ind1, ind2 = np.indices((self.nocc, self.nvirt))
            indices = [ind1, ind2, ind1, ind2]
            tmp = lamb_jbkc._array[indices]
            lamb_jbkc.assign(tmp * 2.0, indices)
        #       lamb_jbkc.assign(0.0)
        #       lamb_jbkc.iscale(0.5)
        #       lamb_jb.assign(0.0)
        #       t2.assign(0.0)

        if self.singles:
            # intermediates:
            # t_kc.l_ka
            tl_ca = t1.contract_two_to_two(
                "ab,ac->bc", lamb_jb, None, 1.0, True, select=tco
            )
            tmp_vv = tl_ca.new()
            # l_ic.t_kc
            tl_ik = lamb_jb.contract_two_to_two(
                "ab,cb->ac", t1, None, 1.0, True, select=tco
            )
            tmp_vv = tl_ca.new()
            tmp_oo = tl_ik.new()

        if self.singles:
            Lambda_ia.clear()
            #####1--------fia------
            Lambda_ia.iadd_slice(fock, 2.0, end0=self.nocc, begin1=self.nocc)
            #####----- ------------------
            #####3---- ---fim*lambda_am----
            # lic . fca
            lamb_jb.contract_two_to_two(
                "ab,bc->ac",
                fock,
                Lambda_ia,
                1.0,
                False,
                begin0=self.nocc,
                begin1=self.nocc,
                select=tco,
            )
            #####----- -------------------
            #####4---- -------------------
            # -lka . fki
            lamb_jb.contract_two_to_two(
                "ab,ac->cb",
                fock,
                Lambda_ia,
                -1.0,
                False,
                end0=self.nocc,
                end1=self.nocc,
                select=tco,
            )
            #####----- ------------------
            #####5---- ------------------
            # ic||ak lkc
            govvo.contract_two_to_two(
                "abcd,db->ac", lamb_jb, Lambda_ia, 2.0, False, select=tco
            )
            govov.contract_two_to_two(
                "abcd,cb->ad", lamb_jb, Lambda_ia, -1.0, False, select=tco
            )
            if False:
                #####----- ------------
                #####2---- ---2*Likac*tkc---- icak | iack
                govvo.contract_two_to_two(
                    "abcd,db->ac", t1, Lambda_ia, 4.0, False, select=tco
                )
                govvo.contract_two_to_two(
                    "abcd,dc->ab", t1, Lambda_ia, -2.0, False, select=tco
                )
                #####----- ------------------
                #####6---- ------------------
                # tkc . fic -> ki . lka
                t1.contract_two_to_two(
                    "ab,cb->ac",
                    fock,
                    tmp_oo,
                    1.0,
                    True,
                    end0=self.nocc,
                    begin1=self.nocc,
                    select=tco,
                )
                tmp_oo.contract_two_to_two(
                    "ab,ac->bc", lamb_jb, Lambda_ia, -1.0, False, select=tco
                )
                #####----- ------------------
                #####7---- -----------------
                # t_kc . fka -> ca . lic
                t1.contract_two_to_two(
                    "ab,ac->bc",
                    fock,
                    tmp_vv,
                    1.0,
                    True,
                    end0=self.nocc,
                    begin1=self.nocc,
                    select=tco,
                )
                lamb_jb.contract_two_to_two(
                    "ab,bc->ac", tmp_vv, Lambda_ia, -1.0, False, select=tco
                )
                #####----- -----------------
                #####8---- -----------------
                # ke||ca tkc -> ea . lie
                govvv.contract_two_to_two(
                    "abcd,ac->bd", t1, tmp_vv, 2.0, True, select=tco
                )
                govvv.contract_two_to_two(
                    "abcd,ad->bc", t1, tmp_vv, -1.0, False, select=tco
                )
                lamb_jb.contract_two_to_two(
                    "ab,bc->ac", tmp_vv, Lambda_ia, 1.0, False, select=tco
                )
                #####----- -----------------
                #####9---- -----------------
                gvooo.contract_two_to_two(
                    "abcd,ca->bd", t1, tmp_oo, -2.0, True, select=tco
                )
                gvooo.contract_two_to_two(
                    "abcd,da->bc", t1, tmp_oo, 1.0, False, select=tco
                )
                tmp_oo.contract_two_to_two(
                    "ab,ac->bc", lamb_jb, Lambda_ia, 1.0, False, select=tco
                )
                #####----- -----------------
                #####10--- -----------------
                # lkd . tkc -> dc
                lamb_jb.contract_two_to_two(
                    "ab,ac->bc", t1, tmp_vv, 1.0, True, select=tco
                )
                # id||ac . dc -> ia
                govvv.contract_two_to_two(
                    "abcd,bd->ac", tmp_vv, Lambda_ia, 2.0, False, select=tco
                )
                govvv.contract_two_to_two(
                    "abcd,bc->ad", tmp_vv, Lambda_ia, -1.0, False, select=tco
                )
                #####----- ---------------
                #####11--- ---------------
                # llc . tkc -> lk
                lamb_jb.contract_two_to_two(
                    "ab,cb->ac", t1, tmp_oo, 1.0, True, select=tco
                )
                # al||ik . lk
                gvooo.contract_two_to_two(
                    "abcd,bd->ca", tmp_oo, Lambda_ia, -2.0, False, select=tco
                )
                gvooo.contract_two_to_two(
                    "abcd,bc->da", tmp_oo, Lambda_ia, 1.0, False, select=tco
                )
            #####----- -------------
            #####12--- --------------
            # kl||ad (kdal) . tkcld -> ac . lic
            govvo.contract_four_to_two(
                "abcd,aedb->ce", t2p, tmp_vv, -2.0, True, select=tco
            )
            govvo.contract_four_to_two(
                "abcd,aedc->be", t2p, tmp_vv, 1.0, False, select=tco
            )
            lamb_jb.contract_two_to_two(
                "ab,cb->ac", tmp_vv, Lambda_ia, 1.0, False, select=tco
            )
            #####----- -------------
            #####13--- ------------
            # il||cd (idcl) . tkcld -> ik . lka
            govvo.contract_four_to_two(
                "abcd,ecdb->ae", t2p, tmp_oo, -2.0, True, select=tco
            )
            govvo.contract_four_to_two(
                "abcd,ebdc->ae", t2p, tmp_oo, 1.0, False, select=tco
            )
            tmp_oo.contract_two_to_two(
                "ab,bc->ac", lamb_jb, Lambda_ia, 1.0, False, select=tco
            )
            #####----- ------------
            #####14-15 ------------
            # tkc||ld . lkc
            t2l_ld = t2p.contract_two_to_two(
                "abcd,cd->ab", lamb_jb, None, 2.0, True, select=tco
            )
            t2p.contract_two_to_two(
                "abcd,ad->cb", lamb_jb, t2l_ld, -1.0, False, select=tco
            )
            # il||ad (idal) . t2l_ld
            govvo.contract_two_to_two(
                "abcd,db->ac", t2l_ld, Lambda_ia, 2.0, False, select=tco
            )
            govvo.contract_two_to_two(
                "abcd,dc->ab", t2l_ld, Lambda_ia, -1.0, False, select=tco
            )
            #####----- -----------
            #####16--- ------------
            govvv.contract_four_to_two(
                "abcd,edac->eb", lamb_jbkc, Lambda_ia, 2.0, False, select=tco
            )
            govvv.contract_four_to_two(
                "abcd,ecbd->eb", lamb_jbkc, Lambda_ia, -1.0, False, select=tco
            )  # art
            #####----- -------------
            #####17--- -------------
            gvooo.contract_four_to_two(
                "abcd,deca->be", lamb_jbkc, Lambda_ia, -2.0, False, select=tco
            )

            gvooo.contract_four_to_two(
                "abcd,ceda->be", lamb_jbkc, Lambda_ia, 1.0, False, select=tco
            )  # art
            if False:
                #####----- -------------
                #####18--- -------------
                tmpiakc = gvvvv.contract_four_to_four(
                    "abcd,ecfd->eafb", lamb_jbkc, None, 1.0, select=tco
                )
                tmpiakc.contract_two_to_two(
                    "abcd,cd->ab", t1, Lambda_ia, 1.0, False, select=tco
                )
                del tmpiakc
                #####----- -------------
                #####20--- -------------
                tmpincm = goooo.contract_two_to_four(
                    "abcd,de->abec", t1, None, 1.0, True, select=tco
                )
                lamb_jbkc.contract_four_to_two(
                    "abcd,eabc->ed", tmpincm, Lambda_ia, 1.0, False
                )
                del tmpincm
                #####----- -------------
                #####19--- -------------
                # idcl . tkc -> ikld . lkald
                tmpikld = govvo.contract_two_to_four(
                    "abcd,ec->aedb", t1, None, -1.0, True, select=tco
                )
                lamb_jbkc.contract_four_to_two(
                    "abcd,eacd->eb", tmpikld, Lambda_ia, 1.0, False, select=tco
                )
                # idlc . tkc -> ikld . llakd
                govov.contract_two_to_four(
                    "abcd,ed->aecb", t1, tmpikld, -1.0, True, select=tco
                )
                lamb_jbkc.contract_four_to_two(
                    "abcd,ecad->eb", tmpikld, Lambda_ia, 1.0, False, select=tco
                )
                del tmpikld
                #####----- -----------
                #####21--- -----------
                # licld . tkc -> ikld . kdal
                tmpikld = lamb_jbkc.contract_two_to_four(
                    "abcd,eb->aecd", t1, None, 1.0, True
                )  # lamb_icme -> ikem
                govvo.contract_four_to_two(
                    "abcd,eadb->ec",
                    tmpikld,
                    Lambda_ia,
                    -1.0,
                    False,
                    select=tco,
                )  # (Lkeam)
                # llcid . tkc -> ikld . kdla
                lamb_jbkc.contract_two_to_four(
                    "abcd,eb->cead", t1, tmpikld, 1.0, True
                )  # lamb_icme -> ikem
                govov.contract_four_to_two(
                    "abcd,eacb->ed",
                    tmpikld,
                    Lambda_ia,
                    -1.0,
                    False,
                    select=tco,
                )
                del tmpikld
            #####----- ----------
            #####22-23 -----------
            lamb_jbkc.contract_four_to_two(
                "abcd,abce->de", t2p, tmp_vv, -1.0, True, select=tco
            )
            tmp_vv.contract_two_to_two(
                "ab,cb->ca",
                fock,
                Lambda_ia,
                1.0,
                False,
                end0=self.nocc,
                begin1=self.nocc,
                select=tco,
            )
            #####-----  -----------
            #####24-25  -----------
            lamb_jbkc.contract_four_to_two(
                "abcd,abed->ce", t2p, tmp_oo, -1.0, True, select=tco
            )
            tmp_oo.contract_two_to_two(
                "ab,bc->ac",
                fock,
                Lambda_ia,
                1.0,
                False,
                end0=self.nocc,
                begin1=self.nocc,
                select=tco,
            )
            #####----- #---------
            #####26-27 #--------
            # llekc . tldkc -> ed
            lamb_jbkc.contract_four_to_two(
                "abcd,aecd->be", t2p, tmp_vv, 1.0, True, select=tco
            )
            # iead | iade . ed -> ia
            govvv.contract_two_to_two(
                "abcd,bd->ac", tmp_vv, Lambda_ia, 2.0, False, select=tco
            )
            govvv.contract_two_to_two(
                "abcd,dc->ab", tmp_vv, Lambda_ia, -1.0, False, select=tco
            )  # iad
            #####----- #---------
            #####28--- #--------
            # iecd . kcld -> iekl
            tmpiekl = govvv.contract_four_to_four(
                "abcd,ecfd->abef", t2p, None, 1.0, True, select=tco
            )
            # kale . iekl
            lamb_jbkc.contract_four_to_two(
                "abcd,edac->eb", tmpiekl, Lambda_ia, -1.0, False, select=tco
            )
            del tmpiekl
            #####----- #-------
            #####29--- #--------
            # icmd . kcld -> imkl
            tmpimkl = lamb_jbkc.contract_four_to_four(
                "abcd,ebfd->acef", t2p, None, 1.0, True, select=tco
            )
            # amkl . imkl
            gvooo.contract_four_to_two(
                "abcd,ebcd->ea", tmpimkl, Lambda_ia, 1.0, False, select=tco
            )
            del tmpimkl
            #####----- #-------
            #####30-31  --------
            # lmdkc . tldkc -> ml
            lamb_jbkc.contract_four_to_two(
                "abcd,ebcd->ae", t2p, tmp_oo, 1.0, True, select=tco
            )
            #####-----  --------
            # amil | ailm . ml -> ia
            gvooo.contract_two_to_two(
                "abcd,bd->ca", tmp_oo, Lambda_ia, -2.0, False, select=tco
            )
            gvooo.contract_two_to_two(
                "abcd,dc->ba", tmp_oo, Lambda_ia, 1.0, False, select=tco
            )
            del tmp_oo
            #####----- #-------
            #####32-34 --------
            # liekc . tlckd -> iedl
            tmpiedl = lamb_jbkc.contract_four_to_four(
                "abcd,edcf->abfe", t2p, None, 1.0, True, select=tco
            )
            # lade . iedl
            govvv.contract_four_to_two(
                "abcd,edca->eb", tmpiedl, Lambda_ia, -1.0, False, select=tco
            )
            #
            # licke . tlckd -> iedl
            lamb_jbkc.contract_four_to_four(
                "abcd,ebcf->adfe", t2p, tmpiedl, 1.0, True, select=tco
            )
            # lead . iedl
            govvv.contract_four_to_two(
                "abcd,ebda->ec", tmpiedl, Lambda_ia, -1.0, False, select=tco
            )
            # 35
            # liekc . tldkc -> iedl
            lamb_jbkc.contract_four_to_four(
                "abcd,efcd->abfe", t2p, tmpiedl, 1.0, True, select=tco
            )
            # lade | lead . iedl
            govvv.contract_four_to_two(
                "abcd,edca->eb", tmpiedl, Lambda_ia, 2.0, False, select=tco
            )
            govvv.contract_four_to_two(
                "abcd,ebda->ec", tmpiedl, Lambda_ia, -1.0, False, select=tco
            )
            del tmpiedl
            #####----- #-----------
            #####36-38 39-----
            #
            # lkamc . tlckd -> amld
            tmpamld = lamb_jbkc.contract_four_to_four(
                "abcd,edaf->bcef", t2p, None, 1.0, True, select=tco
            )
            # dmil . amld
            gvooo.contract_four_to_two(
                "abcd,ebda->ce", tmpamld, Lambda_ia, 1.0, False, select=tco
            )
            #
            # lmakc . tlckd -> amld
            lamb_jbkc.contract_four_to_four(
                "abcd,edcf->baef", t2p, tmpamld, 1.0, True, select=tco
            )
            # dmli . amld
            gvooo.contract_four_to_two(
                "abcd,ebca->de", tmpamld, Lambda_ia, 1.0, False, select=tco
            )
            #
            # lmakc . tldkc -> amld
            lamb_jbkc.contract_four_to_four(
                "abcd,efcd->baef", t2p, tmpamld, 1.0, True, select=tco
            )
            # dmli | dmil . amld
            gvooo.contract_four_to_two(
                "abcd,ebca->de", tmpamld, Lambda_ia, -2.0, False, select=tco
            )
            gvooo.contract_four_to_two(
                "abcd,ebda->ce", tmpamld, Lambda_ia, 1.0, False, select=tco
            )
            del tmpamld
            #####-----  ----------
            #####40---  ----------
            if False:
                # l_kk^ad c_k^c
                tmpadc = lamb_jbkc.contract_two_to_three(
                    "abac,ad->bcd", args[1], None, 1.0, True
                )
                # adc tld
                tmpalc = tmpadc.contract_two_to_three(
                    "abc,eb->aec", t1, None, 1.0, True, select=tco
                )
                # ilcc alc
                gpqrr.contract_three_to_two(
                    "abc,dbc->ad",
                    tmpalc,
                    Lambda_ia,
                    1.0,
                    False,
                    end0=self.nocc,
                    end1=self.nocc,
                    begin2=self.nocc,
                    select=tco,
                )
                #####-----  ----------
                #####41---  --------
                tmpidkc = lamb_jbkc.contract_two_to_four(
                    "abcd,cd->abcd", args[1], None, 1.0, True
                )
                tmpilkc = tmpidkc.contract_two_to_four(
                    "abcd,eb->aecd", t1, None, 1.0, True, select=tco
                )
                del tmpidkc
                govvo.contract_four_to_two(
                    "abcd,eadb->ec",
                    tmpilkc,
                    Lambda_ia,
                    -2.0,
                    False,
                    select=tco,
                )
                govvo.contract_four_to_two(
                    "abcd,eadc->eb", tmpilkc, Lambda_ia, 1.0, False, select=tco
                )
                #####-----  ---------
                #####42---  ---------
                # licke . ckc -> icke
                tmpicke = lamb_jbkc.contract_two_to_four(
                    "abcd,cb->abcd", args[1], None, 1.0, True
                )
                # icke . tme -> ickm
                tmpickm = tmpicke.contract_two_to_four(
                    "abcd,ed->abce", t1, None, 1.0, True, select=tco
                )
                # ickm . ma|ck
                govvo.contract_four_to_two(
                    "abcd,ecda->eb", tmpickm, Lambda_ia, 1.0, False, select=tco
                )
                del tmpickm
                #
                # liekc . ckc -> iekc
                lamb_jbkc.contract_two_to_four(
                    "abcd,cd->abcd", args[1], tmpicke, 1.0, True
                )
                # iekc . tme -> ikcm
                tmpikcm = tmpicke.contract_two_to_four(
                    "abcd,eb->acde", t1, None, 1.0, True, select=tco
                )
                del tmpicke
                # ikcm . mc|ak
                govvo.contract_four_to_two(
                    "abcd,edba->ec", tmpikcm, Lambda_ia, 1.0, False, select=tco
                )
                del tmpikcm
                #####-----  ---------
                #####43---  ---------
                # l_kk^cd c_k^c
                tmpcd = lamb_jbkc.contract_two_to_two(
                    "abac,ab->bc", args[1], None, 1.0, True
                )
                # kcd tld
                tmpcl = tmpcd.contract_two_to_two(
                    "ab,cb->ac", t1, None, 1.0, True, select=tco
                )
                # il||ac cl
                govvo.contract_two_to_two(
                    "abcd,bd->ac", tmpcl, Lambda_ia, -2.0, False, select=tco
                )
                govvo.contract_two_to_two(
                    "abcd,cd->ab", tmpcl, Lambda_ia, 1.0, False, select=tco
                )
                #####-----  -----------
                #####44---  ----------
                tmpic = govvo.contract_two_to_two(
                    "abcd,db->ac", t1, None, -2.0, True, select=tco
                )
                govvo.contract_two_to_two(
                    "abcd,dc->ab", t1, tmpic, 1.0, False, select=tco
                )
                # ic.ca
                tmpic.contract_two_to_two(
                    "ab,bc->ac", tmpcd, Lambda_ia, 1.0, False, select=tco
                )
                #####45---  ---------
                tmpik = lamb_jbkc.contract_two_to_two(
                    "abcb,cb->ac", args[1], None, 1.0, True
                )
                tmpak = govvo.contract_two_to_two(
                    "abcd,db->ca", t1, None, -2.0, True, select=tco
                )
                govvo.contract_two_to_two(
                    "abcd,ab->cd", t1, tmpak, 1.0, False, select=tco
                )
                # ik.ak
                tmpik.contract_two_to_two(
                    "ab,cb->ac", tmpak, Lambda_ia, 1.0, False, select=tco
                )
                #####-----  ----------
                #####46---  ----------
                tmpilk = lamb_jbkc.contract_two_to_three(
                    "abcb,db->acd", args[1], None, 1.0, True
                )
                # ilk.ld
                tmpidk = tmpilk.contract_two_to_three(
                    "abc,bd->adc", t1, None, 1.0, True
                )
                # ad|kk . idk
                gpqrr.contract_three_to_two(
                    "abc,dbc->da",
                    tmpidk,
                    Lambda_ia,
                    1.0,
                    False,
                    begin0=self.nocc,
                    begin1=self.nocc,
                    end2=self.nocc,
                    select=tco,
                )
                #####-----  -----------
                #####47---  ----------
                # l_kl^cc c_k^c
                tmplk = lamb_jbkc.contract_two_to_two(
                    "abcb,cb->ac", args[1], None, 1.0, True
                )
                # tld.lk
                tmpdk = t1.contract_two_to_two(
                    "ab,ac->bc", tmplk, None, 1.0, True, select=tco
                )
                # ik||ad dk
                govvo.contract_two_to_two(
                    "abcd,bd->ac", tmpdk, Lambda_ia, -2.0, False, select=tco
                )
                govvo.contract_two_to_two(
                    "abcd,cd->ab", tmpdk, Lambda_ia, 1.0, False, select=tco
                )
                #####-----  ----------
                #####48---  ----------
                tmpalkc = lamb_jbkc.contract_two_to_four(
                    "abcd,cd->bacd", args[1], None, 1.0, True
                )
                # ik||dc. tld
                tmpilkc = govvo.contract_two_to_four(
                    "abcd,ec->aedb", t1, None, -2.0, True, select=tco
                )
                govvo.contract_two_to_four(
                    "abcd,eb->aedc", t1, tmpilkc, 1.0, False, select=tco
                )
                tmpilkc.contract_four_to_two(
                    "abcd,ebcd->ae", tmpalkc, Lambda_ia, 1.0, False, select=tco
                )
                del tmpalkc
                #####-----  ---------
                #####49---  --------
                # llamd . ld -> adlm
                tmpadlm = lamb_jbkc.contract_two_to_four(
                    "abcd,ad->bdac", args[1], None, 1.0, True
                )
                # adlm . tme -> adle
                tmpadle = tmpadlm.contract_two_to_four(
                    "abcd,de->abce", t1, None, 1.0, True, select=tco
                )
                del tmpadlm
                # ie|dl . adle
                govvo.contract_four_to_two(
                    "abcd,ecdb->ae", tmpadle, Lambda_ia, 1.0, False, select=tco
                )
                del tmpadle
                #
                # lmald . ld -> amld
                tmpamld = lamb_jbkc.contract_two_to_four(
                    "abcd,cd->bacd", args[1], None, 1.0, True
                )
                # amld . tme -> adel
                tmpadel = tmpamld.contract_two_to_four(
                    "abcd,be->adec", t1, None, 1.0, True, select=tco
                )
                del tmpamld
                # id|el . adel
                govvo.contract_four_to_two(
                    "abcd,ebcd->ae", tmpadel, Lambda_ia, 1.0, False, select=tco
                )
                del tmpadel
        # ----------------END of Lambda_ia----------------------------------------------
        # ----------------Lambda_iajb---------------------------------------------------
        # 50------------lamb_jb--
        # Lijab <--ibaj | iabj
        # This is 100% correct
        govvo.slice("abcd->acdb", Lambda_iajb, 2.0, True)
        govvo.slice("abcd->abdc", Lambda_iajb, -1.0, False)
        if self.singles:
            #####----------------
            #####[51...54]
            Lambda_iajb.iadd_expand_two_to_four(
                "ab,cd->abcd",
                fock,
                2.0,
                lamb_jb,
                end0=self.nocc,
                begin1=self.nocc,
            )
            Lambda_iajb.iadd_expand_two_to_four(
                "ab,cd->adcb",
                fock,
                -1.0,
                lamb_jb,
                end0=self.nocc,
                begin1=self.nocc,
            )
            #####---- -------------
            #####55-- -------------
            # ->jeba | jeab<- ie
            govvv.contract_two_to_four(
                "abcd,eb->edac", lamb_jb, Lambda_iajb, 2.0, False, select=tco
            )
            govvv.contract_two_to_four(
                "abcd,eb->ecad", lamb_jb, Lambda_iajb, -1.0, False, select=tco
            )
            #####56-- ----------------
            # amij | amji  . mb
            gvooo.contract_two_to_four(
                "abcd,be->cade", lamb_jb, Lambda_iajb, -2.0, False, select=tco
            )
            gvooo.contract_two_to_four(
                "abcd,be->dace", lamb_jb, Lambda_iajb, 1.0, False, select=tco
            )
            if False:
                #####57-- ---------------
                # ik||ac<-- icak | iack . tkc . ljb
                tmpia = govvo.contract_two_to_two(
                    "abcd,db->ac", t1, None, 2.0, True, select=tco
                )
                govvo.contract_two_to_two(
                    "abcd,dc->ab", t1, tmpia, -1.0, False, select=tco
                )
                Lambda_iajb.iadd_expand_two_to_four(
                    "ab,cd->abcd", tmpia, 1.0, lamb_jb
                )
                Lambda_iajb.iadd_expand_two_to_four(
                    "ab,cd->adcb", lamb_jb, -0.5, tmpia
                )
                del tmpia
                #####---- -----------
                #####58-- -----------
                # ikab | ikba->ibak|iabk . tkc . ljc
                tmpjk = lamb_jb.contract_two_to_two(
                    "ab,cb->ac", t1, None, 1.0, True, select=tco
                )
                govvo.contract_two_to_four(
                    "abcd,ed->aceb",
                    tmpjk,
                    Lambda_iajb,
                    -2.0,
                    False,
                    select=tco,
                )
                govvo.contract_two_to_four(
                    "abcd,ed->abec", tmpjk, Lambda_iajb, 1.0, False, select=tco
                )
                del tmpjk
                #####---- ------------
                #####59-- -----------
                # ijcb|ijbc -> ibcj|icbj . tkc . lka
                tmpac = lamb_jb.contract_two_to_two(
                    "ab,ac->bc", t1, None, 1.0, True, select=tco
                )
                govvo.contract_two_to_four(
                    "abcd,ec->aedb",
                    tmpac,
                    Lambda_iajb,
                    -2.0,
                    False,
                    select=tco,
                )
                govvo.contract_two_to_four(
                    "abcd,eb->aedc", tmpac, Lambda_iajb, 1.0, False, select=tco
                )
                #####67---------------
                # tkc fjc . liakb
                tmpik = t1.contract_two_to_two(
                    "ab,cb->ca",
                    fock,
                    None,
                    1.0,
                    True,
                    end0=self.nocc,
                    begin1=self.nocc,
                    select=tco,
                )
                lamb_jbkc.contract_two_to_four(
                    "abcd,ec->abed",
                    tmpik,
                    Lambda_iajb,
                    -1.0,
                    False,
                    select=tco,
                )
                del tmpik
                #####68---- ----------
                # tkc fkb . liajc
                tmpac = t1.contract_two_to_two(
                    "ab,ac->cb",
                    fock,
                    None,
                    1.0,
                    True,
                    end0=self.nocc,
                    begin1=self.nocc,
                    select=tco,
                )
                lamb_jbkc.contract_two_to_four(
                    "abcd,ed->abce",
                    tmpac,
                    Lambda_iajb,
                    -1.0,
                    False,
                    select=tco,
                )
                del tmpac
                #####69---- ----------
                # abkd tkc licjd
                tmpijdk = lamb_jbkc.contract_two_to_four(
                    "abcd,eb->acde", t1, None, 1.0, True, select=tco
                )
                # kdab
                govvv.contract_four_to_four(
                    "abcd,efba->ecfd",
                    tmpijdk,
                    Lambda_iajb,
                    -1.0,
                    False,
                    select=tco,
                )
                del tmpijdk
                #####70---- -----------
                # cmij . kc  . lkalb
                tmpijkm = gvooo.contract_two_to_four(
                    "abcd,ea->cdeb", t1, None, 1.0, True, select=tco
                )
                # kamb . ijkm
                lamb_jbkc.contract_four_to_four(
                    "abcd,efac->ebfd",
                    tmpijkm,
                    Lambda_iajb,
                    1.0,
                    False,
                    select=tco,
                )
                del tmpijkm
                #####71---- ----------
                # kfca | kfac. tkc . lifjb
                tmpaf = govvv.contract_two_to_two(
                    "abcd,ac->bd", t1, None, 2.0, True
                )
                govvv.contract_two_to_two(
                    "abcd,ad->cb", t1, tmpaf, -1.0, False
                )
                lamb_jbkc.contract_two_to_four(
                    "abcd,eb->aecd", tmpaf, Lambda_iajb, 1.0, False, select=tco
                )
                #####72---- ----------
                # cikn/cnik kc -> in najb
                tmpin = gvooo.contract_two_to_two(
                    "abcd,ca->bd", t1, None, -2.0, True, select=tco
                )
                gvooo.contract_two_to_two(
                    "abcd,da->cb", t1, tmpin, 1.0, False, select=tco
                )
                # najb
                lamb_jbkc.contract_two_to_four(
                    "abcd,ea->ebcd", tmpin, Lambda_iajb, 1.0, False, select=tco
                )
                del tmpin
                #####73---- ---------
                # ljalc.tkc -> jalk
                tmpjalk = lamb_jbkc.contract_two_to_four(
                    "abcd,ed->abce", t1, None, 1.0, True, select=tco
                )
                # blik . jalk -> iajb
                gvooo.contract_four_to_four(
                    "abcd,efbd->cfea",
                    tmpjalk,
                    Lambda_iajb,
                    1.0,
                    False,
                    select=tco,
                )
                del tmpjalk
                #####74---- --------
                # anik.tkc
                tmpianc = gvooo.contract_two_to_four(
                    "abcd,de->cabe", t1, None, -2.0, True, select=tco
                )
                gvooo.contract_two_to_four(
                    "abcd,ce->dabe", t1, tmpianc, 1.0, False, select=tco
                )
                # ianc jbnc
                tmpianc.contract_four_to_four(
                    "abcd,efcd->abef",
                    lamb_jbkc,
                    Lambda_iajb,
                    1.0,
                    False,
                    select=tco,
                )
                del tmpianc
                #####78---- ----------
                # jcna . tkc -> jnak . bikn
                tmpjnak = lamb_jbkc.contract_two_to_four(
                    "abcd,eb->acde", t1, None, 1.0, True, select=tco
                )
                gvooo.contract_four_to_four(
                    "abcd,edfc->bfea",
                    tmpjnak,
                    Lambda_iajb,
                    1.0,
                    False,
                    select=tco,
                )
                del tmpjnak
                #####75----#-----------
                # idbc . tkc -> ibkd . ljakd
                tmpibkd = govvv.contract_two_to_four(
                    "abcd,ed->aceb", t1, None, 1.0, True, select=tco
                )
                lamb_jbkc.contract_four_to_four(
                    "abcd,efcd->ebaf",
                    tmpibkd,
                    Lambda_iajb,
                    -1.0,
                    False,
                    select=tco,
                )
                #####76----#----------
                # ibcf . tkc -> ibfk . jfka
                tmpjbfk = govvv.contract_two_to_four(
                    "abcd,ec->abde", t1, None, 1.0, True, select=tco
                )
                lamb_jbkc.contract_four_to_four(
                    "abcd,efbc->edaf",
                    tmpjbfk,
                    Lambda_iajb,
                    -1.0,
                    False,
                    select=tco,
                )
                del tmpjbfk
                #####77---- ----------
                # ieac/iace . tkc -> iake . jbke
                tmpiake = govvv.contract_two_to_four(
                    "abcd,ed->aceb", t1, None, 2.0, True, select=tco
                )
                govvv.contract_two_to_four(
                    "abcd,ec->abed", t1, tmpiake, -1.0, False, select=tco
                )
                tmpiake.contract_four_to_four(
                    "abcd,efcd->abef",
                    lamb_jbkc,
                    Lambda_iajb,
                    1.0,
                    False,
                    select=tco,
                )
                del tmpiake
        # 60---------------
        # fbc liajc
        lamb_jbkc.contract_two_to_four(
            "abcd,ed->abce",
            fock,
            Lambda_iajb,
            1.0,
            False,
            begin4=self.nocc,
            begin5=self.nocc,
            select=tco,
        )
        # 61---------------
        # fjk liakb
        lamb_jbkc.contract_two_to_four(
            "abcd,ec->abed",
            fock,
            Lambda_iajb,
            -1.0,
            False,
            end4=self.nocc,
            end5=self.nocc,
            select=tco,
        )
        # 62---------------
        # abcd licjd
        gvvvv.contract_four_to_four(
            "abcd,ecfd->eafb", lamb_jbkc, Lambda_iajb, 0.5, False, select=tco
        )
        # 63---------------
        # ijkl lkalb
        goooo.contract_four_to_four(
            "abcd,cedf->aebf", lamb_jbkc, Lambda_iajb, 0.5, False, select=tco
        )
        # 64---------------
        # ci||ka jbkc <- icak | icka
        govvo.contract_four_to_four(
            "abcd,efdb->acef", lamb_jbkc, Lambda_iajb, 2.0, False, select=tco
        )
        govov.contract_four_to_four(
            "abcd,efcb->adef", lamb_jbkc, Lambda_iajb, -1.0, False, select=tco
        )
        # 65----------------
        # icbk jakc
        govvo.contract_four_to_four(
            "abcd,efdb->afec", lamb_jbkc, Lambda_iajb, -1.0, False, select=tco
        )
        # 66----------------
        # jena . lienb -> iajb
        govov.contract_four_to_four(
            "abcd,ebcf->edaf", lamb_jbkc, Lambda_iajb, -1.0, False, select=tco
        )
        # ---------------------
        # 79---- ----------
        # kald . kcld -> ac . ibcj/icbj
        tmpac = lamb_jbkc.contract_four_to_two(
            "abcd,aecd->be", t2p, None, 1.0, True, select=tco
        )
        govvo.contract_two_to_four(
            "abcd,ec->aedb", tmpac, Lambda_iajb, -2.0, False, select=tco
        )
        govvo.contract_two_to_four(
            "abcd,eb->aedc", tmpac, Lambda_iajb, 1.0, False, select=tco
        )
        # 80---- -----------
        # icld . kcld -> ik . jabk/jbak
        tmpik = lamb_jbkc.contract_four_to_two(
            "abcd,ebcd->ae", t2p, None, 1.0, True, select=tco
        )
        govvo.contract_two_to_four(
            "abcd,ed->ebac", tmpik, Lambda_iajb, -2.0, False, select=tco
        )
        govvo.contract_two_to_four(
            "abcd,ed->ecab", tmpik, Lambda_iajb, 1.0, False, select=tco
        )
        # 81----------------
        # idcj . tkcld -> ijkl . kalb
        tmpijkl = govvo.contract_four_to_four(
            "abcd,ecfb->adef", t2p, None, 1.0, True, select=tco
        )
        lamb_jbkc.contract_four_to_four(
            "abcd,efac->ebfd", tmpijkl, Lambda_iajb, 0.5, False, select=tco
        )
        # 82----------------
        # licjd . tkcld -> ijkl . kbal
        lamb_jbkc.contract_four_to_four(
            "abcd,ebfd->acef", t2p, tmpijkl, 1.0, True, select=tco
        )
        govvo.contract_four_to_four(
            "abcd,efad->ecfb", tmpijkl, Lambda_iajb, 0.5, False, select=tco
        )
        # 83----------------
        # ladk/ldak . tkcld -> ac . icjb
        tmpac = govvo.contract_four_to_two(
            "abcd,deac->be", t2p, None, -2.0, True, select=tco
        )
        govvo.contract_four_to_two(
            "abcd,aedc->be", t2p, tmpac, 1.0, False, select=tco
        )
        lamb_jbkc.contract_two_to_four(
            "abcd,eb->aecd", tmpac, Lambda_iajb, 1.0, False, select=tco
        )
        # 84----------------
        # idcl/icdl . tkcld -> ik . kajb
        tmpik = govvo.contract_four_to_two(
            "abcd,ecdb->ae", t2p, None, -2.0, True, select=tco
        )
        govvo.contract_four_to_two(
            "abcd,ebdc->ae", t2p, tmpik, 1.0, False, select=tco
        )
        lamb_jbkc.contract_two_to_four(
            "abcd,ea->ebcd", tmpik, Lambda_iajb, 1.0, False, select=tco
        )
        # 85----------------
        # jacl . tkcld -> jakd . idkb
        # jlda -> jadl . tkdlc -> jakc | lkbic . jakc -> iajb
        tmpjakd = govvo.contract_four_to_four(
            "abcd,ecdf->abef", t2p, None, 1.0, True, select=tco
        )
        lamb_jbkc.contract_four_to_four(
            "abcd,efad->cfeb", tmpjakd, Lambda_iajb, 1.0, False, select=tco
        )
        del tmpjakd
        # 86----------------
        # icbk | ibck . tkcld -> ibld . jald
        tmpibld = govvo.contract_four_to_four(
            "abcd,efdb->acef", t2p, None, -2.0, True, select=tco
        )
        govvo.contract_four_to_four(
            "abcd,efdc->abef", t2p, tmpibld, 1.0, False, select=tco
        )
        lamb_jbkc.contract_four_to_four(
            "abcd,efcd->ebaf", tmpibld, Lambda_iajb, 1.0, False, select=tco
        )
        del tmpibld
        # 87---------------
        # idbl . tlckd -> ibkc . ljakc
        tmpibkc = govvo.contract_four_to_four(
            "abcd,defb->acfe", t2p, None, 1.0, True, select=tco
        )
        lamb_jbkc.contract_four_to_four(
            "abcd,efcd->ebaf", tmpibkc, Lambda_iajb, 1.0, False, select=tco
        )
        del tmpibkc
        # 88-part----------
        # jcbk/jbck . tlckd -> jbld . iald
        tmpjakd = govvo.contract_four_to_four(
            "abcd,ebdf->acef", t2p, None, -2.0, True, select=tco
        )
        govvo.contract_four_to_four(
            "abcd,ecdf->abef", t2p, tmpjakd, 1.0, False, select=tco
        )
        # 89-part----------
        # jcbk/jbck . tkcld -> jbld . iald
        govvo.contract_four_to_four(
            "abcd,dbef->acef", t2p, tmpjakd, 4.0, False, select=tco
        )
        govvo.contract_four_to_four(
            "abcd,dcef->abef", t2p, tmpjakd, -2.0, False, select=tco
        )
        # 88+89-------------
        lamb_jbkc.contract_four_to_four(
            "abcd,efcd->abef", tmpjakd, Lambda_iajb, 1.0, False, select=tco
        )
        del tmpjakd
        # ____________________________SUMMATION_____________________________

        Lambda_iajb.iadd_transpose((2, 3, 0, 1))
        #       Lambda_iajb.iscale(0.5)
        #       Lambda_iajb.assign(0.0)
        #       Lambda_ia.assign(0.0)
        #
        # Set all diagonal amplitudes zero
        #
        if self.multireference:
            ind1, ind2 = np.indices((self.nocc, self.nvirt))
            indices = [ind1, ind2, ind1, ind2]
            Lambda_iajb.assign(0.0, indices)
            # clean up pair amplitudes in t2
            t2.assign(0.0, indices)
            lamb_jbkc.assign(0.0, indices)

        if self.singles:
            t = np.hstack(
                (
                    Lambda_ia._array.ravel(order="C"),
                    Lambda_iajb.get_triu().ravel(order="C"),
                )
            )
            if self.solver == "krylov":
                return t
            else:
                raise NotImplementedError
                out = self.lf.create_one_index(t.shape[0])
                out.assign(t)
                return out
        t = np.hstack(Lambda_iajb.get_triu().ravel(order="C"))
        if self.solver == "krylov":
            return t
        else:
            raise NotImplementedError
            out = self.lf.create_one_index(t.shape[0])
            out.assign(t)
            return out


class RLCCD(RLCC):
    @property
    def nunknowns(self):
        """Single excitations"""
        return self.nocc * self.nvirt * (self.nocc * self.nvirt + 1) // 2

    @property
    def singles(self):
        """Single excitations"""
        return False


class RHFLCCD(RLCCD):
    def print_method(self):
        log("Linearized Coupled Cluster correction on top of RHF")

    def get_model(self):
        return "RHFLCCD"

    @property
    def multireference(self):
        """multireference wavefunction"""
        return False


class RpCCDLCCD(RLCCD):
    def print_method(self):
        log("Linearized Coupled Cluster correction on top of RpCCD")

    def get_model(self):
        return "RpCCDLCCD"

    @property
    def multireference(self):
        """multireference wavefunction"""
        return True


class RLCCSD(RLCC):
    @property
    def nunknowns(self):
        """Single excitations"""
        return (
            self.nocc * self.nvirt * (self.nocc * self.nvirt + 1) // 2
            + self.nocc * self.nvirt
        )

    @property
    def singles(self):
        """Single excitations"""
        return True


class RHFLCCSD(RLCCSD):
    def get_model(self):
        return "RHFLCCSD"

    @property
    def multireference(self):
        """multireference wavefunction"""
        return False


class RpCCDLCCSD(RLCCSD):
    def get_model(self):
        return "RpCCDLCCSD"

    @property
    def multireference(self):
        """multireference wavefunction"""
        return True
