# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file


import numpy as np
import pytest
from scipy import optimize as opt

import pybest.gbasis.gobasis as pybest_basis
from pybest.gbasis import get_gobasis
from pybest.gbasis.gobasis import (
    compute_kinetic,
    compute_nuclear,
    compute_eri,
    compute_cholesky_eri,
    compute_nuclear_repulsion,
    compute_overlap,
)
from pybest.linalg.dense import DenseLinalgFactory
from pybest.linalg.cholesky import CholeskyLinalgFactory
from pybest.scf import AufbauOccModel
from pybest.cc.rlccsd import RHFLCCD, RHFLCCSD, RpCCDLCCD, RpCCDLCCSD
from pybest.geminals.rpccd import ROOpCCD
from pybest.wrappers.hf import RHF
from pybest.io.iodata import IOData
from pybest.context import context
from pybest.test.common import numpy_seed, fda_1order
from pybest.orbital_utils import transform_integrals


@pytest.mark.slow
def test_lccd_one_dm():
    fn_xyz = context.get_fn("test/li2.xyz")
    mol = IOData.from_file(fn_xyz)
    obasis = get_gobasis("cc-pvdz", fn_xyz, print_basis=False)
    lf = DenseLinalgFactory(obasis.nbasis)
    occ_model = AufbauOccModel(3)
    orb = lf.create_orbital(obasis.nbasis)
    olp = compute_overlap(obasis)
    kin = compute_kinetic(obasis)
    na = compute_nuclear(obasis)
    external = {"nn": compute_nuclear_repulsion(obasis)}
    er = compute_eri(obasis)

    rhf = RHF(lf, occ_model)
    rhf_ = rhf(kin, na, er, external, orb, olp)

    one = lf.create_two_index(obasis.nbasis, label="one")
    one.iadd(kin)
    one.iadd(na)

    # Do AP1roG optimization:
    geminal_solver = ROOpCCD(lf, occ_model)
    ap1rog = geminal_solver(
        one,
        er,
        orb,
        olp,
        e_core=external["nn"],
        checkpoint=-1,
        maxiter={"orbiter": 1000},
    )

    lccd_solver = RpCCDLCCD(lf, occ_model)
    lccd = lccd_solver(
        one, er, ap1rog, tco="td", l=True, memoryopt=False, tthreshold=1e-16
    )

    one_mo_ = lf.create_two_index()
    one_mo_.assign_two_index_transform(one, orb)
    two_mo = []
    output = geminal_solver.lf.create_four_index()
    output.assign_four_index_transform(er, orb, orb, orb, orb, "tensordot")
    two_mo.append(output)

    def fun(x):
        one_mo = []
        one_mo.append(geminal_solver.lf.create_two_index())
        one_mo[0].assign(x.reshape(28, 28))
        # pCCD
        geminal_solver.clear_aux_matrix()
        geminal_solver.update_aux_matrix("scf", two_mo, one_mo)
        # eref from pCCD
        eref_pccd = geminal_solver.compute_total_energy()
        eref_hf = geminal_solver.compute_reference_energy()
        t_p = geminal_solver.geminal
        # LCCSD
        lccd_solver.clear_aux_matrix()
        lccd_solver.update_aux_matrix(
            one_mo[0], two_mo[0], t_p, deletemo=False
        )
        l2 = lccd_solver.l_amplitudes[0].get_triu().ravel(order="C")
        coeff = l2
        t2 = lccd_solver.amplitudes[0].get_triu().ravel(order="C")
        coefft = t2

        e_lccd = lccd_solver.calculate_energy(coefft)
        lagrangian = e_lccd["e_corr"] + eref_pccd

        lagrangian += (
            np.dot(coeff, lccd_solver.vfunction(coefft, [0, ap1rog.t_p])),
        )
        return lagrangian

    def fun_deriv(x):

        # LCCD
        lccd_solver_ = RpCCDLCCD(lf, occ_model)
        lccd = lccd_solver_(
            one,
            er,
            ap1rog,
            tco="td",
            l=True,
            memoryopt=False,
            tthreshold=1e-16,
        )
        # Get 1dm
        onedm_lccd = lccd_solver_.one_dm_pq
        onedm_lccd.iscale(2.0)

        # SUM
        a = np.zeros((28, 28))
        a[:] += onedm_lccd._array

        return a.ravel()

    x = one_mo_._array.ravel()
    with numpy_seed():
        dxs = []
        for i in range(5):
            dxs_ = np.random.rand(28, 28) * 0.001
            dxs.append(((dxs_ + dxs_.T) / 2).ravel())
        dxs = np.array(dxs)
        fda_1order(fun, fun_deriv, x, dxs)


@pytest.mark.slow
def test_lccsd_one_dm():
    fn_xyz = context.get_fn("test/li2.xyz")
    mol = IOData.from_file(fn_xyz)
    obasis = get_gobasis("cc-pvdz", fn_xyz, print_basis=False)
    lf = DenseLinalgFactory(obasis.nbasis)
    occ_model = AufbauOccModel(3)
    orb = lf.create_orbital(obasis.nbasis)
    olp = compute_overlap(obasis)
    kin = compute_kinetic(obasis)
    na = compute_nuclear(obasis)
    external = {"nn": compute_nuclear_repulsion(obasis)}
    er = compute_eri(obasis)

    rhf = RHF(lf, occ_model)
    rhf_ = rhf(kin, na, er, external, orb, olp)

    one = lf.create_two_index(obasis.nbasis, label="one")
    one.iadd(kin)
    one.iadd(na)

    # Do AP1roG optimization:
    geminal_solver = ROOpCCD(lf, occ_model)
    ap1rog = geminal_solver(
        one,
        er,
        orb,
        olp,
        e_core=external["nn"],
        checkpoint=-1,
        maxiter={"orbiter": 1000},
    )

    lccsd_solver = RpCCDLCCSD(lf, occ_model)
    lccsd = lccsd_solver(
        one, er, ap1rog, tco="td", l=True, memoryopt=False, tthreshold=1e-16
    )

    one_mo_ = lf.create_two_index()
    one_mo_.assign_two_index_transform(one, orb)
    two_mo = []
    output = geminal_solver.lf.create_four_index()
    output.assign_four_index_transform(er, orb, orb, orb, orb, "tensordot")
    two_mo.append(output)

    def fun(x):
        one_mo = []
        one_mo.append(geminal_solver.lf.create_two_index())
        one_mo[0].assign(x.reshape(28, 28))
        # pCCD
        geminal_solver.clear_aux_matrix()
        geminal_solver.update_aux_matrix("scf", two_mo, one_mo)
        # eref from pCCD
        eref_pccd = geminal_solver.compute_total_energy()
        eref_hf = geminal_solver.compute_reference_energy()
        t_p = geminal_solver.geminal
        # LCCSD
        lccsd_solver.clear_aux_matrix()
        lccsd_solver.update_aux_matrix(
            one_mo[0], two_mo[0], t_p, deletemo=False
        )
        l1 = lccsd_solver.l_amplitudes[0]._array.ravel(order="C")
        l2 = lccsd_solver.l_amplitudes[1].get_triu().ravel(order="C")
        coeff = np.hstack((l1, l2))

        t1 = lccsd_solver.amplitudes[0]._array.ravel(order="C")
        t2 = lccsd_solver.amplitudes[1].get_triu().ravel(order="C")
        coefft = np.hstack((t1, t2))

        e_lccsd = lccsd_solver.calculate_energy(coefft)
        lagrangian = e_lccsd["e_corr"] + eref_pccd

        lagrangian += (
            np.dot(coeff, lccsd_solver.vfunction(coefft, [0, ap1rog.t_p])),
        )
        return lagrangian

    def fun_deriv(x):

        # LCCSD
        lccsd_solver_ = RpCCDLCCSD(lf, occ_model)
        lccsd = lccsd_solver_(
            one,
            er,
            ap1rog,
            tco="td",
            l=True,
            memoryopt=False,
            tthreshold=1e-16,
        )
        # Get 1dm
        onedm_lccsd = lccsd_solver_.one_dm_pq
        onedm_lccsd.iscale(2.0)

        # SUM
        a = np.zeros((28, 28))
        a[:] += onedm_lccsd._array

        return a.ravel()

    x = one_mo_._array.ravel()
    with numpy_seed():
        dxs = []
        for i in range(5):
            dxs_ = np.random.rand(28, 28) * 0.001
            dxs.append(((dxs_ + dxs_.T) / 2).ravel())
        dxs = np.array(dxs)
        fda_1order(fun, fun_deriv, x, dxs)
