# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file

import numpy as np
import pytest

import pybest.gbasis.gobasis as pybest_basis
from pybest.gbasis import get_gobasis
from pybest.gbasis.gobasis import (
    compute_kinetic,
    compute_nuclear,
    compute_eri,
    compute_cholesky_eri,
    compute_nuclear_repulsion,
    compute_overlap,
)
from pybest.linalg.dense import DenseLinalgFactory
from pybest.linalg.cholesky import CholeskyLinalgFactory
from pybest.scf import AufbauOccModel
from pybest.cc.rccs import RHFCCS, RpCCDCCS
from pybest.geminals.rpccd import RpCCD
from pybest.io.iodata import IOData
from pybest.context import context


def prepare_mol(
    basis, linalg, filen="test/hf.xyz", orbfilen="test/hf_ap1rog.txt"
):
    fn_xyz = context.get_fn(filen)
    obasis = get_gobasis(basis, fn_xyz, print_basis=False)
    lf = linalg(obasis.nbasis)
    olp = compute_overlap(obasis)
    kin = compute_kinetic(obasis)
    na = compute_nuclear(obasis)
    one = kin.copy()
    one.iadd(na)
    if isinstance(lf, CholeskyLinalgFactory):
        er = compute_cholesky_eri(obasis, threshold=1e-8)
    elif isinstance(lf, DenseLinalgFactory):
        er = compute_eri(obasis)
    external = compute_nuclear_repulsion(obasis)

    fn_orb = context.get_fn(orbfilen)
    orb_ = np.fromfile(fn_orb, sep=",").reshape(obasis.nbasis, obasis.nbasis)
    orb_a = lf.create_orbital()
    orb_a._coeffs = orb_

    occ_model = AufbauOccModel(5)

    return lf, olp, orb_a, one, er, external, occ_model


core_set_hf = [
    (0, RHFCCS, {"e_ref": -100.00939084, "e_tot": -100.00988281,}),
    (1, RHFCCS, {"e_ref": -100.00939084, "e_tot": -100.00988271,}),
]

core_set_pccd = [
    (0, RpCCDCCS, {"e_ref": -100.08906717, "e_tot": -100.08905945,}),
    (1, RpCCDCCS, {"e_ref": -100.08885944, "e_tot": -100.08885189,}),
]


if pybest_basis.PYBEST_CHOLESKY_ENABLED:
    linalg_set = [DenseLinalgFactory, CholeskyLinalgFactory]
else:
    linalg_set = [DenseLinalgFactory]


@pytest.mark.parametrize("ncore,cls,dict_", core_set_hf)
@pytest.mark.parametrize("linalg_set", linalg_set)
def test_ccs(ncore, cls, dict_, linalg_set):
    lf, olp, orb_a, one, er, external, occ_model = prepare_mol(
        "cc-pvdz", linalg_set
    )

    iodata = IOData(
        **{
            "orb_a": orb_a,
            "olp": olp,
            "e_ref": dict_["e_ref"],
            "e_core": external,
        }
    )
    ccs = cls(lf, occ_model, ncore=ncore)
    ccs_ = ccs(one, er, iodata)
    assert abs(ccs_.e_tot - dict_["e_tot"]) < 1e-6

    ccs_ = ccs(
        one,
        er,
        restart=f"pybest-results/checkpoint_{cls.__name__}.h5",
        solver="pbqn",
    )
    assert abs(ccs_.e_tot - dict_["e_tot"]) < 1e-6


@pytest.mark.parametrize("ncore,cls,dict_", core_set_pccd)
@pytest.mark.parametrize("linalg_set", linalg_set)
def test_pccd_ccs(ncore, cls, dict_, linalg_set):
    lf, olp, orb_a, one, er, external, occ_model = prepare_mol(
        "cc-pvdz", linalg_set
    )

    pccd = RpCCD(lf, occ_model, ncore=ncore)
    pccd_ = pccd(one, er, orb_a, olp, e_core=external)
    assert abs(pccd_.e_tot - dict_["e_ref"]) < 1e-6

    ccs = cls(lf, occ_model, ncore=ncore)
    ccs_ = ccs(one, er, pccd_)

    assert abs(ccs_.e_tot - dict_["e_tot"]) < 1e-6

    ccs_ = ccs(
        one,
        er,
        restart=f"pybest-results/checkpoint_{cls.__name__}.h5",
        solver="pbqn",
    )
    assert abs(ccs_.e_tot - dict_["e_tot"]) < 1e-6
