# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file

import numpy as np
import pytest

import pybest.gbasis.gobasis as pybest_basis
from pybest.gbasis import get_gobasis
from pybest.gbasis.gobasis import (
    compute_kinetic,
    compute_nuclear,
    compute_eri,
    compute_cholesky_eri,
    compute_nuclear_repulsion,
    compute_overlap,
)
from pybest.linalg.dense import DenseLinalgFactory
from pybest.linalg.cholesky import CholeskyLinalgFactory
from pybest.scf import AufbauOccModel
from pybest.cc.rlccsd import RHFLCCD, RHFLCCSD, RpCCDLCCD, RpCCDLCCSD
from pybest.geminals.rpccd import ROOpCCD
from pybest.wrappers.hf import RHF
from pybest.io.iodata import IOData
from pybest.context import context
from pybest.test.common import numpy_seed, fda_1order
from pybest.orbital_utils import transform_integrals


scalingfactor = 1e-6


@pytest.mark.slow
def test_rhflccd_lambda():
    fn_xyz = context.get_fn("test/c2-12.xyz")
    mol = IOData.from_file(fn_xyz)
    obasis = get_gobasis("cc-pvdz", fn_xyz, print_basis=False)
    lf = DenseLinalgFactory(obasis.nbasis)
    occ_model = AufbauOccModel(6)
    orb = lf.create_orbital(obasis.nbasis)
    olp = compute_overlap(obasis)
    kin = compute_kinetic(obasis)
    na = compute_nuclear(obasis)
    external = {"nn": compute_nuclear_repulsion(obasis)}

    er = compute_eri(obasis)

    one = lf.create_two_index(obasis.nbasis, label="one")
    one.iadd(kin)
    one.iadd(na)

    rhf = RHF(lf, occ_model)
    rhf_ = rhf(kin, na, er, external, orb, olp)

    lccd_solver = RHFLCCD(lf, occ_model)
    with numpy_seed():
        lccd = lccd_solver(
            one, er, rhf_, tco="td", l=True, memoryopt=False, tthreshold=1e-6
        )
        lccd_solver.l_amplitudes[0].assign(np.random.rand(6, 22, 6, 22))
        # Symmetrize
        lccd_solver.l_amplitudes[0].iadd_transpose((2, 3, 0, 1))
        lccd_solver.l_amplitudes[0].iscale(0.5)
        xt2 = lccd_solver.amplitudes[0].get_triu().ravel(order="C")

        def fun(x):
            l2 = lccd_solver.l_amplitudes[0].get_triu().ravel(order="C")
            coeff = l2

            lagrangian = lccd_solver.calculate_energy(x)["e_tot"]
            lagrangian += np.dot(
                coeff, lccd_solver.vfunction(x.ravel(order="C"), [0])
            )
            return lagrangian

        def fun_deriv(x):
            l2 = lccd_solver.l_amplitudes[0].get_triu().ravel(order="C")
            coeff = l2
            t2 = lccd_solver.lf.create_four_index(6, 22, 6, 22)
            t2.assign_triu(x)

            gradient = lccd_solver.vector_lagrange(coeff, [0])
            return gradient.ravel(order="C")

        x = xt2
        dx2 = []
        for i in range(5):
            tmp = np.random.rand(6, 22, 6, 22) * scalingfactor
            tmp[:] = tmp + tmp.transpose((2, 3, 0, 1))
            indtriu = np.triu_indices(6 * 22)
            tmp = tmp.reshape(6 * 22, 6 * 22)[indtriu]
            dx2.append(tmp.ravel())
        dx2 = np.array(dx2)
        dxs = dx2
        # transform AOs to MOs:
        ints = transform_integrals(one, er, orb)
        # recalculate auxiliary matrices because they are deleted after call function exists
        lccd_solver.update_aux_matrix(ints.one[0], ints.two[0], None)
        fda_1order(fun, fun_deriv, x, dxs)


@pytest.mark.slow
def test_lccsd_rhf_lambda():
    fn_xyz = context.get_fn("test/c2-12.xyz")
    mol = IOData.from_file(fn_xyz)
    obasis = get_gobasis("cc-pvdz", fn_xyz, print_basis=False)
    lf = DenseLinalgFactory(obasis.nbasis)
    occ_model = AufbauOccModel(6)
    orb = lf.create_orbital(obasis.nbasis)
    olp = compute_overlap(obasis)
    kin = compute_kinetic(obasis)
    na = compute_nuclear(obasis)
    external = {"nn": compute_nuclear_repulsion(obasis)}

    er = compute_eri(obasis)

    one = lf.create_two_index(obasis.nbasis, label="one")
    one.iadd(kin)
    one.iadd(na)

    rhf = RHF(lf, occ_model)
    rhf_ = rhf(kin, na, er, external, orb, olp)

    lccsd_solver = RHFLCCSD(lf, occ_model)
    with numpy_seed():
        lccsd = lccsd_solver(
            one, er, rhf_, tco="td", l=True, memoryopt=False, tthreshold=1e-6
        )
        lccsd_solver.l_amplitudes[0].assign(np.random.rand(6, 22))
        lccsd_solver.l_amplitudes[1].assign(np.random.rand(6, 22, 6, 22))
        # Symmetrize
        lccsd_solver.l_amplitudes[1].iadd_transpose((2, 3, 0, 1))
        lccsd_solver.l_amplitudes[1].iscale(0.5)
        xt1 = lccsd_solver.amplitudes[0]._array.ravel(order="C")
        xt2 = lccsd_solver.amplitudes[1].get_triu().ravel(order="C")

        def fun(x):
            l1 = lccsd_solver.l_amplitudes[0]._array.ravel(order="C")
            l2 = lccsd_solver.l_amplitudes[1].get_triu().ravel(order="C")
            coeff = np.hstack((l1, l2))

            lagrangian = lccsd_solver.calculate_energy(x)["e_tot"]
            lagrangian += np.dot(
                coeff, lccsd_solver.vfunction(x.ravel(order="C"), [0])
            )
            return lagrangian

        def fun_deriv(x):
            l1 = lccsd_solver.l_amplitudes[0]._array.ravel(order="C")
            l2 = lccsd_solver.l_amplitudes[1].get_triu().ravel(order="C")
            coeff = np.hstack((l1, l2)).ravel(order="C")
            t1 = lccsd_solver.lf.create_two_index(6, 22)
            t2 = lccsd_solver.lf.create_four_index(6, 22, 6, 22)
            t1.assign(x[: 6 * 22])
            t2.assign_triu(x[6 * 22 :])

            gradient = lccsd_solver.vector_lagrange(coeff, [0])
            return gradient.ravel(order="C")

        x = np.hstack((xt1, xt2))
        dx1 = np.random.rand(5, (6 * 22)) * scalingfactor
        dx2 = []
        for i in range(5):
            tmp = np.random.rand(6, 22, 6, 22) * scalingfactor
            tmp[:] = tmp + tmp.transpose((2, 3, 0, 1))
            indtriu = np.triu_indices(6 * 22)
            tmp = tmp.reshape(6 * 22, 6 * 22)[indtriu]
            dx2.append(tmp.ravel())
        dx2 = np.array(dx2)
        dxs = np.hstack((dx1, dx2))
        # transform AOs to MOs:
        ints = transform_integrals(one, er, orb)
        # recalculate auxiliary matrices because they are deleted after call function exists
        lccsd_solver.update_aux_matrix(ints.one[0], ints.two[0], None)
        fda_1order(fun, fun_deriv, x, dxs)


@pytest.mark.slow
def test_lccd_lambda():
    fn_xyz = context.get_fn("test/c2_30.xyz")
    mol = IOData.from_file(fn_xyz)
    obasis = get_gobasis("cc-pvdz", fn_xyz, print_basis=False)
    lf = DenseLinalgFactory(obasis.nbasis)
    occ_model = AufbauOccModel(6)
    orb = lf.create_orbital(obasis.nbasis)
    olp = compute_overlap(obasis)
    kin = compute_kinetic(obasis)
    na = compute_nuclear(obasis)
    external = {"nn": compute_nuclear_repulsion(obasis)}

    er = compute_eri(obasis)

    one = lf.create_two_index(obasis.nbasis, label="one")
    one.iadd(kin)
    one.iadd(na)

    fn_orb = context.get_fn("test/ap1rog_c2_30.txt")
    orb_ = np.fromfile(fn_orb, sep=",").reshape(obasis.nbasis, obasis.nbasis)
    orb_a = lf.create_orbital(obasis.nbasis)
    orb_a._coeffs = orb_
    # Do AP1roG optimization:
    geminal_solver = ROOpCCD(lf, occ_model)
    lccd_solver = RpCCDLCCD(lf, occ_model)
    with numpy_seed():
        ap1rog = geminal_solver(
            one,
            er,
            orb_a,
            olp,
            e_core=external["nn"],
            checkpoint=-1,
            maxiter={"orbiter": 0},
        )
        lccd = lccd_solver(
            one, er, ap1rog, tco="td", l=True, memoryopt=False, tthreshold=1e-6
        )
        lccd_solver.l_amplitudes[0].assign(np.random.rand(6, 22, 6, 22))
        # Symmetrize
        lccd_solver.l_amplitudes[0].iadd_transpose((2, 3, 0, 1))
        lccd_solver.l_amplitudes[0].iscale(0.5)
        xt2 = lccd_solver.amplitudes[0].get_triu().ravel(order="C")
        ind1, ind2 = np.indices((6, 22))
        indices = [ind1, ind2, ind1, ind2]
        # Get rid of pairs
        lccd_solver.l_amplitudes[0].assign(0.0, indices)

        def fun(x):
            l2 = lccd_solver.l_amplitudes[0].get_triu().ravel(order="C")
            coeff = l2

            lagrangian = lccd_solver.calculate_energy(x)["e_tot"]
            lagrangian += np.dot(
                coeff,
                lccd_solver.vfunction(x.ravel(order="C"), [0, ap1rog.t_p]),
            )
            return lagrangian

        def fun_deriv(x):
            l2 = lccd_solver.l_amplitudes[0].get_triu().ravel(order="C")
            coeff = l2
            t2 = lccd_solver.lf.create_four_index(6, 22, 6, 22)
            t2.assign_triu(x)

            gradient = lccd_solver.vector_lagrange(coeff, [0, ap1rog.t_p])
            return gradient.ravel(order="C")

        x = xt2
        dx2 = []
        for i in range(5):
            tmp = np.random.rand(6, 22, 6, 22) * scalingfactor
            tmp[:] = tmp + tmp.transpose((2, 3, 0, 1))
            tmp[indices] = 0.0
            indtriu = np.triu_indices(6 * 22)
            tmp = tmp.reshape(6 * 22, 6 * 22)[indtriu]
            dx2.append(tmp.ravel())
        dx2 = np.array(dx2)
        dxs = dx2
        # transform AOs to MOs:
        ints = transform_integrals(one, er, orb_a)
        # recalculate auxiliary matrices because they are deleted after call function exists
        lccd_solver.update_aux_matrix(ints.one[0], ints.two[0], ap1rog.t_p)
        fda_1order(fun, fun_deriv, x, dxs)


@pytest.mark.slow
def test_lccsd_lambda():
    fn_xyz = context.get_fn("test/c2_30.xyz")
    mol = IOData.from_file(fn_xyz)
    obasis = get_gobasis("cc-pvdz", fn_xyz, print_basis=False)
    lf = DenseLinalgFactory(obasis.nbasis)
    occ_model = AufbauOccModel(6)
    orb = lf.create_orbital(obasis.nbasis)
    olp = compute_overlap(obasis)
    kin = compute_kinetic(obasis)
    na = compute_nuclear(obasis)
    external = {"nn": compute_nuclear_repulsion(obasis)}

    er = compute_eri(obasis)

    one = lf.create_two_index(obasis.nbasis, label="one")
    one.iadd(kin)
    one.iadd(na)

    fn_orb = context.get_fn("test/ap1rog_c2_30.txt")
    orb_ = np.fromfile(fn_orb, sep=",").reshape(obasis.nbasis, obasis.nbasis)
    orb_a = lf.create_orbital(obasis.nbasis)
    orb_a._coeffs = orb_
    # Do AP1roG optimization:
    geminal_solver = ROOpCCD(lf, occ_model)
    lccsd_solver = RpCCDLCCSD(lf, occ_model)
    with numpy_seed():
        ap1rog = geminal_solver(
            one,
            er,
            orb_a,
            olp,
            e_core=external["nn"],
            checkpoint=-1,
            maxiter={"orbiter": 0},
        )
        lccsd = lccsd_solver(
            one, er, ap1rog, tco="td", l=True, memoryopt=False, tthreshold=1e-6
        )
        lccsd_solver.l_amplitudes[0].assign(np.random.rand(6, 22))
        lccsd_solver.l_amplitudes[1].assign(np.random.rand(6, 22, 6, 22))
        # Symmetrize
        lccsd_solver.l_amplitudes[1].iadd_transpose((2, 3, 0, 1))
        lccsd_solver.l_amplitudes[1].iscale(0.5)
        xt1 = lccsd_solver.amplitudes[0]._array.ravel(order="C")
        xt2 = lccsd_solver.amplitudes[1].get_triu().ravel(order="C")
        ind1, ind2 = np.indices((6, 22))
        indices = [ind1, ind2, ind1, ind2]
        # Get rid of pairs
        lccsd_solver.l_amplitudes[1].assign(0.0, indices)

        def fun(x):
            l1 = lccsd_solver.l_amplitudes[0]._array.ravel(order="C")
            l2 = lccsd_solver.l_amplitudes[1].get_triu().ravel(order="C")
            coeff = np.hstack((l1, l2))

            lagrangian = lccsd_solver.calculate_energy(x)["e_tot"]
            lagrangian += np.dot(
                coeff,
                lccsd_solver.vfunction(x.ravel(order="C"), [0, ap1rog.t_p]),
            )
            return lagrangian

        def fun_deriv(x):
            l1 = lccsd_solver.l_amplitudes[0]._array.ravel(order="C")
            l2 = lccsd_solver.l_amplitudes[1].get_triu().ravel(order="C")
            coeff = np.hstack((l1, l2)).ravel(order="C")
            t1 = lccsd_solver.lf.create_two_index(6, 22)
            t2 = lccsd_solver.lf.create_four_index(6, 22, 6, 22)
            t1.assign(x[: 6 * 22])
            t2.assign_triu(x[6 * 22 :])

            gradient = lccsd_solver.vector_lagrange(coeff, [0, ap1rog.t_p])
            return gradient.ravel(order="C")

        x = np.hstack((xt1, xt2))
        dx1 = np.random.rand(5, (6 * 22)) * scalingfactor
        dx2 = []
        for i in range(5):
            tmp = np.random.rand(6, 22, 6, 22) * scalingfactor
            tmp[:] = tmp + tmp.transpose((2, 3, 0, 1))
            tmp[indices] = 0.0
            indtriu = np.triu_indices(6 * 22)
            tmp = tmp.reshape(6 * 22, 6 * 22)[indtriu]
            dx2.append(tmp.ravel())
        dx2 = np.array(dx2)
        dxs = np.hstack((dx1, dx2))
        # transform AOs to MOs:
        ints = transform_integrals(one, er, orb_a)
        # recalculate auxiliary matrices because they are deleted after call function exists
        lccsd_solver.update_aux_matrix(ints.one[0], ints.two[0], ap1rog.t_p)
        fda_1order(fun, fun_deriv, x, dxs)


@pytest.mark.skipif(
    not pybest_basis.PYBEST_CHOLESKY_ENABLED,
    reason="Cholesky-decomposition ERI are not available. Build libchol and re-run build --enable-cholesky=1",
)
@pytest.mark.slow
def test_lccsd_chol_lambda():
    fn_xyz = context.get_fn("test/c2_30.xyz")
    mol = IOData.from_file(fn_xyz)
    obasis = get_gobasis("cc-pvdz", fn_xyz, print_basis=False)
    lf = CholeskyLinalgFactory(obasis.nbasis)
    occ_model = AufbauOccModel(6)
    orb = lf.create_orbital(obasis.nbasis)
    olp = compute_overlap(obasis)
    kin = compute_kinetic(obasis)
    na = compute_nuclear(obasis)
    external = {"nn": compute_nuclear_repulsion(obasis)}

    er = compute_cholesky_eri(obasis)

    one = lf.create_two_index(obasis.nbasis, label="one")
    one.iadd(kin)
    one.iadd(na)

    fn_orb = context.get_fn("test/ap1rog_c2_30.txt")
    orb_ = np.fromfile(fn_orb, sep=",").reshape(obasis.nbasis, obasis.nbasis)
    orb_a = lf.create_orbital(obasis.nbasis)
    orb_a._coeffs = orb_
    # Do AP1roG optimization:
    geminal_solver = ROOpCCD(lf, occ_model)
    lccsd_solver = RpCCDLCCSD(lf, occ_model)
    # transform AOs to MOs:
    ints = transform_integrals(one, er, orb_a)
    with numpy_seed():
        ap1rog = geminal_solver(
            one,
            er,
            orb_a,
            olp,
            e_core=external["nn"],
            checkpoint=-1,
            maxiter={"orbiter": 0},
        )
        lccsd = lccsd_solver(
            one, er, ap1rog, tco="td", l=True, memoryopt=False, tthreshold=1e-6
        )
        lccsd_solver.l_amplitudes[0].assign(np.random.rand(6, 22))
        lccsd_solver.l_amplitudes[1].assign(np.random.rand(6, 22, 6, 22))
        # Symmetrize
        lccsd_solver.l_amplitudes[1].iadd_transpose((2, 3, 0, 1))
        lccsd_solver.l_amplitudes[1].iscale(0.5)
        xt1 = lccsd_solver.amplitudes[0]._array.ravel(order="C")
        xt2 = lccsd_solver.amplitudes[1].get_triu().ravel(order="C")
        ind1, ind2 = np.indices((6, 22))
        indices = [ind1, ind2, ind1, ind2]
        # Get rid of pairs
        lccsd_solver.l_amplitudes[1].assign(0.0, indices)

        def fun(x):
            l1 = lccsd_solver.l_amplitudes[0]._array.ravel(order="C")
            l2 = lccsd_solver.l_amplitudes[1].get_triu().ravel(order="C")
            coeff = np.hstack((l1, l2))

            lagrangian = lccsd_solver.calculate_energy(x)["e_tot"]
            lagrangian += np.dot(
                coeff,
                lccsd_solver.vfunction(
                    x.ravel(order="C"), [0, ap1rog.t_p, ints.two[0]]
                ),
            )
            return lagrangian

        def fun_deriv(x):
            l1 = lccsd_solver.l_amplitudes[0]._array.ravel(order="C")
            l2 = lccsd_solver.l_amplitudes[1].get_triu().ravel(order="C")
            coeff = np.hstack((l1, l2)).ravel(order="C")
            t1 = lccsd_solver.lf.create_two_index(6, 22)
            t2 = lccsd_solver.denself.create_four_index(6, 22, 6, 22)
            t1.assign(x[: 6 * 22])
            t2.assign_triu(x[6 * 22 :])

            gradient = lccsd_solver.vector_lagrange(
                coeff, [0, ap1rog.t_p, ints.two[0]]
            )
            return gradient.ravel(order="C")

        x = np.hstack((xt1, xt2))
        dx1 = np.random.rand(5, (6 * 22)) * scalingfactor
        dx2 = []
        for i in range(5):
            tmp = np.random.rand(6, 22, 6, 22) * scalingfactor
            tmp[:] = tmp + tmp.transpose((2, 3, 0, 1))
            tmp[indices] = 0.0
            indtriu = np.triu_indices(6 * 22)
            tmp = tmp.reshape(6 * 22, 6 * 22)[indtriu]
            dx2.append(tmp.ravel())
        dx2 = np.array(dx2)
        dxs = np.hstack((dx1, dx2))
        # recalculate auxiliary matrices because they are deleted after call function exists
        lccsd_solver.update_aux_matrix(ints.one[0], ints.two[0], ap1rog.t_p)
        fda_1order(fun, fun_deriv, x, dxs)
