# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file


import numpy as np
import pytest

import pybest.gbasis.gobasis as pybest_basis
from pybest.gbasis import get_gobasis
from pybest.gbasis.gobasis import (
    compute_kinetic,
    compute_nuclear,
    compute_eri,
    compute_cholesky_eri,
    compute_nuclear_repulsion,
    compute_overlap,
)
from pybest.linalg.dense import DenseLinalgFactory
from pybest.linalg.cholesky import CholeskyLinalgFactory
from pybest.scf import AufbauOccModel
from pybest.geminals.rpccd import RpCCD
from pybest.cc.rlccsd import RHFLCCD, RHFLCCSD, RpCCDLCCD, RpCCDLCCSD
from pybest.context import context
from pybest.io.iodata import IOData


def prepare_mol(test_dict, linalg):
    fn_xyz = context.get_fn(test_dict["xyz"])
    obasis = get_gobasis(test_dict["basis"], fn_xyz, print_basis=False)
    lf = linalg(obasis.nbasis)
    olp = compute_overlap(obasis)
    kin = compute_kinetic(obasis)
    na = compute_nuclear(obasis)
    one = kin.copy()
    one.iadd(na)
    if isinstance(lf, CholeskyLinalgFactory):
        er = compute_cholesky_eri(obasis, threshold=1e-8)
    elif isinstance(lf, DenseLinalgFactory):
        er = compute_eri(obasis)
    external = compute_nuclear_repulsion(obasis)

    fn_orb = context.get_fn(test_dict["orb"])
    orb_ = np.fromfile(fn_orb, sep=",").reshape(obasis.nbasis, obasis.nbasis)
    orb_a = lf.create_orbital()
    orb_a._coeffs = orb_

    occ_model = AufbauOccModel(test_dict["nocc"])

    return lf, olp, orb_a, one, er, external, occ_model


rhf_set = [
    (
        RHFLCCD,
        "nh3/cc-pvdz",
        "krylov",
        {
            "xyz": "test/nh3.xyz",
            "orb": "test/nh3_hf.txt",
            "nocc": 5,
            "basis": "cc-pvdz",
            "solver": "krylov",
            "e_ref": -56.195341335133,
            "e_tot": -56.40363609015413,
            "e_corr": -0.20829475502103717,
            "e_corr_d": -0.20829475502103717,
        },
    ),
    (
        RHFLCCSD,
        "nh3/cc-pvdz",
        "krylov",
        {
            "xyz": "test/nh3.xyz",
            "orb": "test/nh3_hf.txt",
            "nocc": 5,
            "basis": "cc-pvdz",
            "solver": "krylov",
            "e_ref": -56.195341335133,
            "e_tot": -56.404534989197664,
            "e_corr": -0.2091936543105653,
            "e_corr_d": -0.2091936543105653,
            "e_corr_s": 0.0000000000000000,
        },
    ),
]

pccd_set = [
    (
        RpCCDLCCD,
        "nh3/cc-pvdz/can",
        {
            "xyz": "test/nh3.xyz",
            "orb": "test/nh3_hf.txt",
            "nocc": 5,
            "basis": "cc-pvdz",
            "e_tot_pccd": -56.235650029799,
            "e_tot": -56.40112996066194,
            "e_corr": -0.16547993086294,
            "e_corr_d": -0.16547993086294,
        },
    ),
    (
        RpCCDLCCSD,
        "nh3/cc-pvdz/can",
        {
            "xyz": "test/nh3.xyz",
            "orb": "test/nh3_hf.txt",
            "nocc": 5,
            "basis": "cc-pvdz",
            "e_tot_pccd": -56.235650029799,
            "e_tot": -56.40187108073425,
            "e_corr": -0.16622105102133822,
            "e_corr_s": 0.00000000000000000,
            "e_corr_d": -0.16622105102133822,
        },
    ),
    (
        RpCCDLCCD,
        "nh3/cc-pvdz/opt",
        {
            "xyz": "test/nh3.xyz",
            "orb": "test/nh3_ap1rog.txt",
            "nocc": 5,
            "basis": "cc-pvdz",
            "e_tot_pccd": -56.288865946757,
            "e_tot": -56.404001426,
            "e_corr": -0.115135479243,
            "e_corr_d": -0.115135479243,
        },
    ),
    (
        RpCCDLCCSD,
        "nh3/cc-pvdz/opt",
        {
            "xyz": "test/nh3.xyz",
            "orb": "test/nh3_ap1rog.txt",
            "nocc": 5,
            "basis": "cc-pvdz",
            "e_tot_pccd": -56.288865946757,
            "e_tot": -56.40489919845452,
            "e_corr": -0.11603329,
            "e_corr_s": -0.00024555,
            "e_corr_d": -0.11578774,
        },
    ),
]

pbqn_set = [
    (
        "pbqn/6/f/2",
        {"solver": "pbqn", "diismax": 6, "diisreset": False, "jacobian": 2,},
    ),
    (
        "pbqn/6/f/1",
        {"solver": "pbqn", "diismax": 6, "diisreset": False, "jacobian": 1,},
    ),
    (
        "pbqn/6/t/1",
        {"solver": "pbqn", "diismax": 6, "diisreset": True, "jacobian": 1,},
    ),
    (
        "pbqn/6/t/2",
        {"solver": "pbqn", "diismax": 6, "diisreset": True, "jacobian": 2,},
    ),
]

# test only one module, frozen core is assigned using the same code in all modules
frozen_core_set = [
    (
        RpCCDLCCSD,
        "nh3/cc-pvdz",
        0,
        {
            "xyz": "test/nh3.xyz",
            "orb": "test/nh3_ap1rog.txt",
            "nocc": 5,
            "basis": "cc-pvdz",
            "e_tot_pccd": -56.288865946757,
            "e_tot": -56.40489919845452,
            "e_corr": -0.11603329,
            "e_corr_s": -0.00024555,
            "e_corr_d": -0.11578774,
        },
    ),
    (
        RpCCDLCCSD,
        "nh3/cc-pvdz",
        1,
        {
            "xyz": "test/nh3.xyz",
            "orb": "test/nh3_ap1rog.txt",
            "nocc": 5,
            "basis": "cc-pvdz",
            "e_tot_pccd": -56.288582636705,
            "e_tot": -56.40250582,
            "e_corr": -0.11392318,
            "e_corr_s": -0.00024744,
            "e_corr_d": -0.11367575,
        },
    ),
]

if pybest_basis.PYBEST_CHOLESKY_ENABLED:
    linalg_set = [DenseLinalgFactory, CholeskyLinalgFactory]
else:
    linalg_set = [DenseLinalgFactory]


@pytest.mark.parametrize("cls,name,solver,dict_", rhf_set)
@pytest.mark.parametrize("linalg_set", linalg_set)
def test_rhflcc(cls, name, solver, dict_, linalg_set):
    lf, olp, orb_a, one, er, external, occ_model = prepare_mol(
        dict_, linalg_set
    )

    iodata = IOData(
        **{
            "orb_a": orb_a,
            "olp": olp,
            "e_core": external,
            "e_ref": dict_["e_ref"],
        }
    )
    lcc = cls(lf, occ_model)
    lcc_ = lcc(one, er, iodata, tthreshold=1e-8,)
    assert abs(lcc_.e_tot - dict_["e_tot"]) < 1e-6
    assert abs(lcc_.e_corr - dict_["e_corr"]) < 1e-6
    assert abs(lcc_.e_corr_d - dict_["e_corr_d"]) < 1e-6
    if "e_corr_s" in dict_:
        assert abs(lcc_.e_corr_s - dict_["e_corr_s"]) < 1e-6


@pytest.mark.parametrize("cls,name,solver,dict_", rhf_set)
def test_rhflcc_restart(cls, name, solver, dict_):
    lf, olp, orb_a, one, er, external, occ_model = prepare_mol(
        dict_, DenseLinalgFactory
    )

    iodata = IOData(
        **{
            "orb_a": orb_a,
            "olp": olp,
            "e_core": external,
            "e_ref": dict_["e_ref"],
        }
    )
    #
    # Converge not too tight
    #
    lcc = cls(lf, occ_model)
    lcc_ = lcc(one, er, iodata, tthreshold=1e-3,)
    #
    # Test restart
    #
    lcc = cls(lf, occ_model)
    lcc_ = lcc(
        one,
        er,
        restart=f"pybest-results/checkpoint_{cls.__name__}.h5",
        tthreshold=1e-8,
    )
    assert abs(lcc_.e_tot - dict_["e_tot"]) < 1e-6
    assert abs(lcc_.e_corr - dict_["e_corr"]) < 1e-6
    assert abs(lcc_.e_corr_d - dict_["e_corr_d"]) < 1e-6
    if "e_corr_s" in dict_:
        assert abs(lcc_.e_corr_s - dict_["e_corr_s"]) < 1e-6


@pytest.mark.parametrize("solver", ["krylov"])
@pytest.mark.parametrize("cls,name,dict_", pccd_set)
@pytest.mark.parametrize("linalg_set", linalg_set)
def test_pccdlcc(cls, name, dict_, solver, linalg_set):
    lf, olp, orb_a, one, er, external, occ_model = prepare_mol(
        dict_, linalg_set
    )

    # Do pCCD optimization:
    pccd = RpCCD(lf, occ_model)
    pccd_ = pccd(one, er, orb_a, olp, e_core=external)
    assert abs(pccd_.e_tot - dict_["e_tot_pccd"]) < 1e-6

    lcc = cls(lf, occ_model)
    lcc_ = lcc(one, er, pccd_, tthreshold=1e-8)
    assert abs(lcc_.e_tot - dict_["e_tot"]) < 1e-6
    assert abs(lcc_.e_corr - dict_["e_corr"]) < 1e-6
    assert abs(lcc_.e_corr_d - dict_["e_corr_d"]) < 1e-6
    if "e_corr_s" in dict_:
        assert abs(lcc_.e_corr_s - dict_["e_corr_s"]) < 1e-6


@pytest.mark.parametrize("solver", ["krylov"])
@pytest.mark.parametrize("cls,name,dict_", pccd_set)
def test_pccdlcc_restart(cls, name, dict_, solver):
    lf, olp, orb_a, one, er, external, occ_model = prepare_mol(
        dict_, DenseLinalgFactory
    )

    # Do pCCD optimization:
    pccd = RpCCD(lf, occ_model)
    pccd_ = pccd(one, er, orb_a, olp, e_core=external)
    assert abs(pccd_.e_tot - dict_["e_tot_pccd"]) < 1e-6
    #
    # Converge not too tight
    #
    lcc = cls(lf, occ_model)
    lcc_ = lcc(one, er, pccd_, tthreshold=1e-3)
    #
    # Test restart
    #
    lcc = cls(lf, occ_model)
    lcc_ = lcc(
        one,
        er,
        pccd_,
        restart=f"pybest-results/checkpoint_{cls.__name__}.h5",
        tthreshold=1e-8,
    )
    assert abs(lcc_.e_tot - dict_["e_tot"]) < 1e-6
    assert abs(lcc_.e_corr - dict_["e_corr"]) < 1e-6
    assert abs(lcc_.e_corr_d - dict_["e_corr_d"]) < 1e-6
    if "e_corr_s" in dict_:
        assert abs(lcc_.e_corr_s - dict_["e_corr_s"]) < 1e-6


@pytest.mark.parametrize("solver, solver_dict", pbqn_set)
@pytest.mark.parametrize("cls,name,dict_", pccd_set)
def test_pccdlcc_solver(cls, name, dict_, solver, solver_dict):
    lf, olp, orb_a, one, er, external, occ_model = prepare_mol(
        dict_, DenseLinalgFactory
    )

    # Do pCCD optimization:
    pccd = RpCCD(lf, occ_model)
    pccd_ = pccd(one, er, orb_a, olp, e_core=external)
    assert abs(pccd_.e_tot - dict_["e_tot_pccd"]) < 1e-6

    lcc = cls(lf, occ_model)
    lcc_ = lcc(one, er, pccd_, tthreshold=1e-8, **solver_dict)
    assert abs(lcc_.e_tot - dict_["e_tot"]) < 1e-6
    assert abs(lcc_.e_corr - dict_["e_corr"]) < 1e-6
    assert abs(lcc_.e_corr_d - dict_["e_corr_d"]) < 1e-6
    if "e_corr_s" in dict_:
        assert abs(lcc_.e_corr_s - dict_["e_corr_s"]) < 1e-6


@pytest.mark.parametrize("cls,name,ncore,dict_", frozen_core_set)
def test_pccdlcc_frozen_core(cls, name, ncore, dict_):
    lf, olp, orb_a, one, er, external, occ_model = prepare_mol(
        dict_, DenseLinalgFactory
    )

    # Do pCCD optimization:
    pccd = RpCCD(lf, occ_model, ncore=ncore)
    pccd_ = pccd(one, er, orb_a, olp, e_core=external)
    assert abs(pccd_.e_tot - dict_["e_tot_pccd"]) < 1e-6

    lcc = cls(lf, occ_model, ncore=ncore)
    lcc_ = lcc(one, er, pccd_, tthreshold=1e-8)
    assert abs(lcc_.e_tot - dict_["e_tot"]) < 1e-6
    assert abs(lcc_.e_corr - dict_["e_corr"]) < 1e-6
    assert abs(lcc_.e_corr_d - dict_["e_corr_d"]) < 1e-6
    if "e_corr_s" in dict_:
        assert abs(lcc_.e_corr_s - dict_["e_corr_s"]) < 1e-6

    # Overwrite core energy and thus total energy
    # Can only be done in pCCD, affects only total energy
    pccd = RpCCD(lf, occ_model, ncore=ncore)
    pccd_ = pccd(one, er, orb_a, olp, e_core=(external + 10.0))
    assert abs(pccd_.e_tot - dict_["e_tot_pccd"] - 10.0) < 1e-6

    lcc = cls(lf, occ_model, ncore=ncore)
    lcc_ = lcc(one, er, pccd_, tthreshold=1e-8)
    assert abs(lcc_.e_tot - dict_["e_tot"] - 10.0) < 1e-6
    assert abs(lcc_.e_corr - dict_["e_corr"]) < 1e-6
    assert abs(lcc_.e_corr_d - dict_["e_corr_d"]) < 1e-6
    if "e_corr_s" in dict_:
        assert abs(lcc_.e_corr_s - dict_["e_corr_s"]) < 1e-6
