# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file

import numpy as np
import pytest

from pybest import filemanager
from pybest.cc.rccsd import RCCD, RCCSD
from pybest.context import context
from pybest.gbasis import get_gobasis
from pybest.gbasis.gobasis import (
    PYBEST_CHOLESKY_ENABLED,
    compute_kinetic,
    compute_nuclear,
    compute_eri,
    compute_cholesky_eri,
    compute_nuclear_repulsion,
    compute_overlap,
)
from pybest.io.iodata import IOData
from pybest.linalg.dense import DenseLinalgFactory
from pybest.linalg.dense import DenseTwoIndex, DenseFourIndex
from pybest.linalg.cholesky import CholeskyLinalgFactory
from pybest.scf import AufbauOccModel
from pybest.scf.observable import RTwoIndexTerm, RDirectTerm, RExchangeTerm
from pybest.scf.hamiltonian import RScfHam
from pybest.scf.guess import guess_core_hamiltonian
from pybest.scf.utils import compute_1dm_hf
from pybest.scf.scf_ediis2 import EDIIS2SCFSolver


###############################################################################
#  Unit test  #################################################################
###############################################################################


def test_read_input():

    class DenseNIndexMockObject:
        def __init__(self, label):
    	    self.label = label

    # Create input data
    one = DenseNIndexMockObject("one")
    eri = DenseNIndexMockObject("eri")
    orb = DenseNIndexMockObject("orb")
    iodata_ = IOData(eri=eri, orb_a=orb, e_ref="e_ref", x="x")

    # Make RCCD instance and read input
    occ_model = AufbauOccModel(1)
    lf = DenseLinalgFactory(2)
    cc = RCCD(lf, occ_model)
    cc.read_input(iodata_, orb, one=one)
    assert cc.e_ref == "e_ref"


def check_guess(t_1, t_2):
    assert isinstance(t_1, DenseTwoIndex)
    assert isinstance(t_2, DenseFourIndex)
    assert t_1._array.shape == (1, 9)
    assert t_2._array.shape == (1, 9, 1, 9)


@pytest.mark.parametrize("method", [RCCD, RCCSD])
def test_get_random_guess(method):

    coordfile = context.get_fn(TEST_CASE[0]["path"])
    model = Input(
        DenseLinalgFactory,
        coordfile,
        TEST_CASE[0]["nocc"],
        TEST_CASE[0]["basis"],
    )
    cc = method(model.lf, model.occ_model)
    check_guess(*getattr(cc, "get_random_guess")())


@pytest.mark.parametrize("method", [RCCD, RCCSD])
def test_get_constant_guess(method):

    coordfile = context.get_fn(TEST_CASE[0]["path"])
    model = Input(
        DenseLinalgFactory,
        coordfile,
        TEST_CASE[0]["nocc"],
        TEST_CASE[0]["basis"],
    )
    cc = method(model.lf, model.occ_model)
    check_guess(*getattr(cc, "get_constant_guess")(0.5))


def test_get_mp2_guess():

    coordfile = context.get_fn(TEST_CASE[0]["path"])
    model = Input(
        DenseLinalgFactory,
        coordfile,
        TEST_CASE[0]["nocc"],
        TEST_CASE[0]["basis"],
    )
    ao1 = model.kin.copy()
    ao1.iadd(model.ne)
    mo1 = ao1.new()
    mo1.assign_two_index_transform(ao1, model.orb)
    mo2 = model.eri.new()
    mo2.assign_four_index_transform(model.eri, model.orb)

    ccd = RCCD(model.lf, model.occ_model)
    t_1, t_2 = getattr(ccd, "get_mp2_guess")(mo1, mo2, model.orb)
    assert t_1 == None
    assert isinstance(t_2, DenseFourIndex)
    assert t_2._array.shape == (1, 9, 1, 9)

    ccsd = RCCSD(model.lf, model.occ_model)
    check_guess(*getattr(ccsd, "get_mp2_guess")(mo1, mo2, model.orb))


@pytest.mark.parametrize("method, expected", [(RCCD, 45), (RCCSD, 54)])
def test_get_guess(method, expected):

    coordfile = context.get_fn(TEST_CASE[0]["path"])
    model = Input(
        DenseLinalgFactory,
        coordfile,
        TEST_CASE[0]["nocc"],
        TEST_CASE[0]["basis"],
    )
    ao1 = model.kin.copy()
    ao1.iadd(model.ne)
    kwargs = {"orbital": model.orb, "ao1": ao1, "ao2": model.eri}

    cc = method(model.lf, model.occ_model)
    cc.initguess = "mp2"
    tvector = getattr(cc, "get_guess")(**kwargs)
    assert isinstance(tvector, np.ndarray)
    assert len(tvector) == expected


###############################################################################
#  Unit test  #################################################################
###############################################################################


def test_read_input():

    class DenseNIndexMockObject:
        def __init__(self, label):
    	    self.label = label

    # Create input data
    one = DenseNIndexMockObject("one")
    eri = DenseNIndexMockObject("eri")
    orb = DenseNIndexMockObject("orb")
    iodata_ = IOData(eri=eri, orb_a=orb, e_ref="e_ref", x="x")

    # Make RCCD instance and read input
    occ_model = AufbauOccModel(1)
    lf = DenseLinalgFactory(2)
    cc = RCCD(lf, occ_model)
    cc.read_input(iodata_, orb, one=one)
    assert cc.e_ref == "e_ref"


def check_guess(t_1, t_2):
    assert isinstance(t_1, DenseTwoIndex)
    assert isinstance(t_2, DenseFourIndex)
    assert t_1._array.shape == (1, 9)
    assert t_2._array.shape == (1, 9, 1, 9)


@pytest.mark.parametrize("method", [RCCD, RCCSD])
def test_get_random_guess(method):

    coordfile = context.get_fn(TEST_CASE[0]["path"])
    model = Input(
        DenseLinalgFactory,
        coordfile,
        TEST_CASE[0]["nocc"],
        TEST_CASE[0]["basis"],
    )
    cc = method(model.lf, model.occ_model)
    check_guess(*getattr(cc, "get_random_guess")())


@pytest.mark.parametrize("method", [RCCD, RCCSD])
def test_get_constant_guess(method):

    coordfile = context.get_fn(TEST_CASE[0]["path"])
    model = Input(
        DenseLinalgFactory,
        coordfile,
        TEST_CASE[0]["nocc"],
        TEST_CASE[0]["basis"],
    )
    cc = method(model.lf, model.occ_model)
    check_guess(*getattr(cc, "get_constant_guess")(0.5))


def test_get_mp2_guess():

    coordfile = context.get_fn(TEST_CASE[0]["path"])
    model = Input(
        DenseLinalgFactory,
        coordfile,
        TEST_CASE[0]["nocc"],
        TEST_CASE[0]["basis"],
    )
    ao1 = model.kin.copy()
    ao1.iadd(model.ne)
    mo1 = ao1.new()
    mo1.assign_two_index_transform(ao1, model.orb)
    mo2 = model.eri.new()
    mo2.assign_four_index_transform(model.eri, model.orb)

    ccd = RCCD(model.lf, model.occ_model)
    t_1, t_2 = getattr(ccd, "get_mp2_guess")(mo1, mo2, model.orb)
    assert t_1 == None
    assert isinstance(t_2, DenseFourIndex)
    assert t_2._array.shape == (1, 9, 1, 9)

    ccsd = RCCSD(model.lf, model.occ_model)
    check_guess(*getattr(ccsd, "get_mp2_guess")(mo1, mo2, model.orb))


@pytest.mark.parametrize("method, expected", [(RCCD, 45), (RCCSD, 54)])
def test_get_guess(method, expected):

    coordfile = context.get_fn(TEST_CASE[0]["path"])
    model = Input(
        DenseLinalgFactory,
        coordfile,
        TEST_CASE[0]["nocc"],
        TEST_CASE[0]["basis"],
    )
    ao1 = model.kin.copy()
    ao1.iadd(model.ne)
    kwargs = {"orbital": model.orb, "ao1": ao1, "ao2": model.eri}

    cc = method(model.lf, model.occ_model)
    cc.initguess = "mp2"
    tvector = getattr(cc, "get_guess")(**kwargs)
    assert isinstance(tvector, np.ndarray)
    assert len(tvector) == expected


###############################################################################
## E2E tests ##################################################################
###############################################################################


class Input:
    def __init__(self, linalg_factory, xyz_data, npairs, basis):

        self.test_case = xyz_data
        self.occ_model = AufbauOccModel(npairs)
        self._init_input(linalg_factory, xyz_data, basis)
        self._init_hf()

    def _init_input(self, linalg_factory, xyz_file, basis):

        # Construct basis object, orbitals, and overlap
        obasis = get_gobasis(basis, xyz_file, print_basis=False)
        self.lf = linalg_factory(obasis.nbasis)
        self.orb = self.lf.create_orbital(obasis.nbasis)
        self.olp = compute_overlap(obasis)

        # Compute integrals
        self.kin = compute_kinetic(obasis)
        self.ne = compute_nuclear(obasis)
        if isinstance(self.lf, CholeskyLinalgFactory):
            self.eri = compute_cholesky_eri(obasis, threshold=1e-9)
        else:
            self.eri = compute_eri(obasis)
        self.external = {"nn": compute_nuclear_repulsion(obasis)}
        self.io = {"kin": self.kin, "ne": self.ne, "eri": self.eri}

    def _init_hf(self):
        # For HF calculations we need to collect all terms
        terms = [
            RTwoIndexTerm(self.kin, "kin"),
            RTwoIndexTerm(self.ne, "ne"),
            RDirectTerm(self.eri, "hartree"),
            RExchangeTerm(self.eri, "x_hf"),
        ]
        ham = RScfHam(terms, self.external)
        ## Perform initial guess
        guess_core_hamiltonian(self.olp, self.kin, self.ne, self.orb)
        ## Do a Hartree-Fock calculation with DIIS solvers
        self.occ_model.assign(self.orb)
        dm_alpha = compute_1dm_hf(self.orb)
        scf_solver = EDIIS2SCFSolver(threshold=1e-8)
        scf = scf_solver(ham, self.lf, self.olp, self.occ_model, dm_alpha)
        focks = [self.lf.create_two_index() for i in range(ham.ndm)]
        ham.compute_fock(*focks)
        self.orb.from_fock(focks[0], self.olp)
        # HF reference energy
        self.io.update({"e_ref": scf.e_tot, "orb_a": self.orb})

    def rccd(self, guess="const", ncore=0, maxiter=40):
        ccd = RCCD(self.lf, self.occ_model, ncore=ncore)
        result = ccd(IOData(**self.io), maxiter=maxiter, initguess=guess)
        self.method = "RCCD"
        self.energy = result.e_corr
        return result

    def rccsd(self, guess="const", ncore=0, maxiter=40):
        ccd = RCCSD(self.lf, self.occ_model, ncore=ncore)
        result = ccd(IOData(**self.io), maxiter=maxiter, initguess=guess)
        self.method = "RCCSD"
        self.energy = result.e_corr
        return result

    def compare_energy(self, expected, threshold=1e-8):
        status = (
            f"Test of {self.method} vector function\n"
            f"test case         {self.test_case}\n"
            f"linalg factory    {self.lf}\n"
            f"current result    {self.energy:15.9f}\n"
            f"expected result   {expected:15.9f}\n"
            f"difference        {self.energy - expected:15.9f}\n"
        )
        assert abs(self.energy - expected) < threshold, status


TEST_CASE = [
    {
        "path": "test/h2.xyz",
        "nocc": 1,
        "basis": "cc-pvdz",
        "energy_rccd": -0.034582180698988885,
        "energy_rccsd": -0.034709514282981926,
    },
    {
        "path": "test/hf.xyz",
        "nocc": 5,
        "basis": "3-21g",
        "energy_rccd": -0.12911913585251084,
        "energy_rccsd": -0.12987666859843192,
    },
    {
        "path": "test/h2o.xyz",
        "nocc": 5,
        "basis": "3-21g",
        "energy_rccd": -0.1499368041545847,
        "energy_rccsd": -0.15213686280456384,
    },
]


def dotest(method, lf, test_case, **kwargs):

    coordfile = context.get_fn(test_case["path"])
    model = Input(lf, coordfile, test_case["nocc"], test_case["basis"])

    getattr(model, method)(**kwargs)
    model.compare_energy(test_case["energy_" + method])


@pytest.mark.parametrize("method", ["rccd", "rccsd"])
@pytest.mark.parametrize("test_case", TEST_CASE)
def test_rccsd_dense(method, test_case):
    dotest(method, DenseLinalgFactory, test_case)


@pytest.mark.skipif(
    not PYBEST_CHOLESKY_ENABLED,
    reason="Cholesky-decomposition ERI are not available. Build libchol and re-run build --enable-cholesky=1",
)
@pytest.mark.parametrize("method", ["rccd", "rccsd"])
@pytest.mark.parametrize("test_case", TEST_CASE)
def test_rccsd_cholesky(method, test_case):
    dotest(method, CholeskyLinalgFactory, test_case)


@pytest.mark.parametrize("method", ["rccd", "rccsd"])
def test_rccsd_frozencore(method):
    test_frozen_core = {
        "path": "test/hf.xyz",
        "nocc": 5,
        "basis": "3-21g",
        "energy_rccd": -0.0890402717835759,
        "energy_rccsd": -0.08978685181771721,
    }
    dotest(method, DenseLinalgFactory, test_frozen_core, ncore=2)


@pytest.mark.parametrize("method", ["rccd", "rccsd"])
def test_restart(method):

    coordfile = context.get_fn(TEST_CASE[0]["path"])
    model = Input(
        DenseLinalgFactory,
        coordfile,
        TEST_CASE[0]["nocc"],
        TEST_CASE[0]["basis"],
    )

    path = filemanager.temp_path("checkpoint_{method.capitalize()}.h5")
    cc_init = getattr(model, method)(maxiter=1)
    cc_init.to_file(path)
    cc_final = getattr(model, method)(guess=path)
