#!/usr/bin/env python

import numpy as np

from pybest.gbasis import get_gobasis
from pybest.gbasis.gobasis import (
    compute_eri,
    compute_kinetic,
    compute_nuclear,
    compute_nuclear_repulsion,
    compute_overlap,
)
from pybest.io.iodata import IOData
from pybest.linalg.dense import DenseLinalgFactory
from pybest.orbital_utils import transform_integrals
from pybest.scf import AufbauOccModel
from pybest.wrappers.hf import RHF

# Set up Neon dimer, define basis set
# -----------------------------------
coordinates = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 5.0]])
atom = np.array(["Ne", "Ne"])
mol = IOData(coordinates=coordinates, atom=atom)
obasis = get_gobasis("cc-pvdz", mol)
lf = DenseLinalgFactory(obasis.nbasis)
occ_model = AufbauOccModel(2)
orb = lf.create_orbital()

# Construct Hamiltonian
# ---------------------
kin_ao = compute_kinetic(obasis)
ne_ao = compute_nuclear(obasis)
eri_ao = compute_eri(obasis)
e_core = compute_nuclear_repulsion(obasis)
olp = compute_overlap(obasis)

# Get Hartree-Fock MOs
# --------------------
hf = RHF(lf, occ_model)
hf_ = hf(kin_ao, ne_ao, eri_ao, e_core, orb, olp)

# Transform Hamiltinian to MO basis
# ---------------------------------
# transform integrals for restricted orbitals orb
ti_ = transform_integrals(kin_ao, ne_ao, eri_ao, orb)

# transformed one-electron integrals: attribute 'one' (list)
(one,) = ti_.one  # or: one = ti_.one[0]
# transformed two-electron integrals: attribute 'two' (list)
(two,) = ti_.two  # or: two = ti_.two[0]

# Write to a FCIDUMP file
# -----------------------
data = IOData(one=one, two=two, e_core=e_core, nelec=20, ms2=0)
data.to_file("hamiltonian_mo.FCIDUMP")
