#!/usr/bin/env python
# JSON {"lot": "RHF/3-21G",
# JSON  "scf": "PlainSCFSolver",
# JSON  "linalg": "DenseLinalgFactory",
# JSON  "difficulty": 1,
# JSON  "description": "Basic RHF example with dense matrices"}

import numpy as np

from pybest import context
from pybest.gbasis import get_gobasis
from pybest.gbasis.gobasis import (
    compute_eri,
    compute_kinetic,
    compute_nuclear,
    compute_nuclear_repulsion,
    compute_overlap,
)
from pybest.io.iodata import IOData
from pybest.linalg.dense import DenseLinalgFactory
from pybest.scf import AufbauOccModel
from pybest.wrappers.hf import RHF


# Hartree-Fock calculation
# ------------------------

# Load the coordinates from file.
# Use the XYZ file from PyBEST's test data directory.
fn_xyz = context.get_fn("test/water.xyz")

# Create a Gaussian basis set
gobasis = get_gobasis("cc-pvdz", fn_xyz)

# Create a linalg factory
lf = DenseLinalgFactory(gobasis.nbasis)

# Compute integrals
olp = compute_overlap(gobasis)
kin = compute_kinetic(gobasis)
na = compute_nuclear(gobasis)
er = compute_eri(gobasis)
external = compute_nuclear_repulsion(gobasis)

# Create alpha orbitals
orb_a = lf.create_orbital()

# Decide how to occupy the orbitals (5 alpha electrons)
occ_model = AufbauOccModel(5)

# Converge RHF
hf = RHF(lf, occ_model)
# the order of the arguments does not matter
hf_ = hf(kin, na, er, external, olp, orb_a)

# Write SCF results to a molden file
# First, add gobasis as attribute
hf_.gobasis = gobasis
# Ready to dump results
hf_.to_file("water-scf.molden")

# Restart RHF
hf = RHF(lf, occ_model)
# we still need to provide some inital guess orbitals
# results are stored in pybest-results directory
hf_ = hf(kin, na, er, external, olp, orb_a, restart="pybest-results/checkpoint_scf.h5")

# Read in orbitals and modify them
restart = IOData.from_file("pybest-results/checkpoint_scf.h5")

# Swap HOMO-LUMO orbitals (Python indexing, starts with 0)
swaps = np.array([[4, 5]])
# Only swap AO/MO coefficients
restart.orb_a.swap_orbitals(swaps, skip_occs=True, skip_energies=True)

# Do RHF calculation using those orbitals
hf = RHF(lf, occ_model)
hf_ = hf(kin, na, er, external, restart.olp, restart.orb_a)
