#!/usr/bin/env python
# JSON {"lot": "UHF/3-21G",
# JSON  "scf": "PlainSCFSolver",
# JSON  "linalg": "DenseLinalgFactory",
# JSON  "difficulty": 1,
# JSON  "description": "Basic UHF example with dense matrices"}

from pybest import context
from pybest.gbasis import get_gobasis
from pybest.gbasis.gobasis import (
    compute_eri,
    compute_kinetic,
    compute_nuclear,
    compute_nuclear_repulsion,
    compute_overlap,
)
from pybest.io.iodata import IOData
from pybest.linalg.dense import DenseLinalgFactory
from pybest.scf import AufbauOccModel
from pybest.wrappers.hf import UHF


# Load the coordinates from file.
# Use the XYZ file from PyBEST's test data directory.
fn_xyz = context.get_fn("test/methyl.xyz")
mol = IOData.from_file(fn_xyz)

# Create a Gaussian basis set
gobasis = get_gobasis("cc-pvdz", fn_xyz)

# Create a linalg factory
lf = DenseLinalgFactory(gobasis.nbasis)

# Compute integrals
olp = compute_overlap(gobasis)
kin = compute_kinetic(gobasis)
na = compute_nuclear(gobasis)
er = compute_eri(gobasis)
external = compute_nuclear_repulsion(gobasis)

# Create alpha orbitals
orb_a = lf.create_orbital()
orb_b = lf.create_orbital()

# Decide how to occupy the orbitals (5 alpha electrons, 4 beta electrons)
occ_model = AufbauOccModel(5, 4)

# Converge UHF
hf = UHF(lf, occ_model)
hf_ = hf(kin, na, er, external, olp, orb_a, orb_b)

# Write SCF results to a molden file
# First, add gobasis as attribute
hf_.gobasis = gobasis
# Ready to dump results
hf_.to_file("methyl-scf.molden")
