# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Equation of Motion Coupled Cluster implementations of a base class for EOM with
   single and double excitations.

   Variables used in this module:
    :nocc:      number of (active) occupied orbitals in the principle configuration
    :nvirt:     number of (active) virtual orbitals in the principle configuration
    :nbasis:    total number of basis functions

    Indexing convention:
     :i,j,k,..: occupied orbitals of principle configuration
     :a,b,c,..: virtual orbitals of principle configuration
     :p,q,r,..: general indices (occupied, virtual)

    P^bc_jk performs a pair permutation, i.e., P^bc_jk o_(bcjk) = o_(cbkj)

    Abbreviations used (if not mentioned in doc-strings):
     :L_pqrs: 2<pq|rs>-<pq|sr>
     :g_pqrs: <pq|rs>
"""

import numpy as np
from scipy import linalg as LA
from pathlib import Path

from pybest.filemanager import filemanager
from pybest.cache import Cache
from pybest.log import log, timer
from pybest.exceptions import (
    ArgumentError,
    RestartError,
    UnknownOption,
    DirectoryError,
    NonEmptyData,
)
from pybest.io.iodata import IOData
from pybest.io.checkpoint import CheckPoint
from pybest.utils import check_options
from pybest.linalg.cholesky import CholeskyFourIndex, CholeskyLinalgFactory
from pybest.linalg.dense import DenseLinalgFactory
from pybest.auxmat import get_diag_fock_matrix, get_fock_matrix
from pybest.orbital_utils import (
    transform_integrals,
    split_core_active,
    project_orbitals,
)
from pybest.solvers.davidson import Davidson
from pybest.units import invcm, electronvolt


class REOMCC(object):
    """Restricted Equation of Motion Coupled Cluster class

       Purpose:
       Determine the excitation energies from a given EOMCC model
       (a) Build the non-symmetruc EOM hamiltonian
       (b) Diagonalize EOM Hamiltonian

       Currently supported wavefunction models:
        * REOMpCCD
        * REOMpCCD+S
        * REOMpCCDCCS
        * REOMCCS

    """

    def __init__(self, lf, occ_model, ncore=None):
        """
           **Arguments:**

           lf
                A LinalgFactory instance.

           occ_model
                The occupation model.

           **Optional arguments:**

           ncore
                Number of frozen core orbitals. If provided, the core orbitals
                are excluded from the cluster operator.

        """
        self._lf = lf
        self._denself = DenseLinalgFactory(lf.default_nbasis)
        if ncore is None:
            self._nocc = occ_model.noccs[0]
            self._ncore = 0
        else:
            self._nocc = occ_model.noccs[0] - ncore
            self._ncore = ncore
        if self._nocc <= 0 or self._nocc > occ_model.noccs[0]:
            raise ValueError(
                f"Impossible number of frozen core orbitals! You have chosen {ncore} "
                f"frozen core orbitals out of {occ_model.noccs[0]} occupied orbitals"
            )
        self._nbasis = lf.default_nbasis - self._ncore
        self._nvirt = self._nbasis - self._nocc
        if (
            self._nvirt <= 0
            or self._nvirt > lf.default_nbasis - occ_model.noccs[0]
        ):
            raise ValueError(
                f"Impossible number of virtual orbitals: {self._nvirt}!"
            )
        self._cache = Cache()
        # e_core is not required to be assigned
        self._e_core = 0.0
        self._omega = None
        self._gamma = None
        self._tco = "td"
        self._checkpoint = CheckPoint({})
        self._checkpoint_fn = f"checkpoint_EOM{self.get_model()}.h5"
        self._todisk = False

        log.cite(
            "the EOM-pCCD-based methods",
            "boguslawski2016a",
            "boguslawski2017c",
        )

    @property
    def tco(self):
        """The tensor contraction engine used factory"""
        return self._tco

    @tco.setter
    def tco(self, new):
        self._tco = new

    @property
    def lf(self):
        """The linalg factory"""
        return self._lf

    @property
    def denself(self):
        """The dense linalg factory"""
        return self._denself

    @property
    def nbasis(self):
        """The number of basis functions"""
        return self._nbasis

    @property
    def nocc(self):
        """The number of occupied orbitals"""
        return self._nocc

    @property
    def nvirt(self):
        """The number of virtual orbitals"""
        return self._nvirt

    @property
    def ncore(self):
        """The number of core orbitals"""
        return self._ncore

    @property
    def e_core(self):
        """Core energy"""
        return self._e_core

    @property
    def omega(self):
        """Used in LR"""
        return self._omega

    @property
    def gamma(self):
        """Used in LR"""
        return self._gamma

    @property
    def checkpoint(self):
        """The iodata container that contains all data dump to disk"""
        return self._checkpoint

    @property
    def checkpoint_fn(self):
        """The filename that will be dumped to disk"""
        return self._checkpoint_fn

    @property
    def todisk(self):
        """Davidson dumps vectors to disk if True"""
        return self._todisk

    @todisk.setter
    def todisk(self, new):
        self._todisk = new

    def clear_aux_matrix(self):
        """Clear the auxiliary matrices"""
        self._cache.clear(tags="m", dealloc=True)

    def get_aux_matrix(self, select):
        """Get a matrix.

           **Arguments:**

           select
                any auxiliary matrix supported by each module.
        """
        if not f"matrix_{select}" in self._cache:
            raise ValueError(
                f"The aux_matrix {select} not found in cache. Did you use "
                "init_aux_matrix?"
            )
        return self._cache.load(f"matrix_{select}")

    def calculate_aux_matrix(self, mo1, mo2, *args, **kwargs):
        """Compute auxiliary matrices

           **Arguments:**

           mo1, mo2
                One- and two-electron integrals (some Hamiltonian matrix
                elements) in the MO basis.
                List of arguements. Only geminal coefficients [1], one- [2] and
                two-body [3] integrals are used.

        """
        self.clear_aux_matrix()
        self.update_aux_matrix(mo1, mo2, *args, **kwargs)

    def init_aux_matrix(
        self, select, dim0=None, dim1=None, dim2=None, dim3=None
    ):
        """Initialize auxiliary matrices

           **Arguments:**

           select
                (str) label of the auxiliary matrix

           **Optional Arguments:**

           dim0, dim1, dim2, dim3
                The size of the auxiliary matrix in each dimension. The number of given
                arguments not equal to None determines the order of the matrix.
        """
        # Do we need to check at all?
        # self.check_aux_matrix(select)

        # Can we do it nicer without nested if's?
        if dim0 is None:
            raise ArgumentError("At least one dimension has to be specified")
        if dim1 is None:
            alloc_size = (self.lf.create_one_index, dim0)
        else:
            if dim2 is None:
                alloc_size = (self.lf.create_two_index, dim0, dim1)
            else:
                if dim3 is None:
                    alloc_size = (self.lf.create_three_index, dim0, dim1, dim2)
                else:
                    alloc_size = (
                        self.denself.create_four_index,
                        dim0,
                        dim1,
                        dim2,
                        dim3,
                    )
        matrix, new = self._cache.load(
            f"matrix_{select}", alloc=alloc_size, tags="m",
        )
        if not new:
            raise NonEmptyData(
                f"The auxiliary matrix {select} already exists. "
                "Call clear prior to updating the auxiliary matrices."
            )
        return matrix

    def check_input(self, **kwargs):
        """Check input parameters."""
        for name, value in kwargs.items():
            check_options(
                name,
                name,
                "threshold",
                "davidson",
                "maxiter",
                "indextrans",
                "nroot",
                "tolerance",
                "tolerancev",
                "nguessv",
                "tco",
                "deleteao",
                "todisk",
                "maxvectors",
            )
        nroot = kwargs.get("nroot", 2)
        if nroot <= 1:
            raise ValueError("Number of roots must be larger than 1!")
        todisk = kwargs.get("todisk", False)
        if type(todisk) != bool:
            raise ValueError("Option 'todisk' has to be boolean type!")

    def print_info(self, **kwargs):
        dim = self.dimension
        davidson = kwargs.get("davidson", True)
        nroot = kwargs.get("nroot", 2)
        if log.do_medium:
            log(" ")
            log("Entering EOM-CC module")
            log(" ")
            log.hline("~")
            log(f"EOM{self.get_model()} framework selected")
            log.hline("~")
            log("OPTIMIZATION PARAMETERS:")
            log(f"Reference Function:            {self.get_model()}")
            log(f"Number of pairs:               {self.nocc}")
            log(f"Number of virtuals:            {self.nvirt}")
            log(f"Number of roots:               {nroot}")
            log(f"Total number of roots:         {dim}")
            if davidson:
                log("Diagonalization:               Davidson")
            else:
                log("Diagonalization:               Exact Diagonalization")
            log.hline("~")

    def get_index_d(self, index):
        """Get index for double excitation in dense representation from upper
           triangular block
        """
        triu = np.triu_indices(self.nocc * self.nvirt)
        row = triu[0][index]
        col = triu[1][index]
        i, a = self.get_index_s(row)
        j, b = self.get_index_s(col)
        return int(i), int(a), int(j), int(b)

    def get_index_s(self, index):
        """Get index for single excitation"""
        b = index % self.nvirt
        j = ((index - b) / self.nvirt) % self.nocc
        return int(j), int(b)

    def get_range(self, string, start=0):
        """ Returns dictionary with keys beginX, endX, begin(X+1), etc.
            *  string - any sequence of 'o' (occupied) and 'v' (virtual)
        """
        range_ = {}
        ind = start
        for char in string:
            if char == "o":
                range_[f"begin{ind}"] = 0
                range_[f"end{ind}"] = self.nocc
            elif char == "v":
                range_[f"begin{ind}"] = self.nocc
                range_[f"end{ind}"] = self.nbasis
            elif char == "V":
                range_[f"begin{ind}"] = 0
                range_[f"end{ind}"] = self.nvirt
            elif char == "n":
                range_[f"begin{ind}"] = 0
                range_[f"end{ind}"] = self.nbasis
            else:
                raise ValueError(f"Do not know how to handle choice {char}")
            ind += 1
        return range_

    def print_results(self, threshold):
        evals = self.checkpoint["e_ee"]
        evecs = self.checkpoint["civ_ee"]
        if log.do_medium:
            log("Final results:")
            log(" ")
            shift = self.nocc * self.nvirt
            inds = 1 + shift
            for ind in range(len(evals)):
                log(
                    f"Excitation energy:      {evals[ind].real: e} [au]  /  "
                    f"{evals[ind].real / electronvolt: e} [eV]  /  "
                    f"{evals[ind].real / invcm: e} [cm-1]"
                )
                if self.todisk:
                    filename = filemanager.temp_path(f"evecs_{ind}.h5")
                    v = IOData.from_file(str(filename))
                    evecsj = v.vector
                else:
                    evecsj = evecs[:, ind]
                kc = np.where(abs(evecsj) > threshold)[0]
                for ind2 in kc:
                    doubles = ind2 - 1 - shift >= 0
                    if ind2 == 0:
                        # Reference state
                        log(
                            f"             t_0:                      {evecsj[ind2]: 1.5f}  "
                        )
                    elif doubles:
                        index = ind2 - shift - 1
                        # (iajb)
                        if self.doubles:
                            i, a, j, b = self.get_index_d(index)
                            log(
                                f"          t_iajb:  ({i+1:3},{a + 1 + self.nocc:3},"
                                f"{j + 1:3},{b + 1 + self.nocc:3})   {evecsj[ind2]: 1.5f}"
                            )
                        else:
                            i, a = self.get_index_s(index)
                            log(
                                f"          t_iaia:  ({i+1:3},{a + 1 + self.nocc:3},"
                                f"{i + 1:3},{a + 1 + self.nocc:3})   {evecsj[ind2]: 1.5f}"
                            )
                    else:
                        i, a = self.get_index_s(ind2 - 1)
                        if self.singles:
                            log(
                                f"            t_ia:          ({i + 1:3},"
                                f"{a + 1 + self.nocc:3})   {evecsj[ind2]: 1.5f}"
                            )
                        else:
                            log(
                                f"          t_iaia:  ({i+1:3},{a + 1 + self.nocc:3},"
                                f"{i + 1:3},{a + 1 + self.nocc:3})   {evecsj[ind2]: 1.5f}"
                            )
                log(
                    f"          weight(s): {np.dot(evecsj[1:inds], evecsj[1:inds]): 1.5f}"
                )
                log(
                    f"          weight(d): {np.dot(evecsj[inds:], evecsj[inds:]): 1.5f}"
                )
            log.hline()
            log.hline("=")

    def build_guess_vectors(self, nguess, todisk, *args):
        """Build (orthonormal) set of guess vectors for each root.
           Search space contains excitations according to increasing values of
           Hdiag.

           todisk
                if True vectors are stored to disk to files
                {filemanager.temp_dir}/bvector_#int.h5
        """
        bvector = []
        dim = self.dimension
        hdiag = args[0]
        sortedind = hdiag.sort_indices(True)
        if todisk:
            #
            # Guess vector containing ground state (CC reference state)
            #
            b = self.lf.create_one_index(dim)
            b.set_element(0, 1)
            bv = IOData(vector=b)
            filename = filemanager.temp_path(f"bvector_0.h5")
            bv.to_file(filename)
            #
            # Construct Guess vectors according to hdiag
            #
            count = 1
            for ind in range(nguess):
                b.clear()
                if sortedind[ind] == 0:
                    continue
                b.set_element(sortedind[ind], 1)
                bv = IOData(vector=b)
                filename = filemanager.temp_path(f"bvector_{ind}.h5")
                bv.to_file(filename)
                count += 1
            return None, count
        else:
            #
            # Guess vector containing ground state (CC reference state)
            #
            b0 = self.lf.create_one_index(dim)
            b0.set_element(0, 1)
            bvector.append(b0)
            #
            # Construct Guess vectors according to hdiag
            #
            for ind in range(nguess):
                if sortedind[ind] == 0:
                    continue
                bis = self.lf.create_one_index(dim)
                bis.set_element(sortedind[ind], 1)
                bvector.append(bis)
            return bvector, len(bvector)

    @timer.with_section("EomDiagonalize")
    def diagonalize(self, arg, **kwargs):
        """Solve for eomham, return eigenvalues and eigenvectors

           **Arguments:**

           arg
                The one- and two-electron integrals, overlap intgrals, orbitals,
                and some IOData container containing the solution of the 
                reference calculation

           **Keywords**

            :tofile: If True, Davidson diagonalization is performed by writing all
                    vectors to disk (very I/O intensive and thus slow)
            
            :tolerance:

            :tolerancev:

            :maxiter:

            :davidson:

            :nroot:

            :nguessv:

            :maxvectors:
        """
        if log.do_medium:
            log("Starting diagonalization...")
            log.hline()
        #
        # Set default keyword arguements:
        #
        self.todisk = kwargs.get("todisk", False)
        tol = kwargs.get("tolerance", 1e-6)
        tolv = kwargs.get("tolerancev", 1e-4)
        maxiter = kwargs.get("maxiter", 200)
        davidson = kwargs.get("davidson", True)
        nroot = kwargs.get("nroot", 2)
        nguessv = kwargs.get("nguessv", None)
        maxvectors = kwargs.get("maxvectors", None)

        if davidson:
            davidson = Davidson(
                self.lf,
                nroot,
                nguessv,
                maxiter=maxiter,
                maxvectors=maxvectors,
                tolerance=tol,
                tolerancev=tolv,
                todisk=self.todisk,
            )
            e_excitation, eigen_vector_real = davidson(self, arg)
        else:
            e_excitation, eigen_vector_real = LA.eig(
                arg._array,
                b=None,
                left=False,
                right=True,
                overwrite_a=False,
                overwrite_b=False,
                check_finite=True,
            )
            ind = np.argsort(e_excitation)
            e_excitation = e_excitation[ind].real
            eigen_vector_real = eigen_vector_real[:, ind].real

        self.checkpoint.update("e_ee", e_excitation[:nroot])
        if not self.todisk:
            self.checkpoint.update("civ_ee", eigen_vector_real[:, :nroot])
        else:
            # Store 0 (float), otherwise we cannot dump to hdf5
            self.checkpoint.update("civ_ee", 0)

    def dump_eri(self, eri, checkpoint_fn="checkpoint_eri.h5"):
        """Dump orbitals for restart

           **Arguments:**

           eri
                The electron repulsion integrals

           checkpoint_fn
                The filename of the checkpoint file.
        """
        # Use the pybest.io interface to write the checkpoint file
        mol = IOData(eri=eri)
        # Some tmp dir
        filename = filemanager.temp_path(f"{checkpoint_fn}")
        mol.to_file(filename)

    def load_eri(self, eri, checkpoint_fn="checkpoint_eri.h5"):
        """Dump orbitals for restart

           **Arguments:**

           eri
                The electron repulsion integrals

           checkpoint_fn
                The filename of the checkpoint file.
        """
        # Use the pybest.io interface to write the checkpoint file
        filename = filemanager.temp_path(f"{checkpoint_fn}")
        mol = IOData.from_file(filename)
        eri.replace_array(mol.eri)

    def unmask_args(self, *args, **kwargs):
        raise NotImplementedError

    def construct_ham(self, args):
        raise NotImplementedError

    @timer.with_section("REOMCC")
    def __call__(self, *args, **kwargs):
        """Solve the non-symmetric eigenvalue problem in EOM-pCCD

           Currently supported wavefunction models (Psi_0):
            * RpCCD

           Currently supported CI operators:
            * Pair excitation operator

           **Arguments:**

           args
                One- and two-body integrals (some Hamiltonian matrix elements)
                expressed in the AO basis.
                It contains the AO/MO coefficients and the geminal coefficients
                (if Psi_0 = pCCD) as second arguement.

           **Keywords:**

                Contains reference energy and solver specific input parameters:
                * davidson:    default True
                * tolerance:   tolerance for energies (default 1e-6)
                * tolerancev:  tolerance for eigenvectors (default 1e-5)
                * threshold:   printing threshold for amplitudes (default 0.1)
                * maxiter:     maximum number of iterations (default 200)
                * indextrans: (str) 4-index transformation. One of ``einsum``,
                              ``tensordot`` (default ``tensordot``)

           **Returns**

                List of excitation energies and EOM CI coeff

        """
        #
        # Set default keyword arguements:
        #
        thresh = kwargs.get("threshold", 0.1)
        davidson = kwargs.get("davidson", True)
        indextrans = kwargs.get("indextrans", "tensordot")
        self.tco = kwargs.get("tco", "td")

        #
        # Check input parameters:
        #
        self.check_input(**kwargs)

        #
        # Unmask arguments
        #
        one, two, orbs, olp = self.unmask_args(*args, **kwargs)

        #
        # Print method specific information
        #
        self.print_info(**kwargs)

        #
        # Transform integrals:
        #
        if self.ncore > 0:
            cas = split_core_active(
                one,
                two,
                orbs,
                e_core=self.e_core,
                ncore=self.ncore,
                nactive=self.nbasis,
                indextrans=indextrans,
            )
            mo1 = cas.one
            mo2 = cas.two
        else:
            ti = transform_integrals(one, two, orbs, indextrans=indextrans)
            mo1 = ti.one[0]
            mo2 = ti.two[0]
        self.dump_eri(two)
        two.__del__()

        #
        # Construct auxiliary matrices (checks also type of arguments):
        #
        log("Building effective Hamiltonian...")
        self.calculate_aux_matrix(mo1, mo2)
        fargs = []
        if isinstance(self.lf, CholeskyLinalgFactory):
            # append Cholesky vectors to arguments
            fargs.append(mo2)
            mo1.__del__()
        else:
            mo2.__del__()
            mo1.__del__()
        log("...done.")

        if davidson:
            self.diagonalize(fargs, **kwargs)
        else:
            log("Building Hamiltonian matrix...")
            eomham = self.construct_ham(fargs)
            log("...done")
            eomham = self.construct_ham(fargs)
            self.clear_aux_matrix()
            self.diagonalize(eomham, **kwargs)

        #
        # Do Checkpoint
        #
        self.checkpoint.to_file(self.checkpoint_fn)

        #
        # Pint final results
        #
        self.print_results(thresh)
        #
        # Clear up and finish
        #
        self.clear_aux_matrix()
        if isinstance(self.lf, CholeskyLinalgFactory):
            mo2.__del__()
        #
        # Read again eri
        #
        self.load_eri(two)

        return self.checkpoint()
