# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Equation of Motion Coupled Cluster implementations of EOM-CCS

   Child class of REOMCC class.
"""

import gc

from pybest.ee_eom.eom_base import REOMCC
from pybest.log import log, timer
from pybest.utils import (
    check_type,
    check_options,
    unmask,
    unmask_orb,
)
from pybest.linalg import (
    Orbital,
    DenseOrbital,
    TwoIndex,
    DenseTwoIndex,
    FourIndex,
    DenseFourIndex,
)
from pybest.featuredlists import OneBodyHamiltonian, TwoBodyHamiltonian
from pybest.auxmat import get_fock_matrix
from pybest.exceptions import ArgumentError

__all__ = [
    "REOMCCS",
]


class REOMCCS(REOMCC):
    """Performs a EOM-CCS calculation, which is equivalent to CIS."""

    @property
    def dimension(self):
        """Number of unknowns"""
        return self.nocc * self.nvirt + 1

    @property
    def singles(self):
        return True

    @property
    def pairs(self):
        return False

    @property
    def doubles(self):
        return False

    def get_model(self):
        return "CCS"

    def unmask_args(self, *args, **kwargs):
        #
        # olp
        #
        olp = unmask("olp", *args, **kwargs)
        self.checkpoint.update("olp", olp)
        #
        # orb
        #
        orbs = unmask_orb(*args, **kwargs)
        if orbs:
            orbs = orbs[0]
            self.checkpoint.update("orb_a", orbs.copy())
        #
        # t_1
        #
        t_1 = unmask("t_1", *args, **kwargs)
        if t_1 is not None:
            self.checkpoint.update("t_1", t_1)
        #
        # 1-e ints and 2-e ints
        #
        one = self.lf.create_two_index(label="one")
        for arg in args:
            if isinstance(arg, TwoIndex):
                if arg.label in OneBodyHamiltonian:
                    one.iadd(arg)
            elif isinstance(arg, FourIndex):
                if arg.label in TwoBodyHamiltonian:
                    two = arg
        if olp is None:
            raise ArgumentError(
                "Cannot find overlap integrals in LCC function call."
            )
        return one, two, orbs, olp

    @timer.with_section("BuildHCCS")
    def construct_ham(self, args):
        tia = self.checkpoint["t_1"]
        eomham = self.lf.create_two_index(
            self.nvirt * self.nocc + 1, self.nocc * self.nvirt + 1
        )
        #
        # Get auxiliary matrices
        #
        fock = self.get_aux_matrix("fock")
        miaka = self.get_aux_matrix("miaka")
        miaic = self.get_aux_matrix("miaic")
        miakc = self.get_aux_matrix("miakc")
        miakc_ = miakc.copy()

        eomham.assign(
            fock._array[: self.nocc, self.nocc :].ravel(), end0=1, begin1=1
        )
        eomham.iscale(2.0)
        miakc_.iadd_expand_two_to_four("1-3", miaic, 1.0)
        miakc_.iadd_expand_two_to_four("0-2", miaka, 1.0)
        eomham.assign(miakc_._array, begin0=1, begin1=1)

        return eomham

    @timer.with_section("BHDiagpCCDCCS")
    def compute_h_diag(self, args):
        eomhamdiag = self.lf.create_one_index(self.dimension)
        #
        #
        #
        fock = self.get_aux_matrix("fock")
        miaka = self.get_aux_matrix("miaka")
        miaic = self.get_aux_matrix("miaic")
        miakc = self.get_aux_matrix("miakc")

        # Singles
        diags = self.lf.create_two_index(self.nocc, self.nvirt)
        miakc.slice("abab->ab", diags)
        tmp = miaka.copy_diagonal()
        diags.iadd_expand_one_to_two("a->ab", tmp, 1.0)
        tmp = miaic.copy_diagonal()
        diags.iadd_expand_one_to_two("b->ab", tmp, 1.0)

        eomhamdiag.assign(diags.ravel(), begin0=1)
        return eomhamdiag

    @timer.with_section("SubHCCS")
    def build_subspace_hamiltonian(self, bvector, hdiag, args):
        #
        # Get auxiliary matrices
        #
        fock = self.get_aux_matrix("fock")
        miaka = self.get_aux_matrix("miaka")
        miaic = self.get_aux_matrix("miaic")
        miakc = self.get_aux_matrix("miakc")

        #
        # Get ranges
        #
        ov2 = self.get_range("ov", start=2)
        #
        # Calculate sigma vector (H.bvector)_kc
        #
        sigmas = self.lf.create_two_index(self.nocc, self.nvirt)
        bvs = self.denself.create_two_index(self.nocc, self.nvirt)
        #
        # reshape bvector
        #
        bvs.assign(bvector, begin2=1)
        #
        #
        #
        sigma = self.lf.create_one_index(self.dimension)
        #
        # Reference vector R_0
        #
        # X0,kc rkc
        sum0_ = bvs.contract("ab,ab", fock, **ov2) * 2.0
        sigma.set_element(0, sum0_)
        #
        # Single excitations
        #
        # Xkcld rld
        miakc.contract("abcd,cd->ab", bvs, sigmas, clear=True)
        # Xkclc rlc
        miaka.contract("ab,bc->ac", bvs, sigmas)
        # Xkckd rkd
        bvs.contract("ab,cb->ac", miaic, sigmas)

        sigma.assign(sigmas.ravel(), begin0=1)

        return sigma

    @timer.with_section("AuxMatEOMCCS")
    def update_aux_matrix(self, mo1, mo2):
        """Derive all auxiliary matrices.

           **Arguments:**

           mo1, mo2
                one- and two-electron integrals to be sorted.

        """
        tia = self.checkpoint["t_1"]
        check_type("tia", tia, DenseTwoIndex, TwoIndex)
        #
        # Get ranges
        #
        ov2 = self.get_range("ov", start=2)
        oo2 = self.get_range("oo", start=2)
        vv2 = self.get_range("vv", start=2)
        voov = self.get_range("voov")
        vovo = self.get_range("vovo")
        ooov = self.get_range("ooov")
        oovo = self.get_range("oovo")
        ovvv = self.get_range("ovvv")
        #
        # Inactive Fock matrix
        #
        fock = self.init_aux_matrix("fock", self.nbasis, self.nbasis)
        get_fock_matrix(fock, mo1, mo2, self.nocc)
        #
        # <vo||ov>+<vo|ov>
        #
        lvoov = self.denself.create_four_index(
            self.nvirt, self.nocc, self.nocc, self.nvirt
        )
        mo2.slice("abcd->abcd", lvoov, 2.0, True, **voov)
        mo2.slice("abcd->abdc", lvoov, -1.0, False, **vovo)
        #
        # <oo||ov>+<oo|ov>
        #
        looov = self.denself.create_four_index(
            self.nocc, self.nocc, self.nocc, self.nvirt
        )
        mo2.slice("abcd->abcd", looov, 2.0, True, **ooov)
        mo2.slice("abcd->abdc", looov, -1.0, False, **oovo)
        # Improve with Cholesky
        # <ov||vv>+<ov|vv>
        #
        lovvv = self.denself.create_four_index(
            self.nocc, self.nvirt, self.nvirt, self.nvirt
        )
        mo2.slice("abcd->abcd", lovvv, 2.0, True, **ovvv)
        mo2.slice("abcd->abdc", lovvv, -1.0, False, **ovvv)
        #
        # Mia,ka
        #
        aux_matiaka = self.init_aux_matrix("miaka", self.nocc, self.nocc)
        # fik
        aux_matiaka.iadd(fock, -1.0, **oo2)
        # tie fke
        tia.contract("ab,cb->ac", fock, aux_matiaka, **ov2)
        # -L_kmie t_me
        looov.contract("abcd,bd->ca", tia, aux_matiaka, factor=-1.0)
        #
        # Mia,ic
        #
        aux_matiaic = self.init_aux_matrix("miaic", self.nvirt, self.nvirt)
        # fac
        aux_matiaic.iadd(fock, 1.0, **vv2)
        # -tma fmc
        tia.contract("ab,ac->bc", fock, aux_matiaic, factor=-1.0, **ov2)
        # L_maec t_me
        lovvv.contract("abcd,ac->bd", tia, aux_matiaic, factor=-1.0)
        #
        # Aux matrix for Mia,kc
        #
        miakc = self.init_aux_matrix(
            "miakc", self.nocc, self.nvirt, self.nocc, self.nvirt
        )
        # L_akic
        lvoov.slice("abcd->cabd", miakc, 1.0)
        # L_mkic t_ma
        looov.contract("abcd,ae->cebd", tia, miakc, factor=-1.0)
        # L_kace t_ie
        lovvv.contract("abcd,ed->ebac", tia, miakc)

        del lovvv
        del looov
        del lvoov
        gc.collect()
