# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Equation of Motion Coupled Cluster implementations

   Variables used in this module:
    :nocc:      number of (active) occupied orbitals in the principle configuration
    :nvirt:     number of (active) virtual orbitals in the principle configuration
    :nbasis:    total number of basis functions
    :energy:    the energy correction, list that can contain different
                contributions
    :eomcoeff:the CI amplitudes from a given EOM model

    Indexing convention:
     :i,j,k,..: occupied orbitals of principle configuration
     :a,b,c,..: virtual orbitals of principle configuration
     :p,q,r,..: general indices (occupied, virtual)

    P^bc_jk performs a pair permutation, i.e., P^bc_jk o_(bcjk) = o_(cbkj)

    Abbreviations used (if not mentioned in doc-strings):
     :L_pqrs: 2<pq|rs>-<pq|sr>
     :g_pqrs: <pq|rs>
"""

from pybest.log import log, timer
from pybest.utils import (
    check_options,
    unmask,
    unmask_orb,
)
from pybest.auxmat import get_diag_fock_matrix
from pybest.ee_eom.eom_base import REOMCC
from pybest.exceptions import ArgumentError
from pybest.linalg import (
    Orbital,
    DenseOrbital,
    TwoIndex,
    DenseTwoIndex,
    FourIndex,
    DenseFourIndex,
)
from pybest.featuredlists import OneBodyHamiltonian, TwoBodyHamiltonian


__all__ = [
    "REOMpCCD",
]


class REOMpCCD(REOMCC):
    @property
    def dimension(self):
        """Number of unknowns"""
        return self.nocc * self.nvirt + 1

    @property
    def singles(self):
        return False

    @property
    def pairs(self):
        return True

    @property
    def doubles(self):
        return False

    def get_model(self):
        return "pCCD"

    def unmask_args(self, *args, **kwargs):
        #
        # olp
        #
        olp = unmask("olp", *args, **kwargs)
        if olp is None:
            raise ArgumentError(
                "Cannot find overlap integrals in LCC function call."
            )
        self.checkpoint.update("olp", olp)
        #
        # orb
        #
        orbs = unmask_orb(*args, **kwargs)
        if orbs:
            orbs = orbs[0]
            self.checkpoint.update("orb_a", orbs.copy())
        else:
            raise ArgumentError("Cannot find orbitals.")
        #
        # t_p
        #
        t_p = unmask("t_p", *args, **kwargs)
        if t_p is not None:
            self.checkpoint.update("t_p", t_p)
        else:
            raise ArgumentError("Cannot find Tp amplitudes.")
        #
        # 1-e ints and 2-e ints
        #
        one = self.lf.create_two_index(label="one")
        for arg in args:
            if isinstance(arg, TwoIndex):
                if arg.label in OneBodyHamiltonian:
                    one.iadd(arg)
            elif isinstance(arg, FourIndex):
                if arg.label in TwoBodyHamiltonian:
                    two = arg
        return one, two, orbs, olp

    @timer.with_section("BuildHpCCD")
    def construct_ham(self, args):
        eomham = self.lf.create_two_index(
            self.nvirt * self.nocc + 1, self.nocc * self.nvirt + 1
        )

        G_ppqq = self.get_aux_matrix("gppqq")
        mppia = self.get_aux_matrix("mppia")
        mppiaic = self.get_aux_matrix("mppiaic")
        mppiaka = self.get_aux_matrix("mppiaka")

        miakc = self.denself.create_four_index(
            self.nocc, self.nvirt, self.nocc, self.nvirt
        )
        eomham.assign(
            G_ppqq._array[: self.nocc, self.nocc :].ravel(), end0=1, begin1=1
        )
        miakc.iadd_expand_two_to_four("diag", mppia, 1.0)
        miakc.iadd_expand_three_to_four("1-3-1-2", mppiaic, 1.0)
        miakc.iadd_expand_three_to_four("0-2-0-2", mppiaka, 1.0)
        eomham.assign(miakc._array, begin0=1, begin1=1)

        return eomham

    @timer.with_section("BuildHDiagpCCD")
    def compute_h_diag(self, args):
        eomhamdiag = self.lf.create_one_index(self.nvirt * self.nocc + 1)

        mppia = self.get_aux_matrix("mppia")
        mppiaic = self.get_aux_matrix("mppiaic")
        mppiaka = self.get_aux_matrix("mppiaka")

        diagp = mppia.copy()
        mppiaic.slice("abb->ab", diagp, 1.0, False)
        mppiaka.slice("aba->ab", diagp, 1.0, False)
        eomhamdiag.assign(diagp.ravel(), begin0=1)
        return eomhamdiag

    @timer.with_section("SubHpCCD")
    def build_subspace_hamiltonian(self, bvector, hdiag, args):
        #
        # Get auxiliary matrices
        #
        G_ppqq = self.get_aux_matrix("gppqq")
        mppiaka = self.get_aux_matrix("mppiaka")
        mppiaic = self.get_aux_matrix("mppiaic")
        mppia = self.get_aux_matrix("mppia")
        #
        # Get ranges
        #
        ov2 = self.get_range("ov", start=2)
        #
        # Calculate sigma vector (H.bvector)_kc
        #
        sigmap = self.lf.create_two_index(self.nocc, self.nvirt)
        bvp = self.lf.create_two_index(self.nocc, self.nvirt)
        sigma = self.lf.create_one_index(self.nocc * self.nvirt + 1)
        bvp.assign(bvector, begin2=1)
        sum0_ = bvp.contract("ab,ab", G_ppqq, **ov2)
        sigma.set_element(0, sum0_)
        #
        # Pair excitations
        #
        # Xlk rlclc
        mppiaka.contract("abc,cb->ab", bvp, sigmap, clear=True)
        # Xcd rkdkd
        mppiaic.contract("abc,ac->ab", bvp, sigmap)
        # Xkc rkckc
        sigmap.iadd_mult(mppia, bvp, 1.0)
        #
        # Assign new sigma vector
        #
        sigma.assign(sigmap.ravel(), begin0=1)
        return sigma

    @timer.with_section("AuxMatEOMpCCD")
    def update_aux_matrix(self, mo1, mo2):
        """Derive all auxiliary matrices.
           fock_pq:     one_pq + sum_m(2<pm|qm> - <pm|mq>),
           lpqpq:   2<pq|pq>-<pq|qp>,
           gpqpq:   <pq|pq>,

           **Arguments:**

           mo1, mo2
                one- and two-electron integrals to be sorted.

        """
        cia = self.checkpoint["t_p"]
        #
        # Get ranges
        #
        ov = self.get_range("ov")
        vo = self.get_range("vo")
        vo2 = self.get_range("vo", start=2)
        ov2 = self.get_range("ov", start=2)
        oov = self.get_range("oov")
        ovv = self.get_range("ovv")
        #
        # 2<pq|pq>-<pq|pq>
        #
        aux_mat1 = self.init_aux_matrix("lpqpq", self.nbasis, self.nbasis)
        mo2.slice("abab->ab", aux_mat1, 2.0, True)
        mo2.slice("abba->ab", aux_mat1, -1.0, False)
        #
        # Inactive Fock matrix
        #
        aux_mat2 = self.lf.create_one_index(self.nbasis)
        get_diag_fock_matrix(aux_mat2, mo1, mo2, self.nocc)
        #
        # <pp|qq>
        #
        gppqq = self.init_aux_matrix("gppqq", self.nbasis, self.nbasis)
        mo2.slice("aabb->ab", gppqq)
        #
        # Aux matrix for Mppiaia,kaka (iak)
        #
        mppiaka = self.init_aux_matrix(
            "mppiaka", self.nocc, self.nvirt, self.nocc
        )
        # G_iikk
        mppiaka.iadd_expand_two_to_three(
            "ac->abc", gppqq, 1.0, end0=self.nocc, end1=self.nocc
        )
        # Used in pCCD-based methods
        # <kk|aa> cia
        cia.contract(
            "ab,bc->abc", gppqq, mppiaka, factor=-2.0, **vo2, select="einsum"
        )
        # <kk|ee> cie
        tmp = self.lf.create_two_index(self.nocc)
        cia.contract("ab,bc->ac", gppqq, tmp, **vo2)
        mppiaka.iadd_expand_two_to_three("ac->abc", tmp, 1.0)
        #
        # Aux matrix for Mppiaia,icic (iac)
        #
        mppiaic = self.init_aux_matrix(
            "mppiaic", self.nocc, self.nvirt, self.nvirt
        )
        # G_aacc
        mppiaic.iadd_expand_two_to_three(
            "bc->abc", gppqq, 1.0, begin0=self.nocc, begin1=self.nocc
        )
        # <ii|cc> cia
        cia.contract(
            "ab,ac->abc", gppqq, mppiaic, factor=-2.0, **ov2, select="einsum"
        )
        # <mm|cc> cma
        tmp = self.lf.create_two_index(self.nvirt)
        cia.contract("ab,ac->bc", gppqq, tmp, **ov2)
        mppiaic.iadd_expand_two_to_three("bc->abc", tmp, 1.0)
        #
        # Aux matrix for Mppiaia,iaia (ia)
        #
        mppia = self.init_aux_matrix("mppia", self.nocc, self.nvirt)
        # F_aa - F_ii
        mppia.iadd_expand_one_to_two("a->ab", aux_mat2, -2.0, end0=self.nocc)
        mppia.iadd_expand_one_to_two("b->ab", aux_mat2, 2.0, begin0=self.nocc)
        # L_iaia
        mppia.iadd(aux_mat1, -2.0, **ov2)
        # <ii|aa> cia
        mppia.iadd_mult(gppqq, cia, 4.0, **ov)
        # <ii|dd> cid
        tempk = self.lf.create_one_index(self.nocc)
        cia.contract("ab,ab->a", gppqq, tempk, **ov2)
        mppia.iadd_expand_one_to_two("a->ab", tempk, -2.0)
        # <kk|aa> cka
        tempc = self.lf.create_one_index(self.nvirt)
        cia.contract("ab,ab->b", gppqq, tempc, **ov2)
        mppia.iadd_expand_one_to_two("b->ab", tempc, -2.0)
