# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Equation of Motion Coupled Cluster implementations of EOM-pCCD+S and EOM-pCCD-CCS,
   that is, pCCD with single excitations.

   Child class of REOMCC class.
"""

import gc

from pybest.ee_eom.eom_base import REOMCC
from pybest.log import log, timer
from pybest.utils import (
    check_type,
    check_options,
    unmask,
    unmask_orb,
)
from pybest.auxmat import get_fock_matrix
from pybest.linalg import (
    Orbital,
    DenseOrbital,
    TwoIndex,
    DenseTwoIndex,
    FourIndex,
    DenseFourIndex,
)
from pybest.featuredlists import OneBodyHamiltonian, TwoBodyHamiltonian
from pybest.exceptions import ArgumentError


__all__ = [
    "REOMpCCDS",
    "REOMpCCDCCS",
]


class REOMpCCDS(REOMCC):
    @property
    def dimension(self):
        """Number of unknowns"""
        return 2 * self.nocc * self.nvirt + 1

    @property
    def singles(self):
        return True

    @property
    def pairs(self):
        return True

    @property
    def doubles(self):
        return False

    def get_model(self):
        return "pCCD+S"

    def unmask_args(self, *args, **kwargs):
        #
        # olp
        #
        olp = unmask("olp", *args, **kwargs)
        if olp is None:
            raise ArgumentError(
                "Cannot find overlap integrals in LCC function call."
            )
        self.checkpoint.update("olp", olp)
        #
        # orb
        #
        orbs = unmask_orb(*args, **kwargs)
        if orbs:
            orbs = orbs[0]
            self.checkpoint.update("orb_a", orbs.copy())
        else:
            raise ArgumentError("Cannot find orbitals.")
        #
        # t_p
        #
        t_p = unmask("t_p", *args, **kwargs)
        if t_p is not None:
            self.checkpoint.update("t_p", t_p)
        else:
            raise ArgumentError("Cannot find Tp amplitudes.")
        #
        # 1-e ints and 2-e ints
        #
        one = self.lf.create_two_index(label="one")
        for arg in args:
            if isinstance(arg, TwoIndex):
                if arg.label in OneBodyHamiltonian:
                    one.iadd(arg)
            elif isinstance(arg, FourIndex):
                if arg.label in TwoBodyHamiltonian:
                    two = arg
        return one, two, orbs, olp

    @timer.with_section("BuildHpCCDS")
    def construct_ham(self, args):
        cia = self.checkpoint["t_p"]
        eomham = self.lf.create_two_index(
            2 * self.nvirt * self.nocc + 1, 2 * self.nocc * self.nvirt + 1
        )
        #
        # effective Hamiltonian matrix elements
        #
        fock = self.get_aux_matrix("fock")
        G_ppqq = self.get_aux_matrix("gppqq")
        L_pqrq = self.get_aux_matrix("lpqrq")
        miaka = self.get_aux_matrix("miaka")
        miaic = self.get_aux_matrix("miaic")
        miakc = self.get_aux_matrix("miakc")
        m1kald = self.get_aux_matrix("m1kald")
        x1kcda = self.get_aux_matrix("x1kcda")
        m1iald = self.get_aux_matrix("m1iald")
        mpiaka = self.get_aux_matrix("mpiaka")
        mpiaic = self.get_aux_matrix("mpiaic")
        mppiaka = self.get_aux_matrix("mppiaka")
        mppiaic = self.get_aux_matrix("mppiaic")
        mppia = self.get_aux_matrix("mppia")
        m10ia = self.get_aux_matrix("m10ia")

        row = 1
        shift = self.nocc * self.nvirt
        for k in range(self.nocc):
            for c in range(self.nvirt):
                ci = c + self.nocc
                #
                # Single excitations:
                #
                # H_0,kc
                #
                eomham._array[0, row] = 2 * fock._array[k, ci]
                #
                # H_kc,0
                #
                eomham._array[row, 0] = m10ia.get_element(k, c)
                #
                # H_kc,kkcc
                #
                eomham._array[row, row + shift] += m1iald.get_element(k, c)

                #
                # Pair excitations:
                #
                #
                # Diagonal elements H_kc,kc
                #
                eomham._array[row + shift, row + shift] += mppia.get_element(
                    k, c
                )
                #
                # H_0,kc
                #
                eomham._array[0, row + shift] = G_ppqq._array[k, ci]

                #
                # H_kc,kd
                #
                col = k * self.nvirt + 1
                for d in range(self.nvirt):
                    di = d + self.nocc
                    #
                    # H_kkcc,kkdd
                    #
                    eomham._array[
                        row + shift, col + shift
                    ] += mppiaic.get_element(k, c, d)
                    #
                    # H_kc,kd (singles contribution)
                    #
                    eomham._array[row, col] += miaic.get_element(c, d)
                    #
                    # H_kc,kkdd (coupling pair-doubles-singles contribution)
                    #
                    eomham._array[row, col + shift] += x1kcda.get_element(
                        k, c, d, d
                    )
                    #
                    # H_kkdd,kc (coupling pair-doubles-singles contribution)
                    #
                    sumkkddkc = mpiaic.get_element(c, k, d)
                    eomham._array[col + shift, row] += sumkkddkc
                    #
                    col = col + 1

                #
                # H_kc,lc
                #
                col = c + 1
                for l in range(self.nocc):
                    eomham._array[
                        row + shift, col + shift
                    ] += mppiaka.get_element(k, c, l)
                    #
                    # H_kc,lc (singles contribution)
                    #
                    eomham._array[row, col] += miaka.get_element(k, l)
                    #
                    # H_kc,llcc (coupling pair-doubles-singles contribution)
                    #
                    eomham._array[row, col + shift] += m1kald.get_element(
                        l, l, k, c
                    )
                    #
                    # H_llcc,kc (coupling pair-doubles-singles contribution)
                    #
                    sumllcckc = mpiaka.get_element(c, k, l)
                    eomham._array[col + shift, row] += sumllcckc
                    #
                    # H_kc,ld (singles and coupling to pair-doubles)
                    #
                    cols = l * self.nvirt + 1
                    for d in range(self.nvirt):
                        di = d + self.nocc
                        #
                        # H_kc,ld (singles)
                        #
                        eomham._array[row, cols] += miakc.get_element(
                            k, c, l, d
                        )
                        #
                        # H_lldd,kc (coupling to pair-doubles)
                        #
                        eomham._array[cols + shift, row] += (
                            2.0
                            * L_pqrq.get_element(k, di, ci)
                            * cia._array[l, d]
                        )
                        eomham._array[cols + shift, row] -= (
                            2.0
                            * L_pqrq.get_element(k, l, ci)
                            * cia._array[l, d]
                        )

                        cols = cols + 1

                    col = col + self.nvirt
                row = row + 1
        return eomham

    @timer.with_section("BHDiagpCCDS")
    def compute_h_diag(self, args):
        cia = self.checkpoint["t_p"]
        check_type("cia", cia, DenseTwoIndex, TwoIndex)
        lens = self.nvirt * self.nocc
        dim = 2 * lens + 1
        eomhamdiag = self.lf.create_one_index(dim)
        #
        #
        #
        fock = self.get_aux_matrix("fock")
        miaka = self.get_aux_matrix("miaka")
        miaic = self.get_aux_matrix("miaic")
        miakc = self.get_aux_matrix("miakc")
        mppia = self.get_aux_matrix("mppia")
        mppiaic = self.get_aux_matrix("mppiaic")
        mppiaka = self.get_aux_matrix("mppiaka")

        # Singles
        diags = self.lf.create_two_index(self.nocc, self.nvirt)
        miakc.slice("abab->ab", diags)
        tmp = miaka.copy_diagonal()
        diags.iadd_expand_one_to_two("a->ab", tmp, 1.0)
        tmp = miaic.copy_diagonal()
        diags.iadd_expand_one_to_two("b->ab", tmp, 1.0)
        # Pairs:
        diagp = mppia.copy()
        mppiaic.slice("abb->ab", diagp, 1.0, False)
        mppiaka.slice("aba->ab", diagp, 1.0, False)
        eomhamdiag.assign(diags.ravel(), begin0=1, end0=(lens + 1))
        eomhamdiag.assign(diagp.ravel(), begin0=(lens + 1))
        return eomhamdiag

    @timer.with_section("SubHpCCDS")
    def build_subspace_hamiltonian(self, bvector, hdiag, args):
        cia = self.checkpoint["t_p"]
        check_type("cia", cia, DenseTwoIndex, TwoIndex)
        #
        # Get ranges
        #
        ov2 = self.get_range("ov", start=2)
        oov = self.get_range("oov")
        ovv = self.get_range("ovv")
        #
        # Get auxiliary matrices
        #
        fock = self.get_aux_matrix("fock")
        G_ppqq = self.get_aux_matrix("gppqq")
        L_pqrq = self.get_aux_matrix("lpqrq")
        miaka = self.get_aux_matrix("miaka")
        miaic = self.get_aux_matrix("miaic")
        miakc = self.get_aux_matrix("miakc")
        m1kald = self.get_aux_matrix("m1kald")
        x1kcda = self.get_aux_matrix("x1kcda")
        m1iald = self.get_aux_matrix("m1iald")
        mpiaka = self.get_aux_matrix("mpiaka")
        mpiaic = self.get_aux_matrix("mpiaic")
        mppiaka = self.get_aux_matrix("mppiaka")
        mppiaic = self.get_aux_matrix("mppiaic")
        mppia = self.get_aux_matrix("mppia")
        m10ia = self.get_aux_matrix("m10ia")

        #
        # Calculate sigma vector (H.bvector)_kc
        #
        lens = self.nvirt * self.nocc
        shift = lens + 1
        sigmas = self.lf.create_two_index(self.nocc, self.nvirt)
        sigmap = self.lf.create_two_index(self.nocc, self.nvirt)
        bvs = self.denself.create_two_index(self.nocc, self.nvirt)
        bvp = self.denself.create_two_index(self.nocc, self.nvirt)
        #
        # reshape bvector
        #
        bvs.assign(bvector, begin2=1, end2=shift)
        bvp.assign(bvector, begin2=shift)
        #
        #
        #
        sigma = self.lf.create_one_index(self.dimension)
        #
        # Reference vector R_0
        #
        # X0,kc rkc
        sum0_ = 0.0
        sum0_ = bvs.contract("ab,ab", fock, **ov2) * 2.0
        sum0_ += bvp.contract("ab,ab", G_ppqq, **ov2)
        sigma.set_element(0, sum0_)
        #
        # Single excitations
        #
        # Xkcld rld
        miakc.contract("abcd,cd->ab", bvs, sigmas, clear=True)
        # Xkclc rlc
        miaka.contract("ab,bc->ac", bvs, sigmas)
        # Xkckd rkd
        bvs.contract("ab,cb->ac", miaic, sigmas)
        # Xllkc rlclc
        m1kald.contract("aabc,ac->bc", bvp, sigmas)
        # Xclde rkdle
        x1kcda.contract("abcc,ac->ab", bvp, sigmas)
        # Xkc rkckc
        sigmas.iadd_mult(m1iald, bvp, 1.0)
        # Xkc0 r0
        sigmas.iadd(m10ia, factor=bvector.get_element(0))
        #
        # Pair excitations
        #
        mppiaka.contract("abc,cb->ab", bvp, sigmap, clear=True)
        # Xcd rkdkd
        mppiaic.contract("abc,ac->ab", bvp, sigmap)
        # Xkc rkckc
        sigmap.iadd_mult(mppia, bvp, 1.0)
        # Coupling pairs-singles
        mpiaka.contract("abc,ba->ca", bvs, sigmap)
        # Xdkc rkd
        mpiaic.contract("abc,ba->bc", bvs, sigmap)
        # Xldck rld
        tmp = self.lf.create_one_index(self.nvirt)
        L_pqrq.contract("abc,ac->b", bvs, tmp, **ovv)
        cia.contract("ab,b->ab", tmp, sigmap, factor=2.0)
        tmp = self.lf.create_one_index(self.nocc)
        L_pqrq.contract("abc,ac->b", bvs, tmp, **oov)
        cia.contract("ab,a->ab", tmp, sigmap, factor=-2.0)

        sigma.assign(sigmas.ravel(), begin0=1, end0=shift)
        sigma.assign(sigmap.ravel(), begin0=shift)

        return sigma

    @timer.with_section("AuxMatEOMpCCDS")
    def update_aux_matrix(self, mo1, mo2):
        """Derive all auxiliary matrices.

           **Arguments:**

           mo1, mo2
                one- and two-electron integrals to be sorted.

        """
        cia = self.checkpoint["t_p"]
        check_type("cia", cia, DenseTwoIndex, TwoIndex)
        #
        # Get ranges
        #
        oo = self.get_range("oo")
        ov = self.get_range("ov")
        vo = self.get_range("vo")
        ov2 = self.get_range("ov", start=2)
        oo2 = self.get_range("oo", start=2)
        vv2 = self.get_range("vv", start=2)
        vo2 = self.get_range("vo", start=2)
        vv = self.get_range("vv")
        ovo = self.get_range("ovo")
        voo = self.get_range("voo")
        vov = self.get_range("vov")
        voo3 = self.get_range("voo", 3)
        vov3 = self.get_range("vov", 3)
        vvo = self.get_range("vvo")
        oov = self.get_range("oov")
        ovv = self.get_range("ovv")
        oovv = self.get_range("oovv")
        voov = self.get_range("voov")
        vovo = self.get_range("vovo")
        ooov = self.get_range("ooov")
        oovo = self.get_range("oovo")
        ovvv = self.get_range("ovvv")
        #
        # 2<pq|pq>-<pq|pq> and <pq|pq>
        #
        aux_mat1 = self.init_aux_matrix("lpqpq", self.nbasis, self.nbasis)
        mo2.slice("abab->ab", aux_mat1, 2.0, True)
        mo2.slice("abba->ab", aux_mat1, -1.0, False)
        #
        # <pq||rq>+<pq|rq>
        #
        aux_mat2 = self.init_aux_matrix(
            "lpqrq", self.nbasis, self.nbasis, self.nbasis
        )
        mo2.slice("abcb->abc", aux_mat2, 2.0, True)
        # this is really slow and we do not need to do it this way
        # exchange part is added in <pq|rr> block
        # two_mo[0].slice('abbc->abc', matrix5,-1.0, False)
        #
        # <pq|rr>
        #
        aux_mat3 = self.init_aux_matrix(
            "gpqrr", self.nbasis, self.nbasis, self.nbasis
        )
        mo2.slice("abcc->abc", aux_mat3)
        #
        # add exchange part to matrix2
        #
        tmp3 = self.lf.create_three_index(self.nbasis)
        tmp3.assign(aux_mat3._array.transpose(0, 2, 1))
        aux_mat2.iadd(tmp3, factor=-1.0)
        del tmp3
        #
        # Inactive Fock matrix
        #
        aux_mat4 = self.init_aux_matrix("fock", self.nbasis, self.nbasis)
        get_fock_matrix(aux_mat4, mo1, mo2, self.nocc)
        #
        # <pp|qq>
        #
        aux_mat5 = self.init_aux_matrix("gppqq", self.nbasis, self.nbasis)
        mo2.slice("aabb->ab", aux_mat5)
        #
        # <oo||vv>+<oo|vv>
        #
        aux_mat8 = self.denself.create_four_index(
            self.nocc, self.nocc, self.nvirt, self.nvirt
        )
        mo2.slice("abcd->abcd", aux_mat8, 2.0, True, **oovv)
        mo2.slice("abcd->abdc", aux_mat8, -1.0, False, **oovv)
        #
        # <vo||ov>+<vo|ov>
        #
        aux_mat9 = self.denself.create_four_index(
            self.nvirt, self.nocc, self.nocc, self.nvirt
        )
        mo2.slice("abcd->abcd", aux_mat9, 2.0, True, **voov)
        mo2.slice("abcd->abdc", aux_mat9, -1.0, False, **vovo)
        #
        # <oo||ov>+<oo|ov>
        #
        aux_mat10 = self.denself.create_four_index(
            self.nocc, self.nocc, self.nocc, self.nvirt
        )
        mo2.slice("abcd->abcd", aux_mat10, 2.0, True, **ooov)
        mo2.slice("abcd->abdc", aux_mat10, -1.0, False, **oovo)
        # Improve with Cholesky
        # <ov||vv>+<ov|vv>
        #
        aux_mat11 = self.denself.create_four_index(
            self.nocc, self.nvirt, self.nvirt, self.nvirt
        )
        mo2.slice("abcd->abcd", aux_mat11, 2.0, True, **ovvv)
        mo2.slice("abcd->abdc", aux_mat11, -1.0, False, **ovvv)
        #
        # Aux matrix for Mia,0
        #
        aux_matia0 = self.init_aux_matrix("m10ia", self.nocc, self.nvirt)
        # fia
        aux_matia0.iadd(aux_mat4, 1.0, **ov2)
        # fia cia
        aux_matia0.iadd_mult(cia, aux_mat4, 1.0, **ov2)
        # <ia|cc> cic
        aux_mat3.contract("abc,ac->ab", cia, aux_matia0, **ovv)
        # <ia|ll> cla
        aux_mat3.contract("abc,cb->ab", cia, aux_matia0, factor=-1.0, **ovo)
        #
        # Aux matrix for Mia,ka
        #
        miaka = self.init_aux_matrix("miaka", self.nocc, self.nocc)
        # fki
        miaka.iadd(aux_mat4, -1.0, **oo2)
        # -Gikee cie
        aux_mat3.contract("abc,ac->ab", cia, miaka, factor=-1.0, **oov)
        #
        # Aux matrix for Mia,ic
        #
        miaic = self.init_aux_matrix("miaic", self.nvirt, self.nvirt)
        # fac
        miaic.iadd(aux_mat4, 1.0, **vv2)
        # -Gacmm cma
        aux_mat3.contract("abc,ca->ab", cia, miaic, factor=-1.0, **vvo)
        #
        # Aux matrix for Mia,kc
        #
        miakc = self.init_aux_matrix(
            "miakc", self.nocc, self.nvirt, self.nocc, self.nvirt
        )
        # L_akic
        aux_mat9.slice("abcd->cabd", miakc, 1.0)
        # Likac cia
        aux_mat8.contract("abcd,ac->acbd", cia, miakc)
        del aux_mat9
        #
        # Aux matrix for Mia,kald (klid)
        #
        m1kald = self.init_aux_matrix(
            "m1kald", self.nocc, self.nocc, self.nocc, self.nvirt
        )
        # L_klid
        aux_mat10.slice("abcd->abcd", m1kald, -1.0)
        del aux_mat10
        #
        # Aux matrix for Mia,icld (lacd)
        #
        x1kcda = self.init_aux_matrix(
            "x1kcda", self.nocc, self.nvirt, self.nvirt, self.nvirt
        )
        # L_ladc
        aux_mat11.slice("abcd->abdc", x1kcda, 1.0)
        del aux_mat11
        #
        # Aux matrix for Mia,iald (ld)
        #
        m1iald = self.init_aux_matrix("m1iald", self.nocc, self.nvirt)
        # F_ld
        m1iald.iadd(aux_mat4, 1.0, **ov2)
        del aux_mat8
        #
        # Aux matrix for Miaia,ka (aki)
        #
        mpiaka = self.init_aux_matrix(
            "mpiaka", self.nvirt, self.nocc, self.nocc
        )
        # G_akii
        mpiaka.iadd(aux_mat3, -2.0, **voo3)
        # F_ak cia
        aux_mat4.contract(
            "ab,ca->abc", cia, mpiaka, factor=-2.0, **vo, select="einsum"
        )
        # <ak|ee> cie
        aux_mat3.contract("abc,dc->abd", cia, mpiaka, factor=-2.0, **vov)
        # L_aiki cia
        aux_mat2.contract("abc,ba->acb", cia, mpiaka, factor=2.0, **voo)
        #
        # Aux matrix for Miaia,ic (cia)
        #
        mpiaic = self.init_aux_matrix(
            "mpiaic", self.nvirt, self.nocc, self.nvirt
        )
        # G_ciaa
        mpiaic.iadd(aux_mat3, 2.0, **vov3)
        # F_ci cia
        aux_mat4.contract(
            "ab,bc->abc", cia, mpiaic, factor=-2.0, **vo, select="einsum"
        )
        # <ci|ll> cla
        aux_mat3.contract("abc,cd->abd", cia, mpiaic, factor=2.0, **voo)
        # L_caia cia
        aux_mat2.contract("abc,cb->acb", cia, mpiaic, factor=-2.0, **vvo)
        #
        # Aux matrix for Mppiaia,kaka (iak)
        #
        mppiaka = self.init_aux_matrix(
            "mppiaka", self.nocc, self.nvirt, self.nocc
        )
        # G_iikk
        mppiaka.iadd_expand_two_to_three(
            "ac->abc", aux_mat5, 1.0, end0=self.nocc, end1=self.nocc
        )
        # <kk|aa> cia
        cia.contract(
            "ab,bc->abc",
            aux_mat5,
            mppiaka,
            factor=-2.0,
            **vo2,
            select="einsum",
        )
        # <kk|ee> cie
        tmp = self.lf.create_two_index(self.nocc)
        cia.contract("ab,bc->ac", aux_mat5, tmp, **vo2)
        mppiaka.iadd_expand_two_to_three("ac->abc", tmp, 1.0)
        #
        # Aux matrix for Mppiaia,icic (iac)
        #
        mppiaic = self.init_aux_matrix(
            "mppiaic", self.nocc, self.nvirt, self.nvirt
        )
        # G_aacc
        mppiaic.iadd_expand_two_to_three(
            "bc->abc", aux_mat5, 1.0, begin0=self.nocc, begin1=self.nocc
        )
        # <ii|cc> cia
        cia.contract(
            "ab,ac->abc",
            aux_mat5,
            mppiaic,
            factor=-2.0,
            **ov2,
            select="einsum",
        )
        # <mm|cc> cma
        tmp = self.lf.create_two_index(self.nvirt, self.nvirt)
        cia.contract("ab,ac->bc", aux_mat5, tmp, **ov2)
        mppiaic.iadd_expand_two_to_three("bc->abc", tmp, 1.0)
        #
        # Aux matrix for Mppiaia,iaia (ia)
        #
        mppia = self.init_aux_matrix("mppia", self.nocc, self.nvirt)
        # F_aa - F_ii
        fockdiag = aux_mat4.copy_diagonal()
        mppia.iadd_expand_one_to_two("a->ab", fockdiag, -2.0, end0=self.nocc)
        mppia.iadd_expand_one_to_two("b->ab", fockdiag, 2.0, begin0=self.nocc)
        # L_iaia
        mppia.iadd(aux_mat1, -2.0, **ov2)
        # <ii|aa> cia
        mppia.iadd_mult(aux_mat5, cia, 4.0, **ov)
        # <ii|dd> cid
        tmp = self.lf.create_one_index(self.nocc)
        cia.contract("ab,ab->a", aux_mat5, tmp, **ov2)
        mppia.iadd_expand_one_to_two("a->ab", tmp, -2.0)
        # <kk|aa> cka
        tmp = self.lf.create_one_index(self.nvirt)
        cia.contract("ab,ab->b", aux_mat5, tmp, **ov2)
        mppia.iadd_expand_one_to_two("b->ab", tmp, -2.0)

        gc.collect()


class REOMpCCDCCS(REOMCC):
    @property
    def dimension(self):
        """Number of unknowns"""
        return 2 * self.nocc * self.nvirt + 1

    @property
    def singles(self):
        return True

    @property
    def pairs(self):
        return True

    @property
    def doubles(self):
        return False

    def get_model(self):
        return "pCCD-CCS"

    def unmask_args(self, *args, **kwargs):
        #
        # olp
        #
        olp = unmask("olp", *args, **kwargs)
        if olp is None:
            raise ArgumentError(
                "Cannot find overlap integrals in LCC function call."
            )
        self.checkpoint.update("olp", olp)
        #
        # orb
        #
        orbs = unmask_orb(*args, **kwargs)
        if orbs:
            orbs = orbs[0]
            self.checkpoint.update("orb_a", orbs.copy())
        else:
            raise ArgumentError("Cannot find orbitals.")
        #
        # t_p
        #
        t_p = unmask("t_p", *args, **kwargs)
        if t_p is not None:
            self.checkpoint.update("t_p", t_p)
        else:
            raise ArgumentError("Cannot find Tp amplitudes")
        #
        # t_1
        #
        t_1 = unmask("t_1", *args, **kwargs)
        if t_1 is not None:
            self.checkpoint.update("t_1", t_1)
        elif self.singles:
            raise ArgumentError("Cannot find T1 amplitudes")
        #
        # 1-e ints and 2-e ints
        #
        one = self.lf.create_two_index(label="one")
        for arg in args:
            if isinstance(arg, TwoIndex):
                if arg.label in OneBodyHamiltonian:
                    one.iadd(arg)
            elif isinstance(arg, FourIndex):
                if arg.label in TwoBodyHamiltonian:
                    two = arg
        return one, two, orbs, olp

    @timer.with_section("BuildHpCCDCCS")
    def construct_ham(self, args):
        cia = self.checkpoint["t_p"]
        tia = self.checkpoint["t_1"]
        eomham = self.lf.create_two_index(
            2 * self.nvirt * self.nocc + 1, 2 * self.nocc * self.nvirt + 1
        )
        #
        # effective Hamiltonian matrix elements
        #
        fock = self.get_aux_matrix("fock")
        G_ppqq = self.get_aux_matrix("gppqq")
        L_pqrq = self.get_aux_matrix("lpqrq")
        miaka = self.get_aux_matrix("miaka")
        miaic = self.get_aux_matrix("miaic")
        miakc = self.get_aux_matrix("miakc")
        m1kald = self.get_aux_matrix("m1kald")
        x1kcda = self.get_aux_matrix("x1kcda")
        m1iald = self.get_aux_matrix("m1iald")
        mpiaka = self.get_aux_matrix("mpiaka")
        mpiaic = self.get_aux_matrix("mpiaic")
        mppiaka = self.get_aux_matrix("mppiaka")
        mppiaic = self.get_aux_matrix("mppiaic")
        mppia = self.get_aux_matrix("mppia")

        row = 1
        shift = self.nocc * self.nvirt
        for k in range(self.nocc):
            for c in range(self.nvirt):
                ci = c + self.nocc
                #
                # Single excitations:
                #
                # H_0,kc
                #
                eomham._array[0, row] = 2 * fock._array[k, ci]
                #
                # H_kc,kkcc
                #
                eomham._array[row, row + shift] += m1iald.get_element(k, c)

                #
                # Pair excitations:
                #
                #
                # Diagonal elements H_kc,kc
                #
                eomham._array[row + shift, row + shift] += mppia.get_element(
                    k, c
                )
                #
                # H_0,kc
                #
                eomham._array[0, row + shift] = G_ppqq._array[k, ci]

                #
                # H_kc,kd
                #
                col = k * self.nvirt + 1
                for d in range(self.nvirt):
                    di = d + self.nocc
                    #
                    # H_kkcc,kkdd
                    #
                    eomham._array[
                        row + shift, col + shift
                    ] += mppiaic.get_element(k, c, d)
                    #
                    # H_kc,kd (singles contribution)
                    #
                    eomham._array[row, col] += miaic.get_element(c, d)
                    #
                    # H_kc,kkdd (coupling pair-doubles-singles contribution)
                    #
                    eomham._array[row, col + shift] += x1kcda.get_element(
                        k, c, d, d
                    )
                    #
                    # H_kkdd,kc (coupling pair-doubles-singles contribution)
                    #
                    sumkkddkc = mpiaic.get_element(c, k, d)
                    eomham._array[col + shift, row] += sumkkddkc
                    #
                    col = col + 1

                #
                # H_kc,lc
                #
                col = c + 1
                for l in range(self.nocc):
                    eomham._array[
                        row + shift, col + shift
                    ] += mppiaka.get_element(k, c, l)
                    #
                    # H_kc,lc (singles contribution)
                    #
                    eomham._array[row, col] += miaka.get_element(k, l)
                    #
                    # H_kc,llcc (coupling pair-doubles-singles contribution)
                    #
                    eomham._array[row, col + shift] += m1kald.get_element(
                        l, l, k, c
                    )
                    #
                    # H_llcc,kc (coupling pair-doubles-singles contribution)
                    #
                    sumllcckc = mpiaka.get_element(c, k, l)
                    eomham._array[col + shift, row] += sumllcckc
                    #
                    # H_kc,ld (singles and coupling to pair-doubles)
                    #
                    cols = l * self.nvirt + 1
                    for d in range(self.nvirt):
                        di = d + self.nocc
                        #
                        # H_kc,ld (singles)
                        #
                        eomham._array[row, cols] += miakc.get_element(
                            k, c, l, d
                        )
                        #
                        # H_lldd,kc (coupling to pair-doubles)
                        #
                        eomham._array[cols + shift, row] += (
                            2.0
                            * L_pqrq.get_element(k, di, ci)
                            * cia._array[l, d]
                        )
                        eomham._array[cols + shift, row] -= (
                            2.0
                            * L_pqrq.get_element(k, l, ci)
                            * cia._array[l, d]
                        )

                        cols = cols + 1

                    col = col + self.nvirt
                row = row + 1
        return eomham

    @timer.with_section("BHDiagpCCDCCS")
    def compute_h_diag(self, args):
        cia = self.checkpoint["t_p"]
        tia = self.checkpoint["t_1"]
        check_type("cia", cia, DenseTwoIndex, TwoIndex)
        check_type("tia", tia, DenseTwoIndex, TwoIndex)
        lens = self.nvirt * self.nocc
        eomhamdiag = self.lf.create_one_index(self.dimension)
        #
        #
        #
        fock = self.get_aux_matrix("fock")
        miaka = self.get_aux_matrix("miaka")
        miaic = self.get_aux_matrix("miaic")
        miakc = self.get_aux_matrix("miakc")
        mppia = self.get_aux_matrix("mppia")
        mppiaic = self.get_aux_matrix("mppiaic")
        mppiaka = self.get_aux_matrix("mppiaka")

        # Singles
        diags = self.lf.create_two_index(self.nocc, self.nvirt)
        miakc.slice("abab->ab", diags)
        tmp = miaka.copy_diagonal()
        diags.iadd_expand_one_to_two("a->ab", tmp, 1.0)
        tmp = miaic.copy_diagonal()
        diags.iadd_expand_one_to_two("b->ab", tmp, 1.0)
        # Pairs:
        diagp = mppia.copy()
        mppiaic.slice("abb->ab", diagp, 1.0, False)
        mppiaka.slice("aba->ab", diagp, 1.0, False)
        eomhamdiag.assign(diags.ravel(), begin0=1, end0=(lens + 1))
        eomhamdiag.assign(diagp.ravel(), begin0=(lens + 1))
        return eomhamdiag

    @timer.with_section("SubHpCCDCCS")
    def build_subspace_hamiltonian(self, bvector, hdiag, args):
        cia = self.checkpoint["t_p"]
        tia = self.checkpoint["t_1"]
        check_type("cia", cia, DenseTwoIndex, TwoIndex)
        check_type("tia", tia, DenseTwoIndex, TwoIndex)
        #
        # Get ranges
        #
        ov2 = self.get_range("ov", start=2)
        ovv = self.get_range("ovv")
        oov = self.get_range("oov")
        #
        # Get auxiliary matrices
        #
        fock = self.get_aux_matrix("fock")
        G_ppqq = self.get_aux_matrix("gppqq")
        L_pqrq = self.get_aux_matrix("lpqrq")
        miaka = self.get_aux_matrix("miaka")
        miaic = self.get_aux_matrix("miaic")
        miakc = self.get_aux_matrix("miakc")
        m1kald = self.get_aux_matrix("m1kald")
        x1kcda = self.get_aux_matrix("x1kcda")
        m1iald = self.get_aux_matrix("m1iald")
        mpiaka = self.get_aux_matrix("mpiaka")
        mpiaic = self.get_aux_matrix("mpiaic")
        mppiaka = self.get_aux_matrix("mppiaka")
        mppiaic = self.get_aux_matrix("mppiaic")
        mppia = self.get_aux_matrix("mppia")

        #
        # Calculate sigma vector (H.bvector)_kc
        #
        lens = self.nvirt * self.nocc
        shift = lens + 1
        sigmas = self.lf.create_two_index(self.nocc, self.nvirt)
        sigmap = self.lf.create_two_index(self.nocc, self.nvirt)
        bvs = self.denself.create_two_index(self.nocc, self.nvirt)
        bvp = self.denself.create_two_index(self.nocc, self.nvirt)
        #
        # reshape bvector
        #
        bvs.assign(bvector, begin2=1, end2=shift)
        bvp.assign(bvector, begin2=shift)
        #
        #
        #
        sigma = self.lf.create_one_index(self.dimension)
        #
        # Reference vector R_0
        #
        # X0,kc rkc
        sum0_ = 0.0
        sum0_ = bvs.contract("ab,ab", fock, **ov2) * 2.0
        sum0_ += bvp.contract("ab,ab", G_ppqq, **ov2)
        sigma.set_element(0, sum0_)
        #
        # Single excitations
        #
        # Xkcld rld
        miakc.contract("abcd,cd->ab", bvs, sigmas, clear=True)
        # Xkclc rlc
        miaka.contract("ab,bc->ac", bvs, sigmas)
        # Xkckd rkd
        bvs.contract("ab,cb->ac", miaic, sigmas)
        # Xllkc rlclc
        m1kald.contract("aabc,ac->bc", bvp, sigmas)
        # Xclde rkdle
        x1kcda.contract("abcc,ac->ab", bvp, sigmas)
        # Xkc rkckc
        sigmas.iadd_mult(m1iald, bvp, 1.0)
        #
        # Pair excitations
        #
        mppiaka.contract("abc,cb->ab", bvp, sigmap, clear=True)
        # Xcd rkdkd
        mppiaic.contract("abc,ac->ab", bvp, sigmap)
        # Xkc rkckc
        sigmap.iadd_mult(mppia, bvp, 1.0)
        # Coupling pairs-singles
        mpiaka.contract("abc,ba->ca", bvs, sigmap)
        # Xdkc rkd
        mpiaic.contract("abc,ba->bc", bvs, sigmap)
        # Xldck rld
        tmp = self.lf.create_one_index(self.nvirt)
        L_pqrq.contract("abc,ac->b", bvs, tmp, **ovv)
        cia.contract("ab,b->ab", tmp, sigmap, factor=2.0)
        tmp = self.lf.create_one_index(self.nocc)
        L_pqrq.contract("abc,ac->b", bvs, tmp, **oov)
        cia.contract("ab,a->ab", tmp, sigmap, factor=-2.0)

        sigma.assign(sigmas.ravel(), begin0=1, end0=shift)
        sigma.assign(sigmap.ravel(), begin0=shift)

        return sigma

    @timer.with_section("AuxMatEOMpCCDCCS")
    def update_aux_matrix(self, mo1, mo2):
        """Derive all auxiliary matrices.
           fock_pq:     one_pq + sum_m(2<pm|qm> - <pm|mq>),
           lpqpq:   2<pq|pq>-<pq|qp>,
           gpqpq:   <pq|pq>,
           lpqrq:   2<pq|rq>-<pq|qr>,
           gpqrr:   <pq|rr>,

           **Arguments:**

           mo1, mo2
                one- and two-electron integrals to be sorted.

        """
        cia = self.checkpoint["t_p"]
        tia = self.checkpoint["t_1"]
        check_type("cia", cia, DenseTwoIndex, TwoIndex)
        check_type("tia", tia, DenseTwoIndex, TwoIndex)
        #
        # Get ranges
        #
        ov = self.get_range("ov")
        vo = self.get_range("vo")
        ov2 = self.get_range("ov", start=2)
        vo2 = self.get_range("vo", start=2)
        oo2 = self.get_range("oo", start=2)
        vv2 = self.get_range("vv", start=2)
        ovo = self.get_range("ovo")
        oov = self.get_range("oov")
        voo = self.get_range("voo")
        vov = self.get_range("vov")
        vvo = self.get_range("vvo")
        voo3 = self.get_range("voo", start=3)
        vov3 = self.get_range("vov", start=3)
        ooov = self.get_range("ooov")
        oovo = self.get_range("oovo")
        oovv = self.get_range("oovv")
        voov = self.get_range("voov")
        vovo = self.get_range("vovo")
        ovvv = self.get_range("ovvv")
        #
        # 2<pq|pq>-<pq|pq> and <pq|pq>
        #
        # 2<pq|pq>-<pq|pq>
        #
        aux_mat1 = self.init_aux_matrix("lpqpq", self.nbasis, self.nbasis)
        mo2.slice("abab->ab", aux_mat1, 2.0, True)
        mo2.slice("abba->ab", aux_mat1, -1.0, False)
        #
        # <pq||rq>+<pq|rq>
        #
        aux_mat2 = self.init_aux_matrix(
            "lpqrq", self.nbasis, self.nbasis, self.nbasis
        )
        mo2.slice("abcb->abc", aux_mat2, 2.0, True)
        #
        # <pq|rr>
        #
        aux_mat3 = self.init_aux_matrix(
            "gpqrr", self.nbasis, self.nbasis, self.nbasis
        )
        mo2.slice("abcc->abc", aux_mat3)
        #
        # add exchange part to matrix2
        #
        tmp3 = self.lf.create_three_index(self.nbasis)
        tmp3.assign(aux_mat3._array.transpose(0, 2, 1))
        aux_mat2.iadd(tmp3, factor=-1.0)
        del tmp3
        #
        # Inactive Fock matrix
        #
        aux_mat4 = self.init_aux_matrix("fock", self.nbasis, self.nbasis)
        get_fock_matrix(aux_mat4, mo1, mo2, self.nocc)
        #
        # <pp|qq>
        #
        aux_mat5 = self.init_aux_matrix("gppqq", self.nbasis, self.nbasis)
        mo2.slice("aabb->ab", aux_mat5)
        #
        # <oo||vv>+<oo|vv>
        #
        aux_mat8 = self.denself.create_four_index(
            self.nocc, self.nocc, self.nvirt, self.nvirt
        )
        mo2.slice("abcd->abcd", aux_mat8, 2.0, True, **oovv)
        mo2.slice("abcd->abdc", aux_mat8, -1.0, False, **oovv)
        #
        # <vo||ov>+<vo|ov>
        #
        aux_mat9 = self.denself.create_four_index(
            self.nvirt, self.nocc, self.nocc, self.nvirt
        )
        mo2.slice("abcd->abcd", aux_mat9, 2.0, True, **voov)
        mo2.slice("abcd->abdc", aux_mat9, -1.0, False, **vovo)
        #
        # <oo||ov>+<oo|ov>
        #
        aux_mat10 = self.denself.create_four_index(
            self.nocc, self.nocc, self.nocc, self.nvirt
        )
        mo2.slice("abcd->abcd", aux_mat10, 2.0, True, **ooov)
        mo2.slice("abcd->abdc", aux_mat10, -1.0, False, **oovo)
        # Improve with Cholesky
        # <ov||vv>+<ov|vv>
        #
        aux_mat11 = self.denself.create_four_index(
            self.nocc, self.nvirt, self.nvirt, self.nvirt
        )
        mo2.slice("abcd->abcd", aux_mat11, 2.0, True, **ovvv)
        mo2.slice("abcd->abdc", aux_mat11, -1.0, False, **ovvv)
        #
        # Aux matrix for Mia,ka
        #
        miaka = self.init_aux_matrix("miaka", self.nocc, self.nocc)
        # fki
        miaka.iadd(aux_mat4, -1.0, **oo2)
        # tie fke
        tia.contract("ab,cb->ac", aux_mat4, miaka, **ov2)
        # -L_kmie t_me
        aux_mat10.contract("abcd,bd->ca", tia, miaka, factor=-1.0)
        # -Gikee cie
        aux_mat3.contract("abc,ac->ab", cia, miaka, factor=-1.0, **oov)
        #
        # Aux matrix for Mia,ic
        #
        miaic = self.init_aux_matrix("miaic", self.nvirt, self.nvirt)
        # fac
        miaic.iadd(aux_mat4, 1.0, **vv2)
        # tma fmc
        tia.contract("ab,ac->bc", aux_mat4, miaic, factor=-1.0, **ov2)
        # L_maec t_me
        aux_mat11.contract("abcd,ac->bd", tia, miaic)
        # -Gacmm cma
        aux_mat3.contract(
            "abc,ca->ab", cia, miaic, factor=-1.0, **vvo, select="einsum"
        )
        #
        # Aux matrix for Mia,kc
        #
        miakc = self.init_aux_matrix(
            "miakc", self.nocc, self.nvirt, self.nocc, self.nvirt,
        )
        # L_akic
        aux_mat9.slice("abcd->cabd", miakc, 1.0)
        # L_mkic t_ma
        aux_mat10.contract("abcd,ae->cebd", tia, miakc, factor=-1.0)
        # L_kace t_ie
        aux_mat11.contract("abcd,ed->ebac", tia, miakc)
        # Likac cia
        aux_mat8.contract("abcd,ac->acbd", cia, miakc)
        del aux_mat9
        #
        # Aux matrix for Mia,kald (klid)
        #
        m1kald = self.init_aux_matrix(
            "m1kald", self.nocc, self.nocc, self.nocc, self.nvirt,
        )
        # L_klid
        aux_mat10.slice("abcd->abcd", m1kald, -1.0)
        # L_kled tie
        aux_mat8.contract("abcd,ec->abed", tia, m1kald, factor=-1.0)
        del aux_mat10
        #
        # Aux matrix for Mia,icld (lacd)
        #
        x1kcda = self.init_aux_matrix(
            "x1kcda", self.nocc, self.nvirt, self.nvirt, self.nvirt
        )
        # L_ladc
        aux_mat11.slice("abcd->abdc", x1kcda, 1.0)
        # L_mlcd tma
        aux_mat8.contract("abcd,ae->becd", tia, x1kcda, factor=-1.0)
        del aux_mat11
        #
        # Aux matrix for Mia,iald (ld)
        #
        m1iald = self.init_aux_matrix("m1iald", self.nocc, self.nvirt)
        # F_ld
        m1iald.iadd(aux_mat4, 1.0, **ov2)
        # L_mled tme
        aux_mat8.contract("abcd,ac->bd", tia, m1iald)
        del aux_mat8
        #
        # Aux matrix for Miaia,ka (aki)
        #
        mpiaka = self.init_aux_matrix(
            "mpiaka", self.nvirt, self.nocc, self.nocc,
        )
        # G_akii
        mpiaka.iadd(aux_mat3, -2.0, **voo3)
        # F_ak cia
        aux_mat4.contract(
            "ab,ca->abc", cia, mpiaka, factor=-2.0, **vo, select="einsum"
        )
        # <ak|ee> cie
        aux_mat3.contract("abc,dc->abd", cia, mpiaka, factor=-2.0, **vov)
        # L_aiki cia
        aux_mat2.contract("abc,ba->acb", cia, mpiaka, factor=2.0, **voo)
        #
        # Aux matrix for Miaia,ic (cia)
        #
        mpiaic = self.init_aux_matrix(
            "mpiaic", self.nvirt, self.nocc, self.nvirt,
        )
        # G_ciaa
        mpiaic.iadd(aux_mat3, 2.0, **vov3)
        # F_ci cia
        aux_mat4.contract(
            "ab,bc->abc", cia, mpiaic, factor=-2.0, **vo, select="einsum"
        )
        # <ci|ll> cla
        aux_mat3.contract("abc,cd->abd", cia, mpiaic, factor=2.0, **voo)
        # L_caia cia
        aux_mat2.contract("abc,cb->acb", cia, mpiaic, factor=-2.0, **vvo)
        #
        # Aux matrix for Mppiaia,kaka (iak)
        #
        mppiaka = self.init_aux_matrix(
            "mppiaka", self.nocc, self.nvirt, self.nocc,
        )
        # G_iikk
        mppiaka.iadd_expand_two_to_three(
            "ac->abc", aux_mat5, 1.0, end0=self.nocc, end1=self.nocc
        )
        # <kk|aa> cia
        cia.contract(
            "ab,bc->abc",
            aux_mat5,
            mppiaka,
            factor=-2.0,
            **vo2,
            select="einsum",
        )
        # <kk|ee> cie
        tmp = self.lf.create_two_index(self.nocc)
        cia.contract("ab,bc->ac", aux_mat5, tmp, **vo2)
        mppiaka.iadd_expand_two_to_three("ac->abc", tmp, 1.0)
        #
        # Aux matrix for Mppiaia,icic (iac)
        #
        mppiaic = self.init_aux_matrix(
            "mppiaic", self.nocc, self.nvirt, self.nvirt,
        )
        # G_aacc
        mppiaic.iadd_expand_two_to_three(
            "bc->abc", aux_mat5, 1.0, begin0=self.nocc, begin1=self.nocc
        )
        # <ii|cc> cia
        cia.contract(
            "ab,ac->abc",
            aux_mat5,
            mppiaic,
            factor=-2.0,
            **ov2,
            select="einsum",
        )
        # <mm|cc> cma
        tmp = self.lf.create_two_index(self.nvirt)
        cia.contract("ab,ac->bc", aux_mat5, tmp, **ov2)
        mppiaic.iadd_expand_two_to_three("bc->abc", tmp, 1.0)
        #
        # Aux matrix for Mppiaia,iaia (ia)
        #
        mppia = self.init_aux_matrix("mppia", self.nocc, self.nvirt)
        # F_aa - F_ii
        fockdiag = aux_mat4.copy_diagonal()
        mppia.iadd_expand_one_to_two("a->ab", fockdiag, -2.0, end0=self.nocc)
        mppia.iadd_expand_one_to_two("b->ab", fockdiag, 2.0, begin0=self.nocc)
        # L_iaia
        mppia.iadd(aux_mat1, -2.0, **ov2)
        # <ii|aa> cia
        mppia.iadd_mult(aux_mat5, cia, 4.0, **ov)
        # <ii|dd> cid
        tmp = self.lf.create_one_index(self.nocc)
        cia.contract("ab,ab->a", aux_mat5, tmp, **ov2)
        mppia.iadd_expand_one_to_two("a->ab", tmp, -2.0)
        # <kk|aa> cka
        tmp = self.lf.create_one_index(self.nvirt)
        cia.contract("ab,ab->b", aux_mat5, tmp, **ov2)
        mppia.iadd_expand_one_to_two("b->ab", tmp, -2.0)

        gc.collect()
