# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file

import numpy as np

from pybest.linalg.dense import DenseLinalgFactory
from pybest.scf import AufbauOccModel
from pybest.io.molpro import load_fcidump
from pybest.context import context
from pybest.wrappers import RHF
from pybest.linalg.cholesky import CholeskyLinalgFactory
from pybest.gbasis.gobasis import (
    get_gobasis,
    compute_cholesky_eri,
    compute_kinetic,
    compute_nuclear,
    compute_nuclear_repulsion,
    compute_overlap,
)
from pybest.io.iodata import IOData


def prepare_mol(filen, nocc, nbasis, orb=None):
    fn = context.get_fn(filen)
    ###########################################################################
    ## Define Occupation model, expansion coefficients and overlap ############
    ###########################################################################
    lf = DenseLinalgFactory(nbasis)
    occ_model = AufbauOccModel(nocc)
    orb_a = lf.create_orbital(nbasis)
    olp = lf.create_two_index(nbasis, label="olp")
    olp.assign_diagonal(1.0)
    orb_a.assign(olp)
    if orb is not None:
        fn_orb = context.get_fn(orb)
        orb_a.coeffs[:] = np.fromfile(fn_orb, sep=",").reshape(nbasis, nbasis)

    ###########################################################################
    ## Construct Hamiltonian ##################################################
    ###########################################################################
    integrals = load_fcidump(fn, lf)
    one = integrals["one"]
    er = integrals["two"]
    core = integrals["e_core"]

    iodata = IOData(
        **{
            "orb_a": orb_a,
            "olp": olp,
            "e_core": core,
            "e_ref": 0.0,
        }
    )
    return lf, occ_model, one, er, iodata


def prepare_chol(basisfile, molfile, nocc):
    fn_basis = context.get_fn(basisfile)
    fn_mol = context.get_fn(molfile)

    obasis = get_gobasis(fn_basis, fn_mol)
    lf = CholeskyLinalgFactory(obasis.nbasis)

    occ_model = AufbauOccModel(nocc)

    orb_a = lf.create_orbital(obasis.nbasis)
    olp = compute_overlap(obasis)

    kin = compute_kinetic(obasis)
    ne = compute_nuclear(obasis)
    eri = compute_cholesky_eri(obasis, threshold=1e-12)
    external = compute_nuclear_repulsion(obasis)

    ###############################################################################
    ## Do a Hartree-Fock calculation ##############################################
    ###############################################################################
    hf = RHF(lf, occ_model)
    hf_ = hf(kin, ne, eri, external, olp, orb_a)

    return lf, occ_model, kin, ne, eri, hf_
