# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file


import numpy as np
import pytest

from pybest.ee_eom.eom_pccd import REOMpCCD
from pybest.geminals.rpccd import RpCCD
from pybest.ee_eom.test.common import prepare_mol


# Reference energies
# pCCD orbitals
epccd_opt_ref = -37.991225550170
e_ex_pccd_opt_ref = [
    0.000000e00,
    6.889018e-01,
    7.930113e-01,
    1.355189e00,
    1.373127e00,
    1.833921e00,
    2.074985e00,
    2.224260e00,
    2.259740e00,
    2.277095e00,
    2.310193e00,
    2.343555e00,
    2.350504e00,
    2.384404e00,
    2.427066e00,
    2.496954e00,
]
e_ex_pccd_opt_ref_fc = [
    0.000000e00,
    6.889361e-01,
    7.930974e-01,
    1.355166e+00,
]


def test_pccd_opt_exactdiag():
    lf, occ_model, one, er, iodata = prepare_mol(
        "test/chplus.fcidump", 3, 25, "test/chplus_opt.dat"
    )

    # pCCD
    geminal_solver = RpCCD(lf, occ_model)
    pccd = geminal_solver(one, er, iodata)
    energy = pccd.e_tot
    assert abs(energy - epccd_opt_ref) < 1e-6

    # EOM-pCCD
    eom = REOMpCCD(lf, occ_model)
    eom_ = eom(one, er, pccd, nroot=16, davidson=False, todisk=False)

    for i in range(len(e_ex_pccd_opt_ref)):
        assert abs(eom_.e_ee[i] - e_ex_pccd_opt_ref[i]) < 1e-6


def test_pccd_opt_davidson():
    lf, occ_model, one, er, iodata = prepare_mol(
        "test/chplus.fcidump", 3, 25, "test/chplus_opt.dat"
    )

    # pCCD
    geminal_solver = RpCCD(lf, occ_model)
    pccd = geminal_solver(one, er, iodata)
    energy = pccd.e_tot
    assert abs(energy - epccd_opt_ref) < 1e-6

    # EOM-pCCD
    eom = REOMpCCD(lf, occ_model)
    eom_ = eom(one, er, pccd, nroot=16, davidson=True, todisk=False)

    for i in range(len(e_ex_pccd_opt_ref)):
        assert abs(eom_.e_ee[i] - e_ex_pccd_opt_ref[i]) < 1e-6


def test_pccd_opt_davidson_ncore():
    lf, occ_model, one, er, iodata = prepare_mol(
        "test/chplus.fcidump", 3, 25, "test/chplus_opt.dat"
    )

    # pCCD
    geminal_solver = RpCCD(lf, occ_model, ncore=1)
    pccd = geminal_solver(one, er, iodata)

    # EOM-pCCD
    eom = REOMpCCD(lf, occ_model, ncore=1)
    eom_ = eom(one, er, pccd, nroot=4, davidson=True, todisk=False)

    for i in range(len(e_ex_pccd_opt_ref_fc)):
        assert abs(eom_.e_ee[i] - e_ex_pccd_opt_ref_fc[i]) < 1e-6
