# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file


import numpy as np
import pytest

from pybest.ee_eom.eom_pccd_s import REOMpCCDS, REOMpCCDCCS
from pybest.geminals.rpccd import RpCCD
from pybest.cc.rccs import RpCCDCCS
from pybest.ee_eom.test.common import prepare_mol


# Reference energies
# pCCD orbitals
epccd_opt_ref = -37.991225550170
epccdccs_opt_ref = -37.99124837
epccdccs1_opt_ref = -0.00002282
e_ex_pccds_opt_ref = [
    -1.175150e-08,
    1.566229e-01,
    1.566229e-01,
    5.345106e-01,
    6.010138e-01,
    6.010138e-01,
    6.602393e-01,
    6.773517e-01,
    6.802024e-01,
    6.876033e-01,
    6.876033e-01,
    7.651787e-01,
    7.820253e-01,
    7.902141e-01,
    7.902141e-01,
    8.333694e-01,
]
e_ex_pccd_ccs_opt_ref = [
    0.000000e00,
    1.567204e-01,
    1.567204e-01,
    5.343764e-01,
    6.009420e-01,
    6.009420e-01,
    6.602406e-01,
    6.770740e-01,
    6.799853e-01,
    6.873988e-01,
    6.873988e-01,
    7.649222e-01,
    7.819493e-01,
    7.900298e-01,
    7.900298e-01,
    8.332708e-01,
]
e_ex_pccds_opt_ref_fc = [
    -8.059321e-07,
    1.568994e-01,
    1.568995e-01,
    6.076699e-01,
]


def test_pccds_opt_davidson():
    lf, occ_model, one, er, iodata = prepare_mol(
        "test/chplus.fcidump", 3, 25, "test/chplus_opt.dat"
    )

    # pCCD
    geminal_solver = RpCCD(lf, occ_model)
    pccd = geminal_solver(one, er, iodata)
    energy = pccd.e_tot
    assert abs(energy - epccd_opt_ref) < 1e-6

    # EOM-pCCD+S
    eom = REOMpCCDS(lf, occ_model)
    eom_ = eom(one, er, pccd, nroot=16, davidson=True, todisk=False)

    for i in range(len(e_ex_pccds_opt_ref)):
        assert abs(eom_.e_ee[i] - e_ex_pccds_opt_ref[i]) < 1e-6


def test_pccds_opt_exactdiag():
    lf, occ_model, one, er, iodata = prepare_mol(
        "test/chplus.fcidump", 3, 25, "test/chplus_opt.dat"
    )

    # pCCD
    geminal_solver = RpCCD(lf, occ_model)
    pccd = geminal_solver(one, er, iodata)
    energy = pccd.e_tot
    assert abs(energy - epccd_opt_ref) < 1e-6

    # EOM-pCCD+S
    eom = REOMpCCDS(lf, occ_model)
    eom_ = eom(one, er, pccd, nroot=16, davidson=False, todisk=False)

    for i in range(len(e_ex_pccds_opt_ref)):
        assert abs(eom_.e_ee[i] - e_ex_pccds_opt_ref[i]) < 1e-6


def test_pccds_opt_davidson_fc():
    lf, occ_model, one, er, iodata = prepare_mol(
        "test/chplus.fcidump", 3, 25, "test/chplus_opt.dat"
    )

    # pCCD
    geminal_solver = RpCCD(lf, occ_model, ncore=1)
    pccd = geminal_solver(one, er, iodata)

    # EOM-pCCD+S
    eom = REOMpCCDS(lf, occ_model, ncore=1)
    eom_ = eom(one, er, pccd, nroot=4, davidson=True, todisk=False)

    for i in range(len(e_ex_pccds_opt_ref_fc)):
        assert abs(eom_.e_ee[i] - e_ex_pccds_opt_ref_fc[i]) < 1e-6


def test_pccd_ccs_opt_davidson():
    lf, occ_model, one, er, iodata = prepare_mol(
        "test/chplus.fcidump", 3, 25, "test/chplus_opt.dat"
    )

    # pCCD
    geminal_solver = RpCCD(lf, occ_model)
    pccd = geminal_solver(one, er, iodata)
    energy = pccd.e_tot
    assert abs(energy - epccd_opt_ref) < 1e-6

    # pCCD-CCS
    ccs = RpCCDCCS(lf, occ_model)
    ccs_ = ccs(one, er, pccd, tthreshold=1e-10)

    assert abs(ccs_.e_tot - epccdccs_opt_ref) < 1e-6
    assert abs(ccs_.e_corr - epccdccs1_opt_ref) < 1e-6

    # EOM-pCCD-CCS
    eom = REOMpCCDCCS(lf, occ_model)
    eom_ = eom(one, er, ccs_, nroot=16, davidson=True, todisk=False)

    for i in range(len(e_ex_pccd_ccs_opt_ref)):
        assert abs(eom_.e_ee[i] - e_ex_pccd_ccs_opt_ref[i]) < 1e-6


def test_pccd_ccs_opt_exactdiag():
    lf, occ_model, one, er, iodata = prepare_mol(
        "test/chplus.fcidump", 3, 25, "test/chplus_opt.dat"
    )

    # pCCD
    geminal_solver = RpCCD(lf, occ_model)
    pccd = geminal_solver(one, er, iodata)
    energy = pccd.e_tot
    assert abs(energy - epccd_opt_ref) < 1e-6

    # pCCD-CCS
    ccs = RpCCDCCS(lf, occ_model)
    ccs_ = ccs(one, er, pccd, tthreshold=1e-10)

    assert abs(ccs_.e_tot - epccdccs_opt_ref) < 1e-6
    assert abs(ccs_.e_corr - epccdccs1_opt_ref) < 1e-6
    assert abs(ccs_.e_corr_s - epccdccs1_opt_ref) < 1e-6

    # EOM-pCCD-CCS
    eom = REOMpCCDCCS(lf, occ_model)
    eom_ = eom(one, er, ccs_, nroot=16, davidson=False, todisk=False,)

    for i in range(len(e_ex_pccd_ccs_opt_ref)):
        assert abs(eom_.e_ee[i] - e_ex_pccd_ccs_opt_ref[i]) < 1e-6

    # EOM-pCCD-CCS
    eom = REOMpCCDCCS(lf, occ_model)
    eom_ = eom(one, er, ccs_, nroot=16, davidson=False, todisk=False)

    for i in range(len(e_ex_pccd_ccs_opt_ref)):
        assert abs(eom_.e_ee[i] - e_ex_pccd_ccs_opt_ref[i]) < 1e-6
