# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Dumping files to disk

   The dumping manager decides to which directories PyBEST results are written.
   We can choose between two directories
   
    :result_dir: Contains all final results of a PyBEST calculation, for instance,
                 checkpoint files, dat files, etc. These directory contains all
                 checkpoint files that can be used for restarts.
    :temp_dir:   Contains all intermediate files generated during code execution, for
                 instance, Hamiltonian integrals, eigenvectors, etc.
                 It does not contain restartable data, that is, checkpoint files.
"""

from pathlib import Path
import shutil
import atexit

from pybest.exceptions import ArgumentError, DirectoryError


__all__ = ["filemanager"]


class FileManager(object):

    def __init__(self, result_dir, temp_dir):
        self._temp_list = []
        self.result_dir = result_dir
        self.temp_dir = temp_dir
        self.keep_temp = False

    @property
    def result_dir(self):
        return self._result_dir

    @result_dir.setter
    def result_dir(self, new_dir):
        if not isinstance(new_dir, str):
            raise ArgumentError(f"Do not know how to handle argument {new_dir}")
        if new_dir in [""]:
            raise ArgumentError("Directory name cannot be empty.")
        self._result_dir = Path(f"{new_dir}/")
        self.create_dir(self._result_dir)

    @property
    def temp_dir(self):
        return self._temp_dir

    @temp_dir.setter
    def temp_dir(self, new_dir):
        if not isinstance(new_dir, str):
            raise ArgumentError(f"Do not know how to handle argument {new_dir}")
        if new_dir in [""]:
            raise ArgumentError(f"Directory name cannot be empty.")
        self._temp_dir = Path(f"{new_dir}/")
        self.create_dir(self._temp_dir)
        self._temp_list.append(self._temp_dir)

    @property
    def keep_temp(self):
        return self._keep_temp
    
    @keep_temp.setter
    def keep_temp(self, new):
        if not type(new) == bool:
            raise TypeError("keep_temp can only be boolean type")
        self._keep_temp = new

    def create_dir(self, pathlib_dir):
        pathlib_dir.mkdir(parents=True, exist_ok=True)
        if not pathlib_dir.exists() and not pathlib_dir.is_dir():
            raise DirectoryError(f"Cannot create directory {str(pathlib_dir)}!")
    
    def result_path(self, fname):
        # be sure that result_dir exists:
        self.create_dir(self.result_dir)
        return self.result_dir.joinpath(fname)
    
    def temp_path(self, fname):
        # be sure that temp_dir exists:
        self.create_dir(self.temp_dir)
        return self.temp_dir.joinpath(fname)

    def clean_up_temp(self):
        if self.keep_temp:
            return
        for temp_dir in self._temp_list:
            if temp_dir.exists():
                shutil.rmtree(temp_dir)


# Maybe call it fm?
filemanager = FileManager("pybest-results", "pybest-temp")
# If keep_temp is set to False, delete all temp dirs after exit
atexit.register(filemanager.clean_up_temp)
