// PyBEST: Pythonic Black-box Electronic Structure Tool
// Copyright (C) 2016-- The PyBEST Development Team
//
// This file is part of PyBEST.
//
// PyBEST is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// PyBEST is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>
//
// --

//#define DEBUG

#include <cmath>
#include <cstdlib>
#include <stdexcept>
#include <stdlib.h>
#include <cstring>
#include <string>
#include <iostream>
#include <iomanip>
#include "libint_utils.h"
#include "basis.h"
#include <Eigen/Core>
#include <highfive/H5Easy.hpp>

using std::abs;

/*
    Basis

*/


// Overload constructor:
Basis::Basis(const std::string &basisname, const std::string &molfile, const std::string &dirname) :
    basisname(basisname), molfile(molfile), dirname(dirname)
{
    libint2::Shell::do_enforce_unit_normalization(false);
    libint2::initialize();  // safe to use libint now

    // read coordinates and initialize atoms
    std::string xyzfilename = std::string(molfile);
    std::ifstream input_file(xyzfilename);
    atoms = libint2::read_dotxyz(input_file);

    // Change LIBINT_DATA_PATH to read in basis set from correct directory
    // dirname. This change is only locally and does not affect the global
    // enviromnent variable.
    std::string pybestdata = std::string(dirname);
    std::string mypath_ = "LIBINT_DATA_PATH="+pybestdata;
    putenv( strdup(mypath_.c_str()) );

    // Do not print stdout of libint
    // get underlying buffer
    std::streambuf* orig_buf = std::cout.rdbuf();
    std::cout.rdbuf(NULL);
    obs = libint2::BasisSet (basisname, atoms);
    // restore buffer
    std::cout.rdbuf(orig_buf);

    nbasis = obs.nbf();
    nshell = obs.size();
    ncenter = atoms.size();
    shell2atom = obs.shell2atom(obs, atoms);
    for(size_t s=0; s!=obs.size(); ++s) {
        for(const auto& c: obs[s].contr) {
            auto factor = (!c.pure && c.l>1) ? -1 : 1;
            shell_types.push_back(c.l*factor);
            for (auto n = 0ul; n < obs[s].alpha.size(); ++n) {
                // use normalization factor to get back to original coefficient
                double normalization_factor = get_renorm(c.l, s, n);
                contraction.push_back(c.coeff[n]/normalization_factor);
                alpha.push_back(obs[s].alpha[n]);
            }
        }
        nprim.push_back(obs[s].nprim());
    }
    for(auto a=0ul; a<atoms.size(); ++a) {
        atomic_numbers.push_back(atoms[a].atomic_number);
        std::vector<double> xyz;
        xyz.push_back(atoms[a].x);
        xyz.push_back(atoms[a].y);
        xyz.push_back(atoms[a].z);
        coordinates.push_back(xyz);
    }
    libint2::finalize();  // do not use libint after this
}

Basis::Basis(const std::string &molfile, std::vector<long> nprims, std::vector<long> center, std::vector<int> l, std::vector<double> alphas, std::vector<double> contr) :
    basisname(""), molfile(molfile), dirname("")
{
    // Do some standard test:
    if (nprims.size() != center.size())
        throw std::domain_error("Number of primitives and centers does not match.");
    if (nprims.size() != l.size())
        throw std::domain_error("Number of primitives and shell types does not match.");
    if ( std::any_of(nprims.begin(), nprims.end(), [](size_t i){ return i == 0; }) )
        throw std::domain_error("Number of primitives has to be larger than zero.");
    if (alphas.size() != contr.size())
        throw std::domain_error("Number of exponents and contraction coefficients does not match.");
    size_t nprim_tot = 0;
    for (auto& n : nprims) nprim_tot += n;
    if (alphas.size() != nprim_tot)
        throw std::domain_error("Number of exponents and primitives does not match.");
    // Start assigning basis set
    libint2::Shell::do_enforce_unit_normalization(false);
    libint2::initialize();  // safe to use libint now

    // read coordinates and initialize atoms
    std::string xyzfilename = std::string(molfile);
    std::ifstream input_file(xyzfilename);
    atoms = libint2::read_dotxyz(input_file);

    std::vector<std::vector<libint2::Shell>> element_bases(118);
    size_t n = 0;
    for(size_t a=0ul; a<atoms.size(); ++a) {
        auto Z = atoms[a].atomic_number;
        auto x = atoms[a].x;
        auto y = atoms[a].y;
        auto z = atoms[a].z;

        std::vector<libint2::Shell> shells;
        for(size_t a_=0ul; a_<center.size(); ++a_) {
            if (atoms.size() <= size_t(center[a_]))
                throw std::domain_error("More centers than atoms specified.");
            if (a!=size_t(center[a_])) continue;

            libint2::svector<double> contr_;
            libint2::svector<double> alpha_;
            for(auto s=0; s!=nprims[a_]; ++s) {
                contr_.emplace_back(contr[n]);
                alpha_.emplace_back(alphas[n]);
                n++;
            }
            auto pure = (l[a_]>1) ? true : false;
            shells.push_back(
                {
                  std::move(alpha_), // exponents of primitive Gaussians
                  {
                    {abs(l[a_]), pure, std::move(contr_)}
                  },
                  {{x, y, z}}   // origin coordinates
                }
            );
        }
        element_bases[Z] = shells;
    }
    obs = libint2::BasisSet(atoms, element_bases);
    // set to pure to invoke init function of libint basis set
    // check if all l>1 are either cartesian or pure. Currently, we cannot
    // handle mixed basis sets. It is possible with libint, but not supported
    // here.
    bool cartesian = false;
    bool solid = false;
    int l_max = 0;
    for(auto& l_: l) {
        if (l_ < 0) cartesian = true;
        if (l_ > 1) solid = true;
        l_max = (l_max < abs(l_)) ? abs(l_) : l_max;
    }
    if (l_max <= 1) cartesian = true;
    if (solid && cartesian)
        throw std::domain_error("Cannot handle mixed cartesian and solid harminics");
    if (cartesian) obs.set_pure(false);
    else if (solid) obs.set_pure(true);
    else throw std::domain_error("Don't know whether to use cartesians or solid harmonics.");
    pure2cart(0);
    pure2cart(1);
    nbasis = obs.nbf();
    nshell = obs.size();
    ncenter = atoms.size();
    shell2atom = obs.shell2atom(obs, atoms);
    for(size_t s=0; s!=obs.size(); ++s) {
        for(const auto& c: obs[s].contr) {
            auto factor = (!c.pure && c.l>1) ? -1 : 1;
            shell_types.push_back(c.l*factor);
            for (auto n = 0ul; n < obs[s].alpha.size(); ++n) {
                // use normalization factor to get back to original coefficient
                double normalization_factor = get_renorm(c.l, s, n);
                contraction.push_back(c.coeff[n]/normalization_factor);
                alpha.push_back(obs[s].alpha[n]);
            }
        }
        nprim.push_back(obs[s].nprim());
    }
    for(auto a=0ul; a<atoms.size(); ++a) {
        atomic_numbers.push_back(atoms[a].atomic_number);
        std::vector<double> xyz;
        xyz.push_back(atoms[a].x);
        xyz.push_back(atoms[a].y);
        xyz.push_back(atoms[a].z);
        coordinates.push_back(xyz);
    }
    libint2::finalize();  // do not use libint after this
}

Basis::Basis(std::vector<libint2::Atom> atoms_, std::vector<long> nprims, std::vector<long> center, std::vector<int> l, std::vector<double> alphas, std::vector<double> contr) :
    basisname(""), molfile(""), dirname("")
{
    // Do some standard test:
    if (nprims.size() != center.size())
        throw std::domain_error("Number of primitives and centers does not match.");
    if (nprims.size() != l.size())
        throw std::domain_error("Number of primitives and shell types does not match.");
    if ( std::any_of(nprims.begin(), nprims.end(), [](size_t i){ return i == 0; }) )
        throw std::domain_error("Number of primitives has to be larger than zero.");
    if (alphas.size() != contr.size())
        throw std::domain_error("Number of exponents and contraction coefficients does not match.");
    size_t nprim_tot = 0;
    for (auto& n : nprims) nprim_tot += n;
    if (alphas.size() != nprim_tot)
        throw std::domain_error("Number of exponents and primitives does not match.");
    // Start assigning basis set
    libint2::Shell::do_enforce_unit_normalization(false);
    libint2::initialize();  // safe to use libint now

    atoms = atoms_;

    std::vector<std::vector<libint2::Shell>> element_bases(118);
    size_t n = 0;
    for(size_t a=0ul; a<atoms.size(); ++a) {
        auto Z = atoms[a].atomic_number;
        auto x = atoms[a].x;
        auto y = atoms[a].y;
        auto z = atoms[a].z;

        std::vector<libint2::Shell> shells;
        for(size_t a_=0ul; a_<center.size(); ++a_) {
            if (atoms.size() <= size_t(center[a_]))
                throw std::domain_error("More centers than atoms specified.");
            if (a!=size_t(center[a_])) continue;

            libint2::svector<double> contr_;
            libint2::svector<double> alpha_;
            for(auto s=0; s!=nprims[a_]; ++s) {
                contr_.emplace_back(contr[n]);
                alpha_.emplace_back(alphas[n]);
                n++;
            }
            auto pure = (l[a_]>1) ? true : false;
            shells.push_back(
                {
                  std::move(alpha_), // exponents of primitive Gaussians
                  {
                    {abs(l[a_]), pure, std::move(contr_)}
                  },
                  {{x, y, z}}   // origin coordinates
                }
            );
        }
        element_bases[Z] = shells;
    }
    obs = libint2::BasisSet(atoms, element_bases);
    // set to pure to invoke init function of libint basis set
    // check if all l>1 are either cartesian or pure. Currently, we cannot
    // handle mixed basis sets. It is possible with libint, but not supported
    // here.
    bool cartesian = false;
    bool solid = false;
    int l_max = 0;
    for(auto& l_: l) {
        if (l_ < 0) cartesian = true;
        if (l_ > 1) solid = true;
        l_max = (l_max < abs(l_)) ? abs(l_) : l_max;
    }
    if (l_max <= 1) cartesian = true;
    if (solid && cartesian)
        throw std::domain_error("Cannot handle mixed cartesian and solid harminics");
    if (cartesian) obs.set_pure(false);
    else if (solid) obs.set_pure(true);
    else throw std::domain_error("Don't know whether to use cartesians or solid harmonics.");
    pure2cart(0);
    pure2cart(1);
    nbasis = obs.nbf();
    nshell = obs.size();
    ncenter = atoms.size();
    shell2atom = obs.shell2atom(obs, atoms);
    for(size_t s=0; s!=obs.size(); ++s) {
        for(const auto& c: obs[s].contr) {
            auto factor = (!c.pure && c.l>1) ? -1 : 1;
            shell_types.push_back(c.l*factor);
            for (auto n = 0ul; n < obs[s].alpha.size(); ++n) {
                // use normalization factor to get back to original coefficient
                double normalization_factor = get_renorm(c.l, s, n);
                contraction.push_back(c.coeff[n]/normalization_factor);
                alpha.push_back(obs[s].alpha[n]);
            }
        }
        nprim.push_back(obs[s].nprim());
    }
    for(auto a=0ul; a<atoms.size(); ++a) {
        atomic_numbers.push_back(atoms[a].atomic_number);
        std::vector<double> xyz;
        xyz.push_back(atoms[a].x);
        xyz.push_back(atoms[a].y);
        xyz.push_back(atoms[a].z);
        coordinates.push_back(xyz);
    }
    libint2::finalize();  // do not use libint after this
}

/*
    Defintion of getters:
*/
const std::string &Basis::getBasisname() const {
    return basisname;
}

const std::string &Basis::getMolfile() const {
    return molfile;
}

const std::string &Basis::getDirname() const {
    return dirname;
}

size_t Basis::getNBasis() {
    return nbasis;
}

size_t Basis::getNShell() {
    return nshell;
}

size_t Basis::getNCenter() {
    return ncenter;
}

std::vector<long> Basis::getNPrim() {
    return nprim;
}

std::vector<double> Basis::getAlpha() {
    return alpha;
}

std::vector<double> Basis::getContraction() {
    return contraction;
}

std::vector<long> Basis::getShell2atom() {
    return shell2atom;
}

std::vector<int> Basis::getShellTypes() {
    return shell_types;
}

std::vector<long> Basis::getAtomicNumbers() {
    return atomic_numbers;
}

std::vector< std::vector<double> > Basis::getCoordinates() {
    return coordinates;
}

/*
    Defintion of general functions:
*/
void Basis::pure2cart(int l) {
    for(auto& shell: obs)
        for(auto& contraction: shell.contr)
            if (contraction.l == l) contraction.pure = false;
}

void Basis::print_basis_info() {
    libint2::Shell::do_enforce_unit_normalization(false);
    libint2::initialize();  // safe to use libint now

    std::cout << "  Reading basis set information from file: " << basisname << std::endl;
    std::cout << "  Number of contracted basis functions:    " << nbasis << std::endl;
    std::cout << "  Maximum orbital angular momentum:        " << obs.max_l() << std::endl;

    for(auto a=0ul; a<atoms.size(); ++a) {
        auto Z = atoms[a].atomic_number;
        auto x = atoms[a].x;
        auto y = atoms[a].y;
        auto z = atoms[a].z;

        std::cout << " Index:  " << a << '\t'
                  << " Charge: " << Z << '\t'
                  << " Center: " << std::fixed << std::setprecision(9)
                  << x << '\t' << y << '\t' << z << std::endl;
        for(size_t s=0; s!=obs.size(); ++s) {
            auto x_ = obs[s].O[0];
            auto y_ = obs[s].O[1];
            auto z_ = obs[s].O[2];
            if (x!=x_ || y!=y_ || z!=z_) continue;
            std::cout << " " << obs[s].contr[0].l << std::endl;
            for(const auto& c: obs[s].contr) {
                for (auto n = 0ul; n < obs[s].alpha.size(); ++n) {
                    // use normalization factor to get back to original coefficient
                    double normalization_factor = get_renorm(c.l, s, n);
                    std::cout << " " << std::setw(18) << std::right << obs[s].alpha[n] << '\t'
                              << std::setw(16) << std::right << c.coeff[n]/normalization_factor
                              << std::endl;
                }
            }
        }
    }

    libint2::finalize();  // do not use libint after this
}

void Basis::dump_molden(const char* filename, const py::EigenDRef<Eigen::MatrixXd> coeffs, const py::EigenDRef<Eigen::MatrixXd> occs, const py::EigenDRef<Eigen::MatrixXd> energies, const std::vector<bool>& spin, double eps) {
    libint2::Shell::do_enforce_unit_normalization(false);
    libint2::initialize();  // safe to use libint now

    libint2::molden::Export molden(atoms, obs, coeffs, occs, energies, std::vector<std::string>(), spin, eps);
    molden.write(std::string(filename));

    libint2::finalize();  // do not use libint after this
}


void Basis::print_atomic_info() {
    std::cout << "Index \t Charge    \t   x [au]   \t   y [au]   \t   z [au]" << std::endl;
    for(auto a=0ul; a<atoms.size(); ++a) {
        auto Z = atoms[a].atomic_number;
        auto x = atoms[a].x;
        auto y = atoms[a].y;
        auto z = atoms[a].z;

        std::cout << a << "  \t  "
                  << Z << "\t \t"
                  << std::right << std::fixed << std::setw(12) << std::setprecision(8) << x << '\t'
                  << std::right << std::fixed << std::setw(12) << y << '\t'
                  << std::right << std::fixed << std::setw(12) << z << std::endl;
    }
}

size_t Basis::get_shell_size(int shell_type) {
    return (shell_type >= 0) ? (2*shell_type+1) : ((abs(shell_type)+1) * (abs(shell_type)+2) / 2);
}

size_t Basis::get_nbasis_in_shell(size_t s) {
    return obs[s].size();
}

void Basis::set_dummy_atom(int i) {
    atoms[i].atomic_number = 0.0;
}

double Basis::get_renorm(size_t l, size_t s, size_t n) {
    // The following lines are taken from libint's renorm function
    // Note, this only works if libint2::Shell::do_enforce_unit_normalization(false);
    // is set.
    using libint2::math::df_Kminus1;
    using std::pow;
    const auto sqrt_Pi_cubed = double{5.56832799683170784528481798212};
    const auto two_alpha = 2 * obs[s].alpha[n];
    const auto two_alpha_to_am32 = pow(two_alpha,l+1) * sqrt(two_alpha);
    const auto normalization_factor = sqrt(pow(2,l) * two_alpha_to_am32/(sqrt_Pi_cubed * df_Kminus1[2*l] ));
    return normalization_factor;
}

double Basis::get_renorm(size_t l, size_t s, size_t na, std::vector<size_t> n) {
    // The following lines are taken from libint's renorm function
    // Note, this only works if libint2::Shell::do_enforce_unit_normalization(false);
    // is set.
    using libint2::math::df_Kminus1;
    using std::pow;
    const auto sqrt_Pi_cubed = double{5.56832799683170784528481798212};
    const auto two_alpha = 2 * obs[s].alpha[na];
    const auto two_alpha_to_am32 = two_alpha * sqrt(two_alpha);
    const auto normalization_factor = sqrt(pow(2*two_alpha,n[0]+n[1]+n[2]) * two_alpha_to_am32/(sqrt_Pi_cubed * df_Kminus1[2*n[0]] * df_Kminus1[2*n[1]] * df_Kminus1[2*n[2]]));
    return normalization_factor;
}

void Basis::renormalize_contr(int l, std::vector<size_t> n, double factor) {
    libint2::Shell::do_enforce_unit_normalization(false);
    libint2::initialize();  // safe to use libint now
    // loop over all basis functions
    for(size_t s=0; s!=obs.size(); ++s) {
        for(auto& c: obs[s].contr) {
            if (c.l != l) continue;
            for (auto na = 0ul; na < obs[s].alpha.size(); ++na) {
                // use normalization factor to get back to original coefficient
                double normalization_factor = get_renorm(c.l, s, na, n)*factor;
                c.coeff[na] /= normalization_factor;
            }
        }
    }
    libint2::finalize();  // do not use libint after this
}

void Basis::to_hdf5(const std::string &filename) {
    H5Easy::File file(filename, H5Easy::File::ReadWrite);

    H5Easy::dump(file, "/gobasis/basisname", basisname);
    H5Easy::dump(file, "/gobasis/molfile", molfile);
    H5Easy::dump(file, "/gobasis/dirname", dirname);
    H5Easy::dump(file, "/gobasis/nbasis", nbasis);
    H5Easy::dump(file, "/gobasis/nshell", nshell);
    H5Easy::dump(file, "/gobasis/ncenter", ncenter);
    H5Easy::dump(file, "/gobasis/shell_types", shell_types);
    H5Easy::dump(file, "/gobasis/shell2atom", shell2atom);
    H5Easy::dump(file, "/gobasis/nprim", nprim);
    H5Easy::dump(file, "/gobasis/alpha", alpha);
    H5Easy::dump(file, "/gobasis/contraction", contraction);
    H5Easy::dump(file, "/gobasis/atomic_numbers", atomic_numbers);
    H5Easy::dump(file, "/gobasis/coordinates", coordinates);

}

Basis from_hdf5(const std::string &filename) {
    H5Easy::File file(filename, H5Easy::File::ReadOnly);

    const std::string basisname = H5Easy::load<std::string>(file, "/gobasis/basisname");
    const std::string molfile = H5Easy::load<std::string>(file, "/gobasis/molfile");
    const std::string dirname = H5Easy::load<std::string>(file, "/gobasis/dirname");
//  size_t nbasis = H5Easy::load<size_t>(file, "/gobasis/nbasis");
//  size_t nshell = H5Easy::load<size_t>(file, "/gobasis/nshell");
//  size_t ncenter = H5Easy::load<size_t>(file, "/gobasis/ncenter");
    std::vector<int> shell_types = H5Easy::load<std::vector<int>>(file, "/gobasis/shell_types");
    std::vector<long> shell2atom = H5Easy::load<std::vector<long>>(file, "/gobasis/shell2atom");
    std::vector<long> nprim = H5Easy::load<std::vector<long>>(file, "/gobasis/nprim");
    std::vector<double> alpha = H5Easy::load<std::vector<double>>(file, "/gobasis/alpha");
    std::vector<double> contraction = H5Easy::load<std::vector<double>>(file, "/gobasis/contraction");
    std::vector<long> atomic_numbers = H5Easy::load<std::vector<long>>(file, "/gobasis/atomic_numbers");
    std::vector< std::vector<double> > coordinates = H5Easy::load<std::vector< std::vector<double> >>(file, "/gobasis/coordinates");

    libint2::Shell::do_enforce_unit_normalization(false);
    libint2::initialize();  // safe to use libint now
    std::vector<libint2::Atom> atoms;
    // assign to libint2::Atoms
    for(auto a=0ul; a<atomic_numbers.size(); ++a) {
        libint2::Atom atom;
        atom.atomic_number = atomic_numbers[a];
        atom.x = coordinates[a][0];
        atom.y = coordinates[a][1];
        atom.z = coordinates[a][2];

        atoms.push_back(atom);
    }
    libint2::finalize();  // do not use libint after this

    return Basis(atoms, nprim, shell2atom, shell_types, alpha, contraction);
}
