# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --

# Detailed changelog:
# This module has been originaly written and updated by Katharina Boguslawski (see CHANGELOG)
# Its current version contains updates from the PyBEST developer team.
#
# This implementation can also be found in `Horton 2.0.0`.
# However, this file has been updated and debugged. Compatibility with Horton is NOT
# guaranteed.
# 
# Detailed changes:
# See CHANGELOG 

"""Correlated wavefunction implementations

   This is a general geminals class.
"""

import numpy as np
from itertools import permutations
from operator import mul
from functools import reduce

from pybest.filemanager import filemanager
from pybest.log import log
from pybest.cache import Cache
from pybest.featuredlists import OneBodyHamiltonian, TwoBodyHamiltonian
from pybest.exceptions import ArgumentError, UnknownOption
from pybest.utils import check_type, check_options, unmask_orb, unmask
from pybest.io.checkpoint import CheckPoint
from pybest.linalg.dense import DenseLinalgFactory
from pybest.linalg.base import TwoIndex, FourIndex, Orbital
from pybest.pt.perturbation_utils import get_epsilon
from pybest.helperclass import PropertyHelper


class Geminal(object):
    """A collection of geminals and optimization routines.

       This is just a base class that serves as a template for
       specific implementations.
    """

    def __init__(
        self,
        lf,
        occ_model,
        npairs=None,
        nvirt=None,
        ncore=None,
        freezerot=None,
    ):
        """
           **Arguments:**

           lf
                A LinalgFactory instance.

           occ_model
                Occupation model

           **Optional arguments:**

           npairs
                Number of electron pairs, if not specified,
                npairs = number of occupied orbitals

           nvirt
                Number of virtual orbitals, if not specified,
                nvirt = (nbasis-npairs)

           ncore
                Number of frozen core orbitals, if not specified,
                ncore = 0
        """
        log.cite("the pCCD method", "boguslawski2014a", "boguslawski2014b")
        check_type("pairs", npairs, int, type(None))
        check_type("virtuals", nvirt, int, type(None))
        check_type("ncore", ncore, int, type(None))
        self._lf = lf
        self._denself = DenseLinalgFactory(lf.default_nbasis)
        if ncore is None:
            ncore = 0
        if npairs is None:
            npairs = occ_model.noccs[0] - ncore
        self._nbasis = lf.default_nbasis - ncore
        if npairs <= 0:
            raise ValueError(
                f"Number of frozen core orbitals ({ncore}) larger than number of "
                f"occupied orbitals ({occ_model.noccs[0]})"
            )
        if npairs >= (lf.default_nbasis - ncore):
            raise ValueError(
                f"Number of electron pairs ({npairs}) larger than number of basis "
                f"functions ({self._nbasis})"
            )
        if nvirt is None:
            nvirt = lf.default_nbasis - npairs - ncore
        elif nvirt >= (lf.default_nbasis - ncore):
            raise ValueError(
                f"Number of virtuals ({nvirt}) larger than number of basis functions "
                f"({self._nbasis})"
            )
        self._npairs = npairs
        self._nocc = occ_model.noccs[0] - ncore
        assert self.nocc == self.npairs
        self._nvirt = nvirt
        self._ncore = ncore
        self._cache = Cache()
        self._e_core = 0
        self._geminal = lf.create_two_index(npairs, nvirt, label="t_p")
        self._lagrange = lf.create_two_index(npairs, nvirt, label="l_p")
        # TODO: This is a quick hack and should be cleaned up in the future
        self._freezerot = freezerot
        self._norbiter = 0
        if freezerot is not None:
            for el in freezerot:
                check_options("freezerot", el, "vv", "ov", "oo")
        self._energy = {}
        self._checkpoint = CheckPoint({})


    def __call__(self, *args, **kwargs):
        """Optimize geminal coefficients and---if required---find
           optimal set of orbitals.

           **Arguments:**

           one, two
                One- and two-body integrals (some Hamiltonian matrix elements).

           core
                The core energy (not included in 'one' and 'two').

           orb
                An expansion instance. It contains the MO coefficients
                (orbitals).

           olp
                The AO overlap matrix. A TwoIndex instance.

           **Keywords:**
                See :py:meth:`RpCCD.solve`
                and :py:meth:`ROOpCCD.solve`
        """
        one = self.lf.create_two_index(label="one")
        # olp
        olp = unmask("olp", *args, **kwargs)
        # orb; we have to use unmask_orb here
        orbs = unmask_orb(*args, **kwargs)
        if not orbs and kwargs.get("restart", False):
            orbs = [None]
        elif not orbs:
            raise ArgumentError("Cannot find orbitals in function call.")
        # 1-e ints and 2-e ints
        one = self.lf.create_two_index(label="one")
        for arg in args:
            if isinstance(arg, TwoIndex):
                if arg.label in OneBodyHamiltonian:
                    one.iadd(arg)
            elif isinstance(arg, FourIndex):
                two = arg

        # Solve function only works for restricted orbitals
        # FIXME: passing some args twice. Clean-up later (too invasive for now).
        return self.solve(one, two, olp, *orbs, *args, **kwargs)

    def solve(self, one, two, core, orb, olp, **kwargs):
        raise NotImplementedError

    def solve_scf(self, one, two, core, orb, olp, **kwargs):
        raise NotImplementedError

    @property
    def nbasis(self):
        """The number of basis functions"""
        return self._nbasis

    @property
    def nocc(self):
        """The number of occupied orbitals"""
        return self._nocc

    @property
    def nvirt(self):
        """The number of virtual orbitals"""
        return self._nvirt

    @property
    def ncore(self):
        """The number of virtual orbitals"""
        return self._ncore

    # TODO: This is a quick hack and should be cleaned up in the future
    @property
    def freezerot(self):
        """Freeze orbital rotations"""
        return self._freezerot

    @property
    def npairs(self):
        """The number of electron pairs"""
        return self._npairs

    @property
    def denself(self):
        """The LinalgFactory instance"""
        return self._denself

    @property
    def lf(self):
        """The LinalgFactory instance"""
        return self._lf

    @property
    def e_core(self):
        """The core energy"""
        return self._e_core

    @e_core.setter
    def e_core(self, new):
        """Update core energy"""
        self._e_core = new

    @property
    def dimension(self):
        """The number of unknowns (i.e. the number of geminal coefficients)"""
        return self._npairs * self._nvirt

    @property
    def geminal(self):
        """The geminal coefficients"""
        return self._geminal

    @geminal.setter
    def geminal(self, new):
        """Update geminal matrix"""
        self._geminal.assign(new)

    @property
    def lagrange(self):
        """The Lagrange multipliers"""
        return self._lagrange

    @lagrange.setter
    def lagrange(self, new):
        """Update Lagragne multipliers"""
        self._lagrange.assign(new)

    @property
    def norbiter(self):
        """The number of orbital optimization steps"""
        return self._norbiter

    @norbiter.setter
    def norbiter(self, new):
        """Update number of orbital rotation steps"""
        self._norbiter = new

    @property
    def checkpoint(self):
        """The iodata container that contains all data dump to disk"""
        return self._checkpoint

    @property
    def energy(self):
        """The energy dictionary (e_tot, e_corr, e_ref)"""
        return self._energy

    @energy.setter
    def energy(self, energy_dict):
        """Update energies in dictionary"""
        self._energy.update(energy_dict)

    @energy.deleter
    def energy(self):
        """Clear energy information"""
        self._energy = {}

    def __clear__(self):
        self.clear()

    def clear(self):
        """Clear all wavefunction information"""
        self._cache.clear()

    def clear_dm(self):
        """Clear RDM information"""
        self._cache.clear(tags="d", dealloc=True)

    def update_aux_matrix(self, select, two_mo, one_mo=None):
        """Update auxiliary matrices"""
        raise NotImplementedError

    def get_aux_matrix(self, select):
        """Get auxiliary matrices"""
        raise NotImplementedError

    def compute_1dm(self, dmout, *args):
        raise NotImplementedError

    def compute_2dm(self, dmout, *args):
        raise NotImplementedError

    def init_ndm(self, select):
        """Initialize an n-RDM object

           **Arguments**

           select
                'ps2' or 'response'.
        """
        check_options(
            "ndm",
            select,
            "one_dm_ps2",
            "one_dm_response",
            "two_dm_ppqq",
            "two_dm_pqpq",
            "two_dm_rppqq",
            "two_dm_rpqpq",
        )
        alloc_size = {
            "one_dm_ps2": (self.lf.create_one_index, self.nbasis),
            "one_dm_response": (self.lf.create_one_index, self.nbasis),
            "two_dm_ppqq": (self.lf.create_two_index, self.nbasis),
            "two_dm_ppqq": (self.lf.create_two_index, self.nbasis),
            "two_dm_rppqq": (self.lf.create_two_index, self.nbasis),
            "two_dm_rpqpq": (self.lf.create_two_index, self.nbasis),
        }
        alloc_method = {
            "one_dm": self.compute_1dm,
            "two_dm": self.compute_2dm,
        }
        for key in alloc_method:
            if key in select:
                method = alloc_method[key]
        dm, new = self._cache.load(
            select, alloc=alloc_size[select], tags="d",
        )
        if not new:
            # FIXME: Substitute by PyBEST excpetion?
            raise RuntimeError(
                f"The matrix {select} already exists. Call clear prior updating DMs."
            )

        return dm, method

    def get_ndm(self, select):
        """Get a density matrix (n-RDM). If not available, it will be created
           (if possible)

           **Arguments:**

           select
                'ps2', or 'response'.
        """
        if not select in self._cache:
            self.update_ndm(select)
        return self._cache.load(select)

    one_dm_ps2 = PropertyHelper(get_ndm, "one_dm_ps2", "Alpha 1-RDM")
    one_dm_response = PropertyHelper(get_ndm, "one_dm_response", "Alpha 1-RDM")
    two_dm_ppqq = PropertyHelper(
        get_ndm, "two_dm_ppqq", "Alpha-beta PS2 (ppqq) 2-RDM"
    )
    two_dm_pqpq = PropertyHelper(
        get_ndm, "two_dm_pqpq", "Alpha-beta PS2 (pqpq) 2-RDM"
    )
    two_dm_rppqq = PropertyHelper(
        get_ndm, "two_dm_rppqq", "Alpha-beta (ppqq) 2-RDM"
    )
    two_dm_rpqpq = PropertyHelper(
        get_ndm, "two_dm_rpqpq", "Alpha-beta (pqpq) 2-RDM"
    )

    def update_ndm(self, select, ndm=None):
        """Update n-RDM

           **Optional arguments:**

           ndm
                When provided, this n-RDM is stored.
        """
        raise NotImplementedError

    # Initial guess generators:
    def generate_guess(self, guess, dim=None, mo1=None, mo2=None):
        """Generate a guess of type 'guess'.

           **Arguments:**

           guess
               A dictionary, containing the type of guess.

           **Optional arguments:**

           dim
               Length of guess.
        """
        check_options("guess.type", guess["type"], "random", "const", "mp2")
        check_type("guess.factor", guess["factor"], int, float)
        if guess["factor"] == 0:
            raise ValueError("Scaling factor must be different from 0.")
        if dim is None:
            dim = self.dimension
        if guess["type"] == "random":
            return np.random.random(dim) * guess["factor"]
        elif guess["type"] == "const":
            return np.ones(dim) * guess["factor"]
        elif guess["type"] == "mp2":
            self.clear_aux_matrix()
            self.update_aux_matrix("guess", mo2, mo1)
            iiaa = self.get_aux_matrix("gppqq")
            fock = self.get_aux_matrix("fock")
            fi = fock.copy(0, self.nocc)
            fa = fock.copy(self.nocc, self.nbasis)
            eps = get_epsilon(
                self.denself, [fi, fa], singles=True, doubles=False
            )
            guess = iiaa.divide(eps, end0=self.nocc, begin1=self.nocc)
            return guess._array.ravel()

    def compute_rotation_matrix(self, coeff):
        """Compute orbital rotation matrix"""
        raise NotImplementedError

    # Check convergence:
    def check_convergence(self, e0, e1, gradient, thresh):
        """Check convergence.

           **Arguements:**

           e0, e1
                Used to calculate energy difference e0-e1

           gradient
                The gradient, a OneIndex instance

           thresh
                Dictionary containing threshold parameters ('energy', 'gradientmax',
                'gradientnorm')

           **Returns:**
                True if energy difference, norm of orbital gradient, largest
                element of orbital gradient are smaller than some threshold
                values.
        """
        return (
            abs(e0 - e1) < thresh["energy"]
            and gradient.get_max() < thresh["gradientmax"]
            and gradient.norm() < thresh["gradientnorm"]
        )

    def check_stepsearch(self, linesearch):
        """Check trustradius. Abort calculation if trustradius is smaller than
           1e-8
        """
        return (
            linesearch.method == "trust-region"
            and linesearch.trustradius < 1e-8
        )

    def prod(self, lst):
        return reduce(mul, lst)

    def perm(self, a):
        """Calculate the permament of a matrix

           **Arguements**

           a
                A np array
        """
        check_type("matrix", a, np.ndarray)
        n = len(a)
        r = range(n)
        s = permutations(r)
        import math  # FIXME: fsum really needed for accuracy?

        return math.fsum(self.prod(a[i][sigma[i]] for i in r) for sigma in s)

    def get_range(self, string, start=0):
        """ Returns dictionary with keys beginX, endX, begin(X+1), etc.
            *  string - any sequence of 'o' (occupied) and 'v' (virtual)
        """
        range_ = {}
        ind = start
        for char in string:
            if char == "o":
                range_[f"begin{ind}"] = 0
                range_[f"end{ind}"] = self.npairs
            elif char == "v":
                range_[f"begin{ind}"] = self.npairs
                range_[f"end{ind}"] = self.nbasis
            elif char == "V":
                range_[f"begin{ind}"] = 0
                range_[f"end{ind}"] = self.nvirt
            elif char == "n":
                range_[f"begin{ind}"] = 0
                range_[f"end{ind}"] = self.nbasis
            else:
                raise UnknownOption(
                    f"Do not know how to handle choice {char}."
                )
            ind += 1
        return range_
