# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --

# Detailed changelog:
# This module has been originally written and updated by Katharina Boguslawski (see CHANGELOG)
# Its current version contains updates from the PyBEST developer team.
#
# This implementation can also be found in `Horton 2.0.0`.
# However, this file has been updated and debugged. Compatibility with Horton is NOT
# guaranteed.
# 
# Detailed changes:
# See CHANGELOG 

"""Correlated wavefunction implementations

   This module contains restricted pCCD

   Variables used in this module:
    :nocc:      number of (active) occupied orbitals in the principle configuration
    :pairs:     number of electron pairs
    :nvirt:     number of (active) virtual orbitals in the principle configuration
    :nbasis:    total number of basis functions

    Indexing convention:
     :i,j,k,..: occupied orbitals of principle configuration
     :a,b,c,..: virtual orbitals of principle configuration
     :p,q,r,..: general indices (occupied, virtual)

    Abbreviations used (if not mentioned in doc-strings):
     :L_pqrs: 2<pq|rs>-<pq|sr>
     :g_pqrs: <pq|rs>

    For more information see doc-strings.
"""

import numpy as np
from scipy import optimize as opt
from scipy import linalg as linalg
import warnings

from pybest.filemanager import filemanager
from pybest.log import log, timer
from pybest.orbital_utils import (
    compute_unitary_matrix,
    transform_integrals,
    split_core_active,
    project_orbitals,
)
from pybest.geminals.geminal import Geminal
from pybest.steplength.stepsearch import RStepSearch
from pybest.utils import check_type, check_options, unmask
from pybest.linalg.cholesky import CholeskyFourIndex
from pybest.linalg.base import TwoIndex, Orbital
from pybest.exceptions import (
    ArgumentError,
    UnknownOption,
    NoConvergence,
    ConsistencyError,
)

from copy import copy


__all__ = [
    "RpCCD",
    "ROOpCCD",
]


class RpCCDBase(Geminal):
    """Restricted pCCD base class"""

    def clear(self):
        """Clear all wavefunction information"""
        self._cache.clear()

    def clear_aux_matrix(self):
        """Clear auxiliary matrices"""
        self._cache.clear(tags="m", dealloc=True)

    #
    # Density matrices:
    #
    @timer.with_section("nDM-pCCD")
    def update_ndm(self, select, ndm=None):
        """Update n-RDM

           **Arguments:**

           select
                One of ``ps2``, ``response``

           **Optional arguments:**

           ndm
                When provided, this n-RDM is stored.
        """
        cached_ndm, method = self.init_ndm(select)
        if ndm is None:
            if select in ["one_dm_response", "two_dm_rppqq", "two_dm_rpqpq"]:
                method(cached_ndm, self.geminal, self.lagrange, select, 1.0)
            elif select in ["one_dm_ps2", "two_dm_ppqq", "two_dm_pqpq"]:
                method(cached_ndm, self.geminal, self.geminal, select, 1.0)
            else:
                raise UnknownOption(
                    f"Do not know what to do with option {select}"
                )
        else:
            cached_ndm.assign(ndm)

    @timer.with_section("1DM-pCCD")
    def compute_1dm(self, dmout, mat1, mat2, select, factor=1.0):
        """Compute 1-RDM for pCCD

           **Arguments:**

           dmout
                The output DM

           mat1, mat2
                A DenseTwoIndex instance used to calculated 1-RDM. For response
                RDM, mat1 is the geminal matrix, mat2 the Lagrange multipliers

           select
                Switch between response and PS2 1-RDM

           **Optional arguments:**

           factor
                A scalar factor
        """
        summand = 1.0
        if select in ["one_dm_ps2"]:
            summand = 1 + mat1.contract("ab,ab", mat2, select="einsum")

        #
        # Calculate occupied block
        #
        tmpocc = self.lf.create_one_index(self.npairs)
        tmpocc.assign(summand)
        mat1.contract("ab,ab->a", mat2, tmpocc, factor=-1.0, select="einsum")

        #
        # Calculate virtual block
        #
        tmpvir = self.lf.create_one_index(self.nvirt)
        mat1.contract("ab,ab->b", mat2, tmpvir, select="einsum")

        #
        # Combine both blocks and scale
        #
        dmout.assign(tmpocc, 0, self.npairs)
        dmout.assign(tmpvir, self.npairs, self.nbasis)
        dmout.iscale(factor)

    @timer.with_section("2DM-pCCD")
    def compute_2dm(self, dmout, mat1, mat2, select, factor=1.0):
        """Compute 2-RDM for pCCD

           ** Arguments **

           
               A 1DM. A OneIndex instance.

           mat1, mat2
               TwoIndex instances used to calculate the 2-RDM. To get the
               response DM, mat1 is the geminal coefficient matrix, mat2 are the
               Lagrange multipliers

           select
               Either 'two_dm_(r)ppqq' or 'two_dm_(r)pqpq'. Note that the elements
               (iiii), i.e., the 1DM, are stored in pqpq, while the elements (aaaa) are
               stored in ppqq.

           **Optional arguments:**

           factor
                A scalar factor
        """
        lc = mat1.contract("ab,ab", mat2, select="einsum")
        factor1 = 1.0

        if select in ["two_dm_rppqq", "two_dm_rpqpq"]:
            factor1 = factor1 - lc
            dm1 = self.one_dm_response
        else:
            dm1 = self.one_dm_ps2

        if "ppqq" in select:
            #
            # temporary storage
            #
            tmpvv = self.lf.create_two_index(self.nvirt, self.nvirt)
            tmpoo = self.lf.create_two_index(self.npairs, self.npairs)
            tmpvv2 = self.lf.create_two_index(self.nvirt, self.nvirt)
            tmpov = self.lf.create_two_index(self.npairs, self.nvirt)
            tmpo = self.lf.create_one_index(self.npairs)
            tmpv = self.lf.create_one_index(self.nvirt)

            #
            # o-o block
            #
            mat2.contract("ab,cb->ca", mat1, tmpoo, select="einsum")
            tmpoo.assign_diagonal(0.0)
            dmout.iadd(tmpoo, 1.0, 0, self.npairs, 0, self.npairs)
            #
            # v-v block
            #
            mat2.contract("ab,ac->bc", mat1, tmpvv, select="td")
            dmout.iadd(
                tmpvv, 1.0, self.npairs, self.nbasis, self.npairs, self.nbasis
            )
            #
            # v-o block
            #
            dmout.iadd_t(mat2, 1.0, self.npairs, self.nbasis, 0, self.npairs)
            #
            # o-v block
            #
            tmpvv2.iadd_tdot(mat2, mat1)
            tmpov.iadd_dot(mat1, tmpvv2)
            mat2.contract("ab,ab->a", mat1, tmpo, select="einsum")
            mat1.contract("ab,a->ab", tmpo, tmpov, factor=-2.0)
            mat2.contract("ab,ab->b", mat1, tmpv, select="einsum")
            mat1.contract("ab,b->ab", tmpv, tmpov, factor=-2.0)
            mat2_ = mat2.copy()
            mat2_.imul(mat1)
            mat2_.imul(mat1)
            tmpov.iadd(mat2_, 2.0)

            dmout.iadd(
                mat1, (factor1 + lc), 0, self.npairs, self.npairs, self.nbasis
            )
            dmout.iadd(tmpov, 1.0, 0, self.npairs, self.npairs, self.nbasis)
        elif "pqpq" in select:
            #
            # temporary storage
            #
            tmpo = self.lf.create_one_index(self.npairs)
            mat2.contract("ab,ab->a", mat1, tmpo, select="einsum")
            dm1v = dm1.copy(self.npairs, self.nbasis)
            mat2_ = mat2.copy()
            mat2_.imul(mat1)
            for i in range(self.npairs):
                for j in range(i + 1, self.npairs):
                    value = (
                        factor1
                        + lc
                        - tmpo.get_element(i)
                        - tmpo.get_element(j)
                    )
                    dmout.set_element(i, j, value, symmetry=2)
                value = factor1 + lc - tmpo.get_element(i)
                dmout.set_element(i, i, value)
            dmout.iadd_t(dm1v, 1.0, 0, self.npairs, self.npairs, self.nbasis)
            dmout.iadd(mat2_, -1.0, 0, self.npairs, self.npairs, self.nbasis)
            dmout.iadd(dm1v, 1.0, self.npairs, self.nbasis, 0, self.npairs)
            dmout.iadd_t(mat2_, -1.0, self.npairs, self.nbasis, 0, self.npairs)
        # scale
        dmout.iscale(factor)

    def solve_model(self, one, two, orb, **kwargs):
        """Solve for geminal model.

           **Arguments:**

           one, two
                One- and two-body integrals (some Hamiltonian matrix elements).
                A TwoIndex and FourIndex/Cholesky instance

           orb
                An expansion instance which contains the MO coefficients.

           **Keywords:**
                guess: initial guess for wfn (1-dim np.array)
                guesslm: initial guess Lagrange multipliers (1-dim np.array)
                solver: wfn/Lagrange solver (dictionary)
                indextrans: 4-index Transformation (str).
                maxiter: maximum number of iterations (dictionary)
                thresh: thresholds (dictionary)
                orbitaloptimizer: orbital optimization method (str)

                For more details, see :py:meth:`RpCCD.solve_scf`
        """
        guess = kwargs.get("guess", None)
        guesslm = kwargs.get("guesslm", None)
        solver = kwargs.get("solver", None)
        indextrans = kwargs.get("indextrans", "tensordot")
        maxiter = kwargs.get("maxiter", None)
        thresh = kwargs.get("thresh", None)
        orbitaloptimizer = kwargs.get("orbitaloptimizer", "variational")

        #
        # Transform integrals into MO basis
        #
        if self.ncore > 0:
            cas = split_core_active(
                one,
                two,
                orb,
                e_core=self.e_core,
                ncore=self.ncore,
                nactive=self.nbasis,
                indextrans=indextrans,
            )
            one_mo = cas.one
            two_mo = cas.two
            e_core = cas.e_core
        else:
            ti = transform_integrals(one, two, orb, indextrans=indextrans)
            one_mo = ti.one[0]
            two_mo = ti.two[0]
        #
        # Generate auxiliary matrices needed for optimization
        #
        self.clear_aux_matrix()
        self.update_aux_matrix("scf", two_mo, one_mo)
        del one_mo, two_mo

        #
        # Optimize pCCD wavefunction amplitudes:
        #
        coeff = self.solve_geminal(
            guess, solver, thresh["wfn"], maxiter["wfniter"]
        )
        self.geminal = coeff
        #
        # Update IOData container
        #
        self.checkpoint.update("t_p", self.geminal)

        #
        # Optimize pCCD Lagrange multipliers (lambda equations):
        #
        if orbitaloptimizer == "variational":
            lcoeff = self.solve_lagrange(
                guesslm, solver, thresh["wfn"], maxiter["wfniter"]
            )
            self.lagrange = lcoeff
            #
            # Update IOData container
            #
            self.checkpoint.update("l_p", self.lagrange)

            #
            # Update occupation numbers (response 1-RDM)
            #
            self.clear_dm()
            self.update_ndm("one_dm_response")
            self.update_ndm("two_dm_rpqpq")
            self.update_ndm("two_dm_rppqq")
            orb.assign_occupations(self.one_dm_response, self.ncore)
            orb.clear_energies()
            #
            # Update IOData container
            # Copy final orbitals
            self.checkpoint.update("orb_a", orb.copy())
            self.checkpoint.update("dm_1", self.one_dm_response)
            self.checkpoint.update(
                "dm_2", {"pqpq": self.two_dm_rpqpq, "ppqq": self.two_dm_rppqq}
            )

        elif orbitaloptimizer == "ps2":
            self.clear_dm()
            self.update_ndm("one_dm_ps2")
            self.update_ndm("two_dm_pqpq")
            self.update_ndm("two_dm_ppqq")
            #
            # Update IOData container
            #
            self.checkpoint.update("orb_a", orb.copy())
            self.checkpoint.update("dm_1", self.one_dm_ps2)
            self.checkpoint.update(
                "dm_2", {"pqpq": self.two_dm_pqpq, "ppqq": self.two_dm_ppqq}
            )

    @timer.with_section("PSEqRpCCD")
    def solve_geminal(self, guess, solver, wfnthreshold, wfnmaxiter):
        """Solves for geminal matrix

           **Arguments:**

           guess
                The initial guess. A 1-dim np array.

           solver
                The solver used. A dictionary with element 'wfn'.

           wfnthreshold
                The optimization threshold. A float.

           wfnmaxiter
                The maximum number of iterations. An integer.
        """
        iiaa = self.get_aux_matrix("gppqq")
        iaia = self.get_aux_matrix("lpqpq")
        one = self.get_aux_matrix("t")
        fock = self.get_aux_matrix("fock")
        sol = opt.root(
            self.vector_function_geminal,
            guess,
            args=(iiaa, iaia, one, fock),
            method=solver["wfn"],
            options={"xtol": wfnthreshold, "maxiter": wfnmaxiter},
            callback=None,
        )
        if not sol.success:
            raise NoConvergence(
                f"ERROR: program terminated. Error in solving geminal equations: "
                f"{sol.message}"
            )
        if log.do_high:
            log(
                f"Optimization of geminal coefficients converged in {sol.nit} iterations."
            )
        #
        # Update IOData container
        #
        self.checkpoint.update("niter_t_p", sol.nit)
        return sol.x

    @timer.with_section("LagrangeRpCCD")
    def solve_lagrange(self, guess, solver, wfnthreshold, wfnmaxiter):
        """Solves for Lagrange multipliers

           **Arguments:**

           guess
                The initial guess. A 1-dim np array.

           solver
                The solver used. A dictionary with element 'lagrange'.

           wfnthreshold
                The optimization threshold. A float.

           wfnmaxiter
                The maximum number of iterations. An integer.
        """
        iiaa = self.get_aux_matrix("gppqq")
        iaia = self.get_aux_matrix("lpqpq")
        one = self.get_aux_matrix("t")
        fock = self.get_aux_matrix("fock")
        sol = opt.root(
            self.vector_function_lagrange,
            guess,
            args=(self.geminal, iiaa, iaia, one, fock),
            method=solver["lagrange"],
            callback=None,
            options={"xtol": wfnthreshold, "maxiter": wfnmaxiter},
        )
        if not sol.success:
            raise NoConvergence(
                f"ERROR: program terminated. Error in solving Lagrange multipliers: "
                f"{sol.message}"
            )
        if log.do_high:
            log(
                f"Optimization of Lagrange multipliers converged in {sol.nit} iterations."
            )
        #
        # Update IOData container
        #
        self.checkpoint.update("niter_l_p", sol.nit)
        return sol.x

    #
    # Auxiliary matrices
    #
    def init_aux_matrix(self, select):
        """Initialize auxiliary matrices

           **Arguments:**

           select
                One of ``t``, ``fock``, ``gppqq``, ``gpqpq``, ``gpqqp``,
                ``lpqpq``, ``lpqrq``, ``gpqrr``
        """
        check_options(
            "select",
            select,
            "t",
            "fock",
            "gppqq",
            "gpqpq",
            "gpqqp",
            "lpqpq",
            "lpqrq",
            "gpqrr",
        )
        if select in ["lpqrq", "gpqrr"]:
            alloc_size = (self._lf.create_three_index, self.nbasis)
        elif select in ["fock"]:
            alloc_size = (self._lf.create_one_index, self.nbasis)
        else:
            alloc_size = (self._lf.create_two_index, self.nbasis)

        matrix, new = self._cache.load(
            f"matrix_{select}", alloc=alloc_size, tags="m",
        )
        if not new:
            raise RuntimeError(
                f"The auxiliary matrix matrix_{select} already exists. Call clear "
                f"prior to updating the auxiliary matrices."
            )
        return matrix

    @timer.with_section("IntegralSort")
    def update_aux_matrix(self, select, two_mo, one_mo):
        """Derive all auxiliary matrices.
           gppqq:   <pp|qq>,
           gpqpq:   <pq|pq>,
           gpqqp:   <pq|qp>,
           lpqpq:   2<pq|pq>-<pq|qp>,
           lpqrq:   2<pq|rq>-<pq|qr>,
           gpqrr:   <pq|rr>,
           fock:    h_pp + sum_i(2<pi|pi>-<pi|ip>)

           **Arguments:**

           select
                ``scf``.

           two_mo
                two-electron integrals in the MO basis. A FourIndex instance.

           one_mo
                one-electron integrals in the MO basis. A TwoIndex instance.
        """
        # Get ranges:
        no = self.get_range("no")
        try:
            one_mo = one_mo[0]
            two_mo = two_mo[0]
        except:
            pass
        if select == "scf":
            matrix = self.init_aux_matrix("t")
            matrix.assign(one_mo)
            #
            # <pp|qq>
            #
            matrix1 = self.init_aux_matrix("gppqq")
            two_mo.slice("aabb->ab", matrix1)
            #
            # <pq|pq>
            #
            matrix2 = self.init_aux_matrix("gpqpq")
            two_mo.slice("abab->ab", matrix2)
            #
            # <pq|qp>
            #
            matrix3 = self.init_aux_matrix("gpqqp")
            two_mo.slice("abba->ab", matrix3)
            #
            # <pq||pq>+<pq|pq>
            #
            matrix4 = self.init_aux_matrix("lpqpq")
            two_mo.slice("abab->ab", matrix4, 2.0, True)
            matrix4.iadd(matrix3, factor=-1.0)
            #           two_mo.slice('abba->ab', matrix4,-1.0, False)
            #
            # <pq||rq>+<pq|rq>
            #
            matrix5 = self.init_aux_matrix("lpqrq")
            two_mo.slice("abcb->abc", matrix5, 2.0, True)
            # this is really slow and we do not need to do it this way
            # exchange part is added in <pq|rr> block
            # two_mo.slice('abbc->abc', matrix5,-1.0, False)
            #
            # <pq|rr>
            #
            matrix6 = self.init_aux_matrix("gpqrr")
            two_mo.slice("abcc->abc", matrix6)
            #
            # add exchange part to matrix5
            #
            tmp3 = self.lf.create_three_index(self.nbasis)
            tmp3.assign(matrix6._array.transpose(0, 2, 1))
            matrix5.iadd(tmp3, factor=-1.0)
            del tmp3
            #
            # inactive diagonal Fock = h_pp + sum_i (<pi||pi>+<pi|pi>)
            #
            matrix7 = self.init_aux_matrix("fock")
            tmp1 = self.lf.create_one_index(self.nbasis)
            matrix.copy_diagonal(tmp1)
            matrix4.contract("ab->a", matrix7, **no, select="einsum")
            matrix7.iadd(tmp1)
        elif select == "nonscf":
            matrix = self.init_aux_matrix("t")
            matrix.assign(one_mo)
            #
            # <pp|qq>
            #
            matrix1 = self.init_aux_matrix("gppqq")
            two_mo.slice("aabb->ab", matrix1)
            #
            # <pq|pq>
            #
            matrix2 = self.init_aux_matrix("gpqpq")
            two_mo.slice("abab->ab", matrix2)
            #
            # <pq|qp>
            #
            matrix3 = self.init_aux_matrix("gpqqp")
            two_mo.slice("abba->ab", matrix3)
            #
            # <pq||pq>+<pq|pq>
            #
            matrix4 = self.init_aux_matrix("lpqpq")
            two_mo.slice("abab->ab", matrix4, 2.0, True)
            two_mo.slice("abba->ab", matrix4, -1.0, False)
            #
            # <pq||rq>+<pq|rq>
            #
            matrix5 = self.init_aux_matrix("lpqrq")
            two_mo.slice("abcb->abc", matrix5, 2.0, True)
            # this is really slow and we do not need to do it this way
            # exchange part is added in <pq|rr> block
            # two_mo.slice('abbc->abc', matrix5,-1.0, False)
            #
            # <pq|rr>
            #
            matrix6 = self.init_aux_matrix("gpqrr")
            two_mo.slice("abcc->abc", matrix6)
            #
            # add exchange part to matrix5
            #
            tmp3 = self.lf.create_three_index(self.nbasis)
            tmp3.assign(matrix6._array.transpose(0, 2, 1))
            matrix5.iadd(tmp3, factor=-1.0)
            del tmp3
            #
            # inactive diagonal Fock = h_pp + sum_i (<pi||pi>+<pi|pi>)
            #
            matrix7 = self.init_aux_matrix("fock")
            tmp1 = self.lf.create_one_index(self.nbasis)
            matrix.copy_diagonal(tmp1)
            matrix4.contract("ab->a", matrix7, **no, select="einsum")
            matrix7.iadd(tmp1)
        elif select == "guess":
            matrix = self.init_aux_matrix("t")
            matrix.assign(one_mo)
            #
            # <pp|qq>
            #
            matrix1 = self.init_aux_matrix("gppqq")
            two_mo.slice("aabb->ab", matrix1)
            #
            # <pq||pq>+<pq|pq>
            #
            matrix4 = self.init_aux_matrix("lpqpq")
            two_mo.slice("abab->ab", matrix4, 2.0, True)
            two_mo.slice("abba->ab", matrix4, -1.0, False)
            #
            # inactive diagonal Fock = h_pp + sum_i (<pi||pi>+<pi|pi>)
            #
            matrix7 = self.init_aux_matrix("fock")
            tmp1 = self.lf.create_one_index(self.nbasis)
            matrix.copy_diagonal(tmp1)
            matrix4.contract("ab->a", matrix7, **no, select="einsum")
            matrix7.iadd(tmp1)
        else:
            raise ArgumentError(
                f"The auxiliary matrix matrix_{select} is not supported."
            )

    def get_aux_matrix(self, select):
        """Get an auxiliary matrix from cache.

           **Arguments:**

           select
                't', 'gpqpq', 'gpqqp', 'lpqpq', 'pqrq', 'gpqrr', 'fock'.
        """
        if not f"matrix_{select}" in self._cache:
            raise ValueError(
                f"The aux_matrix {select} not found in cache. Did you use init_aux_matrix?"
            )
        return self._cache.load(f"matrix_{select}")

    #
    # Functions for energy evaluation:
    #
    def compute_correlation_energy(self, arg=None):
        """Get correlation energy of restricted pCCD

           **Optional arguments:**

           arg
                The geminal coefficient matrix (np.array or TwoIndex instance).
                If not provided, the correlation energy is calculated
                from self.geminal (default None)
        """
        ov = self.get_range("ov", 2)
        if arg is None:
            coeff = self.geminal
        else:
            coeff = self.lf.create_two_index(self.npairs, self.nvirt)
            coeff.assign(arg)
        kmat = self.get_aux_matrix("gppqq")
        #
        # Ecorr = sum_ia c_ia <ii|aa>
        #
        return coeff.contract("ab,ab", kmat, **ov, select="einsum")

    def compute_reference_energy(self):
        """Get energy of reference determinant for restricted pCCD including core energy"""
        one = self.get_aux_matrix("t")
        fock = self.get_aux_matrix("fock")
        #
        # Eref = sum_i (t_ii + F_ii)
        #
        energy = one.trace(0, self.npairs, 0, self.npairs)
        energy += fock.trace(0, self.npairs)
        return energy + self.e_core

    def compute_total_energy(self, coeff=None):
        """Get total energy (reference+correlation) including nuclear-repulsion/
           core energy for restricted pCCD.

           **Optional arguments:**

           arg
                The geminal coefficient matrix (np.array or TwoIndex instance).
                If not provided, the correlation energy is calculated
                from self.geminal (default None)
        """
        return (
            self.compute_correlation_energy(coeff)
            + self.compute_reference_energy()
        )

    #
    # Functions for orbital optimization (except gradient/Hessian):
    #
    def compute_rotation_matrix(self, coeff):
        """Determine orbital rotation matrix for (oo), (vo), and (vv) blocks

           **Arguments:**

           coeff
                The nonreduntant orbital rotations k_pq (1-dim np.array). The
                elements are sorted w.r.t. lower triangular indices (p>q).
        """
        indl = np.tril_indices(self.nbasis, -1)
        # account for frozen core
        kappa = self.lf.create_two_index(
            self.nbasis + self.ncore, self.nbasis + self.ncore
        )
        kappa.assign(coeff, indl, begin0=self.ncore, begin1=self.ncore)
        #
        # k_pq = -k_qp
        #
        kappa.iadd_t(kappa, -1.0)

        out = compute_unitary_matrix(kappa)
        return out

    def orbital_rotation_step(
        self, lshift=1e-8, pos=True, optimizer="variational"
    ):
        """Get orbital rotation step (Newton--Raphson)

           **Arguments:**


           **Optional arguments:**

           lshift
               Level shift (float) for approximate Hessian added to small elements
               (default 1e-8)

           pos
               (boolean) Make all elements of Hessian positive if set to True
               (default True)

           optimizer
               Orbital otimization method (str) (default 'variational')

        """
        check_options("orbitaloptimizer", optimizer, "variational", "ps2c")
        #
        # Switch between different orbital optimization schemes
        #
        ps2c = optimizer == "ps2c"
        #
        # Calculate orbital gradient and diagonal approximation to the Hessian
        #
        grad = self.compute_orbital_gradient(ps2c)
        #
        # We use a diagonal Hessian
        #
        hessian = self.compute_orbital_hessian(lshift, pos, ps2c)
        #
        # Orbital rotation step
        #
        kappa = grad.divide(hessian, -1.0)

        return kappa, grad, hessian

    #
    # Vector function for pCCD:
    #
    @timer.with_section("VecFctGeminal")
    def vector_function_geminal(self, coeff, miiaa, miaia, one, diagfock):
        """Construct vector function for optimization of geminal coefficients.

           **Arguments:**

           coeff
                The geminal coefficients (np.array)

           miiaa, miaia
                Auxiliary matrices (TwoIndex instances)

           one
                1-electron integrals (TwoIndex instance)

           diagfock
                Diagonal inactive Fock matrix (OneIndex instances)
        """
        # get ranges
        ov = self.get_range("ov")
        o2 = self.get_range("o", 2)
        v2 = self.get_range("v", 2)
        oo = self.get_range("oo")
        vv = self.get_range("vv")
        vo2 = self.get_range("vo", 2)
        ov2 = self.get_range("ov", 2)
        vv2 = self.get_range("vv", 2)
        gmat = self.lf.create_two_index(self.npairs, self.nvirt)
        gmat.assign(coeff)

        #
        # correlation energy for current iteration step
        #
        e_corr = self.compute_correlation_energy(gmat)

        #
        # vectorFunction_ia
        #
        result = self.lf.create_two_index(self.npairs, self.nvirt)

        #
        # Add contributions to vectorFunction_ia:
        #
        # c_ia*e_corr
        #
        result.iadd(gmat, -e_corr)

        #
        # c_0*miiaa
        #
        result.iadd(miiaa, 1.0, **ov2)

        #
        # -2c_ia*f_ii
        #
        gmat.contract("ab,a->ab", diagfock, result, factor=-2.0, **o2)

        #
        # 2c_ia*f_aa
        #
        gmat.contract("ab,b->ab", diagfock, result, factor=2.0, **v2)

        #
        # -2c_ia*(<ia||ia>+<ia|ia>)
        #
        result.iadd_mult(gmat, miaia, -2.0, **ov2)

        #
        # c_ja*(<ii|jj>)
        #
        result.iadd_dot(miiaa, gmat, 1.0, **oo)

        #
        # c_ib*(<bb|aa>)
        #
        result.iadd_dot(gmat, miiaa, 1.0, **vv2)

        #
        # c_ia*(c_jb*<jj|bb>)
        #
        factor = gmat.contract("ab,ab", miiaa, **ov2, select="einsum")
        result.iadd(gmat, factor)

        #
        # c_ib*<bb|jj>*c_ja
        #
        tmp = self.lf.create_two_index(self.npairs, self.npairs)
        tmp.iadd_dot(gmat, miiaa, 1.0, **vo2)
        result.iadd_dot(tmp, gmat, 1.0)

        #
        # -2c_ja*c_ia*<jj|aa>
        #
        tmpv = self.lf.create_one_index(self.nvirt)
        gmat.contract("ab,ab->b", miiaa, tmpv, **ov2, select="einsum")
        gmat.contract("ab,b->ab", tmpv, result, factor=-2.0, select="einsum")

        #
        # -2c_ib*c_ia*<ii|bb>
        #
        tmpo = self.lf.create_one_index(self.nocc)
        gmat.contract("ab,ab->a", miiaa, tmpo, **ov2, select="einsum")
        gmat.contract("ab,a->ab", tmpo, result, factor=-2.0, select="einsum")

        #
        # +2c_ia*c_ia*<ii|aa>
        #
        tmp = gmat.copy()
        tmp.imul(gmat)
        result.iadd_mult(tmp, miiaa, 2.0, **ov2)

        return result._array.ravel(order="C")

    #
    # Jacobian for pCCD:
    #
    def jacobian_pccd(self, coeff, miiaa, miaia, one, fock):
        """Construct Jacobian for optimization of geminal coefficients.
        """
        raise NotImplementedError

    #
    # Vector function for Lagrange multipliers:
    #
    @timer.with_section("VecFctLagrange")
    def vector_function_lagrange(
        self, lagrange, gmat, miiaa, miaia, one, diagfock
    ):
        """Construct vector function for optimization of Lagrange multipliers.

           **Arguments:**

           lagrange
                The lagrange multipliers (np array)

           gmat
                The geminal coefficients (TwoIndex instance)

           miiaa, miaia
                Auxiliary matrices (TwoIndex instances)

           one
                1-electron integrals (TwoIndex instance)

           diagfock
                Diagonal inactive Fock matrix (OneIndex instance)
        """
        ov = self.get_range("ov")
        o2 = self.get_range("o", 2)
        v2 = self.get_range("v", 2)
        oo = self.get_range("oo")
        ov2 = self.get_range("ov", 2)
        vv2 = self.get_range("vv", 2)
        lmat = self.lf.create_two_index(self.npairs, self.nvirt)
        lmat.assign(lagrange)

        #
        # Correlation energy
        #
        e_corr = self.compute_correlation_energy(gmat)
        #
        # intermediate variables:
        #  * lc = sum_ia c_ia l_ia
        #  * ciiaa = sum_ia c_ia <ii|aa>
        #  * diagint = <pp|pp>
        #
        lc = lmat.contract("ab,ab", gmat, select="einsum")
        ciiaa = gmat.contract("ab,ab", miiaa, **ov2, select="einsum")
        diagint = miiaa.copy_diagonal()
        #
        # cgi = sum_b c_ib <ii|bb>
        # cga = sum_j c_ja <jj|aa>
        # lci = sum_b c_ib l_ib
        # lca = sum_j c_ja l_ja
        #
        cgi = self.lf.create_one_index(self.npairs)
        cga = self.lf.create_one_index(self.nvirt)
        lci = self.lf.create_one_index(self.npairs)
        lca = self.lf.create_one_index(self.nvirt)
        gmat.contract("ab,ab->a", miiaa, cgi, **ov2, select="einsum")
        gmat.contract("ab,ab->b", miiaa, cga, **ov2, select="einsum")
        lmat.contract("ab,ab->a", gmat, lci, select="einsum")
        lmat.contract("ab,ab->b", gmat, lca, select="einsum")

        #
        # vectorFunction_ia
        #
        result = self.lf.create_two_index(self.npairs, self.nvirt)

        #
        # miiaa
        #
        result.iadd(miiaa, 1.0, **ov2)

        #
        # -2l_ia*cgi
        #
        lmat.contract("ab,a->ab", cgi, result, factor=-2.0)

        #
        # -2l_ia*cga
        #
        lmat.contract("ab,b->ab", cga, result, factor=-2.0)

        #
        # -2miiaa*lci
        #
        miiaa.contract("ab,a->ab", lci, result, factor=-2.0, **ov)

        #
        # -2miiaa*lca
        #
        miiaa.contract("ab,b->ab", lca, result, factor=-2.0, **ov)

        #
        # -2l_ia*f_ii
        #
        lmat.contract("ab,a->ab", diagfock, result, factor=-2.0, **o2)

        #
        # 2l_ia*f_aa
        #
        lmat.contract("ab,b->ab", diagfock, result, factor=2.0, **v2)

        #
        # -2l_ia*(<ia||ia>+<ia|ia>)
        #
        result.iadd_mult(lmat, miaia, -2.0, **ov2)

        #
        # l_ja*(<ii|jj>)
        #
        result.iadd_dot(miiaa, lmat, 1.0, **oo)

        #
        # l_ib*(<bb|aa>)
        #
        result.iadd_dot(lmat, miiaa, 1.0, **vv2)

        #
        # 4l_ia*c_ia*<ii|aa>
        #
        tmp = lmat.copy()
        tmp.imul(gmat)
        result.iadd_mult(tmp, miiaa, 4.0, **ov2)

        #
        # <ii|bb>*c_jb*l_ja
        #
        tmp = self.lf.create_two_index(self.nvirt, self.nvirt)
        tmp.iadd_tdot(gmat, lmat)
        result.iadd_dot(miiaa, tmp, 1.0, **ov)

        #
        # l_ib*c_jb*<jj|aa>
        #
        tmp = self.lf.create_two_index(self.npairs, self.npairs)
        tmp.iadd_dott(lmat, gmat)
        result.iadd_dot(tmp, miiaa, 1.0, **ov2)

        return result._array.ravel(order="C")

    #
    # Jacobian for Lagrange multipliers of OpCCD:
    #
    def jacobian_lambda(self, lagrange, gmat, miiaa, miaia, one, diagfock):
        """Construct Jacobian for optimization of Lagrange multipliers for
           restricted pCCD.
        """
        raise NotImplementedError

    #
    # Calculate orbital gradient for OpCCD (oo,vo,vv):
    #
    @timer.with_section("OOpCCD grad")
    def compute_orbital_gradient(self, ps2c=False):
        """Determine orbital gradient for all non-reduntant orbital rotations
           (oo,vo,vv).

           **Optional arguments:**

           ps2c
                (boolean) If True, switches to PS2c orbital optimization
                (default False)
        """
        one = self.get_aux_matrix("t")
        vpqrq = self.get_aux_matrix("lpqrq")
        vpqrr = self.get_aux_matrix("gpqrr")

        oo = self.get_range("oo")
        vv = self.get_range("vv")
        ov = self.get_range("ov")
        vo = self.get_range("vo")
        #
        # Get 1- and 2-RDMs
        #
        self.clear_dm()
        if ps2c:
            self.update_ndm("one_dm_ps2")
            self.update_ndm("two_dm_pqpq")
            self.update_ndm("two_dm_ppqq")
            onedm = self.one_dm_ps2
            twodmpqpq = self.two_dm_pqpq
            twodmppqq = self.two_dm_ppqq
        else:
            self.update_ndm("one_dm_response")
            self.update_ndm("two_dm_rpqpq")
            self.update_ndm("two_dm_rppqq")
            onedm = self.one_dm_response
            twodmpqpq = self.two_dm_rpqpq
            twodmppqq = self.two_dm_rppqq
        #
        # Update IOData container
        #
        self.checkpoint.update("dm_1", onedm)
        self.checkpoint.update("dm_2", {"pqpq": twodmpqpq, "ppqq": twodmppqq})

        #
        # Orbital gradient g_ut
        #
        gradient = self.lf.create_two_index(self.nbasis, self.nbasis)

        #
        # Symmetrize G(amma)_ppqq 2-RDM. This reduces the number of operations
        #
        two_dm_av = twodmppqq.copy()
        two_dm_av.symmetrize()
        #
        # L_uptp*G_up
        #
        vpqrq.contract(
            "abc,ab->ac", twodmpqpq, gradient, factor=4.0, select="einsum"
        )
        #
        # <tu|pp>*G_up
        #
        vpqrr.contract(
            "abc,bc->ba", two_dm_av, gradient, factor=4.0, select="einsum"
        )
        #
        # -L_tpup*G_tp
        #
        vpqrq.contract(
            "abc,ab->ca", twodmpqpq, gradient, factor=-4.0, select="einsum"
        )
        #
        # -<ut|pp>*G_tp
        #
        vpqrr.contract(
            "abc,bc->ab", two_dm_av, gradient, factor=-4.0, select="einsum"
        )
        #
        # h_ut g_tt
        #
        one.contract("ab,a->ab", onedm, gradient, factor=4.0)
        #
        # h_ut g_uu
        #
        one.contract("ab,b->ab", onedm, gradient, factor=-4.0)

        ind = np.tril_indices(self.nbasis, -1)

        # TODO: This is a quick hack and should be cleaned up in the future
        if self.freezerot is not None:
            if "vv" in self.freezerot:
                gradient.assign(0.0, **vv)
            if "oo" in self.freezerot:
                gradient.assign(0.0, **oo)
            if "ov" in self.freezerot:
                gradient.assign(0.0, **ov)
                gradient.assign(0.0, **vo)

        #
        # return only lower triangle
        #
        return gradient.ravel(ind=ind)

    #
    # Approximate diagonal Hessian for OpCCD:
    #
    def compute_orbital_hessian(self, lshift=1e-8, pos=False, ps2c=False):
        """Construct diagonal approximation to Hessian for orbital optimization
           of pCCD.

           **Optional arguments:**

           lshift
                Level shift (float) (default 1e-8)

           pos
                Set all elements positive (boolean) (default False)

           ps2c
                (boolean) If True, switches to PS2c orbital optimization
                (default False)
        """
        #
        # Get auxiliary matrices
        #
        one = self.get_aux_matrix("t")
        vpq = self.get_aux_matrix("lpqpq")
        vpp = self.get_aux_matrix("gppqq")
        vmat = self.get_aux_matrix("gpqpq")
        vmata = self.get_aux_matrix("gpqqp")

        #
        # Get 1- and 2-RDMs
        #
        self.clear_dm()
        if ps2c:
            self.update_ndm("one_dm_ps2")
            self.update_ndm("two_dm_pqpq")
            self.update_ndm("two_dm_ppqq")
            onedm = self.one_dm_ps2
            twodmpqpq = self.two_dm_pqpq
            twodmppqq = self.two_dm_ppqq
        else:
            self.update_ndm("one_dm_response")
            self.update_ndm("two_dm_rpqpq")
            self.update_ndm("two_dm_rppqq")
            onedm = self.one_dm_response
            twodmpqpq = self.two_dm_rpqpq
            twodmppqq = self.two_dm_rppqq

        #
        # Symmetrize G(amma)_ppqq 2-RDM
        #
        two_dm_av = twodmppqq.copy()
        two_dm_av.symmetrize()
        #
        # Modify diagonal elements to simplify contractions
        #
        two_dm_av.assign_diagonal(0.0)
        twodmpqpq.assign_diagonal(0.0)

        #
        # Calculate additional auxiliary matrices
        #
        vmatdiag = vmat.copy_diagonal()
        one_diag = one.copy_diagonal()
        two_c = self.lf.create_one_index(self.nbasis)
        two_ca = self.lf.create_one_index(self.nbasis)
        vpq.contract("ab,ab->a", twodmpqpq, two_c, select="einsum")
        vpp.contract("ab,ab->a", two_dm_av, two_ca, select="einsum")

        #
        # Diagonal orbital Hessian hessian_pq = hessian_(pq,pq)
        #
        hessian = self.lf.create_two_index(self.nbasis, self.nbasis)
        #
        # <qt> G_pt
        #
        hessian.iadd_dot(vpq, twodmpqpq, 4.0)
        #
        # <pt> G_qt
        #
        hessian.iadd_dot(twodmpqpq, vpq, 4.0)
        #
        # <qt> G_qt
        #
        hessian.iadd(two_c, -4.0)
        #
        # <pt> G_pt
        #
        hessian.iadd_t(two_c, -4.0)
        #
        # <qt> G_pt
        #
        hessian.iadd_dot(two_dm_av, vpp, 4.0)
        #
        # <pt> G_qt
        #
        hessian.iadd_dot(vpp, two_dm_av, 4.0)
        #
        # <qt> G_qt
        #
        hessian.iadd(two_ca, -4.0)
        #
        # <pt> G_pt
        #
        hessian.iadd_t(two_ca, -4.0)
        #
        # <pq> G_pq
        #
        hessian.iadd_mult(vmat, twodmpqpq, 8.0)
        hessian.iadd_mult(vmata, twodmpqpq, -8.0)
        hessian.iadd_mult(vpp, twodmpqpq, -16.0)
        hessian.iadd_mult(vmat, two_dm_av, -8.0)
        hessian.iadd_mult(vpp, two_dm_av, -8.0)
        #
        # <ppqq> G_pp
        #
        vpp.contract("ab,a->ab", onedm, hessian, factor=8.0)
        vpp.contract("ab,b->ab", onedm, hessian, factor=8.0)
        #
        # <qq> g_pp
        #
        hessian.iadd_one_mult(one_diag, onedm, 4.0, True, False)
        #
        # <pp> g_qq
        #
        hessian.iadd_one_mult(onedm, one_diag, 4.0, True, False)
        #
        # <qq> g_qq
        #
        hessian.iadd_one_mult(one_diag, onedm, -4.0, True, True)
        #
        # <pp> g_pp
        #
        hessian.iadd_one_mult(onedm, one_diag, -4.0, False, False)
        vmat.contract("ab,a->ab", onedm, hessian, factor=4.0)
        vmat.contract("ab,b->ab", onedm, hessian, factor=4.0)
        hessian.iadd_one_mult(vmatdiag, onedm, -4.0, True, True)
        hessian.iadd_one_mult(vmatdiag, onedm, -4.0, False, False)

        #
        # Make everything positive
        #
        if pos:
            hessian.iabs()
        #
        # Add levelshift:
        #
        if lshift:
            hessian.iadd_shift(lshift)

        ind = np.tril_indices(self.nbasis, -1)

        return hessian.ravel(ind=ind)

    @timer.with_section("exact Hessian")
    def get_exact_hessian(self, mo1, mo2):
        """Construct exact Hessian for orbital optimization of restricted OpCCD.

           The usage of the exact orbital Hessian for the orbital optimization
           is currently not supported. The exact Hessian can only be evaluated
           a posteriori.

           **Arguements**

           mo1, mo2
                The 1- and 2-el integrals in the MO basis (TwoIndex and
                FourIndex instances)
        """
        if log.do_medium:
            log("Calculating exact Hessian of pCCD")
        #
        # exact orbital Hessian output hessian_pq,rs
        #
        hessian = self.denself.create_four_index(self.nbasis)
        if isinstance(mo2, CholeskyFourIndex):
            mo2 = mo2.get_dense()

        #
        # Get response DMs
        #
        self.clear_dm()
        self.update_ndm("one_dm_response")
        self.update_ndm("two_dm_rpqpq")
        self.update_ndm("two_dm_rppqq")
        dm1 = self.one_dm_response.copy()
        dm2pqpq = self.two_dm_rpqpq.copy()
        dm2pqpqex = self.two_dm_rpqpq.copy()
        dm2ppqq = self.two_dm_rppqq.copy()
        #
        # Symmetrize 2DM
        #
        dm2av = dm2ppqq.copy()
        dm2av.symmetrize()
        #
        # Reset diagonal elements of DMs
        #
        dm2pqpqex.assign_diagonal(0.0)
        dm2pqpq.assign_diagonal(dm1)
        dm2av.assign_diagonal(0.0)

        #
        # temporary storage
        #
        ind2 = self.lf.create_two_index(self.nbasis)
        ind30 = self.lf.create_three_index(self.nbasis)
        ind31 = self.lf.create_three_index(self.nbasis)

        if log.do_medium:
            log(" Direct part...")
        # Direct part
        # (1)
        # ab
        mo2.contract(
            "abcd,cd->acbd", dm2pqpq, hessian, factor=4.0, select="einsum"
        )
        mo2.contract(
            "abcd,ab->acdb", dm2pqpq, hessian, factor=-4.0, select="einsum"
        )
        mo2.contract(
            "abcd,cd->acdb", dm2pqpq, hessian, factor=-4.0, select="einsum"
        )
        mo2.contract(
            "abcd,ab->acbd", dm2pqpq, hessian, factor=4.0, select="einsum"
        )
        # (2)
        # ab
        mo2.contract(
            "abcd,cb->acdb", dm2pqpq, hessian, factor=4.0, select="einsum"
        )
        mo2.contract(
            "abcd,ad->acbd", dm2pqpq, hessian, factor=-4.0, select="einsum"
        )
        mo2.contract(
            "abcd,cb->acbd", dm2pqpq, hessian, factor=-4.0, select="einsum"
        )
        mo2.contract(
            "abcd,ad->acdb", dm2pqpq, hessian, factor=4.0, select="einsum"
        )
        # (3)
        mo2.contract(
            "abcd,bd->abcd", dm2av, hessian, factor=4.0, select="einsum"
        )
        mo2.contract(
            "abcd,ad->abcd", dm2av, hessian, factor=-4.0, select="einsum"
        )
        mo2.contract(
            "abcd,bd->abdc", dm2av, hessian, factor=-4.0, select="einsum"
        )
        mo2.contract(
            "abcd,ac->abcd", dm2av, hessian, factor=4.0, select="einsum"
        )
        mo2.contract(
            "abcd,bc->abdc", dm2av, hessian, factor=4.0, select="einsum"
        )
        mo2.contract(
            "abcd,ac->abdc", dm2av, hessian, factor=-4.0, select="einsum"
        )
        mo2.contract(
            "abcd,bc->abcd", dm2av, hessian, factor=-4.0, select="einsum"
        )
        mo2.contract(
            "abcd,ad->abdc", dm2av, hessian, factor=4.0, select="einsum"
        )
        # Apq,qw (pw) (qv) (-qw) (-pv)
        mo1.contract("ab,b->ab", dm1, ind2, factor=4.0, select="einsum")
        # ab
        mo2.contract("abcb,cb->ac", dm2pqpq, ind2, factor=2.0, select="einsum")
        mo2.contract("abcb,ab->ac", dm2pqpq, ind2, factor=2.0, select="einsum")
        mo2.contract("abcc,bc->ba", dm2av, ind2, factor=2.0, select="einsum")
        mo2.contract("abcc,bc->ab", dm2av, ind2, factor=2.0, select="einsum")
        # Apq,pw (pq,pw) (-pq,vp)
        mo1.contract("ab,c->cab", dm1, ind30, factor=4, select="einsum")
        # ab
        mo2.contract(
            "abcb,db->dac", dm2pqpq, ind30, factor=4.0, select="einsum"
        )
        mo2.contract("abcc,dc->dab", dm2av, ind30, factor=4.0, select="einsum")
        # Apq,vq (pq,vq) (-pq,qw)
        mo1.contract("ab,c->acb", dm1, ind31, factor=4, select="einsum")
        # ab
        mo2.contract(
            "abcb,db->adc", dm2pqpq, ind31, factor=4.0, select="einsum"
        )
        mo2.contract("abcc,dc->adb", dm2av, ind31, factor=4.0, select="einsum")

        if log.do_medium:
            log(" Exchange part...")
        # Exchange part
        #
        # calculate <pq|rs>-<pq|sr>
        #
        mo2.iadd_exchange()
        # (1)
        # aa
        mo2.contract(
            "abcd,cd->acbd", dm2pqpqex, hessian, factor=4.0, select="einsum"
        )
        mo2.contract(
            "abcd,ab->acdb", dm2pqpqex, hessian, factor=-4.0, select="einsum"
        )
        mo2.contract(
            "abcd,cd->acdb", dm2pqpqex, hessian, factor=-4.0, select="einsum"
        )
        mo2.contract(
            "abcd,ab->acbd", dm2pqpqex, hessian, factor=4.0, select="einsum"
        )
        # (2)
        # aa
        mo2.contract(
            "abcd,cb->acdb", dm2pqpqex, hessian, factor=4.0, select="einsum"
        )
        mo2.contract(
            "abcd,ad->acbd", dm2pqpqex, hessian, factor=-4.0, select="einsum"
        )
        mo2.contract(
            "abcd,cb->acbd", dm2pqpqex, hessian, factor=-4.0, select="einsum"
        )
        mo2.contract(
            "abcd,ad->acdb", dm2pqpqex, hessian, factor=4.0, select="einsum"
        )
        # Apq,qw (pw) (qv) (-qw) (-pv)
        # aa
        mo2.contract(
            "abcb,cb->ac", dm2pqpqex, ind2, factor=2.0, select="einsum"
        )
        mo2.contract(
            "abcb,ab->ac", dm2pqpqex, ind2, factor=2.0, select="einsum"
        )
        # Apq,pw (pq,pw) (-pq,vp)
        # aa
        mo2.contract(
            "abcb,db->dac", dm2pqpqex, ind30, factor=4.0, select="einsum"
        )
        # Apq,vp (pq,vq) (-pq,qw)
        # aa
        mo2.contract(
            "abcb,db->adc", dm2pqpqex, ind31, factor=4.0, select="einsum"
        )

        # release memory
        del mo2

        if log.do_medium:
            log(" Collecting terms...")
        #
        # collect 2- and 3-index terms
        #
        # Apq,vq
        hessian.iadd_expand_two_to_four("0-2", ind2, -1.0)
        # Apq,qw
        hessian.iadd_expand_two_to_four("0-3", ind2, 1.0)
        # Apq,vp
        hessian.iadd_expand_two_to_four("1-2", ind2, 1.0)
        # Apq,pw
        hessian.iadd_expand_two_to_four("1-3", ind2, -1.0)
        # Apq,vw (pq,pw) (-pq,vp)
        hessian.iadd_expand_three_to_four("1-3-1-2", ind30, 1.0)
        hessian.iadd_expand_three_to_four("1-2-1-2", ind30, -1.0)
        # Apq,vw (pq,vq) (-pq,qw)
        hessian.iadd_expand_three_to_four("0-2-0-2", ind31, 1.0)
        hessian.iadd_expand_three_to_four("0-3-0-2", ind31, -1.0)

        #
        # reorder elements (p,q,r,s)->(pq,rs)
        #
        dim = (self.nbasis * (self.nbasis - 1)) // 2
        tril = np.tril_indices(self.nbasis, -1)
        out = self.lf.create_two_index(dim, dim)
        # FIXME: use matrix class
        out._array[:] = (hessian._array[:, :, tril[0], tril[1]])[tril]

        # release memory
        del hessian
        if log.do_medium:
            log("...Done")

        return out._array

    def sort_natural_orbitals(self, orb):
        """Sort orbitals w.r.t. the natural occupation numbers

           **Arguments:**

           orb
                The AO/MO coefficients (natural orbitals) and the natural
                occupation numbers (Orbital instance)
        """
        #
        # Take natural occupation numbers from orb.occupations
        #
        onedm = self.lf.create_one_index()
        onedm.assign(orb.occupations)
        order = onedm.sort_indices(begin0=self.ncore)
        orderref = np.arange(self.nbasis)
        if not (order == orderref).all():
            orb.permute_orbitals(order, begin0=self.ncore)

    def print_solution(
        self,
        cithresh=0.05,
        excitationlevel=2,
        amplitudestofile=False,
        fname="pccd_amplitudes.dat",
    ):
        """Print coefficients :math:`{ci}` and Slater Determinant if :math:`|ci|` > threshold.
           Prints up to hextuply excited pairs.

           **Optional arguments:**

           cithresh
                Upper threshold (a float) for CI coefficients to be
                reconstructed. (Default 0.01)

           excitationslevel
                The maximum number of substitutions w.r.t. the reference
                determinant for which the CI coefficients are reconstructed.
                (Default 2)

           amplitudestofile
                A boolean. If True, the CI coefficients are stored in a separate
                file. (Default False)

           filename
                The file name for the wfn amplitudes.
                (Default pccd_amplitudes)
        """
        it = []
        coeff = self.geminal._array.copy()
        matrix = np.identity(self.npairs)
        solution_dict = {}
        # dump to pybest-results dir
        filename = filemanager.result_path(fname)
        for i in range(excitationlevel):
            it.append(np.nditer(coeff, flags=["multi_index"]))
        if amplitudestofile:
            with open(filename, "w") as filea:
                filea.write(
                    f"{self.get_slater_determinant((-1, -1))} {1.0: 20.16f}"
                )
                filea.write("\n")
        else:
            log(f"{self.get_slater_determinant((-1, -1))} {1.0: 20.16f}")
        for i in range(excitationlevel):
            if i == 0:
                for ci in it[0]:
                    if abs(ci) >= (cithresh):
                        sd = self.get_slater_determinant(it[0].multi_index)
                        if amplitudestofile is True:
                            with open(filename, "a") as filea:
                                filea.write(f"{sd} {ci: 20.16f}")
                                filea.write("\n")
                        else:
                            solution_dict.update({sd: float(ci)})
            if self.npairs > 1:
                for index in range((i + 1)):
                    it[index] = np.nditer(coeff, flags=["multi_index"])
                if i == 1:
                    for ci in it[0]:
                        if abs(ci) >= (cithresh / excitationlevel):
                            it[1] = it[0].copy()
                            for ci2 in it[1]:
                                if it[1].multi_index[0] > it[0].multi_index[0]:
                                    if (
                                        it[1].multi_index[1]
                                        > it[0].multi_index[1]
                                    ):
                                        matrix[
                                            it[0].multi_index[0],
                                            it[0].multi_index[0],
                                        ] = float(ci)
                                        matrix[
                                            it[1].multi_index[0],
                                            it[1].multi_index[0],
                                        ] = float(ci2)
                                        matrix[
                                            it[0].multi_index[0],
                                            it[1].multi_index[0],
                                        ] = coeff[
                                            it[0].multi_index[0],
                                            it[1].multi_index[1],
                                        ]
                                        matrix[
                                            it[1].multi_index[0],
                                            it[0].multi_index[0],
                                        ] = coeff[
                                            it[1].multi_index[0],
                                            it[0].multi_index[1],
                                        ]
                                        amplitude = self.perm(matrix)
                                        if abs(amplitude) >= cithresh:
                                            sd = self.get_slater_determinant(
                                                it[0].multi_index,
                                                it[1].multi_index,
                                            )
                                            if amplitudestofile is True:
                                                with open(
                                                    filename, "a"
                                                ) as filea:
                                                    filea.write(
                                                        f"{sd} {amplitude: 20.16f}"
                                                    )
                                                    filea.write("\n")
                                            else:
                                                solution_dict.update(
                                                    {sd: float(amplitude)}
                                                )
                                        matrix = np.identity(self.npairs)
                if i == 2:
                    for ci in it[0]:
                        if abs(ci) >= (cithresh / excitationlevel):
                            it[1] = it[0].copy()
                            for ci2 in it[1]:
                                if it[1].multi_index[0] > it[0].multi_index[0]:
                                    if (
                                        it[1].multi_index[1]
                                        > it[0].multi_index[1]
                                    ):
                                        it[2] = it[1].copy()
                                        for ci3 in it[2]:
                                            if (
                                                it[2].multi_index[0]
                                                > it[1].multi_index[0]
                                            ):
                                                if (
                                                    it[2].multi_index[1]
                                                    > it[1].multi_index[1]
                                                ):
                                                    matrix[
                                                        it[0].multi_index[0],
                                                        it[0].multi_index[0],
                                                    ] = float(ci)
                                                    matrix[
                                                        it[1].multi_index[0],
                                                        it[1].multi_index[0],
                                                    ] = float(ci2)
                                                    matrix[
                                                        it[2].multi_index[0],
                                                        it[2].multi_index[0],
                                                    ] = float(ci3)

                                                    matrix[
                                                        it[0].multi_index[0],
                                                        it[1].multi_index[0],
                                                    ] = coeff[
                                                        it[0].multi_index[0],
                                                        it[1].multi_index[1],
                                                    ]
                                                    matrix[
                                                        it[0].multi_index[0],
                                                        it[2].multi_index[0],
                                                    ] = coeff[
                                                        it[0].multi_index[0],
                                                        it[2].multi_index[1],
                                                    ]

                                                    matrix[
                                                        it[1].multi_index[0],
                                                        it[0].multi_index[0],
                                                    ] = coeff[
                                                        it[1].multi_index[0],
                                                        it[0].multi_index[1],
                                                    ]
                                                    matrix[
                                                        it[1].multi_index[0],
                                                        it[2].multi_index[0],
                                                    ] = coeff[
                                                        it[1].multi_index[0],
                                                        it[2].multi_index[1],
                                                    ]

                                                    matrix[
                                                        it[2].multi_index[0],
                                                        it[0].multi_index[0],
                                                    ] = coeff[
                                                        it[2].multi_index[0],
                                                        it[0].multi_index[1],
                                                    ]
                                                    matrix[
                                                        it[2].multi_index[0],
                                                        it[1].multi_index[0],
                                                    ] = coeff[
                                                        it[2].multi_index[0],
                                                        it[1].multi_index[1],
                                                    ]
                                                    amplitude = self.perm(
                                                        matrix
                                                    )
                                                    if (
                                                        abs(amplitude)
                                                        >= cithresh
                                                    ):
                                                        sd = self.get_slater_determinant(
                                                            it[0].multi_index,
                                                            it[1].multi_index,
                                                            it[2].multi_index,
                                                        )
                                                        if (
                                                            amplitudestofile
                                                            is True
                                                        ):
                                                            with open(
                                                                filename, "a"
                                                            ) as filea:
                                                                filea.write(
                                                                    f"{sd} {amplitude: 20.16f}"
                                                                )
                                                                filea.write(
                                                                    "\n"
                                                                )
                                                        else:
                                                            solution_dict.update(
                                                                {
                                                                    sd: float(
                                                                        amplitude
                                                                    )
                                                                }
                                                            )
                                                    matrix = np.identity(
                                                        self.npairs
                                                    )
                if i == 3:
                    for ci in it[0]:
                        if abs(ci) >= (cithresh / excitationlevel):
                            it[1] = it[0].copy()
                            for ci2 in it[1]:
                                if it[1].multi_index[0] > it[0].multi_index[0]:
                                    if (
                                        it[1].multi_index[1]
                                        > it[0].multi_index[1]
                                    ):
                                        it[2] = it[1].copy()
                                        for ci3 in it[2]:
                                            if (
                                                it[2].multi_index[0]
                                                > it[1].multi_index[0]
                                            ):
                                                if (
                                                    it[2].multi_index[1]
                                                    > it[1].multi_index[1]
                                                ):
                                                    it[3] = it[2].copy()
                                                    for ci4 in it[3]:
                                                        if (
                                                            it[3].multi_index[
                                                                0
                                                            ]
                                                            > it[
                                                                2
                                                            ].multi_index[0]
                                                        ):
                                                            if (
                                                                it[
                                                                    3
                                                                ].multi_index[
                                                                    1
                                                                ]
                                                                > it[
                                                                    2
                                                                ].multi_index[
                                                                    1
                                                                ]
                                                            ):
                                                                matrix[
                                                                    it[
                                                                        0
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        0
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                ] = float(ci)
                                                                matrix[
                                                                    it[
                                                                        1
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        1
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                ] = float(ci2)
                                                                matrix[
                                                                    it[
                                                                        2
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        2
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                ] = float(ci3)
                                                                matrix[
                                                                    it[
                                                                        3
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        3
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                ] = float(ci4)

                                                                matrix[
                                                                    it[
                                                                        0
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        1
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                ] = coeff[
                                                                    it[
                                                                        0
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        1
                                                                    ].multi_index[
                                                                        1
                                                                    ],
                                                                ]
                                                                matrix[
                                                                    it[
                                                                        0
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        2
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                ] = coeff[
                                                                    it[
                                                                        0
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        2
                                                                    ].multi_index[
                                                                        1
                                                                    ],
                                                                ]
                                                                matrix[
                                                                    it[
                                                                        0
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        3
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                ] = coeff[
                                                                    it[
                                                                        0
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        3
                                                                    ].multi_index[
                                                                        1
                                                                    ],
                                                                ]

                                                                matrix[
                                                                    it[
                                                                        1
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        0
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                ] = coeff[
                                                                    it[
                                                                        1
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        0
                                                                    ].multi_index[
                                                                        1
                                                                    ],
                                                                ]
                                                                matrix[
                                                                    it[
                                                                        1
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        2
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                ] = coeff[
                                                                    it[
                                                                        1
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        2
                                                                    ].multi_index[
                                                                        1
                                                                    ],
                                                                ]
                                                                matrix[
                                                                    it[
                                                                        1
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        3
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                ] = coeff[
                                                                    it[
                                                                        1
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        3
                                                                    ].multi_index[
                                                                        1
                                                                    ],
                                                                ]

                                                                matrix[
                                                                    it[
                                                                        2
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        0
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                ] = coeff[
                                                                    it[
                                                                        2
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        0
                                                                    ].multi_index[
                                                                        1
                                                                    ],
                                                                ]
                                                                matrix[
                                                                    it[
                                                                        2
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        1
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                ] = coeff[
                                                                    it[
                                                                        2
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        1
                                                                    ].multi_index[
                                                                        1
                                                                    ],
                                                                ]
                                                                matrix[
                                                                    it[
                                                                        2
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        3
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                ] = coeff[
                                                                    it[
                                                                        2
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        3
                                                                    ].multi_index[
                                                                        1
                                                                    ],
                                                                ]

                                                                matrix[
                                                                    it[
                                                                        3
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        0
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                ] = coeff[
                                                                    it[
                                                                        3
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        0
                                                                    ].multi_index[
                                                                        1
                                                                    ],
                                                                ]
                                                                matrix[
                                                                    it[
                                                                        3
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        1
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                ] = coeff[
                                                                    it[
                                                                        3
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        1
                                                                    ].multi_index[
                                                                        1
                                                                    ],
                                                                ]
                                                                matrix[
                                                                    it[
                                                                        3
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        2
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                ] = coeff[
                                                                    it[
                                                                        3
                                                                    ].multi_index[
                                                                        0
                                                                    ],
                                                                    it[
                                                                        2
                                                                    ].multi_index[
                                                                        1
                                                                    ],
                                                                ]
                                                                amplitude = self.perm(
                                                                    matrix
                                                                )
                                                                if (
                                                                    abs(
                                                                        amplitude
                                                                    )
                                                                    >= cithresh
                                                                ):
                                                                    sd = self.get_slater_determinant(
                                                                        it[
                                                                            0
                                                                        ].multi_index,
                                                                        it[
                                                                            1
                                                                        ].multi_index,
                                                                        it[
                                                                            2
                                                                        ].multi_index,
                                                                        it[
                                                                            3
                                                                        ].multi_index,
                                                                    )
                                                                    if (
                                                                        amplitudestofile
                                                                        is True
                                                                    ):
                                                                        with open(
                                                                            filename,
                                                                            "a",
                                                                        ) as filea:
                                                                            filea.write(
                                                                                f"{sd} {amplitude: 20.16f}"
                                                                            )
                                                                            filea.write(
                                                                                "\n"
                                                                            )
                                                                    else:
                                                                        solution_dict.update(
                                                                            {
                                                                                sd: float(
                                                                                    amplitude
                                                                                )
                                                                            }
                                                                        )
                                                                matrix = np.identity(
                                                                    self.npairs
                                                                )
                if i == 4:
                    for ci in it[0]:
                        if abs(ci) >= (cithresh / excitationlevel):
                            it[1] = it[0].copy()
                            for ci2 in it[1]:
                                if (
                                    it[1].multi_index[0] > it[0].multi_index[0]
                                ) and (
                                    it[1].multi_index[1] > it[0].multi_index[1]
                                ):
                                    it[2] = it[1].copy()
                                    for ci3 in it[2]:
                                        if (
                                            it[2].multi_index[0]
                                            > it[1].multi_index[0]
                                        ) and (
                                            it[2].multi_index[1]
                                            > it[1].multi_index[1]
                                        ):
                                            it[3] = it[2].copy()
                                            for ci4 in it[3]:
                                                if (
                                                    it[3].multi_index[0]
                                                    > it[2].multi_index[0]
                                                ) and (
                                                    it[3].multi_index[1]
                                                    > it[2].multi_index[1]
                                                ):
                                                    it[4] = it[3].copy()
                                                    for ci5 in it[4]:
                                                        if (
                                                            it[4].multi_index[
                                                                0
                                                            ]
                                                            > it[
                                                                3
                                                            ].multi_index[0]
                                                        ) and (
                                                            it[4].multi_index[
                                                                1
                                                            ]
                                                            > it[
                                                                3
                                                            ].multi_index[1]
                                                        ):
                                                            matrix[
                                                                it[
                                                                    0
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    0
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = float(ci)
                                                            matrix[
                                                                it[
                                                                    1
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    1
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = float(ci2)
                                                            matrix[
                                                                it[
                                                                    2
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    2
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = float(ci3)
                                                            matrix[
                                                                it[
                                                                    3
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    3
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = float(ci4)
                                                            matrix[
                                                                it[
                                                                    4
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    4
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = float(ci5)

                                                            matrix[
                                                                it[
                                                                    0
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    1
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    0
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    1
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]
                                                            matrix[
                                                                it[
                                                                    0
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    2
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    0
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    2
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]
                                                            matrix[
                                                                it[
                                                                    0
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    3
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    0
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    3
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]
                                                            matrix[
                                                                it[
                                                                    0
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    4
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    0
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    4
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]

                                                            matrix[
                                                                it[
                                                                    1
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    0
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    1
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    0
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]
                                                            matrix[
                                                                it[
                                                                    1
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    2
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    1
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    2
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]
                                                            matrix[
                                                                it[
                                                                    1
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    3
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    1
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    3
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]
                                                            matrix[
                                                                it[
                                                                    1
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    4
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    1
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    4
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]

                                                            matrix[
                                                                it[
                                                                    2
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    0
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    2
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    0
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]
                                                            matrix[
                                                                it[
                                                                    2
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    1
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    2
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    1
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]
                                                            matrix[
                                                                it[
                                                                    2
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    3
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    2
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    3
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]
                                                            matrix[
                                                                it[
                                                                    2
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    4
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    2
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    4
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]

                                                            matrix[
                                                                it[
                                                                    3
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    0
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    3
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    0
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]
                                                            matrix[
                                                                it[
                                                                    3
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    1
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    3
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    1
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]
                                                            matrix[
                                                                it[
                                                                    3
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    2
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    3
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    2
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]
                                                            matrix[
                                                                it[
                                                                    3
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    4
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    3
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    4
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]

                                                            matrix[
                                                                it[
                                                                    4
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    0
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    4
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    0
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]
                                                            matrix[
                                                                it[
                                                                    4
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    1
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    4
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    1
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]
                                                            matrix[
                                                                it[
                                                                    4
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    2
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    4
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    2
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]
                                                            matrix[
                                                                it[
                                                                    4
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    3
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                            ] = coeff[
                                                                it[
                                                                    4
                                                                ].multi_index[
                                                                    0
                                                                ],
                                                                it[
                                                                    3
                                                                ].multi_index[
                                                                    1
                                                                ],
                                                            ]
                                                            amplitude = self.perm(
                                                                matrix
                                                            )
                                                            if (
                                                                abs(amplitude)
                                                                >= cithresh
                                                            ):
                                                                if (
                                                                    amplitudestofile
                                                                    is True
                                                                ):
                                                                    sd = self.get_slater_determinant(
                                                                        it[
                                                                            0
                                                                        ].multi_index,
                                                                        it[
                                                                            1
                                                                        ].multi_index,
                                                                        it[
                                                                            2
                                                                        ].multi_index,
                                                                        it[
                                                                            3
                                                                        ].multi_index,
                                                                        it[
                                                                            4
                                                                        ].multi_index,
                                                                    )
                                                                    with open(
                                                                        filename,
                                                                        "a",
                                                                    ) as filea:
                                                                        filea.write(
                                                                            f"{sd} {amplitude: 20.16f}"
                                                                        )
                                                                        filea.write(
                                                                            "\n"
                                                                        )
                                                                else:
                                                                    solution_dict.update(
                                                                        {
                                                                            sd: float(
                                                                                amplitude
                                                                            )
                                                                        }
                                                                    )
                                                            matrix = np.identity(
                                                                self.npairs
                                                            )
                if i == 5:
                    for ci in it[0]:
                        if abs(ci) >= (cithresh / excitationlevel):
                            it[1] = it[0].copy()
                            for ci2 in it[1]:
                                if (
                                    it[1].multi_index[0] > it[0].multi_index[0]
                                ) and (
                                    it[1].multi_index[1] > it[0].multi_index[1]
                                ):
                                    it[2] = it[1].copy()
                                    for ci3 in it[2]:
                                        if (
                                            it[2].multi_index[0]
                                            > it[1].multi_index[0]
                                        ) and (
                                            it[2].multi_index[1]
                                            > it[1].multi_index[1]
                                        ):
                                            it[3] = it[2].copy()
                                            for ci4 in it[3]:
                                                if (
                                                    it[3].multi_index[0]
                                                    > it[2].multi_index[0]
                                                ) and (
                                                    it[3].multi_index[1]
                                                    > it[2].multi_index[1]
                                                ):
                                                    it[4] = it[3].copy()
                                                    for ci5 in it[4]:
                                                        if (
                                                            it[4].multi_index[
                                                                0
                                                            ]
                                                            > it[
                                                                3
                                                            ].multi_index[0]
                                                        ) and (
                                                            it[4].multi_index[
                                                                1
                                                            ]
                                                            > it[
                                                                3
                                                            ].multi_index[1]
                                                        ):
                                                            it[5] = it[
                                                                4
                                                            ].copy()
                                                            for ci6 in it[5]:
                                                                if (
                                                                    it[
                                                                        5
                                                                    ].multi_index[
                                                                        0
                                                                    ]
                                                                    > it[
                                                                        4
                                                                    ].multi_index[
                                                                        0
                                                                    ]
                                                                ) and (
                                                                    it[
                                                                        5
                                                                    ].multi_index[
                                                                        1
                                                                    ]
                                                                    > it[
                                                                        4
                                                                    ].multi_index[
                                                                        1
                                                                    ]
                                                                ):
                                                                    matrix[
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = float(
                                                                        ci
                                                                    )
                                                                    matrix[
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = float(
                                                                        ci2
                                                                    )
                                                                    matrix[
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = float(
                                                                        ci3
                                                                    )
                                                                    matrix[
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = float(
                                                                        ci4
                                                                    )
                                                                    matrix[
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = float(
                                                                        ci5
                                                                    )
                                                                    matrix[
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = float(
                                                                        ci6
                                                                    )

                                                                    matrix[
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]

                                                                    matrix[
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]

                                                                    matrix[
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]

                                                                    matrix[
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]

                                                                    matrix[
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]

                                                                    matrix[
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            0
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            1
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            2
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            3
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]
                                                                    matrix[
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                    ] = coeff[
                                                                        it[
                                                                            5
                                                                        ].multi_index[
                                                                            0
                                                                        ],
                                                                        it[
                                                                            4
                                                                        ].multi_index[
                                                                            1
                                                                        ],
                                                                    ]

                                                                    amplitude = self.perm(
                                                                        matrix
                                                                    )
                                                                    if (
                                                                        abs(
                                                                            amplitude
                                                                        )
                                                                        >= cithresh
                                                                    ):
                                                                        sd = self.get_slater_determinant(
                                                                            it[
                                                                                0
                                                                            ].multi_index,
                                                                            it[
                                                                                1
                                                                            ].multi_index,
                                                                            it[
                                                                                2
                                                                            ].multi_index,
                                                                            it[
                                                                                3
                                                                            ].multi_index,
                                                                            it[
                                                                                4
                                                                            ].multi_index,
                                                                            it[
                                                                                5
                                                                            ].multi_index,
                                                                        )
                                                                        if (
                                                                            amplitudestofile
                                                                            is True
                                                                        ):
                                                                            with open(
                                                                                filename,
                                                                                "a",
                                                                            ) as filea:
                                                                                filea.write(
                                                                                    f"{sd} {amplitude: 20.16f}"
                                                                                )
                                                                                filea.write(
                                                                                    "\n"
                                                                                )
                                                                        else:
                                                                            solution_dict.update(
                                                                                {
                                                                                    sd: float(
                                                                                        amplitude
                                                                                    )
                                                                                }
                                                                            )
                                                                    matrix = np.identity(
                                                                        self.npairs
                                                                    )
                if i == 6:
                    for ci in it[0]:
                        if abs(ci) >= (cithresh / excitationlevel):
                            it[1] = it[0].copy()
                            for ci2 in it[1]:
                                if (
                                    it[1].multi_index[0] > it[0].multi_index[0]
                                ) and (
                                    it[1].multi_index[1] > it[0].multi_index[1]
                                ):
                                    it[2] = it[1].copy()
                                    for ci3 in it[2]:
                                        if (
                                            it[2].multi_index[0]
                                            > it[1].multi_index[0]
                                        ) and (
                                            it[2].multi_index[1]
                                            > it[1].multi_index[1]
                                        ):
                                            it[3] = it[2].copy()
                                            for ci4 in it[3]:
                                                if (
                                                    it[3].multi_index[0]
                                                    > it[2].multi_index[0]
                                                ) and (
                                                    it[3].multi_index[1]
                                                    > it[2].multi_index[1]
                                                ):
                                                    it[4] = it[3].copy()
                                                    for ci5 in it[4]:
                                                        if (
                                                            it[4].multi_index[
                                                                0
                                                            ]
                                                            > it[
                                                                3
                                                            ].multi_index[0]
                                                        ) and (
                                                            it[4].multi_index[
                                                                1
                                                            ]
                                                            > it[
                                                                3
                                                            ].multi_index[1]
                                                        ):
                                                            it[5] = it[
                                                                4
                                                            ].copy()
                                                            for ci6 in it[5]:
                                                                if (
                                                                    it[
                                                                        5
                                                                    ].multi_index[
                                                                        0
                                                                    ]
                                                                    > it[
                                                                        4
                                                                    ].multi_index[
                                                                        0
                                                                    ]
                                                                ) and (
                                                                    it[
                                                                        5
                                                                    ].multi_index[
                                                                        1
                                                                    ]
                                                                    > it[
                                                                        4
                                                                    ].multi_index[
                                                                        1
                                                                    ]
                                                                ):
                                                                    it[6] = it[
                                                                        5
                                                                    ].copy()
                                                                    for (
                                                                        ci7
                                                                    ) in it[6]:
                                                                        if (
                                                                            it[
                                                                                6
                                                                            ].multi_index[
                                                                                0
                                                                            ]
                                                                            > it[
                                                                                5
                                                                            ].multi_index[
                                                                                0
                                                                            ]
                                                                        ) and (
                                                                            it[
                                                                                6
                                                                            ].multi_index[
                                                                                1
                                                                            ]
                                                                            > it[
                                                                                5
                                                                            ].multi_index[
                                                                                1
                                                                            ]
                                                                        ):
                                                                            matrix[
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = float(
                                                                                ci
                                                                            )
                                                                            matrix[
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = float(
                                                                                ci2
                                                                            )
                                                                            matrix[
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = float(
                                                                                ci3
                                                                            )
                                                                            matrix[
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = float(
                                                                                ci4
                                                                            )
                                                                            matrix[
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = float(
                                                                                ci5
                                                                            )
                                                                            matrix[
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = float(
                                                                                ci6
                                                                            )
                                                                            matrix[
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = float(
                                                                                ci7
                                                                            )

                                                                            matrix[
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]

                                                                            matrix[
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]

                                                                            matrix[
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]

                                                                            matrix[
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]

                                                                            matrix[
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]

                                                                            matrix[
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]

                                                                            matrix[
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    0
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    1
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    2
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    3
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    4
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            matrix[
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                            ] = coeff[
                                                                                it[
                                                                                    6
                                                                                ].multi_index[
                                                                                    0
                                                                                ],
                                                                                it[
                                                                                    5
                                                                                ].multi_index[
                                                                                    1
                                                                                ],
                                                                            ]
                                                                            amplitude = self.perm(
                                                                                matrix
                                                                            )
                                                                            if (
                                                                                abs(
                                                                                    amplitude
                                                                                )
                                                                                >= cithresh
                                                                            ):
                                                                                if (
                                                                                    amplitudestofile
                                                                                    is True
                                                                                ):
                                                                                    sd = self.get_slater_determinant(
                                                                                        it[
                                                                                            0
                                                                                        ].multi_index,
                                                                                        it[
                                                                                            1
                                                                                        ].multi_index,
                                                                                        it[
                                                                                            2
                                                                                        ].multi_index,
                                                                                        it[
                                                                                            3
                                                                                        ].multi_index,
                                                                                        it[
                                                                                            4
                                                                                        ].multi_index,
                                                                                        it[
                                                                                            5
                                                                                        ].multi_index,
                                                                                        it[
                                                                                            6
                                                                                        ].multi_index,
                                                                                    )
                                                                                    with open(
                                                                                        filename,
                                                                                        "a",
                                                                                    ) as filea:
                                                                                        filea.write(
                                                                                            f"{sd} {amplitude: 20.16f}"
                                                                                        )
                                                                                        filea.write(
                                                                                            "\n"
                                                                                        )
                                                                                else:
                                                                                    solution_dict.update(
                                                                                        {
                                                                                            sd: float(
                                                                                                amplitude
                                                                                            )
                                                                                        }
                                                                                    )
                                                                            matrix = np.identity(
                                                                                self.npairs
                                                                            )
        from collections import OrderedDict

        sorted_pairs = sorted(
            solution_dict.items(), key=lambda k: abs(k[1]), reverse=True
        )
        ordered_dict = OrderedDict(sorted_pairs)
        for key, value in ordered_dict.items():
            log(f"{key} {value:20.16f}")

    def get_slater_determinant(self, *indices):
        """Return excited Slater Determinant.

           **Arguments:**

           indices
                A (list of) multi-index. First element contains occupied index
                w.r.t. reference determinant, second element contains virtual
                index w.r.t. reference determinant.
        """
        orb_ref = []
        excited = []
        for ind in indices:
            orb_ref.append(ind[0])
            excited.append(ind[1] + self.npairs)

        sd = []
        for i in range(self.nbasis):
            if i in orb_ref:
                sd.append(0)
            elif i in excited:
                sd.append(2)
            else:
                if i < self.npairs:
                    sd.append(2)
                else:
                    sd.append(0)

        return str(sd).translate(str.maketrans("", "", ", "))

    def print_final(self, s="Final"):
        """Print energies

           **Optional arguments:**

           s
                A string.
        """
        log.hline("-")
        log(
            f"{s+' reference Energy:':<30} {self.compute_reference_energy():> 16.12f}"
        )
        log(
            f"{s+' correlation Energy:':<30} {self.compute_correlation_energy():> 16.12f}"
        )
        log.hline("=")
        log(f"{s+' total Energy:':<30} {self.compute_total_energy():> 16.12f}")
        log.hline("=")

    def dump_final(
        self,
        orb,
        olp,
        printoptions,
        dumpci,
        checkpoint,
        checkpoint_fn="checkpoint_pccd.h5",
    ):
        """Dump final solution (orbitals, wavefunction amplitudes, geminal
           coefficients)

           **Arguments:**

           orb
                An expansion instance. AO/MO coefficients stored to disk

           olp
                A TwoIndex instance. The AO overlap matrix. Required for restart

           printoptions, dumpci
                A dictionary. See :py:meth:`RpCCD.solve_scf`

           checkpoint:
                An integer. See :py:meth:`RpCCD.solve_scf`

           checkpoint_fn
                The filename of the checkpoint file.
        """
        if checkpoint > 0:
            if log.do_medium:
                log(
                    f"Writing pCCD output to file {filemanager.result_dir}/"
                    f"{checkpoint_fn}"
                )
            self.checkpoint.to_file(checkpoint_fn)
            if log.do_medium:
                log.hline("-")
                log(
                    f"Final solution for coefficients: (only printed if |c_i| > "
                    f"{printoptions['ci']})"
                )
        self.print_solution(
            printoptions["ci"],
            printoptions["excitationlevel"],
            amplitudestofile=dumpci["amplitudestofile"],
            fname=dumpci["amplitudesfilename"],
        )
        if printoptions["geminal"]:
            log.hline("-")
            log("Geminal matrix: (rows=occupied, columns=virtual)")
            self.print_geminal_coefficients()
        if log.do_medium:
            log.hline("-")
            log("Natural occupation numbers (for spatial orbitals):")
            self.print_occupations(orb)
            log.hline("=")

    def print_geminal_coefficients(self):
        """Print geminal coefficients
        """
        log.hline("")
        s = "".join(f"{(i + 1 + self.npairs):11}" for i in range(self.nvirt))
        log(f"  {s}")
        log.hline("")
        for line in range(self.npairs):
            s = f"{(line + 1):>4}:"
            for row in range(self.nvirt):
                s += f"{self.geminal.get_element(line, row):> 11.6f}"
            log(s)

    def print_occupations(self, orb):
        """Print occupation numbers
        """
        # Print at most 6 in one line
        s = ""
        for i, el in zip(range(self.nbasis), orb.occupations):
            if i % 5 == 0 and i != 0:
                log(s)
                s = ""
            s += f"{(i + 1):>6}: {(el * 2.0): 7.6f}"
        log(s)

    def check_keywords(
        self, guess, solver, maxiter, dumpci, thresh, printoptions
    ):
        """Check dictionaries if they contain proper keys.

           **Arguments:**

           guess, solver, maxiter, dumpci, thresh, printoptions
                See :py:meth:`RpCCD.solve`

        """
        #
        # Check guess, values are checked separately
        #
        for key in guess:
            check_options("guess", key, "type", "factor", "geminal")
        #
        # Check solver
        #
        for key in solver:
            check_options("solver", key, "wfn", "lagrange")
        #
        # Check maxiter
        #
        for key, value in maxiter.items():
            check_options("maxiter", key, "wfniter")
            check_type("maxiter", value, int)
        #
        # Check thresh
        #
        for key, value in thresh.items():
            check_options("thresh", key, "wfn")
            check_type("thresh", value, float)
            if value < 0:
                raise ValueError(
                    f"Negative convergence threshold for {key} is not allowed!"
                )
        #
        # Check printoptions
        #
        for key, value in printoptions.items():
            check_options(
                "printoptions", key, "geminal", "ci", "excitationlevel"
            )
            if key == "geminal":
                check_options("printoptions.geminal", value, False, True, 0, 1)
            elif key == "ci":
                check_type("printoptions.ci", value, float)
            elif key == "excitationlevel":
                check_type("printoptions.excitationlevel", value, int)
        #
        # Check dumpci
        #
        for key, value in dumpci.items():
            check_options(
                "dumpci", key, "amplitudestofile", "amplitudesfilename"
            )
            if key == "amplitudestofile":
                check_options(
                    "dumpci.amplitudestofile", value, False, True, 0, 1
                )
            if key == "amplitudesfilename":
                check_type("dumpci.amplitudesfilename", value, str)

    def check_keywords_scf(
        self, guess, solver, maxiter, dumpci, thresh, printoptions, stepsearch
    ):
        """Check dictionaries if they contain proper keys.

           **Arguments:**

           guess, solver, maxiter, dumpci, thresh, printoptions, stepseach
                See :py:meth:`RpCCD.solve_scf`

        """
        #
        # Check guess, values are checked separately
        #
        for key in guess:
            check_options(
                "guess", key, "type", "factor", "geminal", "lagrange"
            )
        #
        # Check solver
        #
        for key in solver:
            check_options("solver", key, "wfn", "lagrange")
        #
        # Check maxiter
        #
        for key, value in maxiter.items():
            check_options("maxiter", key, "wfniter", "orbiter")
            check_type("maxiter", value, int)
        #
        # Check thresh
        #
        for key, value in thresh.items():
            check_options(
                "thresh", key, "wfn", "energy", "gradientnorm", "gradientmax"
            )
            check_type("thresh", value, float)
            if value < 0:
                raise ValueError(
                    f"Negative convergence threshold for {key} is not allowed!"
                )
        #
        # Check printoptions
        #
        for key, value in printoptions.items():
            check_options(
                "printoptions", key, "geminal", "ci", "excitationlevel"
            )
            if key == "geminal":
                check_options("printoptions.geminal", value, False, True, 0, 1)
            elif key == "ci":
                check_type("printoptions.ci", value, float)
            elif key == "excitationlevel":
                check_type("printoptions.excitationlevel", value, int)
        #
        # Check dumpci
        #
        for key, value in dumpci.items():
            check_options(
                "dumpci", key, "amplitudestofile", "amplitudesfilename"
            )
            if key == "amplitudestofile":
                check_options(
                    "dumpci.amplitudestofile", value, False, True, 0, 1
                )
            if key == "amplitudesfilename":
                check_type("dumpci.amplitudesfilename", value, str)
        #
        # Check stepsearch, values are checked separately
        #
        for key in stepsearch:
            check_options(
                "stepsearch",
                key,
                "method",
                "optimizer",
                "alpha",
                "c1",
                "minalpha",
                "maxiterouter",
                "maxiterinner",
                "maxeta",
                "mineta",
                "upscale",
                "downscale",
                "trustradius",
                "maxtrustradius",
                "threshold",
            )

    def print_options(
        self, guess, solver, maxiter, thresh, printoptions, indextrans
    ):
        """Print optimization options.

           **Arguments:**

           guess, solver, maxiter, thresh, printoptions, indextrans
                See :py:meth:`RpCCD.solve`

        """
        log.hline()
        log(" ")
        log("Entering pCCD optimization (restricted, closed-shell):")
        log(" ")
        log.hline()
        log("OPTIMIZATION PARAMETERS:")
        log(f"Number of pairs:             {self.npairs}")
        log(f"Number of virtuals:          {self.nvirt}")
        log(f"4-index transformation:      {indextrans}")
        log("Initial guess:")
        log(f"  type:                        {guess['type']}")
        log(f"  scaling factor:              {guess['factor']:2.3f}")
        log("Solvers:")
        log(f"  wavefunction amplitudes:   {solver['wfn']}")
        log("Optimization thresholds:")
        log(f"  wavefunction:              {thresh['wfn']:1.2e}")
        log("Printing options:")
        log(f"  c_i:                       {printoptions['ci']:1.2e}")
        log(f"  excitation level:          {printoptions['excitationlevel']}")

    def print_options_scf(
        self,
        guess,
        solver,
        maxiter,
        lshift,
        stepsearch,
        thresh,
        printoptions,
        checkpoint,
        checkpoint_fn,
        indextrans,
        orbitaloptimizer,
        sort,
    ):
        """Print optimization options.

           **Arguments:**

           See :py:meth:`RpCCD.solve_scf`

        """
        if log.do_medium:
            log.hline()
            log(" ")
            log("Entering OOpCCD optimization (restricted, closed-shell):")
            log(" ")
            log.hline()
            log("OPTIMIZATION PARAMETERS:")
            log(f"Number of pairs:               {self.npairs}")
            log(f"Number of virtuals:            {self.nvirt}")
            log("Initial guess:")
            log(f"  type:                        {guess['type']}")
            log(f"  scaling factor:              {guess['factor']:2.3f}")
            log("Solvers:")
            log(f"  wavefunction amplitudes:     {solver['wfn']}")
            log(f"  Lagrange multipliers:        {solver['lagrange']}")
            log(f"  orbital optimization:        {orbitaloptimizer}")
            log(f"Number of iterations:          {maxiter['orbiter']}")
            log(f"Checkpointing:                 {checkpoint}")
            log(f"Checkpoint file:               {checkpoint_fn}")
            log(f"4-index transformation:        {indextrans}")
            log(f"Level shift:                   {lshift:3.3e}")
            log(f"Sorting natural orbitals:      {sort}")
            if stepsearch["method"] == "trust-region":
                log("Apply trust region:")
                log(
                    f"  initial trust radius:        {stepsearch['trustradius']:1.3f}"
                )
                log(
                    f"  maximum turst radius:        {stepsearch['maxtrustradius']:2.3f}"
                )
                log(
                    f"  upper trust ratio bound:     {stepsearch['maxeta']:1.2e}"
                )
                log(
                    f"  lower trust ratio bound:     {stepsearch['mineta']:1.2e}"
                )
                log(
                    f"  upper scaling factor:        {stepsearch['upscale']:1.2e}"
                )
                log(
                    f"  lower scaling factor:        {stepsearch['downscale']:1.2e}"
                )
                log(
                    f"  max number of iterations:    {stepsearch['maxiterouter']}"
                )
                log(
                    f"  max number of optimizations: {stepsearch['maxiterinner']}"
                )
                log(
                    f"  optimization threshold:      {stepsearch['threshold']:1.2e}"
                )
                log(
                    f"  optimizer:                   {stepsearch['optimizer']}"
                )
            elif stepsearch["method"] == "backtracking":
                log("Apply line search:")
                log(f"  line search method:          {stepsearch['method']}")
                log(
                    f"  initial scaling factor:      {stepsearch['alpha']:1.3f}"
                )
                log(
                    f"  contraction factor:          {stepsearch['downscale']:1.3f}"
                )
                log(f"  c1 factor:                   {stepsearch['c1']:1.3e}")
                log(
                    f"  minimum scaling factor:      {stepsearch['minalpha']:1.3e}"
                )
            else:
                log(f"No step search selected:")
                log(
                    f"  scaling factor:              {stepsearch['alpha']:1.3f}"
                )
            log("Optimization thresholds:")
            log(f"  wavefunction:                {thresh['wfn']:1.2e}")
            log(f"  energy:                      {thresh['energy']:1.2e}")
            log(f"  gradient:                    {thresh['gradientmax']:1.2e}")
            log(
                f"  gradient norm:               {thresh['gradientnorm']:1.2e}"
            )
            log("Printing options:")
            log(f"  c_i:                         {printoptions['ci']:1.2e}")
            log(
                f"  excitation level:            {printoptions['excitationlevel']}"
            )

    def compute_objective_function(self, coeff=None):
        """Objective function for line search optimization

           **Optional arguments:**

           coeff
                See :py:meth:`RpCCD.compute_total_energy`
        """
        return self.compute_total_energy(coeff)

    #
    # compute approximate overlap:
    #
    def compute_overlap(self):
        """Compute approximate overlap of pCCD
        """
        return 1 + self.geminal.contract(
            "ab,ab", self.geminal, select="einsum"
        )


class RpCCD(RpCCDBase):
    """Restricted pCCD wavefunction class"""

    @timer.with_section("pCCD")
    def solve(self, one, two, olp, orb, *args, **kwargs):
        """Optimize pCCD coefficients for some Hamiltonian.
           For restricted, closed-shell pCCD.

           **Arguments:**

           one, two
                One- (TwoIndex instance) and two-body integrals (FourIndex or
                Cholesky instance) (some Hamiltonian matrix elements)

           orb
                An expansion instance. It contains the MO coefficients.

           olp
                The AO overlap matrix. A TwoIndex instance.

           *args
                May contain an IOData instance

           **Keywords:**

            :indextrans: 4-index Transformation (str). Choice between
                         ``tensordot`` (default) or ``einsum`` or ``cpp``
            :warning: Print warnings (boolean) (default False)
            :guess: initial guess (dictionary) containing:

                    * type: guess type (str). One of ``mp2`` (MP2 pair
                      amplitudes, default), ``random`` (random numbers),
                      ``const`` (a constant scaled by a factor)
                    * factor: a scaling factor (float) (default -0.1)
                    * geminal: external guess for geminal coefficients
                      (1-d np.array); if provided, 'type' is
                      ignored (default None)

            :solver: wfn solver (dictionary) containing:

                     * wfn: wavefunction solver (str) (default ``krylov``)

            :maxiter: max number of iterations (dictionary) containing:

                      * wfniter: maximum number of iterations (int) for wfn
                        solver (default 200)

            :dumpci: dump ci coefficients (dictionary):

                     * amplitudestofile: write wfn amplitudes to file
                       (boolean) (default False)
                     * amplitudesfilename: (str) (default pccd_amplitudes.dat)

            :thresh: optimization thresholds (dictionary) containing:

                     * wfn: threshold for geminal coefficients (float)
                       (default 1e-12)

            :printoptions: print level (dictionary) containing:

                           * geminal: (boolean), if True, geminal matrix is
                             printed (default True)

                           * ci: threshold for CI coefficients (float) (requires
                             evaluation of a permanent), all coefficients
                             (for a given excitation order) larger than
                             'ci' are printed (default 0.05)
                           * excitationlevel: number of excited pairs w.r.t.
                             reference determinant for which wfn amplitudes
                             are reconstructed (int) (default 1)
        """
        if log.do_medium:
            log.hline("=")
            log("Entering pCCD module")

        #
        # Assign keyword arguements
        #
        names = []

        def _helper(x, y):
            names.append(x)
            return kwargs.get(x, y)

        indextrans = _helper("indextrans", "tensordot")
        warning = _helper("warning", False)
        guess = _helper("guess", {})
        guess.setdefault("type", "mp2")
        guess.setdefault("factor", -0.1)
        guess.setdefault("geminal", None)
        solver = _helper("solver", {})
        solver.setdefault("wfn", "krylov")
        maxiter = _helper("maxiter", {"wfniter": 200})
        dumpci = _helper("dumpci", {})
        dumpci.setdefault("amplitudestofile", False)
        dumpci.setdefault("amplitudesfilename", "pccd_amplitudes.dat")
        thresh = _helper("thresh", {"wfn": 1e-12})
        printoptions = _helper("printoptions", {})
        printoptions.setdefault("geminal", False)
        printoptions.setdefault("ci", 0.05)
        printoptions.setdefault("excitationlevel", 1)
        e_core = _helper("e_core", None)
        #
        # orb_a is NOT assigned here; check for allowed keywords only
        #
        orb_a = _helper("orb_a", None)

        #
        # Check kwargs
        #
        for name, value in kwargs.items():
            if name not in names:
                raise ArgumentError(f"Unknown keyword argument {name}")

        #
        # Check dictionaries in keyword arguments
        #
        self.check_keywords(
            guess, solver, maxiter, dumpci, thresh, printoptions
        )
        check_options("warning", warning, False, True, 0, 1)

        if not warning:
            warnings.filterwarnings("ignore")

        #
        # Print optimization parameters
        #
        if log.do_medium:
            self.print_options(
                guess, solver, maxiter, thresh, printoptions, indextrans
            )

        #
        # Overwrite core energy using unmask function if not found in kwargs
        #
        if e_core is None:
            # Check IOData for e_core
            e_core = unmask("e_core", *args, **kwargs)
            if e_core is None:
                raise ArgumentError("Cannot find core energy in arguments.")
        self.e_core = e_core

        #
        # Generate Guess
        #
        if guess["geminal"] is None:
            if guess["type"] == "mp2":
                #
                # Transform integrals into MO basis
                #
                if self.ncore > 0:
                    cas = split_core_active(
                        one,
                        two,
                        orb,
                        e_core=self.e_core,
                        ncore=self.ncore,
                        nactive=self.nbasis,
                        indextrans=indextrans,
                    )
                    one_mo = cas.one
                    two_mo = cas.two
                    self.e_core = cas.e_core
                else:
                    ti = transform_integrals(
                        one, two, orb, indextrans=indextrans
                    )
                    one_mo = ti.one[0]
                    two_mo = ti.two[0]
                initial_guess = [
                    self.generate_guess(guess, mo1=one_mo, mo2=two_mo)
                ]
                one_mo = None
                two_mo = None
            else:
                initial_guess = [self.generate_guess(guess)]
            #
            # Pass dummy guess
            #
            initial_guess.append(initial_guess[0])
        else:
            check_type("guess.geminal", guess["geminal"], np.ndarray)
            if not guess["geminal"].shape[0] == self.dimension:
                raise ConsistencyError(
                    "Size of geminal guess array does not match number of unknowns."
                )
            initial_guess = [guess["geminal"]]
            #
            # Pass dummy guess
            #
            initial_guess.append(initial_guess[0])
        #
        # Update IOData container
        #
        self.checkpoint.update("e_core", self.e_core)
        self.checkpoint.update("orb_a", orb.copy())
        self.checkpoint.update("olp", olp)

        #
        # Solve for wavefunction
        #
        self.solve_model(
            one,
            two,
            orb,
            **{
                "maxiter": maxiter,
                "thresh": thresh,
                "guess": initial_guess[0],
                "guesslm": initial_guess[1],
                "solver": solver,
                "indextrans": indextrans,
                "orbitaloptimizer": False,
            },
        )

        #
        # Final print statements:
        #
        if log.do_medium:
            self.print_final()
            self.dump_final(orb, olp, printoptions, dumpci, -1)
        #
        # Sanity check for correlation energy:
        #
        if self.compute_correlation_energy() > 0:
            raise ValueError(
                "Warning: Final correlation energy is positive! Improve initial guess!"
            )

        e_tot = self.compute_total_energy()
        e_corr = self.compute_correlation_energy()
        e_ref = self.compute_reference_energy()
        e_dict = {"e_tot": e_tot, "e_corr": e_corr, "e_ref": e_ref}
        self.energy = e_dict
        #
        # Update IOData container
        #
        self.checkpoint.update("e_core", self.e_core)
        self.checkpoint.update("e_tot", e_tot)
        self.checkpoint.update("e_corr", e_corr)
        self.checkpoint.update("e_ref", e_ref)
        self.checkpoint.update("overlap", self.compute_overlap())

        #
        # Clean the Cache because we do not want to keep the auxiliary matrices in memory
        #
        self.clear_aux_matrix()
        #
        # Create final container as return value
        #
        return self.checkpoint()


class ROOpCCD(RpCCDBase):
    """Restricted orbital-optimzied pCCD wavefunction class"""

    @timer.with_section("pCCD-SCF")
    def solve(self, one, two, olp, orb, *args, **kwargs):
        """Find Geminal expansion coefficient for some Hamiltonian.
           For restricted, closed-shell pCCD.
           Perform orbital optimization.

           **Arguments:**

           one, two
                One- (TwoIndex instance) and two-body integrals (FourIndex or
                Cholesky instance) (some Hamiltonian matrix elements)

           orb
                An expansion instance. It contains the MO coefficients.

           olp
                The AO overlap matrix. A TwoIndex instance.

           *args
                May contain an IOData instance

           **Keywords:**

            :indextrans: 4-index Transformation (str). Choice between
                         ``tensordot`` (default) or ``einsum`` or ``cpp``
            :warning: print warnings (boolean) (default False)
            :guess: initial guess (dictionary) containing:

                    * type: guess type (str). One of ``mp2`` (MP2 pair
                      amplitudes, default), ``random`` (random numbers),
                      ``const`` (a constant scaled by a factor)
                    * factor: a scaling factor (float) (default -0.1)
                    * geminal: external guess for geminal coefficients
                      (1-dim np.array) (default None)
                    * lagrange: external guess for Lagrange multipliers
                      (1-dim np.array) (default None)

            :solver: wfn/Lagrange solver (dictionary) containing:

                     * wfn: wavefunction solver (str) (default ``krylov``)
                     * lagrange: Lagrange multiplier solver (str)
                       (default krylov)

            :maxiter: max number of iterations (dictionary) containing:

                      * wfniter: maximum number of iterations for
                                 wfn/lagrange solver (int) (default 200)
                      * orbiter: maximum number of orbital optimization
                                 steps (int) (default 100)
            :dumpci: dump ci coefficient (dictionary) containing:

                     * amplitudestofile: write wfn amplitudes to file
                       (boolean) (default False)
                     * amplitudesfilename: (str) (default pccd_amplitudes.dat)

            :thresh: optimization thresholds (dictionary) containing:

                     * wfn: threshold for geminal coefficients and
                       Lagrange multipliers (float) (defaulf 1e-12)
                     * energy: convergence threshold for energy (float)
                       (default 1e-8)
                     * gradientnorm: convergence threshold for norm of
                       orbital gradient (float) (default 1e-4)
                     * gradientmax: threshold for maximum absolute value of
                       orbital gradient (float) (default 5e-5)

            :printoptions: print level; dictionary containing:

                           * geminal: (boolean) if True, geminal matrix is
                             printed (default True)
                           * ci: threshold for CI coefficients (requires
                             evaluation of a permanent) (float). All
                             coefficients (for a given excitation order)
                             larger than ci are printed (default 0.01)
                           * excitationlevel: number of excited pairs w.r.t.
                             reference determinant for which
                             wfn amplitudes are reconstructed
                             (int) (default 1)

            :stepsearch: step search options (dictionary) containing:

                         * method: step search method used (str). One of
                           ``trust-region`` (default), ``None``,
                           ``backtracking``
                         * alpha: scaling factor for Newton step (float),
                           used in ``backtracking`` and ``None`` method (default
                           1.00)
                         * c1: parameter used in ``backtracking`` (float)
                           (default 1e-4)
                         * minalpha: minimum scaling factor used in
                           ``backracking`` (float) (default 1e-6)
                         * maxiterouter: maximum number of search steps
                           (int) (default 10)
                         * maxiterinner: maximum number of optimization
                           step in each search step (int) (used only in ``pcg``,
                           default 500)
                         * maxeta: upper bound for estimated vs actual
                           change in ``trust-region`` (float) (default 0.75)
                         * mineta: lower bound for estimated vs actual change in
                           ``trust-region`` (float) (default 0.25)
                         * upscale: scaling factor to increase trustradius
                           in ``trust-region`` (float) (default 2.0)
                         * downscale: scaling factor to decrease trustradius
                           in ``trust-region`` and step length in
                           ``backtracking`` (float) (default 0.25)
                         * trustradius: initial trustradius (float) (default
                           0.75)
                         * maxtrustradius: maximum trustradius (float) (default
                           0.75)
                         * threshold: trust-region optimization threshold, only
                           used in ``pcg`` (float) (default 1e-8)
                         * optimizer: optimizes step to boundary of trustradius
                           (str). One of ``pcg``, ``dogleg``, ``ddl`` (default
                           ddl)

            :checkpoint: frequency of checkpointing (int). If > 0, writes
                         orbitals and overlap to a checkpont file (defatul 1)
            :checkpoint_fn: filename to use for the checkpoint file (default
                            "checkpoint_pccd.h5")
            :levelshift: level shift of Hessian (float) (default 1e-8)
            :absolute: (boolean), if True, take absolute values of Hessian
                       (default False)
            :sort: (boolean), if True, orbitals are sorted according to their
                   natural occupation numbers. This requires us to solve for
                   the wavefunction again. Works only if orbitaloptimizer
                   is set to ``variational``. (default True)
            :orbitaloptimizer: (str) switch between variational orbital
                               optimization (``variational``) and PS2c
                               (``ps2c``) (default ``variational``).
        """
        if log.do_medium:
            log.hline("=")
            log("Entering orbital-optimized pCCD module")

        #
        # Assign keyword arguements, checks only name of dictionary, keys and
        # values are checked below.
        #
        names = []

        def _helper(x, y):
            names.append(x)
            return kwargs.get(x, y)

        indextrans = _helper("indextrans", "tensordot")
        warning = _helper("warning", False)
        checkpoint = _helper("checkpoint", 1)
        checkpoint_fn = _helper("checkpoint_fn", "checkpoint_pccd.h5")
        lshift = _helper("levelshift", 1e-8)
        pos = _helper("absolute", False)
        sort = _helper("sort", True)
        guess = _helper("guess", {})
        guess.setdefault("type", "mp2")
        guess.setdefault("factor", -0.1)
        guess.setdefault("geminal", None)
        guess.setdefault("lagrange", None)
        solver = _helper("solver", {})
        solver.setdefault("wfn", "krylov")
        solver.setdefault("lagrange", "krylov")
        maxiter = _helper("maxiter", {})
        maxiter.setdefault("wfniter", 200)
        maxiter.setdefault("orbiter", 100)
        dumpci = _helper("dumpci", {})
        dumpci.setdefault("amplitudestofile", False)
        dumpci.setdefault("amplitudesfilename", "./pccd_amplitudes.dat")
        thresh = _helper("thresh", {})
        thresh.setdefault("wfn", 1e-12)
        thresh.setdefault("energy", 1e-8)
        thresh.setdefault("gradientnorm", 1e-4)
        thresh.setdefault("gradientmax", 5e-5)
        printoptions = _helper("printoptions", {})
        printoptions.setdefault("geminal", False)
        printoptions.setdefault("ci", 0.05)
        printoptions.setdefault("excitationlevel", 1)
        stepsearch = _helper("stepsearch", {})
        stepsearch.setdefault("method", "trust-region")
        stepsearch.setdefault("alpha", 1.0)
        stepsearch.setdefault("c1", 0.0001)
        stepsearch.setdefault("minalpha", 1e-6)
        stepsearch.setdefault("maxiterouter", 10)
        stepsearch.setdefault("maxiterinner", 500)
        stepsearch.setdefault("maxeta", 0.75)
        stepsearch.setdefault("mineta", 0.25)
        stepsearch.setdefault("upscale", 2.0)
        stepsearch.setdefault("downscale", 0.25)
        stepsearch.setdefault("trustradius", 0.75)
        stepsearch.setdefault("maxtrustradius", 0.75)
        stepsearch.setdefault("threshold", 1e-8)
        stepsearch.setdefault("optimizer", "ddl")
        orbitaloptimizer = _helper("orbitaloptimizer", "variational")
        restart = _helper("restart", False)
        e_core = _helper("e_core", None)
        #
        # orb_a is NOT assigned here; check for allowed keywords only
        #
        orb_a = _helper("orb_a", None)

        for name, value in kwargs.items():
            if name not in names:
                raise ArgumentError(f"Unknown keyword argument {name}")

        #
        # Check dictionaries in keyword arguments
        #
        self.check_keywords_scf(
            guess, solver, maxiter, dumpci, thresh, printoptions, stepsearch
        )
        check_options("warning", warning, False, True, 0, 1)
        check_type("checkpoint", checkpoint, int)
        check_type("checkpoint_fn", checkpoint_fn, str)
        check_type("levelshift", lshift, float, int)
        check_options("absolute", pos, False, True, 0, 1)
        check_options("sort", sort, True, False, 0, 1)

        #
        # Set optimization parameters
        #
        if not warning:
            warnings.filterwarnings("ignore")
        if maxiter["orbiter"] < 0:
            raise ValueError("Number of iterations must be greater/equal 0!")

        #
        # Print optimization parameters
        #
        if log.do_medium:
            self.print_options_scf(
                guess,
                solver,
                maxiter,
                lshift,
                stepsearch,
                thresh,
                printoptions,
                checkpoint,
                checkpoint_fn,
                indextrans,
                orbitaloptimizer,
                sort,
            )

        #
        # Overwrite core energy using unmask function if not found in kwargs
        #
        if e_core is None:
            # Check IOData for e_core
            e_core = unmask("e_core", *args, **kwargs)
            if e_core is None and not restart:
                raise ArgumentError("Cannot find core energy in arguments.")
        self.e_core = e_core

        #
        # Generate Guess [geminal, lagrange]
        #
        if restart:
            from pybest.io.iodata import IOData

            data = IOData.from_file(restart)
            if hasattr(data, "t_p"):
                guess["geminal"] = data.t_p._array.ravel()
            if hasattr(data, "l_p"):
                guess["lagrange"] = data.l_p._array.ravel()
            if hasattr(data, "olp") and hasattr(data, "orb_a"):
                if isinstance(olp, TwoIndex) and isinstance(orb, Orbital):
                    project_orbitals(data.olp, olp, data.orb_a, orb)
                else:
                    olp = data.olp
                    orb = data.orb_a
            if hasattr(data, "e_core"):
                self.e_core = unmask("e_core", data, **kwargs)
            if self.e_core is None:
                raise ArgumentError("Cannot find core energy in arguments.")
        if guess["geminal"] is None:
            if guess["type"] == "mp2":
                #
                # Transform integrals into MO basis
                #
                if self.ncore > 0:
                    cas = split_core_active(
                        one,
                        two,
                        orb,
                        e_core=self.e_core,
                        ncore=self.ncore,
                        nactive=self.nbasis,
                        indextrans=indextrans,
                    )
                    one_mo = cas.one
                    two_mo = cas.two
                    self.e_core = cas.e_core
                else:
                    ti = transform_integrals(
                        one, two, orb, indextrans=indextrans
                    )
                    one_mo = ti.one[0]
                    two_mo = ti.two[0]
                initial_guess = [
                    self.generate_guess(guess, mo1=one_mo, mo2=two_mo)
                ]
                one_mo = None
                two_mo = None
            else:
                initial_guess = [self.generate_guess(guess)]
        else:
            check_type("guess.geminal", guess["geminal"], np.ndarray)
            if not guess["geminal"].shape[0] == self.dimension:
                raise ValueError(
                    "Size of geminal guess array does not match number of unknowns."
                )
            initial_guess = [guess["geminal"]]
        if guess["lagrange"] is None:
            if guess["type"] == "mp2":
                initial_guess.append(initial_guess[0])
            else:
                initial_guess.append(self.generate_guess(guess))
        else:
            check_type("guess.lagrange", guess["lagrange"], np.ndarray)
            if not guess["lagrange"].shape[0] == self.dimension:
                raise ValueError(
                    "Size of Lagrange guess array does not match number of unknowns."
                )
            initial_guess.append(guess["lagrange"])

        # create a copy, this adds more flexibility during calculations and
        # the olp files are typically very small
        self.checkpoint.update("olp", olp.copy())
        self.checkpoint.update("orb_a", orb.copy())
        #
        # First iteration:
        #
        self.solve_model(
            one,
            two,
            orb,
            **{
                "maxiter": maxiter,
                "thresh": thresh,
                "guess": initial_guess[0],
                "guesslm": initial_guess[1],
                "solver": solver,
                "indextrans": indextrans,
                "orbitaloptimizer": orbitaloptimizer,
            },
        )

        #
        # Update total/corr energy
        #
        e_tot = self.compute_total_energy()
        e_corr = self.compute_correlation_energy()
        e_ref = self.compute_reference_energy()
        e_dict = {"e_tot": e_tot, "e_corr": e_corr, "e_ref": e_ref}
        self.energy = e_dict
        #
        # Update IOData container
        #
        self.checkpoint.update("e_core", self.e_core)
        self.checkpoint.update("e_tot", e_tot)
        self.checkpoint.update("e_corr", e_corr)
        self.checkpoint.update("e_ref", e_ref)

        if log.do_medium:
            log.hline("~")
            log("Initial step:")
            self.print_final("Initial")
            log("Entering orbital optimization of pCCD...")
            log.hline(" ")
            if stepsearch["method"] == "trust-region":
                log(
                    f"{'step':>6} {'Etot':>10} {'D(Etot)':>14} {'Ecorr':>12} "
                    f"{'D(Ecorr)':>14} {'Max(Grad)':>12} {'|Grad|':>8} "
                    f"{'TrustRegion':>14}"
                )
            else:
                f"{'step':>6} {'Etot':>10} {'D(Etot)':>14} {'Ecorr':>12} "
                f"{'D(Ecorr)':>14} {'Max(Grad)':>12} {'|Grad|':>8} "
                f"{'Step':>14}"

        i = 0

        #
        # Initialize step search
        #
        stepsearch_ = RStepSearch(self.lf, **stepsearch)
        while i < maxiter["orbiter"]:
            #
            # Copy energies from previous iteration step
            #
            e_tot_old = copy(e_tot)
            e_corr_old = copy(e_corr)

            #
            # Calculate orbital gradient and diagonal approximation to the Hessian
            #
            kappa, gradient, hessian = self.orbital_rotation_step(
                lshift, pos, orbitaloptimizer
            )

            #
            # Apply step search to orbital rotation step 'kappa'
            #
            stepsearch_(
                self,
                one,
                two,
                orb,
                **{
                    "kappa": kappa,
                    "thresh": thresh,
                    "maxiter": maxiter,
                    "gradient": gradient,
                    "hessian": hessian,
                    "guess": initial_guess[0],
                    "guesslm": initial_guess[1],
                    "solver": solver,
                    "indextrans": indextrans,
                    "orbitaloptimizer": orbitaloptimizer,
                },
            )

            #
            # reorder orbitals according to natural occupation numbers
            # works only for varitiational orbital optimization
            #
            if sort and orbitaloptimizer == "variational":
                self.sort_natural_orbitals(orb)
                # Recalculate WFN:
                # FIXME: this slows us down.
                self.solve_model(
                    one,
                    two,
                    orb,
                    **{
                        "maxiter": maxiter,
                        "thresh": thresh,
                        "guess": initial_guess[0],
                        "guesslm": initial_guess[1],
                        "solver": solver,
                        "indextrans": indextrans,
                        "orbitaloptimizer": orbitaloptimizer,
                    },
                )

            e_tot = self.compute_total_energy()
            e_corr = self.compute_correlation_energy()
            e_ref = self.compute_reference_energy()
            e_dict = {"e_tot": e_tot, "e_corr": e_corr, "e_ref": e_ref}
            self.energy = e_dict
            #
            # Update IOData container
            #
            self.checkpoint.update("e_core", self.e_core)
            self.checkpoint.update("e_tot", e_tot)
            self.checkpoint.update("e_corr", e_corr)
            self.checkpoint.update("e_ref", e_ref)

            #
            # Print information of iteration step
            #
            if log.do_medium:
                if stepsearch["method"] == "trust-region":
                    log(
                        f"{(i + 1):>5} {e_tot:> 14.8f} {(e_tot - e_tot_old):> 12.8f} "
                        f"{e_corr:> 14.8f} {(e_corr - e_corr_old):> 12.8f} "
                        f"{gradient.get_max():>10.5f} {gradient.norm():>10.2e} "
                        f"{stepsearch_.trustradius:>10.2e}"
                    )
                else:
                    log(
                        f"{(i + 1):>5} {e_tot:> 14.8f} {(e_tot - e_tot_old):> 12.8f} "
                        f"{e_corr:> 14.8f} {(e_corr - e_corr_old):> 12.8f} "
                        f"{gradient.get_max():>10.5f} {gradient.norm():>10.2e} "
                        f"{stepsearch_.alpha:>10.2e}"
                    )

            #
            # Checkpoint for orbitals
            #
            if (i + 1) % checkpoint == 0 and checkpoint > 0:
                self.checkpoint.to_file(checkpoint_fn)

            #
            # Check convergence
            #
            if self.check_convergence(e_tot, e_tot_old, gradient, thresh):
                if log.do_medium:
                    log.hline(" ")
                    log(
                        f"Orbital optimization converged in {(i + 1)} iterations"
                    )
                    log.hline(" ")
                self.checkpoint.update("converged", True)
                break
            elif self.check_stepsearch(stepsearch_):
                if log.do_medium:
                    log.hline(" ")
                    log("Trustradius too small. Orbital optimization aborted!")
                    log.hline(" ")
                self.checkpoint.update("converged", False)
                break
            i = i + 1

        #
        # Check convergence if i = maxorbiter:
        # Don't raise annoying ValueError
        #
        if i >= maxiter["orbiter"] and i > 0:
            if not self.check_convergence(e_tot, e_tot_old, gradient, thresh):
                if log.do_medium:
                    log.hline(" ")
                    log.warn(
                        f"Orbital optimization NOT converged in {(i + 1)} iterations"
                    )
                    log.hline(" ")
                self.checkpoint.update("converged", False)
        self.norbiter = i
        #
        # Update IOData container
        #
        self.checkpoint.update("niter_orb", i)
        self.checkpoint.update("overlap", self.compute_overlap())

        #
        # Print final information
        #
        if log.do_medium:
            self.print_final()
            self.dump_final(
                orb, olp, printoptions, dumpci, checkpoint, checkpoint_fn
            )

        #
        # Clean the Cache because we do not want to keep the auxiliary matrices in memory
        #
        self.clear_aux_matrix()
        #
        # Create final container as return value
        #
        return self.checkpoint()
