# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file


import numpy as np
import pytest

from pybest.log import log
from pybest.context import context
from pybest.io.iodata import IOData
from pybest.gbasis import get_gobasis
from pybest.linalg.dense import DenseLinalgFactory
from pybest.wrappers import RHF
from pybest.gbasis.gobasis import (
    compute_kinetic,
    compute_nuclear,
    compute_eri,
    compute_cholesky_eri,
    compute_nuclear_repulsion,
    compute_overlap,
)
from pybest.scf import AufbauOccModel
from pybest.geminals.rpccd import ROOpCCD
from pybest.orbital_utils import rotate_orbitals
from pybest.orbital_utils import transform_integrals
from pybest.test.common import fda_1order, numpy_seed


@pytest.mark.slow
def test_pccd_one_dm():
    fn_xyz = context.get_fn("test/li2.xyz")
    mol = IOData.from_file(fn_xyz)
    obasis = get_gobasis("cc-pvdz", fn_xyz, print_basis=False)
    lf = DenseLinalgFactory(obasis.nbasis)
    occ_model = AufbauOccModel(3)

    kin = compute_kinetic(obasis)
    na = compute_nuclear(obasis)
    er = compute_eri(obasis)
    external = compute_nuclear_repulsion(obasis)

    orb = lf.create_orbital()
    olp = compute_overlap(obasis)

    hf = RHF(lf, occ_model)
    hf_ = hf(kin, na, er, external, olp, orb)

    one = lf.create_two_index(obasis.nbasis)
    one.iadd(kin)
    one.iadd(na)

    # Do pCCD optimization:
    geminal_solver = ROOpCCD(lf, occ_model)
    pccd = geminal_solver(
        one,
        er,
        orb,
        olp,
        e_core=external,
        checkpoint=-1,
        maxiter={"orbiter": 0},
    )

    one_mo_ = lf.create_two_index()
    one_mo_.assign_two_index_transform(one, orb)
    two_mo = []
    output = geminal_solver.lf.create_four_index()
    output.assign_four_index_transform(er, orb, method="tensordot")
    two_mo.append(output)

    def fun(x):
        one_mo = []
        one_mo.append(geminal_solver.lf.create_two_index())
        one_mo[0].assign(x.reshape(28, 28))

        geminal_solver.clear_aux_matrix()
        geminal_solver.update_aux_matrix("scf", two_mo, one_mo)

        iiaa = geminal_solver.get_aux_matrix("gppqq")
        iaia = geminal_solver.get_aux_matrix("lpqpq")
        fock = geminal_solver.get_aux_matrix("fock")
        one = geminal_solver.get_aux_matrix("t")
        coeff = geminal_solver.geminal._array
        lcoeff = geminal_solver.lagrange._array

        lagrangian = geminal_solver.compute_total_energy()
        lagrangian += np.dot(
            lcoeff.ravel(order="C"),
            geminal_solver.vector_function_geminal(
                coeff, iiaa, iaia, one, fock
            ),
        )
        return lagrangian

    def fun_deriv(x):
        one_mo = []
        one_mo.append(geminal_solver.lf.create_two_index())
        one_mo[0].assign(x.reshape(28, 28))
        geminal_solver.clear_aux_matrix()
        geminal_solver.update_aux_matrix("scf", two_mo, one_mo)

        guesst = geminal_solver.generate_guess(
            {"type": "random", "factor": -0.1}
        )
        # Optimize OpCCD wavefunction amplitudes:
        coeff = geminal_solver.solve_geminal(
            guesst, {"wfn": "krylov"}, 10e-12, 128
        )

        # Optimize OpCCD Lagrange multipliers (lambda equations):
        lcoeff = geminal_solver.solve_lagrange(
            guesst, {"lagrange": "krylov"}, 10e-12, 128
        )
        onebody1 = geminal_solver.lf.create_two_index(3, 25)
        onebody2 = geminal_solver.lf.create_two_index(3, 25)
        onebody1.assign(coeff)
        onebody2.assign(lcoeff)

        onedm = geminal_solver.lf.create_one_index()
        geminal_solver.compute_1dm(onedm, onebody1, onebody2, "one_dm_response", factor=2.0)
        a = np.zeros((28, 28))
        np.fill_diagonal(a, onedm._array.T)
        return a.ravel()

    x = one_mo_._array.ravel()
    with numpy_seed():
        dxs = np.random.rand(50, 28 * 28) * 0.00001
        fda_1order(fun, fun_deriv, x, dxs)
