# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file

import numpy as np
import pytest

from pybest.log import log
from pybest.context import context
from pybest.io.iodata import IOData
from pybest.gbasis import get_gobasis
from pybest.linalg.dense import DenseLinalgFactory
from pybest.wrappers import RHF
from pybest.gbasis.gobasis import (
    compute_kinetic,
    compute_nuclear,
    compute_eri,
    compute_cholesky_eri,
    compute_nuclear_repulsion,
    compute_overlap,
)
from pybest.scf import AufbauOccModel
from pybest.geminals.rpccd import RpCCD
from pybest.orbital_utils import rotate_orbitals
from pybest.orbital_utils import transform_integrals
from pybest.test.common import fda_1order, numpy_seed


@pytest.mark.slow
def test_pccd_lagrange():
    fn_xyz = context.get_fn("test/li2.xyz")
    mol = IOData.from_file(fn_xyz)
    obasis = get_gobasis("cc-pvdz", fn_xyz, print_basis=False)
    lf = DenseLinalgFactory(obasis.nbasis)
    occ_model = AufbauOccModel(3)

    kin = compute_kinetic(obasis)
    na = compute_nuclear(obasis)
    er = compute_eri(obasis)
    external = compute_nuclear_repulsion(obasis)

    orb = lf.create_orbital()
    olp = compute_overlap(obasis)

    hf = RHF(lf, occ_model)
    hf_ = hf(kin, na, er, external, olp, orb)

    one = lf.create_two_index(obasis.nbasis, label="one")
    one.iadd(kin)
    one.iadd(na)

    def fun(x):
        iiaa = geminal_solver.get_aux_matrix("gppqq")
        iaia = geminal_solver.get_aux_matrix("lpqpq")
        fock = geminal_solver.get_aux_matrix("fock")
        one = geminal_solver.get_aux_matrix("t")
        coeff = geminal_solver.lagrange._array

        lagrangian = geminal_solver.compute_total_energy(x.reshape(3, 25))
        lagrangian += np.dot(
            coeff.ravel(order="C"),
            geminal_solver.vector_function_geminal(
                x.ravel(order="C"), iiaa, iaia, one, fock
            ),
        )
        return lagrangian

    def fun_deriv(x):
        iiaa = geminal_solver.get_aux_matrix("gppqq")
        iaia = geminal_solver.get_aux_matrix("lpqpq")
        fock = geminal_solver.get_aux_matrix("fock")
        one = geminal_solver.get_aux_matrix("t")
        coeff = geminal_solver.lagrange._array.ravel(order="C")
        gmat = geminal_solver.lf.create_two_index(3, 25)
        gmat.assign(x)

        gradient = geminal_solver.vector_function_lagrange(
            coeff, gmat, iiaa, iaia, one, fock
        )
        return gradient.ravel(order="C")

    # Do pCCD optimization:
    geminal_solver = RpCCD(lf, occ_model)
    pccd = geminal_solver(one, er, orb, olp, e_core=external)

    with numpy_seed():
        geminal_solver.lagrange.assign(np.random.rand(3, 25))
        x = geminal_solver.geminal._array.ravel(order="C")
        dxs = np.random.rand(200, 3 * 25) * (0.001)
        # transform AOs to MOs:
        ti = transform_integrals(one, er, orb)
        one_mo = ti.one
        two_mo = ti.two
        # recalculate auxiliary matrices because they are deleted after call function exists
        geminal_solver.update_aux_matrix("scf", two_mo, one_mo)
        fda_1order(fun, fun_deriv, x, dxs)

