# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file

import numpy as np
import pytest

from pybest.log import log
from pybest.context import context
from pybest.io.iodata import IOData
from pybest.gbasis import get_gobasis
from pybest.linalg.dense import DenseLinalgFactory
from pybest.wrappers import RHF
from pybest.gbasis.gobasis import (
    compute_kinetic,
    compute_nuclear,
    compute_eri,
    compute_cholesky_eri,
    compute_nuclear_repulsion,
    compute_overlap,
)
from pybest.scf import AufbauOccModel
from pybest.geminals.rpccd import ROOpCCD
from pybest.orbital_utils import rotate_orbitals
from pybest.orbital_utils import transform_integrals
from pybest.test.common import fda_2order, numpy_seed


@pytest.mark.slow
def test_pccd_hessian():
    fn_xyz = context.get_fn("test/c2.xyz")
    mol = IOData.from_file(fn_xyz)
    obasis = get_gobasis("cc-pvdz", fn_xyz, print_basis=False)
    lf = DenseLinalgFactory(obasis.nbasis)
    kin = compute_kinetic(obasis)
    na = compute_nuclear(obasis)
    er = compute_eri(obasis)
    external = compute_nuclear_repulsion(obasis)

    occ_model = AufbauOccModel(6)
    orb = lf.create_orbital()
    olp = compute_overlap(obasis)

    hf = RHF(lf, occ_model)
    hf_ = hf(kin, na, er, external, olp, orb)

    one = lf.create_two_index(obasis.nbasis, label="one")
    one.iadd(kin)
    one.iadd(na)

    # Do pCCD optimization:
    geminal_solver = ROOpCCD(lf, occ_model)
    pccd = geminal_solver(
        one,
        er,
        orb,
        olp,
        e_core=external,
        checkpoint=-1,
        maxiter={"orbiter": 0},
    )

    def fun(x):
        rotation = geminal_solver.compute_rotation_matrix(x)
        orb_ = orb.copy()
        rotate_orbitals(orb_, rotation)

        geminal_solver.clear_aux_matrix()
        ti = transform_integrals(one, er, orb_)
        one_mo = ti.one
        two_mo = ti.two

        geminal_solver.update_aux_matrix("scf", two_mo, one_mo)

        iiaa = geminal_solver.get_aux_matrix("gppqq")
        iaia = geminal_solver.get_aux_matrix("lpqpq")
        fock = geminal_solver.get_aux_matrix("fock")
        t = geminal_solver.get_aux_matrix("t")
        coeff = geminal_solver.geminal._array
        lcoeff = geminal_solver.lagrange._array

        lagrangian = geminal_solver.compute_total_energy()
        lagrangian += np.dot(
            lcoeff.ravel(order="C"),
            geminal_solver.vector_function_geminal(coeff, iiaa, iaia, t, fock),
        )
        return lagrangian

    def fun_deriv(x):
        rotation = geminal_solver.compute_rotation_matrix(x)
        orb_ = orb.copy()
        rotate_orbitals(orb_, rotation)

        geminal_solver.clear_aux_matrix()
        ti = transform_integrals(one, er, orb_)
        one_mo = ti.one
        two_mo = ti.two

        hessian = geminal_solver.get_exact_hessian(one_mo[0], two_mo[0])
        return hessian

    x = np.zeros(28 * 27 // 2)
    with numpy_seed():
        dxs = np.random.rand(50, 28 * 27 // 2) * 0.0001
        fda_2order(fun, fun_deriv, x, dxs)
