# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file

import numpy as np
import pytest

from pybest.log import log
from pybest.wrappers import RHF
from pybest.gbasis import get_gobasis
from pybest.gbasis.gobasis import (
    compute_kinetic,
    compute_nuclear,
    compute_eri,
    compute_cholesky_eri,
    compute_nuclear_repulsion,
    compute_overlap,
)
from pybest.linalg.dense import DenseLinalgFactory
from pybest.linalg.dense import DenseTwoIndex, DenseFourIndex
from pybest.scf import AufbauOccModel
from pybest.geminals.rpccd import RpCCD, ROOpCCD
from pybest.io.iodata import IOData
from pybest.context import context
from pybest.exceptions import ArgumentError
from pybest.orbital_utils import project_orbitals


def prepare_mol(basis, filen="test/h2.xyz"):
    fn_xyz = context.get_fn(filen)
    obasis = get_gobasis(basis, fn_xyz, print_basis=False)
    lf = DenseLinalgFactory(obasis.nbasis)
    olp = compute_overlap(obasis)
    kin = compute_kinetic(obasis)
    na = compute_nuclear(obasis)
    er = compute_eri(obasis)
    external = {"nn": compute_nuclear_repulsion(obasis)}

    return lf, olp, kin, na, er, external


def prepare_hf_container(basis, filen="test/h2.xyz", nocc=1):
    lf, olp, kin, na, er, external = prepare_mol(basis, filen)
    occ_model = AufbauOccModel(nocc)
    orb_a = lf.create_orbital()

    hf = RHF(lf, occ_model)
    hf_ = hf(kin, na, er, external, olp, orb_a, diis="plain")
    return lf, occ_model, kin, na, er, external, hf_


# Prepare molecule only once:
lf, occ, kin, na, er, external, hf = prepare_hf_container("6-31G")

e_tot = -1.143420629378
e_wo_ext = -1.143420629378 - external["nn"]
e_w2_ext = -1.143420629378 + external["nn"]

orb_a_ = hf.orb_a.copy()
orb_a_.clear()


testdata_external = [
    (0.0, e_wo_ext),
    (external["nn"], e_tot),
    (2 * external["nn"], e_w2_ext),
]


testdata_orbs = [
    (hf.orb_a, None, e_tot),
    (None, hf.orb_a, e_tot),
    (orb_a_, hf.orb_a, e_tot),
    (hf.orb_a, orb_a_, None),  # FloatingPointError, ValueError
    (orb_a_, None, None),  # FloatingPointError, ValueError
    (None, None, None),  # ArgumentError
]


def test_pccd_cs():
    # Do pCCD optimization:
    geminal_solver = RpCCD(lf, occ)
    pccd = geminal_solver(kin, na, er, hf)
    assert abs(pccd.e_tot - e_tot) < 1e-6


@pytest.mark.parametrize("core, expected", testdata_external)
def test_pccd_sp_args_external(core, expected):
    # Do pCCD optimization:
    geminal_solver = RpCCD(lf, occ)
    # Take only external energy from kwargs
    pccd = geminal_solver(kin, na, er, hf, e_core=core)
    assert abs(pccd.e_tot - expected) < 1e-6


@pytest.mark.parametrize("orbs1, orbs2, expected", testdata_orbs)
def test_pccd_sp_args_orbs(orbs1, orbs2, expected):
    orb = hf.orb_a.copy()
    # Do pCCD optimization:
    geminal_solver = RpCCD(lf, occ)
    # Take only external energy from args
    del hf.orb_a
    try:
        pccd = geminal_solver(kin, na, er, orbs1, hf, orb_a=orbs2)
        # reset
        hf.orb_a = orb
        assert abs(pccd.e_tot - expected) < 1e-6
    except (ArgumentError, FloatingPointError, ValueError):
        # reset
        hf.orb_a = orb
        pass


def test_pccd_cs_scf():
    # Do pCCD optimization:
    geminal_solver = ROOpCCD(lf, occ)
    pccd = geminal_solver(na, kin, er, hf, checkpoint=-1)
    assert abs(pccd.e_tot - -1.151686291339) < 1e-6


def test_pccd_cs_scf_restart():
    # Do pCCD optimization:
    geminal_solver = ROOpCCD(lf, occ)

    pccd = geminal_solver(kin, na, er, hf)
    assert abs(pccd.e_tot - -1.151686291339) < 1e-6

    old = IOData.from_file("pybest-results/checkpoint_pccd.h5")

    assert hasattr(old, "olp")
    assert hasattr(old, "orb_a")
    assert hasattr(old, "e_tot")
    assert hasattr(old, "e_ref")
    assert hasattr(old, "e_core")
    assert hasattr(old, "e_corr")
    assert hasattr(old, "dm_1")
    assert hasattr(old, "dm_2")

    # Update to slightly stretch geometry of H2
    lf1, olp1, kin1, na1, er1, external1 = prepare_mol(
        "6-31G", "test/h2_2.xyz"
    )
    orb_a = lf1.create_orbital()

    # re-orthogonalize orbitals
    project_orbitals(old.olp, olp1, old.orb_a, orb_a)

    # recompute
    pccd2 = geminal_solver(
        kin1,
        na1,
        er1,
        orb_a,
        olp1,
        e_core=external1["nn"],
        checkpoint_fn="checkpoint_restart.h5",
    )

    assert abs(pccd2.e_tot - -1.150027881389) < 1e-6

    # recompute
    pccd3 = geminal_solver(
        kin1, na1, er1, pccd2, restart="pybest-results/checkpoint_restart.h5",
    )

    assert abs(pccd3.e_tot - -1.150027881389) < 1e-6

    # recompute with wrong core energy
    pccd4 = geminal_solver(
        kin1,
        na1,
        er1,
        pccd3,
        e_core=10.000,
        restart="pybest-results/checkpoint_restart.h5",
    )

    with pytest.raises(AssertionError):
        assert abs(pccd4.e_tot - -1.150027881389) < 1e-6
    assert abs(pccd3.e_corr - pccd4.e_corr) < 1e-6
    assert abs(pccd3.e_ref - pccd3.e_core - pccd4.e_ref + pccd4.e_core) < 1e-6


test_core = [
    (RpCCD, "water", "cc-pvdz", 0, 5, {}, {"e_tot": -76.07225799852085}),
    (RpCCD, "water", "cc-pvdz", 1, 5, {}, {"e_tot": -76.07210055926937}),
    (
        ROOpCCD,
        "water",
        "cc-pvdz",
        1,
        5,
        {"sort": False},
        {"e_tot": -76.0994990025405},
    ),
]


@pytest.mark.parametrize(
    "cls, mol, basis, ncore, nocc, kwargs, result", test_core
)
def test_pccd_core(cls, mol, basis, ncore, nocc, kwargs, result):

    lf_, occ_, kin_, na_, er_, external_, hf_ = prepare_hf_container(
        basis, f"test/{mol}.xyz", nocc
    )

    pccd_ = cls(lf_, occ_, ncore=ncore)
    pccd = pccd_(kin_, na_, er_, hf_, **kwargs)

    assert abs(pccd.e_tot - result["e_tot"]) < 1e-6
