# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --


import numpy as np

from pybest.scf.hamiltonian import RScfHam, UScfHam
from pybest.scf.observable import (
    RTwoIndexTerm,
    RDirectTerm,
    RExchangeTerm,
    UTwoIndexTerm,
    UDirectTerm,
    UExchangeTerm,
)
from pybest.part.mulliken import get_mulliken_operators
from pybest.gbasis.gobasis import (
    compute_overlap,
    compute_kinetic,
    compute_nuclear,
    compute_eri,
    compute_nuclear_repulsion,
)
from pybest.scf.utils import compute_1dm_hf


__all__ = ["compute_mulliken_charges"]


def compute_mulliken_charges(obasis, lf, numbers, dm):
    operators = get_mulliken_operators(obasis)
    populations = np.array(
        [operator.contract("ab,ab", dm) for operator in operators]
    )
    return numbers - np.array(populations)


def compute_hf_energy(mol):
    olp = compute_overlap(mol.gobasis)
    kin = compute_kinetic(mol.gobasis)
    na = compute_nuclear(mol.gobasis)
    er = compute_eri(mol.gobasis)
    external = {"nn": compute_nuclear_repulsion(mol.gobasis)}
    if hasattr(mol, "orb_b"):
        # assuming unrestricted
        terms = [
            UTwoIndexTerm(kin, "kin"),
            UDirectTerm(er, "hartree"),
            UExchangeTerm(er, "x_hf"),
            UTwoIndexTerm(na, "ne"),
        ]
        ham = UScfHam(terms, external)
        dm_a = compute_1dm_hf(mol.orb_a)
        dm_b = compute_1dm_hf(mol.orb_b)
        ham.reset(dm_a, dm_b)
    else:
        # assuming restricted
        terms = [
            RTwoIndexTerm(kin, "kin"),
            RDirectTerm(er, "hartree"),
            RExchangeTerm(er, "x_hf"),
            RTwoIndexTerm(na, "ne"),
        ]
        ham = RScfHam(terms, external)
        dm_a = compute_1dm_hf(mol.orb_a)
        ham.reset(dm_a)
    return ham.compute_energy()
