# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file


import numpy as np, os

from pybest import context
from pybest import periodic
from pybest.scf.utils import compute_1dm_hf
from pybest.test.common import tmpdir
from pybest.gbasis.gobasis import compute_overlap
from pybest.io.iodata import IOData
from pybest.io.test.common import compute_mulliken_charges
from pybest.io.molden import load_molden, dump_molden


def test_load_molden_li2_orca():
    fn_molden = context.get_fn("test/li2.molden.input")
    mol = IOData.from_file(fn_molden)

    # Checkt title
    assert mol.title == "Molden file created by orca_2mkl for BaseName=li2"

    # Check normalization
    olp = compute_overlap(mol.gobasis)
    mol.orb_a.check_normalization(olp, 1e-5)
    mol.orb_b.check_normalization(olp, 1e-5)

    # Check Mulliken charges
    numbers = np.array([periodic[i].number for i in mol.atom])
    dm_full = compute_1dm_hf(mol.orb_a)
    dm_full.iadd(compute_1dm_hf(mol.orb_b))
    charges = compute_mulliken_charges(mol.gobasis, mol.lf, numbers, dm_full)
    expected_charges = np.array([0.5, 0.5])
    if os.path.isdir("./pybest_tmp"):
        import shutil

        shutil.rmtree("./pybest_tmp")
    assert abs(charges - expected_charges).max() < 1e-5


def test_load_molden_h2o_orca():
    fn_molden = context.get_fn("test/h2o.molden.input")
    mol = IOData.from_file(fn_molden)

    # Checkt title
    assert mol.title == "Molden file created by orca_2mkl for BaseName=h2o"

    # Check normalization
    olp = compute_overlap(mol.gobasis)
    mol.orb_a.check_normalization(olp, 1e-5)

    # Check Mulliken charges
    numbers = np.array([periodic[i].number for i in mol.atom])
    dm_full = compute_1dm_hf(mol.orb_a)
    dm_full.iscale(2.0)
    charges = compute_mulliken_charges(mol.gobasis, mol.lf, numbers, dm_full)
    expected_charges = np.array([-0.816308, 0.408154, 0.408154])
    if os.path.isdir("./pybest_tmp"):
        import shutil

        shutil.rmtree("./pybest_tmp")
    assert abs(charges - expected_charges).max() < 1e-5


def test_load_dump_molden_nh3_molden_pure():
    # The file tested here is created with molden. It should be read in
    # properly without altering normalization and sign conventions.
    fn_molden = context.get_fn("test/nh3_molden_pure.molden")
    mol = IOData.from_file(fn_molden)
    # Check normalization
    olp = compute_overlap(mol.gobasis)
    mol.orb_a.check_normalization(olp, 1e-4)
    # Check Mulliken charges. Comparison with numbers from the Molden program
    # output.
    numbers = np.array([periodic[i].number for i in mol.atom])
    dm_full = compute_1dm_hf(mol.orb_a)
    dm_full.iscale(2.0)
    charges = compute_mulliken_charges(mol.gobasis, mol.lf, numbers, dm_full)
    molden_charges = np.array([0.0381, -0.2742, 0.0121, 0.2242])
    # dump it
    if not os.path.exists("pybest_tmp"):
        os.makedirs("pybest_tmp")
    fn = "pybest_tmp/tmp.molden"
    dump_molden(fn, mol)
    # read it again
    mol2 = IOData.from_file(fn)
    assert mol.gobasis.basisname == mol2.gobasis.basisname
    assert mol.gobasis.alpha == mol2.gobasis.alpha
    assert mol.gobasis.contraction == mol2.gobasis.contraction
    assert mol.gobasis.nbasis == mol2.gobasis.nbasis
    assert mol.gobasis.nprim == mol2.gobasis.nprim
    assert mol.gobasis.nshell == mol2.gobasis.nshell
    assert mol.gobasis.ncenter == mol2.gobasis.ncenter
    assert mol.gobasis.shell2atom == mol2.gobasis.shell2atom
    assert mol.gobasis.shell_types == mol2.gobasis.shell_types
    assert mol.gobasis.atom == mol2.gobasis.atom
    assert mol.gobasis.ncenter == mol2.gobasis.ncenter
    assert (
        abs(
            np.array(mol.gobasis.coordinates)
            - np.array(mol2.gobasis.coordinates)
        )
        < 1e-8
    ).all()
    if os.path.isdir("./pybest_tmp"):
        import shutil

        shutil.rmtree("./pybest_tmp")
    assert abs(charges - molden_charges).max() < 1e-3


def test_load_dump_molden_nh3_molden_cart():
    # The file tested here is created with molden. It should be read in
    # properly without altering normalization and sign conventions.
    fn_molden = context.get_fn("test/nh3_molden_cart.molden")
    mol = IOData.from_file(fn_molden)
    # Check normalization
    olp = compute_overlap(mol.gobasis)
    mol.orb_a.check_normalization(olp, 1e-4)
    # Check Mulliken charges. Comparison with numbers from the Molden program
    # output.
    numbers = np.array([periodic[i].number for i in mol.atom])
    dm_full = compute_1dm_hf(mol.orb_a)
    dm_full.iscale(2.0)
    charges = compute_mulliken_charges(mol.gobasis, mol.lf, numbers, dm_full)
    molden_charges = np.array([0.3138, -0.4300, -0.0667, 0.1829])
    # dump it
    if not os.path.exists("pybest_tmp"):
        os.makedirs("pybest_tmp")
    fn = "pybest_tmp/tmp.molden"
    dump_molden(fn, mol)
    # read it again
    mol2 = IOData.from_file(fn)
    assert mol.gobasis.basisname == mol2.gobasis.basisname
    assert mol.gobasis.alpha == mol2.gobasis.alpha
    assert mol.gobasis.contraction == mol2.gobasis.contraction
    assert mol.gobasis.nbasis == mol2.gobasis.nbasis
    assert mol.gobasis.nprim == mol2.gobasis.nprim
    assert mol.gobasis.nshell == mol2.gobasis.nshell
    assert mol.gobasis.ncenter == mol2.gobasis.ncenter
    assert mol.gobasis.shell2atom == mol2.gobasis.shell2atom
    assert mol.gobasis.shell_types == mol2.gobasis.shell_types
    assert mol.gobasis.atom == mol2.gobasis.atom
    assert mol.gobasis.ncenter == mol2.gobasis.ncenter
    assert (
        abs(
            np.array(mol.gobasis.coordinates)
            - np.array(mol2.gobasis.coordinates)
        )
        < 1e-8
    ).all()
    if os.path.isdir("./pybest_tmp"):
        import shutil

        shutil.rmtree("./pybest_tmp")
    assert abs(charges - molden_charges).max() < 1e-3


def test_load_molden_nh3_orca():
    # The file tested here is created with ORCA. It should be read in
    # properly by altering normalization and sign conventions.
    fn_molden = context.get_fn("test/nh3_orca.molden")
    mol = IOData.from_file(fn_molden)
    # Check normalization
    olp = compute_overlap(mol.gobasis)
    mol.orb_a.check_normalization(olp, 1e-5)
    # Check Mulliken charges. Comparison with numbers from the Molden program
    # output.
    numbers = np.array([periodic[i].number for i in mol.atom])
    dm_full = compute_1dm_hf(mol.orb_a)
    dm_full.iscale(2.0)
    charges = compute_mulliken_charges(mol.gobasis, mol.lf, numbers, dm_full)
    molden_charges = np.array([0.0381, -0.2742, 0.0121, 0.2242])
    if os.path.isdir("./pybest_tmp"):
        import shutil

        shutil.rmtree("./pybest_tmp")
    assert abs(charges - molden_charges).max() < 1e-3


def test_load_molden_nh3_psi4():
    # The file tested here is created with PSI4. It should be read in
    # properly by altering normalization conventions.
    fn_molden = context.get_fn("test/nh3_psi4.molden")
    mol = IOData.from_file(fn_molden)
    # Check normalization
    olp = compute_overlap(mol.gobasis)
    mol.orb_a.check_normalization(olp, 1e-5)
    # Check Mulliken charges. Comparison with numbers from the Molden program
    # output.
    numbers = np.array([periodic[i].number for i in mol.atom])
    dm_full = compute_1dm_hf(mol.orb_a)
    dm_full.iscale(2.0)
    charges = compute_mulliken_charges(mol.gobasis, mol.lf, numbers, dm_full)
    molden_charges = np.array([0.0381, -0.2742, 0.0121, 0.2242])
    if os.path.isdir("./pybest_tmp"):
        import shutil

        shutil.rmtree("./pybest_tmp")
    assert abs(charges - molden_charges).max() < 1e-3


def test_load_molden_nh3_molpro2012():
    # The file tested here is created with MOLPRO2012.
    fn_molden = context.get_fn("test/nh3_molpro2012.molden")
    mol = IOData.from_file(fn_molden)
    # Check normalization
    olp = compute_overlap(mol.gobasis)
    mol.orb_a.check_normalization(olp, 1e-5)
    # Check Mulliken charges. Comparison with numbers from the Molden program
    # output.
    numbers = np.array([periodic[i].number for i in mol.atom])
    dm_full = compute_1dm_hf(mol.orb_a)
    dm_full.iscale(2.0)
    charges = compute_mulliken_charges(mol.gobasis, mol.lf, numbers, dm_full)
    molden_charges = np.array([0.0381, -0.2742, 0.0121, 0.2242])
    if os.path.isdir("./pybest_tmp"):
        import shutil

        shutil.rmtree("./pybest_tmp")
    assert abs(charges - molden_charges).max() < 1e-3
