# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --

# Detailed changelog:
#
# This implementation has been taken from `Horton 2.0.0`.
# However, this file has been updated and debugged. Compatibility with Horton is NOT
# guaranteed.
# 
# Detailed changes (see also CHANGELOG):
# 2020-07-01: Update to PyBEST standard, including naming convention
# 2020-07-01: Introduce general tensor contraction engine used for all NIndex objects

"""Base classes"""

import numpy as np
from pybest.utils import check_type
from pybest.log import log
from pybest.linalg.contract import (
    td_helper,
    cholesky_blas_routine,
    cholesky_td_routine,
    parse_contract_input,
    parse_subscripts,
    slice_output,
    get_outshape,
)
from pybest.exceptions import ArgumentError, UnknownOption

try:
    from opt_einsum import contract as oe_contract
except ImportError:
    log.warn("The opt_einsum module has been not detected.")
    oe_contract = np.einsum


__all__ = [
    "LinalgFactory",
    "LinalgObject",
    "OneIndex",
    "Orbital",
    "TwoIndex",
    "ThreeIndex",
    "FourIndex",
    "parse_four_index_transform_exps",
]


# Possible symmetries for four-index objects


class LinalgFactory(object):
    """A collection of compatible matrix and linear algebra routines.

       This is just an abstract base class that serves as a template for
       specific implementations.
    """

    def __init__(self, default_nbasis=None):
        """
           **Optional arguments:**

           default_nbasis
                The default basis size when constructing new
                operators/expansions.
        """
        self.default_nbasis = default_nbasis

    @classmethod
    def from_hdf5(cls, grp):
        """Construct an instance from data previously stored in an h5py.Group.

           **Arguments:**

           grp
                An h5py.Group object.
        """
        default_nbasis = grp.attrs.get("default_nbasis")
        return cls(default_nbasis)

    def to_hdf5(self, grp):
        """Write a LinalgFactory to an HDF5 group

           **Argument:**

           grp
                A h5py.Group instance to write to.
        """
        grp.attrs["class"] = self.__class__.__name__
        if self.default_nbasis is not None:
            grp.attrs["default_nbasis"] = self.default_nbasis

    @property
    def default_nbasis(self):
        return self._default_nbasis

    @default_nbasis.setter
    def default_nbasis(self, nbasis):
        self._default_nbasis = nbasis

    def create_one_index(self, nbasis=None):
        raise NotImplementedError

    def create_orbital(self, nbasis=None):
        raise NotImplementedError

    def create_two_index(self, nbasis=None):
        raise NotImplementedError

    def create_four_index(self, nbasis=None):
        raise NotImplementedError

    def create_three_index(self, nbasis=None):
        raise NotImplementedError


class LinalgObject(object):
    def __check_init_args__(self, nbasis):
        raise NotImplementedError

    def __eq__(self, other):
        raise NotImplementedError

    @classmethod
    def from_hdf5(cls, grp):
        raise NotImplementedError

    def to_hdf5(self, grp):
        raise NotImplementedError

    def new(self):
        raise NotImplementedError

    def __clear__(self):
        """Part of the API specified in pybest.cache"""
        self.clear()

    def clear(self):
        raise NotImplementedError

    def copy(self):
        raise NotImplementedError

    def assign(self, other):
        raise NotImplementedError

    def randomize(self):
        raise NotImplementedError

    def permute_basis(self, permutation):
        raise NotImplementedError

    def change_basis_signs(self, signs):
        raise NotImplementedError


class NIndexObject(LinalgObject):
    def __init__(self, nbasis):
        raise NotImplementedError

    def iadd(self, other, factor=1):
        raise NotImplementedError

    def iscale(self, factor):
        raise NotImplementedError

    @property
    def shape(self):
        raise NotImplementedError

    @property
    def ndim(self):
        """The number of axes in the N-index object."""
        return len(self.shape)

    def _fix_ends(self, *ends):
        shape = self.shape
        if len(shape) != len(ends):
            raise ArgumentError(
                "The argument 'ends' must have the same length as 'self.shape'."
            )
        return tuple([ends[i] or shape[i] for i in range(len(shape))])

    def contract(self, *args, **kwargs):
        """ General NIndex contraction function

            NIndex = CholeskyFourIndex or DenseNIndex
            where N = One, Two, Three, Four

        *** Arguments ***

        * self : NIndex
              This is the first operand.

        * subscript : string
              Specifies the subscripts for summation as comma separated list
              of subscript labels, for example: 'abcd,efcd->abfe'.

        * operands : DenseNIndex
              These are the other arrays for the operation. If out keyword is
              not specified and the result of operation is not scalar value,
              the last operand is treated as out.

        *** Keyword arguments ***

        * out : DenseNIndex
              The product of operation is added to out.

        * factor : float
              The product of operation is multiplied by factor before it is
              added to out.

        * clear : boolean
               The out is cleared before the product of operation is added to
               out if clear is set to True.

        * select : string
               Specifies the contraction algorithm: One of:
               - 'opt_einsum' - opt_einsum.contract
               - 'einsum' - numpy.einsum function,
               - 'td' - operations with utilization of numpy.tensordot function,
               - 'blas' - operations written in c++,
               - None - first tries 'td', 'cpp', or 'blas' routine. In case of
                 failure, does 'einsum' operation.
               'td' is usually much faster but it can provide increased
               memory usage and it is available only for specific cases.

        * optimize : {False, True, 'optimal'}
               Specifies contraction path for operation if select is set to
               'einsum'. For details, see numpy.einsum_path.

        * begin0, end0, begin1, ... :
               The lower and upper bounds for summation.
        """
        subscripts = args[0]
        factor = kwargs.get("factor", 1.0)
        select = kwargs.get("select", None)
        clear = kwargs.get("clear", False)
        opt = kwargs.get("optimize", "optimal")
        out = kwargs.get("out", None)

        # Check if out is scalar
        scripts, outscript = parse_subscripts(subscripts)
        is_out_scalar = (outscript == "") or (outscript == "...")

        # If keyword out is None, use last argument as out. Determine operands.
        if (len(args) == len(scripts) + 1) and (out == None):
            operands = [self] + list(args[1:-1])
            out = args[-1]
            arr = out._array
        else:
            operands = [self] + list(args[1:])

        # Slice arrays of operands and create a list with operands (ndarrays)
        subs_, operands_ = parse_contract_input(subscripts, operands, kwargs)
        args_ = [subs_] + operands_

        # Create output array if not given
        if (not is_out_scalar) and (out == None):
            from pybest.linalg import dense

            shape = get_outshape(subs_, operands_)
            out = dense.DenseLinalgFactory._allocate_check_output(None, shape)

        # Clear out is required and reference to its _array attribute.
        if out != None:
            if clear:
                out.clear()
            arr = out._array

        # 1. The product of operation is a scalar.
        if is_out_scalar:
            if out != None:
                raise ArgumentError(
                    "Output operand should not be specified if the contraction"
                    " result is scalar."
                )
            if select == "td":
                return factor * td_helper(*args_)
            elif select == "opt_einsum":
                return factor * oe_contract(*args_, optimize=opt)
            elif select == "einsum":
                return factor * np.einsum(*args_, optimize=opt)
            elif select == None:
                try:
                    x = factor * td_helper(*args_)
                except ArgumentError:
                    x = factor * np.einsum(*args_, optimize=opt)
                return x
            else:
                raise UnknownOption(
                    f"Unrecognized keyword: select = {select}.\n"
                    "Do you mean one of: 'td', 'opt_einsum' or 'einsum'?\n"
                )

        # 2. The product of operation is DenseNIndexObject.
        slice_ = slice_output(subscripts, kwargs)

        if select == "td" and len(self.arrays) == 2:
            try:
                args_td = [subscripts, operands_, arr, factor]
                cholesky_td_routine(*args_td)
            except NotImplementedError:
                try:
                    arr[:] += factor * td_helper(*args_)
                except ArgumentError:
                    m = f"{subscripts} cannot be done with select='td'."

        elif select == "td":
            arr[slice_] += factor * td_helper(*args_)

        elif select == "opt_einsum":
            arr[slice_] += factor * oe_contract(*args_, optimize=opt)

        elif select == "einsum":
            arr[slice_] += factor * np.einsum(*args_, optimize=opt)

        elif select == "blas":
            args_blas = [
                subscripts,
                [self._array, self._array2] + operands_[2:],
                arr,
                factor,
                kwargs,
            ]
            cholesky_blas_routine(*args_blas)

        elif select == None:
            try:
                if len(self.arrays) == 2:
                    try:
                        args_td = [subscripts, operands_, arr, factor]
                        cholesky_td_routine(*args_td)
                    except NotImplementedError:
                        arr[slice_] += factor * td_helper(*args_)
                else:
                    arr[slice_] += factor * td_helper(*args_)
            except ArgumentError:
                arr[slice_] += factor * np.einsum(*args_, optimize=opt)

        else:
            raise UnknownOption(
                f"Unrecognized keyword: select = {select}.\n"
                "Do you mean one of: 'blas', 'td', 'opt_einsum' or 'einsum'?\n"
            )

        return out

    def get_max_values(self, limit=20, absolute=True):
        """ Returns list of tuples with maximal values and their indices. """

        values = [(index, val) for index, val in np.ndenumerate(self._array)]
        sortkey = {True: lambda x: abs(x[1]), False: lambda x: x[1]}
        values.sort(key=sortkey[absolute], reverse=True)

        if len(values) > limit:
            values = values[:limit]

        return values


class OneIndex(NIndexObject):
    def copy(self, begin=0, end=None):
        raise NotImplementedError

    def get_element(self, i):
        raise NotImplementedError

    def set_element(self, i, value):
        raise NotImplementedError


class Orbital(LinalgObject):
    def __init__(self, nbasis, nfn=None):
        raise NotImplementedError

    def check_normalization(self, olp, eps=1e-4):
        raise NotImplementedError

    def error_eigen(self, fock, overlap):
        raise NotImplementedError


class TwoIndex(NIndexObject):
    def __init__(self, nbasis):
        raise NotImplementedError

    def copy(self, begin0=0, end0=None, begin1=None, end1=None):
        raise NotImplementedError

    def get_element(self, i, j):
        raise NotImplementedError

    def set_element(self, i, j, value):
        raise NotImplementedError

    def sum(self):
        raise NotImplementedError

    def trace(self):
        raise NotImplementedError

    def itranspose(self):
        raise NotImplementedError

    def inner(self, vec0, vec1):
        raise NotImplementedError


class ThreeIndex(NIndexObject):
    def copy(
        self,
        begin0=0,
        end0=None,
        begin1=None,
        end1=None,
        begin2=None,
        end2=None,
    ):
        raise NotImplementedError

    def get_element(self, i, j, k):
        raise NotImplementedError

    def set_element(self, i, j, k, value):
        raise NotImplementedError


class FourIndex(NIndexObject):
    def copy(self, begin=0, end=None):
        raise NotImplementedError

    def get_element(self, i, j, k, l):
        raise NotImplementedError

    def set_element(self, i, j, k, l, value):
        raise NotImplementedError


def parse_four_index_transform_exps(exp0, exp1, exp2, exp3, Class):
    """Parse the optional arguments exp1, exp2 and exp3.

       **Arguments:**

       exp0, exp1, exp2, exp3
            Four sets of orbitals for the mo transformation. Some may be None
            but only the following not None combinations are allowed:

            * ``(exp0,)``: maintain eight-fold symmetry (if any)
            * ``(exp0, exp1)``: maintain four-fold symmetry (if any)
            * ``(exp0, exp2)``: maintain two-fold symmetry (if any)
            * ``(exp0, exp1, exp2, exp3)``: break all symmetry

       Class
            The expected class of the exps objects.


       **Returns:** exp0, exp1, exp2, exp3. (All not None)
    """
    # Four supported situations
    if exp1 is None and exp2 is None and exp3 is None:
        # maintains eight-fold symmetry
        exp1 = exp0
        exp2 = exp0
        exp3 = exp0
    elif exp2 is None and exp3 is None:
        # maintains four-fold symmetry
        exp2 = exp0
        exp3 = exp1
    elif exp1 is None and exp3 is None:
        # maintains two-fold symmetry
        exp1 = exp0
        exp3 = exp2
    elif exp1 is None or exp2 is None or exp3 is None:
        # the only other allowed case is no symmetry.
        raise ArgumentError(
            "It is not clear how to interpret the optional arguments exp1, exp2 and exp3."
        )
    check_type("exp0", exp0, Class)
    check_type("exp1", exp1, Class)
    check_type("exp2", exp2, Class)
    check_type("exp3", exp3, Class)
    return exp0, exp1, exp2, exp3
