# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --

# Detailed changelog:
#
# This implementation has been taken from `Horton 2.0.0`.
# However, this file has been updated and debugged. Compatibility with Horton is NOT
# guaranteed.
# 
# Detailed changes (see also CHANGELOG):
# 2020-07-01: Update to PyBEST standard, including naming convention
# 2020-07-01: Introduce general tensor contraction engine
# 2020-07-01: Update to new python features, including f-strings
# 2020-07-01: Changed to general [slice] function and removed deprecated [slice_to] functions
# 2020-07-01: Included additional tensor contractions
# 2020-07-01: Removed deprecated [contract_] functions

"""Cholesky decomposition of four-index objects"""


import numpy as np
from pybest.log import log, timer
from pybest.utils import check_type, check_options, doc_inherit
from pybest.linalg.base import parse_four_index_transform_exps, FourIndex
from pybest.linalg.pyslice import (
    slice_to_three_abbc_abc,
    slice_to_three_abcc_bac,
    slice_to_three_abcc_abc,
)
from pybest.linalg.pytransform import apply_four_index_transform_cpp_chol
from pybest.linalg.pycontract import (
    contract_two_to_two_abcd_cb_ad_blas,
    contract_four_to_four_abcd_efbd_aecf,
    contract_four_to_four_abcd_efbd_aecf_blas,
)
from pybest.linalg.dense import (
    DenseLinalgFactory,
    DenseOrbital,
    DenseOneIndex,
    DenseTwoIndex,
    DenseThreeIndex,
    DenseFourIndex,
)
from pybest.exceptions import ArgumentError, MatrixShapeError, UnknownOption
from pybest.linalg.contract import get_outshape, parse_contract_input


__all__ = [
    "CholeskyFourIndex",
    "CholeskyLinalgFactory",
]


class CholeskyLinalgFactory(DenseLinalgFactory):
    @doc_inherit(DenseLinalgFactory)
    def create_four_index(
        self,
        nbasis=None,
        nvec=None,
        array=None,
        array2=None,
        copy=False,
        label="",
    ):
        nbasis = nbasis or self.default_nbasis
        return CholeskyFourIndex(nbasis, nvec, array, array2, copy, label)

    @doc_inherit(DenseLinalgFactory)
    def _check_four_index_init_args(
        self, four_index, nbasis=None, nvec=None, array=None
    ):
        nbasis = nbasis or self.default_nbasis
        four_index.__check_init_args__(nbasis, nvec)

    create_four_index.__check_init_args__ = _check_four_index_init_args

    @staticmethod
    def td_helper(subscripts, *operands):
        """tensordot helper for cholesky n-index objects.

           **Arguments:**


        """
        if subscripts.count("->") == 1:
            # Sanity checking
            if len(operands) > 3:
                raise ArgumentError(
                    "Cannot use tensordot to contract more than two arrays at a time"
                )
            # Do this operation again
            inscripts, outscript = subscripts.split("->")
            inscripts = inscripts.split(",")
            # Check if operation is possible with tensordot
            for s in inscripts:
                no_rep = set(s)
                if len(s) > len(no_rep):
                    raise ArgumentError(
                        f"Cannot use np.tensordot for operation {subscripts}"
                    )

            # Get list of input and output indices
            inlistl = list(inscripts[0])
            inlistr = list(inscripts[1])

            axis10 = []
            axis11 = []
            axis20 = []
            axis21 = []
            # number of contractions for each vector
            nops1 = 0
            nops2 = 0
            transpose = []
            left = inscripts[0]
            right = inscripts[1]
            # get indices for both Cholesky vectors
            # We assume that the first four indices correspond to Cholesky indices
            left1 = "x" + left[0] + left[2]
            left1_ = "x" + left[0] + left[2]
            left2 = "x" + left[1] + left[3]
            # Distingues between different cases:
            for s in inlistr:
                if s in left1:
                    nops1 += 1
                else:
                    nops2 += 1

            # Get axes for np.td operation
            # First Cholesky vector
            # Consider only if needed
            if nops1 > 0:
                for i, c in zip(range(len(inlistr)), inlistr):
                    t1 = [pos for pos, char in enumerate(left1_) if char == c]
                    if t1:
                        axis10.append(t1[0])
                        axis11.append(i)
                    left1 = left1.replace(c, "")
                    if t1:
                        right = right.replace(c, "")
                for c in right:
                    left1 += c

            inlistleft2 = list(left2)
            left1_ = left1
            left2_ = left2
            # Second Cholesky vector
            # Consider only if needed
            if nops2 >= 0 and nops1 > 0:
                for i, c in zip(range(len(inlistleft2)), inlistleft2):
                    t2 = [pos for pos, char in enumerate(left1_) if char == c]
                    if t2:
                        axis20.append(t2[0])
                        axis21.append(i)
                    left1 = left1.replace(c, "")
                    if t2:
                        left2 = left2.replace(c, "")
            elif nops2 > 0 and nops1 == 0:
                for i, c in zip(range(len(inlistr)), inlistr):
                    t2 = [pos for pos, char in enumerate(left2_) if char == c]
                    if t2:
                        axis20.append(t2[0])
                        axis21.append(i)
                    left2 = left2.replace(c, "")
                    if t2:
                        right = right.replace(c, "")
                for c in right:
                    left2 += c
                inlistleft2 = list(left2)
                left1_ = left1
                # Do final contraction between Cholesky vectors if needed
                for i, c in zip(range(len(inlistleft2)), inlistleft2):
                    t2 = [pos for pos, char in enumerate(left1_) if char == c]
                    if t2:
                        axis10.append(t2[0])
                        axis11.append(i)
                    left1 = left1.replace(c, "")
                    if t2:
                        left2 = left2.replace(c, "")

            reforder = left1 + left2
            # Get indices for matrix transpose
            if outscript == (left1 + left2):
                transpose = None
            else:
                for c in outscript:
                    t = [pos for pos, char in enumerate(reforder) if char == c]
                    if t:
                        transpose.append(t[0])
                    else:
                        raise ArgumentError(
                            f"Cannot use np.tensordot for operation {subscripts}"
                        )

            if nops1 == 0:
                tmp = np.tensordot(
                    operands[1], operands[2], axes=(axis20, axis21)
                )
                if transpose:
                    return np.tensordot(
                        operands[0], tmp, axes=(axis10, axis11)
                    ).transpose(transpose)
                else:
                    return np.tensordot(
                        operands[0], tmp, axes=(axis10, axis11)
                    )
            else:  # if nops2 == 0:
                tmp = np.tensordot(
                    operands[0], operands[2], axes=(axis10, axis11)
                )
                if transpose:
                    return np.tensordot(
                        tmp, operands[1], axes=(axis20, axis21)
                    ).transpose(transpose)
                else:
                    return np.tensordot(
                        tmp, operands[1], axes=(axis20, axis21)
                    )
        else:
            raise NotImplementedError
            inscripts = subscripts.split(",")

            inlistl = list(inscripts[0])
            inlistr = list(inscripts[1])

            axis0 = np.argsort(inlistl)
            axis1 = np.argsort(inlistr)
            return np.tensordot(*operands, axes=(axis0, axis1))

    @staticmethod
    def tco(subscripts, out, factor, clear, select="td", *operands):
        """tensor contraction wrapper for DenseNBody objects

           **Arguments:**

           subscripts
                The usual subscripts argument for the einsum function, except
                that the format is slightly limited. All indexes must be
                explicitly stated. If the output argument is a scalar, ``->``
                may not be present. In all other cases, ``->`` must be present.

           out
                An output DenseNBody object. This argument is not allowed when
                the result is a scalar.

           factor
                Multiply the contraction by a scalar factor.

           clear
                When set to False, the output argument is not zeroed before
                adding the contraction.

           operands
                A list where each element is (i) a DenseNBody object or (ii) a
                two-tuple of DenseNBody object and ranges to select for slicing.
                The ranges must be a tuple of integers of the form (begin0,
                end0, begin1, end1, ...) where the number of (begin, end) pairs
                depends on the dimensionality of the corresponding tensor.

           **Returns:** the out object. When the out argument is given, this is
           returned with in-place modifications.
        """
        check_type("subscripts", subscripts, str)
        check_type("factor", factor, float, int)
        check_type("clear", clear, bool)

        # Basic checking of operands and adding ranges if omitted
        new_operands = []
        for operand in operands:
            if isinstance(operand, tuple):
                if len(operand) != 2:
                    raise ArgumentError(
                        "When an operand is a tuple, it must count two elements."
                    )
                tensor, ranges = operand
                if len(tensor.shape) * 2 != len(ranges):
                    raise ArgumentError(
                        "The dimensionality of the tensor and the size of ranges does not match."
                    )
            else:
                tensor = operand
                ranges = sum([(0, s) for s in tensor.shape], ())
            new_operands.append((tensor, list(ranges)))
        operands = new_operands

        # Impose explicit format for subscripts and do some basic checking with
        # operands
        if subscripts.count("->") == 1:
            inscripts, outscript = subscripts.split("->")
        elif subscripts.count("->") == 0:
            inscripts = subscripts
            outscript = ""
        else:
            raise ArgumentError(
                "The subscripts argument must contain a single or none '->'."
            )
        inscripts = inscripts.split(",")
        for inscript, (tensor, ranges) in zip(inscripts, operands):
            if len(inscript) == 1:
                if not isinstance(tensor, DenseOneIndex):
                    raise ArgumentError(
                        "Expecting a DenseOneIndex tensor for subscripts " \
                        f"'{inscript}'"
                    )
            elif len(inscript) == 2:
                if not isinstance(tensor, DenseTwoIndex):
                    raise ArgumentError(
                        "Expecting a DenseTwoIndex tensor for subscripts " \
                        f"'{inscript}'"
                    )
            elif len(inscript) == 3:
                if not isinstance(tensor, DenseThreeIndex):
                    raise ArgumentError(
                        "Expecting a DenseThreeIndex tensor for subscripts " \
                        f"'{inscript}'"
                    )
            elif len(inscript) == 4:
                if not isinstance(tensor, (DenseFourIndex, CholeskyFourIndex)):
                    raise ArgumentError(
                        "Expecting a DenseFourIndex tensor for subscripts " \
                        f"'{inscript}'"
                    )
            else:
                raise ArgumentError(
                    "The number of subscripts for one tensor must be " \
                    "1, 2, 3 or 4"
                )

        # take slices of the input arrays
        new_operands = []
        for tensor, ranges in operands:
            slices = []
            # do different slicing for Cholesky
            if isinstance(tensor, CholeskyFourIndex):
                slices.append(slice(0, -1))
                # first vector
                begin = ranges[0]
                end = ranges[1]
                slices.append(slice(begin, end))
                begin = ranges[4]
                end = ranges[5]
                slices.append(slice(begin, end))
                array = tensor._array[tuple(slices)]
                new_operands.append(array)
                # second vector
                slices = []
                slices.append(slice(0, -1))
                begin = ranges[2]
                end = ranges[3]
                slices.append(slice(begin, end))
                begin = ranges[6]
                end = ranges[7]
                slices.append(slice(begin, end))
                array = tensor._array2[tuple(slices)]
                new_operands.append(array)
            else:
                while len(ranges) > 0:
                    begin = ranges.pop(0)
                    end = ranges.pop(0)
                    slices.append(slice(begin, end))
                array = tensor._array[tuple(slices)]
                new_operands.append(array)
        operands = new_operands

        # Determine shape of the output
        outshape = []
        for outchar in outscript:
            size = None
            # Consider Cholesky vectors separately
            # They have to be considered first!
            for inscript in inscripts[0]:
                for i, inchar in enumerate(inscript):
                    if inchar == outchar:
                        if size is None:
                            if inchar == "a":
                                size = operands[0].shape[1]
                            elif inchar == "b":
                                size = operands[1].shape[1]
                            elif inchar == "c":
                                size = operands[0].shape[2]
                            elif inchar == "d":
                                size = operands[1].shape[2]
            for inscript, array in zip(inscripts[1:], operands[2:]):
                for i, inchar in enumerate(inscript):
                    if inchar == outchar:
                        if size is None:
                            size = array.shape[i]
                        else:
                            if size != array.shape[i]:
                                raise ArgumentError(
                                    "Incompatible shapes of input operands."
                                )
            if size is None:
                raise ArgumentError("Could not determine the shape of the output.")
            outshape.append(size)
        outshape = tuple(outshape)

        # Allocate/check the output argument
        out = DenseLinalgFactory._allocate_check_output(out, outshape)

        # do the actual work
        if len(outshape) == 0:
            raise NotImplementedError
            if select in ["td"]:
                out = (
                    float(
                        CholeskyLinalgFactory.td_helper(subscripts, *operands)
                    )
                    * factor
                )
            else:
                out = (
                    float(np.einsum(subscripts + "->...", *operands)) * factor
                )
            assert isinstance(out, float)
        else:
            if clear:
                if select in ["td"]:
                    out._array[:] = CholeskyLinalgFactory.td_helper(
                        subscripts, *operands
                    )
                else:
                    raise NotImplementedError
                    # can be done without temporary
                    out._array[:] = np.einsum(subscripts, *operands)
                    # FIXME: due to a bug in numpy, we can't use the output argument
                    # for all cases. For more details, see:
                    # https://github.com/numpy/numpy/issues/5147
                    # np.einsum(subscripts, *operands, out=out._array)
                out.iscale(factor)
            else:
                # can't be done without temporary
                if select in ["td"]:
                    out._array += (
                        CholeskyLinalgFactory.td_helper(subscripts, *operands)
                        * factor
                    )
                else:
                    raise NotImplementedError
                    out._array += np.einsum(subscripts, *operands) * factor
        return out


class CholeskyFourIndex(FourIndex):
    """Cholesky symmetric four-dimensional matrix.
    """

    #
    # Constructor and destructor
    #

    def __init__(
        self, nbasis, nvec=None, array=None, array2=None, copy=False, label=""
    ):
        """
           **Arguments:**

           nbasis
                The number of basis functions.

           **Optional arguments:**

           nvec
                The number of (2-index) Cholesky vectors.

           array
                The array with Cholesky vectors, shape = (nvec, nbasis, nbasis).

           array2
                The second set of Cholesky vectors, if different from the first.

           Either nvec or array must be given (or both).
        """

        def check_array(a, name):
            if a.ndim != 3:
                raise ArgumentError(
                    f"Argument {name} has {a.ndim} dimensions, expecting 3."
                )
            if nvec is not None and nvec != a.shape[0]:
                raise ArgumentError(f"nvec does not match {name}.shape[0].")
            if not (nbasis == a.shape[1] and nbasis == a.shape[2]):
                raise ArgumentError(
                    f"nbasis does not match {name}.shape[1] or {name}.shape[2]."
                )

        if array is None:
            self._self_alloc = True
            if nvec is None:
                raise ArgumentError(
                    "Either nvec or array must be given (or both)."
                )
            if array2 is not None:
                raise ArgumentError(
                    "Argument array2 only allowed when array is given."
                )
            self._array = np.zeros([nvec, nbasis, nbasis])
            log.mem.announce(self._array.nbytes)
            self._array2 = self._array
        else:
            self._self_alloc = False
            check_array(array, "array")
            if copy:
                self._array = array.copy()
            else:
                self._array = array
            if array2 is None:
                self._array2 = self._array
            else:
                check_array(array2, "array2")
                if copy:
                    self._array2 = array2.copy()
                else:
                    self._array2 = array2

        self._label = label

    def __del__(self):
        if log is not None:
            if hasattr(self, "_array") and hasattr(self, "_self_alloc"):
                if self._self_alloc:
                    log.mem.denounce(self._array.nbytes)
                    if (
                        hasattr(self, "_array2")
                        and self._array2 is not self._array
                    ):
                        log.mem.denounce(self._array2.nbytes)
        if hasattr(self, "_array"):
            if hasattr(self, "_array2") and self._array2 is not self._array:
                del self._array2
            del self._array

    #
    # Properties
    #

    @property
    def shape(self):
        """The shape of the object"""
        return (
            self._array.shape[1],
            self._array2.shape[1],
            self._array.shape[2],
            self._array2.shape[2],
        )

    def einsum_index(self, script):
        """ Returns indices to numpy.einsum summation
            string - str of length equal to four

            Example:
            self.einsum_index('abcd') returns 'xac,xbd'
        """
        assert len(script) == 4 and isinstance(script, str)
        x = [char for char in "xijkl" if char not in script]
        return (
            x[0] + script[0] + script[2] + "," + x[0] + script[1] + script[3]
        )

    @property
    def arrays(self):
        """ Returns list containing self._array and self._array2 """
        return [self._array, self._array2]

    @property
    def label(self):
        """ Returns label of Cholesky object """
        return self._label

    #
    # Methods from base class
    #

    def __check_init_args__(self, nbasis, nvec):
        """Is self compatible with the given constructor arguments?"""
        assert self._array is not None
        assert nbasis == self.nbasis
        assert nvec == self.nvec

    def __eq__(self, other):
        """Compare self with other"""
        return (
            isinstance(other, CholeskyFourIndex)
            and other.nbasis == self.nbasis
            and other.nvec == self.nvec
            and other.is_decoupled == self.is_decoupled
            and (other._array == self._array).all()
            and (other._array2 == self._array2).all()
        )
        # Shall we also compare the label here?

    @classmethod
    def from_hdf5(cls, grp):
        """Construct an instance from data previously stored in an h5py.Group.

           **Arguments:**

           grp
                An h5py.Group object.
        """
        nvec = grp["array"].shape[0]
        nbasis = grp["array"].shape[1]
        label = grp.attrs["label"]
        result = cls(nbasis, nvec)
        grp["array"].read_direct(result._array)
        if "array2" in grp:
            result.decouple_array2()
            grp["array2"].read_direct(result._array2)
        return result

    def to_hdf5(self, grp):
        """Dump this object in an h5py.Group

           **Arguments:**

           grp
                An h5py.Group object.
        """
        grp.attrs["class"] = self.__class__.__name__
        grp["array"] = self._array
        if self._array is not self._array2:
            grp["array2"] = self._array2
        grp.attrs["label"] = self._label

    def replace_array(self, value):
        """Replaces an array with another array, if not present, it will be
           generated.
        """
        if value._array2 is not value._array:
            self._array2 = value._array2
        self._array = value._array

    def new(self):
        """Return a new four-index object with the same nbasis"""
        return CholeskyFourIndex(self.nbasis, self.nvec)

    def _check_new_init_args(self, other):
        """Check whether an already initialized object is compatible"""
        other.__check_init_args__(self.nbasis, self.nvec)

    new.__check_init_args__ = _check_new_init_args

    def clear(self):
        """Reset all elements to zero."""
        self._array[:] = 0.0
        if self._array is not self._array2:
            self._array2[:] = 0.0

    def copy(
        self,
        begin0=0,
        end0=None,
        begin1=0,
        end1=None,
        begin2=0,
        end2=None,
        begin3=0,
        end3=None,
    ):
        """Return a copy of the current four-index operator"""
        end0, end1, end2, end3 = self._fix_ends(end0, end1, end2, end3)
        assert end0 - begin0 == end1 - begin1 == end2 - begin2 == end3 - begin3
        nbasis = end0 - begin0
        result = CholeskyFourIndex(nbasis, self.nvec, label=self.label)
        result.assign(
            self, begin0, end0, begin1, end1, begin2, end2, begin3, end3
        )
        return result

    def assign(
        self,
        other,
        begin4=0,
        end4=None,
        begin5=0,
        end5=None,
        begin6=0,
        end6=None,
        begin7=0,
        end7=None,
    ):
        """Assign with the contents of another object

           **Arguments:**

           other
                Another CholeskyFourIndex object.
        """
        end4, end5, end6, end7 = other._fix_ends(end4, end5, end6, end7)
        check_type("other", other, CholeskyFourIndex)
        self._array[:] = other._array[:, begin4:end4, begin6:end6]
        if other._array is other._array2:
            self.reset_array2()
        else:
            self.decouple_array2()
            self._array2[:] = other._array2[:, begin5:end5, begin7:end7]

    def randomize(self):
        """Fill with random normal data"""
        self._array[:] = np.random.normal(0, 1, self._array.shape)
        if self.is_decoupled:
            self._array2[:] = np.random.normal(0, 1, self._array2.shape)

    def permute_basis(self, permutation):
        """Reorder the coefficients for a given permutation of basis functions.
        """
        # Easy enough but irrelevant
        raise NotImplementedError

    def change_basis_signs(self, signs):
        """Correct for different sign conventions of the basis functions."""
        # Easy enough but irrelevant
        raise NotImplementedError

    def iadd(self, other, factor):
        """This method is not supported due to the Cholesky decomposition."""
        raise NotImplementedError

    def iscale(self, factor):
        """In-place multiplication with a scalar

           **Arguments:**

           factor
                A scalar factor.
        """
        self._array *= np.sqrt(factor)

        if self._array is not self._array2:
            # arrays have been transformed
            self._array2 *= np.sqrt(factor)

    def get_element(self, i, j, k, l):
        """Return a matrix element"""
        return np.dot(self._array[:, i, k], self._array2[:, j, l])

    def set_element(self, i, j, k, l, value):
        """This method is not supported due to the Cholesky decomposition."""
        raise NotImplementedError

    #
    # Properties
    #

    @property
    def nbasis(self):
        """The number of basis functions"""
        return self._array.shape[1]

    @property
    def nvec(self):
        """The number of Cholesky vectors"""
        return self._array.shape[0]

    @property
    def is_decoupled(self):
        return self._array is not self._array2

    #
    # New methods for this implementation
    # TODO: consider adding these to base class
    #

    def decouple_array2(self):
        """Allocates a second Cholesky vector if not done yet"""
        if self._array2 is self._array:
            self._array2 = self._array.copy()
            log.mem.announce(self._array2.nbytes)

    def reset_array2(self):
        """Deallocates the second cholesky vector and sets it to match the first.
        """
        if self._array2 is not self._array:
            log.mem.denounce(self._array2.nbytes)
            self._array2 = self._array

    @timer.with_section("GetDenseFI")
    def get_dense(self, select="td"):
        """Return the DenseFourIndex equivalent. ONLY FOR TESTING. SUPER SLOW.
        """
        result = DenseFourIndex(self.nbasis)
        if select == "td":
            result._array[:] = np.tensordot(
                self._array, self._array2, axes=([0], [0])
            )
            result._array[:] = result._array.transpose(0, 2, 1, 3)
        elif select == "einsum":
            np.einsum(
                "kac,kbd->abcd", self._array, self._array2, out=result._array
            )
        else:
            raise UnknownOption(f"Unknown switch: {select}")
        return result

    def is_symmetric(self, symmetry=2, rtol=1e-5, atol=1e-8):
        """Check the symmetry of the array.

           **Optional arguments:**

           symmetry
                The symmetry to check. See :ref:`dense_matrix_symmetry`
                for more details.

           rtol and atol
                relative and absolute tolerance. See to ``np.allclose``.
        """
        if self.is_decoupled and symmetry in (2, 8):
            return False
        if symmetry in (4, 8):
            if not np.allclose(
                self._array, self._array.swapaxes(1, 2), rtol, atol
            ):
                return False
            if self.is_decoupled and not np.allclose(
                self._array2, self._array2.swapaxes(1, 2), rtol, atol
            ):
                return False
        return True

    def symmetrize(self, symmetry=8):
        check_options("symmetry", symmetry, 1, 2, 4, 8)
        if symmetry in (2, 8) and self.is_decoupled:
            # This is a different type of symmetrization than in the dense case!
            self._array[:] += self._array2
            self._array *= 0.5
            self.reset_array2()
        if symmetry in (4, 8):
            self._array[:] = self._array + self._array.transpose(0, 2, 1)
            if self.is_decoupled:
                self._array2[:] = self._array2 + self._array2.transpose(
                    0, 2, 1
                )

    def itranspose(self):
        """In-place transpose: ``0,1,2,3 -> 1,0,3,2``"""
        if self.is_decoupled:
            self._array, self._array2 = self._array2, self._array

    def sum(self):
        """Return the sum of all elements. EXPENSIVE!"""
        return np.tensordot(
            self._array, self._array2, (0, 0)
        ).sum()  # expensive!!

    def iadd_exchange(self):
        """In-place addition of its own exchange contribution"""
        raise NotImplementedError

    def slice(self, subscripts, out=None, factor=1.0, clear=True, **kwargs):
        """Returns a contraction of the four-index object.

           **Arguments:**

           subscripts
                Any of ``aabb->ab``, ``abab->ab``, ``abba->ab``

           **Optional arguments:**

           out, factor, clear
                See :py:meth:`DenseLinalgFactory.einsum`
        """
        check_options(
            "subscripts",
            subscripts,
            "aabb->ab",
            "abab->ab",
            "abba->ab",
            "abcc->bac",
            "abcc->abc",
            "abcb->abc",
            "abbc->abc",
            "abcd->abcd",
            "abcd->abdc",
            "abcd->cadb",
            "abcd->dacb",
            "abcd->acbd",
            "abcd->acdb",
        )
        begins = [kwargs.get(f"begin{i}", 0) for i in range(4)]
        begin0, begin1, begin2, begin3 = begins
        ends = [kwargs.get(f"end{i}", None) for i in range(4)]
        end0, end1, end2, end3 = self._fix_ends(*ends)

        # Handle output argument
        if out is None:
            subs_, arr_ = parse_contract_input(subscripts, [self], kwargs)
            shape = get_outshape(subs_, arr_)
            out = DenseLinalgFactory._allocate_check_output(None, shape)
        elif clear:
            out.clear()
        
        # Actual computation
        if subscripts == "aabb->ab":
            out._array[:] += factor * np.einsum(
                "xab,xab->ab",
                self._array[:, begin0:end0, begin2:end2],
                self._array2[:, begin1:end1, begin3:end3],
            )
        elif subscripts == "abab->ab":
            out._array[:] += factor * np.einsum(
                "xaa,xbb->ab",
                self._array[:, begin0:end0, begin2:end2],
                self._array2[:, begin1:end1, begin3:end3],
            )
        elif subscripts == "abba->ab":
            out._array[:] += factor * np.einsum(
                "xab,xba->ab",
                self._array[:, begin0:end0, begin2:end2],
                self._array2[:, begin1:end1, begin3:end3],
            )
        elif subscripts == "abbc->abc":
            slice_to_three_abbc_abc(
                self._array, self._array2, out._array, factor, clear
            )
        elif subscripts == "abcc->bac":
            slice_to_three_abcc_bac(
                self._array, self._array2, out._array, factor, clear
            )
        elif subscripts == "abcc->abc":
            slice_to_three_abcc_abc(
                self._array, self._array2, out._array, factor, clear
            )
        elif subscripts == "abcb->abc":
            L_r = np.diagonal(self._array2, axis1=1, axis2=2)
            out._array[:] += factor * np.tensordot(
                self._array, L_r, [(0,), (0,)]
            ).swapaxes(1, 2)
        elif subscripts == "abcd->abcd":
            out._array[:] += factor * np.tensordot(
                self._array[:, begin0:end0, begin2:end2],
                self._array2[:, begin1:end1, begin3:end3],
                axes=([0], [0]),
            ).swapaxes(1, 2)
        elif subscripts == "abcd->abdc":
            out._array[:] += factor * np.tensordot(
                self._array[:, begin0:end0, begin2:end2],
                self._array2[:, begin1:end1, begin3:end3],
                axes=([0], [0]),
            ).transpose(0, 2, 3, 1)
        elif subscripts == "abcd->cadb":
            out._array[:] += factor * np.tensordot(
                self._array[:, begin0:end0, begin2:end2],
                self._array2[:, begin1:end1, begin3:end3],
                axes=([0], [0]),
            ).transpose(1, 0, 3, 2)
        elif subscripts == "abcd->dacb":
            out._array[:] += factor * np.tensordot(
                self._array[:, begin0:end0, begin2:end2],
                self._array2[:, begin1:end1, begin3:end3],
                axes=([0], [0]),
            ).transpose(3, 0, 1, 2)
        elif subscripts == "abcd->acbd":
            out._array[:] += factor * np.tensordot(
                self._array[:, begin0:end0, begin2:end2],
                self._array2[:, begin1:end1, begin3:end3],
                axes=([0], [0]),
            )
        elif subscripts == "abcd->acdb":
            out._array[:] += factor * np.tensordot(
                self._array[:, begin0:end0, begin2:end2],
                self._array2[:, begin1:end1, begin3:end3],
                axes=([0], [0]),
            ).swapaxes(2, 3)
        return out

    def assign_four_index_transform(
        self,
        ao_integrals,
        exp0,
        exp1=None,
        exp2=None,
        exp3=None,
        method="tensordot",
    ):
        """Perform four index transformation.

           **Arguments:**

           oa_integrals
                A CholeskyFourIndex with integrals in atomic orbitals.

           exp0
                A DenseOrbital object with molecular orbitals

           **Optional arguments:**

           exp1, exp2, exp3
                Can be provided to transform each index differently. See
                ``parse_four_index_transform_exps`` for details.

           method
                Either ``einsum`` or ``tensordot`` (default).
        """
        check_type("ao_integrals", ao_integrals, CholeskyFourIndex)
        exp0, exp1, exp2, exp3 = parse_four_index_transform_exps(
            exp0, exp1, exp2, exp3, DenseOrbital
        )
        if method == "einsum":
            if ao_integrals.is_decoupled or not (
                exp0 is exp1 and exp2 is exp3
            ):
                self.decouple_array2()
                self._array2[:] = np.einsum(
                    "bi,kbd->kid", exp1.coeffs, ao_integrals._array2
                )
                self._array2[:] = np.einsum(
                    "dj,kid->kij", exp3.coeffs, self._array2
                )
            self._array[:] = np.einsum(
                "ai,kac->kic", exp0.coeffs, ao_integrals._array
            )
            self._array[:] = np.einsum("cj,kic->kij", exp2.coeffs, self._array)
        elif method == "tensordot":
            if ao_integrals.is_decoupled or not (
                exp0 is exp1 and exp2 is exp3
            ):
                self.decouple_array2()
                self._array2[:] = np.tensordot(
                    ao_integrals._array2, exp1.coeffs, axes=([1], [0])
                )
                self._array2[:] = np.tensordot(
                    self._array2, exp3.coeffs, axes=([1], [0])
                )
            self._array[:] = np.tensordot(
                ao_integrals._array, exp0.coeffs, axes=([1], [0])
            )
            self._array[:] = np.tensordot(
                self._array, exp2.coeffs, axes=([1], [0])
            )
        elif method == "blas":
            assert exp0 == exp1 == exp2 == exp3
            if ao_integrals.is_decoupled or not (
                exp0 is exp1 and exp2 is exp3
            ):
                self.decouple_array2()
                apply_four_index_transform_cpp_chol(
                    ao_integrals._array2, exp1.coeffs, self._array2
                )
            apply_four_index_transform_cpp_chol(
                ao_integrals._array, exp0.coeffs, self._array
            )
        else:
            raise UnknownOption(
                "The method must either be 'einsum' or 'tensordot' or 'blas'."
            )
