# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
""" Functions used in NIndex.contract function"""

import numpy as np
from numpy import tensordot as td
from pybest.utils import check_type

from pybest.linalg.pyslice import (
    slice_to_three_abbc_abc,
    slice_to_three_abcc_bac,
    slice_to_three_abcc_abc,
)

from pybest.linalg.pycontract import (
    contract_two_to_two_abcd_cb_ad_blas,
    contract_four_to_four_abcd_efbd_aecf,
    contract_four_to_four_abcd_efbd_aecf_blas,
)
from pybest.exceptions import ArgumentError


__all__ = [
    "td_helper",
    "cholesky_td_routine",
    "cholesky_blas_routine",
    "parse_contract_input",
    "reduce_subscript",
    "slice_output",
]


################################################################################
# Routines
################################################################################


def td_helper(subscripts, *operands):
    """ Performes numpy.tensordot operations with numpy.einsum input format.

        **Arguments**

        subscripts : str
            e.g. 'xac,xbd,defc->abfe'

        operands : numpy.ndarray
            first and second are arrays of CholeskyFourIndex or DenseNIndex,
            the third is an array of DenseNIndex object
    """
    if "->" in subscripts:
        inscripts, outscript = subscripts.split("->")
    else:
        inscripts = subscripts
        outscript = None
    scripts = inscripts.split(",")

    # Sanity check.
    error_message = f"'td' not available for operation {subscripts}."
    if len(scripts) < 2:
        raise ArgumentError(error_message)
    for char in subscripts:
        if (subscripts.count(char) > 2) and char.isalpha():
            raise ArgumentError(error_message)
    for script in scripts:
        for char in script:
            if script.count(char) > 1:
                raise ArgumentError(error_message)

    operands = list(operands)

    # Do contractions.
    path, pinfo = np.einsum_path(subscripts, *operands)

    # Sometimes einsum_path does not return a list of two-element tuples.
    # Since there is no optimized path, we provide some not-optimized path.
    if len(path) < len(operands):
        path = [(0, 1) for i in range(len(operands))]

    for step in path[1:]:
        # Take 1st and 2nd operands and corresponding subscripts
        op0 = operands.pop(step[0])
        op1 = operands.pop(step[1] - 1)
        script0 = scripts.pop(step[0])
        script1 = scripts.pop(step[1] - 1)
        # Find summation axes
        sum_ind = [i for i in script0 if i in script1]
        axis0 = [script0.index(i) for i in sum_ind]
        axis1 = [script1.index(i) for i in sum_ind]
        axis = (axis0, axis1)
        # Find default outscript
        outscript_ = [i for i in script0 if i not in script1]
        outscript_ += [i for i in script1 if i not in script0]
        outscript_ = "".join(outscript_)
        # Do contraction
        outmat = np.tensordot(op0, op1, axes=axis)
        # Update the list of operands and subscripts
        operands.append(outmat)
        scripts.append(outscript_)
    # Do transposition if required.
    if (outscript != outscript_) and (outscript != None):
        trans = tuple([outscript_.index(i) for i in outscript])
        outmat = outmat.transpose(trans)
    return outmat


def cholesky_blas_routine(subscripts, operands, out, factor, kwargs):
    """ All existing c++/blas contractions from CholeskyFourIndex class
    are collected here.

    *** Arguments ***
    * subscripts : str
    * operands : list of numpy.ndarray (cholesky must be not sliced)
    * arr : numpy.ndarray
        the result of contraction is added to arr
    * factor : float or int
    """

    op0 = operands[0]
    op1 = operands[1]
    clear = False
    arr = np.ndarray((op0.shape[1], op0.shape[1], op0.shape[1]), float)

    # Gets information how to slice CholeskyFourIndex array.
    slice_cholesky = [
        subscripts.index(j)
        for i, j in enumerate(subscripts[:4])
        if j not in subscripts[:i]
    ]

    range_ = [
        slice(
            kwargs.get("begin" + str(i), 0), kwargs.get("end" + str(i), None)
        )
        for i in slice_cholesky
    ]

    # Does slice or contraction.
    if subscripts[4:6] == "->":
        if subscripts == "abbc->abc":
            slice_to_three_abbc_abc(op0, op1, arr, factor, clear)
        elif subscripts == "abcc->bac":
            slice_to_three_abcc_bac(op0, op1, arr, factor, clear)
        elif subscripts == "abcc->abc":
            slice_to_three_abcc_abc(op0, op1, arr, factor, clear)
        out[:] += arr[tuple(range_)]

    elif subscripts == "abcd,cb->ad":
        transpose = False
        # If we do not use the full range, we have to create a copy to use blas
        array2 = op1
        if end1 != array2.shape[1] or end3 != array2.shape[2]:
            arrayT = np.ascontiguousarray(op1.transpose(2, 0, 1))
            transpose = True
        else:
            arrayT = array2
        contract_blas = contract_two_to_two_abcd_cb_ad_blas
        contract_blas(
            op0,
            arrayT,
            op1,
            arr,
            factor,
            clear,
            begin0,
            end0,
            begin1,
            end1,
            begin2,
            end2,
            begin3,
            end3,
            begin4,
            end4,
            begin5,
            end5,
            transpose,
        )
        out[:] += arr[tuple(range_)]

    elif subscripts in [
        "abcc,cb->ab",
        "abcc,ca->ab",
        "abcc,bc->ab",
        "abcc,ac->ab",
        "abcc,ab->ab",
        "abcc,cd->abd",
        "abcc,dc->dab",
    ]:
        slice_to_three_abcc_abc(op0, op1, arr, 1.0, True)
        subscripts = reduce_subscript(subscripts)
        out[:] += factor * np.einsum(
            subscripts, arr[tuple(range_)], operands[2]
        )

    else:
        raise NotImplementedError(f"{subscripts}")
    arr = None


def cholesky_td_routine(subscripts, operands_, arr, factor):
    """ All existing tensordot contractions from CholeskyFourIndex class
    are collected here.
        It is a temporary solution for the lack of td_helper method in
    CholeskyLinalgFactory class.

    *** Arguments ***
    * subscripts : str
    * operands : list of numpy.ndarray
    * arr : numpy.ndarray
        the result of contraction is added to arr
    * factor : float or int
    """
    op0 = operands_[0]
    op1 = operands_[1]
    if subscripts == "abcb->abc":
        L_r = np.diagonal(op1, axis1=1, axis2=2)
        arr[:] += factor * td(op0, L_r, [(0,), (0,)]).swapaxes(1, 2)
    elif subscripts == "abcd->abcd":
        arr[:] += factor * td(op0, op1, axes=([0], [0])).swapaxes(1, 2)
    elif subscripts == "abcd->abdc":
        arr[:] += factor * td(op0, op1, axes=([0], [0])).transpose(0, 2, 3, 1)
    elif subscripts == "abcd->cadb":
        arr[:] += factor * td(op0, op1, axes=([0], [0])).transpose(1, 0, 3, 2)
    elif subscripts == "abcd->dacb":
        arr[:] += factor * td(op0, op1, axes=([0], [0])).transpose(3, 0, 1, 2)
    elif subscripts == "abcd->acbd":
        arr[:] += factor * td(op0, op1, axes=([0], [0]))
    elif subscripts == "abcd->acdb":
        arr[:] += factor * td(op0, op1, axes=([0], [0])).swapaxes(2, 3)
    elif subscripts == "abcd,ecfd->eafb":
        for i in range(op0.shape[1]):
            tmp = td(op1, op0[:, i, :], axes=([0, 0]))
            arr[:, i, :, :] += factor * td(
                operands_[2], tmp, axes=([1, 3], [2, 1])
            )
    elif subscripts == "abcd,efcd->efab":
        for i in range(op0.shape[1]):
            tmp = td(op1, op0[:, i, :], axes=([0, 0]))
            arr[:, :, i, :] += factor * td(
                operands_[2], tmp, axes=([2, 3], [2, 1])
            )
    elif subscripts == "abcd,efcd->eafb":
        for i in range(op0.shape[1]):
            tmp = td(op1, op0[:, i, :], axes=([0, 0]))
            arr[:, i, :, :] += factor * td(
                operands_[2], tmp, axes=([2, 3], [2, 1])
            )
    elif subscripts == "abcd,ecfd->efab":
        for i in range(op0.shape[1]):
            tmp = td(op1, op0[:, i, :], axes=([0, 0]))
            arr[:, :, i, :] += factor * td(
                operands_[2], tmp, axes=([1, 3], [2, 1])
            )
    elif subscripts == "abcd,cefd->faeb":
        for i in range(op0.shape[1]):
            tmp = td(op1, op0[:, i, :], axes=([0, 0]))
            arr[:, i, :, :] += factor * td(
                operands_[2], tmp, axes=([0, 3], [2, 1])
            ).transpose((1, 0, 2))
    elif subscripts == "abcd,efcd->abef":
        for i in range(op0.shape[1]):
            tmp = td(op1, op0[:, i, :], axes=([0, 0]))
            arr[i, :, :, :] += factor * td(
                operands_[2], tmp, axes=([2, 3], [2, 1])
            ).transpose((2, 0, 1))
    elif subscripts == "abcd,cedf->aebf":
        for i in range(op0.shape[1]):
            tmp = td(op1, op0[:, i, :], axes=([0, 0]))
            arr[i, :, :, :] += factor * td(
                operands_[2], tmp, axes=([0, 2], [2, 1])
            ).transpose((0, 2, 1))
    elif subscripts == "abcd,edaf->efbc":
        for i in range(op1.shape[1]):
            tmp = td(op0, op1[:, i, :], axes=([0, 0]))
            arr[:, :, i, :] += factor * td(
                operands_[2], tmp, axes=([1, 2], [2, 0])
            )
    elif subscripts in (
        "abcd,cefd->afeb",
        "abcd,cedf->aefb",
        "abcd,cedf->abfe",
    ):
        # Should work for any contraction of type 'abcd,cefd->a...'.
        scripts, outscript = subscripts.split("->")
        trans = tuple(["efb".index(char) for char in outscript[1:]])
        # Do bcd,cedf->efb, the transpose efb and add it to output.
        for i in range(op0.shape[1]):
            tmp = td(op0[:, i, :], op1, axes=([0, 0]))
            arr[i, :, :, :] += factor * td(
                operands_[2], tmp, axes=([0, 3], [0, 2])
            ).transpose(trans)
    else:
        raise NotImplementedError(f"{subscripts}")

################################################################################
# Parser
################################################################################


def parse_contract_input(subscripts, operands, kwargs):
    """ Returns subscripts corresponding to operand._array
        and list of sliced arrays

        Arguments:
        subscripts - 'abcd,cd->ab'
        operands - list of NIndex objects without output operand
    """
    scripts, outscript = parse_subscripts(subscripts)
    assert len(scripts) == len(operands)  # Sanity check

    subscripts_ = ""
    operands_ = []
    i = 0

    for operand, script in zip(operands, scripts):
        scripts_ = operand.einsum_index(script)
        subscripts_ += scripts_ + ","
        for script_, array in zip(scripts_.split(","), operand.arrays):
            slice_ = [
                slice(
                    kwargs.get("begin" + str(i + script.find(char)), 0),
                    kwargs.get("end" + str(i + script.find(char)), None),
                    None,
                )
                for char in script_
            ]
            operands_.append(array[tuple(slice_)])
        i += len(script)

    subscripts_ = subscripts_[: len(subscripts_) - 1]
    if "->" in subscripts:
        subscripts_ += "->" + outscript

    return subscripts_, operands_


def reduce_subscript(subscript):
    from itertools import groupby

    new = "".join(c for c, _ in groupby(subscript))
    return new


def slice_output(subscripts, kwargs):
    """ Returns indices of the block of the output array based on the begin0,
    end0, begin1... keywords
    """
    inscripts, outscript = parse_subscripts(subscripts)
    outlen = len(outscript)
    start = len("".join(inscripts))
    slice_ = []
    for i in range(start, start + outlen):

        slice_.append(
            slice(
                kwargs.get("begin" + str(i), 0),
                kwargs.get("end" + str(i), None),
                None,
            )
        )
    return tuple(slice_)


def parse_subscripts(subscripts):
    """ Returns list of subscripts and outscript.

    Examples:
        'abcd,cd->ba' -> ['abcd', 'cd'], 'ba'
        'abcd,cd' -> ['abcd', 'cd'], 'ab'
        'ab,ac,cd' -> ['ab', 'ac', 'cd'], 'bd'
    """
    if "->" in subscripts:
        inscripts, outscript = subscripts.split("->")
    else:
        outscript = ""
        inscripts = subscripts
    if "," in inscripts:
        scripts = inscripts.split(",")
    else:
        scripts = [inscripts]

    if outscript == "":
        inscripts = inscripts.replace(",", "")
        for char in inscripts:
            if inscripts.count(char) == 1:
                outscript += char
    return scripts, outscript


def get_outshape(subscripts, operands):
    """ Returns shape matching the output of the operation or None if
    output is scalar.

    Arguments

    subscripts : str
        Specifies the subscripts for summation as comma separated subscript
        labels.

    operands : list of numpy.ndarray
    """

    scripts, outscript = parse_subscripts(subscripts)

    # Determine the default shape of the output
    shape = []
    for char in outscript:
        for script, operand in zip(scripts, operands):
            xdim = script.find(char)
            if xdim != -1:
                shape.append(operand.shape[xdim])
                break

    return tuple(shape)
