// PyBEST: Pythonic Black-box Electronic Structure Tool
// Copyright (C) 2016-- The PyBEST Development Team
//
// This file is part of PyBEST.
//
// PyBEST is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// PyBEST is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>
//
// --
#ifdef _OPENMP
   #include <omp.h>
#else
   #define omp_get_thread_num() 0
   #define omp_get_num_threads() 1
   #define omp_get_max_threads() 1
#endif
#include <cstring>
#include <vector>
#include "slicing.h"

void slice_to_three_abbc_abc(py::array_t<double, py::array::c_style | py::array::forcecast> inp1,
                             py::array_t<double, py::array::c_style | py::array::forcecast> inp2,
                             py::array_t<double, py::array::c_style | py::array::forcecast> &out,
                             double factor, bool clear) {
    // Check if number of dimensions of np.ndarray is correct, if not, raise ValueError
    out.mutable_unchecked<3>();
    // Create buffer pointer to input and output
    py::buffer_info bufinp1 = inp1.request();
    py::buffer_info bufinp2 = inp2.request();
    py::buffer_info bufout = out.request();
    double *ptrinp1 = (double *) bufinp1.ptr;
    double *ptrinp2 = (double *) bufinp2.ptr;
    double *ptrout = (double *) bufout.ptr;
    // Get dimensions and check for consistencies
    if (inp1.shape(0) != inp2.shape(0) || 
        inp1.shape(1) != inp2.shape(1) ||
        inp1.shape(2) != inp2.shape(2))
        throw std::runtime_error("Number of dimensions of input arguments does not match");
    if (inp1.shape(1) != out.shape(0) || 
        inp1.shape(1) != out.shape(1) ||
        inp1.shape(1) != out.shape(2))
        throw std::runtime_error("Number of dimensions of input and output arguments does not match");
    long nbasis = inp1.shape(1);
    long nvec = inp1.shape(0);
    if (clear) {
        memset(ptrout, 0, sizeof(double)*nbasis*nbasis*nbasis);
    }
    for (long k=0; k<nvec; k++){
        for (long a=0; a<nbasis; a++){
            for (long c=0; c<nbasis; c++){
                for (long b=0; b<nbasis; b++){
                    ptrout[(a*nbasis + b)*nbasis + c] +=
                        factor *
                        ptrinp1[(k*nbasis + a)*nbasis + b] *
                        ptrinp2[(k*nbasis + b)*nbasis + c];
                }
            }
        }
    }
}


void slice_to_three_abcc_bac(py::array_t<double, py::array::c_style | py::array::forcecast> inp1,
                             py::array_t<double, py::array::c_style | py::array::forcecast> inp2,
                             py::array_t<double, py::array::c_style | py::array::forcecast> &out,
                             double factor, bool clear) {
    // Check if number of dimensions of np.ndarray is correct, if not, raise ValueError
    out.mutable_unchecked<3>();
    // Create buffer pointer to input and output
    py::buffer_info bufinp1 = inp1.request();
    py::buffer_info bufinp2 = inp2.request();
    py::buffer_info bufout = out.request();
    double *ptrinp1 = (double *) bufinp1.ptr;
    double *ptrinp2 = (double *) bufinp2.ptr;
    double *ptrout = (double *) bufout.ptr;
    // Get dimensions and check for consistencies
    if (inp1.shape(0) != inp2.shape(0) || 
        inp1.shape(1) != inp2.shape(1) ||
        inp1.shape(2) != inp2.shape(2))
        throw std::runtime_error("Number of dimensions of input arguments does not match");
    if (inp1.shape(1) != out.shape(0) || 
        inp1.shape(1) != out.shape(1) ||
        inp1.shape(1) != out.shape(2))
        throw std::runtime_error("Number of dimensions of input and output arguments does not match");
    long nbasis = inp1.shape(1);
    long nvec = inp1.shape(0);
    if (clear) {
        memset(ptrout, 0, sizeof(double)*nbasis*nbasis*nbasis);
    }
    for (long k=0; k<nvec; k++){
        for (long a=0; a<nbasis; a++){
            for (long b=0; b<nbasis; b++){
                for (long c=0; c<nbasis; c++){
                    ptrout[(b*nbasis + a)*nbasis + c] +=
                        factor *
                        ptrinp1[(k*nbasis + a)*nbasis + c] *
                        ptrinp2[(k*nbasis + b)*nbasis + c];
                }
            }
        }
    }
}


void slice_to_three_abcc_abc(py::array_t<double, py::array::c_style | py::array::forcecast> inp1,
                             py::array_t<double, py::array::c_style | py::array::forcecast> inp2,
                             py::array_t<double, py::array::c_style | py::array::forcecast> &out,
                             double factor, bool clear) {
    // Check if number of dimensions of np.ndarray is correct, if not, raise ValueError
    out.mutable_unchecked<3>();
    // Create buffer pointer to input and output
    py::buffer_info bufinp1 = inp1.request();
    py::buffer_info bufinp2 = inp2.request();
    py::buffer_info bufout = out.request();
    double *ptrinp1 = (double *) bufinp1.ptr;
    double *ptrinp2 = (double *) bufinp2.ptr;
    double *ptrout = (double *) bufout.ptr;
    // Get dimensions and check for consistencies
    if (inp1.shape(0) != inp2.shape(0) || 
        inp1.shape(1) != inp2.shape(1) ||
        inp1.shape(2) != inp2.shape(2))
        throw std::runtime_error("Number of dimensions of input arguments does not match");
    if (inp1.shape(1) != out.shape(0) || 
        inp1.shape(1) != out.shape(1) ||
        inp1.shape(1) != out.shape(2))
        throw std::runtime_error("Number of dimensions of input and output arguments does not match");
    long nbasis = inp1.shape(1);
    long nvec = inp1.shape(0);

    if (clear) {
        memset(ptrout, 0, sizeof(double)*nbasis*nbasis*nbasis);
    }

    #if defined(BLAS_MKL)
        if (1 < mkl_domain_get_max_threads(MKL_DOMAIN_BLAS)){
            #if defined(_OPENMP)
                omp_set_num_threads(1);
                mkl_set_dynamic(0);
            #endif
        }
    #endif
    #if defined(_OPENMP)
        #pragma omp parallel
    #endif
{
    #if defined(_OPENMP)
        #pragma omp for
    #endif
        for (long a=0; a<nbasis; ++a){
            for (long b=0; b<nbasis; ++b){
                for (long c=0; c<nbasis; ++c){
                    double res = cblas_ddot(nvec, &(ptrinp1)[c+a*nbasis], nbasis*nbasis, &(ptrinp2)[c+b*nbasis], nbasis*nbasis);
                    ptrout[(a*nbasis + b)*nbasis + c] += factor*res;
                }
            }
        }
   
}
    #if defined(BLAS_MKL)
        if (1 < mkl_domain_get_max_threads(MKL_DOMAIN_BLAS)){
            #if defined(_OPENMP)
                mkl_set_dynamic(1);
            #endif
        }
    #endif
} 

int get_index(int r, int c, long n){
    return r*n+c-r*(r+1)/2;
}
