# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Perturbation theory module

   Variables used in this module:
    :nocc:       number of occupied orbitals in the principle configuration
    :nvirt:      number of virtual orbitals in the principle configuration
    :nbasis:     total number of basis functions
    :energy:     the energy correction, list that can contain different
                 contributions
    :amplitudes: the optimized amplitudes, list that can contain different
                 contributions
    :singles:    the singles amplitudes, bool used to determine whether singles
                 are to be included or not

    Indexing convention:
     :i,j,k,..: occupied orbitals of principle configuration
     :a,b,c,..: virtual orbitals of principle configuration
     :p,q,r,..: general indices (occupied, virtual)
"""

import numpy as np
from scipy import optimize as opt
from scipy import linalg as linalg

from pybest.auxmat import get_fock_matrix
from pybest.cache import Cache
from pybest.exceptions import (
    ConsistencyError,
    EmptyData,
    NoConvergence,
    NonEmptyData,
    SymmetryError,
    UnknownOption,
)
from pybest.featuredlists import OneBodyHamiltonian, TwoBodyHamiltonian
from pybest.linalg.base import (
    FourIndex,
    OneIndex,
    Orbital,
    ThreeIndex,
    TwoIndex,
)
from pybest.log import log, timer
from pybest.orbital_utils import transform_integrals
from pybest.pt.perturbation import Perturbation
from pybest.pt.perturbation_utils import get_pt2_amplitudes
from pybest.utils import check_options, check_type, unmask, unmask_orb

__all__ = [
    "PT2SDd",
    "PT2MDd",
    "PT2SDo",
    "PT2MDo",
    "PT2b",
]


class pCCDPT2(Perturbation):
    """Perturbation class for pCCD
    """

    def unmask_args(self, *args, **kwargs):
        # orb
        orbs = unmask_orb(*args)
        if orbs:
            orbs = orbs[0]
            self.checkpoint.update("orb_a", orbs.copy())
        # t_p
        t_p = unmask("t_p", *args, **kwargs)
        if t_p is not None:
            self.checkpoint.update("t_p", t_p)
        # 1-e ints and 2-e ints
        one = self.lf.create_two_index(label="one")
        for arg in args:
            if isinstance(arg, TwoIndex):
                if arg.label in OneBodyHamiltonian:
                    one.iadd(arg)
            elif isinstance(arg, FourIndex):
                two = arg
        # overlap
        overlap = unmask("overlap", *args, **kwargs)
        if overlap is not None:
            self.checkpoint.update("overlap", overlap)
        return one, two, orbs

    @timer.with_section("PT2Solver")
    def solve(self, *args, **kwargs):
        """Solve for energy and amplitudes

           **Arguments:**

           args
                Contains AO/MO and geminal coefficients

           **Keywords**
                :e_ref:      reference energy
                :threshold: threshold when checking symmetry of amplitudes
                :overlap:   approximate overlap used in PT2b-type PT
                :singles:   add single excitations (bool, default False)
        """

        log.cite("the PTX methods", "boguslawski2017a")

        #
        # Calculate amplitudes
        #
        amplitudes = self.calculate_amplitudes(*args, **kwargs)
        if self.singles:
            self.checkpoint.update("t_1", amplitudes[1])
        self.checkpoint.update("t_2", amplitudes[0])
        #
        # Calculate energy contributions of different seniority sectors
        # and total energy
        #
        energy = self.calculate_energy(amplitudes, *args, **kwargs)

        return energy, amplitudes

    def vfunction_0(self, *args):
        """Elements of <bcjk|H|0>.

           **Arguments:**

           args
                All function arguments needed to calculate the vector
                function:
        """
        #
        # B_jb|kc
        #
        govvo = self.get_aux_matrix("govvo")
        out = govvo.slice("abcd->acdb", factor=1.0, clear=True)
        return out

    def calculate_aux_matrix(self, mo1, mo2, *args):
        """Compute auxiliary matrices

           **Arguments:**

           args
                List of arguements. Only geminal coefficients [1], one- [mo1]
                and two-body [mo2] integrals are used.
        """
        t_p = self.checkpoint["t_p"]
        self.clear_aux_matrix()
        self.update_aux_matrix(mo1, mo2, t_p)

    def init_aux_matrix(self, select):
        """Initialize auxiliary matrices

           **Arguments:**

           select
                One of ``fock``, ``ocjbc``, ``vcjkb``, ``ocjkb``,
                ``vcjbc``, ``dcjb``, ``'ocjb``, ``vcjb``

        """
        alloc_size = {
            "fock": (self.lf.create_two_index, self.nbasis),
            "ocjbc": (
                self.lf.create_three_index,
                self.nocc,
                self.nvirt,
                self.nvirt,
            ),
            "vcjbc": (
                self.lf.create_three_index,
                self.nocc,
                self.nvirt,
                self.nvirt,
            ),
            "vcjkb": (
                self.lf.create_three_index,
                self.nocc,
                self.nocc,
                self.nvirt,
            ),
            "ocjkb": (
                self.lf.create_three_index,
                self.nocc,
                self.nocc,
                self.nvirt,
            ),
            "dcjb": (self.lf.create_two_index, self.nocc, self.nvirt),
            "ocjb": (self.lf.create_two_index, self.nocc, self.nvirt),
            "vcjb": (self.lf.create_two_index, self.nocc, self.nvirt),
            "govov": (
                self.denself.create_four_index,
                self.nocc,
                self.nvirt,
                self.nocc,
                self.nvirt,
            ),
            "govvo": (
                self.denself.create_four_index,
                self.nocc,
                self.nvirt,
                self.nvirt,
                self.nocc,
            ),
        }
        matrix, new = self._cache.load(
            select, alloc=alloc_size[select], tags="m",
        )
        if not new:
            raise NonEmptyData(
                f"The matrix {select} already exists. "
                "Call clear prior to updating the wfn."
            )
        return matrix

    @timer.with_section("UpdatingAuxMat")
    def update_aux_matrix(self, mo1, mo2, cia):
        """Derive all matrices.
           fock_pq:     one_pq + sum_m(2<pm|qm> - <pm|mq>),
           oc_jbc:      sum_m(<mm|bc> c_jm^bc),
           vc_jkb:      sum_d(<dd|jk> c_jk^bd),
           vc_jbc:      sum_d(<dd|bc> c_j^d),
           oc_jkb:      sum_m(<mm|jk> c_m^b),
           dc_jb:       sum_md(<mm|dd> c_j^d*c_k^b),
           oc_jb:       sum_i(<jb|ii> c_i^b),
           vc_jb:       sum_a(<jb|aa> c_j^a),
           g_ovov:      <ov|ov>,
           g_ovvo:      <ov|vo>,

           **Arguments:**

           mo1, mo2
                one- and two-electron integrals to be sorted.

           cia
                The geminal coefficients. A TwoIndex instance
        """
        non = self.get_range("non")
        vvo = self.get_range("vvo")
        ovv = self.get_range("ovv")
        oov = self.get_range("oov")
        ovo = self.get_range("ovo")
        vvv = self.get_range("vvv")
        ooo = self.get_range("ooo")
        ov2 = self.get_range("ov", 2)
        ov = self.get_range("ov")
        #
        # Inactive Fock matrix
        #
        auxmat1 = self.init_aux_matrix("fock")
        # need to go over three index intermediates to make it work with cholesky
        tmp = self.lf.create_three_index(self.nbasis)
        mo2.slice("abcb->abc", tmp, 2.0, True)
        mo2.slice("abbc->abc", tmp, -1.0, False)
        tmp.contract("abc->ac", auxmat1, **non)
        auxmat1.iadd(mo1)
        #
        # tmp storage
        #
        tmp = self.denself.create_two_index(self.nvirt, self.nvirt)
        # use 3-index intermediate (will be used several times)
        # This also works with Cholesky
        # tmpabc
        tmpabc = self.denself.create_three_index(self.nbasis)
        mo2.slice("abcc->abc", tmpabc, 1.0, False)
        #
        # oc_jbc = sum_m <mm|bc> c_jm^bc
        #
        auxmat2 = self.init_aux_matrix("ocjbc")
        tmpabc.contract("abc,cb->ab", cia, tmp, clear=True, **vvo)
        cia.contract("ab,bc->abc", tmp, auxmat2, select="einsum")
        # P_jm
        tmpabc.contract("abc,ca->ab", cia, tmp, clear=True, **vvo)
        cia.contract("ac,bc->abc", tmp, auxmat2, select="einsum")
        #
        # vc_jkb = sum_d <dd|jk> c_jk^bd
        #
        auxmat3 = self.init_aux_matrix("vcjkb")
        # tmp storage
        tmp = self.denself.create_two_index(self.nocc, self.nocc)
        tmpabc.contract("abc,bc->ab", cia, tmp, clear=True, **oov)
        tmp.contract("ab,ac->abc", cia, auxmat3, select="einsum")
        # P_jm
        tmpabc.contract("abc,ac->ab", cia, tmp, clear=True, **oov)
        tmp.contract("ab,bc->abc", cia, auxmat3, select="einsum")
        #
        # vc_jbc = sum_d <bc|dd> c_j^d
        #
        auxmat4 = self.init_aux_matrix("vcjbc")
        tmpabc.contract("abc,dc->dab", cia, auxmat4, clear=True, **vvv)
        #
        # oc_jkb = sum_m <mm|jk> c_m^b
        #
        auxmat5 = self.init_aux_matrix("ocjkb")
        tmpabc.contract("abc,cd->abd", cia, auxmat5, clear=True, **ooo)
        #
        # OLD:
        # dc_jb = sum_md <mm|dd> c_jm^bd
        # NEW:
        # dc_jb = sum_md <mm|dd> c_m^b c_j^d
        #
        auxmat6 = self.init_aux_matrix("dcjb")
        tmp = self.lf.create_two_index(self.nbasis, self.nbasis)
        tmp2 = self.lf.create_two_index(self.nocc, self.nocc)
        # There is a bug in np.einsum that forces us to slice first
        # Slice nevertheless to make it work with cholesky
        mo2.slice("aabb->ab", tmp, 1.0, True)
        cia.contract("ab,cb->ac", tmp, tmp2, clear=True, **ov2)
        cia.contract("ab,ca->cb", tmp2, auxmat6)
        #
        # oc_jb = sum_i <jb|ii> c_j^b
        #
        auxmat7 = self.init_aux_matrix("ocjb")
        tmpabc.contract("abc,ab->ab", cia, auxmat7, clear=True, **ovo)
        #
        # vc_jb = sum_i <jb|aa> c_j^b
        #
        auxmat8 = self.init_aux_matrix("vcjb")
        tmpabc.contract("abc,ab->ab", cia, auxmat8, clear=True, **ovv)
        #
        # g_ovov
        #
        auxmat9 = self.init_aux_matrix("govov")
        ovov = {
            "end0": self.nocc,
            "begin1": self.nocc,
            "end2": self.nocc,
            "begin3": self.nocc,
        }
        mo2.slice("abcd->abcd", auxmat9, 1.0, True, **ovov)
        #
        # g_ovvo
        #
        auxmat10 = self.init_aux_matrix("govvo")
        ovvo = {
            "end0": self.nocc,
            "begin1": self.nocc,
            "begin2": self.nocc,
            "end3": self.nocc,
        }
        mo2.slice("abcd->abcd", auxmat10, 1.0, True, **ovvo)
        del tmp, tmpabc, tmp2

    def update_ndm(self, one_dm=None):
        """Update 1-RDM

           **Optional arguments:**

           one_dm
                When provided, this 1-RDM is stored.
        """
        # FIXME
        pass

    def print_results(self, **kwargs):
        if log.do_medium:
            log(
                f"E_{self.get_model()+'(Seniority 0):':<22} "
                f"{self.energy['e_corr_s0']:> 16.12f} a.u."
            )
            log(
                f"E_{self.get_model()+'(Seniority 2):':<22} "
                f"{self.energy['e_corr_s2']:> 16.12f} a.u."
            )
            log(
                f"E_{self.get_model()+'(Seniority 4):':<22} "
                f"{self.energy['e_corr_s4']:> 16.12f} a.u."
            )
            log.hline("-")
            log(
                f"E_{self.get_model()+'(singles):':<22} {self.energy['e_corr_s']:> 16.12f} a.u."
            )
            log(
                f"E_{self.get_model()+'(doubles):':<22} {self.energy['e_corr_d']:> 16.12f} a.u."
            )
            log.hline("-")
            log(f"E_{'ref:':<22} {self.e_ref:> 16.12f} a.u.")
            log(
                f"E_{self.get_model()+':':<22} "
                f"{self.energy['e_tot']:> 16.12f} a.u."
            )
            log.hline("-")
            log(
                f"E_{'tot(d):':<22} {self.energy['e_corr_d'] + self.e_ref:> 16.12f} a.u."
            )
            if self.singles:
                log(
                    f"E_{'tot(s+d):':<22} "
                    f"{self.energy['e_tot']:> 16.12f} a.u."
                )

    def print_info(self, **kwargs):
        if log.do_medium:
            log(f"{self.get_model()} perturbation module")
            log(" ")
            log("OPTIMIZATION PARAMETERS:")
            log("Reference Function           pCCD")
            log(f"Number of pairs:             {self.nocc}")
            log(f"Number of virtuals:          {self.nvirt}")
            if self.overlap:
                olp = self.checkpoint["overlap"]
                log(f"Approximate overlap:         {olp:f}")
            log.hline()

    def check_input(self, **kwargs):
        """Check input parameters."""
        for name, value in kwargs.items():
            check_options(name, name, "e_ref", "threshold", "indextrans")

    def check_result(self, **kwargs):
        """Check if amplitudes are reasonable."""
        thresh = 1e-6

        # check symmetry of amplitudes:
        if not self.amplitudes[0].is_symmetric("cdab", atol=thresh):
            raise SymmetryError("Warning: Cluster amplitudes not symmetric.")

        # check if diagonal amplitudes are zero:
        tmp = self.amplitudes[0].slice("abab->ab")
        if tmp.sum() > thresh:
            raise ConsistencyError(
                "Warning: Diagonal cluster amplitudes not negligible."
            )

    @timer.with_section("VecFctPT2")
    def vfunction_psi0(self, *args, **kwargs):
        """Elements of <bcjk|H|pCCD>.

           **Arguments:**

           args
                All function arguments needed to calculate the vector
                function:

                * [0]:  wfn expansion coefficients
                * [1]:  geminal coefficients
        """
        check_type("args[0]", args[0], Orbital)
        t_p = self.checkpoint["t_p"]
        excludepairs = self.exclude_pairs(**kwargs)
        includefock = self.include_fock(**kwargs)
        #
        # Get ranges
        #
        vv2 = self.get_range("vv", 2)
        oo2 = self.get_range("oo", 2)
        ov2 = self.get_range("ov", 2)
        #
        # Get aux matrices
        #
        fockmod = self.get_aux_matrix("fock")
        vcjkb = self.get_aux_matrix("vcjkb")
        ocjbc = self.get_aux_matrix("ocjbc")
        vcjbc = self.get_aux_matrix("vcjbc")
        ocjkb = self.get_aux_matrix("ocjkb")
        dcjb = self.get_aux_matrix("dcjb")
        ocjb = self.get_aux_matrix("ocjb")
        vcjb = self.get_aux_matrix("vcjb")
        govov = self.get_aux_matrix("govov")
        govvo = self.get_aux_matrix("govvo")
        #
        # Get proper Fock matrix (depends on PT model used)
        #
        if includefock:
            fockmod = self.compute_modified_fock(fockmod, **kwargs)

        #
        # output
        #
        out = self.denself.create_four_index(
            self.nocc, self.nvirt, self.nocc, self.nvirt
        )
        outs = self.denself.create_two_index(self.nocc, self.nvirt)
        #
        # temporary storage
        #
        tmp4ind = self.denself.create_four_index(
            self.nocc, self.nvirt, self.nocc, self.nvirt
        )
        tmp3ind0 = self.denself.create_three_index(
            self.nocc, self.nvirt, self.nvirt
        )
        tmp3ind1 = self.denself.create_three_index(
            self.nocc, self.nocc, self.nvirt
        )

        #
        # <jk|bc>
        #
        govvo.slice("abcd->acdb", out, 1.0, True)
        #
        # c_kc <jc||bk>
        #
        govvo.contract("abcd,db->acdb", t_p, out)
        govov.contract("abcd,cb->adcb", t_p, out, factor=-1.0)
        #
        # c_jb <jc||bk>
        #
        govvo.contract("abcd,ac->acdb", t_p, out)
        govov.contract("abcd,ad->adcb", t_p, out, factor=-1.0)
        #
        # c_jc <bk|cj>
        #
        govov.contract("abcd,ab->adcb", t_p, out, factor=-1.0)
        #
        # c_kb <bk|cj>
        #
        govov.contract("abcd,cd->adcb", t_p, out, factor=-1.0)
        #
        # c_jkbc <bk|jc>
        #
        govvo.contract("abcd,ac->acdb", t_p, tmp4ind, clear=True)
        tmp4ind.contract("abcd,cd->abcd", t_p, out)
        govvo.contract("abcd,ab->acdb", t_p, tmp4ind, clear=True)
        tmp4ind.contract("abcd,cb->abcd", t_p, out)
        #
        # delta_jk [ c_jc F_bc ]
        #
        if includefock:
            t_p.contract(
                "ac,bc->abc", fockmod, tmp3ind0, **vv2, select="einsum"
            )
        #
        # delta_jk [ c_jb F_bc ]
        #
        if includefock:
            t_p.contract(
                "ab,bc->abc", fockmod, tmp3ind0, **vv2, select="einsum"
            )
        #
        # delta_jk [ oc_jbc ]
        #
        tmp3ind0.iadd(ocjbc, -1.0)
        #
        # delta_jk [ vc_jbc ]
        #
        tmp3ind0.iadd(vcjbc, 1.0)
        #
        # Add delta_jk-contribution
        #
        out.iadd_expand_three_to_four("1-3-1-2", tmp3ind0, 1.0)
        #
        # delta_bc [ c_jb F_jk ]
        #
        if includefock:
            t_p.contract(
                "ab,ac->acb",
                fockmod,
                tmp3ind1,
                factor=-1.0,
                **oo2,
                select="einsum",
            )
        #
        # delta_bc [ c_kb F_jk ]
        #
        if includefock:
            t_p.contract(
                "cb,ac->acb",
                fockmod,
                tmp3ind1,
                factor=-1.0,
                **oo2,
                select="einsum",
            )
        #
        # delta_bc [ vc_jkb ]
        #
        tmp3ind1.iadd(vcjkb, -1.0)
        #
        # delta_bc [ oc_jkb ]
        #
        tmp3ind1.iadd(ocjkb, 1.0)
        #
        # Add delta_bc-contribution
        #
        out.iadd_expand_three_to_four("0-2-0-1", tmp3ind1, 1.0)
        #
        #
        # delta_bc,jk [ (sum_md <mm|dd> c_jm c_mb) ]
        #
        out.iadd_expand_two_to_four("diag", dcjb, 1.0)
        #
        # Single excitations
        #
        #
        # F_jb*c_jb
        #
        if includefock:
            outs.assign(t_p)
            outs.imul(fockmod, 1.0, **ov2)
        #
        # F_jb
        #
        if includefock:
            outs.iadd(fockmod, 1.0, **ov2)
        #
        # (sum_a <jb|aa> c_ja)
        #
        outs.iadd(vcjb)
        #
        # (sum_i <jb|ii> c_ib)
        #
        outs.iadd(ocjb, -1.0)

        # Get rid of pair amplitudes:
        if excludepairs:
            ind1, ind2 = np.indices((self.nocc, self.nvirt))
            indices = [ind1, ind2, ind1, ind2]
            out.assign(0.0, indices)

        return [out, outs]

    def compute_modified_fock(self, oldfock, **kwargs):
        raise NotImplementedError

    def callback(self, x, f):
        """Prints progress of convergence and dumps current solution"""
        #
        # Convergence progress
        #
        norm2 = linalg.norm(f)
        normmax = linalg.norm(f, ord=np.inf)
        if log.do_medium:
            log(f"{norm2:< 18.12f} {normmax:< 18.12f}")
        #
        # Dump current solution
        #
        begin = 0
        if self.singles:
            begin = self.nocc * self.nvirt
            self.checkpoint.update("t_1", x[:begin])
            self.checkpoint.update("t_2", x[begin:])
        else:
            self.checkpoint.update("t_2", x[begin:])
        checkpoint_fn = self.get_checkpoint_fn()
        self.checkpoint.to_file(checkpoint_fn)


class pCCDPT2D(pCCDPT2):
    """Diagonal one-body zero-order Hamiltonian
    """

    @timer.with_section("PT2DAmplitudes")
    def calculate_amplitudes(self, *args, **kwargs):
        """Calculate amplitudes (singles and doubles)

           **Returns:**
                List of double and single amplitudes sorted as [D,S]
        """
        #
        # By default, calculate singles
        #
        self.singles = True
        #
        # Get "excitation" matrices w.r.t. |pCCD> (psi0): <jbkc|VN|pCCD>
        #
        matrix = self.vfunction_psi0(*args, **kwargs)

        fock = self.get_aux_matrix("fock")
        #
        # Get diagonal part of Fock matrix
        #
        fdiago = fock.copy_diagonal(end=self.nocc)
        fdiagv = fock.copy_diagonal(begin=self.nocc)

        ts, td = get_pt2_amplitudes(
            self.denself,
            [matrix[1], matrix[0]],
            [fdiago, fdiagv],
            singles=True,
        )
        return [td, ts]

    def calculate_energy(self, amplitudes, *args, **kwargs):
        """Calculate PT energy and energy contribution of seniority sectors.
           Depends on PT model used.

           **Arguments:**

           amplitudes
                PT amplitudes. A FourIndex instance

           args
                List containing the geminal coefficients, 1- and 2-el
                integrals, and auxiliary matrices
        """
        #
        # Get "excitation" matrices w.r.t. |HF> (0)
        #
        exjbkc0 = self.calculate_exmatrix(*args, **kwargs)
        #
        # Single excitations:
        #
        e_singles = self.calculate_singles(
            amplitudes, exjbkc0, *args, **kwargs
        )
        #
        # Seniority-0 sector
        #
        e_seniority_0 = amplitudes[0].contract("abab,abab", exjbkc0[0])
        #
        # Seniority-2 sector
        #
        e_seniority_2 = amplitudes[0].contract("abad,abad", exjbkc0[0])
        e_seniority_2 += amplitudes[0].contract("abdb,abdb", exjbkc0[0])
        energy = amplitudes[0].contract("abcd,abcd", exjbkc0[0]) * 2.0
        energy -= amplitudes[0].contract("abcd,adcb", exjbkc0[0])
        #
        # Seniority-4 sector
        #
        e_seniority_4 = energy - e_seniority_2 - e_seniority_0
        #
        # Update for checkpointing
        #
        energy_dict = {}
        energy_dict.update({"e_corr_d": energy})
        energy_dict.update({"e_corr_s": e_singles})
        energy_dict.update({"e_corr_s0": e_seniority_0})
        energy_dict.update({"e_corr_s2": e_seniority_2})
        energy_dict.update({"e_corr_s4": e_seniority_4})
        e_corr = energy + e_singles
        e_tot = e_corr + self.e_ref
        energy_dict.update({"e_tot": e_tot})
        energy_dict.update({"e_corr": e_corr})
        #
        # Force update singles
        #
        return energy_dict


class PT2SDd(pCCDPT2D):
    """Diagonal one-body zero-order Hamiltonian with single determinant as dual
    """

    def overlap(self):
        return False

    def calculate_singles(self, amplitudes, exjbkc0, *args, **kwargs):
        """Calculate energy contribution of singles

           **Arguments:**

           amplitudes
                PT amplitudes. A FourIndex instance

           exjbkc0
                Excitation matrix (not needed here). A FourIndex instance

           args
                List containing the geminal coefficients, 1- and 2-el
                integrals, and auxiliary matrices
        """
        fock = self.get_aux_matrix("fock")
        ov = self.get_range("ov", 2)
        #
        # Single excitations:
        #
        e_singles = 2.0 * amplitudes[1].contract("ab,ab", fock, **ov)
        return e_singles

    def calculate_exmatrix(self, *args, **kwargs):
        """Calculate "excitation" matrices w.r.t. |HF> (0)

           **Arguments:**

           args
                List containing the geminal coefficients, 1- and 2-el
                integrals, and auxiliary matrices
        """
        exjbkc0 = self.vfunction_0(*args)
        return [exjbkc0]

    def compute_modified_fock(self, oldfock, **kwargs):
        newfock = oldfock.copy()
        newfock.assign_diagonal(0.0)
        return newfock

    def include_fock(self, **kwargs):
        return True

    def exclude_pairs(self, **kwargs):
        return True

    def get_model(self):
        return "PT2SDd"


class PT2MDd(pCCDPT2D):
    """Diagonal one-body zero-order Hamiltonian with pCCD determinant as dual
    """

    def overlap(self):
        return True

    def calculate_singles(self, amplitudes, exjbkc0, *args, **kwargs):
        """Calculate energy contribution of singles

           **Arguments:**

           amplitudes
                PT amplitudes. A FourIndex instance

           exjbkc0
                Excitation matrix (not needed here). A FourIndex instance

           args
                List containing the geminal coefficients, 1- and 2-el
                integrals, and auxiliary matrices
        """
        #
        # Single excitations:
        #
        e_singles = 2.0 * amplitudes[1].contract("ab,ab", exjbkc0[1])
        return e_singles

    def calculate_exmatrix(self, *args, **kwargs):
        """Calculate "excitation" matrices w.r.t. |pCCD> (0)

           **Arguments:**

           args
                List containing the geminal coefficients, 1- and 2-el
                integrals, and auxiliary matrices
        """
        #
        # Get "excitation" matrices w.r.t. |psi_0> (0)
        #
        exjbkc0 = self.vfunction_psi0(*args, **kwargs)
        return exjbkc0

    def compute_modified_fock(self, oldfock, **kwargs):
        """Scale diagonal of Fock matrix using wfn overlap"""
        newfock = oldfock.copy()
        olp = self.checkpoint["overlap"]
        # copy diagonal of old fock and scale
        diagfock = oldfock.copy_diagonal()
        scale = 1 - 1 / olp
        diagfock.iscale(scale)
        # assign new diagonal
        newfock.assign_diagonal(diagfock)
        return newfock

    def check_input(self, **kwargs):
        """Check input parameters."""
        for name, value in kwargs.items():
            check_options(
                name, name, "e_ref", "threshold", "indextrans", "overlap"
            )

    def include_fock(self, **kwargs):
        return True

    def exclude_pairs(self, **kwargs):
        return True

    def get_model(self):
        return "PT2MDd"


class pCCDPT2O(pCCDPT2):
    @timer.with_section("PT2OSolver")
    def calculate_amplitudes(self, *args, **kwargs):
        """Solve for amplitudes

           **Arguments:**

           args
                Contains geminal coefficients, 1- and 2-el integrals, and
                auxiliary matrices

           **Keywords**
                :e_ref: reference energy
                :ecore: core energy
                :threshold: threshold when checking symmetry of amplitudes
                :guess: initial guess

           For more details, see :py:meth:`Pertubation.__call__`
        """
        if log.do_medium:
            log(f"{'residual 2-norm':<18s} {'residual max norm':<18s}")

        #
        # Check argument type prior optimization:
        #
        check_type("args[0]", args[0], Orbital)
        #
        # Keyword arguments
        #
        thresh = kwargs.get("threshold", 1e-7)
        maxiter = kwargs.get("maxiter", 200)
        initguess = kwargs.get("guess", None)
        singles = kwargs.get("singles", False)
        excludepairs = self.exclude_pairs(**kwargs)
        includefock = self.include_fock(**kwargs)

        #
        # Update PT parameters
        #
        self.singles = singles
        #
        #
        # Append overlap to function arguments:
        #
        args += ((self.checkpoint["t_p"]),)
        args += ((self.checkpoint["overlap"]),)
        args += ((excludepairs),)
        check_type("args[2]", args[2], float)
        check_type("args[4]", args[3], bool)

        #
        # Get initial guess:
        #
        if initguess is not None:
            guess = initguess
            guess = self.fix_guess(initguess, singles, excludepairs)
        else:
            guess = self.get_guess(singles, excludepairs)

        #
        # Solve matrix equation using scipy.optimize.root routine:
        #
        amplitudes = opt.root(
            self.vfunction,
            guess,
            args=(args),
            method="krylov",
            callback=self.callback,
            options={"fatol": thresh, "maxiter": maxiter},
        )
        if not amplitudes.success:
            raise NoConvergence(
                f"ERROR: program terminated. Error in solving \
                              amplitude equations: {amplitudes.message}"
            )
        if log.do_medium:
            log(
                f"Optimization of cluster amplitudes converged in {amplitudes.nit} "
                f"iterations."
            )
            log.hline("~")
            log("Calculating energy correction:")

        #
        # Store amplitudes
        #
        t = self.amplitudes2index(amplitudes.x)

        return t

    def amplitudes2index(self, t):
        """Transforms the np.array into dense objects"""
        end = self.nocc * self.nvirt * (self.nocc * self.nvirt + 1) // 2
        td = self.denself.create_four_index(
            self.nocc, self.nvirt, self.nocc, self.nvirt
        )
        # assign triu block
        td.assign_triu(t[:end])
        # add missing block
        tp_ = td.slice("abab->ab")
        td.iadd_transpose((2, 3, 0, 1))
        ind1, ind2 = np.indices((self.nocc, self.nvirt))
        indp = [ind1, ind2, ind1, ind2]
        # assign diagonal
        td.assign(tp_, indp)
        if self.singles:
            ts = self.denself.create_two_index(self.nocc, self.nvirt)
            ts.assign(t[end:])
            return [td, ts]
        return [td]

    @timer.with_section("VecFctPT2O")
    def vfunction(self, amplitudes, *args):
        """.

           **Arguments:**

           amplitudes
                Cluster amplitudes. Need to be determined.

           args
                All function arguments needed to calculated the vector
                function:

                * [0]:  wfn expansion coefficients
                * [1]:  geminal coefficients
                * [2]:  overlap (float)
                * [3]:  singles (bool)
                * [4]:  pairs (bool)
        """
        ov = self.get_range("ov")
        vv2 = self.get_range("vv", 2)
        oo2 = self.get_range("oo", 2)
        oo4 = self.get_range("oo", 4)
        ov4 = self.get_range("ov", 4)
        vv4 = self.get_range("vv", 4)
        excludepairs = args[3]
        fock = self.get_aux_matrix("fock")
        #
        # output
        #
        matrix = self.vfunction_psi0(
            *args, **{"overlap": args[2], "excludepairs": excludepairs}
        )
        out = matrix[0]
        outs = matrix[1]
        #
        # Scale due to P_ijab operation
        #
        out.iscale(0.5)
        #
        # PT amplitudes
        #
        end = self.nocc * self.nvirt * (self.nocc * self.nvirt + 1) // 2
        ptamplitudes = self.denself.create_four_index(
            self.nocc, self.nvirt, self.nocc, self.nvirt
        )
        #
        # Expand triangular blocks
        #
        ptamplitudes.assign_triu(amplitudes[:end])
        tp_ = ptamplitudes.slice("abab->ab")
        ptamplitudes.iadd_transpose((2, 3, 0, 1))
        ind1, ind2 = np.indices((self.nocc, self.nvirt))
        indp = [ind1, ind2, ind1, ind2]
        # assign diagonal
        ptamplitudes.assign(tp_, indp)
        if self.singles:
            samplitudes = self.lf.create_two_index(self.nocc, self.nvirt)
            samplitudes.assign(amplitudes[end:])

        # Singles only
        if self.singles:
            #
            #
            #
            samplitudes.contract("ab,bc->ac", fock, outs, **vv2)
            #
            #
            #
            samplitudes.contract("ab,ac->cb", fock, outs, factor=-1.0, **oo2)

        # coupling singles-doubles
        if self.singles:
            #
            #
            #
            ptamplitudes.contract("abcd,ab->cd", fock, outs, **ov4)
            # P_jb,kc
            ptamplitudes.contract("abcd,cd->ab", fock, outs, **ov4)
            #
            #
            #
            ptamplitudes.contract(
                "abcd,cb->ad", fock, outs, factor=-0.5, **ov4
            )
            # P_jb,kc
            ptamplitudes.contract(
                "abcd,ad->cb", fock, outs, factor=-0.5, **ov4
            )

        # coupling doubles-singles
        if self.singles:
            #
            #
            #
            fock.contract(
                "ab,cd->abcd", samplitudes, out, **ov, select="einsum"
            )

        # Doubles only
        # sum_aifo A^jbkc_iaof
        #
        # sum_c F_ac t_icjb
        #
        ptamplitudes.contract("abcd,eb->aecd", fock, out, factor=0.5, **vv4)
        #
        # sum_c F_ac t_jbic
        #
        ptamplitudes.contract("abcd,ed->ceab", fock, out, factor=0.5, **vv4)
        #
        # sum_l F_lj t_lbia
        #
        ptamplitudes.contract("abcd,ae->cdeb", fock, out, factor=-0.5, **oo4)
        #
        # sum_l F_lj t_ialb
        #
        ptamplitudes.contract("abcd,ce->edab", fock, out, factor=-0.5, **oo4)

        #
        # P_ijab
        #
        out.iadd_transpose((2, 3, 0, 1))
        # Get rid of pair amplitudes:
        if excludepairs:
            ind1, ind2 = np.indices((self.nocc, self.nvirt))
            indices = [ind1, ind2, ind1, ind2]
            out.assign(0.0, indices)

        if self.singles:
            return np.hstack(
                (out.get_triu().ravel(order="C"), outs._array.ravel(order="C"))
            )
        return out.get_triu().ravel(order="C")

    def get_guess(self, singles, excludepairs):
        """Generate initial guess for amplitudes. By default, an MP2 guess
           is calculated.
        """
        #
        # Get auxmatrices
        #
        iajb = self.get_aux_matrix("govov")
        liajb = iajb.copy()
        liajb.iadd_transpose((0, 3, 2, 1), -1.0)
        liajb.iadd(iajb)
        fock = self.get_aux_matrix("fock")
        #
        # Get diagonal part of Fock matrix
        #
        fdiago = fock.copy_diagonal(end=self.nocc)
        fdiagv = fock.copy_diagonal(begin=self.nocc)

        ts, td = get_pt2_amplitudes(
            self.denself,
            [fock, liajb],
            [fdiago, fdiagv],
            singles=True,
            lagrange=False,
        )
        if excludepairs:
            ind1, ind2 = np.indices((self.nocc, self.nvirt))
            indp = [ind1, ind2, ind1, ind2]
            td.assign(0.0, indp)
        # get only upper triangular block
        if self.singles:
            return np.hstack(
                (td.get_triu().ravel(order="C"), ts.ravel()._array)
            )
        return td.get_triu().ravel(order="C")

    def fix_guess(self, guess, singles, excludepairs):
        """Generate initial guess for amplitudes"""
        if singles:
            endd = self.nocc * self.nvirt * (self.nocc * self.nvirt + 1) // 2
            ends = self.nocc * self.nvirt
            end = endd + ends
            if guess.ndim != 1:
                raise UnknownOption(
                    f"Guess has to be a 1-dimensional vector containing first doubles "
                    f"amplitudes, then singles"
                )
        else:
            end = self.nocc * self.nvirt * (self.nocc * self.nvirt + 1) // 2
            if guess.ndim != 1:
                raise UnknownOption(
                    "Guess has to be a 1-dimensional vector containing doubles amplitudes"
                )
        if guess.shape[0] == end:
            return guess
        else:
            # assuming that guess is a dense four index object
            endd = self.nocc * self.nvirt * self.nocc * self.nvirt
            guessd = self.denself.create_four_index(
                self.nocc, self.nvirt, self.nocc, self.nvirt
            )
            #
            # Expand triangular blocks
            #
            guessd.assign(guess[:endd])
            guessd.iadd_transpose((2, 3, 0, 1))
            guessd.iscale(0.5)
            if excludepairs:
                ind1, ind2 = np.indices((self.nocc, self.nvirt))
                indp = [ind1, ind2, ind1, ind2]
                guessd.assign(0.0, indp)
            # get only upper triangular block
            if self.singles:
                return np.hstack((guessd.get_triu(), guess[endd:]))
            return guessd.get_triu()

    def print_info(self, **kwargs):
        thresh = kwargs.get("threshold", 1e-7)
        maxiter = kwargs.get("maxiter", 200)
        guess = kwargs.get("guess", None)
        singles = kwargs.get("singles", False)
        exclude = self.exclude_pairs(**kwargs)
        if log.do_medium:
            log(f"{self.get_model()} perturbation module")
            log(" ")
            log("OPTIMIZATION PARAMETERS:")
            log("Reference Function           pCCD")
            log(f"Number of pairs:             {self.nocc}")
            log(f"Number of virtuals:          {self.nvirt}")
            if self.overlap:
                log(
                    f"Approximate overlap:         {self.checkpoint['overlap']}"
                )
            if singles:
                log("Projection manifold          T1 + T2")
            else:
                log("Projection manifold          T2")
            if exclude:
                log("  Electron pairs             excluded")
            else:
                log("  Electron pairs             included")
            if guess is None:
                log("Initial guess:               MP2")
            else:
                log("Initial guess:               user-defined")
            log("Solvers:")
            log(f"  {self.get_model()+'amplitudes:':<26s} krylov")
            log("Optimization thresholds:")
            log(f"  {self.get_model()+'amplitudes:':<26s} {thresh:.2e}")
            log(f"  maxiter:                   {maxiter}")
            log.hline()

    def check_input(self, **kwargs):
        """Check input parameters."""
        for name, value in kwargs.items():
            check_options(
                name,
                name,
                "e_ref",
                "threshold",
                "maxiter",
                "singles",
                "guess",
                "indextrans",
            )
        guess = kwargs.get("guess", None)
        if guess is not None:
            check_type("guess", guess, np.ndarray)
            if (
                not len(guess)
                == self.nocc * self.nocc * self.nvirt * self.nvirt
            ):
                raise ConsistencyError(
                    "Length of guess array does not agree with number of unknowns"
                )

    def check_result(self, **kwargs):
        """Check if amplitudes are reasonable."""
        thresh = 1e-6

        # check symmetry of amplitudes:
        if not self.amplitudes[0].is_symmetric("cdab", atol=thresh):
            raise SymmetryError("Warning: Cluster amplitudes not symmetric!")

    def calculate_energy(self, amplitudes, *args, **kwargs):
        """Calculate PT energy and energy contribution of seniority sectors

           **Arguments:**

           amplitudes
                PT amplitudes. A FourIndex instance

           args
                List containing the geminal coefficients, 1- and 2-el
                integrals, and auxiliary matrices
        """
        #
        # Get "excitation" matrices w.r.t. |HF> (0)
        #
        exjbkc0 = self.calculate_exmatrix(*args, **kwargs)
        #
        # Single excitations:
        #
        e_singles = self.calculate_singles(
            amplitudes, exjbkc0, *args, **kwargs
        )
        #
        # Seniority-0 sector
        #
        e_seniority_0 = amplitudes[0].contract("abab,abab", exjbkc0[0])
        #
        # Seniority-2 sector
        #
        e_seniority_2 = amplitudes[0].contract("abad,abad", exjbkc0[0])
        e_seniority_2 += amplitudes[0].contract("abdb,abdb", exjbkc0[0])
        energy = amplitudes[0].contract("abcd,abcd", exjbkc0[0]) * 2.0
        energy -= amplitudes[0].contract("abcd,adcb", exjbkc0[0])
        #
        # Seniority-4 sector
        #
        e_seniority_4 = energy - e_seniority_2 - e_seniority_0
        #
        # Update for checkpointing
        #
        energy_dict = {}
        energy_dict.update({"e_corr_d": energy})
        energy_dict.update({"e_corr_s": e_singles})
        energy_dict.update({"e_corr_s0": e_seniority_0})
        energy_dict.update({"e_corr_s2": e_seniority_2})
        energy_dict.update({"e_corr_s4": e_seniority_4})
        e_corr = energy + e_singles
        e_tot = e_corr + self.e_ref
        energy_dict.update({"e_tot": e_tot})
        energy_dict.update({"e_corr": e_corr})

        return energy_dict


class PT2SDo(pCCDPT2O):
    def overlap(self):
        return False

    def calculate_singles(self, amplitudes, exjbkc0, *args, **kwargs):
        """Calculate energy contribution of singles
        """
        return 0.0

    def calculate_exmatrix(self, *args, **kwargs):
        """Calculate "excitation" matrices w.r.t. |HF> (0)

           **Arguments:**

           args
                List containing the geminal coefficients, 1- and 2-el
                integrals, and auxiliary matrices
        """
        exjbkc0 = self.vfunction_0(*args)
        return [exjbkc0]

    def get_model(self):
        return "PT2SDo"

    def include_fock(self, **kwargs):
        return False

    def exclude_pairs(self, **kwargs):
        return True


class PT2MDo(pCCDPT2O):
    def overlap(self):
        return True

    def calculate_singles(self, amplitudes, exjbkc0, *args, **kwargs):
        """Calculate energy contribution of singles

           **Arguments:**

           amplitudes
                PT amplitudes. A FourIndex instance

           exjbkc0
                Excitation matrix (not needed here). A FourIndex instance

           args
                List containing the geminal coefficients, 1- and 2-el
                integrals, and auxiliary matrices
        """
        #
        # Single excitations:
        #
        if self.singles:
            e_singles = 2.0 * amplitudes[1].contract("ab,ab", exjbkc0[1])
            return e_singles
        return 0.0

    def calculate_exmatrix(self, *args, **kwargs):
        """Calculate "excitation" matrices w.r.t. |pCCD> (0)

           **Arguments:**

           args
                List containing the geminal coefficients, 1- and 2-el
                integrals, and auxiliary matrices
        """
        #
        # Get "excitation" matrices w.r.t. |psi_0> (0)
        #
        exjbkc0 = self.vfunction_psi0(*args, **kwargs)
        return exjbkc0

    def compute_modified_fock(self, oldfock, **kwargs):
        newfock = oldfock.copy()
        olp = self.checkpoint["overlap"]
        scale = 1 - 1 / olp
        newfock.iscale(scale)
        return newfock

    def check_input(self, **kwargs):
        """Check input parameters."""
        for name, value in kwargs.items():
            check_options(
                name,
                name,
                "e_ref",
                "threshold",
                "indextrans",
                "overlap",
                "singles",
                "guess",
            )
        try:
            olp = self.checkpoint["overlap"]
        except KeyError:
            raise EmptyData("Warning: Cannot find wavefunction overlap!")

    def get_model(self):
        return "PT2MDo"

    def include_fock(self, **kwargs):
        return True

    def exclude_pairs(self, **kwargs):
        return True


class PT2b(pCCDPT2O):
    def overlap(self):
        return False

    def calculate_singles(self, amplitudes, exjbkc0, *args, **kwargs):
        """Calculate energy contribution of singles

           **Arguments:**

           amplitudes
                PT amplitudes. A FourIndex instance

           exjbkc0
                Excitation matrix (not needed here). A FourIndex instance

           args
                List containing the geminal coefficients
        """
        #
        # Single excitations:
        #
        if self.singles:
            e_singles = 2.0 * amplitudes[1].contract("ab,ab", exjbkc0[1])
            return e_singles
        return 0.0

    def calculate_exmatrix(self, *args, **kwargs):
        """Calculate "excitation" matrices w.r.t. |pCCD> (0)

           **Arguments:**

           args
                List containing the geminal coefficients, 1- and 2-el
                integrals, and auxiliary matrices
        """
        #
        # Get "excitation" matrices w.r.t. |psi_0> (0)
        #
        exjbkc0 = self.vfunction_psi0(*args, **kwargs)
        return exjbkc0

    def compute_modified_fock(self, oldfock, **kwargs):
        return oldfock

    def check_input(self, **kwargs):
        """Check input parameters."""
        for name, value in kwargs.items():
            check_options(
                name,
                name,
                "e_ref",
                "threshold",
                "indextrans",
                "singles",
                "excludepairs",
                "guess",
            )

    def get_model(self):
        return "PT2b"

    def exclude_pairs(self, **kwargs):
        exclude = kwargs.get("excludepairs", False)
        return exclude

    def include_fock(self, **kwargs):
        return True
