# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file

import numpy as np
import pytest

from pybest.pt.mp2 import RMP2
from pybest.io.iodata import IOData
from pybest.context import context
from pybest.pt.test.common import prepare_mol


def prepare_mp2(
    lf,
    occ_model,
    one,
    er,
    orb_a,
    olp,
    e_ref,
    e_core,
    fos=1.0,
    fss=1.0,
    natorb=False,
    relaxation=False,
    passorbs=False,
    ncore=0,
):
    hf = IOData(e_tot=e_ref, orb_a=orb_a.copy(), olp=olp, e_core=e_core)
    pt2 = RMP2(lf, occ_model, ncore=ncore)
    if passorbs:
        mp2 = pt2(
            one,
            er,
            orb_a,
            hf,
            fos=fos,
            fss=fss,
            natorb=natorb,
            relaxation=relaxation,
        )
    else:
        mp2 = pt2(
            one, er, hf, fos=fos, fss=fss, natorb=natorb, relaxation=relaxation
        )
    return mp2, hf


test_cases = [
    (
        "H2O",
        "cc-pvdz",
        0,
        {
            "mol": "h2o_ccdz",
            "orb_file": "h2o_ccdz_hf.txt",
            "e_ref": -76.02644658,
            "e_tot": -76.23075702,
        },
    ),
    (
        "H2O",
        "cc-pvdz",
        1,
        {
            "mol": "h2o_ccdz",
            "orb_file": "h2o_ccdz_hf.txt",
            "e_ref": -76.02644658,
            "e_tot": -76.22843001,
        },
    ),
]


@pytest.mark.parametrize("mol, basis, ncore, dict_", test_cases)
def test_MP2(mol, basis, ncore, dict_):
    lf, olp, orb_a, one, er, external, occ_model = prepare_mol(
        basis, f"test/{dict_['mol']}.xyz", f"test/{dict_['orb_file']}"
    )

    mp2, hf = prepare_mp2(
        lf,
        occ_model,
        one,
        er,
        orb_a,
        olp,
        dict_["e_ref"],
        external,
        fos=1.0,
        fss=1.0,
        ncore=ncore,
    )
    assert abs(mp2.e_tot - dict_["e_tot"]) < 1e-6


def test_MP2_update_orb():
    lf, olp, orb_a, one, er, external, occ_model = prepare_mol(
        "cc-pvdz", "test/h2o_ccdz.xyz", "test/h2o_ccdz_hf.txt"
    )
    eref = -76.02644658
    orb_ac = orb_a.copy()

    mp2, hf = prepare_mp2(
        lf,
        occ_model,
        one,
        er,
        orb_a,
        olp,
        eref,
        external,
        fos=1.0,
        fss=1.0,
        natorb=True,
        relaxation=True,
    )
    assert abs(mp2.e_tot - -76.23075702) < 1e-6
    assert np.allclose(hf.orb_a.coeffs, orb_ac.coeffs)
    assert np.allclose(hf.orb_a.energies, orb_ac.energies)
    assert np.allclose(hf.orb_a.occupations, orb_ac.occupations)

    mp2, hf = prepare_mp2(
        lf,
        occ_model,
        one,
        er,
        orb_a,
        olp,
        eref,
        external,
        fos=1.0,
        fss=1.0,
        natorb=True,
        relaxation=True,
        passorbs=True,
    )
    assert abs(mp2.e_tot - -76.23075702) < 1e-6
    assert np.allclose(hf.orb_a.coeffs, orb_ac.coeffs)
    assert np.allclose(hf.orb_a.energies, orb_ac.energies)
    assert np.allclose(hf.orb_a.occupations, orb_ac.occupations)
    assert np.allclose(mp2.orb_a.coeffs, orb_a.coeffs)
    assert np.allclose(mp2.orb_a.energies, orb_a.energies)
    assert np.allclose(mp2.orb_a.occupations, orb_a.occupations)


def test_hf_MP2_fos_d():
    lf, olp, orb_a, one, er, external, occ_model = prepare_mol(
        "cc-pvdz", "test/h2o_ccdz.xyz", "test/h2o_ccdz_hf.txt"
    )
    eref = -76.02644658

    mp2, hf = prepare_mp2(
        lf, occ_model, one, er, orb_a, olp, eref, external, fos=1.0, fss=0.0
    )
    assert abs(mp2.e_corr_os - -0.15274109) < 1e-6

    mp2, hf = prepare_mp2(
        lf,
        occ_model,
        one,
        er,
        orb_a,
        olp,
        eref,
        external,
        fos=1.0,
        fss=0.0,
        natorb=True,
        relaxation=True,
    )
    fn_dm = context.get_fn("test/h2o_ccdz_dm_mp2_os.txt")
    one_dm_ref = np.fromfile(fn_dm, sep=",")
    assert abs(mp2.e_corr_os - -0.15274109) < 1e-6
    assert np.allclose(one_dm_ref, mp2.dm_1._array.diagonal() * 2.0)
    assert abs(np.sum(mp2.dm_1._array.diagonal() * 2.0) - 10.0) < 1e-8


def test_hf_MP2_fss_d():
    lf, olp, orb_a, one, er, external, occ_model = prepare_mol(
        "cc-pvdz", "test/h2o_ccdz.xyz", "test/h2o_ccdz_hf.txt"
    )
    eref = -76.02644658

    mp2, hf = prepare_mp2(
        lf, occ_model, one, er, orb_a, olp, eref, external, fos=0.0, fss=1.0
    )
    assert abs(mp2.e_corr_ss - -0.05156935) < 1e-6

    mp2, hf = prepare_mp2(
        lf,
        occ_model,
        one,
        er,
        orb_a,
        olp,
        eref,
        external,
        fos=0.0,
        fss=1.0,
        natorb=True,
        relaxation=True,
    )
    fn_dm = context.get_fn("test/h2o_ccdz_dm_mp2_ss.txt")
    one_dm_ref = np.fromfile(fn_dm, sep=",")
    assert abs(mp2.e_corr_ss - -0.05156935) < 1e-6
    assert np.allclose(one_dm_ref, mp2.dm_1._array.diagonal() * 2.0)
    assert abs(np.sum(mp2.dm_1._array.diagonal() * 2.0) - 10.0) < 1e-8


def test_hf_MP2_fos_fss_d():
    lf, olp, orb_a, one, er, external, occ_model = prepare_mol(
        "cc-pvdz", "test/h2o_ccdz.xyz", "test/h2o_ccdz_hf.txt"
    )
    eref = -76.02644658

    mp2, hf = prepare_mp2(
        lf, occ_model, one, er, orb_a, olp, eref, external, fos=0.3, fss=1.2
    )
    assert abs(mp2.e_corr - -0.10770555) < 1e-6
    assert abs(mp2.e_corr_ss - -0.06188322) < 1e-6
    assert abs(mp2.e_corr_os - -0.04582233) < 1e-6


def test_hf_MP2_scs_defaults_d():
    lf, olp, orb_a, one, er, external, occ_model = prepare_mol(
        "cc-pvdz", "test/h2o_ccdz.xyz", "test/h2o_ccdz_hf.txt"
    )
    eref = -76.02644658

    mp2, hf = prepare_mp2(
        lf, occ_model, one, er, orb_a, olp, eref, external, fos=1.2, fss=1 / 3
    )
    assert abs(mp2.e_corr - -0.200479095072) < 1e-6
    assert abs(mp2.e_corr_ss - -0.01718978) < 1e-6
    assert abs(mp2.e_corr_os - -0.18328931) < 1e-6
