# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file


import numpy as np
import pytest

from pybest.pt.pccdpt2 import PT2SDd, PT2MDd, PT2SDo, PT2MDo, PT2b
from pybest.geminals.rpccd import RpCCD
from pybest.exceptions import EmptyData
from pybest.pt.test.common import prepare_mol


test_cases = [
    (
        "NH3",
        "cc-pvdz",
        PT2SDd,
        "(sd)",
        {
            "mol": "nh3",
            "orb_file": "nh3_ap1rog.txt",
            "e_pccd": -56.288865946757,
            "e_tot": -56.39067080,
            "e_corr_s": -0.00023982,
            "e_corr_d": -0.101565033243,
            "kwargs": {},
        },
    ),
    (
        "NH3",
        "cc-pvdz",
        PT2MDd,
        "(sd)",
        {
            "mol": "nh3",
            "orb_file": "nh3_ap1rog.txt",
            "e_pccd": -56.288865946757,
            "e_tot": -56.38820919,
            "e_corr_s": -0.000349570000,
            "e_corr_d": -0.098993673243,
            "kwargs": {},
        },
    ),
    (
        "NH3",
        "cc-pvdz",
        PT2SDo,
        "(d)",
        {
            "mol": "nh3",
            "orb_file": "nh3_ap1rog.txt",
            "e_pccd": -56.288865946757,
            "e_tot": -56.40050443,
            "e_corr_d": -0.111638483243,
            "kwargs": {"singles": False},
        },
    ),
    (
        "NH3",
        "cc-pvdz",
        PT2SDo,
        "(sd)",
        {
            "mol": "nh3",
            "orb_file": "nh3_ap1rog.txt",
            "e_pccd": -56.288865946757,
            "e_tot": -56.40050594662758672,
            "e_corr_s": 0.00000000000000000,
            "e_corr_d": -0.11163999987058672,
            "kwargs": {"singles": True},
        },
    ),
    (
        "NH3",
        "cc-pvdz",
        PT2MDo,
        "(d)",
        {
            "mol": "nh3",
            "orb_file": "nh3_ap1rog.txt",
            "e_pccd": -56.288865946757,
            "e_tot": -56.395285377381,
            "e_corr_s": -0.000000000000,
            "e_corr_d": -0.106419469618,
            "kwargs": {"singles": False},
        },
    ),
    (
        "NH3",
        "cc-pvdz",
        PT2MDo,
        "(d)",
        {
            "mol": "nh3",
            "orb_file": "nh3_ap1rog.txt",
            "e_pccd": -56.288865946757,
            "e_tot": -56.395549517241,
            "e_corr_s": -0.000264322783,
            "e_corr_d": -0.106419286696,
            "kwargs": {"singles": True},
        },
    ),
    (
        "NH3",
        "cc-pvdz",
        PT2b,
        "(d)",
        {
            "mol": "nh3",
            "orb_file": "nh3_ap1rog.txt",
            "e_pccd": -56.288865946757,
            "e_tot": -56.392294846446,
            "e_corr_s": -0.00000000000,
            "e_corr_d": -0.10342893868,
            "kwargs": {"singles": False},
        },
    ),
    (
        "NH3",
        "cc-pvdz",
        PT2b,
        "(d)",
        {
            "mol": "nh3",
            "orb_file": "nh3_ap1rog.txt",
            "e_pccd": -56.288865946757,
            "e_tot": -56.392694604520,
            "e_corr_s": -0.000394757465,
            "e_corr_d": -0.103433939292,
            "kwargs": {"singles": True},
        },
    ),
    (
        "NH3",
        "cc-pvdz",
        PT2b,
        "(d)",
        {
            "mol": "nh3",
            "orb_file": "nh3_ap1rog.txt",
            "e_pccd": -56.288865946757,
            "e_tot": -56.392117900844,
            "e_corr_s": -0.00000000000,
            "e_corr_d": -0.103251993081,
            "kwargs": {"singles": False, "excludepairs": True},
        },
    ),
    (
        "NH3",
        "cc-pvdz",
        PT2b,
        "(d)",
        {
            "mol": "nh3",
            "orb_file": "nh3_ap1rog.txt",
            "e_pccd": -56.288865946757,
            "e_tot": -56.392517698981,
            "e_corr_s": -0.000394778161,
            "e_corr_d": -0.103257013058,
            "kwargs": {"singles": True, "excludepairs": True},
        },
    ),
]


@pytest.mark.parametrize("mol, basis, cls, t_, dict_", test_cases)
def test_pccd_pt2(mol, basis, cls, t_, dict_):
    lf, olp, orb_a, one, er, external, occ_model = prepare_mol(
        basis, f"test/{dict_['mol']}.xyz", f"test/{dict_['orb_file']}"
    )

    # Do pCCD optimization:
    geminal_solver = RpCCD(lf, occ_model)
    pccd = geminal_solver(one, er, orb_a, olp, e_core=external)
    assert abs(pccd.e_tot - dict_["e_pccd"]) < 1e-6

    pt2 = cls(lf, occ_model)
    #
    # using IOData
    #
    pt2_ = pt2(one, er, pccd, **dict_["kwargs"])
    assert abs(pt2_.e_tot - dict_["e_tot"]) < 1e-6
    assert abs(pt2_.e_corr_d - dict_["e_corr_d"]) < 1e-6
    if hasattr(dict_, "e_corr_s"):
        assert abs(pt2_.e_corr_s - dict_["e_corr_s"]) < 1e-6
    #
    # using arguments
    #
    try:
        pt2_ = pt2(
            one, er, orb_a, olp, pccd.t_p, e_ref=pccd.e_tot, **dict_["kwargs"]
        )
    except (EmptyData):
        pt2_ = pt2(
            one,
            er,
            orb_a,
            olp,
            pccd.t_p,
            e_ref=pccd.e_tot,
            overlap=pccd.overlap,
            **dict_["kwargs"],
        )
    assert abs(pt2_.e_tot - dict_["e_tot"]) < 1e-6
    assert abs(pt2_.e_corr_d - dict_["e_corr_d"]) < 1e-6
    if hasattr(dict_, "e_corr_s"):
        assert abs(pt2_.e_corr_s - dict_["e_corr_s"]) < 1e-6


test_fc = [
    (
        "NH3",
        "cc-pvdz",
        PT2b,
        1,
        {
            "mol": "nh3",
            "orb_file": "nh3_ap1rog.txt",
            "e_pccd": -56.288582636705,
            "e_tot": -56.389994390499,
            "e_corr_s": -0.000354288573,
            "e_corr_d": -0.101057465221,
            "kwargs": {"singles": True, "excludepairs": True},
        },
    ),
]


@pytest.mark.parametrize("mol, basis, cls, ncore, dict_", test_fc)
def test_pccd_pt2_ncore(mol, basis, cls, ncore, dict_):
    lf, olp, orb_a, one, er, external, occ_model = prepare_mol(
        basis, f"test/{dict_['mol']}.xyz", f"test/{dict_['orb_file']}"
    )

    # Do pCCD optimization:
    geminal_solver = RpCCD(lf, occ_model, ncore=ncore)
    pccd = geminal_solver(one, er, orb_a, olp, e_core=external)

    pt2 = cls(lf, occ_model, ncore=ncore)
    #
    # using IOData
    #
    pt2_ = pt2(one, er, pccd, **dict_["kwargs"])
    assert abs(pt2_.e_tot - dict_["e_tot"]) < 1e-6
    assert abs(pt2_.e_corr_d - dict_["e_corr_d"]) < 1e-6
    if hasattr(dict_, "e_corr_s"):
        assert abs(pt2_.e_corr_s - dict_["e_corr_s"]) < 1e-6
    #
    # using arguments
    #
    try:
        pt2_ = pt2(
            one, er, orb_a, olp, pccd.t_p, e_ref=pccd.e_tot, **dict_["kwargs"]
        )
    except (EmptyData):
        pt2_ = pt2(
            one,
            er,
            orb_a,
            olp,
            pccd.t_p,
            e_ref=pccd.e_tot,
            overlap=pccd.overlap,
            **dict_["kwargs"],
        )
    assert abs(pt2_.e_tot - dict_["e_tot"]) < 1e-6
    assert abs(pt2_.e_corr_d - dict_["e_corr_d"]) < 1e-6
    if hasattr(dict_, "e_corr_s"):
        assert abs(pt2_.e_corr_s - dict_["e_corr_s"]) < 1e-6
