# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Optimization methods

    PBQuasiNewton:  Perturbation-Based Quasi-Newton method
"""


import numpy as np

from pybest.log import log, timer
from pybest.solvers.diis import DIIS
from pybest.utils import check_options
from pybest.exceptions import UnknownOption, ConsistencyError


__all__ = [
    "PBQausiNewton",
]


class PBQausiNewton:
    def __init__(self, lf, maxiter=200, ethreshold=1e-6, tthreshold=1e-6):
        """
        Perturbation-Based Quasi-Newton method with a diagonal Jacobian.

        Vectorfunction has to be provided in different class (called obj here)
        as vfunction.

        Both vector function and jacobian have to be instances of DenseOneIndex.

        The energy is calcuated in obj class using the function calculate_energy
        and has to return a list of energies, with the first element being the
        total energy.

        ** Arguements **

        maxiter
            Maximum number of iterations (int)

        ethreshold
            Convergence tolerance for energy (float)

        tthreshold
            Convergence tolerance for amplitudes (float)

        ** Returns **

          :eval:   energy
          :tvec:   solution vector

        """
        self.lf = lf
        self.maxiter = maxiter
        self.etol = ethreshold
        self.ttol = tthreshold

    @property
    def lf(self):
        """The linalg factory"""
        return self._lf

    @lf.setter
    def lf(self, new):
        self._lf = new

    @property
    def maxiter(self):
        return self._maxiter

    @maxiter.setter
    def maxiter(self, new):
        if new < 0:
            raise UnknownOption("Number of iterations has to be positive")
        else:
            self._maxiter = new

    @property
    def etol(self):
        return self._etol

    @etol.setter
    def etol(self, new):
        self._etol = new

    @property
    def ttol(self):
        return self._ttol

    @ttol.setter
    def ttol(self, new):
        self._ttol = new

    def print_info(self, **kwargs):
        """Print optimization options.
        """
        diismax = kwargs.get("diismax", 9)
        diisstart = kwargs.get("diisstart", 0)
        reset = kwargs.get("diisreset", True)
        if log.do_medium:
            log.hline()
            log("Entering perturbation-based quasi-Newton solver")
            log.hline()
            log("OPTIMIZATION PARAMETERS:")
            log("Optimization thresholds:")
            log(f"  wfn amplitudes:    {self.ttol:.2e}")
            log(f"  energy:            {self.etol:.2e}")
            log(f"  maxiter:           {self.maxiter:3}")
            log("DIIS:")
            log(f"  max vectors:       {diismax}")
            log(f"  start:             {diisstart}")
            log(f"  reset:             {reset}")
            log.hline()

    def check_input(self, **kwargs):
        for name, value in kwargs.items():
            check_options(
                name, name, "diismax", "diisstart", "diisreset", "suffix",
            )

    def compute_vfunction(self, obj, guess, *args):
        """Vector function used in optimization. It is defined in instance obj as
           vfunction.

        ** Returns **

            OneIndex instance
        """
        vfunction = obj.vfunction(guess, *args)
        return vfunction

    def compute_jacobian(self, obj, guess, *args):
        """Jacobian function used in optimization. It is defined in instance obj as
           jacobian."

        ** Returns **

            OneIndex instance
        """
        jacobian = obj.jacobian(guess, *args)
        return jacobian

    def compute_energy(self, obj, tk):
        """Energy calculated in class obj. Function is defined as calculate_energy.

        ** Returns **

            A list of energies
        """
        e = obj.calculate_energy(tk)
        return e

    @timer.with_section("PBQN")
    def __call__(self, obj, guess, *args, **kwargs):
        """The optimization method."""
        diismax = kwargs.get("diismax", 9)
        diisstart = kwargs.get("diisstart", 0)
        suffix = kwargs.get("suffix", "")
        reset = kwargs.get("diisreset", True)
        #
        # Check input
        #
        self.check_input(**kwargs)
        #
        # Print some information
        #
        self.print_info(**kwargs)
        #
        # Compute Jacobian used in optimization (each iteration uses the same Jacobian)
        #
        eps = self.compute_jacobian(obj, guess, *args)

        tk = self.lf.create_one_index(guess.shape[0])
        tk.assign(guess)
        ek = self.compute_energy(obj, tk)
        de = {"e_tot": 0.0}

        diis = DIIS(self.lf, diismax, diisstart)
        iter_ = 0
        if log.do_medium:
            log(
                f"{'iter':>6} {'Etot':>12} {'E(new)-E(old)':>20} {'||dt||_2':>11} "
                f"{'max(dt)':>11} {'Ndiis':>7}"
            )
        while True:
            #
            # compute vector function for zeroth iteration
            #
            if iter_ == 0:
                vf = self.compute_vfunction(obj, tk, *args)
            else:
                vf = self.compute_vfunction(obj, diis.tk[-1], *args)
                tk = diis.tk[-1]
            #
            # Calculate correction as Dt = -vf/eps
            #
            deltatk = vf.divide(eps, -1.0)
            #
            # Calculate average solution vector
            #
            diis(tk, deltatk, iter_, de["e_tot"])
            #
            # calculate new energy
            #
            ekn = self.compute_energy(obj, diis.tk[-1])
            #
            # check convergence
            #
            dek = {}
            for e_ in ek:
                dek.update({e_: (ekn[e_] - ek[e_])})
            # reset diis if necessary
            if abs(dek["e_tot"]) > abs(de["e_tot"]):
                diis.reset(deltatk, reset)
                # Recalculate energy
                ekn = self.compute_energy(obj, diis.tk[-1])
                dek = {}
                for e_ in ek:
                    dek.update({e_: (ekn[e_] - ek[e_])})
                # for en, e in zip(ekn, ek):
                    # dek.append(en - e)

            dtkmax = diis.dtk[-1].get_max()
            dtknorm = diis.dtk[-1].norm()
            #
            #
            #
            if log.do_medium:
                log(
                    f"{iter_:>4} {ekn['e_tot']:>19.12f}  {dek['e_tot']: .6e} "
                    f"{dtknorm:>12.3e} {dtkmax:>12.3e} {len(diis.dtk):>5}"
                )
            #
            #
            #
            if abs(dek["e_tot"]) < self.etol:
                if dtknorm < self.ttol:  # or dtknorm < self.ttol:
                    if log.do_medium:
                        log.hline("~")
                        log(f"   Quasi-Netwon converged in {iter_} iterations")
                        log.hline("=")
                    return ekn, diis.tk[-1]
            # check if norm is reasonable, otherwise abort
            if abs(dtknorm) > 10:
                raise ConsistencyError(
                    "Unreasonable L2 norm. Aborting calculation!"
                )
            #
            # Try to dump current solution for amplitudes
            #
            try:
                obj.dump_solution(diis.tk[-1], suffix)
            except:
                pass
            iter_ += 1
            if iter_ >= self.maxiter:
                if log.do_medium:
                    log.hline("~")
                    log("   Quasi-Netwon NOT converged")
                    log.hline("=")
                return ekn, diis.tk[-1]
            ek = ekn
            de = dek
