# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file


import numpy as np
import pytest
from pathlib import Path, PurePath

from pybest.filemanager import filemanager

test_dirs = [
    ("foo0"),
    ("foo1"),
    ("foo2"),
    ("foo3"),
    ("foo4"),
]

test_join = [
    ("foo", "foobar0"),
    ("foo", "foobar1"),
    ("foo", "foobar2"),
]

@pytest.fixture(autouse=True)
def reset_filemanager():
    """Reset filemanager to default dirs otherwise other tests will break"""
    filemanager.temp_dir = "pybest-temp"
    filemanager.result_dir = "pybest-results"


@pytest.mark.parametrize("temp_dir", test_dirs)
def test_temp_dir(temp_dir, reset_filemanager):
    filemanager.temp_dir = temp_dir
    assert str(PurePath(filemanager.temp_dir).name) == temp_dir
    assert filemanager.temp_dir.exists()
    assert filemanager.temp_dir.is_dir()
    assert Path(temp_dir).exists()
    assert Path(temp_dir).is_dir()

    filemanager.temp_dir.rmdir()
    assert not filemanager.temp_dir.exists()
    assert not Path(temp_dir).exists()

@pytest.mark.parametrize("result_dir", test_dirs)
def test_result_dir(result_dir, reset_filemanager):
    filemanager.result_dir = result_dir
    assert str(PurePath(filemanager.result_dir).name) == result_dir
    assert filemanager.result_dir.exists()
    assert filemanager.result_dir.is_dir()
    assert Path(result_dir).exists()
    assert Path(result_dir).is_dir()

    filemanager.result_dir.rmdir()
    assert not filemanager.result_dir.exists()
    assert not Path(result_dir).exists()

@pytest.mark.parametrize("temp_dir", test_dirs)
def test_clean_up_temp(temp_dir, reset_filemanager):
    filemanager.temp_dir = temp_dir
    assert str(PurePath(filemanager.temp_dir).name) == temp_dir
    assert filemanager.temp_dir.exists()
    assert filemanager.temp_dir.is_dir()

    filemanager.clean_up_temp()
    assert not filemanager.temp_dir.exists()
    assert not Path(temp_dir).exists()

@pytest.mark.parametrize("dir_, filen", test_join)
def test_result_path(dir_, filen, reset_filemanager):
    filemanager.result_dir = dir_
    assert str(PurePath(filemanager.result_dir).name) == dir_
    assert filemanager.result_dir.exists()
    assert filemanager.result_dir.is_dir()

    dir_filen = filemanager.result_path(filen)
    assert dir_filen == Path(dir_).joinpath(filen)

    filemanager.result_dir.rmdir()
    assert not filemanager.result_dir.exists()
    assert not Path(dir_).exists()

@pytest.mark.parametrize("dir_, filen", test_join)
def test_temp_path(dir_, filen, reset_filemanager):
    filemanager.temp_dir = dir_
    assert str(PurePath(filemanager.temp_dir).name) == dir_
    assert filemanager.temp_dir.exists()
    assert filemanager.temp_dir.is_dir()

    dir_filen = filemanager.temp_path(filen)
    assert dir_filen == Path(dir_).joinpath(filen)

    filemanager.temp_dir.rmdir()
    assert not filemanager.temp_dir.exists()
    assert not Path(dir_).exists()
