# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file


import numpy as np
import pytest

from pybest.context import context
from pybest.gbasis.gobasis import (
    get_gobasis,
    compute_kinetic,
    compute_nuclear,
    compute_nuclear_repulsion,
    compute_overlap,
    compute_eri,
)
from pybest.linalg import DenseLinalgFactory
from pybest.wrappers.hf import RHF
from pybest.orbital_utils import (
    print_ao_mo_coeffs,
    split_core_active,
    transform_integrals,
)
from pybest.scf.occ import AufbauOccModel


def prepare_hf(basis_str, mol, occ_model):
    #
    # Run a simple HF calculation on the given IOData in the given basis
    #

    # Input structure
    obasis = get_gobasis(basis_str, mol, print_basis=False)
    lf = DenseLinalgFactory(obasis.nbasis)

    # Compute Gaussian integrals
    olp = compute_overlap(obasis)
    kin = compute_kinetic(obasis)
    na = compute_nuclear(obasis)
    er = compute_eri(obasis)
    core = compute_nuclear_repulsion(obasis)

    # Initial guess
    orb = lf.create_orbital()

    hf = RHF(lf, occ_model)
    hf_ = hf(kin, na, er, olp, orb, core)

    return hf_, kin, na, er


def do_check(
    nocc, ncore, nactive, one_small, two_small, ecore, kin, na, er, hf
):
    #
    # Verify (parts of) the RHF energy using the active space integrals
    #
    # Get the integrals in the mo basis
    ti = transform_integrals(kin, na, er, hf, indextrans="tensordot")
    (one_mo,) = ti.one
    (two_mo,) = ti.two

    # Check the core energy
    ecore_check = hf.e_core + 2 * one_mo.trace(0, ncore, 0, ncore)
    ranges = {"end0": ncore, "end1": ncore, "end2": ncore, "end3": ncore }
    ecore_check += two_mo.slice("abab->ab", None, 2.0, True, **ranges).sum()
    ecore_check += two_mo.slice("abba->ab", None, -1.0, True, **ranges).sum()
    assert abs(ecore - ecore_check) < 1e-10

    # Check the one-body energy of the active space
    nocc_small = nocc - ncore
    e_one_active = 2 * one_small.trace(0, nocc_small, 0, nocc_small)
    e_one_active_check = 2 * one_mo.trace(ncore, nocc, ncore, nocc)
    coxx = {"end0": ncore, "begin1": ncore, "end1": nocc}
    xxco = {"end2": ncore, "begin3": ncore, "end3": nocc}
    xxoc = {"begin2": ncore, "end2": nocc, "end3": ncore}
    e_one_active_check += (
        2 * two_mo.slice("abab->ab", None, 2.0, True, **coxx, **xxco).sum()
    )
    e_one_active_check += (
        2 * two_mo.slice("abba->ab", None, -1.0, True, **coxx, **xxoc).sum()
    )
    assert abs(e_one_active - e_one_active_check) < 1e-10

    # Check the two-body energy of the active space
    ssss = {f"end{i}":nocc_small for i in range(4)}
    e_two_active = (
        two_small.slice("abab->ab", None, 2.0, True, **ssss).sum()
        + two_small.slice("abba->ab", None, -1.0, True, **ssss).sum()
    )
    oooo = {
        **{f"begin{i}":ncore for i in range(4)},
        **{f"end{i}":nocc for i in range(4)},
    }
    e_two_active_check = (
        two_mo.slice("abab->ab", None, 2.0, True, **oooo).sum()
        + two_mo.slice("abba->ab", None, -1.0, True, **oooo).sum()
    )
    assert abs(e_two_active - e_two_active_check) < 1e-10

    # Check the total RHF energy
    e_rhf = ecore + e_one_active + e_two_active
    assert abs(e_rhf - hf.e_tot) < 1e-10


def check_core_active(mol, basis_str, nocc, ncore, nactive):

    # Decide how to occupy the orbitals
    assert ncore + nactive > nocc
    occ_model = AufbauOccModel(nocc)
    #
    # Do the HF calculation
    #
    hf, kin, na, er = prepare_hf(basis_str, mol, occ_model)
    #
    # Get integrals for the active space
    #
    cas = split_core_active(kin, na, er, hf, ncore=ncore, nactive=nactive)
    one_small = cas.one
    two_small = cas.two
    ecore = cas.e_core

    do_check(
        nocc, ncore, nactive, one_small, two_small, ecore, kin, na, er, hf
    )
    #
    # Get integrals for the active space, give twice external energy
    #
    cas = split_core_active(
        kin, na, er, hf, e_core=hf.e_core, ncore=ncore, nactive=nactive
    )
    one_small = cas.one
    two_small = cas.two
    ecore = cas.e_core

    do_check(
        nocc, ncore, nactive, one_small, two_small, ecore, kin, na, er, hf
    )
    #
    # Get integrals for the active space, give only one e integrals
    #
    one = kin.copy()
    one.iadd(na)
    cas = split_core_active(
        one, er, hf, ncore=ncore, nactive=nactive
    )
    one_small = cas.one
    two_small = cas.two
    ecore = cas.e_core

    do_check(
        nocc, ncore, nactive, one_small, two_small, ecore, kin, na, er, hf
    )


def prepare_mol(
    basis, filen="test/h2o_ccdz.xyz", orbfilen="test/h2o_ccdz_hf.txt"
):
    fn_xyz = context.get_fn(filen)
    obasis = get_gobasis(basis, fn_xyz, print_basis=False)

    lf = DenseLinalgFactory(obasis.nbasis)
    fn_orb = context.get_fn(orbfilen)
    orb_ = np.fromfile(fn_orb, sep=",").reshape(obasis.nbasis, obasis.nbasis)
    orb_a = lf.create_orbital()
    orb_a._coeffs = orb_

    return obasis, orb_a


test_cases = [
    ("ne", "cc-pvdz", 5, 2, 7),
    ("water", "cc-pvdz", 5, 1, 8),
    ("2h-azirine", "cc-pvdz", 11, 3, 15),
]


@pytest.mark.parametrize("mol, basis, nocc, ncore, nactive", test_cases)
def test_core_active(mol, basis, nocc, ncore, nactive):
    mol = context.get_fn(f"test/{mol}.xyz")
    check_core_active(mol, basis, nocc, ncore, nactive)


def test_print_ao_mo_coeff_full():
    obasis, orb_a = prepare_mol("cc-pvdz")

    # Dry run
    print_ao_mo_coeffs(obasis, orb_a)


def test_print_ao_mo_coeff_view():
    obasis, orb_a = prepare_mol("cc-pvdz")

    # Dry runs
    print_ao_mo_coeffs(obasis, orb_a, 1, 20)
    print_ao_mo_coeffs(obasis, orb_a, 1, 1)
    print_ao_mo_coeffs(obasis, orb_a, 11, 11)
    # first orbital index exceeds
    with pytest.raises(ValueError):
        print_ao_mo_coeffs(obasis, orb_a, 26)
    # last orbital index exceeds
    with pytest.raises(ValueError):
        print_ao_mo_coeffs(obasis, orb_a, end=26)
    # first orbital index larger than last
    with pytest.raises(ValueError):
        print_ao_mo_coeffs(obasis, orb_a, 0)
