# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Wrapper for mutipole calculations"""

import numpy as np

from pybest.log import log, timer
from pybest.exceptions import ArgumentError
from pybest.utils import (
    check_type,
    unmask_orb,
    unmask,
)
from pybest.linalg.dense import TwoIndex, OneIndex, Orbital
from pybest.units import debye


__all__ = ["compute_dipole_moment"]


def compute_dipole_moment(*args, **kwargs):
    """Calculates the dipole moment for any density matrix in AO or MO basis.

       **Arguments:**

       args: arguments containing:
           integrals
                The electric dipole moment integrals. A list of TwoIndex objects
                for each cartesian coordinate vector. In Libint, also the overlap
                integrals are calculated and returned. Thus, by default PyBEST
                also returns them. The default order of the dipole moment integrals
                is [olp, mux, muy, muz]. The integrals are recognized wrt their
                labels.

           dm
                A density matrix. For restricted orbitals, dm has to correspond to
                the full density matrix. A TwoIndex object.

           orb_a
                The molecular orbitals. Only required if mo=True

       **Keyword Arguments:**

       mo
            (boolean) if True, the dipole moment is transformed into new MO
            basis of orb.

       name
            (str) if provided prints 'name' in the header.

       scale
            (float) a scaling factor for the 1DM. If provided, the 1DM is scaled
            accordingly. Default=2.0 (assuming restricted orbitals)
    """
    mo = kwargs.get("mo", False)
    name = kwargs.get("name", None)
    scale = kwargs.get("scale", 2.0)

    # orb
    orb = unmask_orb(*args)
    if mo and not orb:
        raise ArgumentError("Cannot find orbitals in function call.")
    # dm from IOData, create a copy as we scale the dm afterwards
    dm_ = unmask("dm_1", *args) or unmask("dm_1_a", *args)
    if dm_ is None:
        raise ArgumentError("Cannot find density matrix")
    else:
        dm = dm_.copy()
    # Check arguments
    check_type("dm", dm, OneIndex, TwoIndex)
    # scale dm
    dm.iscale(scale)

    # dipole ints
    ints = {}
    mus = ["mux", "muy", "muz"]
    for arg in args:
        if isinstance(arg, TwoIndex):
            if isinstance(arg_, TwoIndex):
                if arg_.label in mus:
                    ints[label] = arg_
        # overwrite with list
        if isinstance(arg, (list, tuple)):
            for arg_ in arg:
                if isinstance(arg_, TwoIndex):
                    if arg_.label in mus:
                        ints[arg_.label] = arg_
            break
    # Check if all are found:
    if not all([mus_ in ints for mus_ in mus]):
        raise ArgumentError(
            "Cannot find all components of the dipole moment integrals"
        )

    if log.do_medium:
        log.hline("=")
        if name is not None:
            log(f"Calculating dipole moment for {name}")
        else:
            log("Calculating dipole moment")

    # DM is given in AO basis:
    if not mo:
        # DM is TwoIndex object
        if isinstance(dm, TwoIndex):
            ex = -ints["mux"].contract("ab,ab", dm)
            ey = -ints["muy"].contract("ab,ab", dm)
            ez = -ints["muz"].contract("ab,ab", dm)
        elif isinstance(dm, OneIndex):
            dm_ = ints["mux"].new()
            dm_.assign_diagonal(dm)
            ex = -ints["mux"].contract("ab,ab", dm_)
            ey = -ints["muy"].contract("ab,ab", dm_)
            ez = -ints["muz"].contract("ab,ab", dm_)
    # DM is given in MO basis
    else:
        orb_ = orb[0].copy()
        orb_.itranspose()
        # DM is TwoIndex object
        if isinstance(dm, TwoIndex):
            # Transform DM to AO basis
            dm_ = dm.new()
            dm_.assign_two_index_transform(dm, orb_)
            ex = -ints["mux"].contract("ab,ab", dm_)
            ey = -ints["muy"].contract("ab,ab", dm_)
            ez = -ints["muz"].contract("ab,ab", dm_)
        elif isinstance(dm, OneIndex):
            dm_mo = ints["mux"].new()
            dm_ao = ints["mux"].new()
            dm_mo.assign_diagonal(dm)
            dm_ao.assign_two_index_transform(dm_mo, orb_)
            ex = -ints["mux"].contract("ab,ab", dm_ao)
            ey = -ints["muy"].contract("ab,ab", dm_ao)
            ez = -ints["muz"].contract("ab,ab", dm_ao)

    # Print output
    if log.do_medium:
        log.hline("~")
        log(f"{'Component':>10s} {'Hartree':>14s} {'Debye':>14s}")
        log.hline("-")
        log(f"{'x':>6s} {ex:>19.7f} {(ex / debye):>15.7f}")
        log(f"{'y':>6s} {ey:>19.7f} {(ey / debye):>15.7f}")
        log(f"{'z':>6s} {ez:>19.7f} {(ez / debye):>15.7f}")
        log.hline("=")

    # Return in Debye
    return ex / debye, ey / debye, ez / debye
