# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file

import numpy as np
import pytest

from pybest.wrappers.hf import RHF, UHF
from pybest.exceptions import ArgumentError
from pybest.io.iodata import IOData
from pybest.wrappers.test.common import prepare_hf


ref_e_water_dz = {
    "kin": 75.966167929737,
    "hartree": 46.858246886237,
    "x_hf": -8.970595331038,
    "ne": -199.019146537165,
    "nn": 9.138880476031,
    "total": -76.026446576199,
}

ref_e_orb_water_dz = np.array(
    [
        -20.55116636,
        -1.33342941,
        -0.69713262,
        -0.56454528,
        -0.49256408,
        0.18469191,
        0.25535539,
        0.78687255,
        0.84681671,
        1.1635631,
        1.20043873,
        1.25314784,
        1.44983028,
        1.47325247,
        1.67621312,
        1.86958835,
        1.92588011,
        2.4407728,
        2.47526631,
        3.27814662,
        3.33511586,
        3.50068549,
        3.8635714,
        4.13826109,
    ]
)

ref_e_o2_dz = {
    "kin": 149.575743761456,
    "hartree": 100.421460958164,
    "x_hf": -16.348044690036,
    "ne": -411.500936926153,
    "nn": 28.222784582400,
    "total": -149.628992314170,
}


ref_e_orb_o2_a_dz = np.array(
    [
        -20.74916396,
        -20.74850338,
        -1.72456926,
        -1.19861276,
        -0.84002744,
        -0.84002738,
        -0.7602589,
        -0.54608643,
        -0.54608636,
        0.43760164,
        1.05429176,
        1.0542918,
        1.06855672,
        1.1417417,
        1.16489307,
        1.16489311,
        1.31275889,
        1.95442075,
        2.36727483,
        2.36727484,
        2.62190187,
        2.62190187,
        2.94050493,
        2.94050493,
        3.16738333,
        3.65575723,
        3.65575723,
        4.17196446,
    ]
)

ref_e_orb_o2_b_dz = np.array(
    [
        -20.69452164,
        -20.69330828,
        -1.59465534,
        -0.99169551,
        -0.69892562,
        -0.57595919,
        -0.57595916,
        0.11835314,
        0.11835316,
        0.51443822,
        1.099567,
        1.15964718,
        1.17464882,
        1.17464884,
        1.29135609,
        1.2913561,
        1.34484282,
        2.0066591,
        2.44144759,
        2.4414476,
        2.73910336,
        2.73910336,
        3.08879253,
        3.08879253,
        3.23948335,
        3.73269132,
        3.73269133,
        4.21795759,
    ]
)


test_args = [
    (RHF, (5,), "h2o_ccdz",),
    (UHF, (9, 7), "o2_ccdz",),
]


@pytest.mark.parametrize("cls, occ, mol", test_args)
def test_hf_input_args(cls, occ, mol):

    lf, occ_model, olp, orb, kin, na, er, external, gobasis = prepare_hf(
        "cc-pvdz", "test/" + mol + ".xyz", *occ
    )

    scf_ = cls(lf, occ_model)
    with pytest.raises(ValueError):
        scf = scf_(kin, na, er, external, olp, *orb, diis="None")
    with pytest.raises(ValueError):
        scf = scf_(kin, na, er, external, olp, *orb, nvector=-1)
    with pytest.raises(ArgumentError):
        scf = scf_(kin, na, er, external, olp, orb, orb)
    with pytest.raises(ArgumentError):
        scf = scf_(kin, na, er, external, olp, orb, "wrong argument")
    with pytest.raises(ArgumentError):
        scf = scf_(kin, na, er, external, olp, orb, IOData())


test_class = [
    (
        RHF,
        (5,),
        "h2o_ccdz",
        "plain",
        [ref_e_water_dz["total"], ref_e_orb_water_dz],
    ),
    (
        RHF,
        (5,),
        "h2o_ccdz",
        "cdiis",
        [ref_e_water_dz["total"], ref_e_orb_water_dz],
    ),
    (
        RHF,
        (5,),
        "h2o_ccdz",
        "ediis2",
        [ref_e_water_dz["total"], ref_e_orb_water_dz],
    ),
    (
        UHF,
        (9, 7),
        "o2_ccdz",
        "plain",
        [ref_e_o2_dz["total"], ref_e_orb_o2_a_dz, ref_e_orb_o2_b_dz],
    ),
    (
        UHF,
        (9, 7),
        "o2_ccdz",
        "cdiis",
        [ref_e_o2_dz["total"], ref_e_orb_o2_a_dz, ref_e_orb_o2_b_dz],
    ),
    (
        UHF,
        (9, 7),
        "o2_ccdz",
        "ediis2",
        [ref_e_o2_dz["total"], ref_e_orb_o2_a_dz, ref_e_orb_o2_b_dz],
    ),
]


@pytest.mark.parametrize("cls, occ, mol, diis, expected", test_class)
def test_hf_restart(cls, occ, mol, diis, expected):

    lf, occ_model, olp, orb, kin, na, er, external, gobasis = prepare_hf(
        "cc-pvdz", "test/" + mol + ".xyz", *occ
    )

    scf_ = cls(lf, occ_model)
    scf = scf_(kin, na, er, external, olp, *orb, diis=diis, threshold=1e-12)

    assert abs(expected[0] - scf.e_tot) < 1e-8
    assert np.allclose(expected[1], orb[0].energies)
    assert scf.orb_a == orb[0]
    try:
        assert np.allclose(expected[2], orb[1].energies)
        assert scf.orb_b == orb[1]
    except (IndexError):
        pass

    scf_ = cls(lf, occ_model)
    scf = scf_(
        kin,
        na,
        er,
        external,
        olp,
        *orb,
        diis=diis,
        threshold=1e-12,
        restart="pybest-results/checkpoint_scf.h5",
    )

    assert abs(expected[0] - scf.e_tot) < 1e-8
    assert np.allclose(expected[1], orb[0].energies)
    assert scf.orb_a == orb[0]
    try:
        assert np.allclose(expected[2], orb[1].energies)
        assert scf.orb_b == orb[1]
    except (IndexError):
        pass


testdata_diis = [
    ("plain", 1e-12, 1e-8),
    ("cdiis", 1e-12, 1e-8),
    ("ediis", 1e-5, 1e-6),
    ("ediis2", 1e-12, 1e-8),
]


@pytest.mark.parametrize("diis, thresh, compare", testdata_diis)
def test_diis_rhf_water(diis, thresh, compare):

    lf, occ_model, olp, orb, kin, na, er, external, gobasis = prepare_hf(
        "cc-pvdz", "test/h2o_ccdz.xyz", 5
    )

    scf_ = RHF(lf, occ_model)

    scf = scf_(kin, na, er, external, olp, *orb, diis=diis, threshold=thresh)
    assert abs(ref_e_water_dz["total"] - scf.e_tot) < compare
    assert np.allclose(ref_e_orb_water_dz, orb[0].energies)


@pytest.mark.parametrize("diis, thresh, compare", testdata_diis)
def test_diis_uhf_o2(diis, thresh, compare):

    lf, occ_model, olp, orb, kin, na, er, external, gobasis = prepare_hf(
        "cc-pvdz", "test/o2_ccdz.xyz", 9, 7
    )

    scf_ = UHF(lf, occ_model)

    scf = scf_(kin, na, er, external, olp, *orb, diis=diis, threshold=thresh)

    assert abs(ref_e_o2_dz["total"] - scf.e_tot) < compare
    assert np.allclose(ref_e_orb_o2_a_dz, scf.orb_a.energies)
    assert np.allclose(ref_e_orb_o2_b_dz, scf.orb_b.energies)
    assert scf.orb_a == orb[0]
    assert scf.orb_b == orb[1]
