# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file

import numpy as np
import pytest

from pybest.exceptions import ArgumentError
from pybest.wrappers.hf import RHF
from pybest.wrappers.multipole import compute_dipole_moment

from pybest.wrappers.test.common import prepare_hf, prepare_dipole

testdata_mol = [
    (RHF, "h2o", (5,), [-75.943961863294, 0.04129661, 2.25353156, 0.00000000]),
]

# Test can be extended to more molecules or to UHF in the future
@pytest.mark.parametrize("cls, mol, occ, expected", testdata_mol)
def test_dip_hf(cls, mol, occ, expected):

    lf, occ_model, olp, orb, kin, na, er, external, gobasis = prepare_hf(
        "cc-pvdz", f"test/{mol}.xyz", *occ
    )

    scf_ = cls(lf, occ_model)
    scf = scf_(kin, na, er, external, olp, *orb)

    assert abs(expected[0] - scf.e_tot) <= 1e-8
    assert abs(expected[0] - scf.e_ref) <= 1e-8

    dipole = prepare_dipole(gobasis)
    dip_hf = compute_dipole_moment(dipole, scf)

    assert abs(dip_hf[0] - expected[1]) < 1e-3
    assert abs(dip_hf[1] - expected[2]) < 1e-3
    assert abs(dip_hf[2] - expected[3]) < 1e-3


# Test can be extended to more molecules or to UHF in the future
@pytest.mark.parametrize("cls, mol, occ, expected", testdata_mol)
def test_dip_hf_args(cls, mol, occ, expected):
    lf, occ_model, olp, orb, kin, na, er, external, gobasis = prepare_hf(
        "cc-pvdz", f"test/{mol}.xyz", *occ
    )

    scf_ = cls(lf, occ_model)
    scf = scf_(kin, na, er, external, olp, *orb)

    assert abs(expected[0] - scf.e_tot) <= 1e-8
    assert abs(expected[0] - scf.e_ref) <= 1e-8

    dipole = prepare_dipole(gobasis)
    dip_hf = compute_dipole_moment(dipole, scf)

    dm = scf.dm_1.copy()
    del scf.dm_1

    # missing dm
    with pytest.raises(ArgumentError):
        dip_hf = compute_dipole_moment(dipole, scf)

    scf.dm_1 = dm
    del scf.orb_a

    # missing orb
    with pytest.raises(ArgumentError):
        dip_hf = compute_dipole_moment(dipole, scf, mo=True)

    # different scaling factor
    dip_hf = compute_dipole_moment(dipole, scf, scale=1.0)

    assert abs(dip_hf[0] * 2.0 - expected[1]) < 1e-3
    assert abs(dip_hf[1] * 2.0 - expected[2]) < 1e-3
    assert abs(dip_hf[2] * 2.0 - expected[3]) < 1e-3
