# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --

# Detailed changelog:
#
# This implementation has been taken from `Horton 2.0.0`.
# However, this file has been updated and debugged. Compatibility with Horton is NOT
# guaranteed.
# Its current version contains updates from the PyBEST developer team.
# 
# Detailed changes (see also CHANGELOG):
# 2020-07-01: Update to PyBEST standard, including naming convention
# 2020-07-01: Update to new python features, including f-strings
# 2020-07-01: Changed to general [slice] function and removed deprecated [slice_to] functions
# 2020-07-01: Removed deprecated [contract_] functions
# 2020-07-01: Introduce labels for all NIndex objects for book keeping

r"""Dense matrix implementations


   Naming scheme for the expand methods
   -------------------------------------------------------

   The name of ``expand`` methods is as follows::

        [iadd_]{expand}[_X][_Y][_to_Z]

   where each part between square brackets is optional. ``X``, ``Y`` and ``Z``
   can be any of ``one``, ``two``, ``three`` or ``four``. The name ``expand``
   have the following meaning:

   ``expand``
        Products of elements are computed but these products are not added.
        Similar to an outer product but more general.

   When ``iadd_`` is used as a prefix, the result of the contraction is added
   in-place to the self object. In this case, the ``_to_Z`` part is never
   present. A contraction of all input arguments is made. The dimensionality
   of the input arguments is indicated by ``_X`` and ``_Y``.

   When ``_to_Z`` is present, the contraction involves self and possibly other
   arguments whose dimensionality is indicated by ``_X`` and ``_Y``. In this
   case, ``iadd_`` can not be present. The result is written to an output
   argument. If the output argument is not provided, fresh memory is allocated
   to compute the contraction and the result is returned. (This allocation is
   provided for convenience but should not be used in critical situations.)

   Some remarks:

   * Similar conventions apply to an ``expand`` method.
   * All ``expand`` methods are implemented with the driver
     method ``DenseLinalgFactory.tco``. However, other implementations than
     `Dense` are free to implement things differently.
   * All ``expand`` methods never touch the internals of
     higher-index objects.

   For more specific information, read the documentation of the individual
   classes and methods below.


   .. _dense_matrix_symmetry:

   Handling of index symmetry
   --------------------------

   The dense matrix classes do not exploit matrix symmetry to reduce memory
   needs. Instead they will happily store non-symmetric data if need be. There
   are however a few methods in the :py:class:`DenseTwoIndex` and
   :py:class:`DenseFourIndex` classes below that take a ``symmetry`` argument to
   check or enforce a certain index symmetry.

   The symmetry argument is always an integer that corresponds to the redundancy
   of the off-diagonal matrix elements in the dense storage. In practice this
   means the following:

   * :py:class:`DenseTwoIndex`

     * ``symmetry=1``: Nothing is checked/enforced

     * ``symmetry=2``: Hermitian index symmetry is
       checked/enforced (default), i.e. :math:`\langle i \vert A \vert j
       \rangle = ` :math:`\langle j \vert A \vert i \rangle`

   * :py:class:`DenseFourIndex`

     * ``symmetry=1``: Nothing is checked/enforced

     * ``symmetry=2``: Dummy index symmetry is
       checked/enforced, i.e.
       :math:`\langle ij \vert B \vert kl \rangle =`
       :math:`\langle ji \vert B \vert lk \rangle`

     * ``symmetry=4``: Hermitian and real index symmetry are checked/enforced,
       i.e.
       :math:`\langle ij \vert B \vert kl \rangle =`
       :math:`\langle kl \vert B \vert ij \rangle =`
       :math:`\langle kj \vert B \vert il \rangle =`
       :math:`\langle il \vert B \vert kj \rangle`.
       (This only makes sense because the basis functions are assumed to be
       real.)

     * ``symmetry=8``: All possible symmetries are checked/enforced, i.e.
       :math:`\langle ij \vert B \vert kl \rangle =`
       :math:`\langle kl \vert B \vert ij \rangle =`
       :math:`\langle kj \vert B \vert il \rangle =`
       :math:`\langle il \vert B \vert kj \rangle =`
       :math:`\langle ji \vert B \vert lk \rangle =`
       :math:`\langle lk \vert B \vert ji \rangle =`
       :math:`\langle jk \vert B \vert li \rangle =`
       :math:`\langle li \vert B \vert jk \rangle`.
       (This only makes sense because the basis functions are assumed to be
       real.)

   Dense matrix classes
   --------------------
"""


import numpy as np
from functools import reduce
from scipy.linalg import eigh, sqrtm, inv, eig
from pybest.log import log
from pybest.utils import check_type, check_options
from pybest.linalg.base import (
    LinalgFactory,
    OneIndex,
    Orbital,
    TwoIndex,
    ThreeIndex,
    FourIndex,
    parse_four_index_transform_exps,
)
from pybest.linalg.pytransform import apply_four_index_transform_cpp
from pybest.exceptions import (
    ArgumentError,
    UnknownOption,
    SymmetryError,
    MatrixShapeError
)


__all__ = [
    "DenseLinalgFactory",
    "DenseOneIndex",
    "DenseOrbital",
    "DenseTwoIndex",
    "DenseThreeIndex",
    "DenseFourIndex",
]


class DenseLinalgFactory(LinalgFactory):

    #
    # DenseOneIndex constructor with default arguments
    #

    def create_one_index(self, nbasis=None, label=""):
        """Create a DenseOneIndex with defaults from the LinalgFactory

           **Optional arguments:**

           nbasis
                The number of basis functions. When not given, the
                default_nbasis value of the DenseLinalgFactory instance will be
                used.
        """
        nbasis = nbasis or self.default_nbasis
        return DenseOneIndex(nbasis, label)

    def _check_one_index_init_args(self, other, nbasis=None):
        """Is an object is compatible the constructor arguments?"""
        nbasis = nbasis or self.default_nbasis
        other.__check_init_args__(nbasis)

    create_one_index.__check_init_args__ = _check_one_index_init_args

    #
    # DenseOrbital constructor with default arguments
    #

    def create_orbital(self, nbasis=None, nfn=None):
        """Create a DenseOrbital with defaults from the LinalgFactory

           **Optional arguments:**

           nbasis
                The number of basis functions. When not given, the
                default_nbasis value of the DenseLinalgFactory instance will be
                used.

           nfn
                The number of orbitals. When not given, the default_nbasis
                value of the DenseLinalgFactory instance will be used.
        """
        nbasis = nbasis or self.default_nbasis
        nfn = nfn or nbasis
        return DenseOrbital(nbasis, nfn)

    def _check_orbital_init_args(self, expansion, nbasis=None, nfn=None):
        """Is an object is compatible the constructor arguments?"""
        nbasis = nbasis or self.default_nbasis
        nfn = nfn or nbasis
        expansion.__check_init_args__(nbasis, nfn)

    create_orbital.__check_init_args__ = _check_orbital_init_args

    #
    # DenseTwoIndex constructor with default arguments
    #

    def create_two_index(self, nbasis=None, nbasis1=None, label=""):
        """Create a DenseTwoIndex with defaults from the LinalgFactory

           **Optional arguments:**

           nbasis
                The number of basis functions. When not given, the
                default_nbasis value of the DenseLinalgFactory instance will be
                used.

           nbasis1
                The number of basis functions for the second axis if it differes
                from ``nbasis``.
        """
        nbasis = nbasis or self.default_nbasis
        # Don't replace nbasis1 by self.default_nbasis when it is None! It is
        # a genuine optional argument.
        return DenseTwoIndex(nbasis, nbasis1, label)

    def _check_two_index_init_args(self, other, nbasis=None, nbasis1=None):
        """Is an object compatible the constructor arguments?"""
        nbasis = nbasis or self.default_nbasis
        # Don't replace nbasis1 by self.default_nbasis when it is None! It is
        # a genuine optional argument.
        other.__check_init_args__(nbasis, nbasis1)

    create_two_index.__check_init_args__ = _check_two_index_init_args

    #
    # DenseThreeIndex constructor with default arguments
    #

    def create_three_index(
        self, nbasis=None, nbasis1=None, nbasis2=None, label=""
    ):
        """Create a DenseThreeIndex with defaults from the LinalgFactory

           **Optional arguments:**

           nbasis
                The number of basis functions. When not given, the
                default_nbasis value of the DenseLinalgFactory instance will be
                used.
        """
        nbasis = nbasis or self.default_nbasis
        # Don't replace nbasis1, nbasis2 by self.default_nbasis when None! They
        # are genuine optional arguments.
        return DenseThreeIndex(nbasis, nbasis1, nbasis2, label)

    def _check_three_index_init_args(
        self, other, nbasis=None, nbasis1=None, nbasis2=None
    ):
        """Is an object is compatible the constructor arguments?"""
        nbasis = nbasis or self.default_nbasis
        # Don't replace nbasis1, nbasis2 by self.default_nbasis when None! They
        # are genuine optional arguments.
        other.__check_init_args__(nbasis, nbasis1, nbasis2)

    create_three_index.__check_init_args__ = _check_three_index_init_args

    #
    # DenseFourIndex constructor with default arguments
    #

    def create_four_index(
        self, nbasis=None, nbasis1=None, nbasis2=None, nbasis3=None, label=""
    ):
        """Create a DenseFourIndex with defaults from the LinalgFactory

           **Optional arguments:**

           nbasis
                The number of basis functions. When not given, the
                default_nbasis value of the DenseLinalgFactory instance will be
                used.
        """
        nbasis = nbasis or self.default_nbasis
        # Don't replace nbasis1, nbasis2, nbasis3 by self.default_nbasis when None! They
        # are genuine optional arguments.
        return DenseFourIndex(nbasis, nbasis1, nbasis2, nbasis3, label)

    def _check_four_index_init_args(
        self, other, nbasis=None, nbasis1=None, nbasis2=None, nbasis3=None
    ):
        """Is an object is compatible the constructor arguments?"""
        nbasis = nbasis or self.default_nbasis
        other.__check_init_args__(nbasis, nbasis1, nbasis2, nbasis3)

    create_four_index.__check_init_args__ = _check_four_index_init_args

    #
    # Other code
    #

    @staticmethod
    def _allocate_check_output(out, outshape):
        """Allocate/check the output for the wrappers below

           **Arguments:**

           out
                The output argument.

           outshape
                The expected shape of the output argument.

           **Returns:** the output argument.
        """
        #       if len(outshape) == 3 and not (outshape[0] == outshape[1] and
        #                                      outshape[0] == outshape[2]):
        #           raise TypeError('A 3-index object must have the same size in all indexes.')
        #       if len(outshape) == 4 and not (outshape[0] == outshape[1] and
        #                                      outshape[0] == outshape[2] and
        #                                      outshape[0] == outshape[3]):
        #           raise TypeError('A 4-index object must have the same size in all indexes.')

        # Handle the output argument
        if out is None:
            if len(outshape) == 0:
                pass
            elif len(outshape) == 1:
                out = DenseOneIndex(outshape[0])
            elif len(outshape) == 2:
                out = DenseTwoIndex(*outshape)
            elif len(outshape) == 3:
                out = DenseThreeIndex(*outshape)
            #               out = DenseThreeIndex(outshape[0])
            elif len(outshape) == 4:
                out = DenseFourIndex(*outshape)
            #               out = DenseFourIndex(outshape[0])
            else:
                raise ArgumentError(
                    "The outshape must have length 0, 1, 2, 3 or 4."
                )
        else:
            if len(outshape) == 0:
                raise ArgumentError(
                    "No output argument can be given when contracting to a scalar."
                )
            elif len(outshape) == 1:
                check_type("out", out, DenseOneIndex)
            elif len(outshape) == 2:
                check_type("out", out, DenseTwoIndex)
            elif len(outshape) == 3:
                check_type("out", out, DenseThreeIndex)
            elif len(outshape) == 4:
                check_type("out", out, DenseFourIndex)
            else:
                raise ArgumentError(
                    "The outshape must have length 0, 1, 2, 3 or 4."
                )
            if out.shape != outshape:
                raise MatrixShapeError(
                    "The output argument does not have the right shape."
                )
        return out

    @staticmethod
    def tco(subscripts, out, factor, clear, select="einsum", *operands):
        """tensor contraction wrapper for DenseNBody objects

           **Arguments:**

           subscripts
                The usual subscripts argument for the einsum function, except
                that the format is slightly limited. All indexes must be
                explicitly stated. If the output argument is a scalar, ``->``
                may not be present. In all other cases, ``->`` must be present.

           out
                An output DenseNBody object. This argument is not allowed when
                the result is a scalar.

           factor
                Multiply the contraction by a scalar factor.

           clear
                When set to False, the output argument is not zeroed before
                adding the contraction.

           operands
                A list where each element is (i) a DenseNBody object or (ii) a
                two-tuple of DenseNBody object and ranges to select for slicing.
                The ranges must be a tuple of integers of the form (begin0,
                end0, begin1, end1, ...) where the number of (begin, end) pairs
                depends on the dimensionality of the corresponding tensor.

           **Returns:** the out object. When the out argument is given, this is
           returned with in-place modifications.
        """
        check_type("subscripts", subscripts, str)
        check_type("factor", factor, float, int)
        check_type("clear", clear, bool)

        # Basic checking of operands and adding ranges if omitted
        new_operands = []
        for operand in operands:
            if isinstance(operand, tuple):
                if len(operand) != 2:
                    raise ArgumentError(
                        "When an operand is a tuple, it must count two elements."
                    )
                tensor, ranges = operand
                if len(tensor.shape) * 2 != len(ranges):
                    raise ArgumentError(
                        "The dimensionality of the tensor and the size of ranges does not match."
                    )
            else:
                tensor = operand
                ranges = sum([(0, s) for s in tensor.shape], ())
            new_operands.append((tensor, list(ranges)))
        operands = new_operands

        # Impose explicit format for subscripts and do some basic checking with
        # operands
        if subscripts.count("->") == 1:
            inscripts, outscript = subscripts.split("->")
        elif subscripts.count("->") == 0:
            inscripts = subscripts
            outscript = ""
        else:
            raise ArgumentError(
                "The subscripts argument must contain a single or none '->'."
            )
        inscripts = inscripts.split(",")
        for inscript, (tensor, ranges) in zip(inscripts, operands):
            if len(inscript) == 1:
                if not isinstance(tensor, DenseOneIndex):
                    raise MatrixShapeError(
                        "Expecting a DenseOneIndex tensor for subscripts "
                        f"'{inscript}'."
                    )
            elif len(inscript) == 2:
                if not isinstance(tensor, DenseTwoIndex):
                    raise MatrixShapeError(
                        "Expecting a DenseTwoIndex tensor for subscripts "
                        f"'{inscript}'."
                    )
            elif len(inscript) == 3:
                if not isinstance(tensor, DenseThreeIndex):
                    raise MatrixShapeError(
                        "Expecting a DenseThreeIndex tensor for subscripts "
                        f"'{inscript}'."
                    )
            elif len(inscript) == 4:
                if not isinstance(tensor, DenseFourIndex):
                    raise MatrixShapeError(
                        "Expecting a DenseFourIndex tensor for subscripts "
                        f"'{inscript}'."
                    )
            else:
                raise ArgumentError(
                    "The number of subscripts for one tensor must be 1, 2, 3 or 4"
                )

        # take slices of the input arrays
        new_operands = []
        for tensor, ranges in operands:
            slices = []
            while len(ranges) > 0:
                begin = ranges.pop(0)
                end = ranges.pop(0)
                slices.append(slice(begin, end))
            array = tensor._array[tuple(slices)]
            new_operands.append(array)
        operands = new_operands

        # Determine shape of the output
        outshape = []
        for outchar in outscript:
            size = None
            for inscript, array in zip(inscripts, operands):
                for i, inchar in enumerate(inscript):
                    if inchar == outchar:
                        if size is None:
                            size = array.shape[i]
                        else:
                            if size != array.shape[i]:
                                raise MatrixShapeError(
                                    "Incompatible shapes of input operands."
                                )
            if size is None:
                raise MatrixShapeError(
                    "Could not determine the shape of the output."
                )
            outshape.append(size)
        outshape = tuple(outshape)

        # Allocate/check the output argument
        out = DenseLinalgFactory._allocate_check_output(out, outshape)

        # do the actual work
        if len(outshape) == 0:
            if select in ["td"]:
                raise NotImplementedError
                out = (
                    float(DenseLinalgFactory.td_helper(subscripts, *operands))
                    * factor
                )
            else:
                out = (
                    float(np.einsum(subscripts + "->...", *operands)) * factor
                )
            assert isinstance(out, float)
        else:
            if clear:
                if select in ["td"]:
                    raise NotImplementedError
                    out._array[:] = DenseLinalgFactory.td_helper(
                        subscripts, *operands
                    )
                else:
                    # can be done without temporary
                    out._array[:] = np.einsum(subscripts, *operands)
                    # FIXME: due to a bug in numpy, we can't use the output argument
                    # for all cases. For more details, see:
                    # https://github.com/numpy/numpy/issues/5147
                    # np.einsum(subscripts, *operands, out=out._array)
                out.iscale(factor)
            else:
                # can't be done without temporary
                if select in ["td"]:
                    raise NotImplementedError
                    out._array += (
                        DenseLinalgFactory.td_helper(subscripts, *operands)
                        * factor
                    )
                else:
                    out._array += np.einsum(subscripts, *operands) * factor
        return out


class DenseOneIndex(OneIndex):
    """Dense one-dimensional matrix (vector)

       This is also used for (diagonal) density matrices.
    """

    #
    # Constructor and destructor
    #

    def __init__(self, nbasis, label=""):
        """
           **Arguments:**

           nbasis
                The number of basis functions.
        """
        self._array = np.zeros((nbasis,), float)
        self._label = label
        log.mem.announce(self._array.nbytes)

    def __del__(self):
        if log is not None:
            if hasattr(self, "_array"):
                log.mem.denounce(self._array.nbytes)
        if hasattr(self, "_array"):
            del self._array

    #
    # Methods from base class
    #

    def __check_init_args__(self, nbasis):
        """Is self compatible with the given constructor arguments?"""
        assert nbasis == self.nbasis

    def __eq__(self, other):
        """Compare self with other"""
        return (
            isinstance(other, DenseOneIndex)
            and other.nbasis == self.nbasis
            and (other._array == self._array).all()
        )

    def __lt__(self, other):
        """Compare self with other"""
        if isinstance(other, DenseOneIndex):
            assert other.nbasis == self.nbasis
            return (abs(self._array <= other._array)).all()
        elif isinstance(other, float):
            return (abs(self._array <= other)).all()

    @classmethod
    def from_hdf5(cls, grp):
        """Construct an instance from data previously stored in an h5py.Group.

           **Arguments:**

           grp
                An h5py.Group object.
        """
        nbasis = grp["array"].shape[0]
        label = grp.attrs["label"]
        result = cls(nbasis, label)
        grp["array"].read_direct(result._array)
        return result

    def to_hdf5(self, grp):
        """Dump this object in an h5py.Group

           **Arguments:**

           grp
                An h5py.Group object.
        """
        grp.attrs["class"] = self.__class__.__name__
        grp["array"] = self._array
        grp.attrs["label"] = self._label

    # FIXME: rename into clean_copy
    def new(self):
        """Return a new one-index object with the same nbasis"""
        return DenseOneIndex(self.nbasis)

    def einsum_index(self, script):
        """ Returns indices to numpy.einsum summation. """
        assert len(script) == 1 and isinstance(script, str)
        return script

    @property
    def arrays(self):
        """ Returns list containing self._array """
        return [self._array]

    @property
    def label(self):
        """ Returns label of instance"""
        return self._label

    @label.setter
    def label(self, label):
        """ Sets label of instance"""
        self._label = label

    def _check_new_init_args(self, other):
        """Check whether an already initialized object is compatible"""
        other.__check_init_args__(self.nbasis)

    new.__check_init_args__ = _check_new_init_args

    def replace_array(self, value):
        """Replaces an array with another array, if not present, it will be
           generated.
        """
        if isinstance(value, DenseOneIndex):
            self._array = value._array
        elif isinstance(value, np.ndarray):
            self._array = value
        else:
            raise ArgumentError(
                f"Do not know how to assign object of type {type(value)}."
            )

    def clear(self):
        """Reset all elements to zero."""
        self._array[:] = 0.0

    def copy(self, begin=0, end=None):
        """Return a copy of (a part of) the object

           **Optional arguments:**

           begin, end
                Can be used to select a subblock of the object. When not given,
                the full range is used.
        """
        if end is None:
            end = self.nbasis
        result = DenseOneIndex(end - begin)
        result._array[:] = self._array[begin:end]
        result._label = self._label
        return result

    def assign(self, other, begin0=0, end0=None, begin1=0, end1=None):
        """Assign with the contents of another object

           **Arguments:**

           other
                Another DenseOneIndex object or a scalar.

           **Optional arguments:**

           begin0, end0
                If specified, only a sublock of self is assigned.
        """
        end0 = self._fix_ends(end0)[0]
        if isinstance(other, DenseOneIndex):
            end1 = other._fix_ends(end1)[0]
            self._array[begin0:end0] = other._array[begin1:end1]
        elif isinstance(other, float) or isinstance(other, int):
            self._array[begin0:end0] = other
        elif isinstance(other, np.ndarray):
            if end1 is None:
                end1 = other.size
            self._array[begin0:end0] = other[begin1:end1]
        elif isinstance(other, DenseThreeIndex):
            if end1 is None:
                end1 = other._array.size
            self._array[begin0:end0] = (other._array.ravel(order="C"))[
                begin1:end1
            ]
        else:
            raise ArgumentError(
                f"Do not know how to assign object of type {type(other)}."
            )

    def randomize(self):
        """Fill with random normal data"""
        self._array[:] = np.random.normal(0, 1, self.shape)

    def permute_basis(self, permutation):
        """Reorder the coefficients for a given permutation of basis functions.

           **Arguments:**

           permutation
                An integer numpy array that defines the new order of the basis
                functions.
        """
        self._array[:] = self._array[permutation]

    def change_basis_signs(self, signs):
        """Correct for different sign conventions of the basis functions.

           **Arguments:**

           signs
                A numpy array with sign changes indicated by +1 and -1.
        """
        self._array *= signs

    def iadd(
        self, other, factor=1.0, begin0=0, end0=None, begin1=0, end1=None
    ):
        """Add another DenseOneIndex object in-place, multiplied by factor

           **Arguments:**

           other
                A DenseOneIndex instance to be added

           **Optional arguments:**

           factor
                A scalar factor
        """
        check_type("other", other, DenseOneIndex, float)
        check_type("factor", factor, float, int)
        end0 = self._fix_ends(end0)[0]
        if isinstance(other, DenseOneIndex):
            end1 = other._fix_ends(end1)[0]
            self._array[begin0:end0] += other._array[begin1:end1] * factor
        elif isinstance(other, float):
            self._array[begin0:end0] += other * factor

    def iscale(self, factor):
        """In-place multiplication with a scalar

           **Arguments:**

           factor
                A scalar factor.
        """
        check_type("factor", factor, float, int)
        self._array *= factor

    def norm(self):
        """Calculate L2 norm"""
        return np.linalg.norm(self._array)

    def trace(self, begin0=0, end0=None):
        """Calculate trace

           **Optional arguments:**

           begin0, end0
                Can be used to select a subblock of the object. When not given,
                the full range is used.
        """
        end0 = self._fix_ends(end0)[0]
        return np.sum(self._array[begin0:end0])

    def get_element(self, i):
        """Return a matrix element"""
        return self._array[i]

    def get_max(self, abs=True):
        """Return maximum (absolute) element"""
        return np.max(np.abs(self._array))

    def set_element(self, i, value):
        """Set a matrix element"""
        self._array[i] = value

    def get_nonzero(self):
        """Get indices of all non-zero elements"""
        return np.nonzero(self._array)[0]

    def sort_indices(self, reverse=False, begin0=0, end0=None):
        """Return indices of sorted arguments in decreasing order

           **Optional arguements**

           reverse
                If True search order is reversed to increasing arguements
        """
        end0 = self.nbasis
        if reverse:
            return np.argsort(
                self._array[begin0:end0], axis=-1, kind="mergesort", order=None
            )
        else:
            return np.argsort(
                self._array[begin0:end0], axis=-1, kind="mergesort", order=None
            )[::-1]

    def mult(self, other, out=None, factor=1.0):
        """Muliply with other DenseOneIndex object, multiplied by factor

           **Arguments:**

           other
                A DenseOneIndex instance to be added

           **Optional arguments:**

           out
                The output argument (DenseOneIndex with proper size).

           factor
                A scalar factor
        """
        check_type("other", other, DenseOneIndex)
        check_type("factor", factor, float, int)
        if out is None:
            out = DenseOneIndex(self.shape[0])
        else:
            check_type("out", out, DenseOneIndex)
            if out.shape != self.shape:
                raise MatrixShapeError(
                    "The output argument has the incorrect shape."
                )
        out._array[:] = self._array * other._array * factor
        return out

    def dot(self, other, factor=1.0):
        """Dot product with other DenseOneIndex object, multiplied by factor

           **Arguments:**

           other
                A DenseOneIndex instance to be added

           **Optional arguments:**

           factor
                A scalar factor
        """
        check_type("other", other, DenseOneIndex)
        check_type("factor", factor, float, int)
        return np.dot(self._array, other._array) * factor

    def idivide(self, other, factor=1.0, out=None):
        """Divide self by two DenseOneIndex object, multiplied by factor

           **Arguments:**

           other
                A DenseOneIndex instance to be divided

           **Optional arguments:**

           factor
                A scalar factor

        """
        check_type("other", other, DenseOneIndex)
        check_type("factor", factor, float, int)
        if self.shape != other.shape:
            raise MatrixShapeError(
                "The output argument has the incorrect shape."
            )
        self._array[:] = np.divide(self._array, other._array) * factor

    def divide(self, other, factor=1.0, out=None):
        """Divide two DenseOneIndex object, multiplied by factor, and return
           output

           **Arguments:**

           other
                A DenseOneIndex instance to be divided

           **Optional arguments:**

           factor
                A scalar factor

           out
                The output DenseOneIndex
        """
        check_type("other", other, DenseOneIndex)
        check_type("factor", factor, float, int)
        if out is None:
            out = DenseOneIndex(other.shape[0])
        else:
            check_type("out", out, DenseOneIndex)
            if out.shape != self.shape or out.shape != other.shape:
                raise MatrixShapeError(
                    "The output argument has the incorrect shape."
                )
        out._array[:] = np.divide(self._array, other._array) * factor
        return out

    def sqrt(self):
        """Return the real part of the square root of one-index object"""
        out = DenseOneIndex(self.nbasis)
        out._array[:] = np.sqrt(self._array).real
        return out

    def isqrt(self):
        """Real part of the square root of one-index object"""
        self._array[:] = np.sqrt(self._array).real

    def inverse(self):
        """Return the inverse of one-index object"""
        out = DenseOneIndex(self.nbasis)
        out._array[:] = 1.0 / (self._array)
        return out

    #
    # Properties
    #

    @property
    def nbasis(self):
        """The number of basis functions"""
        return self._array.shape[0]

    @property
    def shape(self):
        """The shape of the object"""
        return self._array.shape


class DenseOrbital(Orbital):
    """An expansion of several functions in a basis with a dense matrix of
       coefficients. The implementation is such that the columns of self._array
       contain the orbitals.
    """

    #
    # Constructor and destructor
    #

    def __init__(self, nbasis, nfn=None):
        """
           **Arguments:**

           nbasis
                The number of basis functions.

           **Optional arguments:**

           nfn
                The number of functions to store. Defaults to nbasis.

           do_energies
                Also allocate an array to store an energy corresponding to each
                function.
        """
        if nfn is None:
            nfn = nbasis
        self._coeffs = np.zeros((nbasis, nfn))
        self._energies = np.zeros(nfn)
        self._occupations = np.zeros(nfn)
        log.mem.announce(
            self._coeffs.nbytes
            + self._energies.nbytes
            + self._occupations.nbytes
        )

    def __del__(self):
        if log is not None:
            if (
                hasattr(self, "_coeffs")
                and hasattr(self, "_energies")
                and hasattr(self, "_occupations")
            ):
                log.mem.denounce(
                    self._coeffs.nbytes
                    + self._energies.nbytes
                    + self._occupations.nbytes
                )

    #
    # Methods from base class
    #

    def __check_init_args__(self, nbasis, nfn=None):
        """Is self compatible with the given constructor arguments?"""
        if nfn is None:
            nfn = nbasis
        assert nbasis == self.nbasis
        assert nfn == self.nfn

    def __eq__(self, other):
        """Compare self with other"""
        return (
            isinstance(other, DenseOrbital)
            and other.nbasis == self.nbasis
            and other.nfn == self.nfn
            and (other._coeffs == self._coeffs).all()
            and (other._energies == self._energies).all()
            and (other._occupations == self._occupations).all()
        )

    @classmethod
    def from_hdf5(cls, grp):
        """Construct an instance from data previously stored in an h5py.Group.

           **Arguments:**

           grp
                An h5py.Group object.
        """
        if grp.attrs["class"] != cls.__name__:
            raise ArgumentError(
                "The class of the expansion in the HDF5 file does not match."
            )
        nbasis, nfn = grp["coeffs"].shape
        result = cls(nbasis, nfn)
        grp["coeffs"].read_direct(result._coeffs)
        grp["energies"].read_direct(result._energies)
        grp["occupations"].read_direct(result._occupations)
        return result

    def to_hdf5(self, grp):
        """Dump this object in an h5py.Group

           **Arguments:**

           grp
                An h5py.Group object.
        """
        grp.attrs["class"] = self.__class__.__name__
        grp["coeffs"] = self._coeffs
        grp["energies"] = self._energies
        grp["occupations"] = self._occupations

    # FIXME: rename into clean_copy
    def new(self):
        """Return a new expansion object with the same nbasis and nfn"""
        return DenseOrbital(self.nbasis, self.nfn)

    def _check_new_init_args(self, other):
        """Check whether an already initialized object is compatible"""
        other.__check_init_args__(self.nbasis, self.nfn)

    new.__check_init_args__ = _check_new_init_args

    def clear(self):
        """Reset all elements to zero."""
        self._coeffs[:] = 0.0
        self._energies[:] = 0.0
        self._occupations[:] = 0.0

    def clear_energies(self):
        """Reset all elements to zero."""
        self._energies[:] = 0.0

    def copy(self):
        """Return a copy of the object"""
        result = DenseOrbital(self.nbasis, self.nfn)
        result._coeffs[:] = self._coeffs
        result._energies[:] = self._energies
        result._occupations[:] = self._occupations
        return result

    def any(self):
        """Check if orbitals are empty"""
        return (
            np.any(self._coeffs)
            and np.any(self._energies)
            and np.any(self._occupations)
        )

    def assign(self, other):
        """Assign with the contents of another object

           **Arguments:**

           other
                Another DenseOrbital or DenseTwoIndex object.
                If DenseTwoIndex, energies and occupations are set to zero.
        """
        check_type("other", other, DenseOrbital, DenseTwoIndex)
        if isinstance(other, DenseOrbital):
            self._coeffs[:] = other._coeffs
            self._energies[:] = other._energies
            self._occupations[:] = other._occupations
        else:
            self._coeffs[:] = other._array
            self._energies[:] = 0.0
            self._occupations[:] = 0.0

    def itranspose(self):
        """In-place transpose"""
        self._coeffs[:] = self._coeffs.T

    def imul(self, other):
        """Inplace multiplication with other DenseOneIndex.

           The attributes ``energies`` and ``occupations`` are not altered.

           **Arguments:**

           other
                A DenseOneIndex object.
        """
        check_type("other", other, DenseOneIndex)
        self._coeffs[:] *= other._array

    def iscale_basis(self, other):
        """Inplace scaling of MO coeffs with other (np array).

           The attributes ``energies`` and ``occupations`` are not altered.

           **Arguments:**

           other
                A np.array object.
        """
        check_type("other", other, np.ndarray)
        self._coeffs[:] = other[np.newaxis].T * self._coeffs[:]

    def randomize(self):
        """Fill with random normal data"""
        self._coeffs[:] = np.random.normal(0, 1, self._coeffs.shape)
        self._energies[:] = np.random.normal(0, 1, self._energies.shape)
        self._occupations[:] = np.random.normal(0, 1, self._occupations.shape)

    def permute_basis(self, permutation):
        """Reorder the coefficients for a given permutation of basis functions
           (rows).

           **Arguments:**

           permutation
                An integer numpy array that defines the new order of the basis
                functions.
        """
        self._coeffs[:] = self.coeffs[permutation]

    def permute_orbitals(self, permutation, begin0=0, end0=None):
        """Reorder the coefficients for a given permutation of orbitals
           (columns).

           **Arguments:**

           permutation
                An integer numpy array that defines the new order of the
                orbitals.
        """
        end0 = self.nfn
        self._coeffs[:, begin0:end0] = self.coeffs[:, begin0:end0][
            :, permutation
        ]
        self._occupations[begin0:end0] = self.occupations[begin0:end0][
            permutation
        ]
        self._energies[begin0:end0] = self.energies[begin0:end0][permutation]

    def change_basis_signs(self, signs):
        """Correct for different sign conventions of the basis functions.

           **Arguments:**

           signs
                A numpy array with sign changes indicated by +1 and -1.
        """
        self._coeffs *= signs.reshape(-1, 1)

    def check_normalization(self, overlap, eps=1e-4):
        """Check that the occupied orbitals are normalized.

           **Arguments:**

           overlap
                The overlap two-index operator

           **Optional arguments:**

           eps
                The allowed deviation from unity, very loose by default.
        """
        check_type("overlap", overlap, DenseTwoIndex)
        for i in range(self.nfn):
            if self.occupations[i] == 0:
                continue
            norm = overlap.inner(self._coeffs[:, i], self._coeffs[:, i])
            # print i, norm
            assert abs(norm - 1) < eps, "The orbitals are not normalized!"

    def check_orthonormality(self, overlap, eps=1e-4):
        """Check that the occupied orbitals are orthogonal and normalized.

           **Arguments:**

           overlap
                The overlap two-index operator

           **Optional arguments:**

           eps
                The allowed deviation from unity, very loose by default.
        """
        check_type("overlap", overlap, DenseTwoIndex)
        for i0 in range(self.nfn):
            if self.occupations[i0] == 0:
                continue
            for i1 in range(i0 + 1):
                if self.occupations[i1] == 0:
                    continue
                dot = overlap.inner(self.coeffs[:, i0], self.coeffs[:, i1])
                if i0 == i1:
                    assert abs(dot - 1) < eps
                else:
                    assert abs(dot) < eps

    def error_eigen(self, fock, overlap):
        """Compute the error of the orbitals with respect to the eigenproblem

           **Arguments:**

           fock
                A DenseTwoIndex Hamiltonian (or Fock) operator.

           overlap
                A DenseTwoIndex overlap operator.

           **Returns:** the RMSD error on the orbital energies
        """
        check_type("fock", fock, DenseTwoIndex)
        check_type("overlap", overlap, DenseTwoIndex)
        errors = np.dot(fock._array, (self.coeffs)) - self.energies * np.dot(
            overlap._array, (self.coeffs)
        )
        return np.sqrt((abs(errors) ** 2).mean())

    #
    # Properties
    #

    @property
    def nbasis(self):
        """The number of basis functions"""
        return self._coeffs.shape[0]

    @property
    def nfn(self):
        """The number of orbitals (or functions in general)"""
        return self._coeffs.shape[1]

    @property
    def coeffs(self):
        """The matrix with the expansion coefficients"""
        return self._coeffs.view()

    @property
    def energies(self):
        """The orbital energies"""
        return self._energies.view()

    @property
    def occupations(self):
        """The orbital occupations"""
        return self._occupations.view()

    #
    # New methods for this implementation
    # TODO: consider adding these to base class
    # FIXME: These methods are specific for HF and should be moved to the HF module
    # or to the orbital_utils module

    def from_fock(self, fock, overlap):
        """Diagonalize a Fock matrix to obtain orbitals and energies

           This method updated the attributes ``coeffs`` and ``energies``
           in-place.

           **Arguments:**

           fock
                The fock matrix, an instance of DenseTwoIndex.

           overlap
                The overlap matrix, an instance of DenseTwoIndex.
        """
        check_type("fock", fock, DenseTwoIndex)
        check_type("overlap", overlap, DenseTwoIndex)
        evals, evecs = eigh(fock._array, overlap._array)
        self._energies[:] = evals[: self.nfn]
        self._coeffs[:] = evecs[:, : self.nfn]

    def from_fock_and_dm(self, fock, dm, overlap, epstol=1e-8):
        """Combined Diagonalization of a Fock and a density matrix

           This routine first diagonalizes the Fock matrix to obtain orbitals
           and orbital energies. Then, using first order (degenerate)
           perturbation theory, the occupation numbers are computed and, if
           needed, the the degeneracies of the Fock orbitals are lifted.
           It is assumed that the Fock and the density matrices commute.
           This method updated the attributes ``coeffs``, ``energies`` and
           ``occupations`` in-place.

           **Arguments:**

           fock
                The fock matrix, an instance of DenseTwoIndex.

           dm
                The density matrix, an instance of DenseTwoIndex.

           overlap
                The overlap matrix, an instance of DenseTwoIndex.

           **Optional arguments:**

           epstol
                The threshold for recognizing degenerate energy levels. When two
                subsequent energy levels are separated by an orbital energy less
                than ``epstol``, they are considered to be degenerate. When a
                series of energy levels have an orbital energy spacing between
                subsequent levels that is smaller than ``epstol``, they are all
                considered to be part of the same degenerate group. For every
                degenerate set of orbitals, the density matrix is used to (try
                to) lift the degeneracy.
        """
        # Diagonalize the Fock Matrix
        self.from_fock(fock, overlap)

        # Build clusters of degenerate orbitals. Rely on the fact that the
        # energy levels are sorted (one way or the other).
        clusters = []
        begin = 0
        for ifn in range(1, self.nfn):
            if abs(self.energies[ifn] - self.energies[ifn - 1]) > epstol:
                end = ifn
                clusters.append([begin, end])
                begin = ifn
        end = self.nfn
        clusters.append([begin, end])

        # Lift degeneracies using the density matrix
        sds = overlap.copy()
        sds.itranspose()
        sds.idot(dm)
        sds.idot(overlap)
        for begin, end in clusters:
            if end - begin == 1:
                self.occupations[begin] = sds.inner(
                    self.coeffs[:, begin], self.coeffs[:, begin]
                )
            else:
                # Build matrix
                mat = np.zeros((end - begin, end - begin), float)
                for i0 in range(end - begin):
                    for i1 in range(i0 + 1):
                        mat[i0, i1] = sds.inner(
                            self.coeffs[:, begin + i0],
                            self.coeffs[:, begin + i1],
                        )
                        mat[i1, i0] = mat[i0, i1]
                # Diagonalize and reverse order
                evals, evecs = np.linalg.eigh(mat)
                evals = evals[::-1]
                evecs = evecs[:, ::-1]
                # Rotate the orbitals
                self.coeffs[:, begin:end] = np.dot(
                    self.coeffs[:, begin:end], evecs
                )
                # Compute expectation values
                for i0 in range(end - begin):
                    self.occupations[begin + i0] = evals[i0]
                    self.energies[begin + i0] = fock.inner(
                        self.coeffs[:, begin + i0], self.coeffs[:, begin + i0]
                    )

    def derive_naturals(self, dm, overlap=None, sort=False, backtrafo=False):
        """
           **Arguments**:

           dm
                A DenseTwoIndex object with the density matrix

           **Optional arguments:**

           overlap
                A DenseTwoIndex object with the overlap matrix

           sort
                Sort natural orbitals according to occupation numbers

           backtrafo
                Transform back to original AO basis (dm represented as MO/MO
                is transformed back to AO/MO representation)
        """
        check_type("dm", dm, DenseTwoIndex)
        # diagonalize and compute eigenvalues
        if overlap is None:
            if not dm.is_symmetric():
                raise SymmetryError("Density matrix not symmetric.")
            evals, evecs = eigh(dm._array)
            # do backtransformation
            if backtrafo:
                evecs = np.dot(self._coeffs, evecs)
        else:
            check_type("overlap", overlap, DenseTwoIndex)
            # Construct a level-shifted operator
            occ = overlap.copy()
            occ.idot(dm)
            occ.idot(overlap)
            evals, evecs = eigh(occ._array, overlap._array)
        self._coeffs[:] = evecs[:, : self.nfn]
        self._occupations[:] = evals
        self._energies[:] = 0.0
        if sort:
            order = np.argsort(evals, axis=-1, kind="mergesort", order=None)[
                ::-1
            ]
            self.permute_orbitals(order)

    def get_homo_index(self, offset=0):
        """Return the index of a HOMO orbital.

           **Optional arguments**:

           offset
                By default, the (highest) homo energy is returned. When this
                index is above zero, the corresponding lower homo energy is
                returned.
        """
        if offset < 0:
            raise ArgumentError("Offset must be zero or positive.")
        homo_indexes = self.occupations.nonzero()[0]
        if len(homo_indexes) > offset:
            return homo_indexes[len(homo_indexes) - offset - 1]

    def get_homo_energy(self, offset=0):
        """Return a homo energy

           **Optional arguments**:

           offset
                By default, the (highest) homo energy is returned. When this
                index is above zero, the corresponding lower homo energy is
                returned.
        """
        index = self.get_homo_index(offset)
        if index is not None:
            return self.energies[index]

    homo_energy = property(get_homo_energy)

    def get_lumo_index(self, offset=0):
        """Return the index of a LUMO orbital.

           **Optional arguments**:

           offset
                By default, the (lowest) lumo energy is returned. When this
                index is above zero, the corresponding higher homo energy is
                returned.
        """
        if offset < 0:
            raise ArgumentError("Offset must be zero or positive.")
        lumo_indexes = (self.occupations == 0.0).nonzero()[0]
        if len(lumo_indexes) > offset:
            return lumo_indexes[offset]

    def get_lumo_energy(self, offset=0):
        """Return a lumo energy

           **Optional arguments**:

           offset
                By default, the (lowest) lumo energy is returned. When this
                index is above zero, the corresponding higher homo energy is
                returned.
        """
        index = self.get_lumo_index(offset)
        if index is not None:
            return self.energies[index]

    lumo_energy = property(get_lumo_energy)

    def assign_dot(self, left, right):
        """Dot product of orbitals in a DenseOrbital and TwoIndex object

           **Arguments:**

           left, right:
                An expansion and a two-index object, or a two-index and an expansion
                object.

           The transformed orbitals are stored in self.
        """
        if isinstance(left, DenseOrbital):
            check_type("left", left, DenseOrbital)
            check_type("right", right, DenseTwoIndex)
            if self.nbasis != left.nbasis:
                raise MatrixShapeError(
                    "Both expansions must have the same number of basis functions."
                )
            if right.shape[0] != left.nfn or right.shape[1] != self.nfn:
                raise MatrixShapeError(
                    "The shape of the two-index object is incompatible with that of the expansions."
                )
            self._coeffs[:] = np.dot(left.coeffs, right._array)
        elif isinstance(right, DenseOrbital):
            check_type("left", left, DenseTwoIndex)
            check_type("right", right, DenseOrbital)
            if self.nfn != right.nfn:
                raise MatrixShapeError(
                    "Both expansions must have the same number of orbitals."
                )
            if left.shape[1] != right.nbasis or left.shape[1] != self.nbasis:
                raise MatrixShapeError(
                    "The shape of the two-index object is incompatible with that of the expansions."
                )
            self._coeffs[:] = np.dot(left._array, right.coeffs)

    def assign_occupations(self, occupation, begin0=0, end0=None):
        """Assign orbital occupations

           **Arguments:**

           occupation
                The orbital occupations to be updated. An OneIndex instance
        """
        end0 = self.nbasis
        check_type("occupation", occupation, DenseOneIndex, DenseOrbital)
        if end0 - begin0 != occupation.nbasis:
            raise MatrixShapeError(
                "The expansion and one-index object must have the same number of basis functions."
            )
        if isinstance(occupation, DenseOneIndex):
            self._occupations[begin0:end0] = occupation._array
        elif isinstance(occupation, DenseOrbital):
            self._occupations[begin0:end0] = occupation._occupations[:]

    def rotate_random(self):
        """Apply random unitary transformation distributed with Haar measure

           The attributes ``energies`` and ``occupations`` are not altered.
        """
        z = np.random.normal(0, 1, (self.nfn, self.nfn))
        q, r = np.linalg.qr(z)
        self.coeffs[:] = np.dot(self.coeffs, q)

    def rotate_2orbitals(
        self, angle=0.7853981633974483, index0=None, index1=None
    ):
        """Rotate two orbitals

           **Optional arguments:**

           angle
                The rotation angle, defaults to 45 deg.

           index0, index1
                The orbitals to rotate, defaults to HOMO and LUMO,

           The attributes ``energies`` and ``occupations`` are not altered.
        """
        if index0 == None:
            index0 = self.get_homo_index()
        if index1 == None:
            index1 = self.get_lumo_index()
        old0 = self.coeffs[:, index0].copy()
        old1 = self.coeffs[:, index1].copy()
        self.coeffs[:, index0] = np.cos(angle) * old0 - np.sin(angle) * old1
        self.coeffs[:, index1] = np.sin(angle) * old0 + np.cos(angle) * old1

    def swap_orbitals(self, swaps, skip_occs=False, skip_energies=False):
        """Change the order of the orbitals using pair-exchange

           **Arguments:**

           swaps
                An integer numpy array with two columns where every row
                corresponds to one swap.

           The attributes ``energies`` and ``occupations`` are also reordered.
        """
        if not (
            swaps.shape[1] == 2
            and swaps.ndim == 2
            and issubclass(swaps.dtype.type, np.int_)
        ):
            raise ArgumentError("The argument swaps has the wrong shape/type.")
        for iswap in range(len(swaps)):
            index0, index1 = swaps[iswap]
            if log.do_medium:
                log(f"  Swapping orbitals {index0 + 1} and {index1 + 1}.")
            tmp = self.coeffs[:, index0].copy()
            self.coeffs[:, index0] = self.coeffs[:, index1]
            self.coeffs[:, index1] = tmp
            if not skip_energies:
                self.energies[index0], self.energies[index1] = (
                    self.energies[index1],
                    self.energies[index0],
                )
            if not skip_occs:
                self.occupations[index0], self.occupations[index1] = (
                    self.occupations[index1],
                    self.occupations[index0],
                )


class DenseTwoIndex(TwoIndex):
    """Dense two-dimensional matrix.
    """

    #
    # Constructor and destructor
    #

    def __init__(self, nbasis, nbasis1=None, label=""):
        """
           **Arguments:**

           nbasis
                The number of basis functions. (Number of rows. Also number of
                columns, unless nbasis1 is given.)

           **Optional arguments:**

           nbasis1
                When given, this is the number of columns (second index).

           Note that by default the two-index object is assumed to be Hermitian.
           Only when nbasis1 is given, this assumption is dropped.
        """
        if nbasis1 is None:
            nbasis1 = nbasis
        self._array = np.zeros((nbasis, nbasis1), float)
        self._label = label
        log.mem.announce(self._array.nbytes)

    def __del__(self):
        if log is not None:
            if hasattr(self, "_array"):
                log.mem.denounce(self._array.nbytes)
        if hasattr(self, "_array"):
            del self._array

    #
    # Methods from base class
    #

    def __check_init_args__(self, nbasis, nbasis1=None):
        """Is self compatible with the given constructor arguments?"""
        if nbasis1 is None:
            nbasis1 = nbasis
        assert nbasis == self.nbasis
        assert nbasis1 == self.nbasis1

    def __eq__(self, other):
        """Compare self with other"""
        return (
            isinstance(other, DenseTwoIndex)
            and other.nbasis == self.nbasis
            and other.nbasis1 == self.nbasis1
            and (other._array == self._array).all()
        )

    @classmethod
    def from_hdf5(cls, grp):
        """Construct an instance from data previously stored in an h5py.Group.

           **Arguments:**

           grp
                An h5py.Group object.
        """
        nbasis = grp["array"].shape[0]
        nbasis1 = grp["array"].shape[1]
        label = grp.attrs["label"]
        result = cls(nbasis, nbasis1, label)
        grp["array"].read_direct(result._array)
        return result

    def to_hdf5(self, grp):
        """Dump this object in an h5py.Group

           **Arguments:**

           grp
                An h5py.Group object.
        """
        grp.attrs["class"] = self.__class__.__name__
        grp["array"] = self._array
        grp.attrs["label"] = self._label

    # FIXME: rename into clean_copy
    def new(self):
        """Return a new two-index object with the same nbasis (and nbasis1)"""
        return DenseTwoIndex(self.nbasis, self.nbasis1)

    def einsum_index(self, script):
        """ Returns indices to numpy.einsum summation. """
        assert len(script) == 2 and isinstance(script, str)
        return script

    @property
    def arrays(self):
        """ Returns list containing self._array """
        return [self._array]

    @property
    def label(self):
        """ Returns label of instance"""
        return self._label

    @label.setter
    def label(self, label):
        """ Sets label of instance"""
        self._label = label

    def _check_new_init_args(self, other):
        """Check whether an already initialized object is compatible"""
        other.__check_init_args__(self.nbasis, self.nbasis1)

    new.__check_init_args__ = _check_new_init_args

    def replace_array(self, value):
        """Replaces an array with another array, if not present, it will be
           generated.
        """
        if isinstance(value, DenseTwoIndex):
            self._array = value._array
        elif isinstance(value, np.ndarray):
            self._array = value
        else:
            raise ArgumentError(
                f"Do not know how to assign object of type {type(value)}."
            )

    def clear(self):
        """Reset all elements to zero."""
        self._array[:] = 0.0

    def copy(self, begin0=0, end0=None, begin1=0, end1=None):
        """Return a copy of (a part of) the object

           **Optional arguments:**

           begin0, end0, begin1, end1
                Can be used to select a subblock of the object. When not given,
                the full range is used. Only when the ranges are equal for both
                axes, an Hermitian two-index may be returned.
        """
        end0, end1 = self._fix_ends(end0, end1)
        nbasis = end0 - begin0
        nbasis1 = end1 - begin1
        result = DenseTwoIndex(nbasis, nbasis1)
        result._array[:] = self._array[begin0:end0, begin1:end1]
        result._label = self._label
        return result

    def ravel(self, begin0=0, end0=None, begin1=0, end1=None, ind=None):
        """Return a copy of (a part of) the object as a one-index object

           **Optional arguments:**

           begin0, end0, begin1, end1
                Can be used to select a subblock of the object. When not given,
                the full range is used. Only when the ranges are equal for both
                axes, an Hermitian two-index may be returned.

           ind
                2-Tuple of 1-dim arrays with row and column indices of TwoIndex
                object to be copied.
        """
        end0, end1 = self._fix_ends(end0, end1)
        nbasis = end0 - begin0
        nbasis1 = end1 - begin1
        if ind is None:
            result = DenseOneIndex(nbasis * nbasis1)
            result._array[:] = self._array[begin0:end0, begin1:end1].ravel(
                order="C"
            )
        else:
            if not ind[0].shape == ind[1].shape:
                raise ArgumentError("Number of array indices not identical.")
            result = DenseOneIndex(len(ind[0]))
            result._array[:] = (self._array[begin0:end0, begin1:end1])[
                ind
            ].ravel(order="C")
        return result

    def assign(
        self,
        other,
        ind=None,
        begin0=0,
        end0=None,
        begin1=0,
        end1=None,
        begin2=0,
        end2=None,
        begin3=0,
        end3=None,
    ):
        """Assign a new contents to the two-index object

           **Arguments:**

           other
                The new data, may be DenseTwoIndex, DenseOneIndex, a scalar
                value, or an ndarray.

           ind
                If provided, only these elements (of DenseOneIndex) are
                assigned

           begin0, end0, begin1, end1
                When given, specify the ranges where the contribution will be
                added. When not given, the full range is used.
        """
        end0, end1 = self._fix_ends(end0, end1)
        dim0 = end0 - begin0
        dim1 = end1 - begin1
        if isinstance(other, DenseTwoIndex):
            end2, end3 = other._fix_ends(end2, end3)
            self._array[begin0:end0, begin1:end1] = other._array[
                begin2:end2, begin3:end3
            ]
        elif isinstance(other, float) or isinstance(other, int):
            self._array[begin0:end0, begin1:end1] = other
        elif isinstance(other, np.ndarray):
            if ind is None:
                if other.shape == self.shape:
                    self._array[:] = other
                else:
                    if end2 is None:
                        end2 = other.size
                    self._array[begin0:end0, begin1:end1] = other[
                        begin2:end2
                    ].reshape((dim0, dim1))
            else:
                self._array[begin0:end0, begin1:end1][ind] = other
        elif isinstance(other, DenseOneIndex):
            end2 = other._fix_ends(end2)[0]
            if ind is None:
                self._array[begin0:end0, begin1:end1] = (
                    other._array[begin2:end2]
                ).reshape((dim0, dim1))
            else:
                self._array[begin0:end0, begin1:end1][ind] = other._array[
                    begin2:end2
                ]
        else:
            raise ArgumentError(
                f"Do not know how to assign object of type {type(other)}"
            )

    def assign_dot(self, other, tf2):
        """Dot product of orbitals in a DenseOrbital and TwoIndex object

           **Arguments:**

           other
                An expansion object with input orbitals

           tf2
                A two-index object

           The transformed array is stored in self.
        """
        check_type("other", other, DenseOrbital)
        check_type("tf2", tf2, DenseTwoIndex)
        if not (self.nbasis == other.nbasis):
            raise MatrixShapeError(
                "Both expansions must have the same number of basis functions."
            )
        if not (tf2.shape[0] == other.nfn and tf2.shape[1] == self.shape[1]):
            raise MatrixShapeError(
                "The shape of the two-index object is incompatible with that of the expansions."
            )
        self._array[:] = np.dot(other.coeffs, tf2._array)

    def iadd(
        self,
        other,
        factor=1.0,
        begin0=0,
        end0=None,
        begin1=0,
        end1=None,
        begin2=0,
        end2=None,
        begin3=0,
        end3=None,
        transpose=False,
    ):
        """Add another DenseTwoIndex object in-place, multiplied by factor. If
           begin0, end0, begin1, end1 are specified, other is added to the
           selected range.

           **Arguments:**

           other
                A DenseTwoIndex, DenseOneIndex instance or float to be added

           **Optional arguments:**

           factor
                A scalar factor

           begin0, end0, begin1, end1
                When given, specify the ranges where the contribution will be
                added. When not given, the full range is used.
        """
        check_type("factor", factor, float, int)
        end0, end1 = self._fix_ends(end0, end1)
        if isinstance(other, DenseTwoIndex):
            end2, end3 = other._fix_ends(end2, end3)
            if transpose:
                self._array[begin0:end0, begin1:end1] += (
                    other._array[begin2:end2, begin3:end3]
                ).T * factor
            else:
                self._array[begin0:end0, begin1:end1] += (
                    other._array[begin2:end2, begin3:end3] * factor
                )
        elif isinstance(other, float):
            self._array[begin0:end0, begin1:end1] += other * factor
        elif isinstance(other, DenseOneIndex):
            end2 = other._fix_ends(end2)[0]
            if transpose:
                self._array[begin0:end0, begin1:end1] += (
                    other._array[begin2:end2] * factor
                )
            else:
                self._array[begin0:end0, begin1:end1] += (
                    other._array[begin2:end2]
                )[np.newaxis].T * factor
        else:
            raise ArgumentError(
                "Do not know how to add in-place an object of type"
                f" {type(other)}."
            )

    def iadd_t(
        self,
        other,
        factor=1.0,
        begin0=0,
        end0=None,
        begin1=0,
        end1=None,
        begin2=0,
        end2=None,
        begin3=0,
        end3=None,
    ):
        """See :py:meth:`DenseTwoIndex.iadd`, transpose=True"""
        self.iadd(
            other,
            factor,
            begin0,
            end0,
            begin1,
            end1,
            begin2,
            end2,
            begin3,
            end3,
            True,
        )

    def iadd_diagonal(self, other, factor=1.0, begin0=0, end0=None):
        """Add to diagonal of DenseTwoIndex object in-place, multiplied by
           factor.

           **Arguments:**

           other
                A DenseOneIndex instance to be added

           **Optional arguments:**

           factor
                A scalar factor

           begin0, end0, begin1, end1
                When given, specify the ranges of contribution to be
                added. When not given, the full range is used.
        """
        check_type("factor", factor, float, int)
        check_type("other", other, DenseOneIndex, float)
        end1 = None
        end0, end1 = self._fix_ends(end0, end1)
        if isinstance(other, DenseOneIndex):
            for i, j in zip(range(begin0, end0), range(end0 - begin0)):
                self._array[i, i] += other._array[j] * factor
        elif isinstance(other, float):
            for i in range(begin0, end0):
                self._array[i, i] += other * factor

    def iscale(self, factor):
        """In-place multiplication with a scalar

           **Arguments:**

           factor
                A scalar factor.
        """
        check_type("factor", factor, float, int)
        self._array *= factor

    def iortho(self):
        """In-place orthogonalization

           **Arguments:**

        """
        q, r = np.linalg.qr(self._array, mode="full")
        self._array[:] = q

    def randomize(self):
        """Fill with random normal data"""
        self._array[:] = np.random.normal(0, 1, self.shape)

    def permute_basis(self, permutation):
        """Reorder the coefficients for a given permutation of basis functions.

           The same permutation is applied to all indexes.

           **Arguments:**

           permutation
                An integer numpy array that defines the new order of the basis
                functions.
        """
        self._array[:] = self._array.take(permutation, axis=0)
        self._array[:] = self._array.take(permutation, axis=1)

    def permute(self, permutation):
        """Reorder the coefficients for a given permutation of columns.

           **Arguments:**

           permutation
                An integer numpy array that defines the new order of the
                orbitals.
        """
        self._array[:] = self._array[:, permutation]

    def change_basis_signs(self, signs):
        """Correct for different sign conventions of the basis functions.

           **Arguments:**

           signs
                A numpy array with sign changes indicated by +1 and -1.
        """
        self._array *= signs
        self._array *= signs.reshape(-1, 1)

    def get_element(self, i, j):
        """Return a matrix element"""
        return self._array[i, j]

    def set_element(self, i, j, value, symmetry=2):
        """Set a matrix element

           **Arguments:**

           i, j
                The matrix indexes to be set

           value
                The value to be assigned to the matrix element.

           **Optional arguments:**

           symmetry
                When 2 (the default), the element (j,i) is set to the same
                value. When set to 1 the opposite off-diagonal is not set. See
                :ref:`dense_matrix_symmetry` for more details.
        """
        check_options("symmetry", symmetry, 1, 2)
        if not self.is_shape_symmetric(symmetry):
            raise SymmetryError(
                "TwoIndex object does not have the right shape to impose the selected symmetry."
            )
        self._array[i, j] = value
        if symmetry == 2:
            self._array[j, i] = value

    def sum(self, begin0=0, end0=None, begin1=0, end1=None):
        """Return the sum of all elements (in the selected range)

           **Optional arguments:**

           begin0, end0, begin1, end1
                Can be used to select a subblock of the object to be contracted.
        """
        end0, end1 = self._fix_ends(end0, end1)
        return self._array[begin0:end0, begin1:end1].sum()

    def trace(self, begin0=0, end0=None, begin1=0, end1=None):
        """Return the trace of the two-index object.

           **Optional arguments:**

           begin0, end0, begin1, end1
                Can be used to select a subblock of the object to be contracted.
        """
        end0, end1 = self._fix_ends(end0, end1)
        if end0 - begin0 != end1 - begin1:
            raise ArgumentError(
                "Only the trace of a square (part of a) two-index object can be computed."
            )
        return np.trace(self._array[begin0:end0, begin1:end1])

    def itranspose(self):
        """In-place transpose"""
        self._array[:] = self._array.T

    def inner(self, vec0, vec1):
        """Compute an inner product of two vectors using the two-index as a metric

           **Arguments:**

           vec0, vec1
                The vectors, either DenseOneIndex or numpy arrays.
        """
        if vec0.shape != (self.shape[0],):
            raise MatrixShapeError(
                "The length of vec0 does not match the shape of the two-index object."
            )
        if vec1.shape != (self.shape[1],):
            raise MatrixShapeError(
                "The length of vec1 does not match the shape of the two-index object."
            )
        if isinstance(vec0, DenseOneIndex) and isinstance(vec1, DenseOneIndex):
            return np.dot(vec0._array, np.dot(self._array, vec1._array))
        elif isinstance(vec0, np.ndarray) and isinstance(vec1, np.ndarray):
            return np.dot(vec0, np.dot(self._array, vec1))
        else:
            raise ArgumentError(
                "Do not know how to compute inner product with objects of "
                f"type {type(vec0)} and {type(vec1)}"
            )

    def sqrt(self):
        """Return the real part of the square root of two-index object"""
        out = DenseTwoIndex(self.nbasis, self.nbasis1)
        out._array[:] = sqrtm(self._array).real
        return out

    def inverse(self):
        """Return the inverse of two-index object"""
        out = DenseTwoIndex(self.nbasis, self.nbasis1)
        out._array[:] = inv(self._array)
        return out

    def diagonalize(self, eigvec=False):
        """Return eigenvalues of two-index object"""
        from scipy.linalg import eigvals

        if eigvec:
            evalue, evector = eig(self._array)
            return evalue, evector
        else:
            evalue = eigvals(self._array)
            return evalue

    def det(self):
        """Return determinant of two-index object"""
        if self.nbasis != self.nbasis1:
            raise MatrixShapeError(
                f"Array has the wrong shape {self.nbasis} x {self.nbasis2}"
            )
        (sign, logdet) = np.linalg.slogdet(self._array)
        out = sign * np.exp(logdet)
        return out

    def divide(
        self,
        other,
        factor=1.0,
        out=None,
        begin0=0,
        end0=None,
        begin1=0,
        end1=None,
    ):
        """Divide two DenseTwoIndex object, multiplied by factor, and return
           output

           **Arguments:**

           other
                A DenseTwoIndex instance to be divided. If DenseOneIndex,
                the object will be reshaped assuming C-order.

           **Optional arguments:**

           factor
                A scalar factor

           out
                The output DenseTwoIndex
        """
        end0, end1 = self._fix_ends(end0, end1)
        check_type("other", other, DenseTwoIndex, DenseOneIndex)
        check_type("factor", factor, float, int)
        if out is None:
            out = DenseTwoIndex((end0 - begin0), (end1 - begin1))
        else:
            check_type("out", out, DenseTwoIndex)
            if out.shape != self._array[begin0:end0, begin1:end1].shape:
                raise MatrixShapeError(
                    "The output argument has the incorrect shape."
                )
            if isinstance(other, DenseTwoIndex):
                if out.shape != other.shape:
                    raise MatrixShapeError(
                        "The output argument has the incorrect shape."
                    )
        if isinstance(other, DenseOneIndex):
            out._array[:] = (
                np.divide(
                    self._array[begin0:end0, begin1:end1],
                    other._array.reshape(self.nbasis, self.nbasis1),
                )
                * factor
            )
        else:
            out._array[:] = (
                np.divide(self._array[begin0:end0, begin1:end1], other._array)
                * factor
            )
        return out

    #
    # Properties
    #

    @property
    def nbasis(self):
        """The number of basis functions"""
        return self._array.shape[0]

    @property
    def nbasis1(self):
        """The other size of the two-index object"""
        return self.shape[1]

    @property
    def shape(self):
        """The shape of the object"""
        return self._array.shape

    #
    # New methods for this implementation
    # TODO: consider adding these to base class
    #

    def assign_diagonal(self, value, factor=1.0, begin=0, end=None):
        """Set diagonal elements to value

           **Arguments:**

           value
                Either a scalar or a DenseOneIndex object
        """
        end, end1 = self._fix_ends(end, end)
        if isinstance(value, DenseOneIndex):
            np.fill_diagonal(
                self._array[begin:end, begin:end], value._array * factor
            )
        elif isinstance(value, np.ndarray):
            np.fill_diagonal(self._array[begin:end, begin:end], value * factor)
        elif isinstance(value, float):
            np.fill_diagonal(self._array[begin:end, begin:end], value * factor)
        elif isinstance(value, DenseTwoIndex):
            np.fill_diagonal(
                self._array[begin:end, begin:end],
                value._array.ravel(order="C") * factor,
            )
        else:
            raise ArgumentError(
                "Do not know how to set diagonal with object of type "
                f"{type(value)}"
            )

    def copy_diagonal(self, out=None, begin=0, end=None):
        """Copy (part of) the diagonal of the two-index object

           **Optional arguments:**

           out
                The output argument (DenseOneIndex with proper size).

           begin, end
                Can be used to select a range of the diagonal. If not given,
                then the entire diagonal is copied.
        """
        if not self.shape[0] == self.shape[1]:
            raise MatrixShapeError(
                "The diagonal can only be copied when the two-index object is squared."
            )
        end, end1 = self._fix_ends(end, end)
        if out is None:
            out = DenseOneIndex(end - begin)
        else:
            check_type("out", out, DenseOneIndex)
            if out.shape != (end - begin,):
                raise MatrixShapeError("The output argument has the incorrect shape.")
        out._array[:] = np.diagonal(self._array[begin:end, begin:end])
        return out

    def is_symmetric(self, symmetry=2, rtol=1e-5, atol=1e-8):
        """Check the symmetry of the array.

           **Optional arguments:**

           symmetry
                The symmetry to check. See :ref:`dense_matrix_symmetry`
                for more details.

           rtol and atol
                relative and absolute tolerance. See to ``np.allclose``.
        """
        check_options("symmetry", symmetry, 1, 2)
        if not self.is_shape_symmetric(symmetry):
            return False
        if symmetry == 2:
            return np.allclose(self._array, self._array.T, rtol, atol)
        return True

    def is_shape_symmetric(self, symmetry):
        """Check whether the symmetry argument matches the shape"""
        return symmetry == 1 or self.nbasis == self.nbasis1

    def check_symmetric(self, rtol=1e-5, atol=1e-8):
        """Check the symmetry of the array.

           **Optional arguments:**

           rtol and atol
                relative and absolute tolerance. See to ``np.allclose``.
        """
        return np.allclose(self._array, self._array.T, rtol, atol)

    def symmetrize(self, symmetry=2):
        """Symmetrize in-place

           **Optional arguments:**

           symmetry
                The symmetry to impose. See :ref:`dense_matrix_symmetry` for
                more details.
        """
        check_options("symmetry", symmetry, 1, 2)
        if not self.is_shape_symmetric(symmetry):
            raise SymmetryError(
                "TwoIndex object does not have the right shape for symmetrization."
            )
        if symmetry == 2:
            self._array[:] = self._array + self._array.T
            self.iscale(0.5)

    def iadd_outer(self, other0, other1, factor=1.0):
        """In-place addition of outer product of two other DenseTwoIndex

           **Arguments:**

           other0, other1
                Two-index objects that go into the outer product. They are
                raveled before taking the outer product.

           **Optional arguments:**

           factor
                The term added is scaled by this factor.
        """
        check_type("other0", other0, DenseTwoIndex)
        check_type("other1", other1, DenseTwoIndex)
        check_type("factor", factor, float, int)
        self._array += (
            np.outer(other0._array.ravel(), other1._array.ravel()) * factor
        )

    def iadd_kron(self, other0, other1, factor=1.0):
        """In-place addition of kronecker product of two other DenseTwoIndex

           **Arguments:**

           other0, other1
                Two-index objects that go into the kronecker product.

           **Optional arguments:**

           factor
                The term added is scaled by this factor.
        """
        check_type("other0", other0, DenseTwoIndex)
        check_type("other1", other1, DenseTwoIndex)
        check_type("factor", factor, float, int)
        self._array += np.kron(other0._array, other1._array) * factor

    def iadd_dot(
        self,
        other0,
        other1,
        factor=1.0,
        begin0=0,
        end0=None,
        begin1=0,
        end1=None,
        begin2=0,
        end2=None,
        begin3=0,
        end3=None,
    ):
        """In-place addition of dot product: ``other0 * other1``

           **Arguments:**

           other0, other1
                Two-index objects that go into the kronecker product.

           **Optional arguments:**

           factor
                The term added is scaled by this factor.

           begin0, end0, begin1, end1
                Can be used to select a subblock of the other0 object. When
                not given, the full range is used.

           begin2, end2, begin3, end3
                Can be used to select a subblock of the other1 object. When
                not given, the full range is used.
        """
        check_type("other0", other0, DenseTwoIndex)
        check_type("other1", other1, DenseTwoIndex)
        check_type("factor", factor, float, int)
        end0, end1 = other0._fix_ends(end0, end1)
        end2, end3 = other1._fix_ends(end2, end3)
        self._array[:] += (
            np.dot(
                other0._array[begin0:end0, begin1:end1],
                other1._array[begin2:end2, begin3:end3],
            )
            * factor
        )

    def iadd_tdot(self, other0, other1, factor=1.0):
        """In-place addition of dot product: ``other0.T * other1``

           **Arguments:**

           other0, other1
                Two-index objects that go into the kronecker product.

           **Optional arguments:**

           factor
                The term added is scaled by this factor.
        """
        check_type("other0", other0, DenseTwoIndex)
        check_type("other1", other1, DenseTwoIndex)
        check_type("factor", factor, float, int)
        self._array[:] += np.dot(other0._array.T, other1._array) * factor

    def iadd_dott(self, other0, other1, factor=1.0):
        """In-place addition of dot product: ``other0 * other1.T``

           **Arguments:**

           other0, other1
                Two-index objects that go into the kronecker product.

           **Optional arguments:**

           factor
                The term added is scaled by this factor.
        """
        check_type("other0", other0, DenseTwoIndex)
        check_type("other1", other1, DenseTwoIndex)
        check_type("factor", factor, float, int)
        self._array[:] += np.dot(other0._array, other1._array.T) * factor

    def iadd_transform(self, other, omega, factor=1.0, transpose=False):
        """Perform transformation: ``self += omega.T * other * omega``

           **Arguments:**

           other
                Something computed in some basis

           omega
                The transformation matrix.

           **Optional arguments:**

           transpose
                When given, the following is computed: ``omega * other * omega.T``
        """
        if transpose:
            self._array[:] += (
                reduce(np.dot, [omega._array, other._array, omega._array.T])
                * factor
            )
        else:
            self._array[:] += (
                reduce(np.dot, [omega._array.T, other._array, omega._array])
                * factor
            )

    def iadd_mult(
        self,
        other0,
        other1,
        factor=1.0,
        begin0=0,
        end0=None,
        begin1=0,
        end1=None,
        begin2=0,
        end2=None,
        begin3=0,
        end3=None,
        transpose0=False,
        transpose1=False,
    ):
        """In-place addition of multiplication: ``other0 * other1``

           **Arguments:**

           other0, other1
                Two-index objects that go into the product.

           **Optional arguments:**

           factor
                The term added is scaled by this factor.

           begin0, end0, begin1, end1
                Can be used to select a subblock of the other0 object. When
                not given, the full range is used.

           begin2, end2, begin3, end3
                Can be used to select a subblock of the other1 object. When
                not given, the full range is used.

           transpose0, transpose1
                Can be used to select transpose of other0, other1 objects
        """
        check_type("other0", other0, DenseTwoIndex)
        check_type("other1", other1, DenseTwoIndex)
        check_type("factor", factor, float, int)
        check_type("transpose0", transpose0, bool)
        check_type("transpose1", transpose1, bool)
        end0, end1 = other0._fix_ends(end0, end1)
        end2, end3 = other1._fix_ends(end2, end3)
        if transpose0 and transpose1:
            self._array[:] += (
                other0._array[begin0:end0, begin1:end1].T
                * other1._array[begin2:end2, begin3:end3].T
            ) * factor
        elif transpose0:
            self._array[:] += (
                other0._array[begin0:end0, begin1:end1].T
                * other1._array[begin2:end2, begin3:end3]
            ) * factor
        elif transpose1:
            self._array[:] += (
                other0._array[begin0:end0, begin1:end1]
                * other1._array[begin2:end2, begin3:end3].T
            ) * factor
        else:
            self._array[:] += (
                other0._array[begin0:end0, begin1:end1]
                * other1._array[begin2:end2, begin3:end3]
            ) * factor

    def iadd_one_mult(
        self, other0, other1, factor=1.0, transpose0=False, transpose1=False
    ):
        """In-place addition of multiplication: ``other0 * other1``

           **Arguments:**

           other0, other1
                One-index objects that go into the product.

           **Optional arguments:**

           factor
                The term added is scaled by this factor.

           transpose0, transpose1
                Can be used to select transpose of one-index objects.
        """
        check_type("other0", other0, DenseOneIndex)
        check_type("other1", other1, DenseOneIndex)
        check_type("factor", factor, float, int)
        check_type("transpose0", transpose0, bool)
        check_type("transpose1", transpose1, bool)
        if transpose0 and transpose1:
            self._array[:] += (
                other0._array[np.newaxis].T * other1._array[np.newaxis].T
            ) * factor
        elif transpose0:
            self._array[:] += (
                other0._array[np.newaxis].T * other1._array
            ) * factor
        elif transpose1:
            self._array[:] += (
                other0._array * other1._array[np.newaxis].T
            ) * factor
        else:
            self._array[:] += (other0._array * other1._array) * factor

    def iadd_shift(self, lshift):
        """Add positive shift to elements. If negative subtract shift

           **Arguments:**

           lshift
                A scalar used to augment the matrix elements.
        """
        check_type("lshift", lshift, float, int)
        self._array[self._array >= 0] += lshift
        self._array[self._array < 0] -= lshift

    def iadd_expand_one_to_two(
        self,
        axis,
        one,
        factor=1.0,
        begin0=0,
        end0=None,
        abegin0=0,
        aend0=None,
        abegin1=0,
        aend1=None,
    ):
        """Expand one-index object along one axes and add it to two-index
           object.

           **Arguments:**

           axis
                Any of

           one
                An instance of DenseOneIndex.

           **Optional arguments:**

           factor
                See :py:meth:`DenseLinalgFactory.tco`

           begin0, end0
                Can be used to select a subblock of the one-index object (one).
                When not given, the full range is used.
        """
        check_options("axis", axis, "a->ab", "b->ab")
        end0 = one._fix_ends(end0)[0]
        aend0, aend1 = self._fix_ends(aend0, aend1)
        if axis == "a->ab":
            for i in range(abegin1, aend1):
                self._array[:, i] += one._array[begin0:end0] * factor
        if axis == "b->ab":
            for i in range(abegin0, aend0):
                self._array[i, :] += one._array[begin0:end0] * factor

    def idot(self, other, transpose=False):
        """In-place dot product: self = self * other

           **Arguments:**

           other
                The other array
        """
        check_type("other", other, DenseTwoIndex, DenseOrbital)
        if isinstance(other, DenseTwoIndex):
            if transpose:
                self._array = np.dot(self._array, other._array.T)
            else:
                self._array = np.dot(self._array, other._array)
        else:
            if transpose:
                self._array = np.dot(self._array, other.coeffs.T)
            else:
                self._array = np.dot(self._array, other.coeffs)

    def idot_t(self, other):
        """In-place dot product: self = self * other

           **Arguments:**

           other
                The other array
        """
        self.idot(other, transpose=True)

    def imul(
        self, other, factor=1.0, begin2=0, end2=None, begin3=0, end3=None
    ):
        """In-place element-wise multiplication: ``self *= other * factor``

           **Arguments:**

           other
                A DenseTwoIndex instance.

           **Optional arguments:**

           factor
                The two-index object is scaled by this factor.

           begin0, end0, begin1, end1
                Can be used to select a subblock of the (other) object. When
                not given, the full range is used.
        """
        check_type("other", other, DenseOneIndex, DenseTwoIndex)
        check_type("factor", factor, float, int)
        if isinstance(other, DenseTwoIndex):
            end2, end3 = other._fix_ends(end2, end3)
            self._array *= other._array[begin2:end2, begin3:end3]
        else:
            end2 = other._fix_ends(end2)[0]
            self._array *= other._array[begin2:end2]
        self.iscale(factor)

    def imul_t(
        self, other, factor=1.0, begin2=0, end2=None, begin3=0, end3=None
    ):
        """In-place element-wise multiplication: ``self *= other.T * factor``

           **Arguments:**

           other
                A DenseTwoIndex instance.

           **Optional arguments:**

           factor
                The two-index object is scaled by this factor.

           begin0, end0, begin1, end1
                Can be used to select a subblock of the (other) object. When
                not given, the full range is used.
        """
        check_type("other", other, DenseTwoIndex)
        check_type("factor", factor, float, int)
        end2, end3 = other._fix_ends(end2, end3)
        self._array *= other._array[begin2:end2, begin3:end3].T
        self.iscale(factor)

    def itransform(self, omega, factor=1.0, transpose=False):
        """Perform transformation: ``self = omega.T * self * omega``

           **Arguments:**

           other
                Something computed in some basis

           omega
                The transformation matrix.

           **Optional arguments:**

           transpose
                When given, the following is computed: ``omega * other * omega.T``
        """
        if transpose:
            self._array[:] = (
                reduce(np.dot, [omega._array, self._array, omega._array.T])
                * factor
            )
        else:
            self._array[:] = (
                reduce(np.dot, [omega._array.T, self._array, omega._array])
                * factor
            )

    def distance_inf(self, other):
        """The infinity norm distance between self and other

           **Arguments:**

           other
                A DenseTwoIndex instance.
        """
        check_type("other", other, DenseTwoIndex)
        return abs(self._array.ravel() - other._array.ravel()).max()

    def iabs(self):
        """In-place absolute values"""
        self._array[:] = abs(self._array)

    def assign_two_index_transform(self, ao_integrals, exp0, exp1=None):
        """Perform two index transformation: ``exp0.T * ao_integrals * exp0``

           **Arguments:**

           ao_integrals
                Something computed in the atomic orbital basis

           exp0
                The molecular orbitals.

           **Optional arguments:**

           exp1
                When given, the following is computed: ``exp0.T * ao_integrals *
                exp1``
        """
        if exp1 is None:
            exp1 = exp0
        self._array[:] = reduce(
            np.dot, [exp0.coeffs.T, ao_integrals._array, exp1.coeffs]
        )


class DenseThreeIndex(ThreeIndex):
    """Dense three-dimensional object.

       This is the most inefficient implementation in terms of memory usage and
       computer time. Due to its simplicity, it is trivial to implement. This
       implementation mainly serves as a reference for testing purposes.
    """

    #
    # Constructor and destructor
    #

    def __init__(self, nbasis, nbasis1=None, nbasis2=None, label=""):
        """
           **Arguments:**

           nbasis
                The number of basis functions.
        """
        if nbasis1 is None:
            nbasis1 = nbasis
        if nbasis2 is None:
            nbasis2 = nbasis
        self._array = np.zeros((nbasis, nbasis1, nbasis2), float)
        self._label = label
        log.mem.announce(self._array.nbytes)

    def __del__(self):
        if log is not None:
            if hasattr(self, "_array"):
                log.mem.denounce(self._array.nbytes)
        if hasattr(self, "_array"):
            del self._array

    #
    # Methods from base class
    #

    def __check_init_args__(self, nbasis, nbasis1=None, nbasis2=None):
        """Is self compatible with the given constructor arguments?"""
        if nbasis1 is None:
            nbasis1 = nbasis
        if nbasis2 is None:
            nbasis2 = nbasis
        assert nbasis == self.nbasis
        assert nbasis1 == self.nbasis1
        assert nbasis2 == self.nbasis2

    def __eq__(self, other):
        """Compare self with other"""
        return (
            isinstance(other, DenseThreeIndex)
            and other.nbasis == self.nbasis
            and other.nbasis1 == self.nbasis1
            and other.nbasis2 == self.nbasis2
            and (other._array == self._array).all()
        )

    @classmethod
    def from_hdf5(cls, grp):
        """Construct an instance from data previously stored in an h5py.Group.

           **Arguments:**

           grp
                An h5py.Group object.
        """
        nbasis = grp["array"].shape[0]
        nbasis1 = grp["array"].shape[1]
        nbasis2 = grp["array"].shape[2]
        label = grp.attrs["label"]
        result = cls(nbasis, nbasis1, nbasis2, label)
        grp["array"].read_direct(result._array)
        return result

    def to_hdf5(self, grp):
        """Dump this object in an h5py.Group

           **Arguments:**

           grp
                An h5py.Group object.
        """
        grp.attrs["class"] = self.__class__.__name__
        grp["array"] = self._array
        grp.attrs["label"] = self._label

    # FIXME: rename into clean_copy
    def new(self):
        """Return a new three-index object with the same nbasis"""
        return DenseThreeIndex(self.nbasis, self.nbasis1, self.nbasis2)

    def einsum_index(self, script):
        """ Returns indices to numpy.einsum summation. """
        assert len(script) == 3 and isinstance(script, str)
        return script

    @property
    def arrays(self):
        """ Returns list containing self._array """
        return [self._array]

    @property
    def label(self):
        """ Returns label of instance"""
        return self._label

    @label.setter
    def label(self, label):
        """ Sets label of instance"""
        self._label = label

    def _check_new_init_args(self, other):
        """Check whether an already initialized object is compatible"""
        other.__check_init_args__(self.nbasis, self.nbasis1, self.nbasis2)

    new.__check_init_args__ = _check_new_init_args

    def replace_array(self, value):
        """Replaces an array with another array, if not present, it will be
           generated.
        """
        if isinstance(value, DenseThreeIndex):
            self._array = value._array
        elif isinstance(value, np.ndarray):
            self._array = value
        else:
            raise ArgumentError(
                f"Do not know how to assign object of type {type(value)}."
            )

    def clear(self):
        """Reset all elements to zero."""
        self._array[:] = 0.0

    def copy(
        self, begin0=0, end0=None, begin1=0, end1=None, begin2=0, end2=None
    ):
        """Return a copy of (a part of) the object

           **Optional arguments:**

           begin0, end0, begin1, end1, begin2, end2
                Can be used to select a subblock of the object. When not given,
                the full range is used.
        """
        end0, end1, end2 = self._fix_ends(end0, end1, end2)
        result = DenseThreeIndex(end0 - begin0, end1 - begin1, end2 - begin2)
        result._array[:] = self._array[begin0:end0, begin1:end1, begin2:end2]
        result._label = self._label
        return result

    def assign(
        self,
        other,
        begin3=0,
        end3=None,
        begin4=0,
        end4=None,
        begin5=0,
        end5=None,
    ):
        """Assign with the contents of another object

           **Arguments:**

           other
                Another DenseThreeIndex object or an array.
        """
        check_type("other", other, DenseThreeIndex, np.ndarray, DenseOneIndex)
        if isinstance(other, DenseThreeIndex):
            end3, end4, end5 = other._fix_ends(end3, end4, end5)
            self._array[:] = other._array[
                begin3:end3, begin4:end4, begin5:end5
            ]
        elif isinstance(other, np.ndarray):
            self._array[:] = other
        elif isinstance(other, DenseOneIndex):
            shape = other.shape
            ends = [end3]
            end3 = ends[0] or shape[0]
            self._array[:] = other._array[begin3:end3].reshape(
                self.nbasis, self.nbasis1, self.nbasis2
            )

    def randomize(self):
        """Fill with random normal data"""
        self._array[:] = np.random.normal(0, 1, self.shape)

    def permute_basis(self, permutation):
        """Reorder the coefficients for a given permutation of basis functions.

           **Arguments:**

           permutation
                An integer numpy array that defines the new order of the basis
                functions.
        """
        self._array[:] = self._array.take(permutation, axis=0)
        self._array[:] = self._array.take(permutation, axis=1)
        self._array[:] = self._array.take(permutation, axis=2)

    def change_basis_signs(self, signs):
        """Correct for different sign conventions of the basis functions.

           **Arguments:**

           signs
                A numpy array with sign changes indicated by +1 and -1.
        """
        self._array *= signs
        self._array *= signs.reshape(-1, 1)
        self._array *= signs.reshape(-1, 1, 1)

    def iadd(
        self,
        other,
        factor=1.0,
        begin0=0,
        end0=None,
        begin1=0,
        end1=None,
        begin2=0,
        end2=None,
        begin3=0,
        end3=None,
        begin4=0,
        end4=None,
        begin5=0,
        end5=None,
    ):
        """Add another DenseThreeIndex object in-place, multiplied by factor

           **Arguments:**

           other
                A DenseThreeIndex instance to be added

           **Optional arguments:**

           factor
                The added term is scaled by this factor.

           begin0, end0, begin1, end1, begin2, end2
                Can be used to add only a part of the three-index object
        """
        check_type("other", other, DenseThreeIndex)
        check_type("factor", factor, float, int)
        end0, end1, end2 = self._fix_ends(end0, end1, end2)
        end3, end4, end5 = other._fix_ends(end3, end4, end5)
        self._array[begin0:end0, begin1:end1, begin2:end2] += (
            other._array[begin3:end3, begin4:end4, begin5:end5] * factor
        )

    def iscale(self, factor):
        """In-place multiplication with a scalar

           **Arguments:**

           factor
                A scalar factor.
        """
        check_type("factor", factor, float, int)
        self._array *= factor

    def itranspose(self, *shape):
        """In-place transpose according to shape
        """
        self._array = self._array.transpose(shape)

    def get_element(self, i, j, k):
        """Return a matrix element"""
        return self._array[i, j, k]

    def set_element(self, i, j, k, value):
        """Set a matrix element"""
        self._array[i, j, k] = value

    #
    # Properties
    #

    @property
    def nbasis(self):
        """The number of basis functions"""
        return self._array.shape[0]

    @property
    def nbasis1(self):
        """The number of basis functions"""
        return self._array.shape[1]

    @property
    def nbasis2(self):
        """The number of basis functions"""
        return self._array.shape[2]

    @property
    def shape(self):
        """The shape of the object"""
        return self._array.shape

    def slice(self, subscripts, out=None, factor=1.0, clear=True, **kwargs):
        """Returns a slice of the three-index object.

           **Arguments:**

           subscripts
                Any of ``abc->ab``, ``abc->abc``

           **Optional arguments:**

           out, factor, clear
                See :py:meth:`DenseLinalgFactory.tco`

           begin0, end0, begin1, end1, begin2, end2, begin3, end3
                Can be used to select a subblock of the object. When not given,
                the full range is used.
        """
        ends = [kwargs.get(f"end{i}", None) for i in range(self.ndim)]
        ends = self._fix_ends(*ends)

        ranges = []
        for i in range(self.ndim):
            ranges.append(kwargs.get(f"begin{i}", 0))
            ranges.append(ends[i])

        return DenseLinalgFactory.tco(
            subscripts,
            out,
            factor,
            clear,
            "einsum",
            (self, tuple(ranges)),
        )

    def iadd_expand_two_to_three(
        self,
        axis,
        two,
        factor=1.0,
        begin0=0,
        end0=None,
        begin1=0,
        end1=None,
        abegin0=0,
        aend0=None,
        abegin1=0,
        aend1=None,
        abegin2=0,
        aend2=None,
    ):
        """Expand two-index object along two axes and add it to three-index
           object.

           **Arguments:**

           axis
                Any of ``1-3`` (abac += bc), ``0-2`` (abcb += ac),
                ``0-3`` (abbc += ac), ``1-2`` (abca += bc),
                ``diag`` (abab += ab). The expansion is performed for the
                repeated indices in the four-index object

           two
                An instance of DenseTwoIndex.

           **Optional arguments:**

           factor
                See :py:meth:`DenseLinalgFactory.tco`

           begin0, end0, begin1, end1
                Can be used to select a subblock of the two-index object (two).
                When not given, the full range is used.
        """
        check_options(
            "axis",
            axis,
            "ac->abc",
            "bc->abc",
            "ab->aab",
            "aa->abc",
            "cc->abc",
            "bb->abc",
        )
        end0, end1 = two._fix_ends(end0, end1)
        aend0, aend1, aend2 = self._fix_ends(aend0, aend1, aend2)
        if axis == "ac->abc":
            for i in range(abegin1, aend1):
                self._array[:, i, :] += (
                    two._array[begin0:end0, begin1:end1] * factor
                )
        elif axis == "bc->abc":
            for i in range(abegin0, aend0):
                self._array[i, :, :] += (
                    two._array[begin0:end0, begin1:end1] * factor
                )
        elif axis == "ab->aab":
            assert self.nbasis == self.nbasis1
            for i in range(abegin0, aend0):
                self._array[i, i, :] += two._array[i, begin1:end1] * factor
        elif axis == "aa->abc":
            assert two.nbasis == two.nbasis1
            for i in range(abegin1, aend1):
                for j in range(abegin2, aend2):
                    self._array[:, i, j] += (
                        two._array[begin0:end0, begin1:end1].diagonal()
                        * factor
                    )
        elif axis == "cc->abc":
            assert two.nbasis == two.nbasis1
            for i in range(abegin0, aend0):
                for j in range(abegin1, aend1):
                    self._array[i, j, :] += (
                        two._array[begin0:end0, begin1:end1].diagonal()
                        * factor
                    )
        elif axis == "bb->abc":
            assert two.nbasis == two.nbasis1
            for i in range(abegin0, aend0):
                for j in range(abegin2, aend2):
                    self._array[i, :, j] += (
                        two._array[begin0:end0, begin1:end1].diagonal()
                        * factor
                    )


class DenseFourIndex(FourIndex):
    """Dense symmetric four-dimensional matrix.

       This is the most inefficient implementation in terms of memory usage and
       computer time. Due to its simplicity, it is trivial to implement. This
       implementation mainly serves as a reference for testing purposes.
    """

    #
    # Constructor and destructor
    #

    def __init__(
        self, nbasis, nbasis1=None, nbasis2=None, nbasis3=None, label=""
    ):
        """
           **Arguments:**

           nbasis
                The number of basis functions.

           **Optional arguments:**

           nbasis1, nbasis2, nbasis3
        """
        if nbasis1 is None:
            nbasis1 = nbasis
        if nbasis2 is None:
            nbasis2 = nbasis
        if nbasis3 is None:
            nbasis3 = nbasis
        self._array = np.zeros((nbasis, nbasis1, nbasis2, nbasis3), float)
        self._label = label
        log.mem.announce(self._array.nbytes)

    def __del__(self):
        if log is not None:
            if hasattr(self, "_array"):
                log.mem.denounce(self._array.nbytes)
        if hasattr(self, "_array"):
            del self._array

    #
    # Methods from base class
    #

    def __check_init_args__(
        self, nbasis, nbasis1=None, nbasis2=None, nbasis3=None
    ):
        """Is self compatible with the given constructor arguments?"""
        if nbasis1 is None:
            nbasis1 = nbasis
        if nbasis2 is None:
            nbasis2 = nbasis
        if nbasis3 is None:
            nbasis3 = nbasis
        assert nbasis == self.nbasis
        assert nbasis1 == self.nbasis1
        assert nbasis2 == self.nbasis2
        assert nbasis3 == self.nbasis3

    def __eq__(self, other):
        """Compare self with other"""
        return (
            isinstance(other, DenseFourIndex)
            and other.nbasis == self.nbasis
            and other.nbasis1 == self.nbasis1
            and other.nbasis2 == self.nbasis2
            and other.nbasis3 == self.nbasis3
            and (other._array == self._array).all()
        )

    @classmethod
    def from_hdf5(cls, grp):
        """Construct an instance from data previously stored in an h5py.Group.

           **Arguments:**

           grp
                An h5py.Group object.
        """
        nbasis = grp["array"].shape[0]
        nbasis1 = grp["array"].shape[1]
        nbasis2 = grp["array"].shape[2]
        nbasis3 = grp["array"].shape[3]
        label = grp.attrs["label"]
        result = cls(nbasis, nbasis1, nbasis2, nbasis3, label)
        grp["array"].read_direct(result._array)
        return result

    def to_hdf5(self, grp):
        """Dump this object in an h5py.Group

           **Arguments:**

           grp
                An h5py.Group object.
        """
        grp.attrs["class"] = self.__class__.__name__
        grp["array"] = self._array
        grp.attrs["label"] = self._label

    # FIXME: rename into clean_copy
    def new(self):
        """Return a new four-index object with the same nbasis"""
        return DenseFourIndex(
            self.nbasis, self.nbasis1, self.nbasis2, self.nbasis3
        )

    def einsum_index(self, script):
        """ Returns indices to numpy.einsum summation. """
        assert len(script) == 4 and isinstance(script, str)
        return script

    @property
    def arrays(self):
        """ Returns list containing self._array """
        return [self._array]

    @property
    def label(self):
        """ Returns label of instance"""
        return self._label

    @label.setter
    def label(self, label):
        """ Sets label of instance"""
        self._label = label

    def _check_new_init_args(self, other):
        """Check whether an already initialized object is compatible"""
        other.__check_init_args__(
            self.nbasis, self.nbasis1, self.nbasis2, self.nbasis3
        )

    new.__check_init_args__ = _check_new_init_args

    def clear(self):
        """Reset all elements to zero."""
        self._array[:] = 0.0

    def replace_array(self, value):
        """Replaces an array with another array, if not present, it will be
           generated.
        """
        if isinstance(value, DenseFourIndex):
            self._array = value._array
        elif isinstance(value, np.ndarray):
            self._array = value
        else:
            raise ArgumentError(
                f"Do not know how to assign object of type {type(value)}."
            )

    def copy(
        self,
        begin0=0,
        end0=None,
        begin1=0,
        end1=None,
        begin2=0,
        end2=None,
        begin3=0,
        end3=None,
    ):
        """Return a copy of (a part of) the object

           **Optional arguments:**

           begin0, end0, begin1, end1, begin2, end2, begin3, end3
                Can be used to select a subblock of the object. When not given,
                the full range is used.
        """
        end0, end1, end2, end3 = self._fix_ends(end0, end1, end2, end3)
        result = DenseFourIndex(
            end0 - begin0, end1 - begin1, end2 - begin2, end3 - begin3
        )
        result._array[:] = self._array[
            begin0:end0, begin1:end1, begin2:end2, begin3:end3
        ]
        result._label = self._label
        return result

    def reshape(self, shape):
        """Reshape array

           **Optional arguments:**

           shape
                List containing the new dimension of each axis.
        """
        if len(shape) == 0:
            raise ArgumentError("No array shape given.")
        elif len(shape) == 1:
            out = DenseOneIndex(shape[0])
        elif len(shape) == 2:
            out = DenseTwoIndex(shape[0], shape[1])
        elif len(shape) == 3:
            out = DenseThreeIndex(shape[0], shape[1], shape[2])
        elif len(shape) == 4:
            out = DenseFourIndex(shape[0], shape[1], shape[2], shape[3])
        out._array[:] = np.reshape(self._array, shape, order="C")
        return out

    def assign(self, other, ind=None, begin4=0, end4=None):
        """Assign with the contents of another object

           **Arguments:**

           other
                Another DenseFourIndex object or a np ndarrray or float.
        """
        check_type(
            "other",
            other,
            DenseFourIndex,
            np.ndarray,
            float,
            DenseTwoIndex,
            DenseOneIndex,
        )
        if isinstance(other, DenseFourIndex):
            if ind is not None:
                self._array[tuple(ind)] = other._array
            else:
                self._array[:] = other._array
        elif isinstance(other, float):
            if ind is not None:
                self._array[tuple(ind)] = other
            else:
                self._array[:] = other
        elif isinstance(other, np.ndarray):
            if other.shape == self.shape:
                self._array[:] = other
            elif ind is not None:
                self._array[ind] = other
            else:
                self._array[:] = other.reshape(
                    (self.nbasis, self.nbasis1, self.nbasis2, self.nbasis3)
                )
        elif isinstance(other, DenseTwoIndex):
            if ind is not None:
                self._array[tuple(ind)] = other._array
            else:
                self._array[:] = other._array.reshape(
                    (self.nbasis, self.nbasis1, self.nbasis2, self.nbasis3)
                )
        elif isinstance(other, DenseOneIndex):
            if ind is not None:
                self._array[ind] = other._array
            else:
                end4 = other._fix_ends(end4)[0]
                self._array[:] = other._array[begin4:end4].reshape(
                    (self.nbasis, self.nbasis1, self.nbasis2, self.nbasis3)
                )
        else:
            raise ArgumentError(
                f"Do not know how to assign object of type {type(other)}."
            )

    def assign_triu(self, other, begin4=0, end4=None, shape=None):
        """Assign upper triangular with the contents of another object
           Four-index object will be reshaped as (self.nbasis*self.nbasis1,self.nbasis2*self.nbasis3)
           during that operation.

           **Arguments:**

           other
                Another DenseFourIndex object or a np ndarrray or float.

           begin0, end0, begin1, end1, begin2, end2, begin3, end3, ...
                Can be used to select a subblock of the object. When not given,
                the full range is used.
        """
        check_type("other", other, DenseOneIndex, np.ndarray)
        if isinstance(other, DenseOneIndex):
            end4 = other._fix_ends(end4)[0]
            if shape is None:
                indtriu = np.triu_indices(self.nbasis * self.nbasis1)
                self._array.reshape(
                    self.nbasis * self.nbasis1, self.nbasis2 * self.nbasis3
                )[indtriu] = other._array[begin4:end4]
            else:
                raise NotImplementedError
        elif isinstance(other, np.ndarray):
            if other.ndim != 1:
                raise ArgumentError(
                    f"Do not know how to assign object of type {type(other)}."
                )
            if shape is None:
                indtriu = np.triu_indices(self.nbasis * self.nbasis1)
                self._array.reshape(
                    self.nbasis * self.nbasis1, self.nbasis2 * self.nbasis3
                )[indtriu] = other[begin4:end4]
            else:
                raise NotImplementedError
        else:
            raise ArgumentError(
                f"Do not know how to assign object of type {type(other)}."
            )

    def randomize(self):
        """Fill with random normal data"""
        self._array[:] = np.random.normal(0, 1, self.shape)

    def permute_basis(self, permutation):
        """Reorder the coefficients for a given permutation of basis functions.

           **Arguments:**

           permutation
                An integer numpy array that defines the new order of the basis
                functions.
        """
        self._array[:] = self._array.take(permutation, axis=0)
        self._array[:] = self._array.take(permutation, axis=1)
        self._array[:] = self._array.take(permutation, axis=2)
        self._array[:] = self._array.take(permutation, axis=3)

    def change_basis_signs(self, signs):
        """Correct for different sign conventions of the basis functions.

           **Arguments:**

           signs
                A numpy array with sign changes indicated by +1 and -1.
        """
        self._array *= signs
        self._array *= signs.reshape(-1, 1)
        self._array *= signs.reshape(-1, 1, 1)
        self._array *= signs.reshape(-1, 1, 1, 1)

    def get_triu(self):
        """Convert a four-index object into two-index with shape (nbasis*nbasis1,
           nbasis2*nbasis3) and return corresponding upper triangular matrix.

           **Arguments:**
        """
        if self.nbasis * self.nbasis1 != self.nbasis2 * self.nbasis3:
            raise MatrixShapeError(
                "Do not know how to reshape four-index object."
            )
        indtriu = np.triu_indices(self.nbasis * self.nbasis1)
        return self._array.reshape(
            self.nbasis * self.nbasis1, self.nbasis2 * self.nbasis3
        )[indtriu]

    def iadd(
        self,
        other,
        factor=1.0,
        begin0=0,
        end0=None,
        begin1=0,
        end1=None,
        begin2=0,
        end2=None,
        begin3=0,
        end3=None,
        begin4=0,
        end4=None,
        begin5=0,
        end5=None,
        begin6=0,
        end6=None,
        begin7=0,
        end7=None,
        ind=None,
    ):
        """Add another DenseFourIndex object in-place, multiplied by factor

           **Arguments:**

           other
                A DenseFourIndex instance to be added

           **Optional arguments:**

           factor
                The added term is scaled by this factor.

           begin0, end0, begin1, end1, begin2, end2, begin3, end3
                Can be used to select a subblock of the object. When not given,
                the full range is used.

           ind
                If ind is provided, other will be added to view of array as
                indicated by ind. This only works for 2 index objects.
        """
        check_type("other", other, DenseFourIndex, DenseTwoIndex)
        check_type("factor", factor, float, int)
        end0, end1, end2, end3 = self._fix_ends(end0, end1, end2, end3)
        if ind is None:
            # only works with 4-index
            check_type("other", other, DenseFourIndex)
            end4, end5, end6, end7 = other._fix_ends(end4, end5, end6, end7)
            self._array[
                begin0:end0, begin1:end1, begin2:end2, begin3:end3
            ] += (
                other._array[
                    begin4:end4, begin5:end5, begin6:end6, begin7:end7
                ]
                * factor
            )
        else:
            # only works with 2-index
            check_type("other", other, DenseTwoIndex)
            end4, end5 = other._fix_ends(end4, end5)
            self._array[tuple(ind)] += (
                other._array[begin4:end4, begin5:end5] * factor
            )

    def imul(self, other, factor=1.0):
        """In-place element-wise multiplication: ``self *= other * factor``

           **Arguments:**

           other
                A DenseFourIndex instance.

           **Optional arguments:**

           factor
                The four-index object is scaled by this factor.
        """
        check_type("other", other, DenseFourIndex)
        check_type("factor", factor, float, int)

        def from_mask(mask):
            return {0: 1, 1: 2, 2: 4, 3: 8}[mask]

        def to_mask(sym):
            return {1: 0, 2: 1, 4: 2, 8: 3}[sym]

        self._array *= other._array
        self.iscale(factor)

    def iscale(self, factor):
        """In-place multiplication with a scalar

           **Arguments:**

           factor
                A scalar factor.
        """
        check_type("factor", factor, float, int)
        self._array *= factor

    def get_element(self, i, j, k, l):
        """Return a matrix element"""
        return self._array[i, j, k, l]

    def set_element(self, i, j, k, l, value, symmetry=8):
        """Set a matrix element

           **Arguments:**

           i, j, k, l
                The matrix indexes to be set

           value
                The value to be assigned to the matrix element.

           **Optional arguments:**

           symmetry
                The level of symmetry to be enforced when setting the matrix
                element. See :ref:`dense_matrix_symmetry` for more details.
        """
        check_options("symmetry", symmetry, 1, 2, 4, 8)
        if not self.is_shape_symmetric(symmetry):
            raise SymmetryError(
                "FourIndex object does not have the right shape to impose the selected symmetry."
            )
        self._array[i, j, k, l] = value
        if symmetry in (2, 8):
            self._array[j, i, l, k] = value
        if symmetry in (4, 8):
            self._array[k, j, i, l] = value
            self._array[i, l, k, j] = value
            self._array[k, l, i, j] = value
        if symmetry == 8:
            self._array[l, k, j, i] = value
            self._array[j, k, l, i] = value
            self._array[l, i, j, k] = value

    #
    # Properties
    #

    @property
    def nbasis(self):
        """The number of basis functions"""
        return self._array.shape[0]

    @property
    def nbasis1(self):
        """The number of basis functions"""
        return self._array.shape[1]

    @property
    def nbasis2(self):
        """The number of basis functions"""
        return self._array.shape[2]

    @property
    def nbasis3(self):
        """The number of basis functions"""
        return self._array.shape[3]

    @property
    def shape(self):
        """The shape of the object"""
        return self._array.shape

    #
    # New methods for this implementation
    # TODO: consider adding these to base class
    #

    def is_symmetric(self, symmetry=8, rtol=1e-5, atol=1e-8):
        """Check the symmetry of the array.

           **Optional arguments:**

           symmetry
                The symmetry to check. See :ref:`dense_matrix_symmetry`
                for more details. In addition to 1, 2, 4, 8, also 'cdab' is
                supported.

           rtol and atol
                relative and absolute tolerance. See to ``np.allclose``.
        """
        if not self.is_shape_symmetric(symmetry):
            return False
        result = True
        if symmetry in (2, 8):
            result &= np.allclose(
                self._array, self._array.transpose(1, 0, 3, 2), rtol, atol
            )
        if symmetry in (4, 8):
            result &= np.allclose(
                self._array, self._array.transpose(2, 3, 0, 1), rtol, atol
            )
            result &= np.allclose(
                self._array, self._array.transpose(2, 1, 0, 3), rtol, atol
            )
            result &= np.allclose(
                self._array, self._array.transpose(0, 3, 2, 1), rtol, atol
            )
        if symmetry == 8:
            result &= np.allclose(
                self._array, self._array.transpose(3, 2, 1, 0), rtol, atol
            )
            result &= np.allclose(
                self._array, self._array.transpose(3, 0, 1, 2), rtol, atol
            )
            result &= np.allclose(
                self._array, self._array.transpose(1, 2, 3, 0), rtol, atol
            )
        if symmetry == "cdab":
            result &= np.allclose(
                self._array, self._array.transpose(2, 3, 0, 1), rtol, atol
            )
        return result

    def is_shape_symmetric(self, symmetry):
        """Check whether the symmetry argument matches the shape"""
        result = True
        if symmetry in (2, 8):
            result &= self.nbasis == self.nbasis1
            result &= self.nbasis2 == self.nbasis3
        if symmetry in (4, 8):
            result &= self.nbasis == self.nbasis2
            result &= self.nbasis1 == self.nbasis3
            result &= self.nbasis == self.nbasis3
            result &= self.nbasis1 == self.nbasis2
        return result

    def symmetrize(self, symmetry=8):
        """Symmetrize in-place

           **Optional arguments:**

           symmetry
                The symmetry to impose. See :ref:`dense_matrix_symmetry` for
                more details.
        """
        check_options("symmetry", symmetry, 1, 2, 4, 8)
        # The implementation is relatively expensive (in terms of memory) but
        # results in exactly symmetrized four-index objects.
        if not self.is_shape_symmetric(symmetry):
            raise SymmetryError(
                "FourIndex object does not have the right shape for symmetrization."
            )
        if symmetry in (2, 8):
            self._array[:] = self._array + self._array.transpose(1, 0, 3, 2)
            self.iscale(0.5)
        if symmetry in (4, 8):
            self._array[:] = self._array + self._array.transpose(2, 3, 0, 1)
            self._array[:] = self._array + self._array.transpose(0, 3, 2, 1)
            self.iscale(0.25)

    def itranspose(self, transpose=None):
        """In-place transpose: ``0,1,2,3 -> 1,0,3,2``"""
        if transpose is None:
            self._array = self._array.transpose(1, 0, 3, 2)
        else:
            self._array = self._array.transpose(transpose)

    def sum(self):
        """Return the sum of all elements"""
        return np.sum(self._array)

    def iadd_exchange(self):
        """In-place addition of its own exchange contribution"""
        # Broken code, don't use
        # self._array -= np.einsum('abcd->abdc', self._array)
        # We cannot do inplace einsum. Instead use (and don't change):
        self._array = self._array - np.einsum("abcd->abdc", self._array)

    def iadd_transpose(self, transpose, factor=1.0):
        """In-place addition of its own transpose"""
        check_type("factor", factor, float, int)
        self._array[:] = (
            self._array + self._array.transpose(transpose) * factor
        )

    def slice(self, subscripts, out=None, factor=1.0, clear=True, **kwargs):
        """Returns a slice of the four-index object.

           **Arguments:**

           subscripts
                Any of ``abcd->ab``, ``abcd->abc``, ``abcd->abcd`` 

           **Optional arguments:**

           out, factor, clear
                See :py:meth:`DenseLinalgFactory.tco`

           begin0, end0, begin1, end1, begin2, end2, begin3, end3
                Can be used to select a subblock of the object. When not given,
                the full range is used.
        """
        ends = [kwargs.get(f"end{i}", None) for i in range(self.ndim)]
        ends = self._fix_ends(*ends)

        ranges = []
        for i in range(self.ndim):
            ranges.append(kwargs.get(f"begin{i}", 0))
            ranges.append(ends[i])

        return DenseLinalgFactory.tco(
            subscripts,
            out,
            factor,
            clear,
            "einsum",
            (self, tuple(ranges)),
        )


    def iadd_expand_one_to_four(self, axis, one, factor=1.0):
        """Expand two-index object along two axes and add it to four-index
           object.

           **Arguments:**

           axis

           one
                An instance of DenseOneIndex.

           **Optional arguments:**

           factor
                See :py:meth:`DenseLinalgFactory.tco`

           begin0, end0, begin1, end1
                Can be used to select a subblock of the two-index object (two).
                When not given, the full range is used.
        """
        check_options("axis", axis, "0", "1", "2", "3", "b->abcb", "a->abac")
        if axis == "0":
            for i in range(self.nbasis):
                self._array[i, :, :, :] += one._array[i] * factor
        if axis == "1":
            for i in range(self.nbasis1):
                self._array[:, i, :, :] += one._array[i] * factor
        if axis == "2":
            for i in range(self.nbasis2):
                self._array[:, :, i, :] += one._array[i] * factor
        if axis == "3":
            for i in range(self.nbasis3):
                self._array[:, :, :, i] += one._array[i] * factor
        elif axis == "a->abac":
            for i in range(self.nbasis):
                self._array[i, :, i, :] += one._array[i] * factor
        elif axis == "b->abcb":
            for i in range(self.nbasis1):
                self._array[:, i, :, i] += one._array[i] * factor

    def iadd_expand_two_to_four(
        self,
        axis,
        two,
        factor=1.0,
        two1=None,
        begin0=0,
        end0=None,
        begin1=0,
        end1=None,
    ):
        """Expand two-index object along two axes and add it to four-index
           object.

           **Arguments:**

           axis
                Any of ``1-3`` (abac += bc), ``0-2`` (abcb += ac),
                ``0-3`` (abbc += ac), ``1-2`` (abca += bc),
                ``diag`` (abab += ab). The expansion is performed for the
                repeated indices in the four-index object

           two
                An instance of DenseTwoIndex.

           **Optional arguments:**

           factor
                See :py:meth:`DenseLinalgFactory.tco`

           begin0, end0, begin1, end1
                Can be used to select a subblock of the two-index object (two).
                When not given, the full range is used.
        """
        check_options(
            "axis",
            axis,
            "1-3",
            "0-2",
            "0-3",
            "1-2",
            "diag",
            "ab->abcd",
            "cd->abcd",
            "cb->abcd",
            "ad->abcd",
            "ac->abcd",
            "ca->abcd",
            "bd->abcd",
            "db->abcd",
            "ab,cd->adcb",
            "ab,cd->abcd",
        )
        end0, end1 = two._fix_ends(end0, end1)
        if axis == "0-2":
            for i in range(self.nbasis1):
                self._array[:, i, :, i] += (
                    two._array[begin0:end0, begin1:end1] * factor
                )
        elif axis == "0-3":
            for i in range(self.nbasis1):
                self._array[:, i, i, :] += (
                    two._array[begin0:end0, begin1:end1] * factor
                )
        elif axis == "1-2":
            for i in range(self.nbasis):
                self._array[i, :, :, i] += (
                    two._array[begin0:end0, begin1:end1] * factor
                )
        elif axis == "1-3":
            for i in range(self.nbasis):
                self._array[i, :, i, :] += (
                    two._array[begin0:end0, begin1:end1] * factor
                )
        elif axis == "ab->abcd":
            for i in range(self.nbasis2):
                for j in range(self.nbasis3):
                    self._array[:, :, i, j] += (
                        two._array[begin0:end0, begin1:end1] * factor
                    )
        elif axis == "cd->abcd":
            for i in range(self.nbasis):
                for j in range(self.nbasis1):
                    self._array[i, j, :, :] += (
                        two._array[begin0:end0, begin1:end1] * factor
                    )
        elif axis == "ad->abcd":
            for i in range(self.nbasis1):
                for j in range(self.nbasis2):
                    self._array[:, i, j, :] += (
                        two._array[begin0:end0, begin1:end1] * factor
                    )
        elif axis == "ac->abcd":
            for i in range(self.nbasis1):
                for j in range(self.nbasis3):
                    self._array[:, i, :, j] += (
                        two._array[begin0:end0, begin1:end1] * factor
                    )
        elif axis == "ca->abcd":
            for i in range(self.nbasis1):
                for j in range(self.nbasis3):
                    self._array[:, i, :, j] += (
                        two._array[begin0:end0, begin1:end1].T * factor
                    )
        elif axis == "bd->abcd":
            for i in range(self.nbasis):
                for j in range(self.nbasis2):
                    self._array[i, :, j, :] += (
                        two._array[begin0:end0, begin1:end1] * factor
                    )
        elif axis == "db->abcd":
            for i in range(self.nbasis):
                for j in range(self.nbasis2):
                    self._array[i, :, j, :] += (
                        two._array[begin0:end0, begin1:end1].T * factor
                    )
        elif axis == "cb->abcd":
            for i in range(self.nbasis):
                for j in range(self.nbasis3):
                    self._array[i, :, :, j] += (
                        two._array[begin0:end0, begin1:end1].T * factor
                    )
        elif axis == "diag":
            for i in range(self.nbasis):
                for a in range(self.nbasis1):
                    self._array[i, a, i, a] += two._array[i, a] * factor
        else:
            return DenseLinalgFactory.tco(
                axis,
                self,
                factor,
                False,
                "einsum",
                two1,
                (two, (begin0, end0, begin1, end1)),
            )

    def iadd_expand_three_to_four(self, axis, three, factor=1.0):
        """Expand three-index object along one axis and add it to four-index
           object.

           **Arguments:**

           axis
                Any of ``1-3-1-2`` (abac += abc), ``1-2-1-2`` (abca += abc),
                ``0-2-0-2`` (abcb += abc), ``0-3-0-2`` (abbc += abc),
                ``0-2-0-1`` (abcb += acb). The expansion is performed for
                repeated indices in the four-index object

           three
                An instance of DenseThreeIndex.

           **Optional arguments:**

           factor
                See :py:meth:`DenseLinalgFactory.tco`
        """
        check_options(
            "axis",
            axis,
            "1-3-1-2",
            "0-2-0-1",
            "1-2-1-2",
            "0-2-0-2",
            "0-3-0-2",
            "0",
            "1",
            "2",
            "3",
            "abc->aabc",
        )
        if axis == "1-3-1-2":
            assert self.nbasis == self.nbasis2
            for i in range(self.nbasis):
                self._array[i, :, i, :] += three._array[i, :, :] * factor
        elif axis == "1-2-1-2":
            assert self.nbasis == self.nbasis3
            for i in range(self.nbasis):
                self._array[i, :, :, i] += three._array[i, :, :] * factor
        elif axis == "0-2-0-2":
            assert self.nbasis1 == self.nbasis3
            for i in range(self.nbasis1):
                self._array[:, i, :, i] += three._array[:, i, :] * factor
        elif axis == "0-3-0-2":
            assert self.nbasis1 == self.nbasis2
            for i in range(self.nbasis1):
                self._array[:, i, i, :] += three._array[:, i, :] * factor
        elif axis == "0-2-0-1":
            assert self.nbasis1 == self.nbasis3
            for i in range(self.nbasis1):
                self._array[:, i, :, i] += three._array[:, :, i] * factor
        elif axis == "abc->aabc":
            assert self.nbasis == self.nbasis1
            for i in range(self.nbasis):
                self._array[i, i, :, :] += three._array[i, :, :] * factor
        elif axis == "0":
            assert self.shape[1:] == three.shape
            for i in range(self.nbasis):
                self._array[i, :, :, :] += three._array[:, :, :] * factor
        elif axis == "1":
            assert tuple(np.take(self.shape, [0, 2, 3])) == three.shape
            self._array[:] -= np.expand_dims(three._array, int(axis)) * factor
        #           for i in range(self.nbasis1):
        #               self._array[:,i,:,:] += three._array[:,:,:]*factor
        elif axis == "2":
            assert tuple(np.take(self.shape, [0, 1, 3])) == three.shape
            for i in range(self.nbasis2):
                self._array[:, :, i, :] += three._array[:, :, :] * factor
        elif axis == "3":
            assert self.shape[:3] == three.shape
            for i in range(self.nbasis3):
                self._array[:, :, :, i] += three._array[:, :, :] * factor

    def iadd_expand_four_to_three(
        self,
        axis,
        three,
        factor=1.0,
        begin0=0,
        end0=None,
        begin1=0,
        end1=None,
        begin2=0,
        end2=None,
        begin3=0,
        end3=None,
    ):
        """Expand diagonal parts of four-index object along one axes and add it
           to three-index object.

           **Arguments:**

           axis
                Any of ``1-3`` (abac += bc), ``0-2`` (abcb += ac),
                ``0-3`` (abbc += ac), ``1-2`` (abca += bc),
                ``diag`` (abab += ab). The expansion is performed for the
                repeated indices in the four-index object

           two
                An instance of DenseTwoIndex.

           **Optional arguments:**

           factor
                See :py:meth:`DenseLinalgFactory.tco`

           begin0, end0, begin1, end1
                Can be used to select a subblock of the two-index object (two).
                When not given, the full range is used.
        """
        check_options(axis, axis, "bcbc->abc", "acac->abc", "abab->abc")
        end0, end1, end2, end3 = self._fix_ends(end0, end1, end2, end3)
        if axis == "bcbc->abc":
            for i in range(three.shape[0]):
                three._array[i, :, :] += (
                    np.einsum(
                        "abab->ab",
                        self._array[
                            begin0:end0, begin1:end1, begin2:end2, begin3:end3
                        ],
                    )
                    * factor
                )
        elif axis == "acac->abc":
            for i in range(three.shape[1]):
                three._array[:, i, :] += (
                    np.einsum(
                        "abab->ab",
                        self._array[
                            begin0:end0, begin1:end1, begin2:end2, begin3:end3
                        ],
                    )
                    * factor
                )
        elif axis == "abab->abc":
            for i in range(three.shape[2]):
                three._array[:, :, i] += (
                    np.einsum(
                        "abab->ab",
                        self._array[
                            begin0:end0, begin1:end1, begin2:end2, begin3:end3
                        ],
                    )
                    * factor
                )

    def assign_four_index_transform(
        self,
        ao_integrals,
        exp0,
        exp1=None,
        exp2=None,
        exp3=None,
        method="tensordot",
    ):
        """Perform four index transformation.

           **Arguments:**

           oa_integrals
                A DenseFourIndex with integrals in atomic orbitals.

           exp0
                A DenseOrbital object with molecular orbitals

           **Optional arguments:**

           exp1, exp2, exp3
                Can be provided to transform each index differently.

           method
                Either ``einsum`` or ``tensordot`` (default) or ``cpp``.
        """
        # parse arguments
        check_type("ao_integrals", ao_integrals, DenseFourIndex)
        exp0, exp1, exp2, exp3 = parse_four_index_transform_exps(
            exp0, exp1, exp2, exp3, DenseOrbital
        )
        # actual transform
        if method == "einsum":
            # The order of the dot products is according to literature
            # conventions.
            self._array[:] = np.einsum(
                "sd,pqrs->pqrd",
                exp3.coeffs,
                ao_integrals._array,
                casting="no",
                order="C",
            )
            self._array[:] = np.einsum(
                "rc,pqrd->pqcd",
                exp2.coeffs,
                self._array,
                casting="no",
                order="C",
            )
            self._array[:] = np.einsum(
                "qb,pqcd->pbcd",
                exp1.coeffs,
                self._array,
                casting="no",
                order="C",
            )
            self._array[:] = np.einsum(
                "pa,pbcd->abcd",
                exp0.coeffs,
                self._array,
                casting="no",
                order="C",
            )
        elif method == "tensordot":
            # because the way tensordot works, the order of the dot products is
            # not according to literature conventions.
            self._array[:] = np.tensordot(
                ao_integrals._array, exp0.coeffs, axes=([0], [0])
            )
            self._array[:] = np.tensordot(
                self._array, exp1.coeffs, axes=([0], [0])
            )
            self._array[:] = np.tensordot(
                self._array, exp2.coeffs, axes=([0], [0])
            )
            self._array[:] = np.tensordot(
                self._array, exp3.coeffs, axes=([0], [0])
            )
        elif method == "cpp":
            # only works for restricted orbitals so far
            apply_four_index_transform_cpp(
                ao_integrals._array, exp0.coeffs, self._array
            )
        else:
            raise UnknownOption(
                "The method must either be 'einsum' or 'tensordot'."
            )
