# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --

# Detailed changelog:
# This module has been originally written and updated by Paweł Tecmer (see CHANGELOG).
# Its current version contains updates from the PyBEST developer team.
#
# This implementation can also be found in `Horton 2.0.0`.
# However, this file has been updated and debugged. Compatibility with Horton is NOT
# guaranteed.
# 
# Detailed changes (see also CHANGELOG):
# 2020-07-01: use abs class
# 2020-07-01: rename [compute_] functions
# 2020-07-01: exploit labels of NIndex objects

"""Physical Model Hamiltonian"""


import abc

from pybest.linalg.dense import DenseLinalgFactory
from pybest.log import log

__all__ = ["Hubbard"]


class PhysModHam(abc.ABC):
    """Base class for the Physical Model Hamiltonians: 1-D Hubbard,
       etc.
    """

    def __init__(self, lf, pbc=True):
        """
           **Attributes:**

            pdb
                Periodic boundary conditions. Default, pdb=true

        """
        self._lf = lf
        self._denself = DenseLinalgFactory(lf.default_nbasis)
        self._pbc = pbc

        log.cite(
            "the implementation of model hamiltonians", "boguslawski2014a"
        )

    @property
    def pbc(self):
        """The periodic boundary conditions"""
        return self._pbc

    @property
    def lf(self):
        """The LinalgFactory"""
        return self._lf

    @property
    def denself(self):
        """The dense LinalgFactory"""
        return self._denself

    @abc.abstractmethod
    def compute_one_body(self, tparam):
        """Calculate the one-body term of the 1D Hubbard Hamiltonian"""
        raise NotImplementedError

    @abc.abstractmethod
    def compute_two_body(self, uparam):
        """Calculate the the-body term of the 1D Hubbard Hamiltonian"""
        raise NotImplementedError

    @abc.abstractmethod
    def compute_overlap(self):
        """Calculate overlap of the 1D Hubbard Hamiltonian, (identity matrix)"""
        raise NotImplementedError


class Hubbard(PhysModHam):
    """The 1-D Hubbard class Hamiltonian"""

    def compute_one_body(self, tparam):
        """Calculate the one-body term of the 1D Hubbard Hamiltonian"""
        result = self.lf.create_two_index(label="kin")
        for i in range(self.lf.default_nbasis - 1):
            result.set_element(i, i + 1, tparam, 2)
        if self.pbc == True:
            result.set_element(self.lf.default_nbasis - 1, 0, tparam, 2)
        return result

    def compute_two_body(self, uparam):
        """Calculate the two-body term of the 1D Hubbard Hamiltonian"""
        result = self.denself.create_four_index(label="eri")
        for i in range(self.denself.default_nbasis):
            result.set_element(i, i, i, i, uparam)
        return result

    def compute_overlap(self):
        """Calculate overlap of the 1D Hubbard Hamiltonian, (identity matrix)"""
        result = self.lf.create_two_index(label="olp")
        result.assign_diagonal(1.0)
        return result
